Package: rgl
Version: 1.3.36
Title: 3D Visualization Using OpenGL
Authors@R: c(person("Duncan", "Murdoch", role = c("aut", "cre"),
                     email = "murdoch.duncan@gmail.com"),
              person("Daniel", "Adler", role = "aut",
                     email = "dadler@dyncall.org"),
              person("Oleg", "Nenadic", role = "ctb"),
              person("Simon", "Urbanek", role = "ctb"),
              person("Ming", "Chen", role = "ctb"),
              person("Albrecht", "Gebhardt", role = "ctb"),
              person("Ben", "Bolker", role = "ctb"),
              person("Gabor", "Csardi", role = "ctb"),
              person("Adam", "Strzelecki", role = "ctb"),
              person("Alexander", "Senger", role = "ctb"),
              person("The R Core Team", role = c("ctb", "cph")),
              person("Dirk","Eddelbuettel", role = "ctb"),
              person("The authors of Shiny", role = "cph"),
              person("The authors of knitr", role = "cph"),
              person("Jeroen", "Ooms", role = "ctb"),
              person("Yohann", "Demont", role = "ctb"),
              person("Joshua", "Ulrich", role = "ctb"),
              person("Xavier", "Fernandez i Marin", role = "ctb"),
              person("George", "Helffrich", role = "ctb"),
              person("Ivan", "Krylov", role = "ctb"),
              person("Michael", "Sumner", role = "ctb"),
              person("Mike", "Stein", role = "ctb"),
              person("Jonathon", "Love", role = "ctb"),
              person("Mapbox team", role = c("ctb", "cph")))
Depends: R (>= 3.6.0)
Suggests: MASS, rmarkdown (>= 2.16), deldir (>= 1.0-4), orientlib,
        lattice, misc3d, magick, plotrix (>= 3.7-3), interp,
        alphashape3d, tcltk, js (>= 1.2), webshot2 (>= 0.1.0), downlit
        (>= 0.4.0), pkgdown (>= 2.0.0), extrafont, shiny,
        manipulateWidget (>= 0.9.0), testthat, crosstalk, V8, chromote,
        jpeg, png, markdown, xdvir, grid
Imports: graphics, grDevices, stats, utils, htmlwidgets (>= 1.6.0),
        htmltools, knitr (>= 1.33), jsonlite (>= 0.9.20), magrittr, R6,
        base64enc, mime
Enhances: waldo
Description: Provides medium to high level functions for 3D interactive graphics, including
    functions modelled on base graphics (plot3d(), etc.) as well as functions for
    constructing representations of geometric objects (cube3d(), etc.).  Output
    may be on screen using OpenGL, or to various standard 3D file formats including
    WebGL, PLY, OBJ, STL as well as 2D image formats, including PNG, Postscript, SVG, PGF.
License: GPL
URL: https://github.com/dmurdoch/rgl, https://dmurdoch.github.io/rgl/
SystemRequirements: OpenGL and GLU Library (Required for display in R.
        See "Installing OpenGL support" in README.md. Not needed if
        only browser displays using rglwidget() are wanted.), zlib
        (optional), libpng (>=1.2.9, optional), FreeType (optional),
        pandoc (>=1.14, needed for vignettes), LaTeX needed for
        optional latex3d() examples.
BugReports: https://github.com/dmurdoch/rgl/issues
VignetteBuilder: knitr, rmarkdown
Biarch: true
RoxygenNote: 7.3.3
NeedsCompilation: yes
Packaged: 2026-03-05 16:27:45 UTC; murdoch
Author: Duncan Murdoch [aut, cre],
  Daniel Adler [aut],
  Oleg Nenadic [ctb],
  Simon Urbanek [ctb],
  Ming Chen [ctb],
  Albrecht Gebhardt [ctb],
  Ben Bolker [ctb],
  Gabor Csardi [ctb],
  Adam Strzelecki [ctb],
  Alexander Senger [ctb],
  The R Core Team [ctb, cph],
  Dirk Eddelbuettel [ctb],
  The authors of Shiny [cph],
  The authors of knitr [cph],
  Jeroen Ooms [ctb],
  Yohann Demont [ctb],
  Joshua Ulrich [ctb],
  Xavier Fernandez i Marin [ctb],
  George Helffrich [ctb],
  Ivan Krylov [ctb],
  Michael Sumner [ctb],
  Mike Stein [ctb],
  Jonathon Love [ctb],
  Mapbox team [ctb, cph]
Maintainer: Duncan Murdoch <murdoch.duncan@gmail.com>
Repository: CRAN
Date/Publication: 2026-03-06 05:00:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-28 22:40:55 UTC; unix
Archs: rgl.so.dSYM

export(.check3d, 
  abclines3d, addNormals, addToSubscene3d, arc3d,
  arrow3d, as.mesh3d, as.rglscene, as.tmesh3d, as.triangles3d, asRow,
  asEuclidean, asEuclidean2, asHomogeneous, asHomogeneous2,
  aspect3d, axes3d, axis3d, box3d,
  bbox3d, bg3d, bgplot3d, 
  Buffer,
  checkDeldir,

  clear3d, clearSubsceneList, clipplanes3d, 
  clipMesh3d, clipObj3d, close3d, compare_proxy.mesh3d,
  contourLines3d,
  cube3d, cuboctahedron3d, cur3d, 
  currentSubscene3d, cylinder3d,
  decorate3d, deform.mesh3d, delFromSubscene3d, divide.mesh3d, dodecahedron3d, dot3d, drape3d, ellipse3d, 
  expect_known_scene, extrude3d,
  facing3d, figWidth, figHeight, filledContour3d,
  gc3d, getBoundary3d,
  getr3dDefaults, getShaders, getWidgetId, gltfTypes,
  GramSchmidt, grid3d, 
  highlevel, hook_rgl, hook_webgl, hover3d,
  icosahedron3d, identify3d, identityMatrix, ids3d,
  in_pkgdown, in_pkgdown_example, latex3d,
  layout3d, legend3d, light3d, lines3d, lowlevel, makeDependency, material3d, 
  mergeVertices, mesh3d, mfrow3d, movie3d, mtext3d, 
  newSubscene3d, next3d, normalize.mesh3d, observer3d, octahedron3d, oh3d, open3d, 
  par3d, par3dinterp, par3dinterpControl,  
  particles3d, pch3d, persp3d, planes3d, 
  play3d, plot3d, plotmath3d, points3d, 
  polygon3d, pop3d, projectDown, 
  qmesh3d, quads3d, readOBJ, readSTL, rgl.abclines, rgl.bbox, rgl.bg, rgl.bringtotop, rgl.clear,
  rgl.getAxisCallback, rgl.getMouseCallbacks, rgl.getWheelCallback, 
  rgl.close, rgl.cur, rgl.ids, rgl.init, rgl.light, rgl.lines, rgl.linestrips,
  rgl.clipplanes, rgl.material, 
  rgl.material.names, rgl.material.readonly,
  rgl.par3d.names, rgl.par3d.readonly,
  rgl.open, rgl.pixels, rgl.planes, rgl.points, rgl.pop, rgl.postscript,
  rgl.primitive, rgl.projection, rgl.quads, rgl.quit, rgl.Sweave, rgl.Sweave.off,
  rgl.select, rgl.select3d, rgl.set, rgl.snapshot, rgl.spheres, rgl.sprites,
  rgl.surface, rgl.texts, rgl.triangles, rgl.user2window,
  rgl.attrib, rgl.attrib.count, rgl.attrib.info, 
  rgl.dev.list, rgl.incrementID, rgl.useNULL,
  rgl.viewpoint, rgl.window2user, rglExtrafonts,
  rglFonts, rglId, rglMouse, rglShared, rglToLattice, rglToBase,
  r3dDefaults, rotate3d, rotationMatrix,
  scale3d, scaleMatrix, scene3d, segments3d, 
  select3d, selectionFunction3d, selectpoints3d,
  rgl.setAxisCallback, rgl.setMouseCallbacks, rgl.setWheelCallback,
  safe.dev.off,
  set3d, setAxisCallbacks, setGraphicsDelay, setupKnitr, 
  setUserCallbacks, setUserShaders, shade3d, shadow3d,

  shapelist3d, shinyGetPar3d, shinySetPar3d, shinyResetBrush, 
  show2d, snapshot3d, 
  spheres3d, spin3d, sprites3d, subdivision3d, 
  subsceneInfo, subsceneList, Sweave.snapshot,
  surface3d, tagged3d, terrain3d, 
  tetrahedron3d, text3d, texts3d, textureSource,
  thigmophobe3d, title3d, 
  tkpar3dsave, tkspinControl, tkspin3d, 
  toggleWidget, triangulate, 
  tmesh3d, transform3d, translate3d, translationMatrix, triangles3d,
  turn3d, useSubscene3d, view3d, wire3d, 
  writeASY, writeOBJ, writePLY, writeSTL, writeWebGL)

 S3method(dot3d, shapelist3d)
 S3method(wire3d, shapelist3d)
 S3method(shade3d, shapelist3d)  
 S3method(translate3d, shapelist3d)
 S3method(rotate3d, shapelist3d)
 S3method(scale3d, shapelist3d)
 S3method(addNormals, shapelist3d)
 
 S3method(dot3d, mesh3d)
 S3method(translate3d, mesh3d)
 S3method(rotate3d, mesh3d)
 S3method(scale3d, mesh3d)
 S3method(merge, mesh3d)
 
 S3method(wire3d, mesh3d)
 S3method(shade3d, mesh3d) 
 S3method(subdivision3d, mesh3d)
 S3method(addNormals, mesh3d)
 
 S3method(plot3d, mesh3d)
 
 S3method(all.equal, mesh3d)
 S3method(all.equal, rglscene)
 
 S3method(as.mesh3d, deldir)
 S3method(as.mesh3d, triSht)
 S3method(as.mesh3d, ashape3d)
 S3method(as.mesh3d, rglId)
 S3method(as.mesh3d, rglobject)
 S3method(as.mesh3d, rglsubscene)
 S3method(as.mesh3d, rglscene)
 S3method(as.mesh3d, mesh3d)
 
 S3method(as.mesh3d, default)
 
 S3method(as.tmesh3d, mesh3d)
 S3method(as.tmesh3d, default)
 
 S3method(as.triangles3d, mesh3d)
 S3method(as.triangles3d, rglId)
 
 S3method(translate3d, default)
 S3method(rotate3d, default)
 S3method(scale3d, default)
 
 S3method(ellipse3d, default)
 S3method(ellipse3d, lm)
 S3method(ellipse3d, glm)
 S3method(ellipse3d, nls)
 
 S3method(plot3d, default)
 S3method(persp3d, default)
 S3method(persp3d, "function")
 S3method(persp3d, deldir)
 S3method(persp3d, triSht)
 S3method(persp3d, ashape3d)
 S3method(persp3d, formula)
 
 S3method(plot3d, rglscene)
 S3method(plot3d, rglobject)
 S3method(plot3d, rglbboxdeco)
 S3method(plot3d, rglbackground)
 S3method(plot3d, rglsubscene)
 S3method(plot3d, rglWebGL)
 S3method(plot3d, "function")
 S3method(plot3d, deldir)
 S3method(plot3d, triSht)
 S3method(plot3d, ashape3d)
 S3method(plot3d, formula)
 S3method(plot3d, lm)
 
 S3method(print, rglscene)
 S3method(print, rglobject)
 S3method(print, rglsubscene)
 S3method(print, rglId)
 S3method(print, rglOpen3d)
 S3method(print, mesh3d)
 S3method(print, shapelist3d)
 S3method(print, rglMouseSelection)
 S3method(print, rglshaders)
 
 S3method(knit_print, rglId)
 S3method(knit_print, rglOpen3d)
 
 S3method(summary, rglscene)
 S3method(summary, rglsubscene)
 
 S3method(drape3d, default)
 S3method(drape3d, mesh3d)

 S3method(contourLines3d, rglId)
 S3method(contourLines3d, mesh3d)
 S3method(filledContour3d, rglId)
 S3method(filledContour3d, mesh3d)
 
 S3method(sew, rglRecordedplot)
 S3method(is_low_change, rglRecordedplot)
 
 if(.Platform$OS.type == "windows") {
   importFrom(utils, getWindowsHandle)
 }
 
 importFrom(graphics, legend, par, plot, plot.new, polygon,
       strwidth, strheight)
 importFrom(grDevices, as.raster, col2rgb, colorRamp, dev.cur, dev.list, dev.new,
       dev.off, dev.prev, dev.set, png, postscript, rgb, xy.coords, xyz.coords)
 importFrom(stats, approxfun, get_all_vars, model.frame, 
      qchisq, qf, splinefun, terms, var)
 importFrom(tools, file_ext)
 importFrom(utils, capture.output, count.fields, file_test, 
            flush.console, packageVersion, read.table, head, tail)
 importFrom(R6, R6Class)
 
# These were in rglwidget
 
export(rglwidget, renderRglwidget, rglwidgetOutput,
       playwidget, renderPlaywidget, playwidgetOutput,
       subsetControl, propertyControl, clipplaneControl, ageControl, vertexControl,
       elementId2Prefix, registerSceneChange, sceneChange,
       "%>%")

importFrom(htmlwidgets, createWidget, prependContent, saveWidget, 
           shinyRenderWidget, shinyWidgetOutput, sizingPolicy)
importFrom(htmltools, css, HTML, htmlDependency, htmlPreserve, img,
           includeScript, tags, tagAppendAttributes,
     tagHasAttribute, tagList, browsable, resolveDependencies)
importFrom(jsonlite, toJSON, base64_dec)
importFrom(knitr, asis_output, fig_path, hook_plot_custom, include_graphics, 
           is_low_change, knit_hooks, 
           knit_meta_add, knit_print, opts_current, opts_hooks, opts_knit, 
           pandoc_to, sew)
importFrom(magrittr, "%>%")
importFrom(stats, coef, predict, residuals)
importFrom(base64enc, base64encode)

RGL - 3D visualization device system for R using OpenGL
=======================================================

INTRODUCTION
------------
The RGL package is a visualization device system for R, using 
OpenGL as the rendering backend. An rgl device at its core 
is a real-time 3D engine written in C++. It provides an 
interactive viewpoint navigation facility (mouse + wheel support)
and an R programming interface.


LICENSE
-------
The software is released under the GNU Public License. 
See "COPYING" textfile for details.


FEATURES
--------
- portable R Plug-in using OpenGL on MacOSX, Win32 and X11
- R programming interface 
- interactive viewpoint navigation
- automatic data focus
- geometry primitives: 
  points, lines, triangles, quads, texts, point sprites
- high-level geometry:
  surface, spheres
- up to 8 light sources
- alpha-blending (transparency)
- side-dependant fill-mode rendering (dots,wired and filled)
- texture-mapping with mipmapping and environment mapping support
- environmental effects:
  fogging, background sphere
- bounding box with axis ticks marks
- undo operation:
  shapes and light-sources are managed on type stacks,
  where the top-most objects can be pop'ed, or any item specified by
  an identifier can be removed


PLATFORMS
---------
Mac OS X
Windows 2000/XP/Vista
Unix-derivatives


BUILD TOOLS
-----------
R recommended tools (gcc toolchain)


REQUIREMENTS
------------
Windowing System (osx/carbon, unix/x11 or win32)
OpenGL Library 
OpenGL Utility Library (GLU)

Installation on Debian:
-----------------------
aptitude install libgl1-mesa-dev libglu1-mesa-dev


OPTIONS
-------
libpng library version 1.0.4 or newer (pixmap import/export support)
freetype library for resizable anti-aliased fonts


BUILDING/INSTALLING FROM SOURCE PACKAGE
---------------------------------------
The R build tool is the primary tool to build the RGL package.


BUILDING ON MICROSOFT WINDOWS
-----------------------------

Libpng support:

Make sure bitmap dll has been built as described in the 
R Build Instructions for Windows.

These two files are needed:

%R_HOME%\src\gnuwin32\bitmap\libpng\libpng.a
%R_HOME%\bin\Rzlib.dll (or %R_HOME%\src\extra\zlib\libz.a, in older R versions)

Freetype support:

You need to obtain a Windows build of the Freetype library, version 2.3.5 or better.
A version is available at http://gnuwin32.sourceforge.net/packages/freetype.htm.
Set the environment variable LIB_FREETYPE to give the full path to the
install directory.

BUILDING ON UNIX-STYLE OS (MacOS X, Linux, FreeBSD, ... ) 
---------------------------------------------------------

The build is controlled by an autoconf configure script.
You provide the options through the R CMD build/INSTALL command

  e.g. 
    $ R CMD INSTALL --configure-args="<configure args...>" rgl


  COMMON UNIX-STYLE OS OPTIONS
  ----------------------------

  --with-gl-includes=<path>
    GL C header files include path

  --with-gl-libraries=<path>
    GL library linkage path

  --with-gl-prefix=<libprefix>
    GL library prefix (e.g. Mesa)
  
  --disable-libpng
    disable libpng support

  --with-libpng-prefix=<install location>
    force LibPNG library install prefix (e.g. /usr/local)

  --disable-libpng-config
    explicitly disable libpng-config
  
  --disable-libpng-dynamic
    use static libpng library
    
You may find that your distro doesn't have all necessary development
libraries installed:  read the error messages for hints!  This line
has been reported to be sufficient on FC 5 or 6:

 yum install mesa-libGL-devel mesa-libGLU-devel libpng-devel



  MAC OS X OPTIONS
  ----------------

  --disable-cocoa
    force a X11 port


  X11 WINDOWING SYSTEM OPTIONS
  ----------------------------

  --x-includes=<path>
    X11 C header files include path

  --x-libraries=<path>
    X11 library linkage path


FIRST DEMOS: LOADING LIBRARY AND RUNNING DEMONSTRATIONS
-------------------------------------------------------

> library(rgl)
> demo(rgl)


CREDITS
-------
Daniel Adler <dadler@uni-goettingen.de>
Duncan Murdoch <murdoch@stats.uwo.ca>
Oleg Nenadic <onenadi@uni-goettingen.de>
Simon Urbanek <simon.urbanek@math.uni-augsburg.de>
Ming Chen <mchen34@uwo.ca>
Albrecht Gebhardt <albrecht.gebhardt@uni-klu.ac.at>
Ben Bolker <bolker@zoo.ufl.edu>
Gabor Csardi <csardi@rmki.kfki.hu>
Adam Strzelecki <ono@java.pl>
Alexander Senger <senger@physik.hu-berlin.de>

WEBSITE
-------
http://rgl.neoscientists.org

<!-- README.md is generated from README.Rmd. Please edit that file -->

# RGL - 3D visualization device system for R using OpenGL

![](man/figures/READMEpolyhedra-1.-rgl.png)<!-- -->

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/rgl)](https://CRAN.R-project.org/package=rgl)
[![R-CMD-check](https://github.com/dmurdoch/rgl/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/dmurdoch/rgl/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

## INTRODUCTION

The RGL package is a visualization device system for R, using OpenGL or
WebGL as the rendering backend. An OpenGL rgl device at its core is a
real-time 3D engine written in C++. It provides an interactive viewpoint
navigation facility (mouse + wheel support) and an R programming
interface. WebGL, on the other hand, is rendered in a web browser; rgl
produces the input file, and the browser shows the images.

## WEBSITE

A `pkgdown` website is here:

<https://dmurdoch.github.io/rgl/>

The unreleased development version website is here:

<https://dmurdoch.github.io/rgl/dev/>

See [this
vignette](https://dmurdoch.github.io/rgl/dev/articles/pkgdown.html) for
details on producing your own `pkgdown` website that includes `rgl`
graphics.

The currently active development site is here:

<https://github.com/dmurdoch/rgl>

## INSTALLATION

Most users will want to install the latest CRAN release. For Windows,
macOS and some Linux platforms, installation can be easy, as CRAN
distributes binary versions:

    # Install latest release from CRAN
    install.packages("rgl")

To install the latest development version from Github, you’ll need to do
a source install. Those aren’t easy! Try

    # Install development version from Github
    remotes::install_github("dmurdoch/rgl")

If that fails, read the instructions below.

Currently installs are tested on older R versions back to R 3.5.x, but
this version of `rgl` may work back as far as R 3.3.0.

## LICENSE

The software is released under the GNU Public License. See
[COPYING](./COPYING) for details.

## FEATURES

- portable R package using OpenGL (if available) on macOS, Win32 and X11
- can produce 3D graphics in web pages using WebGL
- R programming interface
- interactive viewpoint navigation
- automatic data focus
- geometry primitives: points, lines, triangles, quads, texts, point
  sprites
- high-level geometry: surface, spheres
- up to 8 light sources
- alpha-blending (transparency)
- side-dependent fill-mode rendering (dots, wired and filled)
- texture-mapping with mipmapping and environment mapping support
- environmental effects: fogging, background sphere
- bounding box with axis ticks marks
- undo operation: shapes and light-sources are managed on type stacks,
  where the top-most objects can be popped, or any item specified by an
  identifier can be removed

## PLATFORMS

macOS (with only partial support on Tahoe, as OpenGL support is
currently deficient on that platform)  
Windows 7+  
Unix-derivatives

## BUILD TOOLS

R recommended tools (gcc toolchain) On Windows, Rtools is needed.

## REQUIREMENTS

**For OpenGL display:**

Windowing System (unix/x11 or Windows)  
OpenGL Library  
OpenGL Utility Library (GLU)

**For WebGL display:**

A browser with WebGL enabled. See <https://get.webgl.org>.

## Installing OpenGL support

**Debian and variants including Ubuntu:**

    aptitude install libgl1-mesa-dev libglu1-mesa-dev

**Fedora:**

    yum install mesa-libGL-devel mesa-libGLU-devel libpng-devel

**macOS:**

Install XQuartz.  
`rgl` should work with XQuartz 2.7.11 or newer, but it will probably
need rebuilding if the XQuartz version changes. XQuartz normally needs
re-installation whenever the macOS version changes.

Currently Tahoe, the most recent version of macOS, does not support
OpenGL in XQuartz, and it appears unlikely such support will be coming.

**Windows:**

Windows normally includes OpenGL support, but to get the appropriate
include files etc., you will need the appropriate version of
[Rtools](https://cran.r-project.org/bin/windows/Rtools/) matched to your
R version.

## Options

The **libpng** library version 1.2.9 or newer is needed for pixmap
import/export support.

The **freetype** library is needed for resizable anti-aliased fonts.

## BUILDING/INSTALLING

Binary builds of `rgl` are available for some platforms on CRAN.

For source builds, install the prerequisites as described above,
download the tarball and at the command line run

    R CMD INSTALL rgl_1.3.30.tar.gz

(with the appropriate version of the tarball). The build uses an
`autoconf` configure script; to see the options, expand the tarball and
run `./configure --help`.

Alternatively, in R run

    install.packages("rgl")

to install from CRAN, or

    remotes::install_github("dmurdoch/rgl")

to install the development version from Github.

### BUILDING ON MACOS

To build on MacOS using one of the ARM64 chips (currently M1, M2 or M3),
follow the instructions on <https://mac.r-project.org/tools/> to install
the tools and libraries into `/opt/R/arm64`. It is important that
`/opt/R/arm64/bin` appear in your PATH before `/usr/local/bin` if the
latter directory has been used for x86_64 installs. If you don’t do
this, or have some other error in setting things up, you’ll get a
warning during `rgl` installation saying that some configure test
failed, and `rgl` will be installed without OpenGL support.

Some versions of RStudio (including 2024.04.2+764) have a bug that
modifies your PATH on startup and again after every package
installation, putting `/usr/local/bin` at the head of the PATH. If you
are building `rgl` in such a system you need to remove files from
`/usr/local/bin` if there’s a file with the same name in
`/opt/R/arm64/bin`. Hopefully this bug will be fixed soon!

### BUILDING WITHOUT OPENGL

As of version 0.104.1, it is possible to build the package without
OpenGL support on Unix-alikes (including macOS) with the configure
option –disable-opengl For example,

    R CMD INSTALL --configure-args="--disable-opengl" rgl_1.3.30.tar.gz 

On Windows, OpenGL support cannot currently be disabled.

## DOCUMENTATION and DEMOS:

    library(rgl)
    browseVignettes("rgl")
    demo(rgl)

## CREDITS

Daniel Adler <dadler@uni-goettingen.de>  
Duncan Murdoch <murdoch@stats.uwo.ca>  
Oleg Nenadic <onenadi@uni-goettingen.de>  
Simon Urbanek <simon.urbanek@math.uni-augsburg.de>  
Ming Chen <mchen34@uwo.ca>  
Albrecht Gebhardt <albrecht.gebhardt@uni-klu.ac.at>  
Ben Bolker <bolker@zoo.ufl.edu>  
Gabor Csardi <csardi@rmki.kfki.hu>  
Adam Strzelecki <ono@java.pl>  
Alexander Senger <senger@physik.hu-berlin.de>  
The R Core Team for some code from R.  
Dirk Eddelbuettel <edd@debian.org>  
The authors of Shiny for their private RNG code.  
The authors of `knitr` for their graphics inclusion code. Jeroen Ooms
for `Rtools40` and `FreeType` help.  
Yohann Demont for Shiny code, suggestions, and testing.  
Joshua Ulrich for a lot of help with the Github migration. Xavier
Fernandez i Marin for help debugging the build.  
George Helffrich for draping code.  
Ivan Krylov for window_group code in X11.  
Michael Sumner for as.mesh3d.default enhancement.  
Tomas Kalibera for `winutf8` and other help.  
David Hugh-Jones for documentation improvements.  
Trevor Davis for a `snapshot3d` patch.  
Mike Stein for pointer-handling code.  
Jonathon Love for the `uname` patch.  
Volodymyr Agafonkin and many others for the `earcut` triangulation code.

The currently shipping OpenGL32.DLL from Microsoft only has entry points
for OpenGL 1.1. If an application wants to use OpenGL {1.2, 1.3, 1.4, 1.5}
functions, it has to use wglGetProcAddress() in order to obtain the entry
points from the driver. The files in this distribution enable the application
pretend that there is full support for OpenGL {1.2/1.3/1.4/1.5} if the
underlying implementation supports OpenGL {1.2/1.3/1.4/1.5}.

A useful feature of the framework is that it provides a safety layer by
initializing each function to a "default" initialization function with
matching arguments and return type. On debug builds, the initialization
function adds an ASSERT to catch unwitting usage of NULL function pointers
by the application.

The framework is also readily usable for multi-context and multi-threaded
applications, including application code that demonstrates how to
achieve this.

Additionally, this distribution provides full support for all the 
extension entry points in glext.h and wglext.h.

In order to use this scheme, the application needs to do the following:

=====================================================================

For applications that do not create rendering contexts on different devices 
(e.g. multi-display with different vendors supporting the two displays):

1. Include "glprocs.h"  in addition to including "gl.h"
2. Compile "glprocs.c" and ensure that glprocs.obj is linked with the
   application code.
3. Call OpenGL{1.2,1.3,1.4,1.5} and extension procs after verifying that
   the implementation has support for the underlying feature. Example:

   if (supports12) {
       glDrawRangeElements (.....)
   }
   
   NOTE: The application can call "glGetString(GL_VERSION)" to determine
   the version of OpenGL supported on a particular implementation.

=====================================================================

For aplications that create rendering contexts on different devices, the
application has to perform the following additional steps:

1. Allocate memory for _GLextensionProcs in the application data
   structures for each device. Example:

   typedef struct {
        appType             appData;

        _GLextensionProcs   extensionProcs;
   } ContextData;

   ContextData cdev1, cdev2;

2. Initialize the extension procs.

   _InitExtensionProcs (&cdev1.extensionProcs);
   _InitExtensionProcs (&cdev2.extensionProcs);

3. Provide a mechanism for accessing the current device. Example:

   ContextData *currentContextData;

   static _inline _GLextensionProcs *_GET_TLS_PROCTABLE(void)
   {
       return (&currentContxtData->extensionProcs);
   }

   Another Example (when the two different devices are being used
   on two different contexts):

   static _inline _GLextensionProcs *_GET_TLS_PROCTABLE(void)
   {
       ContextData *tla = GetThreadSpecificData();
       return (&tla->extensionProcs);
   }

   (NOTE: Aplication has to replace this function in glprocs.h with 
    its own customized function).
        
=====================================================================

This distribution additionally contains the following:

Windows Demo Source:
--------------------
windemo:
    demogl.{c|h} - sample initialization, rendering functions and a sample
    data structure for tracking per-window data (useful for apps. with 
    multiple windows).
    Makefile:Builds all the demos (singlewin, multiwin, and multithread)

singlewin:
    demo.c: Shows how to use the SDK for an application with a single window.
    Makefile: Builds the singlewin sample (demo.exe).

multiwin:
    demo.c: Shows how to use the SDK for an application with two windows
    Makefile: Builds the multiwin sample (demo.exe)

multithread:
    demo.c: Shows how to use the SDK for a multi-threaded application.
    Makefile: Builds the multithread sample (demo.exe).

GLX Demo Source:
--------------------
glxdemo:
GLX demo similar to the singlewin Windows demo.

demogl.{c,h} - OpenGL rendering routines (this directory. Shared by
               programs in threee directories below).

The following three directories have the same "demo.c" program which
uses demogl, but is setup three different ways.

singlewin   - Window setup to render in a single window. Most normal 
              aplications work this way and don't have do anything 
              special for OpenGL 1.2, 1.3 and extension procs.

multiwin    - Multiple windows with a single message loop. At any given
              time, there is only context which is current. The application
              defines its own method for accessing the proc table 
              (see _APP_PROCTABLE flag in the Makefile). Since there is
              only window which is current, a global variable can be used
              to provide reference to the current proc table.

multithread - Multiple windows with a multiple message loops using 
              multiple threads. At any given time, there can be multiple
              current contexts.  The application defines its own method 
              for accessing the proc table (see _APP_PROCTABLE flag in 
              the Makefile). However, there can me multiple current contexts
              on different thread, the application uses thread local 
              storage to define proper access to the current proc table.

## ----setup, echo=FALSE, results="asis"----------------------------------------
source("setup.R")
setupKnitr(autoprint = FALSE)
set.seed(123)

## ----elementId----------------------------------------------------------------
library(rgl)
plotids <- with(iris, plot3d(Sepal.Length, Sepal.Width, Petal.Length, 
                  type="s", col=as.numeric(Species)))
rglwidget(elementId = "plot3drgl")

## -----------------------------------------------------------------------------
toggleWidget(sceneId = "plot3drgl", ids = plotids["data"], label = "Data")

## -----------------------------------------------------------------------------
names(plotids)
unclass(plotids)

## ----Pipes--------------------------------------------------------------------
rglwidget() %>%
toggleWidget(ids = plotids["data"], label = "Data")

## ----eval=FALSE---------------------------------------------------------------
# rglwidget() |>
# toggleWidget(ids = plotids["data"], label = "Data")

## ----"Control before widget"--------------------------------------------------
toggleWidget(NA, ids = plotids["data"], label = "Data") %>%
rglwidget(controllers = .) 

## ----eval=FALSE---------------------------------------------------------------
# toggleWidget(NA, ids = plotids["data"], label = "Data") |>
#   w => rglwidget(controllers = w)

## ----"Toggle subsets"---------------------------------------------------------
clear3d() # Remove the earlier display

with(subset(iris, Species == "setosa"), 
     spheres3d(Sepal.Length, Sepal.Width, Petal.Length, 
                  col=as.numeric(Species),
                  radius = 0.211,
                  tag = "setosa"))
with(subset(iris, Species == "versicolor"), 
     spheres3d(Sepal.Length, Sepal.Width, Petal.Length, 
               col=as.numeric(Species),
     	       radius = 0.211, 
     	       tag = "versicolor"))
with(subset(iris, Species == "virginica"), 
     spheres3d(Sepal.Length, Sepal.Width, Petal.Length, 
               col=as.numeric(Species),
     	       radius = 0.211,
     	       tag = "virginica"))
aspect3d(1,1,1)
decorate3d(tag = "axes")
rglwidget() %>%
toggleWidget(tags = "setosa") %>%
toggleWidget(tags = "versicolor") %>%
toggleWidget(tags = "virginica") %>%
toggleWidget(tags = "axes") %>% 
asRow(last = 4)

## ----Slider-------------------------------------------------------------------
rglwidget() %>%
playwidget(start = 0, stop = 3, interval = 1,
	   subsetControl(1, subsets = list(
	   			 Setosa = tagged3d("setosa"),
	   			 Versicolor = tagged3d("versicolor"),
	   			 Virginica = tagged3d("virginica"),
	   			 All = tagged3d(c("setosa", "versicolor", "virginica"))
	   			 )))

## ----"par3dinterpControl()"---------------------------------------------------
M <- r3dDefaults$userMatrix
fn <- par3dinterp(time = (0:2)*0.75, userMatrix = list(M,
                                      rotate3d(M, pi/2, 1, 0, 0),
                                      rotate3d(M, pi/2, 0, 1, 0)) )
rglwidget() %>%
playwidget(par3dinterpControl(fn, 0, 3, steps=15),
 	   step = 0.01, loop = TRUE, rate = 0.5)

## ----"vertexControl()"--------------------------------------------------------
setosavals <- subset(iris, Species == "setosa")
which <- which.min(setosavals$Sepal.Width)
init <- setosavals$Sepal.Length[which]
rglwidget() %>%
playwidget(
  vertexControl(values = matrix(c(init,   0,   0,   0, 
                                     8,   1,   1,   1), 
                                nrow = 2, byrow = TRUE),
                attributes = c("x", "red", "green", "blue"),
                vertices = which, tag = "setosa"),
	step = 0.01)

## ----"ageControl()"-----------------------------------------------------------
time <- 0:500
xyz <- cbind(cos(time/20), sin(time/10), time)
lineid <- plot3d(xyz, type="l", col = "black")["data"]
sphereid <- spheres3d(xyz[1, , drop=FALSE], radius = 8, col = "red")
rglwidget() %>%
playwidget(list(
  ageControl(births = time, ages = c(0, 0, 50),
		colors = c("gray", "red", "gray"), objids = lineid),
  ageControl(births = 0, ages = time,
		vertices = xyz, objids = sphereid)),
  start = 0, stop = max(time) + 20, rate = 50,
  components = c("Reverse", "Play", "Slower", "Faster",
                 "Reset", "Slider", "Label"),
  loop = TRUE)

## ----crosstalk,eval = requireNamespace("crosstalk", quietly=TRUE)-------------
# This example requires the crosstalk package
# We skip it if crosstalk is not available. 

ids <- with(iris, plot3d(Sepal.Length, Sepal.Width, Petal.Length, 
                  type="s", col=as.numeric(Species)))
par3d(mouseMode = "selecting")
rglwidget(shared = rglShared(ids["data"])) %>% 
rglMouse()

## ----"rglShared()",eval=requireNamespace("crosstalk", quietly = TRUE)---------
# This example requires the crosstalk package.  
# We skip it if crosstalk is not available. 

library(crosstalk)
sd <- SharedData$new(mtcars)
ids <- plot3d(sd$origData(), col = mtcars$cyl, type = "s")
# Copy the key and group from existing shared data
rglsd <- rglShared(ids["data"], key = sd$key(), group = sd$groupName())
rglwidget(shared = rglsd) %>%
asRow("Mouse mode: ", rglMouse(getWidgetId(.)), 
      "Subset: ", filter_checkbox("cylinderselector", 
		                "Cylinders", sd, ~ cyl, inline = TRUE),
      last = 4, colsize = c(1,2,1,2), height = 60)

## ----plot3d2------------------------------------------------------------------
plotids <- with(iris, plot3d(Sepal.Length, Sepal.Width, Petal.Length, 
                  type="s", col=as.numeric(Species)))
subid <- currentSubscene3d()
rglwidget(elementId="plot3drgl2")

## ----echo=FALSE, results="asis"-----------------------------------------------
writeIndex(cols = 5)

# Use the mouse to select points
# Original version written by Yohann Demont

if (!require("shiny"))
  stop("This demo requires shiny.")

if (!requireNamespace("crosstalk"))
  stop("This demo requires crosstalk.")

library(rgl)

ui <- fluidPage(
  sidebarLayout(
    mainPanel(tabsetPanel(id = "navbar",
                          selected = "3D",
                          tabPanel(title = "2D",
                                   plotOutput("plot_2D", brush = brushOpts(id = "plot_2D_brush",
                                                                           resetOnNew = TRUE,
                                                                           direction = "xy")),
                                   verbatimTextOutput("brush_info_2D")),
                          tabPanel(title = "3D",
                                   uiOutput("plot_3D_mousemode"),
                                   rglwidgetOutput("plot_3D"),
                                   verbatimTextOutput("brush_info_3D"),
                                   verbatimTextOutput("selected"))
    )),
    sidebarPanel(selectInput("plot_x", label = "x feature", choices = colnames(iris)[-5], selected = colnames(iris)[1]),
                 selectInput("plot_y", label = "y feature", choices = colnames(iris)[-5], selected = colnames(iris)[2]),
                 selectInput("plot_z", label = "z feature", choices = colnames(iris)[-5], selected = colnames(iris)[3]),
                 actionButton(inputId = "reset_brush", label = "reset brush"))
  ))

server <- function(input, output, session) {
  # 2D
  output$plot_2D <- renderPlot({
    plot(x = iris[, input$plot_x],
         y = iris[, input$plot_y],
         col = as.integer(iris[, "Species"]))
  })
  output$brush_info_2D <- renderPrint(str(input$plot_2D_brush))
  
  # 3D
  sharedData <- NULL
  output$brush_info_3D <- renderPrint(print(input$rgl_3D_brush, verbose = TRUE))
  
  # How to use selectionFunction3d ?
  output$selected <- renderPrint({
    if(length(input$rgl_3D_brush) == 0 || input$rgl_3D_brush$state == "inactive") return(NULL)
    cat("Selections from crosstalk:\n")
    # Need as.logical because selection() might return NULL
    print(which(as.logical(sharedData$selection())))
    cat("Selections using function:\n")
    f <- selectionFunction3d(input$rgl_3D_brush)
    which(f(iris[, c(input$plot_x, input$plot_y, input$plot_z)]))
  })
  
  output$plot_3D_mousemode <-
    renderUI({
      rglMouse( default = "trackball",
                stayActive = FALSE,
                choices = c("trackball", "selecting"),
                sceneId = "plot_3D")
    })
  open3d(useNULL = TRUE)
  output$plot_3D <- renderRglwidget({
    clear3d()
    dat <- iris[, c(input$plot_x, input$plot_y, input$plot_z, "Species")]
    dat$id <-as.character(seq_len(nrow(iris)))
    plot3d(x = dat[, 1:3], type = "s", size = 1, col = as.integer(iris[, "Species"]), aspect = TRUE)
    sharedData <<- rglShared(id = text3d(dat[, 1:3], text = dat[, "id"], adj = -0.5),
                             group = "SharedData_plot_3D_ids",
                             deselectedFade = 0,
                             selectedIgnoreNone = FALSE)
    shinyResetBrush(session, "rgl_3D_brush")
    rglwidget(shared = sharedData,
              shinyBrush = "rgl_3D_brush")
  })
  observeEvent(input$reset_brush, {
    session$resetBrush("plot_2D_brush")
    shinyResetBrush(session, "rgl_3D_brush")
  })
}

shinyApp(ui, server)

if (!require("shiny"))
  stop("This demo requires shiny.")

library(rgl)
library(misc3d)

options(rgl.useNULL = TRUE)
set.seed(123)

ui <- fluidPage(
  registerSceneChange(),
  titlePanel("Nelder-Mead"),
  sidebarLayout(
    sidebarPanel(
      helpText("The Nelder-Mead algorithm evaluates the function",
               "on the vertices of a simplex.  At each step it",
               "moves one vertex of the simplex to a better value."),
      sliderInput("Slider", min=0, max=59, step=1, value=0, label="Steps",
                  animate=animationOptions(200, loop=TRUE)),
      sliderInput("Slider2", min=0, max=59, step=1, value=0, label="Cumulative",
                  animate=animationOptions(200, loop=TRUE)),
      playwidgetOutput('thecontroller'),
      playwidgetOutput('thecontroller2'),
      actionButton('newStart', 'Restart')),
    mainPanel(
      rglwidgetOutput('thewidget', width = "100%", height = 512))
  )
)

u1 <- runif(1)
u2 <- runif(1)*(1-u1)
u3 <- 1 - u1 - u2
# Example modified from ?contour3d
#Example 2: Nested contours of mixture of three tri-variate normal densities
nmix3 <- function(x, y, z, m, s) {
  u1 * dnorm(x, m, s) * dnorm(y, m, s) * dnorm(z, m, s) +
    u2 * dnorm(x, -m, s) * dnorm(y, -m, s) * dnorm(z, -m, s) +
    u3 * dnorm(x, m, s) * dnorm(y, -1.5 * m, s) * dnorm(z, m, s)
}
f <- function(x,y,z) nmix3(x,y,z,.5,.5)
g <- function(n = 40, k = 5, alo = 0.1, ahi = 0.5, cmap = heat.colors) {
  th <- seq(0.05, 0.2, len = k)
  col <- rev(cmap(length(th)))
  al <- seq(alo, ahi, len = length(th))
  x <- seq(-2, 2, len=n)
  bg3d(col="white")
  contour3d(f,th,x,x,x,color=col,alpha=al) # nolint
}

f3 <- function(x) -f(x[1], x[2], x[3])

g(20,3)
surface <- scene3d()
close3d()

neldermead <- function(x, f) {
  n <- nrow(x)
  p <- ncol(x)
  
  if (n != p + 1) stop(paste('Need', p + 1, 'starting points'))
  
  fx <- rep(NA, n)
  for (i in 1:n) fx[i] <- f(x[i,])
  
  o <- order(fx)
  fx <- fx[o]
  x <- x[o,]
  xmid <- apply(x[1:p,], 2, mean)
  z1 <- xmid - (x[n,] - xmid)
  fz1 <- f(z1)
  
  if (fz1 < fx[1]) {
    z2 <- xmid - 2*(x[n,] - xmid)
    fz2 <- f(z2)
    if (fz2 < fz1) {
      x[n,] <- z2
    } else {
      x[n,] <- z1
    }
  } else if (fz1 < fx[p]) {
    x[n,] <- z1
  } else {
    if (fz1 < fx[n]) {
      x[n,] <- z1
      fx[n] <- fz1
    }
    z3 <- xmid + (x[n,] - xmid)/2
    fz3 <- f(z3)
    if (fz3 < fx[n]) {
      x[n,] <- z3
    } else {
      for (i in 2:n) {
        x[i,] <- x[1,] + (x[i,] - x[1,])/2
      }
    }
  }
  return(x)
}

showsimplex <- function(x, f, col="blue") {
  n <- nrow(x)
  z <- numeric(n)
  for (i in 1:n) z[i] <- f(x[i,])
  xyz <- cbind(x, z)
  
  # This is tricky:
  
  # 1. draw all lines, taking vertices two at a time:
  c(segments3d(xyz[as.numeric(combn(n, 2)),], col="black", depth_test = "lequal"),
    # 2. draw all faces, taking vertices three at a time:
    triangles3d(xyz[as.numeric(combn(n, 3)),], col=col, alpha=0.3))
}

setStartPoint <- function() {
  xyz <- matrix(rnorm(12, sd=0.1) + rep(rnorm(3,sd=2), each=4), 4, 3)
  subsets <-list()
  for (i in 1:60) {
    xyz <- neldermead(xyz,f3)
    subset <- showsimplex(xyz,f3)
    subsets <-c(subsets,list(subset))
  }
  names(subsets) <- seq_along(subsets)
  subsets
}

server <- function(input, output, session) {
  
  plot3d(surface)
  dev <- cur3d()
  save <- options(rgl.inShiny = TRUE)
  on.exit(options(save))
  
  session$onSessionEnded(function() {
    set3d(dev)
    close3d()
  })
  
  path <- reactiveValues(subsets = setStartPoint())
  
  observeEvent(input$newStart, {
    set3d(dev)
    
    deletes <- unique(unlist(path$subsets))
    
    if (length(deletes))
      delFromSubscene3d(deletes)
    subsets <- setStartPoint()
    adds <- unique(unlist(subsets))
    session$sendCustomMessage("sceneChange",
                              sceneChange("thewidget", delete = deletes, add = adds,
                                          skipRedraw = TRUE))
    path$subsets <- subsets
    updateSliderInput(session, "Slider", value=0)
    updateSliderInput(session, "Slider2", value=0)
    session$onFlushed(function()
      session$sendCustomMessage("sceneChange",
                                sceneChange("thewidget", skipRedraw = FALSE)))
  })
  
  output$thewidget <- renderRglwidget({
    rglwidget(controllers=c("thecontroller", "thecontroller2"))
  })
  
  output$thecontroller <-
    renderPlaywidget({
      if (length(path$subsets))
        playwidget("thewidget", respondTo = "Slider",
                   subsetControl(1, path$subsets),
                   start = 1, stop = length(path$subsets))
    })
  
  output$thecontroller2 <-
    renderPlaywidget({
      if (length(path$subsets))
        playwidget("thewidget", respondTo = "Slider2",
                   subsetControl(1, path$subsets, accumulate = TRUE))
    })
}

if (interactive())
  shinyApp(ui = ui, server = server)

wave <- function(time) {
  x <- seq(0,2, len=100)
  
  wavefn <- function(x) x * sin(-5*time + 1.5 * (x/2) * 2*pi)/10
  deriv <- function(x) (wavefn(x + 0.01) - wavefn(x - 0.01))/0.02

  arclen <- cumsum(sqrt(1 + deriv(x)^2))*(x[2]-x[1])
  keep <- arclen < 2
  x <- x[keep]
  y <- matrix(wavefn(x), length(x),20)
  z <- matrix(seq(0,1, len=20), length(x), 20, byrow=TRUE)
  arclen <- arclen[keep]
  
  par3d(skipRedraw = TRUE)
  if (nrow(ids3d())) pop3d()
  surface3d(x,y,z, texture_s=matrix(arclen/2, length(x), 20), texture_t=z, col="white")
  c(list(skipRedraw = FALSE), spin(time))
}

open3d()
material3d(texture = system.file("textures","rgl2.png", package="rgl"))
spin <- spin3d(rpm=6,axis=c(0,0,1))
if (!rgl.useNULL())
  play3d(wave, 10, startTime = 5)

library(rgl)

# regression :  this failed for headless tests
par3d(userMatrix = diag(4))

if (!rgl.useNULL())
  for(demo in demo(package="rgl")$results[,"Item"]) 
    if (!(demo %in% c("rgl", "lsystem", "rglExamples", 
  	            	    "shinyDemo", "simpleShinyRgl",
  	            	    "shinyMouse")))
      demo(demo, package="rgl", character.only=TRUE)

xprod <- function(a, b) 
    c(a[2]*b[3] - a[3]*b[2],
       a[3]*b[1] - a[1]*b[3],
       a[1]*b[2] - a[2]*b[1])
       
vlen <- function(a) sqrt(sum(a^2))

angle <- function(a,b) {
    dot <- sum(a*b)
    acos(dot/vlen(a)/vlen(b))
}
 
clamp <- function(x, min, max)  pmin(pmax(x, min), max)

mouseNone <- function(button = 1, dev = cur3d() ) {

    cur <- cur3d()
    
    for (i in dev) {
        set3d(i, TRUE)
        rgl.setMouseCallbacks(button, begin = NULL, update = NULL, end = NULL, dev = dev)
    }
    set3d(cur)
}

mouseTrackball <- function(button = 1, dev = cur3d() ) {
    width <- height <- rotBase <- NULL
    userMatrix <- list()
    cur <- cur3d()
    
    screenToVector <- function(x, y) {
      radius <- max(width, height)/2
      centre <- c(width, height)/2
      pt <- (c(x, y) - centre)/radius
      len <- vlen(pt)

      if (len > 1.e-6) pt <- pt/len

      maxlen <- sqrt(2)
      angle <- (maxlen - len)/maxlen*pi/2
      z <- sin(angle)
      len <- sqrt(1 - z^2)
      pt <- pt * len
      return(c(pt, z))
    }
    
    trackballBegin <- function(x, y) {
        vp <- par3d("viewport")
        width <<- vp[3]
        height <<- vp[4]
        cur <<- cur3d()
        for (i in dev) {
            if (inherits(try(set3d(i, TRUE)), "try-error")) dev <<- dev[dev != i]
            else userMatrix[[i]] <<- par3d("userMatrix")
        }
        set3d(cur, TRUE)
        rotBase <<- screenToVector(x, height - y)
    }
    
    trackballUpdate <- function(x,y) {
        rotCurrent <- screenToVector(x, height - y)
        angle <- angle(rotBase, rotCurrent)
        axis <- xprod(rotBase, rotCurrent)
        mouseMatrix <- rotationMatrix(angle, axis[1], axis[2], axis[3])
        for (i in dev) {
            if (inherits(try(set3d(i, TRUE)), "try-error")) dev <<- dev[dev != i]
            else par3d(userMatrix = mouseMatrix %*% userMatrix[[i]])
        }
        set3d(cur, TRUE)
    }
    
    for (i in dev) {
        set3d(i, TRUE)
        rgl.setMouseCallbacks(button, begin = trackballBegin, update = trackballUpdate, end = NULL,
                              dev = dev)
    }
    set3d(cur, TRUE)
}

mouseXAxis<- function(button = 1, dev = cur3d() , left=TRUE) {
    mouseOneAxis(button, dev, axis=c(1,0,0), left=left) 
}

mouseYAxis<- function(button = 1, dev = cur3d(), left = TRUE ) {
    mouseOneAxis(button, dev, axis=c(0,1,0), left=left) 
}

mouseZAxis<- function(button = 1, dev = cur3d(), left=TRUE) {
    mouseOneAxis(button, dev, axis=c(0,0,1), left=left) 
}

mouseOneAxis <- function(button = 1, dev = cur3d(), axis = c(1,0,0), left = TRUE ) {
    width <- height <- rotBase <- NULL
    userMatrix <- list()
    cur <- cur3d()
    
    screenToVector <- function(x, y) {
      radius <- max(width, height)/2
      centre <- c(width, height)/2
      pt <- (c(x, y) - centre)/radius
      len <- vlen(pt)

      if (len > 1.e-6) pt <- pt/len

      maxlen <- sqrt(2)
      angle <- (maxlen - len)/maxlen*pi/2
      z <- sin(angle)
      len <- sqrt(1 - z^2)
      pt <- pt * len
      return(c(pt, z))
    }

    oneAxisBegin <- function(x, y) {
        vp <- par3d("viewport")
        width <<- vp[3]
        height <<- vp[4]
        cur <<- cur3d()
        for (i in dev) {
            if (inherits(try(set3d(i, TRUE)), "try-error")) dev <<- dev[dev != i]
            else userMatrix[[i]] <<- par3d("userMatrix")
        }
        set3d(cur, TRUE)
        rotBase <<- screenToVector(x, height/2)
    }
    
    oneAxisUpdate <- function(x,y) {
        rotCurrent <- screenToVector(x, height/2)
        angle <- rotCurrent[1] - rotBase[1]
        mouseMatrix <- rotationMatrix(angle, axis[1], axis[2], axis[3])
        for (i in dev) {
            if (inherits(try(set3d(i, TRUE)), "try-error")) dev <<- dev[dev != i]
            else {
		if (left)
		    par3d(userMatrix = mouseMatrix %*% userMatrix[[i]])
		else
		    par3d(userMatrix = userMatrix[[i]] %*% mouseMatrix)
	    }
        }
        set3d(cur, TRUE)
    }
    
    for (i in dev) {
        set3d(i, TRUE)
        rgl.setMouseCallbacks(button, begin = oneAxisBegin, update = oneAxisUpdate, end = NULL, dev = dev)
    }
    set3d(cur, TRUE)
}

mousePolar <- function(button = 1, dev = cur3d()) {
   
    screenToPolar <- function(x,y) {
    	r <- min(width, height)/2
    	dx <- clamp(x - width/2, -r, r)
    	dy <- clamp(y - height/2, -r, r)
    	return( asin( c(dx, -dy)/r ) )
    }

    cur <- cur3d()
    width <- height <- dragBase <- dragCurrent <- NULL
    camBase <- list()
    
    polarBegin <- function(x, y) {
        vp <- par3d("viewport")
        width <<- vp[3]
        height <<- vp[4]
        dragBase <<- screenToPolar(x, y)
    	cur <<- cur3d()
    	for (i in dev) {
	    if (inherits(try(set3d(i, TRUE)), "try-error")) dev <<- dev[dev != i]
            else {
		m <- par3d("userMatrix")
		svd <- svd(m[1:3, 1:3])
		m[1:3, 1:3] <- svd$u %*% t(svd$v)
		theta <- atan2(-m[1,3], m[1,1])
		m <-  m %*% rotationMatrix(theta, 0,1,0)
		svd <- svd(m[1:3, 1:3])
		m[1:3,1:3] <- svd$u %*% t(svd$v)	
		phi <- atan2(-m[2,3], m[3,3])
		camBase[[i]] <<- c(theta, phi)
	    }
	}
        set3d(cur, TRUE)
    }   
    
    polarUpdate <- function(x,y) {
        dragCurrent <<- screenToPolar(x, y)        
        for (i in dev) {
            if (inherits(try(set3d(i, TRUE)), "try-error")) dev <<- dev[dev != i]
            else {
		newpos <- camBase[[i]] - ( dragCurrent - dragBase )
		newpos[2] <- clamp(newpos[2], -pi/2, pi/2)
		mouseMatrix <- rotationMatrix(newpos[2], 1, 0, 0) %*% rotationMatrix(-newpos[1], 0, 1, 0)
		par3d(userMatrix = mouseMatrix) 
	    }
        }
        set3d(cur, TRUE)
    }
    
    for (i in dev) {
	set3d(i, TRUE)
	rgl.setMouseCallbacks(button, begin = polarBegin, update = polarUpdate, end = NULL, dev = dev)
    }
    set3d(cur, TRUE)
}

# Set background colour based on x,y position in the window

mouseBG <- function(button = 1, dev = cur3d(), init = "white", rate = cbind(c(1,0,1),c(0,1,1)), space = c("rgb", "hsv")) {
    cur <- cur3d()
    space <- match.arg(space)
    init <- col2rgb(init)/255
    if (space == "hsv")
    	init <- rgb2hsv(init*255)

    width <- height <- lambda0 <- lambda <- NULL
    
    bgBegin <- function(x, y) {
        lambda0 <<- c(x/width, 1-y/height)    # nolint
        vp <- par3d("viewport")
    	width <<- vp[3]
    	height <<- vp[4]
    }
        
    bgUpdate <- function(x,y) {
        lambda <<- c(x/width, 1-y/height) - lambda0
        color <- clamp(init + rate %*% lambda, 0, 1)
        x <- color[1]
        y <- color[2]
        z <- color[3]
        if (space == "rgb")
            color <- rgb(x,y,z)    
        else
            color <- hsv(x,y,z)
        for (i in dev) {
            if (inherits(try(set3d(i, TRUE)), "try-error")) dev <<- dev[dev != i]
            else bg3d(color=color)  
        }
        set3d(cur, TRUE)
    }
    
    bgEnd <- function() {
        init <<- clamp(init + rate %*% lambda, 0, 1)
    }
    
    for (i in dev) {
        set3d(i, TRUE)
        rgl.setMouseCallbacks(button, begin = bgBegin, update = bgUpdate, end = bgEnd, dev = dev)
    }
    set3d(cur, TRUE)
}

# Set time using an arbitrary par3dinterp function

mouseInterp <- function(button = 1, dev = cur3d(), fn, init = 0, range = NULL, direction=c(1,0)) {
    cur <- cur3d()
    time <- init
    x0 <- width <- NULL
    
    interpBegin <- function(x, y) {
    	vp <- par3d("viewport")
        width <<- vp[3]
        x0 <<- sum(direction*c(x,y))
    }
        
    interpUpdate <- function(x,y) {
        time <<- init + (sum(direction*c(x,y)) - x0)/width
        if (!is.null(range)) time <<- clamp(time, range[1], range[2])
        for (i in dev) {
            if (inherits(try(set3d(i, TRUE)), "try-error")) dev <<- dev[dev != i]
            else par3d(fn(time))
        }
        set3d(cur, TRUE)
    }
    
    interpEnd <- function() {
        init <<- time
    }
    
    for (i in dev) {
        set3d(i, TRUE)
        rgl.setMouseCallbacks(button, begin = interpBegin, update = interpUpdate, end = interpEnd, dev = dev)
    }
    set3d(cur, TRUE)
}

mouseZoom <- function(button = 1, dev = cur3d()) 
    mouseInterp(button,dev=dev,fn=par3dinterp(times=c(-4,4)/4, zoom=c(10^(-4),10^4),method="linear"),
                      init=log10(par3d("zoom"))/4,range=c(-4,4)/4,direction=c(0,-1))
 
mouseFOV <- function(button = 1, dev = cur3d())
    mouseInterp(button,dev=dev,fn=par3dinterp(times=c(1,179)/180, FOV=c(1,179),method="linear"), 
                      init=par3d("FOV")/180, range = c(1,179)/180, direction=c(0,1))
                      
# Synchronize mouse control of two windows for stereo view
example(surface3d, echo = FALSE)
par3d(windowRect= c(0,32,512,544), userMatrix = rotationMatrix(5*pi/180, 0,1,0) %*% par3d("userMatrix") )
w1 <- cur3d()
example(surface3d, echo = FALSE)
par3d(windowRect = c(512,32,1024,544))
w2 <- cur3d()
mouseTrackball(dev=c(w1,w2))
mouseZoom(2,dev=c(w1,w2))
mouseFOV(3,dev=c(w1,w2))

# A simple Shiny demo written by Dieter Menne

options(rgl.useNULL = TRUE)
if (!require(shiny))
  stop("This demo requires shiny.")

library(rgl)

app <- shinyApp(
  ui = bootstrapPage(
    checkboxInput("rescale", "Rescale"),
    rglwidgetOutput("rglPlot")
  ),
  server = function(input, output) {
    output$rglPlot <- renderRglwidget({
      try(close3d(), silent = TRUE)
      if (input$rescale) aspect3d(1,1,10) else aspect3d(1,1,1)
      
      altText <- if (input$rescale) "rescaled" else "not rescaled"

      spheres3d(rnorm(100), rnorm(100), rnorm(100,sd = 0.1), col = "red",
          radius = 0.1)
      axes3d()
      rglwidget(altText = altText)
    })
  })
runApp(app)

if (!require("shiny"))
  stop("This demo requires shiny.")

library(rgl)

options(rgl.useNULL = TRUE)

ui <- fluidPage(
  mainPanel(
    tabsetPanel(
      tabPanel("red",
               rglwidgetOutput('thewidget1')),
      tabPanel("green",
               rglwidgetOutput('thewidget2'))
    ))
)

server <- function(input, output, session) {
  
  x <- rnorm(100)
  y <- 2*rnorm(100)
  z <- 10*rnorm(100)
  open3d()
  plot3d(x, y, z, col = "red")
  scene1 <- scene3d()
  plot3d(z, y, x, col = "green")
  scene2 <- scene3d()
  close3d()
  
  save <- options(rgl.inShiny = TRUE)
  on.exit(options(save))
  
  output$thewidget1 <- renderRglwidget(
    rglwidget(scene1)
  )
  
  output$thewidget2 <- renderRglwidget(
    rglwidget(scene2)
  )

}

if (interactive())
  shinyApp(ui = ui, server = server)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = FALSE,  # Bug in knitr 1.42 requires FALSE here
  comment = "#>",
  snapshot = FALSE,
  screenshot.force = FALSE)
library(rgl)
setupKnitr(autoprint = TRUE)

## -----------------------------------------------------------------------------

##########
### 3D HIST EXAMPLE:
##########

################################################################################

##### Required functions 'binplot' and 'hist3d':

binplot.3d<-function(x,y,z,alpha=1,topcol="#ff0000",sidecol="#aaaaaa") {
  save <- par3d(skipRedraw=TRUE)
  on.exit(par3d(save))
    
  x1<-c(rep(c(x[1],x[2],x[2],x[1]),3),rep(x[1],4),rep(x[2],4))
  z1<-c(rep(0,4),rep(c(0,0,z,z),4))
  y1<-c(y[1],y[1],y[2],y[2],rep(y[1],4),rep(y[2],4),rep(c(y[1],y[2],y[2],y[1]),2))
  x2<-c(rep(c(x[1],x[1],x[2],x[2]),2),rep(c(x[1],x[2],rep(x[1],3),rep(x[2],3)),2))
  z2<-c(rep(c(0,z),4),rep(0,8),rep(z,8) )
  y2<-c(rep(y[1],4),rep(y[2],4),rep(c(rep(y[1],3),rep(y[2],3),y[1],y[2]),2) )
  quads3d(x1,z1,y1,col=rep(sidecol,each=4),alpha=alpha)
  quads3d(c(x[1],x[2],x[2],x[1]),rep(z,4),c(y[1],y[1],y[2],y[2]),
              col=rep(topcol,each=4),alpha=1) 
  segments3d(x2,z2,y2,col="#000000")
}

hist3d<-function(x,y=NULL,nclass="auto",alpha=1,col="#ff0000",scale=10) {
  save <- par3d(skipRedraw=TRUE)
  on.exit(par3d(save))
  xy <- xy.coords(x,y)
  x <- xy$x
  y <- xy$y
  n<-length(x)
  if (nclass == "auto") nclass<-ceiling(sqrt(nclass.Sturges(x)))
  breaks.x <- seq(min(x),max(x),length=(nclass+1))
  breaks.y <- seq(min(y),max(y),length=(nclass+1))
  z<-matrix(0,(nclass),(nclass))
  for (i in seq_len(nclass)) {
    for (j in seq_len(nclass)) {
      z[i, j] <- (1/n)*sum(x < breaks.x[i+1] & y < breaks.y[j+1] & 
                            x >= breaks.x[i] & y >= breaks.y[j])
      binplot.3d(c(breaks.x[i],breaks.x[i+1]),c(breaks.y[j],breaks.y[j+1]),
                 scale*z[i,j],alpha=alpha,topcol=col)
    }
  }
}
################################################################################

open3d()

bg3d(color="gray")
light3d(0, 0)

# Drawing a 'bin' for given coordinates:
binplot.3d(c(-0.5,0.5),c(4.5,5.5),2,alpha=0.6)

# Setting the viewpoint ('theta' and 'phi' have the same meaning as in persp):
view3d(theta=40,phi=40)

##### QUADS FORMING BIN

open3d()

# Defining transparency and colors:
alpha<-0.7; topcol<-"#ff0000"; sidecol<-"#aaaaaa"

# Setting up coordinates for the quads and adding them to the scene:
y<-x<-c(-1,1) ; z<-4; of<-0.3
x12<-c(x[1],x[2],x[2],x[1]); x11<-rep(x[1],4); x22<-rep(x[2],4)
z00<-rep(0,4); z0z<-c(0,0,z,z); zzz<-rep(z,4); y11<-rep(y[1],4)
y1122<-c(y[1],y[1],y[2],y[2]); y12<-c(y[1],y[2],y[2],y[1]); y22<-rep(y[2],4)

quads3d(c(x12,x12,x11-of,x12,x22+of,x12),
          c(z00-of,rep(z0z,4),zzz+of),
          c(y1122,y11-of,y12,y22+of,y12,y1122),
          col=rep(c(rep(sidecol,5),topcol),each=4),alpha=c(rep(alpha,5),1))

# Setting up coordinates for the border-lines of the quads and drawing them:
yl1<-c(y[1],y[2],y[1],y[2]); yl2<-c(y[1]-of,y[1]-of)
xl<-c(rep(x[1],8),rep(x[1]-of,8),rep(c(x[1],x[2]),8),rep(x[2],8),rep(x[2]+of,8))
zl<-c(0,z,0,z,z+of,z+of,-of,-of,0,0,z,z,0,z,0,z,rep(0,4),rep(z,4),rep(-of,4),
      rep(z+of,4),z+of,z+of,-of,-of,rep(c(0,z),4),0,0,z,z)
yl<-c(yl2,y[2]+of,y[2]+of,rep(c(y[1],y[2]),4),y[1],y[1],y[2],y[2],yl2,
      rep(y[2]+of,4),yl2,y[2],y[2],rep(y[1],4),y[2],y[2],yl1,yl2,y[2]+of,
      y[2]+of,y[1],y[1],y[2],y[2],yl1)

lines3d(xl,zl,yl,col="#000000")
view3d(theta=40,phi=40)

##### COMPLETE HISTOGRAM:

open3d()

# Setting the rng to a fixed value:
set.seed(1000)

# Drawing a 3d histogramm of 2500 normaly distributed observations:
hist3d(rnorm(2500),rnorm(2500),alpha=0.4,nclass=7,scale=30)
# Choosing a lightgrey background:
bg3d(col="#cccccc")
view3d(theta=40,phi=40)

## -----------------------------------------------------------------------------
# rgl demo: rgl-bivar.r
# author: Daniel Adler

rgl.demo.bivar <- function() {
  if (!requireNamespace("MASS", quietly = TRUE))
    stop("This demo requires MASS")
  
  # parameters:
  n<-50; ngrid<-40
  
  # generate samples:
  set.seed(31415)
  x<-rnorm(n); y<-rnorm(n)
  
  # estimate non-parameteric density surface via kernel smoothing
  denobj <- MASS::kde2d(x, y, n=ngrid)
  den.z <-denobj$z
  
  # generate parametric density surface of a bivariate normal distribution
  xgrid <- denobj$x
  ygrid <- denobj$y
  bi.z <- dnorm(xgrid)%*%t(dnorm(ygrid))
  
  # visualize:
  zscale<-20
  
  # New window
  open3d()
  
  # clear scene:
  clear3d("all")
  
  # setup env:
  bg3d(color="#887777")
  light3d()
  
  # Draws the simulated data as spheres on the baseline
  spheres3d(x,y,rep(0,n),radius=0.1,color="#CCCCFF")
  
  # Draws non-parametric density
  surface3d(xgrid,ygrid,den.z*zscale,color="#FF2222",alpha=0.5)
  
  # Draws parametric density
  surface3d(xgrid,ygrid,bi.z*zscale,color="#CCCCFF",front="lines") 
}

rgl.demo.bivar()

## -----------------------------------------------------------------------------
# RGL-Demo: animal abundance
# Authors: Oleg Nenadic, Daniel Adler

rgl.demo.abundance <- function() {
  open3d()
  clear3d("all")               # remove all shapes, lights, bounding-box, and restore viewpoint
  
  # Setup environment:
  bg3d(col="#cccccc")     # setup background
  light3d()               # setup head-light
  
  # Importing animal data (created with wisp)
  terrain<-dget(system.file("demodata/region.dat",package="rgl"))
  pop<-dget(system.file("demodata/population.dat",package="rgl"))
  
  # Define colors for terrain
  zlim <- range(terrain)
  colorlut <- terrain.colors(82) 
  col1 <- colorlut[9*sqrt(3.6*(terrain-zlim[1])+2)]
  
  # Set color to (water-)blue for regions with zero 'altitude' 
  col1[terrain==0]<-"#0000FF"
  
  # Add terrain surface shape (i.e. population density):
  surface3d( 
      1:100,seq(1,60,length=100),terrain,
      col=col1,spec="#000000", ambient="#333333", back="lines"
  )
  
  # Define colors for simulated populations (males:blue, females:red):
  col2<-pop[,4]
  col2[col2==0]<-"#3333ff"
  col2[col2==1]<-"#ff3333"
  
  # Add simulated populations as sphere-set shape
  spheres3d(
    pop[,1],
    pop[,2],
    terrain[cbind( ceiling(pop[,1]),ceiling(pop[,2]*10/6) )]+0.5,
    radius=0.2*pop[,3], col=col2, alpha=(1-(pop[,5])/10 )
  )

}
rgl.demo.abundance()

## -----------------------------------------------------------------------------
# demo: lsystem.r
# author: Daniel Adler

#
# geometry 
#

deg2rad <- function( degree ) {
  return( degree*pi/180 )
}

rotZ.m3x3 <- function( degree ) {
  kc <- cos(deg2rad(degree))
  ks <- sin(deg2rad(degree))
  return( 
    matrix(
      c(
        kc, -ks,   0, 
        ks,  kc,   0,
         0,   0,   1
      ),ncol=3,byrow=TRUE
    ) 
  )
}

rotX.m3x3 <- function( degree ) {
  kc <- cos(deg2rad(degree))
  ks <- sin(deg2rad(degree))
  return(
    matrix(
      c(
         1,   0,   0,
         0,  kc, -ks,
         0,  ks,  kc
      ),ncol=3,byrow=TRUE
    )
  )
}

rotY.m3x3 <- function( degree ) {
  kc <- cos(deg2rad(degree))
  ks <- sin(deg2rad(degree))
  return(
    matrix(
      c(
        kc,   0,   ks,
         0,   1,    0,
       -ks,   0,   kc
      ),ncol=3,byrow=TRUE
    )
  )
}

rotZ <- function( v, degree ) {
  return( rotZ.m3x3(degree) %*% v)
}

rotX <- function( v, degree ) {
  return( rotX.m3x3(degree) %*% v)
}

rotY <- function( v, degree ) {
  return( rotY.m3x3(degree) %*% v)
}


#
# turtle graphics, rgl implementation:
#

turtle.init <- function(pos=c(0,0,0),head=0,pitch=90,roll=0,level=0) {
  clear3d("all")
  bg3d(color="gray")
  light3d()
  return( list(pos=pos,head=head,pitch=pitch,roll=roll,level=level) )
}


turtle.move <- function(turtle, steps, color) {
  
  rm <- rotX.m3x3(turtle$pitch) %*% rotY.m3x3(turtle$head) %*% rotZ.m3x3(turtle$roll)
  
  from <- as.vector( turtle$pos )  
  dir  <- rm %*% c(0,0,-1)
  to   <- from + dir * steps
    
  x <- c( from[1], to[1] )
  y <- c( from[2], to[2] )
  z <- c( from[3], to[3] )
  lines3d(x,y,z,col=color,size=1.5,alpha=0.5)
  turtle$pos <- to
  return(turtle)
}

turtle.pitch <- function(turtle, degree) {
  turtle$pitch <- turtle$pitch + degree
  return(turtle)
}

turtle.head <- function(turtle, degree) {
  turtle$head <- turtle$head + degree
  return(turtle)
}

turtle.roll <- function(turtle, degree) {
  turtle$roll <- turtle$roll + degree
  return(turtle)
}

#
# l-system general
#


lsystem.code <- function( x )
  substitute( x )

lsystem.gen <- function( x, grammar, levels=0 ) {
  code <- eval( substitute( substitute( REPLACE , grammar ), list(REPLACE=x) ) )
  if (levels)
    return( lsystem.gen( code , grammar , levels-1 ) )
  else
    return( code )
}

#
# l-system plot
#

lsystem.plot <- function( expr, level ) {
  turtle <- turtle.init(level=level)
  lsystem.eval( expr, turtle )
}

lsystem.eval <- function( expr, turtle ) {
  if ( length(expr) == 3 ) {
    turtle <- lsystem.eval( expr[[2]], turtle )
    turtle <- lsystem.eval( expr[[3]], turtle )
    turtle <- lsystem.eval( expr[[1]], turtle )
  } else if ( length(expr) == 2 ) {
    saved <- turtle
    turtle <- lsystem.eval( expr[[1]], turtle )
    turtle <- lsystem.eval( expr[[2]], turtle )
    turtle <- saved
  } else if ( length(expr) == 1 ) {
    if ( as.name(expr) == "stem" )      turtle <- turtle.move(turtle, 5, "brown")
    else if ( as.name(expr) == "short") turtle <- turtle.move(turtle, 5, "brown")
    else if ( as.name(expr) == "leaf" ) {
      spheres3d(turtle$pos[1],turtle$pos[2],turtle$pos[3],radius=0.1+turtle$level*0.3,color="green")
      sprites3d(turtle$pos[1],turtle$pos[2],turtle$pos[3],radius=0.5+turtle$level*0.3 ,color="green",texture=system.file("textures/particle.png",package="rgl"),textype="alpha",alpha=0.5)     
    }
    else if ( as.name(expr) == "roll" ) turtle <- turtle.head(turtle, 60)
    else if ( as.name(expr) == "down" ) turtle <- turtle.pitch(turtle,10)
    else if ( as.name(expr) == "up" )   turtle <- turtle.pitch(turtle,-10)
    else if ( as.name(expr) == "left" ) turtle <- turtle.head(turtle, 1)
    else if ( as.name(expr) == "right") turtle <- turtle.head(turtle,-1.5)
    else if ( as.name(expr) == "turnleft") turtle <- turtle.head(turtle,20)
    else if ( as.name(expr) == "turnright") turtle <- turtle.head(turtle,-20)
    else if ( as.name(expr) == "turn") turtle <- turtle.roll(turtle,180)
  }
  return(turtle)
}


#
# example
#

simple <- function(level=0) {
  grammar <- list(
    stem=lsystem.code(
      stem-(up-stem-leaf)-stem-(down-stem-leaf)-stem-leaf
    )
  )
  plant <- lsystem.gen(lsystem.code(stem), grammar, level )
  lsystem.plot(plant,level)
}

rgl.demo.lsystem <- function(level=0) {
  gen   <- list(
    stem=lsystem.code( 
      stem-left-stem-branch( turnleft-down-short-turnleft-down-stem-leaf)-right-right-stem--branch( turnright-up-short-turnright-up-short-turnright-short-stem-leaf)-left-left-left-stem-branch( turnleft-down-short-turnright-down-stem-leaf )-branch( up-turnright-short-up-turnleft-up-stem-leaf ) 
    )
  )    
  plant <- lsystem.gen(lsystem.code(stem), gen, level )
  lsystem.plot(plant,level)  
}

open3d()
rgl.demo.lsystem(level=1)
rglwidget()

## -----------------------------------------------------------------------------
# RGL-demo: subdivision surfaces
# author: Daniel Adler

rgl.demo.subdivision <- function() {
  # setup environment
  clear3d("all")
  view3d()
  bg3d(color="gray")
  light3d()

  # generate basic mesh
  obj <- oh3d()

  part <- function( level, tx, ... ) {
    shade3d( translate3d( obj, tx, 0, 0 )
    , color="gray30", front="lines",alpha=0.5,back="lines", override=TRUE
    )
    shade3d( translate3d( subdivision3d( obj, depth=level ), tx, 0, 0 )
    , override=TRUE, ... )
  }
  
  part(0, -5.50, color="blue"   )
  part(1, -1.75, color="yellow" )
  part(2,  1.75, color="red"    )
  part(3,  5.50, color="green"  )

}

open3d()
rgl.demo.subdivision()

## -----------------------------------------------------------------------------
# RGL-Demo: environment mapping
# Author: Daniel Adler

rgl.demo.envmap <- function() {
  open3d()
  # Clear scene:
  clear3d("all")
  light3d()
  bg3d(sphere=TRUE, color="white", back="filled"
  , texture=system.file("textures/refmap.png",package="rgl")
  )
  data(volcano)

  surface3d( 10*seq_len(nrow(volcano)),10*seq_len(ncol(volcano)),5*volcano
  , texture=system.file("textures/refmap.png",package="rgl")
  , texenvmap=TRUE
  , texmode="modulate"
  , color = "white"
  )
}
rgl.demo.envmap()

## -----------------------------------------------------------------------------

cone3d <- function(base=c(0,0,0),tip=c(0,0,1),rad=1,n=30,draw.base=TRUE,qmesh=FALSE,
                   trans = par3d("userMatrix"), ...) {
  ax <- tip-base
  if (missing(trans) && !cur3d()) trans <- diag(4)
  ### is there a better way?
  if (ax[1]!=0) {
    p1 <- c(-ax[2]/ax[1],1,0)
    p1 <- p1/sqrt(sum(p1^2))
    if (p1[1]!=0) {
      p2 <- c(-p1[2]/p1[1],1,0)
      p2[3] <- -sum(p2*ax)
      p2 <- p2/sqrt(sum(p2^2))
    } else {
      p2 <- c(0,0,1)
    }
  } else if (ax[2]!=0) {
    p1 <- c(0,-ax[3]/ax[2],1)
    p1 <- p1/sqrt(sum(p1^2))
    if (p1[1]!=0) {
      p2 <- c(0,-p1[3]/p1[2],1)
      p2[3] <- -sum(p2*ax)
      p2 <- p2/sqrt(sum(p2^2))
    } else {
      p2 <- c(1,0,0)
    }
  } else {
    p1 <- c(0,1,0); p2 <- c(1,0,0)
  }
  degvec <- seq(0,2*pi,length=n+1)[-1]
  ecoord2 <- function(theta) {
    base+rad*(cos(theta)*p1+sin(theta)*p2)
  }
  i <- rbind(1:n,c(2:n,1),rep(n+1,n))
  v <- cbind(sapply(degvec,ecoord2),tip)
  if (qmesh) 
    ## minor kluge for quads -- draw tip twice
    i <- rbind(i,rep(n+1,n))
  if (draw.base) {
    v <- cbind(v,base)
    i.x <- rbind(c(2:n,1),1:n,rep(n+2,n))
    if (qmesh)  ## add base twice
      i.x <-  rbind(i.x,rep(n+2,n))
    i <- cbind(i,i.x)
  }
  if (qmesh) v <- rbind(v,rep(1,ncol(v))) ## homogeneous
  if (!qmesh)
    triangles3d(v[1,i],v[2,i],v[3,i],...)
  else
    return(rotate3d(qmesh3d(v,i,material=list(...)), matrix=trans))
}     


ellipsoid3d <- function(rx=1,ry=1,rz=1,n=30,ctr=c(0,0,0),
                        qmesh=FALSE,
                        trans = par3d("userMatrix"),...) {
  if (missing(trans) && !cur3d()) trans <- diag(4)
  degvec <- seq(0,pi,length=n)
  ecoord2 <- function(p)
    c(rx*cos(p[1])*sin(p[2]),ry*sin(p[1])*sin(p[2]),rz*cos(p[2]))
  v <- apply(expand.grid(2*degvec,degvec),1,ecoord2)
  if (qmesh) v <- rbind(v,rep(1,ncol(v))) ## homogeneous
  e <- expand.grid(1:(n-1),1:n)
  i1 <- apply(e,1,function(z)z[1]+n*(z[2]-1))
  i2 <- i1+1
  i3 <- (i1+n-1) %% n^2 + 1
  i4 <- (i2+n-1) %% n^2 + 1
  i <- rbind(i1,i2,i4,i3)
  if (!qmesh)
    quads3d(v[1,i],v[2,i],v[3,i],...)
  else return(rotate3d(qmesh3d(v,i,material=list(...)),matrix=trans))
}

############
open3d()
ellipsoid3d(ctr=c(2,2,2),rx=3,ry=2,col="red",alpha=0.4)
cone3d(base=c(-2,-2,-2),rad=0.5,tip=c(-3,0,-4),col="blue",front="lines",back="lines")
shade3d(translate3d(cube3d(),3,-2,3,col="purple"))

### now with qmesh()
open3d()
q1 <- cone3d(qmesh=TRUE,trans=diag(4))  ## the "unit cone";
                                        ## height=1,radius=1, base at (0,0,0)
shade3d(q1)
## various transformations and rotations
wire3d(translate3d(q1,3,0,0),col="green")
wire3d(translate3d(scale3d(q1,1,1,2),6,0,0),col="green")
dot3d(translate3d(q1,0,3,0),col="green")
dot3d(translate3d(scale3d(q1,2,1,1),0,6,0),col="green")
shade3d(translate3d(q1,0,0,3),col="red")
shade3d(translate3d(rotate3d(scale3d(q1,1,1,2),pi/4,0,1,0),0,0,6),col="red")

axes3d()
open3d()

s1 <- ellipsoid3d(qmesh=TRUE,trans=diag(4))  ## the "unit sphere";
                                        ## radius=1, ctr at (0,0,0)
shade3d(s1)
## various transformations and rotations
wire3d(translate3d(s1,3,0,0),col="green")
wire3d(translate3d(scale3d(s1,1,1,2),6,0,0),col="green")
dot3d(translate3d(s1,0,3,0),col="green")
dot3d(translate3d(scale3d(s1,2,1,1),0,6,0),col="green")
shade3d(translate3d(s1,0,0,3),col="red")
shade3d(translate3d(rotate3d(scale3d(s1,1,1,2),pi/4,0,1,0),0,0,6),col="red")

axes3d()

## -----------------------------------------------------------------------------

cone3d <- function(base,tip,rad,n=30,...) {
  degvec <- seq(0,2*pi,length=n)
  ax <- tip-base
  ## what do if ax[1]==0?
  if (ax[1]!=0) {
    p1 <- c(-ax[2]/ax[1],1,0)
    p1 <- p1/sqrt(sum(p1^2))
    if (p1[1]!=0) {
      p2 <- c(-p1[2]/p1[1],1,0)
      p2[3] <- -sum(p2*ax)
      p2 <- p2/sqrt(sum(p2^2))
    } else {
      p2 <- c(0,0,1)
    }
  } else if (ax[2]!=0) {
    p1 <- c(0,-ax[3]/ax[2],1)
    p1 <- p1/sqrt(sum(p1^2))
    if (p1[1]!=0) {
      p2 <- c(0,-p1[3]/p1[2],1)
      p2[3] <- -sum(p2*ax)
      p2 <- p2/sqrt(sum(p2^2))
    } else {
      p2 <- c(1,0,0)
    }
  } else {
    p1 <- c(0,1,0); p2 <- c(1,0,0)
  }
  ecoord2 <- function(theta) {
    base+rad*(cos(theta)*p1+sin(theta)*p2)
  }
    for (i in seq_len(n-1)) {
      li <- ecoord2(degvec[i])
      lj <- ecoord2(degvec[i+1])
      triangles3d(c(li[1],lj[1],tip[1]),c(li[2],lj[2],tip[2]),c(li[3],lj[3],tip[3]),...)
    }   
}

lollipop3d <- function(data.x,data.y,data.z,surf.fun,surf.n=50,
                         xlim=range(data.x),
			 ylim=range(data.y),
			 zlim=range(data.z),
                         asp=c(y=1,z=1),
			 xlab=deparse(substitute(x)),
			 ylab=deparse(substitute(y)),
			 zlab=deparse(substitute(z)),
			 alpha.surf=0.4,
                         col.surf=fg,col.stem=c(fg,fg),
                         col.pt="gray",type.surf="line",ptsize,
                         lwd.stem=2,lit=TRUE,bg="white",fg="black",
                         col.axes=fg,col.axlabs=fg,
                         axis.arrow=TRUE,axis.labels=TRUE,
                         box.col=bg,
                         axes=c("lines","box")) {
  axes <- match.arg(axes)
  col.stem <- rep(col.stem,length=2)
  x.ticks <- pretty(xlim)
  x.ticks <- x.ticks[x.ticks>=min(xlim) & x.ticks<=max(xlim)]
  x.ticklabs <- if (axis.labels) as.character(x.ticks) else NULL
  y.ticks <- pretty(ylim)
  y.ticks <- y.ticks[y.ticks>=min(ylim) & y.ticks<=max(ylim)]
  y.ticklabs <- if (axis.labels) as.character(y.ticks) else NULL
  z.ticks <- pretty(zlim)
  z.ticks <- z.ticks[z.ticks>=min(zlim) & z.ticks<=max(zlim)]
  z.ticklabs <- if (axis.labels) as.character(z.ticks) else NULL
  if (!missing(surf.fun)) {
    surf.x <- seq(xlim[1],xlim[2],length=surf.n)
    surf.y <- seq(ylim[1],ylim[2],length=surf.n)
    surf.z <- outer(surf.x,surf.y,surf.fun)  ## requires surf.fun be vectorized
    z.interc <- surf.fun(data.x,data.y)
    zdiff <- diff(range(c(surf.z,data.z)))
  } else {
    z.interc <- rep(min(data.z),length(data.x))
    zdiff <- diff(range(data.z))
  }
  xdiff <- diff(xlim)
  ydiff <- diff(ylim)
  y.adj <- if (asp[1]<=0) 1 else asp[1]*xdiff/ydiff
  data.y <- y.adj*data.y
  y.ticks <- y.adj*y.ticks
  ylim <- ylim*y.adj
  ydiff <- diff(ylim)
  z.adj <- if (asp[2]<=0) 1 else asp[2]*xdiff/zdiff
  data.z <- z.adj*data.z
  if (!missing(surf.fun)) {
    surf.y <- y.adj*surf.y
    surf.z <- z.adj*surf.z
  }
  z.interc <- z.adj*z.interc
  z.ticks <- z.adj*z.ticks
  zlim <- z.adj*zlim
  open3d()
  clear3d("all")
  light3d()
  bg3d(color=c(bg,fg))
  if (!missing(surf.fun)) 
    surface3d(surf.x,surf.y,surf.z,alpha=alpha.surf,
                front=type.surf,back=type.surf,
                col=col.surf,lit=lit)
  if (missing(ptsize)) ptsize <- 0.02*xdiff
  ## draw points
  spheres3d(data.x,data.y,data.z,r=ptsize,lit=lit,color=col.pt)
  ## draw lollipops
  apply(cbind(data.x,data.y,data.z,z.interc),1,
        function(X) {
          lines3d(x=rep(X[1],2),
                  y=rep(X[2],2),
                  z=c(X[3],X[4]),
                  col=ifelse(X[3]>X[4],col.stem[1],
                    col.stem[2]),lwd=lwd.stem)
        })
  if (axes=="box") {
    bbox3d(xat=x.ticks,xlab=x.ticklabs,
             yat=y.ticks,ylab=y.ticklabs,
             zat=z.ticks,zlab=z.ticklabs,lit=lit)
  } else if (axes=="lines") { ## set up axis lines
    axis3d(edge="x",at=x.ticks,labels=x.ticklabs,
           col=col.axes,arrow=axis.arrow)
    axis3d(edge="y",at=y.ticks,labels=y.ticklabs,
           col=col.axes,arrow=axis.arrow)
    axis3d(edge="z",at=z.ticks,labels=z.ticklabs,
           col=col.axes,arrow=axis.arrow)
    box3d(col=col.axes)
  }
  decorate3d(xlab=xlab, ylab=ylab, zlab=zlab, box=FALSE, axes=FALSE, col=col.axlabs)
}

x <- 1:5
y <- x*10
z <- (x+y)/20
open3d()
spheres3d(x,y,z)
axes3d()
set.seed(1001)
x <- runif(30)
y <- runif(30,max=2)
dfun <- function(x,y)  2*x+3*y+2*x*y+3*y^2 
z <- dfun(x,y)+rnorm(30,sd=0.5)
## lollipops only
lollipop3d(x,y,z)

## lollipops plus theoretical surface

open3d()
lollipop3d(x,y,z,dfun,col.pt="red",col.stem=c("red","blue"))

## lollipops plus regression fit

linmodel <- lm(z~x+y)
dfun <- function(x,y) predict(linmodel,newdata=data.frame(x=x,y=y))
open3d()
lollipop3d(x,y,z,dfun,col.pt="red",col.stem=c("red","blue"))

####

if (!require("shiny"))
  stop("This demo requires shiny.")

library(rgl)

open3d(useNULL = TRUE)
ids <- plot3d(rnorm(100), rnorm(100), rnorm(100))["data"]
scene <- scene3d()
close3d()

ui <- fluidPage(
  checkboxInput("chk", label = "Display", value = FALSE),
  playwidgetOutput("control"),
  rglwidgetOutput("wdg")
)

server <- function(input, output, session) {
  options(rgl.useNULL = TRUE)
  save <- options(rgl.inShiny = TRUE)
  on.exit(options(save))
  
  output$wdg <- renderRglwidget({
    rglwidget(scene, controllers = "control")
  })
  
  output$control <- renderPlaywidget({
    toggleWidget("wdg", respondTo = "chk",
                 ids = ids)
  })
}

if (interactive())
  shinyApp(ui = ui, server = server)

## ----setup, echo=FALSE, results="asis"----------------------------------------
source("setup.R")
setupKnitr(autoprint = TRUE)
set.seed(123)

## ----"plot3d()"---------------------------------------------------------------
with(iris, plot3d(Sepal.Length, Sepal.Width, Petal.Length, 
                  type="s", col=as.numeric(Species)))

## ----"persp3d()", fig.height=3, fig.width=6, fig.keep="last", eval=requireNamespace("MASS",quietly=TRUE)----
# This example requires the MASS package
library(MASS)
# from the fitdistr example
set.seed(123)
x <- rgamma(100, shape = 5, rate = 0.1)
fit <- fitdistr(x, dgamma, list(shape = 1, rate = 0.1), lower = 0.001)
loglik <- function(shape, rate) sum(dgamma(x, shape=shape, rate=rate, 
                                           log=TRUE))
loglik <- Vectorize(loglik)
xlim <- fit$estimate[1]+4*fit$sd[1]*c(-1,1)
ylim <- fit$estimate[2]+4*fit$sd[2]*c(-1,1)

mfrow3d(1, 2, sharedMouse = TRUE)
persp3d(loglik, 
        xlim = xlim, ylim = ylim,
        n = 30)
zlim <- fit$loglik + c(-qchisq(0.99, 2)/2, 0)
next3d()
persp3d(loglik, 
        xlim = xlim, ylim = ylim, zlim = zlim,
        n = 30)

## -----------------------------------------------------------------------------
methods(plot3d)
methods(persp3d)

## ----"triangles3d()", fig.width=3, fig.height=3-------------------------------
triangles3d(cbind(x=rnorm(9), y=rnorm(9), z=rnorm(9)), col = "green")
decorate3d()
bg3d("lightgray")
aspect3d(1,1,1)

## ----eval = FALSE-------------------------------------------------------------
# myview <- par3d("userMatrix")
# # ... later ...
# par3d(userMatrix = myview)

## ----Texture------------------------------------------------------------------
filename <- tempfile(fileext = ".png")
png(filename = filename)
plot(rnorm(1000), rnorm(1000))
safe.dev.off()
open3d()
xyz <- cbind(c(0,1,1,0), 0, c(0,0,1,1))
quads3d(xyz, texture = filename, texcoords = xyz[,c(1, 3)]*2, 
        col = "white", specular = "black")

## ----Solids-------------------------------------------------------------------
cols <- rainbow(7)
layout3d(matrix(1:16, 4,4), heights=c(1,3,1,3))
text3d(0,0,0,"tetrahedron3d"); next3d()
shade3d(tetrahedron3d(col=cols[1])); next3d()
text3d(0,0,0,"cube3d"); next3d()
shade3d(cube3d(col=cols[2])); next3d()
text3d(0,0,0,"octahedron3d"); next3d()
shade3d(octahedron3d(col=cols[3])); next3d()
text3d(0,0,0,"dodecahedron3d"); next3d()
shade3d(dodecahedron3d(col=cols[4])); next3d()
text3d(0,0,0,"icosahedron3d"); next3d()
shade3d(icosahedron3d(col=cols[5])); next3d()
text3d(0,0,0,"cuboctahedron3d"); next3d()
shade3d(cuboctahedron3d(col=cols[6])); next3d()
text3d(0,0,0,"oh3d"); next3d()
shade3d(oh3d(col=cols[7]))

## ----Autoprint----------------------------------------------------------------
xyz <- matrix(rnorm(27), ncol = 3)
triangles3d(xyz, col = rainbow(9))
spheres3d(xyz, col = rainbow(9), radius = 0.1)

## ----eval = FALSE-------------------------------------------------------------
# plots <- NULL
# for (i in 1:3) {
#   plot3d(rnorm(10), rnorm(10), rnorm(10))
#   plots <- htmltools::tagList(plots, rglwidget())
#   close3d()
# }
# plots

## ----eval = FALSE-------------------------------------------------------------
# foreignHigh()   # Produces a high level plot, but doesn't return
#                 # an appropriate value
# highlevel()
# foreignLow()    # Modifies the previous plot
# lowlevel()

## -----------------------------------------------------------------------------
par3d("mouseMode")

## ----fig.alt="Volcano in rgl", echo = 2:3-------------------------------------
close3d()
persp3d(volcano, col = "green")

## ----eval=requireNamespace("orientlib", quietly = TRUE) && requireNamespace("lattice", quietly = TRUE), fig.alt=paste0("Volcano in ", c("lattice", "base"), "graphics.")----
# Only evaluated if the lattice & orientlib packages are installed
lattice::wireframe(volcano, col = "green", 
		   screen = rglToLattice())
angles <- rglToBase()
persp(volcano, col = "green", shade = TRUE,
      theta = angles$theta, phi = angles$phi)

## ----echo=FALSE---------------------------------------------------------------
setdiff(ls("package:rgl"), c(documentedfns, deprecatedfns))

## ----echo=FALSE, results="asis"-----------------------------------------------
writeIndex(cols = if (knitr::is_html_output()) 5 else 4)

randomDot <- function(left, right, rightOffset=c(200, 0), n=3000, ...) {
  old <- cur3d()
  on.exit(set3d(old))  
  force(left)
  force(right)
  set3d(left)
  leftViewport <- par3d("viewport")
  leftSize <- leftViewport[3:4]
  leftProj <- rgl.projection()
  leftDepth <- rgl.pixels("depth")  
  leftUR <- leftViewport[1:2] + leftSize - 1
  set3d(right)
  rightViewport <- par3d("viewport")
  rightSize <- rightViewport[3:4]
  rightProj <- rgl.projection()
  rightDepth <- rgl.pixels("depth")
  rightUR <- rightViewport[1:2] + rightSize - 1

  size <- pmax(leftViewport[3:4], rightViewport[3:4]+rightOffset)
  
  pts <- matrix(c(sample(leftSize[1], n, replace=TRUE),
                  sample(leftSize[2], n, replace=TRUE)), n, 2)
  cols <- 1:n
  startpt <- pts
  startcols <- cols
  keep <- startpt[,1] < leftSize[1] & startpt[,2] < leftSize[2]
  pt <- startpt[keep,,drop=FALSE]
  cl <- startcols[keep]
  while (length(pt)) {
    depth <- leftDepth[pt]

    user <- rgl.window2user((pt[,1]-0.5)/leftSize[1], (pt[,2]-0.5)/leftSize[2],
                          depth, projection=leftProj)
    win <- rgl.user2window(user, projection=rightProj)
    bkgd <- cbind((pt[,1] - 0.5)/rightSize[1], (pt[,2] - 0.5)/rightSize[2], 1)
    usewin <- rep(depth < 1, 3)
    rightPt <- structure(ifelse(usewin, win, bkgd), dim=dim(win))
    rightPti <- round(cbind(rightSize[1]*rightPt[,1], rightSize[2]*rightPt[,2]) + 0.5)
    
    keep <- rightPti[,1] >= 1 & rightPti[,1] <= rightUR[1] &
            rightPti[,2] >= 1 & rightPti[,2] <= rightUR[2]
            
    rightPti <- rightPti[keep,,drop=FALSE]
    rightPt <- rightPt[keep,,drop=FALSE]
    cl <- cl[keep]            
    keep <- TRUE | rightPt[,3] <= rightDepth[ rightPti ]+0.001
    rightPti <- rightPti[keep,,drop=FALSE]            
    cl <- cl[keep]

    pt <- cbind(rightPti[,1] + rightOffset[1], rightPti[,2] + rightOffset[2])
    pts <- rbind(pts, pt)
    cols <- c(cols, cl)
    keep <- apply(pt, 1, min) >= 1 &
            pt[,1] <= leftUR[1] &
            pt[,2] <= leftUR[2] 
    pt <- pt[keep,,drop=FALSE]
    cl <- cl[keep]
  } 
  
  pt <- cbind(startpt[,1] - rightOffset[1], startpt[,2] - rightOffset[2])
  keep <- pt[,1] > 1 & pt[,1] < rightSize[1] & 
          pt[,2] > 1 & pt[,2] < rightSize[2] 
  pt <- pt[ keep,,drop=FALSE ]
  cl <- startcols[ keep ]
  
  while (length(pt)) {
    depth <- rightDepth[pt]
    user <- rgl.window2user((pt[,1]-0.5)/rightSize[1], (pt[,2]-0.5)/rightSize[2],
                          depth, projection=rightProj)
    win <- rgl.user2window(user, projection=leftProj)
    bkgd <- cbind((pt[,1] - 0.5)/leftSize[1], (pt[,2] - 0.5)/leftSize[2], 1)
    usewin <- rep(depth < 1, 3)
    leftPt <- structure(ifelse(usewin, win, bkgd), dim=dim(win))
    leftPti <- round(cbind(leftSize[1]*leftPt[,1], leftSize[2]*leftPt[,2]) + 0.5)

    keep <- leftPti[,1] >= 1 & leftPti[,1] <= leftUR[1] &
            leftPti[,2] >= 1 & leftPti[,2] <= leftUR[2]
            
    leftPti <- leftPti[keep,,drop=FALSE]
    leftPt <- leftPt[keep,,drop=FALSE]
    cl <- cl[keep]            
    keep <- TRUE | leftPt[,3] <= leftDepth[ leftPti ]+0.001
    leftPti <- leftPti[keep,,drop=FALSE]            
    cl <- cl[keep]
    
    pt <- leftPti
    pts <- rbind(pts, pt)
    cols <- c(cols, cl)
    pt <- cbind(pt[,1] - rightOffset[1], pt[,2] - rightOffset[2])
    keep <- apply(pt, 1, min) >= 1 &
            pt[,1] <= rightUR[1] &
            pt[,2] <= rightUR[2]
    pt <- pt[keep,,drop=FALSE]
    cl <- cl[keep] 
  }
  plot(pts, col = cols, pch=16, axes=FALSE,cex=0.25+cols/n/2,xlab="",ylab="",...)
  rug((size[1] + c(-1,1)*rightOffset[1])/2, side=1)
  rug((size[1] + c(-1,1)*rightOffset[1])/2, side=3)
  rug((size[2] + c(-1,1)*rightOffset[2])/2, side=2)
  rug((size[2] + c(-1,1)*rightOffset[2])/2, side=4)
}

                                                #red                   #cyan 
anaglyph <- function(left, right, leftColor = c(1,0,0), rightColor = c(0,1,1),
                     dimens = dim(leftPixels)) {
  old <- cur3d()
  on.exit(set3d(old))  
  force(left)
  force(right)

  set3d(left)
  vp <- par3d("viewport")
  leftPixels <- rgl.pixels(viewport=vp)
  leftPixels <- t((leftPixels[,,1]+leftPixels[,,2]+leftPixels[,,3])/3)
  leftPixels <- leftPixels[rev(seq_len(dimens[1])), seq_len(dimens[2])]
  set3d(right)
  rightPixels <- rgl.pixels(viewport=vp)
  rightPixels <- t((rightPixels[,,1]+rightPixels[,,2]+rightPixels[,,3])/3)
  rightPixels <- rightPixels[rev(seq_len(dimens[1])), seq_len(dimens[2])]
  red <- pmin(leftPixels*leftColor[1] + rightPixels*rightColor[1], 1)
  green <- pmin(leftPixels*leftColor[2] + rightPixels*rightColor[2], 1)
  blue <- pmin(leftPixels*leftColor[3] + rightPixels*rightColor[3], 1)
  z <- as.raster(array(c(red, green, blue), dim = c(dimens, 3)))
  if (length(z)) {
    par(mar = c(0,0,0,0))
    plot(z)
  } else {
    cat("Unable to read pixels:\nstr(leftPixels):\n")
    str(leftPixels)
    cat("str(rightPixels):\n")
    str(rightPixels)
  }
}

if (!rgl.useNULL()) {
  source(system.file("demo/mouseCallbacks.R", package="rgl"), echo=FALSE )

# This version assumes the eyes diverge for the stereo view.  
# Reverse the two arguments for the cross-eyed view.

  dev.new(width=9, height=7)
  randomDot(cur3d()-1, cur3d())

# A red-cyan anaglyph (for 3D glasses).  Use optional args to anaglyph for other glasses.

  dev.new()
  anaglyph(cur3d()-1, cur3d())
  
} else
  cat("Can't read pixels from a NULL device\n")

## ----setup, include=FALSE-----------------------------------------------------
if (!requireNamespace("rmarkdown", quietly = TRUE) ||
    !rmarkdown::pandoc_available("1.14")) {
  warning(call. = FALSE, "These vignettes assume rmarkdown and Pandoc
          version 1.14.  These were not found. Older versions will not work.")
  knitr::knit_exit()
}
knitr::opts_chunk$set(echo = TRUE, snapshot = FALSE, screenshot.force = FALSE)
library(rgl)
options(rgl.useNULL = TRUE)
setupKnitr(autoprint = TRUE)

M <- structure(c(0.997410774230957, 0.0707177817821503, -0.0130676832050085, 
0, -0.0703366547822952, 0.99714070558548, 0.02762770652771, 0, 
0.0149840852245688, -0.0266370177268982, 0.999532878398895, 0, 
0, 0, 0, 1), .Dim = c(4L, 4L))

## -----------------------------------------------------------------------------
theta <- 2*pi*c(0:2, 4:6, 8:10)/12
x <- cos(theta)
y <- sin(theta)
z <- rep(c(0,0,1), 3)
xyz <- cbind(x, y, z)
xyz <- xyz[c(1,2,6, 4,5,9, 7,8,3),]
open3d()
par3d(userMatrix = M)
triangles3d(xyz, col = rep(c("red", "green", "blue"), each = 3))

## ----fig.width=8--------------------------------------------------------------
open3d()
par3d(userMatrix = M)
layout3d(matrix(1:9, ncol = 3, byrow=TRUE),
         widths = c(1,2,2), heights = c(1, 3,3), 
         sharedMouse = TRUE)
text3d(0,0,0, " ")
next3d()
text3d(0,0,0, "depth_mask = TRUE")
next3d()
text3d(0,0,0, "depth_mask = FALSE")
next3d()
text3d(0,0,0, "alpha = 0.7")
next3d()
triangles3d(xyz, col = rep(c("red", "green", "blue"), each = 3), alpha = 0.7, depth_mask = TRUE)
next3d()
triangles3d(xyz, col = rep(c("red", "green", "blue"), each = 3), alpha = 0.7, depth_mask = FALSE)
next3d()
text3d(0,0,0, "alpha = 0.3")
next3d()
triangles3d(xyz, col = rep(c("red", "green", "blue"), each = 3), alpha = 0.3, depth_mask = TRUE)
next3d()
triangles3d(xyz, col = rep(c("red", "green", "blue"), each = 3), alpha = 0.3, depth_mask = FALSE)

## -----------------------------------------------------------------------------
open3d()
par3d(userMatrix = M)
triangles3d(xyz[1:3,], col = "red", alpha = 0.3, depth_mask = FALSE)
triangles3d(xyz[4:6,], col = "green", alpha = 0.7, depth_mask = TRUE)
triangles3d(xyz[7:9,], col = "blue", depth_mask = TRUE)

## ----fig.width=8--------------------------------------------------------------
open3d()
par3d(userMatrix = M)
layout3d(matrix(1:9, ncol = 3, byrow=TRUE),
         widths = c(1,2,2), heights = c(1, 3,3), 
         sharedMouse = TRUE)
text3d(0,0,0, " ")
next3d()
text3d(0,0,0, "depth_mask = TRUE")
next3d()
text3d(0,0,0, "depth_mask = FALSE")
next3d()
text3d(0,0,0, "alpha = 0.7")
next3d()
triangles3d(xyz, col = "red", alpha = 0.7, depth_mask = TRUE)
next3d()
triangles3d(xyz, col = "red", alpha = 0.7, depth_mask = FALSE)
next3d()
text3d(0,0,0, "alpha = 0.3")
next3d()
triangles3d(xyz, col = "red", alpha = 0.3, depth_mask = TRUE)
next3d()
triangles3d(xyz, col = "red", alpha = 0.3, depth_mask = FALSE)

## ----setup, include=FALSE-----------------------------------------------------
if (!requireNamespace("rmarkdown", quietly = TRUE) ||
    !rmarkdown::pandoc_available("1.14")) {
  warning(call. = FALSE, "These vignettes assume rmarkdown and Pandoc
          version 1.14.  These were not found. Older versions will not work.")
  knitr::knit_exit()
}
knitr::opts_chunk$set(echo = TRUE, snapshot = FALSE, screenshot.force = FALSE)
library(rgl)
options(rgl.useNULL = TRUE)
setupKnitr(autoprint = TRUE)

## ----eval=FALSE---------------------------------------------------------------
# setupKnitr(autoprint = TRUE)

## -----------------------------------------------------------------------------
# Show regression plane with z as dependent variable
library(rgl)
open3d()
x <- rnorm(100)
y <- rnorm(100)
z <- 0.2*x - 0.3*y + rnorm(100, sd = 0.3)
fit <- lm(z ~ x + y)

plot3d(x, y, z, type = "s", col = "red", size = 1)

# No plot here, because of the planes3d() call below

coefs <- coef(fit)
a <- coefs["x"]
b <- coefs["y"]
c <- -1
d <- coefs["(Intercept)"]
planes3d(a, b, c, d, alpha = 0.5)

##########
### 3D HIST EXAMPLE:
##########

################################################################################

##### Required functions 'binplot' and 'hist3d':

binplot.3d<-function(x,y,z,alpha=1,topcol="#ff0000",sidecol="#aaaaaa") {
  save <- par3d(skipRedraw=TRUE)
  on.exit(par3d(save))
    
  x1<-c(rep(c(x[1],x[2],x[2],x[1]),3),rep(x[1],4),rep(x[2],4))
  z1<-c(rep(0,4),rep(c(0,0,z,z),4))
  y1<-c(y[1],y[1],y[2],y[2],rep(y[1],4),rep(y[2],4),rep(c(y[1],y[2],y[2],y[1]),2))
  x2<-c(rep(c(x[1],x[1],x[2],x[2]),2),rep(c(x[1],x[2],rep(x[1],3),rep(x[2],3)),2))
  z2<-c(rep(c(0,z),4),rep(0,8),rep(z,8) )
  y2<-c(rep(y[1],4),rep(y[2],4),rep(c(rep(y[1],3),rep(y[2],3),y[1],y[2]),2) )
  rgl.quads(x1,z1,y1,col=rep(sidecol,each=4),alpha=alpha)
  rgl.quads(c(x[1],x[2],x[2],x[1]),rep(z,4),c(y[1],y[1],y[2],y[2]),
              col=rep(topcol,each=4),alpha=1) 
  rgl.lines(x2,z2,y2,col="#000000")
}

hist3d<-function(x,y=NULL,nclass="auto",alpha=1,col="#ff0000",scale=10) {
  save <- par3d(skipRedraw=TRUE)
  on.exit(par3d(save))
  xy <- xy.coords(x,y)
  x <- xy$x
  y <- xy$y
  n<-length(x)
  if (nclass == "auto") nclass<-ceiling(sqrt(nclass.Sturges(x)))
  breaks.x <- seq(min(x),max(x),length=(nclass+1))
  breaks.y <- seq(min(y),max(y),length=(nclass+1))
  z<-matrix(0,(nclass),(nclass))
  for (i in seq_len(nclass)) {
    for (j in seq_len(nclass)) {
      z[i, j] <- (1/n)*sum(x < breaks.x[i+1] & y < breaks.y[j+1] & 
                            x >= breaks.x[i] & y >= breaks.y[j])
      binplot.3d(c(breaks.x[i],breaks.x[i+1]),c(breaks.y[j],breaks.y[j+1]),
                 scale*z[i,j],alpha=alpha,topcol=col)
    }
  }
}
################################################################################

rgl.open()

rgl.bg(color="gray")
rgl.light()

# Drawing a 'bin' for given coordinates:
binplot.3d(c(-0.5,0.5),c(4.5,5.5),2,alpha=0.6)

# Setting the viewpoint ('theta' and 'phi' have the same meaning as in persp):
rgl.viewpoint(theta=40,phi=40)

# Choosing a lightgrey background:
rgl.bg(col="#cccccc")


##### QUADS FORMING BIN:

rgl.open()

# Defining transparency and colors:
alpha<-0.7; topcol<-"#ff0000"; sidecol<-"#aaaaaa"

# Setting up coordinates for the quads and adding them to the scene:
y<-x<-c(-1,1) ; z<-4; of<-0.3
x12<-c(x[1],x[2],x[2],x[1]); x11<-rep(x[1],4); x22<-rep(x[2],4)
z00<-rep(0,4); z0z<-c(0,0,z,z); zzz<-rep(z,4); y11<-rep(y[1],4)
y1122<-c(y[1],y[1],y[2],y[2]); y12<-c(y[1],y[2],y[2],y[1]); y22<-rep(y[2],4)

rgl.quads(c(x12,x12,x11-of,x12,x22+of,x12),
          c(z00-of,rep(z0z,4),zzz+of),
          c(y1122,y11-of,y12,y22+of,y12,y1122),
          col=rep(c(rep(sidecol,5),topcol),each=4),alpha=c(rep(alpha,5),1))

# Setting up coordinates for the border-lines of the quads and drawing them:
yl1<-c(y[1],y[2],y[1],y[2]); yl2<-c(y[1]-of,y[1]-of)
xl<-c(rep(x[1],8),rep(x[1]-of,8),rep(c(x[1],x[2]),8),rep(x[2],8),rep(x[2]+of,8))
zl<-c(0,z,0,z,z+of,z+of,-of,-of,0,0,z,z,0,z,0,z,rep(0,4),rep(z,4),rep(-of,4),
      rep(z+of,4),z+of,z+of,-of,-of,rep(c(0,z),4),0,0,z,z)
yl<-c(yl2,y[2]+of,y[2]+of,rep(c(y[1],y[2]),4),y[1],y[1],y[2],y[2],yl2,
      rep(y[2]+of,4),yl2,y[2],y[2],rep(y[1],4),y[2],y[2],yl1,yl2,y[2]+of,
      y[2]+of,y[1],y[1],y[2],y[2],yl1)

rgl.lines(xl,zl,yl,col="#000000")


##### COMPLETE HISTOGRAM:

rgl.open()

# Choosing a lightgrey background:
rgl.bg(col="#cccccc")

# Setting the rng to a fixed value:
set.seed(1000)

# Drawing a 3d histogramm of 2500 normaly distributed observations:
hist3d(rnorm(2500),rnorm(2500),alpha=0.4,nclass=7,scale=30)

# RGL-Demo: environment mapping
# Author: Daniel Adler

rgl.demo.envmap <- function() {
  open3d()
  # Clear scene:
  clear3d("all")
  light3d()
  bg3d(sphere=TRUE, color="white", back="filled"
  , texture=system.file("textures/refmap.png",package="rgl")
  )
  data(volcano)

  surface3d( 10*seq_len(nrow(volcano)),10*seq_len(ncol(volcano)),5*volcano
  , texture=system.file("textures/refmap.png",package="rgl")
  , texenvmap=TRUE
  , color = "white"
  )
}
rgl.demo.envmap()

# RGL-Demo: animal abundance
# Authors: Oleg Nenadic, Daniel Adler

rgl.demo.abundance <- function() {
  open3d()
  clear3d("all")               # remove all shapes, lights, bounding-box, and restore viewpoint
  
  # Setup environment:
  bg3d(col="#cccccc")     # setup background
  light3d()               # setup head-light
  
  # Importing animal data (created with wisp)
  terrain<-dget(system.file("demodata/region.dat",package="rgl"))
  pop<-dget(system.file("demodata/population.dat",package="rgl"))
  
  # Define colors for terrain
  zlim <- range(terrain)
  colorlut <- terrain.colors(82) 
  col1 <- colorlut[9*sqrt(3.6*(terrain-zlim[1])+2)]
  
  # Set color to (water-)blue for regions with zero 'altitude' 
  col1[terrain==0]<-"#0000FF"
  
  # Add terrain surface shape (i.e. population density):
  surface3d( 
      1:100,seq(1,60,length=100),terrain,
      col=col1,spec="#000000", ambient="#333333", back="lines"
  )
  
  # Define colors for simulated populations (males:blue, females:red):
  col2<-pop[,4]
  col2[col2==0]<-"#3333ff"
  col2[col2==1]<-"#ff3333"
  
  # Add simulated populations as sphere-set shape
  spheres3d(
    pop[,1],
    pop[,2],
    terrain[cbind( ceiling(pop[,1]),ceiling(pop[,2]*10/6) )]+0.5,
    radius=0.2*pop[,3], col=col2, alpha=(1-(pop[,5])/10 )
  )

}
rgl.demo.abundance()

addNormals <- function(x, ...) UseMethod("addNormals")
  
addNormals.mesh3d <- function(x, angleWeighted = TRUE, ...) {
  v <- x$vb
  
  # Make sure v is homogeneous with unit w
  if (nrow(v) == 3) v <- rbind(v, 1)
  else v <- t( t(v)/v[4,] )
  
  v <- v[1:3,]
  normals <- v*0
  
  if (is.na(angleWeighted)) {
    reproduceBug <- TRUE
    angleWeighted <- FALSE
  } else
    reproduceBug <- FALSE
  
  dopolys <- function(it, normals) {
    n <- nrow(it)
    for (i in seq_len(ncol(it))) {
      normal <- xprod( v[, it[1, i]] - v[, it[3, i]], 
                       v[, it[2, i]] - v[, it[1, i]])
      if (reproduceBug)
        normal <- normalize(normal)
      if (!any(is.na(normal))) {
        if (angleWeighted) 
          normal <- normalize(normal)
        
        for (j in seq_len(n)) {
          if (angleWeighted) {
            jm1 <- (j + n - 2) %% n + 1
            jp1 <- j %% n + 1
            weight <- angle(v[, it[jm1, i]] - v[, it[j, i]],
                           v[, it[jp1, i]] - v[, it[j, i]])
          } else
            weight <- 1
          normals[, it[j,i]] <- normals[, it[j,i]] + normal*weight
        }
      }
    }
    normals
  }
  
  if (length(x$it)) 
    normals <- dopolys(x$it, normals)
  
  if (length(x$ib)) 
    normals <- dopolys(x$ib, normals)
  normals <- rbind(apply(normals, 2, function(n) n/veclen(n)), 1)
  x$normals <- normals
  x
}

veclen <- function(v) sqrt(sum(v^2))

normalize <- function(v) v/veclen(v)

xprod <- function(v, w) c( v[2]*w[3] - v[3]*w[2],
                           v[3]*w[1] - v[1]*w[3],
                           v[1]*w[2] - v[2]*w[1] )

angle <- function(a,b) {
  dot <- sum(a*b)
  acos(pmin(1, pmax(-1, dot/veclen(a)/veclen(b))))
}

arc3d <- function(from, to, center, radius, n, circle = 50, base = 0, plot = TRUE, ...) {
  fixarg <- function(arg) {
    if (is.matrix(arg))
      arg[, 1:3, drop = FALSE]
    else
      matrix(arg, 1, 3)
  }
  normalize <- function(v)
    v / veclen(v)
  getrow <- function(arg, i) {
    arg[1 + (i - 1) %% nrow(arg),]
  }
  from <- fixarg(from)
  to <- fixarg(to)
  center <- fixarg(center)
  
  m <- max(nrow(from), nrow(to), nrow(center), length(base))
  base <- rep_len(base, m)
  
  result <- matrix(NA_real_, nrow = 1, ncol = 3)
  
  for (j in seq_len(m)) {
  	from1 <- getrow(from, j)
  	to1 <- getrow(to, j)
  	center1 <- getrow(center, j)
  	# The "arc" might be a straight line
  	if (isTRUE(all.equal(from1, center1)) || 
  			isTRUE(all.equal(to1, center1)) ||
  			isTRUE(all.equal(normalize(from1 - center1), 
  											 normalize(to1 - center1)))) {
  		result <- rbind(result, from1, to1) 	
  	} else {
  		base1 <- base[j]
  		logr1 <- log(veclen(from1 - center1))
  		logr2 <- log(veclen(to1 - center1))
  		A <- normalize(from1 - center1)
  		B <- normalize(to1 - center1)
  		steps <- if (base1 <= 0) 4*abs(base1) + 1 else 4*base1 - 1
  		for (k in seq_len(steps)) {
  			if (k %% 2) {
  				A1 <- A * (-1)^(k %/% 2)
  				B1 <- B * (-1)^(k %/% 2 + (base1 > 0))
  			} else {
  				A1 <- B * (-1)^(k %/% 2 + (base1 <= 0))
  				B1 <- A * (-1)^(k %/% 2)
  			}
  			theta <- acos(sum(A1*B1))
  			if (isTRUE(all.equal(theta, pi)))
  				warning("Arc ", j, " points are opposite each other!  Arc is not well defined.")
  			if (missing(n))
  				n1 <- ceiling(circle*theta/(2*pi))
  			else
  				n1 <- n
  			
  			if (missing(radius)) {
  				pretheta <- (k %/% 2)*pi - (k %% 2 == 0)*theta
  				if (k == 1)
  					totaltheta <- (steps %/% 2)*pi - (steps %% 2 == 0)*theta + theta
  				p1 <- pretheta/totaltheta
  				p2 <- (pretheta + theta)/totaltheta
  				radius1 <- exp(seq(from = (1 - p1)*logr1 + p1*logr2,
  													 to   = (1 - p2)*logr1 + p2*logr2,
  													 length.out = n1 + 1))
  			} else
  				radius1 <- rep_len(radius, n1 + 1)
  			arc <- matrix(NA_real_, nrow = n1 + 1, ncol = 3)
  			p <- seq(0, 1, length.out = n1 + 1)
  			arc[1,] <- center1 + radius1[1]*A1
  			arc[n1 + 1,] <- center1 + radius1[n1 + 1]*B1
  			AB <- veclen(A1 - B1)
  			for (i in seq_len(n1)[-1]) {
  				ptheta <- p[i]*theta
  				phi <- pi/2 + (0.5 - p[i])*theta
  				q <- (sin(ptheta) / sin(phi))/AB
  				D <- (1-q)*A1 + q*B1
  				arc[i,] <- center1 + radius1[i] * normalize(D)
  			}
  			if (k == 1)
  				result <- rbind(result, arc)
  			else
  				result <- rbind(result[-nrow(result), ,drop = FALSE], arc)
  		}
  	}
  	result <- rbind(result, result[1,])  
  }
  if (plot)
  	lines3d(result[c(-1, -nrow(result)), , drop = FALSE], ...)
  else
  	result[c(-1, -nrow(result)), , drop = FALSE]
}

saveURI <- function(uri, con) {
  if (!grepl("^data:", uri))
    stop("Does not appear to be a data URI")
  header <- sub(",.*", "", uri)
  type <- sub(";.*", "", sub("^data:", "", header))
  encoding <- sub("^.*;", "", header)
  if (encoding != "base64")
    stop("Not encoded in base64")
  payload <- sub(paste0(header, ","), "", uri, fixed = TRUE)
  writeBin(base64_dec(payload), con)
  invisible(type)
}

writeOBJ <- function(con,
                     pointRadius=0.005, 
                     pointShape = icosahedron3d(),
                     lineRadius = pointRadius,
                     lineSides = 20,
                     pointsAsPoints = FALSE,
                     linesAsLines = FALSE,
                     withNormals = TRUE,
                     withTextures = TRUE,
                     separateObjects = TRUE,
                     ids = tagged3d(tags),
                     tags = NULL) {
 
  writeHeader <- function() {
    ident <- paste(filename, " produced by RGL")
    cat("#", ident, "\n", file=con)
  }   
    
  Vertices <- 0
  Normals <- 0
  Texcoords <- 0
  
  writeData <- function(id) {
    vbase <- Vertices
    tbase <- Texcoords
    nbase <- Normals
    vertices <- rgl.attrib(id, "vertices")
    cat(paste("v", vertices[,1], vertices[,2], vertices[,3]), 
        sep="\n", file=con)
    n <- nrow(vertices)
    Vertices <<- Vertices + n
    if (withTextures) {
      textures <- rgl.attrib(id, "texcoords")
      if (nrow(textures)) 
        cat(paste("vt", textures[,1], textures[,2]),
          sep="\n", file=con)
      Texcoords <<- Texcoords + nrow(textures)
    }
    if (withNormals) {
      normals <- rgl.attrib(id, "normals")
      if (nrow(normals))
        cat(paste("vn", normals[,1], normals[,2], normals[,3]),
          sep="\n", file=con)
      Normals <<- Normals + nrow(normals)
    }
    list(n=n, 
         ntexcoords=if (withTextures) nrow(textures) else 0,
         nnormals=if (withNormals) nrow(normals) else 0,
         vbase=vbase,
         tbase=tbase,
         nbase=nbase)
  }
  
  refnum <- function(n) sprintf("%d", n)
  
  writeTriangles <- function(id) {
    if (separateObjects)
      cat("o triangles", id, "\n", sep="", file=con)
    x <- writeData(id)
    indices <- refnum(x$vbase + getIndices(id))
    if (x$ntexcoords)
      indices <- paste0(indices, "/", refnum(x$tbase + getIndices(id)))
    if (x$nnormals)
      indices <- paste0(indices, if (!x$ntexcoords) "/", 
      			"/", refnum(x$nbase + getIndices(id)))
    indices <- matrix(indices, ncol=3, byrow=TRUE)
    cat(paste("f", indices[,1], indices[,2], indices[,3]), 
        sep="\n", file=con)
  }
  
  writeQuads <- function(id) {
    if (separateObjects)
      cat("o quads", id, "\n", sep="", file=con)
    x <- writeData(id)
    indices <- refnum(x$vbase + getIndices(id))
    if (x$ntexcoords)
      indices <- paste0(indices, "/", refnum(x$tbase + getIndices(id)))
    if (x$nnormals)
      indices <- paste0(indices, if (!x$ntexcoords) "/", 
      			"/", refnum(x$nbase + getIndices(id)))
    indices <- matrix(indices, ncol=4, byrow=TRUE)
    cat(paste("f", indices[,1], indices[,2], indices[,3], indices[,4]), 
        sep="\n", file=con)
  }
      
  writeSurface <- function(id) {
    if (separateObjects)
      cat("o surface", id, "\n", sep="", file=con)
    x <- writeData(id)
    dims <- rgl.attrib(id, "dim")
    nx <- dims[1]
    nz <- dims[2]
    rows <- seq_len(nx)
    
    vertices <- matrix(character(0), ncol=3)
    for (i in seq_len(nz)[-nz]) {
      indices <- getIndices(id)[(i-1)*nx + 
      		c(rows[-nx],rows[-nx],
                  rows[-1]+nx,rows[-nx]+nx,
                  rows[-1],rows[-1]+nx)]
      cindices <- refnum(x$vbase + indices)
      if (x$ntexcoords)
        cindices <- paste0(cindices, "/", refnum(x$tbase + indices))
      if (x$nnormals)
        cindices <- paste0(cindices, if (!x$ntexcoords) "/",
                           "/", refnum(x$nbase + indices))
      vertices <- rbind(vertices, matrix(cindices, ncol=3))
    }
    cat(paste("f", vertices[,1], vertices[,2], vertices[,3]),
        sep="\n", file=con)
  }      
  
  writeMesh <- function(mesh, scale=1, offset=c(0,0,0)) {
    vertices <- asEuclidean(t(mesh$vb))*scale 
    n <- nrow(vertices)
    vertices <- vertices + rep(offset, each=n)
    vbase <- Vertices
    cat(paste("v", vertices[,1], vertices[,2], vertices[,3]), 
        sep="\n", file=con)
    Vertices <<- Vertices + n
    if (withTextures && length(textures <- mesh$texcoords)) {
      tbase <- Texcoords
      textures <- asEuclidean(t(textures))
      cat(paste("vt", textures[,1], textures[,2]),
          sep="\n", file=con)
      Texcoords <<- Texcoords + nrow(textures)
    } else
      withTextures <- FALSE
    if (withNormals && length(normals <- mesh$normals)) {
      nbase <- Normals
      normals <- asEuclidean(t(normals))
      cat(paste("vn", normals[,1], normals[,2], normals[,3]),
          sep="\n", file=con)
      Normals <<- Normals + nrow(normals)
    } else
      withNormals <- FALSE
    nt <- length(mesh$it)/3
    nq <- length(mesh$ib)/4
    if (nt) {
      indices <- t(mesh$it)
      cindices <- refnum(vbase + indices)
      if (withTextures)
        cindices <- paste0(cindices, "/", refnum(tbase + indices))
      if (withNormals)
        cindices <- paste0(cindices, if (!withTextures) "/",
                           "/", refnum(nbase + indices))
      cindices <- matrix(cindices, ncol=3)
      cat(paste("f", cindices[,1], cindices[,2], cindices[,3]), 
          sep="\n", file=con)
    }
    if (nq) {
      indices <- t(mesh$ib)
      cindices <- refnum(vbase + indices)
      if (withTextures)
        cindices <- paste0(cindices, "/", refnum(tbase + indices))
      if (withNormals)
        cindices <- paste0(cindices, if (!withTextures) "/",
                           "/", refnum(nbase + indices))
      cindices <- matrix(cindices, ncol=4)
      cat(paste("f", cindices[,1], cindices[,2], 
                     cindices[,3], cindices[,4]), 
          sep="\n", file=con)
    }
  }

  writeSpheres <- function(id) {
    if (separateObjects)
      cat("o sphere", id, "\n", sep="", file=con) 
    vertices <- expandVertices(id)
    n <- nrow(vertices)    
    radii <- expandAttrib(id, "radii")
    radii <- rep(radii, length.out=n)
    x <- subdivision3d(icosahedron3d(),3)
    r <- sqrt(x$vb[1,]^2 + x$vb[2,]^2 + x$vb[3,]^2)
    x$vb[4,] <- r
    x$normals <- x$vb
    for (i in seq_len(n)) 
      writeMesh(x, radii[i], vertices[i,])
  }  
  
  avgScale <- function() {
    bbox <- par3d("bbox")
    ranges <- c(bbox[2]-bbox[1], bbox[4]-bbox[3], bbox[6]-bbox[5])
    if (prod(ranges) == 0) 1
    else exp(mean(log(ranges)))
  }  
   
  writePoints <- function(id) {
    if (separateObjects)
      cat("o points", id, "\n", sep="", file=con)
    if (pointsAsPoints) {
      x <- writeData(id)
      cat("p", refnum(x$vbase + getIndices(id)), "\n", file=con)
    } else {
      vertices <- expandVertices(id)
      n <- nrow(vertices)
      radius <- pointRadius*avgScale()
      if (withNormals && is.null(pointShape$normals))
        pointShape <- addNormals(pointShape)
      for (i in seq_len(n)) 
        writeMesh(pointShape, radius, vertices[i,])
    }
  }
  
  writeSegments <- function(id) {
    if (separateObjects)
      cat("o segments", id, "\n", sep="", file=con)
    if (linesAsLines) {
      x <- writeData(id)
      indices <- matrix(refnum(x$vbase + getIndices(id)), ncol=2, byrow=TRUE)
      cat(paste("l", indices[,1], indices[,2]), sep="\n", file=con)
    } else {
      vertices <- expandVertices(id)
      n <- nrow(vertices)    
      n <- n/2
      radius <- lineRadius*avgScale()
      for (i in seq_len(n)) {
        cyl <- cylinder3d( vertices[(2*i-1):(2*i),1:3],
     			   radius = radius,
     			   sides = lineSides, 
     			   closed = -2 )
        if (withNormals)
          cyl <- addNormals(cyl)
        writeMesh(cyl)
      }
    }
  }
  
  writeLines <- function(id) {
    if (separateObjects)
      cat("o lines", id, "\n", sep="", file=con)
    if (linesAsLines) {
      x <- writeData(id)
      indices <- refnum(x$vbase + getIndices(id))
      cat("l", indices, "\n", file=con)
    } else {
      vertices <- expandVertices(id)
      n <- nrow(vertices) - 1
      radius <- lineRadius*avgScale()
      for (i in seq_len(n)) {
        cyl <- cylinder3d( vertices[i:(i+1),],
     			   radius = radius,
     			   sides = lineSides, 
     			   closed = -2 )
        if (withNormals) 
          cyl <- addNormals(cyl)
        writeMesh(cyl)
      }
    }
  }
  
  knowntypes <- c("triangles", "quads", #,
                  "surface", "spheres", "points", 
                  "linestrip", "lines", "planes")
  
  #  Execution starts here!

  if (is.character(con)) {
    con <- file(con, "w")
    on.exit(close(con))
  }
  filename <- summary(con)$description
  
  if (NROW(bbox <- ids3d("bboxdeco")) && (is.null(ids) || bbox$id %in% ids)) {
    ids <- setdiff(ids, bbox$id)
    save <- par3d(skipRedraw = TRUE)
    bbox <- convertBBox(bbox$id)
    on.exit({ pop3d(id=bbox); par3d(save) }, add=TRUE) # nolint
    dobbox <- TRUE
  } else dobbox <- FALSE 
  
  if (is.null(ids)) {
    ids <- ids3d()
    types <- as.character(ids$type)
    ids <- ids$id
  } else {
    if (dobbox) ids <- c(ids, bbox)
    allids <- ids3d()
    ind <- match(ids, allids$id)
    keep <- !is.na(ind)
    if (any(!keep)) warning(gettextf("Object(s) with id %s not found", paste(ids[!keep], collapse=" ")),
    			    domain = NA)
    ids <- ids[keep]
    types <- allids$type[ind[keep]]
  }  
    
  unknowntypes <- setdiff(types, knowntypes)
  if (length(unknowntypes))
    warning(gettextf("Object type(s) %s not handled", 
      paste("'", unknowntypes, "'", sep="", collapse=", ")), domain = NA)

  keep <- types %in% knowntypes
  ids <- ids[keep]
  types <- types[keep]
  
  writeHeader()

  for (i in seq_along(ids))
    switch(types[i],
      planes =,
      triangles = writeTriangles(ids[i]),
      quads = writeQuads(ids[i]),
      surface = writeSurface(ids[i]),
      spheres = writeSpheres(ids[i]),
      points = writePoints(ids[i]),
      lines = writeSegments(ids[i]),
      linestrip = writeLines(ids[i])
    )
  
  invisible(filename)
}

readOBJ <- function(con, ...) {
  lines <- readLines(con)
  lines <- sub("^[[:blank:]]*", "", lines)
  instrs <- sub("[[:blank:]].*", "", lines)
  instrs <- sub("^#.*", "#", instrs)
  vertices <- read.table(textConnection(lines[instrs == "v"]),
                         col.names = c("instr", "x", "y", "z"),
                         colClasses = c(instr = "character", 
                                        x="numeric",
                                        y="numeric",
                                        z="numeric"))
  vertices <- with(vertices, rbind(x, y, z)) # nolint
  
  subset <- lines[instrs == "vn"]
  if (length(subset)) {
    fields <- count.fields(textConnection(subset))
    if (!all(fields == 4))
      stop("Normals must have 4 fields")
    vn <- read.table(textConnection(subset),
  			    col.names = c("instr", "x", "y", "z"),
  			    colClasses = c(instr = "character", 
  			    	       x="numeric",
  			    	       y="numeric",
  			    	       z="numeric"))
    vn <- rbind(t(vn[,2:4]), 1)
  } else
    vn <- matrix(numeric(), nrow = 4, ncol = 0)
  
  subset <- lines[instrs == "vt"]
  if (length(subset)) {
    fields <- count.fields(textConnection(subset))
    if (length(unique(fields)) != 1)
      stop("Textures must have consistent field count") 
    fields <- fields[1]
    colClasses <- c("character", rep("numeric", fields - 1))
    vt <- read.table(textConnection(subset),
  		            colClasses = colClasses)
    if (fields == 2)
      vt <- cbind(vt, 0)
    vt <- t(vt[, 2:3])
  } else
    vt <- matrix(numeric(), nrow = 2, ncol = 0)
  
  # Get rid of texture and normal info
  polys <- gsub("/[^[:blank:]]*", "", lines[instrs == "f"])
  polys <- strsplit(polys, "[[:blank:]]+")
  polys <- lapply(polys, function(poly) as.numeric(poly[-1]))
  
  # Regexp suggested by Bill Dunlap -- thanks!  Modified 
  # by DJM to replace " " with "[[:blank:]]"
  normals <- gsub("(^|[[:blank:]]*)([^/[:blank:]]*/?){0,2}", "\\1", lines[instrs == "f"]) # nolint
  normals <- strsplit(normals, "[[:blank:]]+")
  normals <- lapply(normals, function(normal) as.numeric(normal[nchar(normal) > 0]))

  textures <- gsub("(^|[[:blank:]]*)([^/[:blank:]]*/?){0,1}", "\\1", lines[instrs == "f"]) # nolint
  textures <- gsub("/[^[:blank:]]*", "", textures)
  textures <- strsplit(textures, "[[:blank:]]+")
  textures <- lapply(textures, function(texture) as.numeric(texture[nchar(texture) > 0]))
  
  nverts <- sapply(polys, length)
  nnorms <- sapply(normals, length) 
  ntexts <- sapply(textures, length)
  
  hasnormals <- nnorms == nverts
  hastextures <- ntexts == nverts
  if (any(hasnormals) || any(hastextures)) {
    # OBJ format allows different normals to be associated
    # with a single vertex in different polygons.  rgl 
    # doesn't, so may need to replicate some vertices.
    # This could be slow...
  	
    vlinks <- vector("list", ncol(vertices))
    for (i in seq_along(polys)) {
      nvec <- tvec <- NA
      if (hasnormals[i])
        nvec <- as.numeric(normals[[i]])
      if (hastextures[i])
      	tvec <- as.numeric(textures[[i]])
      vvec <- as.numeric(polys[[i]])
      for (j in seq_along(vvec))
        vlinks[[vvec[j]]] <- rbind(vlinks[[vvec[j]]], c(nvec[j], tvec[j], i, j))
    }
    total <- 0
    for (i in seq_along(vlinks)) {
      # Sort by texture 
      vlinks[[i]] <- vlinks[[i]][order(vlinks[[i]][,2]),,drop=FALSE]
      total <- total + max(1, length(unique(vlinks[[i]][,2])))
    }
    last <- ncol(vertices)
    vertices <- cbind(vertices, matrix(NA_real_, 3, total - ncol(vertices)))
    vnormals <- matrix(0, 4, total)
    vtexcoords <- matrix(NA_real_, 2, total)
    for (i in seq_along(vlinks)) {
      links <- vlinks[[i]]
      if (nrow(links)) {
      	# Average the normals at this vertex by
      	# summing the homogeneous coordinates
      	for (j in seq_len(nrow(links)))
          if (!is.na(links[j,1]))
      	    vnormals[,i] <- vnormals[,i] + vn[,links[1,1]]
        # A given vertex may have more than one texture
        # coordinate; rgl doesn't allow this, so we
        # duplicate vertices where that happened
        if (!is.na(links[1,2]))
      	  vtexcoords[,i] <- vt[,links[1,2]]
        same <- duplicated(links[,2])
        duped <- FALSE
        for (j in seq_len(nrow(links))[-1]) {
          if (!same[j]) {
            last <- last + 1          
            vertices[,last] <- vertices[,i]
            vnormals[,last] <- vnormals[,i]
            duped <- TRUE
          }
          # Update the polygon and texture links to the new copy
          if (duped) {
            polys[[links[j, 3]]][links[j, 4]] <- last
            if (!is.na(links[j, 2]))
              vtexcoords[,last] <- vt[,links[j, 2]]
          }
        }
      }
    }
  }

  triangles <- do.call(cbind, polys[nverts == 3])
  if (!length(triangles))
    triangles <- matrix(numeric(), 3, 0)
  
  # We build quads transposed, because we're going to stick
  # it directly into the structure
  quads <- do.call(cbind, polys[nverts == 4])

  others <- which(!(nverts %in% 3:4))
  
  # FIXME:  this will be really slow if there are a lot of others
  # Should pre-allocate extra space.
  for (i in seq_along(others)) {
    v <- polys[[others[i]]]
    tri <- triangulate(t(vertices[,v]))
    tri <- structure(v[tri], dim = dim(tri))
    triangles <- cbind(triangles, tri)
  }
  ignored <- unique(instrs)
  ignored <- ignored[!(ignored %in% c("v", "vn", "vt", "f", "", "#"))]
  if (length(ignored))
    warning(gettextf("Instructions %s ignored", paste0('"', ignored, '"', collapse = ", ")),
    	    domain = NA)
  result <- tmesh3d(vertices, triangles, homogeneous = FALSE, ...)
  if (length(quads)) 
    result$ib <- quads
  if (any(hasnormals))
    result$normals <- vnormals[1:3,]/rep(vnormals[4,], each=3)
  if (any(hastextures))
    result$texcoords <- vtexcoords
  
  result
}

set.seed(123)
test_that("readOBJ works", {
	# textured rectangular prism
	obj_file <- "# geometric vertices
v 1 1 -0.125
v 1 -1 -0.125
v -1 -1 -0.125
v -1 1 -0.125
v 1 1 0.125
v 1 -1 0.125
v -1 -1 0.125
v -1 1 0.125
# texture coordinates
vt 0.025 1
vt 0.025 0
vt 0.425 0
vt 0.425 1
vt 0.575 1
vt 0.575 0
vt 0.975 0
vt 0.975 1
vt 0.52 0
vt 0.48 0
vt 0.48 1
vt 0.52 1
# Textured polygonal face element
f 1/1 2/2 3/3 4/4
f 5/5 8/8 7/7 6/6
f 6/9 7/10 3/11 2/12
f 7/9 8/10 4/11 3/12
f 8/9 5/10 1/11 4/12
f 5/9 6/10 2/11 1/12"
	filename <- tempfile(fileext = ".obj")
	writeLines(obj_file, filename)
	mesh <- readOBJ(filename)
	unlink(filename)
	open3d()
	shade3d(mesh)
	expect_known_scene("obj")
})

pch3d <- function(x, y = NULL, z = NULL, pch = 1, bg = material3d("color")[1], 
		  cex = 1, radius = avgscale*cex/8, color = "black", lit = FALSE, ...) {
  xyz <- xyz.coords(x, y, z)
  xyz <- cbind(xyz$x, xyz$y, xyz$z)
  pch <- rep(pch, length.out = nrow(xyz))
  color <- rep(color, length.out = nrow(xyz))
  bg <- rep(bg, length.out = nrow(xyz))
  
  basic <- function(p) 
  	switch(as.character(p),
  	       '0' = { # square
  	       	theta <- seq(pi/4, 2*pi + pi/4, length.out = 5)
  	       	cbind(cos(theta), sin(theta), 0)
  	       },
  	       '1' = { # circle (actually 16-gon)
  	       	theta <- seq(0, 2*pi, length.out = 17)
  	       	cbind(cos(theta), sin(theta), 0)*sqrt(0.75)
  	       },
  	       '2' = { # triangle
  	       	theta <- seq(pi/2, 2*pi + pi/2, length.out = 4)
  	       	cbind(cos(theta), sin(theta), 0)
  	       },
  	       '-2' = { # triangle in square
  	       	cbind(c(0, -1, 1, 0), c(1, -1, -1, 1), 0)*sqrt(0.5)
  	       },
  	       '3' = { # plus
  	       	theta <- seq(0, 2*pi, length.out = 5)
  	       	rbind(cbind(cos(theta[c(1,3)]), sin(theta[c(1,3)]), 0),
  	       	      c(NA, NA, NA),
  	       	      cbind(cos(theta[c(2,4)]), sin(theta[c(2,4)]), 0))
  	       },
  	       '4' = { # cross
  	       	theta <- seq(pi/4, 2*pi + pi/4, length.out = 5)
  	       	rbind(cbind(cos(theta[c(1,3)]), sin(theta[c(1,3)]), 0),
  	       	      c(NA, NA, NA),
  	       	      cbind(cos(theta[c(2,4)]), sin(theta[c(2,4)]), 0))
  	       },      	
  	       '5' = { # diamond
  	       	theta <- seq(0, 2*pi, length.out = 5)
  	       	cbind(cos(theta), sin(theta), 0)
  	       },
  	       '6' = { # nabla
  	       	theta <- seq(3*pi/2, 2*pi + 3*pi/2, length.out = 4)
  	       	cbind(cos(theta), sin(theta), 0)
  	       })*radius
  
  join <- function(s1, s2) 
    rbind(s1, c(NA, NA, NA), s2)
  
  if (missing(radius))
    avgscale <- sqrt(sum(c(diff(range(xyz[,1],na.rm=TRUE)), 
  			   diff(range(xyz[,2],na.rm=TRUE)), 
  			   diff(range(xyz[,3],na.rm=TRUE)))^2/3))
  
  rot <- rotate3d(r3dDefaults$userMatrix, pi/2, 1, 0, 0)
  
  draw <- function(s, color) {
    sprites3d(thisxyz, shapes = lines3d(s, color = color, lit = lit, ...), userMatrix = rot)
  }
  
  fill <- function(s, color)
    sprites3d(thisxyz, shapes = polygon3d(s, color = color, lit = lit, ...), userMatrix = rot)
  
  outline <- function(s, color, bg) {
    dots <- list(...)
    dots$color <- bg
    sprites3d(thisxyz, shapes = c(lines3d(s, color = color, lit = lit, ...), do.call(polygon3d, c(list(s, lit = lit), dots))),
    	      userMatrix = rot)
  }
  
  save <- par3d(skipRedraw = TRUE)
  on.exit(par3d(save))
  
  ids <- numeric()   
  if (is.numeric(pch)) {
    bg[!(pch %in% 21:25)] <- bg[1]
    pchcol <- data.frame(pch, color, bg)
    uniquePchcol <- unique(pchcol)
    for (i in seq_len(nrow(uniquePchcol))) {
      p <- uniquePchcol$pch[i]
      cl <- uniquePchcol$color[i]
      b <- uniquePchcol$bg[i]
      thisxyz <- xyz[p == pch & cl == color & b == bg,, drop = FALSE]
      ids <- c(ids, switch(as.character(p),
        '0' =,
        '1' =,
        '2' =, 
        '3' =,
        '4' =,
        '5' =,
        '6' = draw(basic(p), color = cl),
        '7' = draw(join(basic(0), basic(4)), color = cl),
        '8' = draw(join(basic(3), basic(4)), color = cl),
        '9' = draw(join(basic(3), basic(5)), color = cl),
        '10' = draw(join(basic(1), basic(3)*sqrt(0.75)), color = cl),
        '11' = draw(join(basic(2), basic(6)), color = cl),
        '12' = draw(join(basic(0), basic(3)*sqrt(0.5)), color = cl),
        '13' = draw(join(basic(1), basic(4)), color = cl),
        '14' = draw(join(basic(0), basic(-2)), color = cl),
        '15' = fill(basic(0), color = cl),
        '16' = fill(basic(1), color = cl),
        '17' = fill(basic(2), color = cl),
        '18' = fill(basic(5)/sqrt(2), color = cl),
        '19' = fill(basic(1), color = cl),
        '20' = fill(basic(1)*(2/3), color = cl),
        '21' = outline(basic(1), color = cl, bg = b),
        '22' = outline(basic(0), color = cl, bg = b),
        '23' = outline(basic(5), color = cl, bg = b),
        '24' = outline(basic(2), color = cl, bg = b),
        '25' = outline(basic(6), color = cl, bg = b)))
    }
    letters <- 32 <= pch & pch <= 255
    xyz <- xyz[letters,, drop = FALSE]
    if (nrow(xyz)) {
      pch <- rawToChar(as.raw(pch[letters]), multiple = TRUE)
      color <- color[letters]
    }
  }
  if (nrow(xyz)) {
    pch <- substr(pch, start = 1, stop = 1)
    ids <- c(ids, text3d(xyz, texts = pch, cex = cex, color = color, ...))
  }
  lowlevel(ids)
}

# all rgl demos
demo(hist3d)
demo(abundance)
demo(lsystem)
demo(subdivision)
# requires MASS package
demo(bivar)
demo(envmap)
demo(flag)
demo(lollipop3d)
demo(mouseCallbacks)
demo(shapes3d)
demo(shinyDemo)
demo(shinyMouse)
demo(shinyTabs)
demo(shinyToggle)
demo(simpleShinyRgl)
demo(stereo)

rglMouse <- function(sceneId, 
		     choices = c("trackball", "selecting", 
		     	         "xAxis", "yAxis", "zAxis", 
				 "polar", "zoom", "fov", 
				 "none"),
	       labels = choices,
	       button = 1, 
		     dev = cur3d(), 
		     subscene = currentSubscene3d(dev),
	       default = par3d("mouseMode", dev = dev, subscene = subscene)[button + 1],
		     stayActive = FALSE,
		     height = 40,
		     ...) {
  stopifnot(length(choices) == length(labels))
  stopifnot(length(button == 1) && button %in% 1:3)
  options <- mapply(function(x, y) tags$option(x, value = y), labels, choices,
  		    SIMPLIFY = FALSE)
  for (i in seq_along(choices)) 
    options[[i]] <- tags$option(labels[i], value = choices[i])
  default <- which(choices == default)
  options[[default]] <- tagAppendAttributes(options[[default]], selected = NA)
  changecode <- 'document.getElementById(this.attributes.rglSceneId.value).rglinstance.
                   setMouseMode(this.value, 
                                button = parseInt(this.attributes.rglButton.value), 
                                subscene = parseInt(this.attributes.rglSubscene.value),
                                stayActive = parseInt(this.attributes.rglStayActive.value))' 
  result <- tags$select(tagList(options), 
  		        onchange = HTML(changecode), 
  		        rglButton = button,
  		        rglSubscene = subscene,
  		        rglStayActive = as.numeric(stayActive),
  		        ...)
  
  if (!missing(sceneId) && !is.null(sceneId)) {
    upstream <- processUpstream(sceneId)
    if (!is.null(upstream$prevRglWidget))
      result <- tagAppendAttributes(result, rglSceneId = upstream$prevRglWidget)
  } else
    upstream <- list()
  
  if (is.list(upstream$objects)) {
  	if (requireNamespace("manipulateWidget", quietly = TRUE))
      return(do.call(manipulateWidget::combineWidgets, c(upstream$objects, 
                              list(result, 
                                   rowsize = c(upstream$rowsizes, height), 
                                   ncol = 1))))
    else
    	warning("Combining widgets requires the 'manipulateWidget' package.", call. = FALSE)
  }
  browsable(result)
}

subst <- function(strings, ..., digits=7) {
  substitutions <- list(...)
  names <- names(substitutions)
  if (is.null(names)) names <- rep("", length(substitutions))
  for (i in seq_along(names)) {
    if ((n <- names[i]) == "")
      n <- as.character(sys.call()[[i+2]])
    value <- substitutions[[i]]
    if (is.numeric(value)) 
      value <- formatC(unclass(value), digits=digits, width=1)
    strings <- gsub(paste("%", n, "%", sep=""), value, strings)
  }
  strings
}


convertBBox <- function(id,
  verts = rgl.attrib(id, "vertices"),
  text = rgl.attrib(id, "text"),
  mat = rgl.getmaterial(id = id)) {
  if (!length(text))
    text <- rep("", NROW(verts))
  if (length(mat$color) > 1)
    mat$color <- mat$color[2] # We ignore the "box" colour

  if(any(missing <- text == ""))
    text[missing] <- apply(verts[missing,], 1, function(row) format(row[!is.na(row)]))

  res <- integer(0)
  if (any(inds <- is.na(verts[,2]) & is.na(verts[,3])))
    res <- c(res, do.call(axis3d, c(list(edge = "x", at = verts[inds, 1], labels = text[inds]), mat)))
  if (any(inds <- is.na(verts[,1]) & is.na(verts[,3])))
    res <- c(res, do.call(axis3d, c(list(edge = "y", at = verts[inds, 2], labels = text[inds]), mat)))
  if (any(inds <- is.na(verts[,1]) & is.na(verts[,2])))
    res <- c(res, do.call(axis3d, c(list(edge = "z", at = verts[inds, 3], labels = text[inds]), mat)))
  res <- c(res, do.call(box3d, mat))
  res
}

rootSubscene <- function() {
  id <- currentSubscene3d()
  repeat {
    info <- subsceneInfo(id)
    if (is.null(info$parent)) return(id)
    else id <- info$parent
  }
}

writeWebGL <- function(dir="webGL", filename=file.path(dir, "index.html"),
                       template = system.file(file.path("WebGL", "template.html"), package = "rgl"),
                       prefix = "",
                       snapshot = TRUE, commonParts = TRUE, reuse = NULL,
		       font="Arial",
                       width = NULL, height = NULL) {
  
  .Defunct("rglwidget")

}

# This function is modelled on similar functions in 
# the extrafont package.

loadfonts_rgl <- function(..., quiet = TRUE) {
	
	makeRglFont <- function(family) {
		getIndex <- function(Bold, Italic)
			1 + Bold + 2*Italic
		fontdata <- fontdata[fontdata$FamilyName == family,,drop = FALSE]
		result <- rep(NA, 4)
		for (i in seq_len(nrow(fontdata))) {
			index <- with(fontdata, getIndex(Bold[i], Italic[i]))
			if (is.na(result[index]))
				result[index] <- fontdata$fontfile[i]
		}
		value <- result[1]
		# Propagate to more conditions
		for (i in 2:4)
			if (is.na(result[i]))
				result[i] <- result[i-1]
		# Propagate to fewer
		for (i in 3:1)
			if (is.na(result[i]))
				result[i] <- result[i+1]
		result
	}
	
	register_family_rgl <- function(family) {
		# Now we can register the font with rgl with something like this:
		# rglFonts("Arial" = rglFont("Arial"))
		if (family %in% cfonts) {
			if (!quiet) {
				message(family, " already registered with rglFonts().")
			}
			return(NULL)
		}
		if (!quiet) {
			message("Registering font with R using rglFonts(): ", family)
		}
		# Since 'family' is a string containing the name of the argument, we
		# need to use do.call
		args <- list()
		args[[family]] <- makeRglFont(family)
		if (!is.null(args[[family]]))
			do.call(rglFonts, args)
	}
	
	if (!requireNamespace("extrafont", quietly = TRUE))
		stop("This function requires the extrafont package.")
	
	fontdata <- extrafont::fonttable()
	# remove empty FamilyNames
	fontdata <- fontdata[fontdata$FamilyName != "", , drop = FALSE]
	families <- unique(fontdata$FamilyName)
	# If args were given, limit attention to those
	args <- list(...)
	if (length(args))
		families <- intersect(families, args)
	cfonts <- names(rglFonts())
	lapply(families, register_family_rgl)
	allfonts <- rglFonts()
	names <- names(args)
	for (i in seq_along(names)) {
		origname <- args[[names[i]]]
		if (origname %in% names(allfonts)) {
			font <- allfonts[[origname]]
			arg <- list()
			arg[[names[i]]] <- font
			do.call(rglFonts, arg)
		}
	}
}

rglExtrafonts <- function(..., quiet = TRUE) {
	if (!requireNamespace("extrafont", quietly = TRUE))
		return()
	
	args <- list(...)
	names <- names(args)
	result <- character()
	if (is.null(names))
		names <- rep("", length(args))
	for (i in seq_along(args)) {
		choices <- args[[i]]
		if (length(choices)) {
		  font <- extrafont::choose_font(choices)
		  result[i] <- font
		  if (nchar(font)) {
		  	arg <- list(quiet = quiet)
		  	if (!nchar(names[i]))
		  		names[i] <- font
		  	arg[[names[i]]] <- font
		  	do.call(loadfonts_rgl, arg)
		  } else
		  	warning("Fonts ", paste0('"', choices, '"', collapse = ", "), " not found.")
		}
	}
	names(result) <- names
  invisible(result)
}

typeSignedByte    <- 5120
typeUnsignedByte  <- 5121
typeSignedShort   <- 5122
typeUnsignedShort <- 5123
typeSignedInt     <- 5124  # Not supported in glTF
typeUnsignedInt   <- 5125
typeFloat         <- 5126
typeDouble        <- 5130  # Not supported in glTF

gltfTypes <- c(byte = 5120, ubyte = 5121,
                   short = 5122, ushort = 5123,
                   uint = 5125, float = 5126,                     int = 5124, double = 5130)

getType <- function(x, types = "anyGLTF") {
  types <- match.arg(types, 
                     c(names(gltfTypes), c("anyGLTF", "any")), 
                     several.ok = TRUE)
  if ("anyGLTF" %in% types)
    types <- c(types, names(gltfTypes)[1:6])
  if ("any" %in% types)
    types <- c(types, names(gltfTypes))
  types <- unique(setdiff(types, c("anyGLTF", "any")))
  r <- suppressWarnings(range(x, na.rm = TRUE))
  if (is.integer(x) && 
      !any(is.na(x)) &&
      (r[1] >= 0 && 
       any(c("byte", "short", "int", "ubyte", "ushort", "uint") %in% types) ||
      (r[1] < 0 && 
       any(c("byte", "short", "int") %in% types))))
      {
    if (r[1] < 0 && ("byte" %in% types)) {
      if (-128 <= r[1] && r[2] <= 127)
        "byte"
      else if (-32768 <= r[1] && r[2] <= 32767 && ("short" %in% types))
        "short"
      else
        "int"
    } else {
      if (r[2] <= 255 && ("ubyte" %in% types))
        "ubyte"
      else if (r[2] <= 65535 && ("ushort" %in% types))
        "ushort"
      else
        "uint"
    }
  } else if (is.numeric(x)) {
    if ((-32768 <= r[1] && r[2] <= 32767 ||
         0 <= r[1] && r[2] <= 65535) && 
        isTRUE(all(x == as.integer(x))) &&
        any(c("byte", "short", "ubyte", "ushort") %in% types))
      getType(as.integer(x), types)
    else if ("float" %in% types)
      "float"
    else if ("double" %in% types)
      "double"
  } else
    stop('Unrecognized or disallowed type')
}

#' @title R6 Class for binary buffers in glTF files.
#'
#' @description
#' These files typically have one buffer holding all the
#' binary data for a scene.

Buffer <- R6Class("Buffer",
    public = list(

#' @param json
#'   list read from glTF file.
#' @param binfile
#'   optional External binary filename, or raw vector
#'
      initialize = function(json = NULL, binfile = NULL) {
        if (!is.null(json)) {
          private$buffers <- json$buffers
          private$bufferViews <- json$bufferViews
          private$accessors <- json$accessors
        }
        buffer <- self$getBuffer(0)
        if (is.null(buffer$uri)) {
          if (is.character(binfile))
            buffer$uri <- binfile
          else if (is.raw(binfile))
            buffer$bytes <- binfile
        }
        self$setBuffer(0, buffer)
      },

#' @description
#'   Load from file.
#'
#' @param uri Which file to load.
#' @param buf Which buffer number to load.
#'
      load = function(uri, buf = 0) {
        buffer <- self$getBuffer(buf)
        if (is.null(buffer))
          buffer <- list(byteLength = 0)
        self$closeBuffer(buf)
        if (is.character(uri)) {
          bytes <- readBin(uri, "raw", n = file.size(uri))
          buffer$uri <- uri
        } else if (is.raw(uri))
          bytes <- uri
        buffer$byteLength <- length(bytes)
        buffer$con <- rawConnection(bytes, open = "r+b")
        self$setBuffer(buf, buffer)
      },

#' @description
#'   Write open buffer to connection.
#'
#' @param con
#'   Output connection.
#' @param buf
#'   Buffer number.
#'
      saveOpenBuffer = function(con, buf = 0) {
        buffer <- self$getBuffer(buf)
        if (is.null(buffer) ||
            is.null(con0 <- buffer$con) ||
            !inherits(con0, "connection") ||
            !isOpen(con0))
          stop("buffer ", buf, " is not open.")
        bytes <- rawConnectionValue(con0)
        writeBin(bytes, con)
      },

#' @description
#'   Get buffer object.
#'
#' @param buf Buffer number.
#' @param default Default buffer object if `buf` not found.
#'
#' @return A list containing components described here:
#' \url{https://registry.khronos.org/glTF/specs/2.0/glTF-2.0.html#reference-buffer}.
#'
      getBuffer = function(buf, default = list(byteLength = 0)) {
        buffer <- if (buf + 1 <= length(private$buffers))
          private$buffers[[buf + 1]]
        if (is.null(buffer))
          default
        else
          structure(buffer, class = "gltfBuffer")
      },

#' @description
#'   Set buffer object.
#'
#' @param buf Buffer number.
#' @param buffer New value to insert.
#'
      setBuffer = function(buf, buffer)
        private$buffers[[buf + 1]] <- unclass(buffer),

#' @description
#'   Open a connection for the data in a buffer.
#'
#' @param buf Buffer number.
#'
#' @return An open binary connection.
#'
      openBuffer = function(buf) {
        buffer <- self$getBuffer(buf)
        if (is.null(buffer))
          stop("no such buffer")
        if (is.null(buffer$con)) {
          if (!is.null(bytes <- buffer$bytes)) {
            buffer$con <- rawConnection(bytes, open = "r+b")
            buffer$bytes <- NULL
            self$setBuffer(buf, buffer)
          } else if (is.null(buffer$uri)) {
            buffer$con <- rawConnection(raw(0), open = "r+b")
            self$setBuffer(buf, buffer)
          } else
            self$load(buffer$uri, buf = buf)
        }
        self$getBuffer(buf)$con
      },

#' @description
#'   Write data to buffer.
#'
#' @param values Values to write.
#' @param type Type to write.
#' @param size Byte size of each value.
#' @param buf Which buffer to write to.
#'
#' @return Byte offset of start of bytes written.
#'
      writeBuffer = function(values, type, size, buf = 0) {
        if (is.null(buffer <- self$getBuffer(buf)))
          self$setBuffer(buf, buffer <- list(byteLength = 0))
        byteLength <- buffer$byteLength
        byteOffset <- byteLength
        con <- self$openBuffer(buf)
        seek(con, byteOffset)
        byteOffset <- bitwAnd(byteOffset + size - 1, bitwNot(size - 1))
        if (is.null(byteLength))
          browser()
        if (byteOffset > byteLength) {
          writeBin(raw(byteOffset - byteLength), con)
        }
        if (type %in% c(typeFloat, typeDouble))
          values <- as.numeric(values)
        else
          values <- as.integer(values)
        writeBin(values, con, size = size, endian = "little")
        buffer <- self$getBuffer(buf)
        buffer$byteLength <- byteOffset + length(values)*size
        self$setBuffer(buf, buffer)
        byteOffset
      },

#' @description
#'   Close the connection in a buffer.
#'
#' If there was a connection open, this will save the
#' contents in the raw vector `bytes` within the buffer object.
#'
#' @param buf The buffer number.
#'
      closeBuffer = function(buf) {
        buffer <- self$getBuffer(buf)
        if (!is.null(buffer) &&
            !is.null(buffer$con)) {
          buffer$bytes <- rawConnectionValue(buffer$con)
          close(buffer$con)
          buffer$con <- NULL
          self$setBuffer(buf, buffer)
        }
      },

#' @description
#'   Close any open buffers.
#'
#'   Call this after working with a GLTF file to avoid warnings
#'   from R about closing unused connections.
#'
      closeBuffers = function() {
        for (i in seq_along(private$buffers)) {
          self$closeBuffer(i - 1)
        }
      },

#' @description
#'   Get `bufferView` object.
#'
#' @param bufv `bufferView` number.
#'
#' @return A list containing components described here:
#' \url{https://registry.khronos.org/glTF/specs/2.0/glTF-2.0.html#reference-bufferview}.
#'
      getBufferview = function(bufv) {
        bufferview <- private$bufferViews[[bufv+1]]
        if (is.null(bufferview))
          stop("bufferView ", bufv, " not found.")
        structure(bufferview, class = "gltfBufferview")
      },

#' @description
#'   Add a new buffer view.
#'
#' @param values Values to put in the view.
#' @param type Type of values.
#' @param size Size of values in bytes.
#' @param target Optional target use for values.
#' @param buf Which buffer to write to.
#'
#' @return New `bufferView` number.
#'
      addBufferView = function(values, type, size, target = NULL, buf = 0) {
        bufferview <- list()
        bufferview$buffer <- buf
        bufferview$byteLength <- size*length(values)

        buffer <- self$getBuffer(buf)

        bufferview$byteOffset <- self$writeBuffer(values, type, size, buf)
        if (!is.null(target))
          bufferview$target <- target
        self$setBufferview(length(private$bufferViews), bufferview)
        length(private$bufferViews) - 1
      },

#' @description
#'   Open a connection to a buffer view.
#'
#' @param bufv Which `bufferView`.
#'
#' @return A connection.

      openBufferview = function(bufv) {
        bufferview <- self$getBufferview(bufv)
        con <- self$openBuffer(bufferview$buffer)
        seek(con, bufferview$byteOffset)
        con
      },

#' @description
#'   Set `bufferView` object.
#'
#' @param bufv `bufferView` number.
#' @param bufferView New value to insert.

      setBufferview = function(bufv, bufferView)
        private$bufferViews[[bufv + 1]] <- unclass(bufferView),

#' @description
#'   Get accessor object
#'
#' @param acc Accessor number
#'
#' @return A list containing components described here:
#' \url{https://registry.khronos.org/glTF/specs/2.0/glTF-2.0.html#reference-accessor}
#'
      getAccessor = function(acc)
        structure(private$accessors[[acc + 1]], class = "gltfAccessor"),

#' @description
#'   Set accessor object.
#'
#' @param acc Accessor number.
#' @param accessor New value to insert.
#'
      setAccessor = function(acc, accessor)
        private$accessors[[acc + 1]] <- unclass(accessor),

#' @description
#'   Read data given by accessor number.
#'
#' @param acc Accessor number.
#'
#' @return A vector or array as specified in the accessor.  For the `MATn` types, the 3rd index
#' indexes the element.
#'
   readAccessor = function(acc) {
     if (acc + 1 > length(private$accessors))
       stop("No such accessor")
     accessor <- self$getAccessor(acc)
     self$readAccessor0(accessor)
   },

#' @description
#'   Read data given by accessor object.
#'
#' @param accessor Accessor object
#'
#' @return A vector or array as specified in the accessor.  For the `MATn` types, the 3rd index
#' indexes the element.
#'
      readAccessor0 = function(accessor) {
        typenames <- c("5120" = "byte", "5121" = "unsigned_byte",
                       "5122" = "short", "5123" = "unsigned_short",
                       "5125" = "unsigned_int", "5126" = "float")
        types <- c("5120" = "int", "5121" = "int",
                   "5122" = "int", "5123" = "int",
                   "5125" = "int", "5126" = "double")
        sizes <- c("5120" = 1, "5121" = 1,
                   "5122" = 2, "5123" = 2,
                   "5125" = 4, "5126" = 4)
        signeds <- c("5120" = TRUE, "5121" = FALSE,
                     "5122" = TRUE, "5123" = FALSE,
                     "5125" = TRUE, # not really, but make readBin happy
                     "5126" = TRUE)
        lens <- c(SCALAR = 1, VEC2 = 2, VEC3 = 3, VEC4 = 4,
                  MAT2 = 4, MAT3 = 9, MAT4 = 16)
        ctype <- as.character(accessor$componentType)
        atype <- accessor$type
        type <- types[ctype]
        len <- lens[atype]
        size <- sizes[ctype]
        signed <- signeds[ctype]
        count <- accessor$count
        if (is.null(accessor$bufferView)) {
          values <- numeric(count*len) # initialized to zero
        } else {
          view <- self$getBufferview(accessor$bufferView)
          con <- self$openBufferview(accessor$bufferView)
          if (is.null(view$byteStride)) {
            skip <- 0
          } else
            skip <- len*size - view$byteStride
          if (is.null(byteOffset <- accessor$byteOffset))
            byteOffset <- 0
          start <- seek(con) + byteOffset

          if (skip == 0) {
            seek(con, start)
            values <- readBin(con, type, n = len*count,  size = size,
                              signed = signed, endian = "little")
          } else {
            values <- numeric(count*len)
            for (i in seq_len(count)) {
              seek(con, start + (i-1)*view$byteStride)
              values[(i-1)*len + seq_len(len)] <-
                readBin(con, type, n = len,  size = size,
                      signed = signed, endian = "little")
            }
          }
          if (ctype == "5125") { # fix up unsigned integers
            values[is.na(values)] <- 2^31
            values[values < 0] <- values[values < 0] + 2^32
          }
        }
        if (!is.null(sparse <- accessor$sparse)) {
          indexobj <- sparse$indices
          indexobj$type <- "SCALAR"
          indexobj$count <- sparse$count
          index <- self$readAccessor0(indexobj)
          
          valueobj <- sparse$values
          valueobj$type <- "SCALAR"
          valueobj$componentType <- accessor$componentType
          valueobj$count <- len*sparse$count
          newvalues <- self$readAccessor0(valueobj)
          
          for (i in seq_len(sparse$count)) 
            for (j in seq_len(len))
              values[len*index[i] + j] <-
                newvalues[len*(i-1) + j]
        }
        
        if (!is.null(accessor$normalized) && accessor$normalized)
          values <- switch(ctype,
                           "5120" = (values + 128)/255 - 1, # byte
                           "5121" = values/255,             # u byte
                           "5122" = (values + 2^15)/65535 - 1, # short
                           "5123" = values/65535,           # u short
                           values)                 # default
        if (len > 1)
          if (grepl("MAT", atype)) {
            values <- array(values, dim=c(sqrt(len), sqrt(len), count))
          } else
            values <- matrix(values, ncol = len, byrow = TRUE)
        values
      },

#' @description
#'   Write values to accessor, not including `min` and `max`.
#'
#' @param values Values to write.
#' @param target Optional target use for values.
#' @param types Which types can be used?
#' @param normalized Are normalized integers allowed?
#' @param useDouble Whether to write doubles or singles.
#'
#' @return New accessor number

      addAccessor = function(values, target = NULL, types = "anyGLTF", normalized = FALSE) {
        componentTypeName <- getType(values, types)
        size <- switch(componentTypeName,
            "byte" =,       # typeSignedByte
            "ubyte" = 1,     # typeUnsignedByte = 1,
            "short" =,       # typeSignedShort =,
            "ushort" = 2,     # typeUnsignedShort = 2,
            "uint" =,       # typeUnsignedInt =,
            "int" =,       # typeSignedInt =,
            "float" = 4,     # typeFloat = 4,
            "double" = 8)     # typeDouble = 8)

        bufferView <- self$addBufferView(c(values), gltfTypes[componentTypeName],
                                    size = size, target = target)
        if (is.matrix(values) && nrow(values) > 1) {
          count <- ncol(values)
          type <- paste0("VEC", nrow(values))
        } else {
          count <- length(values)
          type <- "SCALAR"
        }
        if (normalized && 
            !(componentTypeName %in% 
              c("byte", "ubyte", "short", "ushort")))
          stop("Only bytes and short values can be normalized")
        accessor <- list(bufferView = bufferView,
                         componentType = gltfTypes[[componentTypeName]], # double to drop name
                         count = count,
                         type = type)
        if (normalized)
          accessor$normalized <- TRUE

        private$accessors <- c(private$accessors, list(accessor))
        length(private$accessors) - 1
      },

#' @description
#'   Convert buffer to data URI.
#'
#' @param buf Buffer to convert.
#'
#' @return String containing data URI.

      dataURI = function(buf = 0) {
        self$closeBuffer(buf)
        buffer <- self$getBuffer(buf)
        if (is.null(buffer))
          stop("Buffer ", buf, " does not exist.")
        bytes <- buffer$bytes
        if (is.null(bytes)) {
          if (is.null(buffer$uri))
            return(base64enc::dataURI(raw(0), mime = "application/octet-stream"))
          else {
            self$load(buffer$uri, buf)
            self$closeBuffer(buf)
            buffer <- self$getBuffer(buf)
            bytes <- buffer$bytes
          }
        }
        base64enc::dataURI(bytes, mime = "application/octet-stream")
      },


#' @description Convert to list.
#' @return List suitable for writing using JSON.
      as.list = function() {
        result <- list()
        for (n in names(private)) {
          thelist <- private[[n]]
          if (is.list(thelist) && length(thelist)) {
            for (i in seq_along(thelist))
              thelist[[i]] <- unclass(thelist[[i]])
            result[[n]] <- thelist
          }
        }
        result
      }
  ),

  private = list(
    buffers = list(),
    bufferViews = list(),
    accessors = list()
    # ,
    # 
    # finalize = function() {
    #   self$closeBuffers()
    # }

  )
)

as.mesh3d.default <- function(x, y = NULL, z = NULL,
                              type = c("triangles", "quads", "segments", "points"),
                              smooth = FALSE, 
                              tolerance = sqrt(.Machine$double.eps),
                              notEqual = NULL,
                              merge = TRUE,
                              ..., 
                              triangles) {
  if (missing(x)) {
    x <- rglId(ids3d()$id)
    return(as.mesh3d(x, ...))
  }
  verts <- rgl.vertex(x, y, z)
  if (!missing(triangles)) {
    warning("Argument 'triangles' is deprecated; please use 'type' instead.")
    if (missing(type)) {
      if (triangles)
        type <- "triangles"
      else
        type <- "quads"
    }
  }
  type <- match.arg(type, several.ok = TRUE)
  pcs <- c(triangles = 3, quads = 4, segments = 2, points = 1)
  nvert <- ncol(verts)
  okay <- FALSE
  for (i in seq_along(type)) {
    if (nvert %% pcs[type[i]] == 0) {
      okay <- TRUE
      break
    }
  }
  if (okay) type <- type[i]
  else stop("Wrong number of vertices")
    
  indices <- matrix(seq_len(nvert), nrow = pcs[type])
  if (merge) {
    if (!is.null(notEqual)) {
      dim <- dim(notEqual)
      if (length(dim) != 2 || dim[1] != nvert || dim[2] != nvert)
        stop("'notEqual' should be a ", nvert, " by ", nvert, " logical matrix.")
      notEqual <- notEqual | t(notEqual) # Make it symmetric
    } else
      notEqual <- matrix(FALSE, nvert, nvert)
    o <- order(verts[1,], verts[2,], verts[3,])
    i1 <- seq_len(nvert)[o]
    for (i in seq_len(nvert)[-1]) {
      if (isTRUE(all.equal(verts[,i1[i-1]], verts[,i1[i]], tolerance = tolerance))
          && (!isTRUE(notEqual[i1[i-1], i1[i]]))) {
        indices[indices == i1[i]] <- i1[i-1]
        notEqual[i1[i], ] <- notEqual[i1[i-1], ] <- notEqual[i1[i], ] | notEqual[i1[i-1], ]
        notEqual[, i1[i]] <- notEqual[, i1[i-1]] <- notEqual[i1[i], ] | notEqual[, i1[i-1]]
        i1[i] <- i1[i-1]
      }
    }
    i1 <- sort(unique(i1))
    keep <- seq_along(i1)
    for (i in keep) {
      verts[,i] <- verts[,i1[i]]
      indices[indices == i1[i]] <- i
    }
  } else
    keep <- seq_len(ncol(verts))

  mesh <- mesh3d(vertices = rbind(verts[,keep, drop = FALSE], 1), points = if (type == "points") indices, 
                  triangles = if (type == "triangles") indices,
                  quads = if (type == "quads") indices,
                  segments = if (type == "segments") indices,
                  material = list(...))
  if (smooth && type != "points")
    mesh <- addNormals(mesh)
  mesh
}

as.mesh3d.rglId <- function(x, type = NA, subscene = NA, 
                            ...) {
  local_t <- function(x) {
    if (!is.null(x)) t(x)
  }
  
  colorByFaces <- function(mesh) {
    constColumns <- function(m)
      all(apply(m, 2, function(col) length(unique(col))) == 1)
          
    # See if we can use meshColor = "faces"
    if (!is.null(mesh$material) && !is.null(mesh$meshColor)
        && mesh$meshColor == "vertices") {
      cols <- mesh$material$color
      alpha <- mesh$material$alpha
      texcoords <- mesh$texcoords
      if (length(cols) == 1 && length(alpha) == 1 && is.null(texcoords)) {
        mesh$meshColor <- "faces"
        return(mesh)
      }
      if (length(cols) == 1)
        cols <- rep(cols, length.out = ncol(mesh$vb))
      if (length(alpha) == 1)
        alpha <- rep(alpha, length.out = ncol(mesh$vb))
      prev <- 0
      newcols <- NULL
      newalpha <- NULL
      newtexcoords <- NULL
      
      if (!is.null(mesh$ip)) {
        inds <- mesh$ip
        newcols <- cols[inds]
        newalpha <- alpha[inds]
      }
      if (!is.null(mesh$is)) {
        inds <- as.numeric(mesh$is) 
        cols <- matrix(cols[inds], nrow = 2)
        alpha <- matrix(alpha[inds], nrow = 2)
        if (!constColumns(cols) || !constColumns(alpha))
          return(mesh)
        newcols <- c(newcols, cols[1,])
        newalpha <- c(newalpha, alpha[1,])
      }
      if (!is.null(mesh$it)) {
        inds <- as.numeric(mesh$it)
        cols <- matrix(cols[inds], nrow = 3)
        alpha <- matrix(alpha[inds], nrow = 3)
        if (!constColumns(cols) || !constColumns(alpha))
          return(mesh)
        if (!is.null(texcoords)) {
          tex1 <- matrix(texcoords[1, inds], nrow = 3)
          tex2 <- matrix(texcoords[2, inds], nrow = 3)
          if (!constColumns(tex1) || !constColumns(tex2))
            return(mesh)
        }
        newcols <- c(newcols, cols[1,])
        newalpha <- c(newalpha, alpha[1,])
        if (!is.null(texcoords))
          newtexcoords <- cbind(newtexcoords, rbind(tex1[1,], tex2[1,]))
      }
      if (!is.null(mesh$ib)) {
        inds <- as.numeric(mesh$ib)
        cols <- matrix(cols[inds], nrow = 3)
        alpha <- matrix(alpha[inds], nrow = 3)
        tex1 <- matrix(texcoords[1, inds], nrow = 3)
        tex2 <- matrix(texcoords[2, inds], nrow = 3)
        if (!constColumns(cols) || !constColumns(alpha) ||
            !constColumns(tex1) || !constColumns(tex2))
          return(mesh)
        newcols <- c(newcols, cols[1,])
        newalpha <- c(newalpha, alpha[1,])
        newtexcoords <- cbind(newtexcoords, rbind(tex1[1,], tex2[1,]))
      }
      mesh$meshColor <- "faces"
      mesh$material$color <- if (length(unique(newcols)) == 1)
                               newcols[1]
                             else
                               newcols
      mesh$material$alpha <- if (length(unique(newalpha)) == 1)
                               newalpha[1]
                             else
                               newalpha
      mesh$texcoords <- newtexcoords
    }
    mesh
  }
  
  mergeMaterials <- function(oldmat, newmat, n, type) {
    if (length(newmat$color) == 1)
      newmat$color <- rep(newmat$color, length.out = n)
    if (length(newmat$alpha) == 1)
      newmat$alpha <- rep(newmat$alpha, length.out = n)
    # meshColor isn't a material property, but put it here
    # for now...
    newmat$meshColor <- "vertices"
    if (is.null(oldmat))
      result <- newmat
    else {
      cols <- c(oldmat$color, newmat$color)
      oldmat$color <- newmat$color <- NULL
      alpha <- c(oldmat$alpha, newmat$alpha)
      oldmat$alpha <- newmat$alpha <- NULL
      same <- all.equal(oldmat, newmat)
      if (!isTRUE(same))
        warning(same, "\nOnly last material used")
      result <- newmat
      result$color <- cols
      result$alpha <- alpha
      result$meshColor <- "vertices"
    }
    result
  }

  ids <- ids3d(subscene = subscene)
  ids <- ids[ids$id %in% x,]
  # Parts will be drawn in order ip, is, it, ib,
  # so sort that way now
  ordered <- c(points = 1, lines = 2, linestrip = 3,
               triangles = 4, planes = 5, quads = 6, 
               surface = 7)
  if (!is.na(type))
    ids <- ids[ids$type %in% type,]
  ids <- ids[order(ordered[ids$type]),]
  ip <- NULL
  is <- NULL
  it <- NULL
  vertices <- NULL
  normals <- NULL
  texcoords <- NULL
  material <- NULL
  for (i in seq_len(NROW(ids))) {
    id <- ids[i, "id"]
    verts <- rgl.attrib(id, "vertices")
    nvert <- NROW(verts)
    if (nvert) {
      type <- ids[i, "type"]
      prev <- length(vertices)/3
      inds <- switch(as.character(type),
                     points = seq_len(nvert) + prev,
                     lines = # from segments3d
                       matrix(1:nvert + prev, nrow = 2),
                     linestrip = # from lines3d
                       rbind(seq_len(nvert - 1), 
                             seq_len(nvert - 1) + 1) + prev,
                     triangles =,
                     planes = matrix(1:nvert + prev, nrow = 3),
                     quads = {
                       nquads <- nvert/4
                       matrix(4*rep(seq_len(nquads) - 1, each = 6) + c(1,2,3,1,3,4) + prev, nrow = 3)
                     },
                     surface = {
                       dim <- rgl.attrib(id, "dim")
                       ul <- rep(2:dim[1], dim[2]-1) + dim[1]*rep(0:(dim[2]-2), each=dim[1]-1) + prev
                       if (rgl.attrib(id, "flags")["flipped",])
                         rbind(c(ul-1, ul-1+dim[1]),
                                 c(ul, ul),
                                 c(ul-1+dim[1], ul+dim[1]))
                       else
                         rbind(c(ul, ul),
                                 c(ul-1, ul-1+dim[1]),
                                 c(ul-1+dim[1], ul+dim[1]))
                     },
                     NULL)
      if (length(inds)) {
        if (type == "points") {
          ip <- c(ip, inds)
          normals <- rbind(normals, matrix(NA, ncol = 3, nrow = nvert))
        } else if (type %in% c("lines", "linestrip")) {
          is <- cbind(is, inds)
          normals <- rbind(normals, matrix(NA, ncol = 3, nrow = nvert))
        } else {
          it <- cbind(it, inds)
          normals <- rbind(normals, rgl.attrib(id, "normals"))
        }  
        vertices <- cbind(vertices, local_t(rgl.attrib(id, "vertices")))
        if (rgl.attrib.count(id,"texcoords"))
          texcoords <- rbind(texcoords, rgl.attrib(id, "texcoords"))
        else
          texcoords <- rbind(texcoords, matrix(NA, ncol = 2, nrow = nvert))
        mat <- rgl.getmaterial(nvert, id = id)
        
        material <- mergeMaterials(material, mat, nvert, type)
      }
    }
  }
  if (length(vertices)) {
    if (length(unique(material$color)) == 1)
      material$color <- material$color[1]
    if (length(unique(material$alpha)) == 1)
      material$alpha <- material$alpha[1]
    meshColor <- material$meshColor
    material$meshColor <- NULL
    colorByFaces(mesh3d(vertices = vertices, 
                        points = ip, segments = is, triangles = it, 
                        material = material, normals = normals, 
                        texcoords = if (any(!is.na(texcoords))) texcoords, 
                        meshColor = meshColor))
  }
}

as.mesh3d.rglobject <- function(x, ...) {
  vertices <- x$vertices
  if (is.null(indices <- x$indices))
    if (x$type == "surface") {
      dim <- x$dim
      ul <- rep(2:dim[1], dim[2]-1) + dim[1]*rep(0:(dim[2]-2), each=dim[1]-1)
      if (x$flipped)
        indices <- rbind(c(ul-1, ul-1+dim[1]),
                         c(ul, ul),
                         c(ul-1+dim[1], ul+dim[1]))
      else
        indices <- rbind(c(ul, ul),
                         c(ul-1, ul-1+dim[1]),
                         c(ul-1+dim[1], ul+dim[1]))
    } else
      indices <- seq_len(NROW(vertices))
  material <- x$material
  if (length(indices))
    switch(x$type,
      points = mesh3d(x = vertices,
                      normals = x$normals,
                      texcoords = x$texcoords,
                      points = indices,
                      material = x$material),
      linestrip = {
        indices <- as.numeric(rbind(indices[-length(indices)],
                                    indices[-1]))
        mesh3d(x = vertices,
               segments = indices,
               material = x$material)
      }, 
      lines = mesh3d(x = vertices,
                        normals = x$normals,
                        texcoords = x$texcoords,
                        segments = indices,
                        material = x$material),
      surface =,
      triangles = mesh3d(x = vertices,
                         normals = x$normals,
                         texcoords = x$texcoords,
                         triangles = indices,
                         material = x$material),
      quads = mesh3d(x = vertices,
                     normals = x$normals,
                     texcoords = x$texcoords,
                     quads = indices,
                     material = x$material),
        NULL
      )
  else
    NULL
}

as.mesh3d.rglsubscene <- function(x, rglscene, transform = diag(4), simplify = TRUE, ...) {
  outmeshes <- list()
  objects <- rglscene$objects
  for (i in x$objects) {
    m <- as.mesh3d(objects[[as.character(i)]])
    if (!is.null(m))
      outmeshes <- c(outmeshes, list(rotate3d(m, matrix = transform)))
  }
  for (i in seq_along(x$subscenes)) {
    child <- x$subscenes[[i]]
    outmeshes <- c(outmeshes, as.mesh3d(child, rglscene, transform %*% child$par3d$userMatrix, simplify = FALSE))
  }
  if (simplify && length(outmeshes) == 1)
    outmeshes[[1]]
  else if (length(outmeshes))
    shapelist3d(outmeshes, plot = FALSE)
  else
    NULL
}

as.mesh3d.rglscene <- function(x, ...) {
  result <- as.mesh3d(x$rootSubscene, rglscene = x, ...)
}

mergeVertices <- function(mesh, notEqual = NULL, attribute = "vertices",
                          tolerance = sqrt(.Machine$double.eps)) {

  dropNormals <- function(mesh) {
    if (is.null(mesh$normals))
      return(FALSE)
    
    normals <- mesh$normals
    v <- with(mesh, rbind(vb[1,]/vb[4,], vb[2,]/vb[4,], vb[3,]/vb[4,]))
    it <- mesh$it
    if (!is.null(it))
      for (i in seq_len(ncol(it))) {
        normal <- xprod( v[, it[1, i]] - v[, it[3, i]], 
                         v[, it[2, i]] - v[, it[1, i]])
        normal <- normal/sqrt(sum(normal^2))
        for (j in 1:3) {
          normij <- normals[, it[j, i]]
          normij <- normij/sqrt(sum(normij^2))
          same <- all.equal(normal, normij, tolerance = tolerance,
                            check.attributes = FALSE)
          if (!isTRUE(same)) 
            return(FALSE)
        }
      }
    ib <- mesh$ib
    if (!is.null(ib)) 
      for (i in seq_len(ncol(ib))) {
        norm1 <- xprod( v[, it[1, i]] - v[, it[3, i]], 
                        v[, it[2, i]] - v[, it[1, i]])
        norm1 <- norm1/sqrt(sum(norm1^2))
        norm2 <- xprod( v[, it[2, i]] - v[, it[4, i]], 
                        v[, it[3, i]] - v[, it[2, i]])
        norm2 <- norm1/sqrt(sum(norm2^2))
        for (j in 1:4) {
          normij <- normals[, it[j, i]]
          normij <- normij/sqrt(sum(normij^2))
          same1 <- all.equal(norm1, normij, tolerance = tolerance,
                             check.attributes = FALSE)
          same2 <- all.equal(norm2, normij, tolerance = tolerance,
                             check.attributes = FALSE)
          if (!isTRUE(same1) || !isTRUE(same2)) 
            return(FALSE)
        }
      }      
    TRUE
  }
  
  if (dropNormals(mesh))
    mesh$normals <- NULL
  
  if (is.null(mesh$vb))
    return(mesh)
  nvert <- ncol(mesh$vb)
  if (!is.null(notEqual)) {
    dim <- dim(notEqual)
    if (length(dim) != 2 || dim[1] != nvert || dim[2] != nvert)
      stop("'notEqual' should be a ", nvert, " by ", nvert, " logical matrix.")
    notEqual <- notEqual | t(notEqual) # Make it symmetric
  } else
    notEqual <- matrix(FALSE, nvert, nvert)
  attribute <- match.arg(attribute, 
                          choices = c("vertices", "colors", "normals", "texcoords"),
                          several.ok = TRUE)
  
  
  verts <- matrix(numeric(), ncol = 0, nrow = nvert)
  if ("vertices" %in% attribute) 
    verts <- cbind(mesh$vb[1,]/mesh$vb[4,],
                   mesh$vb[2,]/mesh$vb[4,],
                   mesh$vb[3,]/mesh$vb[4,])
  if (!is.null(normals <- mesh$normals) && "normals" %in% attribute)
    if (nrow(normals) == 3)
      verts <- cbind(verts, t(normals))
    else
      verts <- cbind(verts, normals[1,]/normals[4,], 
                            normals[2,]/normals[4,],
                            normals[3,]/normals[4,])
  colors <- NULL
  if ("colors" %in% attribute &&
      !is.null(mesh$meshColor) && mesh$meshColor == "vertices" &&
      !is.null(mesh$material) && !is.null(colors <- mesh$material$color))
      verts <- cbind(verts, t(col2rgb(colors)))
  
  if (!is.null(texcoords <- mesh$texcoords) && "texcoords" %in% attribute)
    verts <- cbind(verts, t(texcoords))
  
  o <- do.call(order, as.data.frame(verts))
  indices <- c(mesh$ip, mesh$is, mesh$it, mesh$ib)
  i1 <- seq_len(nvert)[o]
  for (i in seq_len(nvert)[-1]) {
    if (isTRUE(all.equal(verts[i1[i-1],], verts[i1[i],], tolerance = tolerance))
        && (!isTRUE(notEqual[i1[i-1], i1[i]]))) {
      notEqual[i1[i], ] <- notEqual[i1[i-1], ] <- notEqual[i1[i], ] | notEqual[i1[i-1], ]
      notEqual[, i1[i]] <- notEqual[, i1[i-1]] <- notEqual[i1[i], ] | notEqual[, i1[i-1]]
      i1[i] <- i1[i-1]
    }
  }
  indices <- i1[order(o)][indices]
  keep <- sort(unique(i1))
  newvals <- numeric(nvert)
  newvals[keep] <- seq_along(keep)
  indices <- newvals[indices]
  mesh$vb <- mesh$vb[,keep]
  if (!is.null(normals))
    mesh$normals <- normals[, keep]
  if (!is.null(texcoords))
    mesh$texcoords <- texcoords[, keep]
  if (!is.null(colors))
    mesh$material$color <- colors[keep]
  
  if (np <- length(mesh$ip))
    mesh$ip <- indices[seq_len(np)]
  if (ns <- length(mesh$is))
    mesh$is <- matrix(indices[seq_len(ns) + np])
  if (nt <- length(mesh$it)) 
    mesh$it <- matrix(indices[seq_len(nt) + np + ns], nrow = 3)
  if (nq <- length(mesh$ib)) 
    mesh$ib <- matrix(indices[seq_len(nq) + np + ns + nt], nrow = 4)
  mesh
}

##
## R source file
## This file is part of rgl
##
##

##
## ===[ SECTION: package entry/exit point ]===================================
##

##
## entry-point
##
##
  
.onLoad <- function(lib, pkg) {

  in_pkgload_loadall <- function() {
    caller <- deparse(sys.call(-4))
    isNamespaceLoaded("pkgload") && any(grepl("load_all", caller))
  }
  
  getDir <- function(useNULL) {
    if (in_pkgload_loadall()) {
      dir <- if (useNULL) file.path("inst", "useNULL") else "src"
    } else {
      dir <- if (useNULL) "useNULL" else "libs"
      if (nchar(.Platform$r_arch))
        dir <- file.path(dir, .Platform$r_arch)
    }
    dir
  }
    
  getDynlib <- function(dir)
    system.file(dir, paste0("rgl", .Platform$dynlib.ext), 
                package = pkg, lib.loc = lib,
                mustWork = TRUE)
  
  is_tahoe <- function() {
    mac_version_string <- system("sw_vers -productVersion", intern = TRUE)
    mac_version <- as.numeric_version(mac_version_string)
    tahoe_version <- as.numeric_version("26.0")
    mac_version >= tahoe_version
  }
  
  # OS-specific 
  initValue <- 0  
  
  onlyNULL <- noOpenGL || rgl.useNULL()
  
  useNULL <- onlyNULL && !noOpenGL && .Platform$OS.type != "windows"
  dir <- getDir(useNULL)
  
  unixos <- "none"
  if (.Platform$OS.type == "unix")
    unixos <- Sys.info()["sysname"]
  
  dll <- try(dyn.load(dynlib <- getDynlib(dir)))
  if (inherits(dll, "try-error")) {
    if (unixos == "Darwin" && !onlyNULL)
      warning(paste("\tLoading rgl's DLL failed.", 
    	       if (unixos == "Darwin" && !onlyNULL) {
    	         paste(if (is_tahoe()) "\n\tYou appear to be running macOS Tahoe or newer.
    	               Tahoe 26.1 is known not to support OpenGL." else "\n\tThis build of rgl depends on XQuartz, which failed to load.\n",
    	           "See the discussion in https://stackoverflow.com/a/66127391")
             }),
         call. = FALSE)
    if (!onlyNULL) {
      dir <- getDir(TRUE)
      warning("Trying without OpenGL...", call. = FALSE)
      noOpenGL <<- TRUE
      dll <- try(dyn.load(dynlib <- getDynlib(dir)))
    }
    if (inherits(dll, "try-error"))
      stop("Loading failed.")
  }
  routines <- getDLLRegisteredRoutines(dll, addNames = FALSE)
  ns <- asNamespace(pkg)
  for(i in 1:4)
    lapply(routines[[i]],
      function(sym) assign(sym$name, sym, envir = ns))
      
  if ( !noOpenGL && .Platform$OS.type == "windows" && !onlyNULL) {
    frame <- getWindowsHandle("Frame")  # nolint 
    ## getWindowsHandle was numeric pre-2.6.0 
    if ( !is.null(frame) ) initValue <- getWindowsHandle("Console") # nolint
  } 
 
  if (onlyNULL) {
    rglFonts(serif = rep("serif", 4), sans = rep("sans", 4), mono = rep("mono", 4), symbol = rep("symbol", 4))
  } else {
    rglFonts(serif = rep(system.file("fonts", "FreeSerif.ttf", package="rgl"), 4),
             sans  = rep(system.file("fonts", "FreeSans.ttf", package="rgl"), 4),
             mono  = rep(system.file("fonts", "FreeMono.ttf", package="rgl"), 4),
             symbol = rep(system.file("fonts", "FreeSerif.ttf", package="rgl"), 4))
    if (requireNamespace("extrafont", quietly = TRUE))
      suppressWarnings(
        rglExtrafonts(sans = c("Helvetica", "Arial"), 
                      serif = c("Times", "Times New Roman"), 
                      mono = c("Courier", "Courier New")))
  }
  
  register_compare_proxy()
  
  .rglEnv$subsceneList <- NULL
  
  dir <- tempfile("rgl")
  dir.create(dir)
  .rglEnv$textureDir <- normalizePath(dir)

  # Workaround for incompatibility with quartz device
  # By default only run this if we'll be using the X11 display on macOS
  # and we're not on R.app.  options("rgl.startQuartz") can 
  # override this.
  # Then we need to start quartz() before starting rgl.
  # See https://github.com/dmurdoch/rgl/issues/27
  if (getOption("rgl.startQuartz", 
           !onlyNULL &&
           interactive() &&
           unixos == "Darwin" && 
           !(.Platform$GUI %in% c("AQUA", "RStudio"))) &&
         exists("quartz", getNamespace("grDevices"))) {
    grDevices::quartz()
    safe.dev.off()
  }
  
  ret <- rgl.init(initValue, onlyNULL)
  
  if (!ret) {
    warning("'rgl.init' failed, will use the null device.\nSee '?rgl.useNULL' for ways to avoid this warning.", call. = FALSE)
    options(rgl.useNULL = TRUE)
    rgl.init(initValue, TRUE)	
  }

  if (!rgl.useNULL()) 
    setGraphicsDelay(unixos = unixos)
  
  # Are we running in reprex::reprex?  If so, do
  # the knitr setup so our output appears there.
  
  if (in_reprex()) {
    setupKnitr(autoprint = TRUE)
  }
}

# Do we need a delay opening graphics?    
# Work around bug in MacOS Catalina:  if base plotting happens
# too quickly after first call to quartz, R crashes.
# This inserts a delay after the
# first call to the graphics device.  The default is
# no delay, unless on Catalina with no graphics device
# currently open, when a 1 second delay will be introduced.
# Use "RGL_SLOW_DEV = value" to change the delay from 
# the default to "value" seconds.  

setGraphicsDelay <- function(delay = Sys.getenv("RGL_SLOW_DEV", 0), 
                             unixos = "none") {
  if (unixos == "Darwin") {
    version <- try(numeric_version(Sys.info()["release"]))
    if (missing(delay) &&
        !inherits(version, "try-error") && 
        !is.na(version) && 
        version >= "19.0.0" &&
        dev.cur() == 1 &&
        identical(getOption("device"), grDevices::quartz))
      delay <- Sys.getenv("RGL_SLOW_DEV", 1)
  }
  delay <- suppressWarnings(as.numeric(delay))
  if (is.na(delay))
    delay <- 1
  if (delay > 0) {
    olddev <- getOption("device")
    if (is.character(olddev)) {
      if (exists(olddev, globalenv(), mode = "function"))
        olddev <- get(olddev, envir = globalenv(), mode = "function")
      else if (exists(olddev, asNamespace("grDevices"), mode = "function"))
        olddev <- get(olddev, asNamespace("grDevices"), mode = "function")
    }
    if (is.function(olddev))
      options(device = function(...) {
        olddev(...)
        Sys.sleep(delay)
        options(device = olddev)
      })
  }
}

rgl.init <- function(initValue = 0, onlyNULL = FALSE, debug = getOption("rgl.debug", FALSE)) 
  .Call( rgl_init, 
    initValue, onlyNULL, environment(rgl.init), debug )

.onAttach <- function(libname, pkgname) {
  if (noOpenGL)
    packageStartupMessage(
      "This build of rgl does not include OpenGL functions.  Use
 rglwidget() to display results, e.g. via options(rgl.printRglwidget = TRUE).")
}

##
## exit-point
##
##

.onUnload <- function(libpath) {
  
  unregisterShinyHandlers()
  unlink(.rglEnv$textureDir, recursive = TRUE)
  
  # shutdown
  .C( rgl_quit, success=FALSE )
  
}

in_reprex <- function() 
  !is.null(getOption("reprex.current_venue"))

plot3d <- function(x, ...) UseMethod("plot3d")


plot3d.default <- function(x, y = NULL, z = NULL, 
        xlab = NULL, ylab = NULL, zlab = NULL, type = 'p', 
        col = material3d("color")[1], size = material3d("size"), 
        lwd = material3d("lwd"),
        radius = avgscale*size/60, add = FALSE, aspect = !add, 
        xlim = NULL, ylim = NULL, zlim = NULL,
        forceClipregion = FALSE, decorate = !add, ...) {
    if (!add) next3d()
    skip <- par3d(skipRedraw=TRUE)
    on.exit(par3d(skip))
     	
    xlabel <- if (!missing(x)) deparse(substitute(x))
    ylabel <- if (!missing(y)) deparse(substitute(y))
    zlabel <- if (!missing(z)) deparse(substitute(z))
 
    xyz <- xyz.coords(x,y,z, xlab=xlabel, ylab=ylabel, zlab=zlabel, recycle=TRUE)
    x <- xyz$x
    y <- xyz$y
    z <- xyz$z

    if (is.null(xlab)) xlab <- xyz$xlab
    if (is.null(ylab)) ylab <- xyz$ylab
    if (is.null(zlab)) zlab <- xyz$zlab

    if (type == "s" && missing(radius)) {
      xvals <- x
      yvals <- y
      zvals <- z
      if (add && diff(bbox <- par3d("bbox"))[1] > 0) {
        xvals <- c(x, bbox[1:2])
        yvals <- c(y, bbox[3:4])
        zvals <- c(z, bbox[5:6])
      } 
      if (!add) {
        if (!is.null(xlim)) 
          xvals <- xlim
        if (!is.null(ylim))
          yvals <- ylim
        if (!is.null(zlim))
          zvals <- zlim
      }
	    avgscale <- sqrt(sum(c(diff(range(xvals,na.rm=TRUE)), 
                             diff(range(yvals,na.rm=TRUE)), 
                             diff(range(zvals,na.rm=TRUE)))^2/3))
    }
    savesubscene <- currentSubscene3d()
    result <- setClipregion(xlim, ylim, zlim, forceClipregion)
    result <- c(result, data=switch(type,
		p = points3d(x, y, z, color=col, size=size, ...),
	        s = spheres3d(x, y, z, radius=radius, color=col, ...),
		l = lines3d(x, y, z, color=col, lwd=lwd, ...),
		h = segments3d(rep(x,rep(2,length(x))),
					   rep(y,rep(2,length(y))),
					   rbind(rep(0,length(z)),z),
					   color = rep(col, rep(2,length(col))), lwd=lwd, ...),
	# this is a hack to plot invisible segments
        n = if (!add) segments3d(rep(range(x, na.rm=TRUE), c(2,2)),
                                 rep(range(y, na.rm=TRUE), c(2,2)),
                                 rep(range(z, na.rm=TRUE), c(2,2))))
	)
    useSubscene3d(savesubscene)
    if (decorate)
      result <- c(result, decorate3d(xlab=xlab, ylab=ylab, zlab=zlab, aspect = aspect, 
                               xlim=xlim, ylim=ylim, zlim=zlim, ...))
    if (!add)
    	highlevel(result)
    else 
    	lowlevel(result)
}

plot3d.mesh3d <- function(x, xlab = "x", ylab = "y", zlab = "z", type = c("shade", "wire", "dots"),
	add = FALSE, aspect = !add, ...) {
    if (!add) next3d()
    skip <- par3d(skipRedraw=TRUE)
    on.exit(par3d(skip))
    
    if (missing(xlab) && !is.null(x$xlab)) xlab <- x$xlab
    if (missing(ylab) && !is.null(x$ylab)) ylab <- x$ylab
    if (missing(zlab) && !is.null(x$zlab)) zlab <- x$zlab
    
    result <- switch(match.arg(type),
    	shade = shade3d(x, ...),
    	wire = wire3d(x, ...),
    	dots = dot3d(x, ...))
    
    if (!add) {
    	result <- c(result, decorate3d(xlab = xlab, ylab = ylab, zlab = zlab, aspect = aspect,
    	                               ...))
    	highlevel(result)
    } else 
    	lowlevel(result)
}

decorate3d <- function(xlim = NULL, ylim = NULL, zlim = NULL, 
                       xlab = "x", ylab = "y", zlab = "z", 
                       box = TRUE, axes = TRUE, main = NULL, sub = NULL,
                       top = TRUE, aspect = FALSE, expand = 1.03, tag = material3d("tag"), ...) {
  
  if (is.logical(aspect)) {
    autoscale <- aspect
    aspect <- c(1,1,1)
  } else autoscale <- TRUE	
  
  result <- numeric(0)    
  if (length(c(xlim, ylim, zlim))) {
    ranges <- .getRanges()        
    if (is.null(xlim))
      xlim <- ranges$xlim
    if (is.null(ylim))
      ylim <- ranges$ylim
    if (is.null(zlim))
      zlim <- ranges$zlim
    ind <- c(1,1,2,2)
    result <- c(result, strut=segments3d(xlim[ind], ylim[ind], zlim[ind], tag = tag))
  }
  
  if (autoscale) aspect3d(aspect)
  
  if (axes) result <- c(result, axes=axes3d(box=box, expand=expand, tag = tag))
  result <- c(result, title3d(xlab = xlab, ylab = ylab, zlab = zlab, 
                              main = main, sub = sub, tag = tag))
  
  if (top) rgl.bringtotop()
  
  lowlevel(result)
}

plot3d.function <- function(x, ...) persp3d(x, ...)

plot3d.deldir <- function(x, ...) persp3d(x, ...)

plot3d.triSht <-
plot3d.tri <- function(x, z, ...) persp3d(x, z, ...)

plot3d.formula <- function(x, data = NULL, xlab = xyz$xlab, ylab = xyz$ylab, zlab = xyz$zlab, ...) {
  if (!is.null(data))
    environment(x) <- list2env(data, envir = environment(x))
  xyz <- xyz.coords(x)
  plot3d(xyz, xlab = xlab, ylab = ylab, zlab = zlab, ...)
}

plot3d.lm <- function(x, which = 1, 
                      plane.col = "gray", plane.alpha = 0.5,
                      sharedMouse = TRUE,
                      use_surface3d, do_grid = TRUE,
                      grid.col = "black", grid.alpha = 1,
                      grid.steps = 5,
                      sub.steps = 4,
                      vars = get_all_vars(terms(x), x$model),
                      clip_to_density = 0,
                      ...) {
  stopifnot(which %in% 1:3)
  dots <- list(...)
  n <- length(which)
  if (clip_to_density > 0) {
    if (!requireNamespace("MASS", quietly = TRUE)) {
      warning("'clip_to_density' requires the MASS package.")
      clip_to_density <- 0
    }
    if (!requireNamespace("interp", quietly = TRUE)) {
      warning("'clip_to_density' requires the interp package.")
      clip_to_density <- 0
    }
  }
  result <- NULL
  if (n > 1) {
    cols <- ceiling(sqrt(n))
    rows <- ceiling(n/cols)
    mfrow3d(rows, cols, sharedMouse = sharedMouse)
  }
  fit <- x
  missing_vars <- missing(vars)
  cols <- ncol(vars)
  if (cols < 3)
    stop("Model has only ", cols, " variables.")
  if (cols > 3)
    warning("Model has ", cols, " variables; first 3 used.")
  observed <- vars[, c(2, 3, 1)]
  names <- colnames(observed)
  if (is.null(dots$xlab)) dots$xlab <- names[1]
  if (is.null(dots$ylab)) dots$ylab <- names[2]
  if (is.null(dots$zlab)) dots$zlab <- names[3]

  if (missing(use_surface3d)) 
    use_surface3d <- !identical(class(fit), "lm") || ncol(as.matrix(model.frame(fit))) > 3
  
  if (clip_to_density > 0) {
    densityVals <- MASS::kde2d(observed[,1], observed[,2])
    densityVals$z <- with(densityVals, z/max(z)) # nolint
    density <- function(xyz) {
      with(densityVals, interp::bilinear(x, y, z, xyz[,1], xyz[,2])$z) # nolint
    }
  }
  
  plotGrid <- function(i, x, y, x0, y0, z) {
    dots$color <- grid.col
    dots$alpha <- grid.alpha
    dots$color <- grid.col
    dots$alpha <- grid.alpha
    dots$front <- dots$back <- dots$polygon_offset <- 
      dots$type <- NULL

    lenx <- length(x)
    lenx0 <- length(x0)
    leny <- length(y)
    leny0 <- length(y0)
    
    x <- c(rep(x0, each = leny + 1),
           rep(c(x, NA), leny0))
    y <- c(rep(c(y, NA), lenx0),
           rep(y0, each = lenx + 1))
    if (missing(z)) {
      newdat <- data.frame(x, y)
      names(newdat) <- names[1:2]
      z <- predict(fit, newdata = newdat)
    }
    grid <- do.call(lines3d, c(list(x, y, z), dots))
    if (clip_to_density > 0)
      grid <- clipObj3d(grid, density, clip_to_density,
                        minVertices = 1000)
    names(grid) <- paste0("grid.", i)
    grid
  }
  
  plotPoints <- function(i, points, zlab) {
    dots$zlab <- zlab
    plot <- do.call(plot3d, c(list(x = points), dots))
    names(plot) <- paste0(names(plot), ".", i)
    plot
  }
  
  plotSurface <- function(i) {
    bbox <- par3d("bbox")
    xlim <- c(bbox[1], bbox[2])
    x0 <- pretty(xlim, grid.steps)
    ylim <- c(bbox[3], bbox[4])
    y0 <- pretty(ylim, grid.steps)
    if (sub.steps > 1) {
      x <- rep(x0, each = sub.steps) + 
           seq(0, diff(x0[1:2]), length.out = sub.steps + 1)[-(sub.steps + 1)]
      y <- rep(y0, each = sub.steps) + 
        seq(0, diff(y0[1:2]), length.out = sub.steps + 1)[-(sub.steps + 1)]
    } else {
      x <- x0
      y <- y0
    }
    x <- c(xlim[1], x[xlim[1] < x & x < xlim[2]], xlim[2])
    y <- c(ylim[1], y[ylim[1] < y & y < ylim[2]], ylim[2])
    newdat <- expand.grid(x = x, y = y)
    names(newdat) <- names[1:2]
    z <- try(matrix(predict(fit, newdat), length(x), length(y)))
    if (inherits(z, "try-error") && !missing_vars) {
      stop("vars should be in order: response, pred1, pred2", call. = FALSE)
    }
    dots$color <- plane.col
    dots$alpha <- plane.alpha
    if (is.null(dots$polygon_offset)) 
      dots$polygon_offset <- 1
    dots$type <- NULL
    surface <- do.call("surface3d", c(list(x, y, z), dots))
    if (clip_to_density > 0)
      surface <- clipObj3d(surface, density, clip_to_density,
                        minVertices = 1000)
    names(surface) <- paste0("surface.", i)
    grid <- if (do_grid) {
      x0 <- x0[xlim[1] <= x0 & x0 <= xlim[2]]
      y0 <- y0[ylim[1] <= y0 & y0 <= ylim[2]]
      plotGrid(i, x, y, x0, y0)
    }
    c(surface, grid)
  }
  plotPlane <- function(i, a, b, d) {
    c <- -1
    if (is.na(d))
      d <- 0
    dots$color <- plane.col
    dots$alpha <- plane.alpha
    if (is.null(dots$polygon_offset)) 
      dots$polygon_offset <- 1
    dots$type <- NULL
    plane <- do.call("planes3d", c(list(a = a, b = b, c = c, d = d), dots))
    if (clip_to_density > 0)
      plane <- clipObj3d(plane, density, clip_to_density,
                        minVertices = 1000)
    names(plane) <- paste0("plane.", i)
    grid <- if (do_grid) {
      bbox <- par3d("bbox")
      xlim <- c(bbox[1], bbox[2])
      x0 <- pretty(xlim, grid.steps)
      ylim <- c(bbox[3], bbox[4])
      y0 <- pretty(ylim, grid.steps)
      x0 <- x0[xlim[1] <= x0 & x0 <= xlim[2]]
      y0 <- y0[ylim[1] <= y0 & y0 <= ylim[2]]
      if (isTRUE(all.equal(c(a,b,d), c(0, 0, 0))))
        plotGrid(i, xlim, ylim, x0, y0, z = 0)
      else
        plotGrid(i, xlim, ylim, x0, y0)      
    }
    c(plane, grid)
  }
  for (i in seq_along(which)) {
    type <- which[i]
    if (type == 1L) {
      plot <- plotPoints(i, observed, dots$zlab)
      if (use_surface3d) {
        plane <- plotSurface(i)
      } else {
        coefs <- coef(fit)
        plane <- plotPlane(i, coefs[names[1]], coefs[names[2]], coefs["(Intercept)"])
      }
    } else if (type == 2L) {
      plot <- plotPoints(i, cbind(observed[,1:2], residuals(fit)), "Residuals")
      plane <- plotPlane(i, 0, 0, 0)
    } else if (type == 3L) {
      plot <- plotPoints(i, cbind(observed[,1:2], predict(fit)), dots$zlab)
      if (use_surface3d) {
        plane <- plotSurface(i)
      } else {
        coefs <- coef(fit)
        plane <- plotPlane(i, coefs[names[1]], coefs[names[2]], coefs["(Intercept)"])
      } 
    }
    result <- c(result, plot, plane)
  }
  highlevel(result)
}

tkspinControl <- function(base, dev = cur3d(), continue=FALSE,
                          speed=30, scale=100, ...) {
  
  if (!requireNamespace("tcltk", quietly = TRUE))
    stop("This function requires 'tcltk'.")
  
  slider <- tcltk::tclVar(speed)
  getZooms <- function() {
    old <- cur3d()
    on.exit(set3d(old))
    result <- numeric(max(dev))
    for (device in dev) {
      set3d(device)
      result[device] <- par3d("zoom")
    }
    result
  }
  zooms <- getZooms()
  
  scale <- tcltk::tclVar(scale)
  continuous <- tcltk::tclVar(as.numeric(continue))
  
  buttonPress <- NULL
  direction <- NULL
  
  lastTime <- proc.time()[3]
  timeDiff <- 0
  
  rotateUp <- function() {
    angle <- timeDiff*as.numeric(tcltk::tclObj(slider))*pi/180
    par3d(userMatrix = rotationMatrix(-angle, 1,0,0) %*% par3d("userMatrix"))
  }
  
  rotateLeft <- function() {
    angle <- timeDiff*as.numeric(tcltk::tclObj(slider))*pi/180
    par3d(userMatrix = rotationMatrix(-angle, 0,1,0) %*% par3d("userMatrix"))
  }
  
  rotateRight <- function() {
    angle <- timeDiff*as.numeric(tcltk::tclObj(slider))*pi/180
    par3d(userMatrix = rotationMatrix(angle, 0,1,0) %*% par3d("userMatrix"))
  }
  
  rotateSpin <- function() {
    angle <- timeDiff*as.numeric(tcltk::tclObj(slider))*pi/180
    par3d(userMatrix = rotationMatrix(-angle, 0,0,1) %*% par3d("userMatrix"))
  }
  
  rotateDown <- function() {
    angle <- timeDiff*as.numeric(tcltk::tclObj(slider))*pi/180
    par3d(userMatrix = rotationMatrix(angle, 1,0,0) %*% par3d("userMatrix"))
  }	
  
  rotate <- function() {
    old <- cur3d()
    on.exit(set3d(old))	
    if (buttonPress) {
      if ((currentTime <- proc.time()[3]) > lastTime) {
        timeDiff <<- currentTime - lastTime
        lastTime <<- currentTime
        for (device in dev) {
          set3d(device)
          if (direction == "up")
            rotateUp()
          if (direction == "left")
            rotateLeft()
          if (direction == "spin")
            rotateSpin()
          if (direction == "right")
            rotateRight()
          if (direction == "down")
            rotateDown()
        }
      }
      tcltk::tcl("after",5,rotate)
    }
    
  }
  
  # rotation button callback functions
  # note that "..." argument is necessary
  upButtonPress <- function(...) {
    buttonPress <<- TRUE
    lastTime <<- proc.time()[3]
    direction <<- "up"
    rotate()
  }
  
  leftButtonPress <- function(...) {
    buttonPress <<- TRUE
    lastTime <<- proc.time()[3]
    direction <<- "left"
    rotate()
  }
  
  spinButtonPress <- function(...) {
    buttonPress <<- TRUE
    lastTime <<- proc.time()[3]
    direction <<- "spin"
    rotate()
  }
  
  rightButtonPress <- function(...) {
    buttonPress <<- TRUE
    lastTime <<- proc.time()[3]
    direction <<- "right"
    rotate()
  }
  
  downButtonPress <- function(...) {
    buttonPress <<- TRUE
    lastTime <<- proc.time()[3]
    direction <<- "down"
    rotate()
  }
  
  onIdle <- function(...) {
    buttonPress <<- TRUE
    rotate()
    buttonPress <<- FALSE
    if (as.numeric(tcltk::tclObj(continuous)))
      tcltk::tcl("after", "idle", onIdle)
  }
  
  buttonRelease <- function(...) {
    buttonPress <<- FALSE
    if (as.numeric(tcltk::tclObj(continuous)))
      tcltk::tcl("after", "idle", onIdle)
  }
  
  
  resetAxis <- function(...) {
    old <- cur3d()
    on.exit(set3d(old))
    for (device in dev) {
      set3d(device)
      par3d(userMatrix = diag(4))
    }
  }
  
  setScale <- function(...) {
    old <- cur3d()
    on.exit(set3d(old))		    
    scale <- as.numeric(tcltk::tclObj(scale))
    for (device in dev) {
      set3d(device)
      par3d(zoom = 10^((scale - 100)/50)*zooms[device])
    }
  }
  
  spec.frm <- tcltk::tkframe(base, ...)
  first.frm <- tcltk::tkframe(spec.frm)
  second.frm <- tcltk::tkframe(spec.frm)
  third.frm <- tcltk::tkframe(spec.frm)
  fourth.frm <- tcltk::tkframe(spec.frm)
  
  # rotations buttons
  upButton <- tcltk::tkbutton(first.frm, text="  ^  ")
  leftButton <- tcltk::tkbutton(first.frm, text="  <  ")
  spinButton <- tcltk::tkbutton(first.frm, text="  O  ")
  rightButton <- tcltk::tkbutton(first.frm, text="  >  ")
  downButton <- tcltk::tkbutton(first.frm, text="  v  ")
  
  
  tcltk::tkgrid(tcltk::tklabel(first.frm, text=" "), upButton,	tcltk::tklabel(first.frm, text=" "))
  tcltk::tkgrid(leftButton,spinButton,rightButton)
  tcltk::tkgrid(tcltk::tklabel(first.frm, text=" "), downButton, tcltk::tklabel(first.frm, text=" "))
  
  tcltk::tkbind(upButton, "<Button-1>", upButtonPress)
  tcltk::tkbind(leftButton, "<Button-1>", leftButtonPress)
  tcltk::tkbind(spinButton, "<Button-1>", spinButtonPress)
  tcltk::tkbind(rightButton, "<Button-1>", rightButtonPress)
  tcltk::tkbind(downButton, "<Button-1>", downButtonPress)
  
  tcltk::tkbind(upButton,"<ButtonRelease-1>", buttonRelease)
  tcltk::tkbind(leftButton,"<ButtonRelease-1>", buttonRelease)
  tcltk::tkbind(spinButton,"<ButtonRelease-1>", buttonRelease)
  tcltk::tkbind(rightButton,"<ButtonRelease-1>", buttonRelease)
  tcltk::tkbind(downButton,"<ButtonRelease-1>", buttonRelease)
  
  # control buttons
  frameAxis <- tcltk::tkframe(second.frm,borderwidth=2)
  tcltk::tkpack(frameAxis)
  
  buttonAxis <- tcltk::tkbutton(frameAxis,text="     Reset Axis     ",command=resetAxis)
  contBox <- tcltk::tkcheckbutton(frameAxis,text="Continue rotating", variable=continuous)
  
  tcltk::tkpack(contBox, buttonAxis,fill="x")
  
  #control scale frame
  frameControl <- tcltk::tkframe(third.frm,borderwidth=4)
  tcltk::tkpack(frameControl)
  frameControlSpeed <- tcltk::tkframe(frameControl)
  sliderSpeed <- tcltk::tkscale(frameControlSpeed, showvalue=FALSE,
                                orient="horiz", from=0, to=400,resolution=1,variable=slider)
  tcltk::tkpack(tcltk::tklabel(frameControlSpeed,text="Speed:"),sliderSpeed,side="left")
  
  frameControlScale <- tcltk::tkframe(frameControl)
  sliderScale <- tcltk::tkscale(frameControlScale,showvalue=FALSE,orient="horiz",
                                from=0, to=200,resolution=1, variable=scale, command=setScale)
  tcltk::tkpack(tcltk::tklabel(frameControlScale,text="Scale:"),sliderScale,side="left")
  tcltk::tkpack(frameControlSpeed,frameControlScale)
  
  
  tcltk::tkpack(first.frm, second.frm, third.frm, fourth.frm)
  tcltk::tkpack(spec.frm,expand=TRUE)
  
  spec.frm
}


tkspin3d <- function(dev = cur3d(), ...) {
  
  if (!requireNamespace("tcltk", quietly = TRUE))
    stop("This function requires 'tcltk'.")
  
  args <- list(...)
  controlargs <- list()
  if (length(args)) {
    controlargindices <- which(names(args) %in% names(formals(tkspinControl)))
    if (length(controlargindices)) {
      args <- args[-controlargindices]
      controlargs <- args[controlargindices]
    }
  }
  
  base <- do.call(tcltk::tktoplevel, args)
  tcltk::tkwm.title(base, "Spin")
  
  spin <- do.call(tkspinControl, c(list(base = base, dev = dev), controlargs))		
  
  quit <- tcltk::tkbutton(base,text="Quit", command=function()tcltk::tkdestroy(base))
  
  tcltk::tkpack(spin, quit)
  
}

turn3d <- function(x, y = NULL, n = 12, smooth = FALSE, ...) {
  xy <- xy.coords(x, y)
  x <- xy$x
  y <- zapsmall(xy$y)
  stopifnot(all(y >= 0))
  len <- length(x)
  inds <- seq_len(len)
  
  if (smooth) {
    nx <- -diff(y)
    nx <- c(nx, 0) + c(0, nx)
    ny <- diff(x)
    ny <- c(ny, 0) + c(0, ny)
    nlen <- sqrt(nx^2 + ny^2)
    nlen[nlen == 0] <- 1
    nx <- nx/nlen
    ny <- ny/nlen
    normals <- matrix(nrow=4, ncol=0)
  }
  zero <- y == 0
  vb <- matrix(nrow=4, ncol=0)
  ib <- matrix(nrow=4, ncol=0)
  it <- matrix(nrow=3, ncol=0)
  
  theta <- seq(0, 2*pi, length.out = n + 1)[-(n + 1)]
  for (i in inds) {
    vb <- cbind(vb, rbind(x[i], sin(theta)*y[i], cos(theta)*y[i], 1))
    if (smooth)
      normals <- cbind(normals, rbind(nx[i], sin(theta)*ny[i], cos(theta)*ny[i], 1))
    if (i > 1) {
      if (zero[i] && zero[i-1]) { # do nothing 
      } else if (!zero[i] && zero[i-1]) { # draw triangles
        prev <- ncol(vb) - n - 0:(n-1)
        curr <- ncol(vb) - 0:(n-1)
        curr2 <- curr + 1
        curr2[1] <- curr2[1] - n
        
        it <- cbind(it, rbind(prev, curr, curr2))
      } else if (zero[i] && !zero[i-1]) { # other triangles
        prev <- ncol(vb) - n - 0:(n-1)
        curr <- ncol(vb) - 0:(n-1)
        prev2 <- prev + 1
        prev2[1] <- prev2[1] - n
        it <- cbind(it, rbind(prev, curr, prev2))
      } else { # quads
        prev <- ncol(vb) - n - 0:(n-1)
        prev2 <- prev + 1
        prev2[1] <- prev2[1] - n
        curr <- ncol(vb) - 0:(n-1)
        curr2 <- curr + 1
        curr2[1] <- curr2[1] - n
        ib <- cbind(ib, rbind(prev, curr, curr2, prev2))
      }
    }
  }
  
  result <- tmesh3d(vb, it, normals=if(smooth) t(normals), ...)
  if (length(ib)) result$ib <- ib
  result
}

# demo: regression
# author: Daniel Adler

rgl.demo.regression <- function(n=100,xa=3,za=8,xb=0.02,zb=0.01,xlim=c(0,100),zlim=c(0,100)) {

  rgl.clear("all")
  rgl.bg(sphere = TRUE, color = c("black", "green"), lit = FALSE, size=2, alpha=0.2, back = "lines")
  rgl.light()
  rgl.bbox()

  x  <- runif(n,min=xlim[1],max=xlim[2])
  z  <- runif(n,min=zlim[1],max=zlim[2])
  ex <- rnorm(n,sd=3)
  ez <- rnorm(n,sd=2)
  
  esty  <- (xa+xb*x) * (za+zb*z) + ex + ez

  rgl.spheres(x,esty,z,color="gray",radius=1.5,specular="green",
              texture=system.file("textures/bump_dust.png",package="rgl"),
              texmipmap=TRUE, texminfilter="linear.mipmap.linear")

  regx  <- seq(xlim[1],xlim[2],len=100)
  regz  <- seq(zlim[1],zlim[2],len=100)
  regy  <- (xa+regx*xb) %*% t(za+regz*zb)
  
  rgl.surface(regx,regz,regy,color="blue",alpha=0.5,shininess=128)

  lx <- c(xlim[1],xlim[2],xlim[2],xlim[1])
  lz <- c(zlim[1],zlim[1],zlim[2],zlim[2])
  f <- function(x,z) (xa+x*xb) * t(za+z*zb)
  ly <- f(lx,lz)
  
  rgl.quads(lx,ly,lz,color="red",size=5,front="lines",back="lines",lit=FALSE)
}

rgl.open()
rgl.demo.regression()

safe.dev.off  <- function(which = dev.cur(), prev = dev.prev()) {
  if (which != 1) {
    force(prev)
    grDevices::dev.off(which)
  }
  if (length(dev.list()))
    dev.set(prev)
  else
    c("null device" = 1)
}

dev.off <- function(...) stop("Use safe.dev.off() instead!")
  
legend3d <- function(...) {
  args <- list(...)
  bgargs <- setdiff(names(formals(bgplot3d)),
                    c("expression", "..."))
  idx <- which(names(args) %in% c("sphere", "fogtype", bgargs))
  if (length(idx)) {
    bgargs <- args[idx]
    args <- args[-idx]
  } else
    bgargs <- NULL
  do.call(bgplot3d, c(list(quote({
    par(mar=c(0,0,0,0))
    plot(0,0, type="n", xlim=0:1, ylim=0:1, xaxs="i", yaxs="i", axes=FALSE, bty="n")
    do.call(legend, args)
  })), bgargs))
}

bgplot3d <- function(expression, bg.color = getr3dDefaults("bg", "color"), 
                     magnify = 1, ...) {
  viewport <- par3d("viewport")
  width <- magnify*viewport["width"]
  height <- magnify*viewport["height"]
  if (width > 0 && height > 0) {
    filename <- tempfile(fileext = ".png")
    png(filename = filename, width = width, height = height,
        bg = bg.color)
    grDevices::devAskNewPage(FALSE)
    value <- try(expression)  
    safe.dev.off()
    result <- bg3d(texture = filename, col = "white", lit = FALSE,
                   front = "filled", back = "filled", ...)
  } else {
    value <- NULL
    result <- bg3d(col = bg.color, ...)
  }
  lowlevel(structure(result, value = value))
}

show2d <- function(expression, 
		   face = "z-",
		   line = 0,
		   reverse = FALSE,
		   rotate = 0,
		   x = NULL, y = NULL, z = NULL, 
		   width = 480,
		   height = 480,
		   filename = NULL,
		   ignoreExtent = TRUE,
		   color = "white", specular = "black",
		   lit = FALSE, 
		   texmipmap = TRUE,
		   texminfilter = "linear.mipmap.linear",
		   expand = 1.03,
		   texcoords = matrix(c(0,1,1,0,0,0,1,1), ncol=2),
		   ...) {
	
  save <- par3d(ignoreExtent = ignoreExtent)
  on.exit(par3d(save))
  
  if (is.null(filename)) {
    stopifnot(width > 0, height > 0)
    filename <- tempfile(fileext = ".png")
    png(filename = filename, width=width, height=height)
    value <- try(expression)  
    safe.dev.off()
  } else
    value <- filename
  face <- c(strsplit(face, '')[[1]], '-')[1:2]
  coord <- tolower(face[1])
  lower <- face[2] == '-'
  
  ranges <- .getRanges(expand = expand)
    switch(coord,
      x = {
      	if (is.null(x)) 
      	  x <- with(ranges, if (lower) 
      	  	               x[1] - 0.075*line*diff(x)
      	  	            else
      	  	               x[2] + 0.075*line*diff(x))
      	if (is.null(y))
      	  y <- with(ranges, c(y[1], y[2], y[2], y[1]))
      	if (is.null(z))
      	  z <- with(ranges, c(z[1], z[1], z[2], z[2]))
      },
      y = {
      	if (is.null(x)) 
      		x <- with(ranges, c(x[1], x[2], x[2], x[1]))
      	if (is.null(y))
      		y <- with(ranges, if (lower) 
      			            y[1] - 0.075*line*diff(y)
      			          else
      				    y[2] + 0.075*line*diff(y))
      	if (is.null(z))
      		z <- with(ranges, c(z[1], z[1], z[2], z[2]))
      },      	
      z = {
      	if (is.null(x)) 
      		x <- with(ranges, c(x[1], x[2], x[2], x[1]))
      	if (is.null(y))
      		y <- with(ranges, c(y[1], y[1], y[2], y[2]))
      	if (is.null(z))
      		z <- with(ranges,  if (lower) 
      			             z[1] - 0.075*line*diff(z)
      			           else
      				     z[2] + 0.075*line*diff(z))
      }) 
  x <- cbind(x, y, z)  
  if (nrow(x) != 4) 
    stop("Exactly 4 corners must be specified.")
  
  if (reverse) {
    temp <- x[2,]
    x[2,] <- x[4,]
    x[4,] <- temp
  }
  if (rotate) 
    x <- x[(0:3 + rotate) %% 4 + 1, ]
  
  result <- quads3d(x, texture=filename, texcoords = texcoords,
  		  color = color, lit = lit, texmipmap = texmipmap, 
  		  texminfilter = texminfilter, ...)
  lowlevel(structure(result, value = value, xyz = x, texcoords = texcoords,
            filename = filename))
}

##
## R source file
## This file is part of rgl
##
##

##
## ===[ SECTION: device management ]==========================================
##

rgl.useNULL <- function() {
  if (noOpenGL)
    return(TRUE)
  opt <- getOption("rgl.useNULL", Sys.getenv("RGL_USE_NULL"))
  if (is.logical(opt)) return(opt)
  opt <- as.character(opt)
  if (nchar(opt)) {
    opt <- pmatch(tolower(opt), c("yes", "true"), nomatch=3)
    return(c(TRUE, TRUE, FALSE)[opt])
  }
  FALSE
}

##
## open device
##
##

rgl.open <- function(useNULL = rgl.useNULL()) {

  .Deprecated("open3d")
  
  ret <- .C( rgl_dev_open, success=FALSE, useNULL=useNULL,
             antialias = as.integer(getOption("rgl.antialias", -1L)))

  if (! ret$success)
    stop("'rgl.open' failed")

}


##
## close device
##
##

rgl.close <- function() {

  .Deprecated("close3d")
  
  if (length(hook <- getHook("on.rgl.close"))) {
    if (is.list(hook)) hook <- hook[[1]]  # test is for compatibility with R < 3.0.0
    hook()
  }
  
  ret <- .C( rgl_dev_close, success=FALSE )

  if (! ret$success)
    stop("No device opened")

}

## 
## get all devices
##
##

rgl.dev.list <- function() 
  .Call( rgl_dev_list )


##
## set current device
##
##

rgl.set <- function(which, silent = FALSE) {
  
  .Deprecated("set3d")

  idata <- c( as.integer(which), as.integer(silent) )

  ret <- .C( rgl_dev_setcurrent, 
    success=FALSE, 
    idata
  )

  if (! ret$success)
    stop(gettextf("No device opened with id %s", which), domain = NA)
}



##
## export image
##
##

rgl.snapshot <- function( filename, fmt="png", top=TRUE ) {
  if (top) rgl.bringtotop()
  
  idata <- as.integer(rgl.enum.pixfmt(fmt))
  if (length(filename) != 1)
    stop("filename is length ", length(filename))
  filename <- normalizePath(filename, mustWork = FALSE)
  ret <- .C( rgl_snapshot,
    success=FALSE,
    idata,
    filename
  )

  if (! ret$success)
    warning("'rgl.snapshot' failed")
  
  invisible(filename)
}

##
## export postscript image
##
##

rgl.postscript <- function( filename, fmt="eps", drawText=TRUE ) {
  idata <- as.integer(c(rgl.enum.gl2ps(fmt), as.logical(drawText)))
  if (length(filename) != 1)
    stop("filename is length ", length(filename))
  ret <- .C( rgl_postscript,
    success=FALSE,
    idata,
    normalizePath(filename, mustWork = FALSE, winslash = "/")
  )

  if (! ret$success)
    warning("Postscript conversion failed")
}

##
## read image
##
##

rgl.pixels <- function(component = c("red", "green", "blue"), viewport = par3d("viewport"), top=TRUE ) {
  if (top) rgl.bringtotop()
  
  compnum <- as.integer(sapply(component, rgl.enum.pixelcomponent))
  stopifnot(length(viewport) == 4)
  ll <- as.integer(viewport[1:2])
  stopifnot(all(!is.na(ll)), all(ll >= 0))
  size <- as.integer(viewport[3:4])
  stopifnot(all(!is.na(size), all(size >= 0)))
  result <- array(NA_real_, dim=c(size[1], size[2], length(component)))
  dimnames(result) <- list(NULL, NULL, component)
  if (length(result) > 0)
    for (i in seq_along(compnum)) {
      ret <- .C( rgl_pixels,
        success=FALSE,
        ll, size, compnum[i],
        values = double(size[1]*size[2]))
 
      if (! ret$success)
        warning(gettextf("Error reading component '%s'", component[i]), domain = NA)
      result[,,i] <- ret$values
    }
  if (length(component) > 1) return(result)
  else return(result[,,1])
}

##
## R source file
## This file is part of rgl
##
##

##
## ===[ SECTION: scene management ]===========================================
##


##
## clear scene
##
##

rgl.clear <- function( type = "shapes", subscene = 0 )  {
  .Deprecated("clear3d")
  
  if (is.na(subscene)) 
    subscene <- currentSubscene3d()

  typeid <- rgl.enum.nodetype(type)
  
  userviewpoint <- 4 %in% typeid
  material  <- 5 %in% typeid
  modelviewpoint <- 8 %in% typeid

  drop <- typeid %in% c(4:5, 8)
  typeid <- typeid[!drop]
  type <- names(typeid)
  
  if (subscene == 0) {
    idata <- as.integer(c(length(typeid), typeid))    	
    ret <- .C( rgl_clear, 
      success = FALSE,
      idata
    )$success
  } else {
    sceneids <- ids3d(type=type, subscene = 0)$id
    thisids <- ids3d(type=type, subscene = subscene)$id
    if (length(thisids)) {
      delFromSubscene3d(ids = thisids, subscene = subscene)
      gc3d(protect = setdiff(sceneids, thisids))
    }
    ret <- 1
  }
  
  if ( userviewpoint || modelviewpoint) 
    view3d(type = c("userviewpoint", "modelviewpoint")[c(userviewpoint, modelviewpoint)])
    
  if ( material ) 
    rgl.material0()
  
  if (! ret)
    stop("'rgl_clear' failed")
  
  lowlevel()
}


##
## pop node
##
##

pop3d <- function( type = "shapes", id = 0, tag = NULL) {
  if (!is.null(tag)) {
    if (!missing(id))
      stop("Only one of 'id' and 'tag' should be specified.")
    allids <- ids3d(intersect(c("shapes", "bboxdeco"), type), subscene = 0, tags = TRUE)
    id <- allids$id[allids$tag %in% tag]
  }
  type <- rgl.enum.nodetype(type)
  save <- par3d(skipRedraw = TRUE)
  on.exit(par3d(save))
  for (i in id) {
    idata <- as.integer(c(type, i))

    ret <- .C( rgl_pop,
      success = FALSE,
      idata
    )

    if (! ret$success)
      stop(gettextf("'rgl_pop' failed for id %d", i), domain = NA)
  }
  lowlevel()
}

rgl.pop <- pop3d

ids3d <- function( type = "shapes", subscene = NA, tags = FALSE) {
  type <- c(rgl.enum.nodetype(type), 0)
  if (is.na(subscene)) 
    subscene <- currentSubscene3d()
  
  count <- .C( rgl_id_count, as.integer(type), count = integer(1), subscene = as.integer(subscene))$count
  
  result <- as.data.frame( .C( rgl_ids, as.integer(type), id=integer(count), 
                               type=rep("",count), subscene = as.integer(subscene) )[2:3] )
  
  if (tags) {
    result$tag <- rep("", nrow(result))
    if (NROW(result)) {
      hasmaterial <- !(result$type %in% c("light", "userviewpoint", "background", "modelviewpoint", "subscene"))
      result$tag[hasmaterial] <- vapply(result$id[hasmaterial],
                                        function(id) {
                                          rgl.getmaterial(0, id = id)$tag
                                        }, "")
    }
  }
  result
}

rgl.ids <- ids3d

rgl.attrib.count <- function( id, attrib ) {
  stopifnot(length(attrib) == 1)
  if (is.character(attrib))
    attrib <- rgl.enum.attribtype(attrib)
  
  result <- integer(length(id))
  for (i in seq_along(id))
    result[i] <- .C( rgl_attrib_count, as.integer(id[i]), as.integer(attrib), 
                     count = integer(1))$count
  names(result) <- names(id)
  result
}

rgl.attrib.ncol.values <- c(vertices=3, normals=3, colors=4, texcoords=2, dim=2,
            texts=1, cex=1, adj=3, radii=1, centers=3, ids=1,
	    usermatrix=4, types=1, flags=1, offsets=1,
	    family=1, font=1, pos=1, fogscale=1, axes=3,
	    indices=1)

rgl.attrib.info <- function( id = ids3d("all", 0)$id, attribs = NULL, showAll = FALSE) {
  ncol <- rgl.attrib.ncol.values
  if (is.null(attribs))
    attribs <- names(ncol)
  else if (is.numeric(attribs))
    attribs <- names(ncol)[attribs]
  na <- length(attribs)
  ni <- length(id)
  if (!ni) 
    result <- data.frame(id = numeric(0), attrib = character(0),
    		         nrow = numeric(0), ncol = numeric(0),
    		         stringsAsFactors = FALSE)
  else
    result <- data.frame(id = rep(id, each = na),
    		         attrib = rep(attribs, ni),
  		         nrow = 0, 
  		         ncol = rep(ncol[attribs], ni),
		         stringsAsFactors = FALSE)
  for (j in seq_len(ni)) 
    for (i in seq_len(na))
      result$nrow[i + na*(j - 1)] <- rgl.attrib.count(id[j], result$attrib[i])
  if (!showAll)
    result <- result[result$nrow != 0,]
  rownames(result) <- NULL
  result
}	

rgl.attrib <- function( id, attrib, first=1, 
                        last=rgl.attrib.count(id, attrib) ) {
  stopifnot(length(attrib) == 1 && length(id) == 1 && length(first) == 1)
  if (is.character(attrib))
    attrib <- rgl.enum.attribtype(attrib)
  ncol <- rgl.attrib.ncol.values[attrib]
  count <- max(last - first + 1, 0)
  if (attrib %in% c(6, 13, 16)) { # texts, types and family
    if (count)
      result <- .C(rgl_text_attrib, as.integer(id), as.integer(attrib), 
                    as.integer(first-1), as.integer(count), 
                result = character(count*ncol))$result
    else
      result <- character(0)
  } else {
    if (count)
      result <- .C(rgl_attrib, as.integer(id), as.integer(attrib), 
                  as.integer(first-1), as.integer(count), 
                  result = numeric(count*ncol))$result
    else
      result <- numeric(0)
  }
  if (attrib == 14) # flags
    result <- as.logical(result)
  result <- matrix(result, ncol=ncol, byrow=TRUE)
  colnames(result) <- list(c("x", "y", "z"), # vertices
                           c("x", "y", "z"), # normals
                           c("r", "g", "b", "a"), # colors
                           c("s", "t"),	     # texcoords
                           c("r", "c"),      # dim
                           "text",	         # texts
                           "cex", 	         # cex
                           c("x", "y", "z"), # adj
                           "r",		     # radii
                           c("x", "y", "z"), # centers
                           "id",	     # ids
                           c("x", "y", "z", "w"), # usermatrix
                           "type",	     # types
                           "flag",	     # flags
			                     "offset",         # offsets
  			                   "family",         # family
  			                   "font",           # font
			                     "pos",            # pos
			                     "fogscale",        # fogscale
			                     c("x", "y", "z"),   # axes
			                     "vertex"          # indices
                           )[[attrib]]
  if (attrib == 14 && count) # flags
    if (id %in% ids3d("lights", subscene = 0)$id)
      rownames(result) <- c("viewpoint", "finite")[first:last]
    else if (id %in% ids3d("background", subscene = 0)$id)
      rownames(result) <- c("sphere", "linear_fog", "exp_fog", "exp2_fog")[first:last]
    else if (id %in% ids3d("bboxdeco", subscene = 0)$id)
      rownames(result) <- c("draw_front", "marklen_rel")[first:last]
    else if (id %in% (ids <- ids3d("shapes", subscene = 0))$id) {
      type <- ids$type[ids$id == id]
      rownames(result) <- c("ignoreExtent", 
                            if (type == "surface") "flipped"
                            else if (type == "spheres") "fastTransparency"
                            else "fixedSize",
                            "rotating")[first:last]
    }
  if (attrib == 20 && count) { # axes
    rownames(result) <- c("mode", "step", "nticks",
                          "marklen", "expand")
    result <- result[first:last,]
    result <- as.data.frame(t(result))
    result$mode <- c("custom", "fixednum", "fixedstep", "pretty", "user", "none")[result$mode + 1]
  }
  result
}

##
## ===[ SECTION: environment ]================================================
##



##
## set viewpoint
##
##

rgl.viewpoint <- function( theta = 0.0, phi = 15.0, fov = 60.0, zoom = 1.0, scale = par3d("scale"),
                           interactive = TRUE, userMatrix, type = c("userviewpoint", "modelviewpoint") ) {
  .Deprecated("view3d")
  
  zoom <- rgl.clamp(zoom,0,Inf)
  phi  <- rgl.clamp(phi,-90,90)
  fov  <- rgl.clamp(fov,0,179)
  
  type <- match.arg(type, several.ok = TRUE)

  polar <- missing(userMatrix)
  if (polar) userMatrix <- diag(4)
  
  idata <- as.integer(c(interactive,polar, "userviewpoint" %in% type, "modelviewpoint" %in% type))
  ddata <- as.numeric(c(theta,phi,fov,zoom,scale,userMatrix[1:16]))

  ret <- .C( rgl_viewpoint,
    success = FALSE,
    idata,
    ddata
  )

  if (! ret$success)
    stop("'rgl_viewpoint' failed")
}

##
## set background
##
##

rgl.bg <- function(sphere=FALSE, fogtype="none", color=c("black","white"), back="lines", 
                   fogScale = 1, ... ) {
  .Deprecated("bg3d")
  
  rgl.material0( color=color, back=back, ... )

  fogtype <- rgl.enum.fogtype(fogtype)

  idata   <- as.integer(c(sphere,fogtype))
  
  fogScale <- as.numeric(fogScale)
  stopifnot(length(fogScale) == 1, fogScale > 0)

  ret <- .C( rgl_bg, 
    success = as.integer(FALSE),
    idata,
    fogScale
  )

  if (! ret$success)
    stop("'rgl_bg' failed")
    
  lowlevel(ret$success)
}


##
## bbox
##
##

rgl.bbox <- function( 
  xat=NULL, xlab=NULL, xunit=0, xlen=5,
  yat=NULL, ylab=NULL, yunit=0, ylen=5,
  zat=NULL, zlab=NULL, zunit=0, zlen=5,
  marklen=15.0, marklen.rel=TRUE, expand=1, draw_front=FALSE,
  ...) {

  .Deprecated("bbox3d")
  
  rgl.material0( ... )

  if (is.null(xat)) 
    xlab <- NULL
  else {
    xlen <- length(xat)
    if (is.null(xlab)) 
      xlab <- format(xat)
    else 
      xlab <- rep(xlab, length.out=xlen)
  }
  if (is.null(yat)) 
    ylab <- NULL
  else {
    ylen <- length(yat)
    if (is.null(ylab)) 
      ylab <- format(yat)
    else 
      ylab <- rep(ylab, length.out=ylen)
  }
  if (is.null(zat)) 
    zlab <- NULL
  else {
    zlen <- length(zat)
    if (is.null(zlab)) 
      zlab <- format(zat)
    else 
      zlab <- rep(zlab,length.out=length(zat))
  }
  xticks <- length(xat)
  yticks <- length(yat)
  zticks <- length(zat)

  if (identical(xunit, "pretty")) xunit <- -1
  if (identical(yunit, "pretty")) yunit <- -1
  if (identical(zunit, "pretty")) zunit <- -1

  length(xlen)        <- 1
  length(ylen)        <- 1
  length(zlen)        <- 1
  length(marklen.rel) <- 1
  length(draw_front)  <- 1
  length(xunit)       <- 1
  length(yunit)       <- 1
  length(zunit)       <- 1
  length(marklen)     <- 1
  length(expand)      <- 1

  idata <- as.integer(c(xticks,yticks,zticks, xlen, ylen, zlen, marklen.rel, draw_front))
  ddata <- as.numeric(c(xunit, yunit, zunit, marklen, expand))

  ret <- .C( rgl_bbox,
    success = as.integer(FALSE),
    idata,
    ddata,
    as.numeric(xat),
    as.character(xlab),
    as.numeric(yat),
    as.character(ylab),
    as.numeric(zat),
    as.character(zlab)
  )

  if (! ret$success)
    stop("'rgl_bbox' failed")
    
  lowlevel(ret$success)

}

##
## set lights
##
##

rgl.light <- function( theta = 0, phi = 0, viewpoint.rel = TRUE, ambient = "#FFFFFF", diffuse = "#FFFFFF", specular = "#FFFFFF", x = NULL, y = NULL, z = NULL) {
  
  .Deprecated("light3d")
  
  ambient  <- rgl.color(ambient)
  diffuse  <- rgl.color(diffuse)
  specular <- rgl.color(specular)
  
  # if a complete set of x, y, z is given, the light source is assumed to be part of the scene, theta and phi are ignored
  # else the light source is infinitely far away and its direction is determined by theta, phi (default) 
  if ( !is.null(x) ) {
    if ( !missing(theta) || !missing(phi) )
      warning("'theta' and 'phi' ignored when 'x' is present")
    xyz <- xyz.coords(x,y,z, recycle = TRUE)
    x <- xyz$x
    y <- xyz$y
    z <- xyz$z
    if (length(x) > 1) stop("A light can only be in one place at a time")
    finite.pos <- TRUE
  }
  else {
    
    if ( !is.null(y) || !is.null(z) ) 
      warning("'y' and 'z' ignored, spherical coordinates used")
    finite.pos <- FALSE
    x <- 0
    y <- 0
    z <- 0
    
  }
    

  idata <- as.integer(c(viewpoint.rel, ambient, diffuse, specular, finite.pos))
  ddata <- as.numeric(c(theta, phi, x, y, z))

  ret <- .C( rgl_light,
    success = as.integer(FALSE),
    idata,
    ddata
  )

  if (! ret$success)
    stop("Too many lights; maximum is 8 sources per scene")
    
  lowlevel(ret$success)
}

##
## ===[ SECTION: shapes ]=====================================================
##

##
## add primitive
##
##

rgl.primitive0 <- function( type, x, y=NULL, z=NULL, normals=NULL, texcoords=NULL, indices=NULL, ... ) {
  rgl.material0( ... )

  type <- rgl.enum.primtype(type)

  vertex  <- rgl.vertex(x,y,z)
  nvertex <- rgl.nvertex(vertex)
  
  if (is.null(indices)) {
    nindices <- 0
    indices <- 0 # to avoid pointing out of range
  } else {
    nindices <- length(indices)
    if (!all(indices > 0 & indices <= nvertex))
      stop("indices out of range")
  }
  
  if (nvertex > 0) {
    
    perelement <- c(points=1, lines=2, triangles=3, quadrangles=4, linestrips=1)[type]
    if (nindices > 0) {
      if (nindices %% perelement)
        stop("Illegal number of indices") 
    } else if (nvertex %% perelement) 
      stop("Illegal number of vertices")
    
    idata   <- as.integer( c(type, nvertex, !is.null(normals), !is.null(texcoords), nindices, indices - 1 ) )
    
    if (is.null(normals)) normals <- 0
    else {
      normals <- xyz.coords(normals, recycle=TRUE)
      x <- rep(normals$x, length.out = nvertex)
      y <- rep(normals$y, length.out = nvertex)
      z <- rep(normals$z, length.out = nvertex)
      normals <- rgl.vertex(x,y,z)
    }
    
    if (is.null(texcoords)) texcoords <- 0
    else {
      texcoords <- xy.coords(texcoords, recycle=TRUE)
      s <- rep(texcoords$x, length.out = nvertex)
      t <- rep(texcoords$y, length.out = nvertex)
      texcoords <- rgl.texcoords(s,t)
    } 
    
    success <- .Call( rgl_primitive,
      as.integer(idata),
      as.numeric(vertex),
      as.numeric(normals),
      as.numeric(texcoords)
    )
        
    if (!success)
      stop("'rgl_primitive' failed")
      
    lowlevel(success)
  }
}

rgl.primitive <- function(...) {
  .Deprecated("", msg = "Please call the primitive directly, e.g. points3d(...).")
  rgl.primitive0(...)
}

rgl.points <- function( x, y=NULL, z=NULL, ... ) {
  .Deprecated("points3d")
  rgl.primitive0( "points", x, y, z, ... )
}

rgl.lines <- function(x, y=NULL, z=NULL, ... ) {
  .Deprecated("segments3d")
  rgl.primitive0( "lines", x, y, z, ... )
}

rgl.triangles <- function(x, y=NULL, z=NULL, normals=NULL, texcoords=NULL, ... ) {
  .Deprecated("triangles3d")
  rgl.primitive0( "triangles", x, y, z, normals, texcoords, ... )
}

rgl.quads <- function( x, y=NULL, z=NULL, normals=NULL, texcoords=NULL, ... ) {
  .Deprecated("quads3d")
  rgl.primitive0( "quadrangles", x, y, z, normals, texcoords, ... )
}

rgl.linestrips<- function( x, y=NULL, z=NULL, ... ) {
  .Deprecated("lines3d")
  rgl.primitive0( "linestrips", x, y, z, ... )
}

##
## add surface
##
##

# Utility function:
# calculates the parity of a permutation of integers

perm_parity <- function(p) {
  x <- seq_along(p)
  result <- 0
  for (i in x) {
    if (x[i] != p[i]) {
      x[x==p[i]] <- x[i]
      result <- result+1
    }
  }
  return(result %% 2)
}

rgl.surface <- function( x, z, y, coords=1:3,  ..., normal_x=NULL, normal_y=NULL, normal_z=NULL,
                         texture_s=NULL, texture_t=NULL) {
  .Deprecated("surface3d")
  
  rgl.material0(...)
  
  flags <- rep(FALSE, 4)
  
  if (is.matrix(x)) {
    nx <- nrow(x)
    flags[1] <- TRUE
    if ( !identical( dim(x), dim(y) ) ) stop(gettextf("Bad dimension for %s", "rows"),
    					     domain = NA)
  } else nx <- length(x)
  
  if (is.matrix(z)) {
    nz <- ncol(z)
    flags[2] <- TRUE
  } else nz <- length(z)
  
  if (is.matrix(y)) {
    if (any(dim(y) != c(nx, nz)))
      stop(gettextf("Bad dimension for %s", "y"))
  } else if (is.matrix(x)) {
    if (length(y) != nx)
      stop(gettextf("Bad length for %s", "y"))
    y <- matrix(y, nx, nz)
  } else 
    y <- matrix(y, nx, nz, byrow = TRUE)
  
  ny <- length(y)

  if ( nx*nz != ny)
    stop("'y' length != 'x' rows * 'z' cols")

  if ( nx < 2 )
    stop("rows < 2")
  
  if ( nz < 2 )   
    stop("cols < 2")
    
  if ( length(coords) != 3 || !identical(all.equal(sort(coords), 1:3), TRUE) )
    stop("'coords' must be a permutation of 1:3")
  
  nulls <- c(is.null(normal_x), is.null(normal_y), is.null(normal_z))
  if (!all( nulls ) ) {
    if (any( nulls )) stop("All normals must be supplied")
    if ( !identical(dim(y), dim(normal_x)) 
      || !identical(dim(y), dim(normal_y))
      || !identical(dim(y), dim(normal_z)) ) stop(gettextf("Bad dimension for %s", "normals"),
      					    domain = NA)
    flags[3] <- TRUE
  }
  
  nulls <- c(is.null(texture_s), is.null(texture_t))
  if (!all( nulls ) ) {
    if (any( nulls )) stop("Both texture coordinates must be supplied")
    if ( !identical(dim(y), dim(texture_s))
      || !identical(dim(y), dim(texture_t)) ) stop(gettextf("Bad dimension for %s", "textures"),
      					     domain = NA)
    flags[4] <- TRUE
  }

  idata <- as.integer( c( nx, nz ) )

  xdecreasing <- diff(x[!is.na(x)][1:2]) < 0
  zdecreasing <- diff(z[!is.na(z)][1:2]) < 0
  parity <- (perm_parity(coords) + xdecreasing + zdecreasing ) %% 2
  
  if (is.na(parity))
    parity <- 0
  
  ret <- .C( rgl_surface,
    success = as.integer(FALSE),
    idata,
    as.numeric(x),
    as.numeric(z),
    as.numeric(y),
    as.numeric(normal_x),
    as.numeric(normal_z),
    as.numeric(normal_y),
    as.numeric(texture_s),
    as.numeric(texture_t),
    as.integer(coords),
    as.integer(parity),
    as.integer(flags),
    NAOK=TRUE
  )

  if (! ret$success)
    stop("'rgl_surface' failed")
    
  lowlevel(ret$success)
}

##
## add spheres
##

rgl.spheres <- function( x, y=NULL, z=NULL, radius=1.0, fastTransparency = TRUE, ...) {
  rgl.material0(...)
  
  vertex  <- rgl.vertex(x,y,z)
  nvertex <- rgl.nvertex(vertex)
  radius  <- rgl.attr(radius, nvertex)
  nradius <- length(radius)
  if (nvertex && nradius) {
    
    idata <- as.integer( c( nvertex, nradius ) )
    
    ret <- .C( rgl_spheres,
               success = as.integer(FALSE),
               idata,
               as.numeric(vertex),    
               as.numeric(radius),
               as.integer(fastTransparency),
               NAOK=TRUE
    )
    
    if (! ret$success)
      stop("'rgl_spheres' failed")
    
    lowlevel(ret$success)
  }
}

##
## add planes
##

rgl.planes <- function( a, b=NULL, c=NULL, d=0,...) {
  .Deprecated("planes3d")
  
  rgl.material0(...)
  
  normals  <- rgl.vertex(a, b, c)
  nnormals <- rgl.nvertex(normals)
  noffsets <- length(d)
  
  if (nnormals && noffsets) {
    
    idata <- as.integer( c( nnormals, noffsets ) )
    
    ret <- .C( rgl_planes,
               success = as.integer(FALSE),
               idata,
               as.numeric(normals),    
               as.numeric(d),
               NAOK=TRUE
    )
    
    if (! ret$success)
      stop("'rgl_planes' failed")
    
    lowlevel(ret$success)
  }
}

##
## add clip planes
##

rgl.clipplanes <- function( a, b=NULL, c=NULL, d=0) {
  .Deprecated("clipplanes3d")
  
  normals  <- rgl.vertex(a, b, c)
  nnormals <- rgl.nvertex(normals)
  noffsets <- length(d)
  if (nnormals && noffsets) {
    idata <- as.integer( c( nnormals, noffsets ) )
    
    ret <- .C( rgl_clipplanes,
               success = as.integer(FALSE),
               idata,
               as.numeric(normals),    
               as.numeric(d),
               NAOK=TRUE
    )
    
    if (! ret$success)
      stop("'rgl_clipplanes' failed")
    
    lowlevel(ret$success)
  }
}


##
## add abclines
##

rgl.abclines <- function(x, y=NULL, z=NULL, a, b=NULL, c=NULL, ...) {
  .Deprecated("abclines3d")
  
  rgl.material0(...)
  
  bases  <- rgl.vertex(x, y, z)
  nbases <- rgl.nvertex(bases)
  directions <- rgl.vertex(a, b, c)
  ndirs <-  rgl.nvertex(directions)
  if (nbases && ndirs) {
    idata <- as.integer( c( nbases, ndirs ) )
    
    ret <- .C( rgl_abclines,
               success = as.integer(FALSE),
               idata,
               as.numeric(bases),    
               as.numeric(directions),
               NAOK=TRUE
    )
    
    if (! ret$success)
      stop("'rgl_abclines' failed")
    
    lowlevel(ret$success)
  }
}


##
## add texts
##

rgl.texts <- function(x, y=NULL, z=NULL, text, adj = 0.5, pos = NULL, offset = 0.5, 
                      family=par3d("family"), font=par3d("font"), 
                      cex=par3d("cex"), useFreeType=par3d("useFreeType"), 
                      ... ) {
  
  .Deprecated("text3d")
  
  rgl.material0( ... )
  
  vertex  <- rgl.vertex(x,y,z)
  nvertex <- rgl.nvertex(vertex)
  
  if (!is.null(pos)) {
    npos <- length(pos)
    stopifnot(all(pos %in% 0:6))
    stopifnot(length(offset) == 1)
    adj <- offset
  } else {
    pos <- 0
    npos <- 1
  }
  if (length(adj) > 3) warning("Only the first three entries of 'adj' are used")
  adj <- c(adj, 0.5, 0.5, 0.5)[1:3]
  
  if (!length(text)) {
    if (nvertex)
      warning("No text to plot")
    return(invisible(integer(0)))
  }
  if (nvertex) {
    text    <- rep(text, length.out=nvertex)
    
    idata <- as.integer(nvertex)
    
    nfonts <- max(length(family), length(font), length(cex)) 
    family <- rep(family, length.out = nfonts)
    font <- rep(font, length.out = nfonts)
    cex <- rep(cex, length.out = nfonts)  
    
    family[font == 5] <- "symbol"
    font <- ifelse( font < 0 | font > 4, 1, font)  
    
    ret <- .C( rgl_texts,
               success = as.integer(FALSE),
               idata,
               as.double(adj),
               as.character(text),
               as.numeric(vertex),
               as.integer(nfonts),
               as.character(family), 
               as.integer(font),
               as.numeric(cex),
               as.integer(useFreeType),
               as.integer(npos),
               as.integer(pos),
               NAOK=TRUE
    )
    
    if (! ret$success)
      stop("'rgl_texts' failed")
    
    lowlevel(ret$success)
  }
}

##
## add sprites
##

rgl.sprites <- function( x, y=NULL, z=NULL, radius=1.0, shapes=NULL, 
                         userMatrix=diag(4), fixedSize = FALSE,
                         adj = 0.5, pos = NULL, offset = 0.25, rotating = FALSE, 
                         ... ) {
  .Deprecated("sprites3d")
  
  rgl.material0(...)
  
  center  <- rgl.vertex(x,y,z)
  ncenter <- rgl.nvertex(center)
  radius  <- rgl.attr(radius, ncenter)
  nradius <- length(radius)
  
  pos <- as.integer(pos)
  npos <- length(pos)
  if (npos) {
    pos <- rep(pos, length.out = ncenter)
    adj <- offset
  }
  adj <- c(adj, 0.5, 0.5, 0.5)[1:3]
  if (ncenter && nradius) {
    if (length(shapes) && length(userMatrix) != 16) stop("Invalid 'userMatrix'")
    if (length(fixedSize) != 1) stop("Invalid 'fixedSize'")
    idata   <- as.integer( c(ncenter,nradius,length(shapes), fixedSize, npos, rotating) )
    
    ret <- .C( rgl_sprites,
               success = as.integer(FALSE),
               idata,
               as.numeric(center),
               as.numeric(radius),
               as.integer(shapes),
               as.numeric(t(userMatrix)),
               as.numeric(adj),
               pos,
               as.numeric(offset),
               NAOK=TRUE
    )
    
    if (! ret$success)
      stop("'rgl_sprites' failed")
    
    lowlevel(ret$success)
  }
}

##
## convert user coordinate to window coordinate
## 

rgl.user2window <- function( x, y=NULL, z=NULL, projection = rgl.projection()) {
  xyz <- xyz.coords(x,y,z,recycle=TRUE)
  points <- rbind(xyz$x,xyz$y,xyz$z,1)
  v <- asEuclidean(with(projection, t(proj %*% model %*% points))) # nolint
  viewport <- projection$view
  cbind( (v[,1]*0.5 + 0.5) + viewport[1]/viewport[3],
         (v[,2]*0.5 + 0.5) + viewport[2]/viewport[4],
         (1 + v[,3])*0.5 )
}

##
## convert window coordinate to user coordinate
## 

rgl.window2user <- function( x, y = NULL, z = 0, projection = rgl.projection()) {
  xyz <- xyz.coords(x,y,z,recycle=TRUE)

  viewport <- projection$view

  normalized <- rbind( 2*xyz$x - 1,
                       2*(xyz$y - viewport[2]/viewport[4]) - 1,
                       2*xyz$z - 1,
                       1 )
  asEuclidean(with(projection, t(solve(proj %*% model, normalized)))) # nolint
}

# Selectstate values
msNONE     <- 1
msCHANGING <- 2
msDONE     <- 3
msABORT    <- 4

rgl.selectstate <- function(dev = cur3d(), subscene = currentSubscene3d(dev)) {
	ret <- .C( rgl_selectstate,
    	as.integer(dev),
    	as.integer(subscene),
    	success = FALSE,
    	state = as.integer(0),
    	mouseposition = double(4)
  	)

  	if (! ret$success)
    	  stop("'rgl_selectstate' failed")
    return(ret)
}


rgl.select <- function(button = c("left", "middle", "right"), 
                       dev = cur3d(), subscene = currentSubscene3d(dev)) {
	if (rgl.useNULL())
	  return(NULL)
	button <- match.arg(button)
	
	newhandler <- par3d("mouseMode", dev = dev, subscene = subscene)
	newhandler[button] <- "selecting"
	oldhandler <- par3d(mouseMode = newhandler, dev = dev, subscene = subscene)
	on.exit(par3d(mouseMode = oldhandler, dev = dev, subscene = subscene))
	
	while ((result <- rgl.selectstate(dev = dev, subscene = subscene))$state < msDONE)
		Sys.sleep(0.1)
	
	rgl.setselectstate("none", dev = dev, subscene = subscene)
	
	if (result$state == msDONE)
	    return(result$mouseposition)
	else
	    return(NULL)
}

rgl.setselectstate <- function(state = "current", 
                               dev = cur3d(), subscene = currentSubscene3d(dev)) {
	state <- rgl.enum(state, current=0, none = 1, middle = 2, done = 3, abort = 4)
	idata <- as.integer(c(state))
	
	  ret <- .C( rgl_setselectstate, 
	    as.integer(dev),
	    as.integer(subscene),
	    success = FALSE,
	    state = idata
	  )
	
	  if (! ret$success)
	    stop("'rgl_setselectstate' failed")

	c("none", "middle", "done", "abort")[ret$state]
}

rgl.projection <- function(dev = cur3d(), subscene = currentSubscene3d(dev)) {
    list(model = par3d("modelMatrix", dev = dev, subscene = subscene),
    	 proj = par3d("projMatrix", dev = dev, subscene = subscene),
    	 view = par3d("viewport", dev = dev, subscene = subscene))
}   
     
rgl.select3d <- function(button = c("left", "middle", "right"), 
                         dev = cur3d(), subscene = currentSubscene3d(dev)) {
  .Deprecated("select3d")
  rect <- rgl.select(button = button, dev = dev, subscene = subscene)
  if (is.null(rect)) return(NULL)
  proj <- rgl.projection(dev = dev, subscene = subscene)
  
  llx <- rect[1]
  lly <- rect[2]
  urx <- rect[3]
  ury <- rect[4]

  selectionFunction3d(proj, region = c(llx, lly, urx, ury))
}

selectionFunction3d <- function(proj, region = proj$region) {
  llx <- region[1]
  lly <- region[2]
  urx <- region[3]
  ury <- region[4]
  
  if ( llx > urx ) {
    temp <- llx
    llx <- urx
    urx <- temp
  }
  if ( lly > ury ) {
    temp <- lly
    lly <- ury
    ury <- temp
  }
  proj$view["x"] <- proj$view["y"] <- 0
  function(x,y=NULL,z=NULL) {
    pixel <- rgl.user2window(x,y,z,projection=proj)
    x <- pixel[,1]
    y <- pixel[,2]
    z <- pixel[,3]
    (llx <= x) & (x <= urx) & (lly <= y) & (y <= ury) & 
    (0 <= z) & (z <= 1)
  }
}

tagged3d <- function(tags = NULL, ids = NULL, full = FALSE, subscene = 0) {
  if ((missing(tags) + missing(ids)) != 1) 
    stop("Exactly one of 'tags' and 'ids' should be specified.")
  all <- ids3d("all", subscene = subscene, tags = TRUE)
  if (!missing(tags))
    all <- all[all$tag %in% tags,]
  else
    all <- all[match(ids, all$id),]
  if (full)
    all
  else
    if (!missing(tags) && !is.null(tags)) 
      all$id
    else if (!missing(ids) && !is.null(ids))
      all$tag
    else
      NULL
}

library(rgl)

test_that("rgl.user2window and rgl.window2user work", {
	xyz <- rbind(x = c(2, 4, 6))
	abc <- rbind(x = c(0.1, 0.2, 0.3))
	open3d()
	mfrow3d(2, 2)
	points3d(1:10, 1:10, 1:10)
	expect_equal(rgl.window2user(rgl.user2window(xyz)), xyz)
	expect_equal(rgl.user2window(rgl.window2user(abc)), abc)
	next3d()
	points3d(1:10, 1:10, 1:10)
	expect_equal(rgl.window2user(rgl.user2window(xyz)), xyz)
	expect_equal(rgl.user2window(rgl.window2user(abc)), abc)
	next3d()
	points3d(1:10, 1:10, 1:10)
	expect_equal(rgl.window2user(rgl.user2window(xyz)), xyz)
	expect_equal(rgl.user2window(rgl.window2user(abc)), abc)
	next3d()
	points3d(1:10, 1:10, 1:10)
	expect_equal(rgl.window2user(rgl.user2window(xyz)), xyz)
	expect_equal(rgl.user2window(rgl.window2user(abc)), abc)
})

library(rgl)

test_that("tags work", {
  material3d(tag = "hello")
  expect_equal(material3d("tag"), "hello")

  file <- normalizePath(system.file("textures/worldsmall.png",                              package = "rgl"))
  material3d(texture = file)
  expect_equal(material3d("texture"), file)
  expect_equal(material3d("tag"), "hello")

  open3d()
  x <- points3d(1,2,3, tag = "hello2")
  expect_equal(unclass(x), tagged3d("hello2"))
  expect_equal(unclass(x), tagged3d("hello2", full = TRUE)$id)
  expect_equal(tagged3d(ids = x), "hello2")
})

test_that("pop3d works on tags", {
  open3d()
  id1 <- points3d(1,1,1, tag = "id1")
  newSubscene3d()
  id2 <- points3d(2,2,2, tag = "id2")
  pop3d(tag = "id1")
  expect_equal(ids3d(subscene = 0, tags = TRUE),
               data.frame(id = as.numeric(id2), 
                          type = "points", tag = "id2"))
})

# These quaternion functions are adapted from the orientlib package

# Convert an array of rotation matrices to a matrix of unit quaternions
toQuaternions <- function(x) {
  nicesqrt <- function(x) sqrt(pmax(x,0))
  q4 <- nicesqrt((1 + x[1,1,] + x[2,2,] + x[3,3,])/4)  # may go negative by rounding
  zeros <- zapsmall(q4) == 0 
  q1 <- ifelse(!zeros, (x[2,3,] - x[3,2,])/4/q4, nicesqrt(-(x[2,2,]+x[3,3,])/2))
  q2 <- ifelse(!zeros, (x[3,1,] - x[1,3,])/4/q4, 
               ifelse(zapsmall(q1) != 0, x[1,2,]/2/q1, nicesqrt((1-x[3,3,])/2)))
  q3 <- ifelse(!zeros, (x[1,2,] - x[2,1,])/4/q4, 
               ifelse(zapsmall(q1) != 0, x[1,3,]/2/q1, 
                      ifelse(zapsmall(q2) != 0, x[2,3,]/2/q2, 1)))
  cbind(q1, q2, q3, q4)
}

# Convert a single quaternion to a rotation matrix

toRotmatrix <- function(x) {
  x <- x/sqrt(sum(x^2))
  matrix(c(1-2*x[2]^2-2*x[3]^2, 
           2*x[1]*x[2]-2*x[3]*x[4],
           2*x[1]*x[3]+2*x[2]*x[4],
           2*x[1]*x[2]+2*x[3]*x[4],
           1-2*x[1]^2-2*x[3]^2,
           2*x[2]*x[3] - 2*x[4]*x[1],
           2*x[1]*x[3] - 2*x[2]*x[4],
           2*x[2]*x[3] + 2*x[1]*x[4],
           1 - 2*x[1]^2 - 2*x[2]^2), 3,3)
}

par3dinterp <- function(times=NULL, userMatrix, scale, zoom, FOV, method=c("spline", "linear"), 
                        extrapolate = c("oscillate","cycle","constant", "natural"),
                        dev = cur3d(), subscene = par3d("listeners", dev = dev)) {
  force(dev)
  force(subscene)
  
  if (is.list(times)) {
    for (n in setdiff(names(times), "times")) assign(n, times[[n]])
    if ("times" %in% names(times)) times <- times[["times"]]
    else times <- NULL
  }
  
  if (!missing(userMatrix) && is.list(userMatrix)) userMatrix <- do.call(cbind, userMatrix)
  if (!missing(scale) && is.list(scale)) scale <- do.call(rbind, scale)
  if (!missing(zoom) && is.list(zoom)) zoom <- unlist(zoom)
  if (!missing(FOV) && is.list(FOV)) FOV <- unlist(FOV)
  
  if (is.null(times)) {
    times <- if (!missing(userMatrix)) 0:(length(userMatrix)/16 - 1)
    else if (!missing(scale)) 0:(dim(scale)[1] - 1)
    else if (!missing(zoom)) 0:(length(zoom) - 1)
    else if (!missing(FOV)) 0:(length(FOV) - 1)
  }
  data <- matrix(0, length(times), 0)
  if (!missing(userMatrix)) {
    stopifnot( length(userMatrix) == 16*length(times) )
    userMatrix <- array(userMatrix, c(4,4,length(times)))
    xlat <- ncol(data) + 1:4
    data <- cbind(data, t(userMatrix[,4,, drop = TRUE]))
    persp <- ncol(data) + 1:3
    data <- cbind(data, t(userMatrix[4,1:3,, drop = TRUE]))
    rot <- ncol(data) + 1:4
    quat <- toQuaternions(userMatrix[1:3, 1:3, , drop = FALSE])
    # Since q and -q are the same rotation, we want to interpolate
    # to the nearer one.
    for (i in seq_len(nrow(quat))[-1]) {
      if (sum((quat[i - 1, ] - quat[i, ])^2) > sum((quat[i - 1, ] + quat[i, ])^2)) 
        quat[i, ] <- -quat[i, ]
    }
    data <- cbind(data, quat)	
  } else {
    xlat <- NULL
  }
  if (!missing(scale)) {
    stopifnot( dim(scale)[1] == length(times) )
    sc <- ncol(data) + 1:3
    data <- cbind(data, log(scale))
  } else sc <- NULL
  if (!missing(zoom)) {
    stopifnot( length(zoom) == length(times) )
    zm <- ncol(data) + 1
    data <- cbind(data, log(zoom))
  } else zm <- NULL
  if (!missing(FOV)) {
    stopifnot( length(FOV) == length(times) )
    fov <- ncol(data) + 1
    data <- cbind(data, FOV)
  } else fov <- NULL
  
  method <- match.arg(method)
  extrapolate <- match.arg(extrapolate)
  if (extrapolate == "oscillate") {
    n <- length(times)
    times <- c(times[-n], -rev(times) + 2*times[length(times)])
    data <- rbind(data[-n,,drop=FALSE], data[n:1,,drop=FALSE])
    n <- 2*n - 1
    extrapolate <- "cycle"
  } else if (extrapolate == "natural" && method != "spline")
    stop("'natural' extrapolation only supported for spline method")
  
  if (method == "spline") {
    fns <- apply(data, 2, function(col) splinefun(times, col, 
                                                  method = ifelse(extrapolate == "cycle", "periodic", "natural")))
  } else {
    fns <- apply(data, 2, function(col) approxfun(times, col, rule=2))
  }
  
  mintime <- min(times)
  maxtime <- max(times)
  
  function(time) {
    if (time < mintime || time > maxtime) {
      if (extrapolate == "constant" || mintime == maxtime)
        time <- ifelse(time < mintime, mintime, maxtime)
      else if (extrapolate == "cycle")
        time <- (time - mintime) %% (maxtime - mintime) + mintime
    }
    data <- sapply(fns, function(f) f(time))
    result <- list(dev = dev, subscene = subscene)
    if (!is.null(xlat)) {
      userMatrix <- matrix(0, 4,4)
      userMatrix[,4] <- data[xlat]
      userMatrix[4,1:3] <- data[persp]
      userMatrix[1:3,1:3] <- toRotmatrix(data[rot])
      result$userMatrix <- userMatrix
    }
    if (!is.null(sc)) 
      result$scale <- exp(data[sc])
    if (!is.null(zm))
      result$zoom <- exp(data[zm])
    if (!is.null(fov))
      result$FOV <- data[fov]
    result
  }
}

spin3d <- function(axis = c(0, 0, 1), rpm = 5, dev = cur3d(), subscene = par3d("listeners", dev = dev)) {
  force(axis)
  force(rpm)    
  force(dev)
  force(subscene)
  M <- par3d("userMatrix", dev = dev, subscene = subscene)
  function(time, base = M) 
    list(userMatrix = rotate3d(base, time*rpm*pi/30, axis[1], axis[2], axis[3]), dev = dev, subscene = subscene)
}

play3d <- function(f, duration = Inf, dev = cur3d(), ..., startTime = 0) {
  # Don't want to start timing until args are known: they may be obtained
  # interactively
  force(f)  
  force(duration)
  force(dev)
  start <- proc.time()[3] - startTime
  rgl.setselectstate("none")
  repeat {
    if(!missing(dev) && cur3d() != dev) set3d(dev)
    time <- proc.time()[3] - start
    if (time > duration || rgl.selectstate()$state == msABORT) return(invisible(NULL))
    stopifnot(cur3d() != 0)
    args <- f(time, ...)
    subs <- args$subscene
    if (is.null(subs))
      subs <- currentSubscene3d(dev)
    else
      args$subscene <- NULL
    for (s in subs)
      par3d(args, subscene = s)
  }
}

movie3d <- function(f, duration, dev = cur3d(), ..., fps=10, 
                    movie = "movie", frames = movie, dir = tempdir(), 
                    convert = NULL, clean = TRUE, verbose=TRUE,
                    top = !rgl.useNULL(), type = "gif", startTime = 0,
                    webshot = TRUE) {
  
  olddir <- setwd(dir)
  on.exit(setwd(olddir))
  
  for (i in round(startTime*fps):(duration*fps)) {
    time <- i/fps        
    if(cur3d() != dev) set3d(dev)
    stopifnot(cur3d() != 0)
    args <- f(time, ...)
    subs <- args$subscene
    if (is.null(subs))
      subs <- currentSubscene3d(dev)
    else
      args$subscene <- NULL
    for (s in subs)
      par3d(args, subscene = s)
    filename <- sprintf("%s%03d.png",frames,i)
    if (verbose) {
      cat(gettextf("Writing '%s'\r", filename))
      flush.console()
    }
    if (top)
      rgl.bringtotop()
    snapshot3d(filename = filename, webshot = webshot)
  }	
  cat("\n")
  if (.Platform$OS.type == "windows") system <- shell  # nolint
  if (is.null(convert) && requireNamespace("magick", quietly = TRUE)) {
    m <- NULL
    for (i in round(startTime*fps):(duration*fps)) {
      filename <- sprintf("%s%03d.png",frames,i)
      frame <- magick::image_read(filename)
      if (is.null(m)) m <- frame
      else m <- c(m, frame)
      if (clean)
        unlink(filename)
    }
    m <- magick::image_animate(m, fps = fps, loop = 1, dispose = "previous")
    magick::image_write(m, paste0(movie, ".", type))
    return(invisible(m))
  } else if (is.null(convert)) {
    warning("R package 'magick' is not installed; trying external package.")
    convert <- TRUE
  }
  if (is.logical(convert) && convert) {
    # Check for ImageMagick
    progname <- "magick"
    version <- try(system2(progname, "--version", stdout = TRUE,
                           stderr = TRUE), silent = TRUE)
    if (inherits(version, "try-error") || !length(grep("ImageMagick", version))) {
      progname <- "convert"
      version <- try(system2(progname, "--version", stdout = TRUE, 
                             stderr = TRUE), silent = TRUE)
    }
    if (inherits(version, "try-error") || !length(grep("ImageMagick", version)))
      stop("'ImageMagick' not found")    
    filename <- paste0(movie, ".", type)
    if (verbose) cat(gettextf("Will create: %s\n", file.path(dir, filename)))
    convert <- paste(progname, "-delay 1x%d %s*.png %s.%s")
  }
  if (is.character(convert)) {
    convert <- sprintf(convert, fps, frames, movie, type, duration, dir)
    if (verbose) {
      cat(gettextf("Executing: '%s'\n", convert))
      flush.console()
    }
    system(convert)
    if (clean) {
      if (verbose)
        cat(gettext("Deleting frames\n"))
      for (i in 0:(duration*fps)) {
        filename <- sprintf("%s%03d.png",frames,i)
        unlink(filename)
      }
    }
  }
  invisible(convert)
}

# Functions for creating 4x4 graphics matrices

identityMatrix <- function() diag(nrow=4)

scaleMatrix <- function(x,y,z) diag(c(x,y,z,1))

translationMatrix <- function(x,y,z) {
    result <- diag(4)
    result[4,1:3] <- c(x,y,z)
    result
}

rotationMatrix <- function(angle,x,y,z,matrix) {
    if (missing(matrix)) {
    	if (angle == 0) return(identityMatrix())
    	
        u <- c(x,y,z)/sqrt(x^2+y^2+z^2)
        cosa <- cos(angle)
        sina <- sin(angle)
        matrix <- (1-cosa)*outer(u,u)
        matrix <- matrix + diag(3)*cosa
        matrix[1,2] <- matrix[1,2] - sina*u[3]
        matrix[1,3] <- matrix[1,3] + sina*u[2]
        matrix[2,1] <- matrix[2,1] + sina*u[3]
        matrix[2,3] <- matrix[2,3] - sina*u[1]
        matrix[3,1] <- matrix[3,1] - sina*u[2]
        matrix[3,2] <- matrix[3,2] + sina*u[1]
    }
    if (identical(all.equal(dim(matrix), c(3,3)), TRUE)) 
        matrix <- cbind(rbind(matrix,c(0,0,0)),c(0,0,0,1))
    return(matrix)    
}

# Coordinate conversions

asHomogeneous <- function(x) {
    if (is.matrix(x)) {
      if (ncol(x) == 3) return(cbind(x,1))
      if (ncol(x) == 4) return(x)
    } else {
      if (length(x) %% 3 == 0) 
        return(cbind(matrix(x, ncol = 3, byrow = TRUE),1))
      if (length(x) %% 4 == 0)
        return(matrix(x, ncol = 4, byrow = TRUE))
    }
    stop("Don't know how to convert x")
}

asHomogeneous2 <- function(x) {
  if (is.matrix(x)) {
    if (nrow(x) == 3) return(rbind(x,1))
    if (nrow(x) == 4) return(x)
  } else {
    if (length(x) %% 3 == 0) 
      return(rbind(matrix(x, nrow = 3), 1))
    if (length(x) %% 4 == 0)
      return(matrix(x, nrow = 4))
  }
  stop("Don't know how to convert x")
}

asEuclidean <- function(x) {
    if (is.matrix(x)) {
      if (ncol(x) == 4) return(x[, 1:3, drop = FALSE]/x[, 4])
      if (ncol(x) == 3) return(x)
    } else {
      if (length(x) %% 4 == 0) 
        return(asEuclidean(matrix(x, ncol = 4, byrow = TRUE)))
      if (length(x) %% 3 == 0)
        return(matrix(x, ncol = 3, byrow = TRUE))
    }
    stop("Don't know how to convert x")
}

asEuclidean2 <- function(x) {
  if (is.matrix(x)) {
    if (nrow(x) == 4) return(rbind(x[1,]/x[4,], x[2,]/x[4,], x[3,]/x[4,]))
    if (nrow(x) == 3) return(x)
  } else {
    if (length(x) %% 4 == 0) 
      return(asEuclidean2(matrix(x, nrow = 4)))
    if (length(x) %% 3 == 0)
      return(matrix(x, nrow = 3))
  }
  stop("Don't know how to convert x")
}

# Default implementations of transformations

translate3d.default <- function(obj,x,y,z,...) {
    if (is.matrix(obj)) n <- dim(obj)[1]
    else n <- 1
        
    if (length(obj) == 3 || (is.matrix(obj) && dim(obj)[2] == 3))
    	return(obj + cbind(rep(x,n), rep(y,n), rep(z,n)))
    else if (length(obj) == 4 || (is.matrix(obj) && dim(obj)[2] == 4))
	return(obj %*% translationMatrix(x,y,z))
    else stop("Unsupported object for translation")
}
	
scale3d.default <- function(obj,x,y,z,...) {
    if (is.matrix(obj)) n <- dim(obj)[1]
    else n <- 1
        
    if (length(obj) == 3 || (is.matrix(obj) && dim(obj)[2] == 3))
    	return(obj * cbind(rep(x,n), rep(y,n), rep(z,n)))
    else if (length(obj) == 4 || (is.matrix(obj) && dim(obj)[2] == 4))
    	return(obj %*% scaleMatrix(x,y,z))
    else stop("Unsupported object for scaling")
}

rotate3d.default <- function(obj,angle,x,y,z,matrix,...) {
    if (length(obj) == 3 || (is.matrix(obj) && dim(obj)[2] == 3))
    	return(asEuclidean(asHomogeneous(obj) %*% rotationMatrix(angle,x,y,z,matrix)))
    else if (length(obj) == 4 || (is.matrix(obj) && dim(obj)[2] == 4))
    	return(obj %*% rotationMatrix(angle,x,y,z,matrix))
    else stop("Unsupported object for rotation")
}    

test_that("asEuclidean works", {
  expect_equal(asEuclidean(c(1,2,3)), asEuclidean(matrix(c(1,2,3), ncol = 3)))
  expect_equal(asEuclidean(c(1,2,3)), asEuclidean(matrix(c(1,2,3,1), ncol = 4)))
  expect_equal(asEuclidean(c(1,2,3)), asEuclidean(c(2,4,6,2)))
  expect_equal(dim(asEuclidean(1:24)), c(6,3))             
})

test_that("asEuclidean2 works", {
  expect_equal(asEuclidean2(c(1,2,3)), asEuclidean2(matrix(c(1,2,3), nrow = 3)))
  expect_equal(asEuclidean2(c(1,2,3)), asEuclidean2(matrix(c(1,2,3,1), nrow = 4)))
  expect_equal(asEuclidean2(c(1,2,3)), asEuclidean2(c(2,4,6,2)))
  expect_equal(dim(asEuclidean2(1:24)), c(3,6))                 
})

test_that("asHomogeneous works", {
  expect_equal(asHomogeneous(c(1,2,3)), asHomogeneous(matrix(c(1,2,3), ncol = 3)))
  expect_equal(asHomogeneous(c(1,2,3)), asHomogeneous(matrix(c(1,2,3,1), ncol = 4)))
  expect_equal(asHomogeneous(c(1,2,3)), asHomogeneous(c(1,2,3,1)))
  expect_equal(dim(asHomogeneous(1:24)), c(8,4))             
})

test_that("asHomogeneous2 works", {
  expect_equal(asHomogeneous2(c(1,2,3)), asHomogeneous2(matrix(c(1,2,3), nrow = 3)))
  expect_equal(asHomogeneous2(c(1,2,3)), asHomogeneous2(matrix(c(1,2,3,1), nrow = 4)))
  expect_equal(asHomogeneous2(c(1,2,3)), asHomogeneous2(c(1,2,3,1)))
  expect_equal(dim(asHomogeneous2(1:24)), c(4,8))
})

## Original by Barry Rowlingson, R-help, 1/10/2010
## Modified by Michael Friendly: added barblen (to specify absolute barb length)
## Modified by DJM: multiple changes

arrow3d <- function(p0=c(1,1,1), p1=c(0,0,0), barblen, s=1/3, theta=pi/12, 
		     type = c("extrusion", "lines", "flat", "rotation"), 
		     n = 3, 
		     width = 1/3, 
		     thickness = 0.618*width,
		     spriteOrigin = NULL, 
		     plot = TRUE, ...) {
 ##      p0: start point
 ##      p1: end point
 ## barblen: length of barb
 ##       s: length of barb as fraction of line length (unless barblen is specified)
 ##   theta: opening angle of barbs
 ##    type: type of arrow to draw 
 ##       n: number of barbs
 ##   width: width of shaft as fraction of barb width
 ##   thickness: thickness of shaft as fraction of barb width	
 ##  spriteOrigin: origin if drawn as sprite
 ##     ...: args passed to lines3d for line styling
 ##
 ## Returns (invisibly): integer ID(s) of the shape added to the scene

 type <- match.arg(type)
 
 nbarbs <- if (type == "lines") n else 2
 
 ## Work in isometric display coordinates
 
 save <- par3d(FOV = 0)
 
 # Compute the center line in window
 # coordinates
 xyz <- rgl.user2window(rbind(p0, p1))
 p0 <- xyz[1,]
 p1 <- xyz[2,]
 
 ## rotational angles of barbs
 phi <- seq(pi/nbarbs, 2*pi-pi/nbarbs, length.out = nbarbs)

 ## length of line
 lp <- sqrt(sum((p1-p0)^2))

 if (missing(barblen))
   barblen <- s*lp
 else
   s <- barblen/lp
 
 ## point down the line where the barb ends line up
 cpt <- p1 + s*cos(theta)*(p0-p1)

 ## need to find a right-angle to the line. 
 gs <- GramSchmidt(p1-p0, c(0,0,-1), c(1,0,0))
 r <- gs[2,]

 ## now compute the barb end points and draw:
 for(i in seq_along(phi)) {
   ptb <- rotate3d(r,phi[i],(p1-p0)[1],(p1-p0)[2],(p1-p0)[3])
   xyz <- rbind(xyz, p1, cpt + barblen*sin(theta)*ptb)
 }
 if (type != "lines") {
   xyz <- xyz[c(3, # 1 head
   	        6, # 2 end of barb 1
   	        6, # 3 end of barb 1 again (to be shrunk)
   	        1, # 4 end of line (to be pushed out)
   	        1, # 5 end of line
   	        1, # 6 end of line (to be pushed the other way)
   	        4, # 7 end of barb 2 (to be shrunk)
   	        4, # 8 end of barb 2
   	        3),] # 9 head
   mid <- (xyz[2,] + xyz[8,])/2
   xyz[3,] <- mid + width*(xyz[2,] - mid)
   xyz[7,] <- mid + width*(xyz[8,] - mid)
   xyz[4,] <- xyz[4,] + xyz[3,] - mid
   xyz[6,] <- xyz[6,] + xyz[7,] - mid
 }
 
 if (type %in% c("extrusion", "rotation")) {
   xyz <- xyz %*% t(gs)
   if (type == "extrusion") {
     thickness <- thickness*sqrt(sum((xyz[2,]-xyz[8,])^2))
     ext <- extrude3d(xyz[,c(1,3)], thickness = thickness)
   } else {
     mid <- xyz[1,3]
     xyz[,3] <- abs(xyz[,3] - mid)
     xyz <- xyz[5:9,]
     ext <- turn3d(xyz[,c(1,3)], n = n)
     ext$vb[2,] <- ext$vb[2,] + mid
     thickness <- 0
   }
   ext$vb <- ext$vb[c(1,3,2,4),]
   ext$vb[2,] <- ext$vb[2,] + xyz[1,2] - thickness/2
   ext$vb[1:3,] <- t(gs) %*% ext$vb[1:3,]
   ext$vb[1:3,] <- t(rgl.window2user(t(ext$vb[1:3,])))
 } else
   xyz <- rgl.window2user(xyz)
 par3d(save)
 if (plot) {
   if (type == "flat")
     id <- polygon3d(xyz, ...)
   else if (type %in% c("extrusion", "rotation"))
     id <- shade3d(ext, ...)
   else
     id <- segments3d(xyz, ...)
 
   if (is.null(spriteOrigin))
     lowlevel(id)
   else
     sprites3d(spriteOrigin, shapes=id)
  } else {
    if (type %in% c("extrusion", "rotation"))
      ext
    else
      xyz
  }
}

currentSubscene3d <- function(dev = cur3d()) {
  .C(rgl_getsubsceneid, id = 1L, dev = as.integer(dev))$id
}

subsceneInfo <- function(id = NA, embeddings, recursive = FALSE) {
  if (is.na(id)) 
    id <- currentSubscene3d()
  else if (is.character(id) && id == "root") 
    id <- .C(rgl_getsubsceneid, id = 0L, dev = as.integer(cur3d()))$id
  if (!id) stop("No subscene info available.")
  id <- as.integer(id)
  result <- list(id = id)
  
  parent <- .C(rgl_getsubsceneparent, id = id)$id
  
  if (is.na(parent)) stop(gettextf("Subscene %s not found", id), domain = NA)
  if (!parent) parent <- NULL
  result[["parent"]] <- parent
  
  n <- .C(rgl_getsubscenechildcount, id = id, n = integer(1))$n
  if (n) {
    children <- .C(rgl_getsubscenechildren, id = id, children=integer(n))$children
    if (recursive) {
      childlist <- list()
      for (i in seq_len(n))
        childlist[[i]] <- subsceneInfo(id = children[i], recursive = TRUE)
      result[["children"]] <- childlist
    } else
      result[["children"]] <- children
  }
  
  embeddingNames <- c("inherit", "modify", "replace")
  if (!missing(embeddings)) {
    embeddings <- pmatch(embeddings, embeddingNames, duplicates.ok = TRUE)
    if (any(is.na(embeddings)) || length(embeddings) != 4)
      stop(gettextf("Four embeddings must be specified; names chosen from %s", 
      	            paste(dQuote(embeddingNames), collapse=", ")), domain = NA)
    if (embeddings[4] == "modify")
      stop("The mouseMode embedding cannot be 'modify'")
    .C(rgl_setEmbeddings, id = id, embeddings = as.integer(embeddings))
  }
  embeddings <- .C(rgl_getEmbeddings, id = id, embeddings = integer(4))$embeddings
  embeddings <- embeddingNames[embeddings]
  names(embeddings) <- c("viewport", "projection", "model", "mouse")
  result[["embeddings"]] <- embeddings
  result
}

newSubscene3d <- function(viewport = "replace", 
                       projection = "replace",
		                   model = "replace",
		                   mouseMode = "inherit",
                       parent = currentSubscene3d(), 
		                   copyLights = TRUE,
                       copyShapes = FALSE,
                       copyBBoxDeco = copyShapes,
                       copyBackground = FALSE,
                       newviewport,
                       ignoreExtent) {
  embedding <- c("inherit", "modify", "replace")
  viewport <- pmatch(viewport, embedding)
  projection <- pmatch(projection, embedding)
  model <- pmatch(model, embedding)
  mouseMode <- pmatch(mouseMode, embedding)
  
  if (missing(ignoreExtent))
    ignoreExtent <- model == 3
  stopifnot(length(viewport) == 1L, length(projection) == 1L, 
            length(model) == 1L, length(mouseMode) ==1L,
            mouseMode != 2L,
            !is.na(viewport), !is.na(projection), !is.na(model))
  embedding <- c(viewport, projection, model, mouseMode)
  
  id <- .C(rgl_newsubscene, id = integer(1), parent = as.integer(parent),
               embedding = as.integer(embedding),
               as.integer(ignoreExtent))$id
               
  if (id) {
    if (copyLights || copyShapes || copyBBoxDeco || copyBackground) {
      useSubscene3d(parent)
      ids <- ids3d(type = 
        c("lights", "shapes", "bboxdeco", "background")[c(copyLights, copyShapes, copyBBoxDeco, copyBackground)])$id
      if (length(ids)) 
        addToSubscene3d(ids, subscene = id)
    }
    useSubscene3d(id)
    if (!missing(newviewport)) {
      embedding <- subsceneInfo(id)$embeddings
      if (embedding[1] > 1)
        par3d(viewport = as.integer(newviewport))
    }
  } else
    stop("Subscene creation failed")
  lowlevel(id)
}

useSubscene3d <- function(subscene) {
  result <- .C(rgl_setsubscene, id=as.integer(subscene))$id
  if (!result) stop(gettextf("Subscene %d not found.", subscene), domain = NA)
  invisible(result)
}

addToSubscene3d <- function(ids = tagged3d(tags), tags, subscene = currentSubscene3d()) {
  ids <- as.integer(ids)
  dups <- intersect(ids, ids3d("all", subscene)$id)
  if (length(dups))
    stop(gettextf("Cannot add %s, already present", paste(dups, collapse = ", ")), domain = NA)
  result <- .C(rgl_addtosubscene, success = as.integer(subscene), 
     n = as.integer(length(ids)), ids = ids)$success
  if (!result)
    stop(gettextf("Failed to add objects to subscene %s", subscene), domain = NA)
  lowlevel(subscene)
}

delFromSubscene3d <- function(ids = tagged3d(tags), tags, subscene = currentSubscene3d()) {
  result <- .C(rgl_delfromsubscene, success = as.integer(subscene), 
     n = as.integer(length(ids)), ids = as.integer(ids))$success
  if (!result)
    stop(gettextf("Failed to delete objects from subscene %s", subscene), domain = NA)
  lowlevel(subscene)
}

# This destroys any objects that are in the scene but
# not in either the protect vector or visible in a subscene

gc3d <- function(protect=NULL) {
  protect <- as.integer(protect)
  invisible( .C(rgl_gc, n = length(protect), protect)$n )
}

subsceneList <- function(value, window = cur3d()) {
  alllists <- .rglEnv$subsceneLists  
  # This cleans up lists for closed windows:
  alllists <- alllists[names(alllists) %in% rgl.dev.list()]
  if (!missing(value)) {
    if (is.null(alllists)) alllists <- list()
    alllists[[as.character(window)]] <- value
    assign("subsceneLists", alllists, envir = .rglEnv)
  }
  if (is.null(alllists)) return(NULL)
  else return(alllists[[as.character(window)]])
}

next3d <- function(current = NA, clear = TRUE, reuse = TRUE) {
  .check3d()
  if (is.na(current))
    current <- currentSubscene3d()
  subscenes <- subsceneList()
  while (!is.null(subscenes) && !(current %in% subscenes))
    subscenes <- attr(subscenes, "prev")
  if (is.null(subscenes)) 
    subscenes <- current
  
  this <- which(current == subscenes)
  if (reuse && !nrow(ids3d(subscene = current))) {
    # do nothing
  } else if (this == length(subscenes)) 
    this <- 1
  else 
    this <- this + 1
  
  repeat{
    current <- subscenes[this]
    result <- try(useSubscene3d(current))
    if (inherits(result, "try-error")) {
      subsceneList(subscenes <- subscenes[-this])
      if (length(subscenes) == 0)
      	stop("'subsceneList()' contained no valid subscenes")
      if (this > length(subscenes)) this <- 1
    } else break
  }
      
  if (clear)
    clear3d(subscene = current)
}
  
clearSubsceneList <- function(delete = currentSubscene3d() %in% subsceneList(),
                              window = cur3d()) {
  if (!missing(window))
    set3d(window)  
  thelist <- subsceneList()
  if (delete && length(thelist)) {
    parent <- subsceneInfo(thelist[1])$parent
    if (is.null(parent))
      parent <- rootSubscene()
    pop3d(type="subscene", id=thelist)
    useSubscene3d(parent)
    gc3d()
  }
  subsceneList(attr(thelist, "prev"))
  invisible(currentSubscene3d())
}

mfrow3d <- function(nr, nc, byrow = TRUE, parent = NA, sharedMouse = FALSE,
                           ...) {
  stopifnot(nr >= 1, nc >= 1)
  .check3d()
  if (missing(parent))
    clearSubsceneList()
  if (is.na(parent))
    parent <- currentSubscene3d()
  useSubscene3d(parent)
  result <- integer(nr*nc)
  parentvp <- par3d("viewport")
  if (byrow)
    for (i in seq_len(nr))
      for (j in seq_len(nc)) {
        newvp <- c(parentvp[1] + (j - 1)*parentvp[3]/nc,
                   parentvp[2] + (nr - i)*parentvp[4]/nr,
                   parentvp[3]/nc, parentvp[4]/nr)
        result[(i-1)*nc + j] <- newSubscene3d(newviewport = newvp, parent = parent, ...)
      }
  else
    for (j in seq_len(nc))
      for (i in seq_len(nr)) {
        newvp <- c(parentvp[1] + (j - 1)*parentvp[3]/nc,
                   parentvp[2] + (nr - i)*parentvp[4]/nr,
                   parentvp[3]/nc, parentvp[4]/nr)
        result[(j-1)*nr + i] <- newSubscene3d(newviewport = newvp, parent = parent, ...)
      }
  if (sharedMouse)
    for (sub in result)
      par3d(listeners = result, subscene = sub)
  useSubscene3d(result[1])
  attr(result, "prev") <- subsceneList()
  subsceneList(result)
  invisible(result)
}

layout3d <- function(mat, widths = rep.int(1, ncol(mat)), 
                          heights = rep.int(1, nrow(mat)),
                          parent = NA, sharedMouse = FALSE,
                          ...) {
  storage.mode(mat) <- "integer"
  mat <- as.matrix(mat)
  num.figures <- max(mat)
  if (!all(seq_len(num.figures) %in% as.integer(mat)))
    stop(gettextf("Layout matrix must contain at least one reference\nto each of the values {1 ... %d}\n", 
            num.figures), domain = NA)

  .check3d()
  if (missing(parent))
    clearSubsceneList()
  if (is.na(parent))
    parent <- currentSubscene3d()
  useSubscene3d(parent)
  parentvp <- par3d("viewport")
  
  widths <- parentvp["width"]*widths/sum(widths)
  heights <- parentvp["height"]*heights/sum(heights)
  xs <- c(0, cumsum(widths))
  ys <- rev(c(0, cumsum(rev(heights))))[-1]
  
  result <- integer(num.figures)
  for (i in seq_len(num.figures)) {
    rows <- range(row(mat)[mat == i])
    cols <- range(col(mat)[mat == i])
    newvp <- c(xs[cols[1]], ys[rows[2]], sum(widths[cols[1]:cols[2]]), sum(heights[rows[1]:rows[2]]))
    result[i] <- newSubscene3d(newviewport = newvp, parent = parent,  ...)
  }
  if (sharedMouse)
    for (sub in result)
      par3d(listeners = result, subscene = sub)
  useSubscene3d(result[1])
  attr(result, "prev") <- subsceneList()
  subsceneList(result)
  invisible(result)
}

library(rgl)

test_that("subscenes work", {
	
	# Show the Earth with a cutout by using clipplanes in subscenes
	
	lat <- matrix(seq(90, -90, len = 50)*pi/180, 50, 50, byrow = TRUE)
	long <- matrix(seq(-180, 180, len = 50)*pi/180, 50, 50)
	
	r <- 6378.1 # radius of Earth in km
	x <- r*cos(lat)*cos(long)
	y <- r*cos(lat)*sin(long)
	z <- r*sin(lat)
	
	open3d()
	obj <- surface3d(x, y, z, col = "white", 
									 texture = system.file("textures/worldsmall.png", package = "rgl"), 
									 specular = "black", axes = FALSE, box = FALSE, xlab = "", ylab = "", zlab = "",
									 normal_x = x, normal_y = y, normal_z = z)
	
	cols <- c(rep("chocolate4", 4), rep("burlywood1", 4), "darkgoldenrod1")
	rs <- c(6350, 5639, 4928.5, 4207, 3486, 
					(3486 + 2351)/2, 2351, (2351 + 1216)/2, 1216)
	for (i in seq_along(rs)) 
		obj <- c(obj, spheres3d(0, 0, col = cols[i], radius = rs[i]))
	
	root <- currentSubscene3d()
	
	newSubscene3d("inherit", "inherit", "inherit", copyShapes = TRUE, parent = root)
	clipplanes3d(1, 0, 0, 0)
	
	newSubscene3d("inherit", "inherit", "inherit", copyShapes = TRUE, parent = root)
	clipplanes3d(0, 1, 0, 0)
	
	newSubscene3d("inherit", "inherit", "inherit", copyShapes = TRUE, parent = root)
	clipplanes3d(0, 0, 1, 0)
	
	# Now delete the objects from the root subscene, to reveal the clipping planes
	useSubscene3d(root)
	delFromSubscene3d(obj)
  expect_known_scene("subscenes")
})

rgl.callback.env <- new.env(parent = emptyenv())

findSubscene <- function(subscene, id) {
  if (subscene$id == id)
    return(subscene)
  subscenes <- subscene$subscenes
  if (!is.null(subscenes)) {
    for (i in seq_along(subscenes)) {
      if (!is.null(result <- findSubscene(subscenes[[i]], id)))
        return(result)
    }
  }
  NULL
}

findBboxdeco <- function(scene, subscene, root = scene$rootSubscene, guess = NULL) {
  active <- root
  for (id in active$objects)
    if (scene$objects[[as.character(id)]]$type == "bboxdeco")
      guess <- id
  if (active$id == subscene)
    if (is.null(guess))
      return(NULL)
    else
      return(scene$objects[[as.character(guess)]])
  for (sub in active$subscenes)
    if (!is.null(bbox <- findBboxdeco(scene, subscene, sub, guess)))
      return(bbox)
  return(NULL)
}

replaceSubscene <- function(subscene, id, newvalue) {
  if (as.character(subscene$id) == id)
    return(newvalue)
  subscenes <- subscene$subscenes
  if (!is.null(subscenes))
    for (i in seq_along(subscenes))
      if (!is.null(result <- replaceSubscene(subscenes[[i]], id, newvalue))) {
        subscene$subscenes[[i]] <- result
        return(subscene)
      }
  NULL
}

getfn <- function(fn, up = 2) {
  if (is.character(fn))
    result <- eval(parse(text = fn), envir = parent.frame(up))
  else
    result <- fn
  if (!is.function(result))
    stop(fn, " is not a function.")
  result
}

setUserCallbacks <- function(button = NULL, begin = NULL, update = NULL, end = NULL,
                             rotate = NULL,
                             javascript = NULL,
                             subscene = scene$rootSubscene$id,
			                       scene = scene3d(minimal = FALSE),
			                       applyToScene = TRUE,
			                       applyToDev = missing(scene)) {

  force(applyToDev)
  stopifnot(inherits(scene, "rglscene"))
  subscene <- as.character(subscene)

  sub <- findSubscene(scene$rootSubscene, subscene)
  if (is.null(sub))
    stop("subscene ", subscene, " not found.")
  if (is.null(sub$par3d) || is.null(sub$par3d$mouseMode) 
      || is.null(sub$embeddings)) 
    stop("Internal error:  subscene missing mouseMode or embeddings")
  if (is.null(sub$callbacks))
    sub$callbacks <- list()
  if (!is.null(button)) {
    if (is.numeric(button))
      button <- c("none", "left", "right", "middle", "wheel")[button + 1]
    sub$par3d$mouseMode[button] <- "user"
    sub$callbacks[[button]] <- list(begin = begin, update = update,
                                    end = end, rotate = rotate)
    sub$embeddings["mouse"] <- "replace"
  } 
  javascript <- paste(c(scene$javascript, javascript),
                      collapse = "\n")
  if (applyToScene) {
    scene$rootSubscene <- replaceSubscene(scene$rootSubscene,
                                        subscene, sub)
    scene$javascript <- javascript
  }
  if (applyToDev) {
    dev <- cur3d()
    devname <- paste0("dev", dev)
    callbacks <- rgl.callback.env[[devname]]
    if (is.null(callbacks))
      callbacks <- list()
    callbacks[[paste0("sub", subscene)]] <- sub$callbacks
    callbacks$javascript <- javascript
    rgl.callback.env[[devname]] <- callbacks
    if (dev > 0) {
      callbacks <- sub$callbacks
      if (!is.null(button)) {
        fns <- callbacks[[button]]
        if (!is.null(fns)) {
          for (f in c("begin", "update", "end", "rotate"))
            if (!is.null(fns[[f]]))
              fns[[f]] <- getfn(fns[[f]])
          callbacks[[button]] <- fns
          if (button == "wheel" && is.function(fns$rotate))
            rgl.setWheelCallback(rotate = fns$rotate, dev = dev, subscene = subscene)
          fns$rotate <- NULL
          if (any(vapply(fns, is.function, TRUE)))
            do.call(rgl.setMouseCallbacks, 
                    c(list(button = c(none = 0, left = 1, right = 2, middle = 3, wheel = 4)[button],
                           dev = dev, subscene = subscene),
                      fns))
        }
      }
    }
  }
  invisible(scene)
}

setAxisCallbacks <- function(axes, fns,
                             javascript = NULL,
                             subscene = scene$rootSubscene$id,
                             scene = scene3d(minimal = FALSE),
                             applyToScene = TRUE,
                             applyToDev = missing(scene)) {
  
  force(applyToDev)
  stopifnot(inherits(scene, "rglscene"))
  subscene <- as.character(subscene)
  
  if (!is.character(axes))
    axes <- c("x", "y", "z")[axes]
  else
    axes <- tolower(axes)
  stopifnot(all(axes %in% c("x", "y", "z")))
  
  if (!is.list(fns))
    fns <- list(fns)
  
  n <- max(length(axes), length(fns))
  axes <- rep(axes, length.out = n)
  fns <- rep(fns, length.out = n)
  
  # Call each function once to set things up before grabbing
  # the scene
  
  for (i in seq_len(n))
    getfn(fns[[i]])(paste0(axes[i], "--"))
  
  sub <- findSubscene(scene$rootSubscene, subscene)
  if (is.null(sub))
    stop("subscene ", subscene, " not found.")
  bbox <- findBboxdeco(scene, subscene)
  if (is.null(bbox)) 
    stop("No bbox decoration found.")
  if (is.null(bbox$callbacks))
    bbox$callbacks <- list()
  for (i in seq_len(n))
    bbox$callbacks[[axes[i]]] <- fns[[i]]

  javascript <- paste(c(scene$javascript, javascript),
                      collapse = "\n")
  if (applyToScene) {
    scene$objects[[as.character(bbox$id)]] <- bbox
    scene$javascript <- javascript
  }
  if (applyToDev) {
    dev <- cur3d()
    devname <- paste0("dev", dev)
    callbacks <- rgl.callback.env[[devname]]
    if (is.null(callbacks))
      callbacks <- list()
    callbacks[[paste0("bbox", bbox$id)]] <- bbox$callbacks
    callbacks$javascript <- javascript
    rgl.callback.env[[devname]] <- callbacks
    if (dev > 0) {
      callbacks <- bbox$callbacks
      for (i in seq_len(n)) {
        fn <- getfn(fns[[i]])
        axis <- match(axes[i], c("x", "y", "z"))
        .Call(rgl_setAxisCallback, fn, as.integer(dev), as.integer(subscene), as.integer(axis - 1))
      }
    }
  }
  invisible(scene)
}

as.triangles3d <- function(obj, ...) 
  UseMethod("as.triangles3d")

as.triangles3d.mesh3d <- function(obj, attribute = c("vertices", "normals", "texcoords",
                                                     "colors"),
                                  ...) {
  indices <- NULL
  if (!is.null(obj$it)) {
    indices <- c(obj$it)
  } 
  if (!is.null(obj$ib)) {
    indices <- c(indices, c(obj$ib[1:3,], obj$ib[c(1,3,4),]))
  }
  if (!is.null(indices)) 
    switch(match.arg(attribute),
           vertices = t(obj$vb[1:3, indices])/obj$vb[4,indices],
           normals  = if (!is.null(obj$normals))
             if (nrow(obj$normals) == 4)
               t(obj$normals[1:3, indices]/obj$normals[4,indices])
             else
               t(obj$normals[, indices]),
           texcoords = if (!is.null(obj$texcoords))
             t(obj$texcoords[, indices]),
           colors = if (!is.null(obj$material) && !is.null(obj$material$color)) {
             col <- t(col2rgb(rep(obj$material$color, length.out = max(indices))))
             alpha <- if (is.null(obj$material$alpha)) 1
             else obj$material$alpha
             alpha <- rep(alpha, length.out = max(indices))
             cbind(col, alpha)[indices,]
           })
}

as.triangles3d.rglId <- function(obj, 
                                 attribute = c("vertices", "normals", "texcoords",
                                               "colors"),
                                 subscene = NA,
                                 ...) {
  attribute <- match.arg(attribute)
  ids <- ids3d(subscene = subscene)
  ids <- ids[ids$id %in% obj,]
  result <- NULL
  for (i in seq_len(nrow(ids))) {
    id <- ids[i, "id"]
    nvert <- rgl.attrib.count(id, "vertices")
    attrib <- rgl.attrib(id, attribute)
    if (nrow(attrib)) {
      if (nrow(attrib) < nvert)
        attrib <- apply(attrib, 2, function(col) rep(col, length.out = nvert))
      type <- ids[i, "type"]
      result <- rbind(result, switch(as.character(type),
        triangles =,              
        planes = attrib,
        quads = {
          nquads <- nrow(attrib)/4
          attrib[4*rep(seq_len(nquads) - 1, each = 6) + c(1,2,3,1,3,4),,drop=FALSE]
        },
        surface = {
          dim <- rgl.attrib(id, "dim")
          ul <- rep(2:dim[1], dim[2]-1) + dim[1]*rep(0:(dim[2]-2), each=dim[1]-1) 
          if (rgl.attrib(id, "flags")["flipped",]) 
            indices <- c(rbind(c(ul-1, ul-1+dim[1]),
                               c(ul, ul), 
                               c(ul-1+dim[1], ul+dim[1])))
          else
            indices <- c(rbind(c(ul, ul), 
                               c(ul-1, ul-1+dim[1]),
                               c(ul-1+dim[1], ul+dim[1])))

          attrib[indices,,drop = FALSE]
        }, 
        NULL))
    }
  }
  result
}

# Tests of bbox3d improvements

library(rgl)

x <- cube3d(col="red", front="culled", back="filled"); open3d(); shade3d(x); rglwidget()

open3d();points3d(1:10, 11:20, 21:30); bbox3d(col="red", lit=FALSE,draw_front=TRUE); rglwidget()

open3d();points3d(1:10, 11:20, 21:30); bbox3d(col="red", lit=TRUE,draw_front=TRUE); rglwidget()

open3d();points3d(1:10, 11:20, 21:30); bbox3d(col="red", lit=TRUE,draw_front=TRUE, front="lines", back = "lines"); rglwidget()

open3d();points3d(1:10, 11:20, 21:30); bbox3d(col="red", lit=TRUE,draw_front=TRUE, front="lines", back = "lines"); rglwidget()

open3d();points3d(1:10, 11:20, 21:30); bbox3d(col="red", lit=TRUE,draw_front=TRUE, front="lines", back = "filled"); rglwidget()

open3d();points3d(1:10, 11:20, 21:30); bbox3d(col="red", lit=TRUE,draw_front=FALSE, front="lines", back = "lines"); rglwidget()

open3d();points3d(1:10, 11:20, 21:30); bbox3d(col="red", lit=TRUE,draw_front=FALSE, front="filled", back = "culled"); rglwidget()

open3d();points3d(1:10, 11:20, 21:30); bbox3d(col="red", lit=TRUE,draw_front=FALSE, front="culled", back = "filled"); rglwidget()


open3d();points3d(1:10, 11:20, 21:30); bbox3d(col="red", lit=FALSE,draw_front=FALSE); rglwidget()

open3d();points3d(1:10, 11:20, 21:30); bbox3d(col="red", lit=TRUE,draw_front=FALSE); rglwidget()

example(bbox3d); rglwidget()

open3d(); spheres3d(rnorm(10), rnorm(10), rnorm(10), 
                    radius = runif(10), color = rainbow(10)); rglwidget()

open3d(); spheres3d(rnorm(10), rnorm(10), rnorm(10), 
                    radius = runif(10), color = rainbow(10),
                    front = "lines", back="culled"); rglwidget()

GramSchmidt <- function(v1, v2, v3, order=1:3) {
  A <- rbind(v1, v2, v3)
  A <- A[order,]
  v1 <- A[1,]
  v2 <- A[2,]
  v3 <- A[3,]
  if (isTRUE(all.equal(as.numeric(v1), c(0,0,0)))) v1 <- xprod(v2, v3)
  v1 <- normalize(v1)
  v2 <- v2 - sum(v2*v1)*v1
  if (isTRUE(all.equal(as.numeric(v2), c(0,0,0)))) v2 <- xprod(v3, v1)
  v2 <- normalize(v2)
  v3 <- v3 - sum(v3*v1)*v1 - sum(v3*v2)*v2
  if (isTRUE(all.equal(as.numeric(v3), c(0,0,0)))) v3 <- xprod(v1, v2)
  v3 <- normalize(v3)
  rbind(v1, v2, v3)[order(order),]
}

cylinder3d <- function(center, radius = 1, twist = 0, e1 = NULL, e2 = NULL, e3 = NULL, 
                       sides = 8, section = NULL, closed = 0, 
		       rotationMinimizing = is.null(e2) && is.null(e3),
		       debug = FALSE, keepVars = FALSE, ...) {
  force(rotationMinimizing) # Need this since e2 and e3 get changed later	
  center <- as.matrix(as.data.frame(xyz.coords(center)[c("x", "y", "z")]))
  n <- nrow(center)  
  inds <- seq_len(n)
  if (closed > 0) {
    ind0 <- c(n-1-closed, n-closed, inds)
    ind1 <- c(n-closed, inds, 1+closed) # nolint
    ind2 <- c(inds, 1+closed, 2+closed)
  } else {
    ind0 <- c(1, 1, inds)
    ind1 <- c(1, inds, n) # nolint
    ind2 <- c(inds, n, n)
  }  
  
  missings <- c(e1=is.null(e1), e2=is.null(e2), e3=is.null(e3))
  
  fixup <- function(coord) {
    usable <- apply(coord, 1, function(v) all(is.finite(v)) & (veclen(v) > 0))
    if (!any(usable) ) stop(gettextf("No usable coordinate values in '%s'", 
    				     deparse(substitute(coord))), domain = NA)
    firstgood <- min(which(usable))
    inds <- seq_len(n)    
    if (firstgood > 1) {
      coord[inds[inds < firstgood],] <- coord[rep(firstgood,firstgood-1),]
      usable[seq_len(firstgood)] <- TRUE
    }
    for (i in inds[-1]) inds[i] <- ifelse(usable[i], inds[i], inds[i-1])
    coord[inds,]
  }
  
  if (!is.null(e1)) {
    e1 <- as.matrix(as.data.frame(xyz.coords(e1)[c("x", "y", "z")]))
    e1 <- e1[rep(seq_len(nrow(e1)), length.out = n),] 
  } else 
    e1 <- (center[ind2,] - center[ind0,])[inds,]
  
  # Fix up degenerate cases by repeating existing ones, or using arbitrary ones
  zeros <- rowSums(e1^2) == 0
  if (all(zeros)) {
    e1[,1] <- 1
    zeros <- FALSE
  } else if (any(zeros)) {
    e1[1,] <- e1[which(!zeros)[1],]
    zeros[1] <- FALSE
    if (any(zeros)) {
      zeros <- which(zeros)
      for (i in zeros)
        e1[i,] <- e1[i-1,]
    }
  }
  if (!is.null(e2)) {
    e2 <- as.matrix(as.data.frame(xyz.coords(e2)[c("x", "y", "z")]))
    e2 <- e2[rep(seq_len(nrow(e2)), length.out = n),] 
  } else
    e2 <- (e1[ind2,] - e1[ind0,])[inds,]
    
  # Fix up degenerate e2's similarly, then force different than e1
  zeros <- rowSums(e2^2) == 0
  if (all(zeros)) {
    e2[,2] <- 1
    zeros <- FALSE
  } else if (any(zeros)) {
    e2[1,] <- e2[which(!zeros)[1],]
    zeros[1] <- FALSE
    if (any(zeros)) {
      zeros <- which(zeros)
      for (i in zeros)
        e2[i,] <- e2[i-1,]
    }
  }
  parallel <- sapply(inds, function(i) all(xprod(e1[i,], e2[i,])  == 0))
  if (any(parallel)) {
    # rotate in the xy plane
    e2[parallel,] <- cbind(-e2[parallel,2], e2[parallel,1], e2[parallel,3])
    parallel <- sapply(inds, function(i) all(xprod(e1[i,], e2[i,])  == 0))
    if (any(parallel)) {
      # if any are still parallel, they must be the z axis
      e2[parallel,1] <- 1
      e2[parallel,3] <- 0
    }
  }
  
  if (!is.null(e3)) {
    e3 <- as.matrix(as.data.frame(xyz.coords(e3)[c("x", "y", "z")]))
    e3 <- e3[rep(seq_len(nrow(e3)), length.out = n),] 
  } else {
    e3 <- matrix(NA_real_, n, 3)
    for (i in inds) e3[i,] <- xprod(e1[i,], e2[i,])
  }

  for (i in inds) {
    A <- GramSchmidt(e1[i,], e2[i,], e3[i,], order=order(missings))
    e1[i,] <- A[1,]
    e2[i,] <- A[2,]
    e3[i,] <- A[3,]
  }
  e1 <- fixup(e1)
  e2 <- fixup(e2)
  e3 <- fixup(e3)
  
  if (rotationMinimizing) {
    # Keep e1 and the first entries in e2 and e3,
    # modify the rest according to Wang et al (2008).
    t <- e1
    r <- e2
    s <- e3
    for (i in seq_len(n-1)) {
      v1 <- center[i+1,] - center[i,]
      c1 <- sum(v1^2)
      rL <- r[i,] - (2/c1)*sum(v1*r[i,])*v1
      tL <- t[i,] - (2/c1)*sum(v1*t[i,])*v1
      v2 <- t[i+1,] - tL
      c2 <- sum(v2^2)
      r[i+1,] <- rL - (2/c2)*sum(v2*rL)*v2
      s[i+1,] <- xprod(t[i+1,], r[i+1,])
    }
    e2 <- r
    e3 <- s
  }

  radius <- rep(radius, length.out = n)
  twist <- rep(twist, length.out = n)
  
  if (debug) {
    for (i in inds) {
      segments3d(rbind(center[i,],center[i,]+e3[i,]*radius[i]*1.5,
                     center[i,],center[i,]+e2[i,]*radius[i]*1.5,
                     center[i,],center[i,]+e1[i,]*radius[i]*1.5), 
                 col = rep(c("red", "green", "blue"), each=2))
      text3d(center, texts=inds)
    }
  }
  
  if (closed > 0) {
    n <- n - closed + 1
    if (isTRUE(all.equal(e1[1,], e1[n,]))) {
      # Need to match starting and ending e2 as
      # well.
      angle <- atan2(sum(xprod(e2[1,], e2[n,])*e1[1,]),
    	             sum(e2[1,]*e2[n,]))
      twist <- twist + (twist[n] - twist[1] - angle)*(seq_len(n) - 1)/(n-1)
    }
  }
  if (is.null(section)) {
    theta <- seq(0, 2*pi, length.out = sides+1)[-1]
    section <- cbind(cos(theta), sin(theta), 0)
  } else 
    sides <- nrow(section)
    
  vertices <- matrix(0, 3, sides*n)
  indices <- matrix(0, 4, sides*(n-1))
  
  if (ncol(section) == 2)
    section <- cbind(section, 0)
    
  for (i in seq_len(n-1)) {
    transform <- rbind(e3[i,], e2[i,], e1[i,])
    p <- rotate3d(section, twist[i], 0,0,1)
    p <- radius[i] * p %*% transform
    p[,1] <- p[,1] + center[i,"x"]
    p[,2] <- p[,2] + center[i,"y"]
    p[,3] <- p[,3] + center[i,"z"]
    vertices[,(i-1)*sides+seq_len(sides)] <- t(p)
    for (j in seq_len(sides)) 
      indices[, (i-1)*sides + j] <- (c(0,0,1,1) + j) %% sides + 1 + 
                                    c((i-1)*sides, i*sides, i*sides, (i-1)*sides)
  }
  transform <- rbind(e3[n,], e2[n,], e1[n,])
  p <- rotate3d(section, twist[n], 0,0,1)
  p <- radius[n] * p %*% transform
  p[,1] <- p[,1] + center[n,"x"]
  p[,2] <- p[,2] + center[n,"y"]
  p[,3] <- p[,3] + center[n,"z"]
  vertices[,(n-1)*sides+seq_len(sides)] <- t(p)
  # Add end cap at start
  if (closed < 0) {
    vertices <- cbind(vertices, center[1,])
    triangles <- rbind(ncol(vertices), seq_len(sides), c(2:sides, 1))
  }
  # Add end cap at end
  if (closed < -1) {
    vertices <- cbind(vertices, center[n,])
    triangles <- cbind(triangles, rbind(ncol(vertices), c(2:sides, 1) + (n-1)*sides, 
                                    seq_len(sides) + (n-1)*sides))
  }
  
  result <- qmesh3d(vertices, indices, homogeneous=FALSE)
  if (closed > 0) { # Look for repeated vertices, and edit the links
    nv <- ncol(result$vb)
    for (i in seq_len(sides)) {
      dupe <- which(apply(result$vb[,(nv-sides+1):nv,drop=FALSE], 2, 
                          function(x) isTRUE(all.equal(x, result$vb[,i]))))+nv-sides
      for (j in dupe) {
        f <- result$ib == j
        result$ib[f] <- i
      }
    }
  } else if (closed < 0)
    result$it <- triangles
    
  if (keepVars)
    attr(result, "vars") <- environment()
  
  result$material <- .fixMaterialArgs2(...)
  
  result
}

drape3d <- function(obj, ...) 
  UseMethod("drape3d")

drape3d.default <- function(obj, ...) 
  drape3d(as.mesh3d(obj), ...)
  
drape3d.mesh3d <- function(obj, x, y = NULL, z = NULL,
    plot = TRUE, up = c(0, 0, 1), P = projectDown(up), ...) 
{
  # Takes segment number as input; returns
  # NULL if in no triangle, otherwise matrix of projected locations and triangle numbers.
  ztri <- function(i) {
    p <- psegs[,i]
    oo <- p[1] < TRI[1,1,] | p[1] > TRI[2,1,] |
      p[2] < TRI[1,2,] | p[2] > TRI[2,2,]
    result <- NULL
    lam <- numeric(3)
    for(j in which(!oo)) {
      ## get barycentric coords of p in projected triangle
      v <- pverts[,obj$it[,j]]  ## v[i,] vertices of projected triangle i
      D <- (v[2,2]-v[2,3]) * (v[1,1]-v[1,3]) +
        (v[1,3]-v[1,2]) * (v[2,1]-v[2,3])
      if (D == 0) next
      l <- (v[2,2]-v[2,3]) * (p[1]  -v[1,3]) +
        (v[1,3]-v[1,2]) * (p[2]  -v[2,3])
      lam[1] <- l/D
      if (lam[1] < 0 || lam[1] > 1) next   ## not in this triangle
      l <- (v[2,3]-v[2,1]) * (p[1]  -v[1,3]) +
        (v[1,1]-v[1,3]) * (p[2]  -v[2,3])
      lam[2] <- l/D
      if (lam[2] < 0 || lam[2] > 1) next   ## not in this triangle
      lam[3] <- 1-sum(lam[1:2])
      if (lam[3] < 0 || lam[3] > 1) next   ## not in this triangle
      v <- matrix(verts[,obj$it[,j]], 3,3)  ## Now v is vertices of original triangle
      result <- rbind(result, c(v %*% lam, j))
    }
    result
  }
  op <- function(v)
    v[if(v[1] > v[2]) c(2,1) else c(1,2)] ## orders pair
  
  obj <- as.tmesh3d(obj)
  
  verts <- obj$vb
  
  segs <- xyz.coords(x, y, z, recycle=TRUE)
  segs <- rbind(segs$x, segs$y, segs$z, 1)
  
  if (length(dim(P)) != 2 || !all(dim(P) == 4))
    stop("P should be a homogeneous coordinate matrix.")
  
  P <- t(P)   # The convention in rgl is row vectors on the left
              # but we'll be using column vectors on the right
  
  # Project the vertices, then get 1st two Euclidean coords
  pverts <- P %*% verts             
  pverts <- pverts[1:2,]/rep(pverts[4,], each = 2)
  # and switch verts to Euclidean:
  verts <- verts[1:3,]/rep(verts[4,], each = 3)
  
  psegs <- P %*% segs               # projected segments
  psegs <- psegs[1:2,]/rep(psegs[4,], each = 2)
  
  ## get unique point pairs making a triangle side
  tri <- matrix(NA,nrow=3*ncol(obj$it),ncol=2)
  n <- 0
  for (j in seq_len(ncol(obj$it))) {
    v <- obj$it[,j]
    tri[n<-n+1,] <- v[c(1,2)]
    tri[n<-n+1,] <- v[c(2,3)]
    tri[n<-n+1,] <- v[c(3,1)]
  }
  
  TRI <- array(NA,c(2,2,ncol(obj$it)))
  for(j in seq_len(ncol(obj$it))) {
    v <- obj$it[,j]       ## vertices of triangle
    TRI[,,j] <- matrix(c(range(pverts[1,v]),range(pverts[2,v])),2,2)
  }
  
  ## now TRI[,1,i] is x coord range for projected triangle i
  ## now TRI[,2,i] is y coord range for projected triangle i
  
  result <- matrix(numeric(), ncol = 3)
  
  p2 <- NA
  p2tri <- NULL
  
  for (i in seq_len(ncol(psegs))) {
    p1 <- p2
    p1tri <- p2tri
    if (!length(p1tri))
      zs <- NULL
    else
      zs <- cbind(p1tri, 0)  # First point
    
    p2 <- psegs[,i]
    if (any(is.na(p2))) {
      p2tri <- NULL
      next
    } else {
      p2tri <- ztri(i)
      zs <- rbind(zs, cbind(p2tri, 1)) # Last point
    }
    
    if (any(is.na(p1)))
      next

    ## add middle points    
    p21 <- p2 - p1

    s <- matrix(c(p1,p2),2,2)  ## speedup: winnow futile intersection calcs
    s <- t(apply(s,1,op))
    ## triangle seg x extent is all below or above line seg x extent
    sx <- (pverts[1,tri[,1]] < s[1,1] & pverts[1,tri[,2]] < s[1,1]) |
      (pverts[1,tri[,1]] > s[1,2] & pverts[1,tri[,2]] > s[1,2])
    ## triangle seg y extent is all below or above line seg y extent
    sy <- (pverts[2,tri[,1]] < s[2,1] & pverts[2,tri[,2]] < s[2,1]) |
      (pverts[2,tri[,1]] > s[2,2] & pverts[2,tri[,2]] > s[2,2])
    for(j in which(!sx & !sy)) {     ## possible intersections
      p3 <- pverts[,tri[j,1]]
      p4 <- pverts[,tri[j,2]]
      p43 <- p4-p3
      p31 <- p3-p1
      D <- -p21[1]*p43[2] + p21[2]*p43[1]
      if (D == 0) next                    ## parallel line segs
      T1<- -p31[1]*p43[2] + p31[2]*p43[1]
      t1 <- T1/D
      if (t1 < 0 || t1 > 1) next          ## not within p1 ... p2
      T2<-  p21[1]*p31[2] - p21[2]*p31[1]
      t2 <- T2/D
      if (t2 < 0 || t2 > 1) next          ## not within p3 ... p4
      v3 <- verts[,tri[j,1]]
      v43 <- verts[,tri[j,2]] - v3
      k <- (j+2)%/%3   # triangle number
      zs <- rbind(zs, c(v3+t2*v43, k, t1))## t1 along line seg & point value
    }
    
    if (length(zs)) {
      # Order by triangle to group them, then by t
      # within triangle 
      o <- order(zs[,4], zs[,5])
      zs <- zs[o, , drop = FALSE]
      # Only keep cases that are on the same triangle
      # Rounding may give us 1 or 3 intersections with a 
      # triangle; discard singletons, use first and last
      # for triples.
      k <- zs[,4]
      dup <- duplicated(k)
      nextdup <- c(dup[-1], FALSE)
      keep <- xor(dup, nextdup) # The first or last for a triangle
      zs <- zs[keep,,drop = FALSE]
      # Order by t value
      finish <- 2*seq_len(nrow(zs)/2)
      start <- finish - 1
      o <- order(zs[start, 5])
      start <- start[o]
      finish <- finish[o]
      # Drop zero length segments
      keep <- zs[start, 5] < zs[finish, 5]
      start <- start[keep]
      finish <- finish[keep]
      both <- as.numeric(rbind(start, finish))
      result <- rbind(result, zs[both, -(4:5), drop = FALSE])
    }
  }
  if (plot)
    segments3d(result, ...)
  else
    result
}

filledContour3d <- function(obj, ...)
  UseMethod("filledContour3d")

filledContour3d.rglId <- function(obj, plot = TRUE, replace = plot, ...) {
  mesh <- as.mesh3d(obj)
  result <- filledContour3d(mesh, plot = plot, ...)
  if (replace)
    pop3d(id = obj)
  result
}

if (getRversion() < "3.6.0") {
  hcl.colors <- function(n, ...) grDevices::cm.colors(n)
} else
  hcl.colors <- grDevices::hcl.colors

filledContour3d.mesh3d <- function(obj, fn = "z",
    nlevels = 20,
    levels = pretty(range(values), nlevels),
    color.palette = function(n) hcl.colors(n, "YlOrRd", rev = TRUE),
    col = color.palette(length(levels) - 1),
    minVertices = 0,
    plot = TRUE, 
    keepValues = FALSE, ...) {
  nverts <- ncol(obj$vb)
  oldnverts <- nverts - 1
  while (nverts < minVertices && oldnverts < nverts) {
    oldnverts <- nverts
    obj <- subdivision3d(obj, deform = FALSE, normalize = TRUE)
    nverts <- ncol(obj$vb)
  }
  
  if (is.null(fn))
    fn <- obj$values
  if (is.null(fn))
    stop("'fn' can only be NULL if 'obj' contains values.")
  
  verts <- asEuclidean(t(obj$vb))
  if (is.numeric(fn))
    values <- fn
  else {
    fn <- .getVertexFn(fn, parent.frame())
    values <- fn(verts)
  }
  if (length(levels) < 2)
    stop("Must have at least 2 levels.")
  
  reverse <- levels[1] > levels[2]
  
  if (any((levels[-length(levels)] > levels[-1]) != reverse))
    stop("Levels must be monotone increasing or decreasing")
  
  if (is.null(obj$material))
    obj$material <- list()
  obj$material$color <- NA
  obj$values <- values
  result <- list()
  ivals <- seq_along(col)
  if (reverse) ivals <- rev(ivals)
  for (i in ivals) {
    r <- range(levels[c(i, i+1)])
    obj <- clipMesh3d(obj, NULL, bound = r[1], keepValues = TRUE)
    strip <- clipMesh3d(obj, NULL, bound = r[2], greater = FALSE, keepValues = keepValues && !plot)
    strip$material$color <- col[i]
    result[[i]] <- cleanMesh3d(strip, rejoin = TRUE)
  }
  result <- do.call(merge, result)
  if (plot)
    shade3d(result, ...)
  else
    result
}

merge.mesh3d <- function(x, y, ..., attributesMustMatch = FALSE) {
  fixmesh <- function(m) {
    stopifnot(inherits(m, "mesh3d"))
    if (nrow(m$vb) == 3)
      m$vb <- rbind(m$vb, 1)
    if (!length(m$ip))
      m$ip <- matrix(numeric(), nrow=1, ncol=0)
    if (!length(m$is))
      m$is <- matrix(numeric(), nrow=2, ncol=0)
    if (!length(m$it))
      m$it <- matrix(numeric(), nrow=3, ncol=0)
    if (!length(m$ib))
      m$ib <- matrix(numeric(), nrow=4, ncol=0)
    if (!is.null(m$normals) && nrow(m$normals) == 3)
      m$normals <- rbind(m$normals, 1)
    if (is.null(m$meshColor))
      m$meshColor <- "vertices"
    if (!is.null(m$material)) {
      n <- switch(m$meshColor,
                  vertices = ncol(m$vb),
                  edges = NCOL(m$is),
                  faces = NCOL(m$it) + NCOL(m$ib),
                  legacy = 3*NCOL(m$it) + 4*NCOL(m$ib))
      if (length(m$material$color) == 1)
        m$material$color <- rep(m$material$color, n)
      if (length(m$material$alpha) == 1)
        m$material$alpha <- rep(m$material$alpha, n)
    } else
      m$material <- list()
    m
  }
  z <- fixmesh(x)
  ylist <- list(...)
  if (!missing(y))
    ylist <- c(list(y), ylist)
  for (i in seq_along(ylist)) {
    x <- z
    y <- fixmesh(ylist[[i]])
    if (!attributesMustMatch) {
      common <- intersect(names(x), names(y))
      x <- x[common]
      y <- y[common]
    }
    stopifnot(is.null(x$normals) == is.null(y$normals),
              is.null(x$texcoords) == is.null(y$texcoords),
              is.null(x$values) == is.null(y$values))
    stopifnot(x$meshColor == y$meshColor)

    nx <- ncol(x$vb)
    z <- list(vb = cbind(x$vb, y$vb),
              ip = if (length(x$ip) || length(y$ip)) cbind(x$ip, y$ip + nx),
              is = if (length(x$is) || length(y$is)) cbind(x$is, y$is + nx),
              it = if (length(x$it) || length(y$it)) cbind(x$it, y$it + nx),
              ib = if (length(x$ib) || length(y$ib)) cbind(x$ib, y$ib + nx),
              normals = cbind(x$normals, y$normals),
              texcoords = cbind(x$texcoords, y$texcoords),
              values = c(x$values, y$values),
              meshColor = x$meshColor)
    if (!attributesMustMatch) {
      common <- intersect(names(x$material), names(y$material))
      x$material <- x$material[common]
      y$material <- y$material[common]
      for (n in common) {
        if (!(n %in% c("color", "alpha")) &&
            !identical(x$material[[n]], y$material[[n]])) 
          x$material[[n]] <- NULL
      }
    } else {
        stopifnot(setequal(names(x$material), names(y$material)))
        for (n in names(x$material))
          if (!identical(x$material[[n]], y$material[[n]]))
            stop("Material ", n, " values don't match")
    }
    z$material <- x$material
    z$material$color <- c(x$material$color, y$material$color)
    z$material$alpha <- c(x$material$alpha, y$material$alpha)
    class(z) <- "mesh3d"
  }
  if (!is.null(z$material)) {
    if (length(unique(z$material$color)) == 1)
      z$material$color <- z$material$color[1]
    if (length(unique(z$material$alpha)) == 1)
      z$material$alpha <- z$material$alpha[1]
  }
  if (!length(z$material$color))
    z$material$color <- NULL
  if (!length(z$material))
    z$material <- NULL
  if (!length(z$ip))
    z$ip <- NULL
  if (!length(z$is))
    z$is <- NULL
  if (!length(z$it))
    z$it <- NULL
  if (!length(z$ib))
    z$ib <- NULL
  z
}

dirname <- tempfile()
dir.create(dirname)
olddir <- setwd(dirname)

show <- c() # list topics to show only those
skip <- c("rgl-package", "shinyGetPar3d", "tkpar3dsave", "tkrgl",
          "tkspin3d", "tkspinControl") # Ones to skip

library(tools)
db <- Rd_db("rgl")
names <- names(db)
if (length(show))
  names <- names[sub("[.]Rd$", "", names) %in% show]

Rmdnames <- sub("[.]Rd$", ".Rmd", names)
htmlnames <- sub("[.]Rd$", ".html", names)

# These functions are based on similar ones from tools
.Rd_deparse <- function (x, tag = TRUE) 
{
    if (!tag) 
        attr(x, "Rd_tag") <- "Rd"
    paste(as.character(x), collapse = "")
}
.Rd_drop_nodes_with_tags <- function (x, tags) 
{
    recurse <- function(e) {
        if (is.list(e)) 
            structure(lapply(e[is.na(match(RdTags(e), tags))], 
                recurse), Rd_tag = attr(e, "Rd_tag"))
        else e
    }
    recurse(x)
}
.Rd_drop_comments <- function (x) 
.Rd_drop_nodes_with_tags(x, "COMMENT")

RdTags <- function (Rd) 
{
    res <- sapply(Rd, attr, "Rd_tag")
    if (!length(res)) 
        res <- character()
    res
}
.Rd_get_section <- function (x, which, predefined = TRUE) 
{
    if (predefined) 
        x <- x[RdTags(x) == paste0("\\", which)]
    else {
        x <- x[RdTags(x) == "\\section"]
        if (length(x)) {
            ind <- sapply(x, function(e) .Rd_get_text(e[[1L]])) == 
                which
            x <- lapply(x[ind], `[[`, 2L)
        }
    }
    if (!length(x)) 
        x
    else structure(x[[1L]], class = "Rd")
}
.Rd_get_example_code <- function (x) 
{
    x <- .Rd_get_section(x, "examples")
    if (!length(x)) 
        return(character())
    x <- .Rd_drop_comments(x)
    recurse <- function(e) {
        if (is.list(e)) {
            unlist(lapply(e[is.na(match(RdTags(e), c(#"\\donttest",
                                                     "\\dontrun")))], 
                recurse))
        }
        else e
    }
    .Rd_deparse(recurse(x), tag = FALSE)
}

writeIndex <- function(names, htmlnames, cols = 4) {
  result <- character()
  if (!is.null(text)) {
    o <- order(names)
    names <- names[o]
    htmlnames <- htmlnames[o]
    entries <- paste0("[", names, "](", htmlnames, ")")
    len <- length(entries)
    padding <- ((len + cols - 1) %/% cols) * cols - len
    if (padding)
      entries <- c(entries, rep("", length.out=padding))
    result <- c(result, '\n<div class="nostripes">\n')
    result <- c(result, knitr::kable(matrix(entries, ncol=cols), format="markdown",
                                     col.names = rep(" ", cols)))
    result <- c(result, "</div>\n")
  }
}

library(rgl)
saveopts <- options(rgl.useNULL = TRUE)
prevlink <- "[Prev](index.html)"
indexlink <- "[Index](index.html)"

for (i in seq_along(names)) {
  Rmd <- file(Rmdnames[i], open = "wt")
  nextlink <- if (i < length(htmlnames)) paste0("[Next](", htmlnames[i+1], ")") else ""
  writeLines(c('---', paste0('title: ', names[i]), 
'output: html_document
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
initialWd <- getwd()
saveopts <- options()
options(rgl.useNULL = TRUE)
library(rgl)
setupKnitr(autoprint = TRUE)
example <- function(...) {
  saveopts <- options(rgl.printRglwidget = FALSE)
  on.exit(options(saveopts))
  utils::example(...)
  lowlevel(numeric())
}
options(ask = FALSE, examples.ask = FALSE, device.ask.default = FALSE)
```
'), Rmd)
  writeLines(paste(prevlink, nextlink, indexlink), Rmd)
  if (file_path_sans_ext(Rmdnames[i]) %in% skip)
    writeLines(
'```{r eval = FALSE}
# This example is skipped in the demo.', Rmd)
  else
    writeLines('```{r}', Rmd)
  
  code <- .Rd_get_example_code(db[[names[i]]])
  if (length(code))
    writeLines(code, Rmd)
  else
    writeLines("# No example code", Rmd)
  writeLines(
'```
```{r echo=FALSE,include=FALSE}
setwd(initialWd)
while(length(rgl.dev.list())) close3d()
rm(examples)
options(saveopts)
```
', Rmd)
  writeLines(paste(prevlink, nextlink, indexlink), Rmd)
  close(Rmd)
  prevlink <- paste0("[Prev](", htmlnames[i], ")")
  rmarkdown::render(Rmdnames[i])
  while(length(rgl.dev.list())) close3d()
  environment(rglwidget)$reuseDF <- NULL
}

indexname <- "index.Rmd"
index <- file(indexname, open = "wt")
writeLines(c(
'---
title: "rgl Examples"
author: "Duncan Murdoch"
output: html_document
---
	
These files show examples from almost every help file
in `rgl`.  
',
writeIndex(names, htmlnames)), index)
close(index)

browseURL(rmarkdown::render(indexname))
options(saveopts)
setwd(olddir)

# 
# subdivision.mesh3d
#
# an *efficient* algorithm for subdivision of mesh3d objects
# using addition operations and homogenous coordinates 
# by Daniel Adler 
#
# 

edgemap <- function( size ) {
  data<-vector( mode="numeric", length=( size*(size+1) )/2 )
  data[] <- -1
  return(data)
}

edgeindex <- function( from, to, size, row=min(from,to), col=max(from,to) )
  return( row*size - ( row*(row+1) )/2 - (size-col) )

getTags <- function(mesh) {
	len <- length(mesh$ip) + length(mesh$is)/2 +
		     length(mesh$it)/3 + length(mesh$ib)/4
	tags <- mesh$tags
	if (is.null(tags))
	  tags <- seq_len(len)
	else if (length(tags) != len)
		stop("mesh$tags length must equal total number of points, segments, triangles and quads")
	tags
}

divide.mesh3d <- function(mesh, vb = mesh$vb, 
													ib = mesh$ib, it = mesh$it, is = mesh$is,
													keepTags = FALSE) {
  nv    <- dim(vb)[2]
  inds <- seq_len(nv)
  np <- length(mesh$ip)
  ns <- length(is)/2
  nt <- length(it)/3
  nq <- length(ib)/4
  
  newnp <- np
  newns <- 2*ns
  newnt <- 4*nt
  newnq <- 4*nq
  
  meshColor <- mesh$meshColor
  if (is.null(meshColor)) meshColor <- "vertices"
  
  hasColors <- !is.null(mesh$material) &&
  		         length(mesh$material$color) > 1 &&
  	           !(meshColor %in% c("edges", "legacy"))
  
  if (hasColors && meshColor == "vertices") {
  	color <- rep_len(mesh$material$color, nv)
  	rgb <- col2rgb(color)
  	interpRGB <- TRUE
  } else
  	interpRGB <- FALSE
  
  hasAlpha <- !is.null(mesh$material) &&
  	length(mesh$material$alpha) > 1 &&
  	!(meshColor %in% c("edges", "legacy"))
  
  if (hasAlpha && meshColor == "vertices") {
  	alpha <- rep_len(mesh$material$alpha, nv)
  	interpAlpha <- TRUE
  } else
    interpAlpha <- FALSE
  
  oldtags <- getTags(mesh)
  newtags <- integer(newnp + newns + newnt + newnq)
 
  nvmax <- nv + nq + ( nv*(nv+1) )/2
  outvb <- matrix(data=0,nrow=4,ncol=nvmax)  
  # copy old points
  outvb[,seq_len(nv)] <- vb  
  vcnt <- nv
  em    <- edgemap( nv )
  if (interpRGB)
  	outrgb <- cbind(rgb, matrix(0, nrow = 3, ncol = nvmax - nv))
  if (interpAlpha)
  	outalpha <- c(alpha, rep(0, nvmax - nv))
  
  if (!is.null(mesh$normals)) {
    if (NROW(mesh$normals) == 4)
      mesh$normals <- t(asEuclidean(t(mesh$normals)))
    newnormals <- matrix(data=0,nrow=3,ncol=nvmax)
    newnormals[,inds] <- mesh$normals
  } else newnormals <- NULL
  
  if (!is.null(mesh$texcoords)) {
    newtexcoords <- matrix(data=0,nrow=2,ncol=nvmax)
    newtexcoords[,inds] <- mesh$texcoords
  } else newtexcoords <- NULL
  
  if (!is.null(newnormals) || !is.null(newtexcoords)) {
    newcount <- rep(0, nvmax)
    newcount[inds] <- 1
  }
  
  result <- structure(list(material=mesh$material), class="mesh3d")

  if (nq) {
  	
  	outib <- matrix(nrow=4,ncol=newnq)
  	vcnt  <- vcnt + nq
  	
  	for (i in 1:nq ) {
  		isurf <- nv + i
  		if (interpRGB)
  			outrgb[, isurf] <- rowSums(rgb[, ib[, i]])/4
  		if (interpAlpha)
  			outalpha[isurf] <- sum(alpha(ib[,i]))/4
  		for (j in 1:4 ) {
  			
  			iprev <- ib[((j+2)%%4) + 1, i]
  			ithis <- ib[j,i]
  			inext <- ib[ (j%%4)    + 1, i]
  			
  			# get or alloc edge-point this->next
  			mindex <- edgeindex(ithis, inext, nv)
  			enext <- em[mindex]
  			if (enext == -1) {
  				vcnt       <- vcnt + 1
  				enext      <- vcnt
  				em[mindex] <- enext
  				if (interpRGB)
  					outrgb[, enext] <- (rgb[, inext] + rgb[, ithis])/2
  				if (interpAlpha)
  					outalpha[enext] <- (alpha[inext] + alpha[ithis])/2
  			}
  			
  			# get or alloc edge-point prev->this
  			mindex <- edgeindex(iprev, ithis, nv)
  			eprev <- em[mindex]
  			if (eprev == -1) {
  				vcnt       <- vcnt + 1
  				eprev      <- vcnt
  				em[mindex] <- eprev
  				if (interpRGB)
  					outrgb[, eprev] <- (rgb[, iprev] + rgb[, ithis])/2
  				if (interpAlpha)
  					outalpha[eprev] <- (alpha[iprev] + alpha[ithis])/2
  			}
  			
  			# gen grid      
  			outib[, (i-1)*4+j ] <- c( ithis, enext, isurf, eprev )
  			
  			newtags[newnp + newns + newnt + (i-1)*4+j] <- np + ns + nt + i
  			
  			# calculate surface point
  			outvb[,isurf] <- outvb[,isurf] + vb[,ithis]
  			
  			# calculate edge point
  			outvb[,enext] <- outvb[,enext] + vb[,ithis] 
  			outvb[,eprev] <- outvb[,eprev] + vb[,ithis] 
  			
  			if (!is.null(newnormals)) {
  				thisnorm <- mesh$normals[,ithis]
  				newnormals[,isurf] <- newnormals[,isurf] + thisnorm
  				newnormals[,enext] <- newnormals[,enext] + thisnorm
  				newnormals[,eprev] <- newnormals[,eprev] + thisnorm
  			}
  			
  			if (!is.null(newtexcoords)) {
  				thistexcoord <- mesh$texcoords[,ithis]
  				newtexcoords[,isurf] <- newtexcoords[,isurf] + thistexcoord
  				newtexcoords[,enext] <- newtexcoords[,enext] + thistexcoord
  				newtexcoords[,eprev] <- newtexcoords[,eprev] + thistexcoord
  			}
  			
  			if (!is.null(newnormals) || !is.null(newtexcoords)) {
  				newcount[isurf] <- newcount[isurf] + 1
  				newcount[enext] <- newcount[enext] + 1
  				newcount[eprev] <- newcount[eprev] + 1
  			}
  		}
  	}
  	result$ib <- outib
  }
  
  if (nt) {
  	
  	outit <- matrix(nrow=3,ncol=newnt)
  	
  	for (i in 1:nt ) {
  		for (j in 1:3 ) {
  			
  			iprev <- it[((j+1)%%3) + 1, i]
  			ithis <- it[j,i]
  			inext <- it[ (j%%3)    + 1, i]
  			
  			# get or alloc edge-point this->next
  			mindex <- edgeindex(ithis, inext, nv)
  			enext <- em[mindex]
  			if (enext == -1) {
  				vcnt       <- vcnt + 1
  				enext      <- vcnt
  				em[mindex] <- enext
  				if (interpRGB)
  					outrgb[, enext] <- (rgb[, inext] + rgb[, ithis])/2
  				if (interpAlpha)
  					outalpha[enext] <- (alpha[inext] + alpha[ithis])/2
  			}
  			
  			# get or alloc edge-point prev->this
  			mindex <- edgeindex(iprev, ithis, nv)
  			eprev <- em[mindex]
  			if (eprev == -1) {
  				vcnt       <- vcnt + 1
  				eprev      <- vcnt
  				em[mindex] <- eprev
  				if (interpRGB)
  					outrgb[, eprev] <- (rgb[, iprev] + rgb[, ithis])/2
  				if (interpAlpha)
  					outalpha[eprev] <- (alpha[iprev] + alpha[ithis])/2
  			}
  			
  			# gen grid      
  			outit[, (i-1)*4+j ] <- c( ithis, enext, eprev )
  			
  			newtags[newnp + newns + (i-1)*4+j ] <- np + ns + i
  			
  			# calculate edge point
  			outvb[,enext] <- outvb[,enext] + vb[,ithis] 
  			outvb[,eprev] <- outvb[,eprev] + vb[,ithis] 
  			
  			if (!is.null(newnormals)) {
  				thisnorm <- mesh$normals[,ithis]
  				newnormals[,enext] <- newnormals[,enext] + thisnorm
  				newnormals[,eprev] <- newnormals[,eprev] + thisnorm
  			}
  			
  			if (!is.null(newtexcoords)) {
  				thistexcoord <- mesh$texcoords[,ithis]
  				newtexcoords[,enext] <- newtexcoords[,enext] + thistexcoord
  				newtexcoords[,eprev] <- newtexcoords[,eprev] + thistexcoord
  			}
  			
  			if (!is.null(newnormals) || !is.null(newtexcoords)) {
  				newcount[enext] <- newcount[enext] + 1
  				newcount[eprev] <- newcount[eprev] + 1
  			}        
  			
  		}
  		# Now central triangle
  		
  		outit[, (i-1)*4+4 ] <- c( em[edgeindex(ithis, inext, nv)],
  															em[edgeindex(inext, iprev, nv)],
  															em[edgeindex(iprev, ithis, nv)] )
  		newtags[newnp + newns + (i-1)*4+4] <- np + ns + i
  	}
  	result$it <- outit
  }

  if (ns) {
  	
  	outis <- matrix(nrow=2,ncol=newns)
  	
  	for (i in 1:ns ) {
  		for (j in 1:2 ) {
  			
  			ithis <- is[j,i]
  			inext <- is[ (j%%2)    + 1, i]
  			
  			# get or alloc edge-point this->next
  			mindex <- edgeindex(ithis, inext, nv)
  			enext <- em[mindex]
  			if (enext == -1) {
  				vcnt       <- vcnt + 1
  				enext      <- vcnt
  				em[mindex] <- enext
  				if (interpRGB)
  					outrgb[, enext] <- (rgb[, inext] + rgb[, ithis])/2
  				if (interpAlpha)
  					outalpha[enext] <- (alpha[inext] + alpha[ithis])/2
  			}
  			
  			# gen grid      
  			outis[, (i-1)*2+j ] <- c( ithis, enext)
  			
  			newtags[newnp + (i-1)*2+j ] <- np + i
  			
  			# calculate edge point
  			outvb[,enext] <- outvb[,enext] + vb[,ithis] 
  			
  			if (!is.null(newnormals)) {
  				thisnorm <- mesh$normals[,ithis]
  				newnormals[,enext] <- newnormals[,enext] + thisnorm
  			}
  			
  			if (!is.null(newtexcoords)) {
  				thistexcoord <- mesh$texcoords[,ithis]
  				newtexcoords[,enext] <- newtexcoords[,enext] + thistexcoord
  			}
  			
  			if (!is.null(newnormals) || !is.null(newtexcoords)) {
  				newcount[enext] <- newcount[enext] + 1
  			}        
  			
  		}
  	}
  	result$is <- outis
  }
  
  if (np) {
  	result$ip <- mesh$ip
  	newtags[seq_len(np)] <- seq_len(np)
  }

  if (!is.null(newnormals) || !is.null(newtexcoords)) 
    newcount <- newcount[seq_len(vcnt)]
    
  if (!is.null(newnormals)) 
    newnormals <- newnormals[, seq_len(vcnt)]/rep(newcount, each=3)
    
  if (!is.null(newtexcoords))
    newtexcoords <- newtexcoords[, seq_len(vcnt)]/rep(newcount, each=2)
    
  result$vb <- outvb[,seq_len(vcnt)]
  result$normals <- newnormals
  result$texcoords <- newtexcoords
  result$meshColor <- mesh$meshColor
  result$material <- mesh$material
  if (interpRGB) 
  	result$material$color <- rgb(t(outrgb[, seq_len(vcnt)]), maxColorValue = 255)
  else if (hasColors && meshColor == "faces") {
  	colors <- rep_len(mesh$material$color, np + ns + nt + nq)
  	result$material$color <- colors[newtags]
  }
  if (interpAlpha)
  	result$material$alpha <- alpha[seq_len(vcnt)]
  else if (hasAlpha && meshColor == "faces") {
  	alpha <- rep_len(mesh$material$alpha, np + ns + nt + nq)
  	result$material$alpha <- alpha[newtags]
  }
  
  if (keepTags)
    result$tags <- oldtags[newtags]

  return( result )
}

normalize.mesh3d <- function(mesh) {
  mesh$vb[1,] <- mesh$vb[1,]/mesh$vb[4,]
  mesh$vb[2,] <- mesh$vb[2,]/mesh$vb[4,]
  mesh$vb[3,] <- mesh$vb[3,]/mesh$vb[4,]
  mesh$vb[4,] <- 1
  return(mesh)
}

deform.mesh3d <- function( mesh, vb = mesh$vb, 
													 ib = mesh$ib, it = mesh$it, is = mesh$is) {
  nv <- dim(vb)[2]
  nq <- if (is.null(ib)) 0 else dim(ib)[2]
  nt <- if (is.null(it)) 0 else dim(it)[2]
  ns <- if (is.null(is)) 0 else dim(is)[2]
  out <- matrix(0, nrow=4, ncol=nv )
  if (nq) {
    for ( i in 1:nq ) {
      for (j in 1:4 ) {
        iprev <- ib[((j+2)%%4) + 1, i]
        ithis <- ib[j,i]
        inext <- ib[ (j%%4)    + 1, i]
        out[ ,ithis ] <- out[ , ithis ] + vb[,iprev] + vb[,ithis] + vb[,inext]
      }
    }
    mesh$vb <- out
  }
  if (nt) {
    for ( i in 1:nt ) {
      for (j in 1:3 ) {
        iprev <- it[((j+1)%%3) + 1, i]
        ithis <- it[j,i]
        inext <- it[ (j%%3)    + 1, i]
        out[ ,ithis ] <- out[ , ithis ] + vb[,iprev] + vb[,ithis] + vb[,inext]
      }
    }
    mesh$vb <- out
  }  
  if (ns) {
  	for ( i in 1:ns ) {
  		for (j in 1:2 ) {
  			ithis <- is[j,i]
  			inext <- is[ (j%%2)    + 1, i]
  			out[ ,ithis ] <- out[ , ithis ] + vb[,ithis] + vb[,inext]
  		}
  	}
  	mesh$vb <- out
  } 
  return(mesh)
}

subdivision3d.mesh3d <- function(x, depth = 1, normalize = FALSE,
																 deform = TRUE, keepTags = FALSE, ...) {
  mesh <- x
  if (depth) {
    mesh <- divide.mesh3d(mesh, keepTags = TRUE)
    if (normalize)
      mesh <- normalize.mesh3d(mesh)
    if (deform)
      mesh <- deform.mesh3d(mesh)      
    mesh<-subdivision3d.mesh3d(mesh, depth-1, normalize, deform,
    													 keepTags = TRUE)
  }
  if (!keepTags)
  	mesh$tags <- NULL
  return(mesh)  
}

cone3d <- function(base,tip,rad,n=30,...) {
  degvec <- seq(0,2*pi,length=n)
  ax <- tip-base
  ## what do if ax[1]==0?
  if (ax[1]!=0) {
    p1 <- c(-ax[2]/ax[1],1,0)
    p1 <- p1/sqrt(sum(p1^2))
    if (p1[1]!=0) {
      p2 <- c(-p1[2]/p1[1],1,0)
      p2[3] <- -sum(p2*ax)
      p2 <- p2/sqrt(sum(p2^2))
    } else {
      p2 <- c(0,0,1)
    }
  } else if (ax[2]!=0) {
    p1 <- c(0,-ax[3]/ax[2],1)
    p1 <- p1/sqrt(sum(p1^2))
    if (p1[1]!=0) {
      p2 <- c(0,-p1[3]/p1[2],1)
      p2[3] <- -sum(p2*ax)
      p2 <- p2/sqrt(sum(p2^2))
    } else {
      p2 <- c(1,0,0)
    }
  } else {
    p1 <- c(0,1,0); p2 <- c(1,0,0)
  }
  ecoord2 <- function(theta) {
    base+rad*(cos(theta)*p1+sin(theta)*p2)
  }
    for (i in seq_len(n-1)) {
      li <- ecoord2(degvec[i])
      lj <- ecoord2(degvec[i+1])
      triangles3d(c(li[1],lj[1],tip[1]),c(li[2],lj[2],tip[2]),c(li[3],lj[3],tip[3]),...)
    }   
}

lollipop3d <- function(data.x,data.y,data.z,surf.fun,surf.n=50,
                         xlim=range(data.x),
			 ylim=range(data.y),
			 zlim=range(data.z),
                         asp=c(y=1,z=1),
			 xlab=deparse(substitute(x)),
			 ylab=deparse(substitute(y)),
			 zlab=deparse(substitute(z)),
			 alpha.surf=0.4,
                         col.surf=fg,col.stem=c(fg,fg),
                         col.pt="gray",type.surf="line",ptsize,
                         lwd.stem=2,lit=TRUE,bg="white",fg="black",
                         col.axes=fg,col.axlabs=fg,
                         axis.arrow=TRUE,axis.labels=TRUE,
                         box.col=bg,
                         axes=c("lines","box")) {
  axes <- match.arg(axes)
  col.stem <- rep(col.stem,length=2)
  x.ticks <- pretty(xlim)
  x.ticks <- x.ticks[x.ticks>=min(xlim) & x.ticks<=max(xlim)]
  x.ticklabs <- if (axis.labels) as.character(x.ticks) else NULL
  y.ticks <- pretty(ylim)
  y.ticks <- y.ticks[y.ticks>=min(ylim) & y.ticks<=max(ylim)]
  y.ticklabs <- if (axis.labels) as.character(y.ticks) else NULL
  z.ticks <- pretty(zlim)
  z.ticks <- z.ticks[z.ticks>=min(zlim) & z.ticks<=max(zlim)]
  z.ticklabs <- if (axis.labels) as.character(z.ticks) else NULL
  if (!missing(surf.fun)) {
    surf.x <- seq(xlim[1],xlim[2],length=surf.n)
    surf.y <- seq(ylim[1],ylim[2],length=surf.n)
    surf.z <- outer(surf.x,surf.y,surf.fun)  ## requires surf.fun be vectorized
    z.interc <- surf.fun(data.x,data.y)
    zdiff <- diff(range(c(surf.z,data.z)))
  } else {
    z.interc <- rep(min(data.z),length(data.x))
    zdiff <- diff(range(data.z))
  }
  xdiff <- diff(xlim)
  ydiff <- diff(ylim)
  y.adj <- if (asp[1]<=0) 1 else asp[1]*xdiff/ydiff
  data.y <- y.adj*data.y
  y.ticks <- y.adj*y.ticks
  ylim <- ylim*y.adj
  ydiff <- diff(ylim)
  z.adj <- if (asp[2]<=0) 1 else asp[2]*xdiff/zdiff
  data.z <- z.adj*data.z
  if (!missing(surf.fun)) {
    surf.y <- y.adj*surf.y
    surf.z <- z.adj*surf.z
  }
  z.interc <- z.adj*z.interc
  z.ticks <- z.adj*z.ticks
  zlim <- z.adj*zlim
  open3d()
  clear3d("all")
  light3d()
  bg3d(color=c(bg,fg))
  if (!missing(surf.fun)) 
    surface3d(surf.x,surf.y,surf.z,alpha=alpha.surf,
                front=type.surf,back=type.surf,
                col=col.surf,lit=lit)
  if (missing(ptsize)) ptsize <- 0.02*xdiff
  ## draw points
  spheres3d(data.x,data.y,data.z,r=ptsize,lit=lit,color=col.pt)
  ## draw lollipops
  apply(cbind(data.x,data.y,data.z,z.interc),1,
        function(X) {
          lines3d(x=rep(X[1],2),
                  y=rep(X[2],2),
                  z=c(X[3],X[4]),
                  col=ifelse(X[3]>X[4],col.stem[1],
                    col.stem[2]),lwd=lwd.stem)
        })
  if (axes=="box") {
    bbox3d(xat=x.ticks,xlab=x.ticklabs,
             yat=y.ticks,ylab=y.ticklabs,
             zat=z.ticks,zlab=z.ticklabs,lit=lit)
  } else if (axes=="lines") { ## set up axis lines
    axis3d(edge="x",at=x.ticks,labels=x.ticklabs,
           col=col.axes,arrow=axis.arrow)
    axis3d(edge="y",at=y.ticks,labels=y.ticklabs,
           col=col.axes,arrow=axis.arrow)
    axis3d(edge="z",at=z.ticks,labels=z.ticklabs,
           col=col.axes,arrow=axis.arrow)
    box3d(col=col.axes)
  }
  decorate3d(xlab=xlab, ylab=ylab, zlab=zlab, box=FALSE, axes=FALSE, col=col.axlabs)
}

x <- 1:5
y <- x*10
z <- (x+y)/20
open3d()
spheres3d(x,y,z)
axes3d()
set.seed(1001)
x <- runif(30)
y <- runif(30,max=2)
dfun <- function(x,y)  2*x+3*y+2*x*y+3*y^2 
z <- dfun(x,y)+rnorm(30,sd=0.5)
## lollipops only
lollipop3d(x,y,z)

## lollipops plus theoretical surface

lollipop3d(x,y,z,dfun,col.pt="red",col.stem=c("red","blue"))
## lollipops plus regression fit

linmodel <- lm(z~x+y)
dfun <- function(x,y) predict(linmodel,newdata=data.frame(x=x,y=y))
lollipop3d(x,y,z,dfun,col.pt="red",col.stem=c("red","blue"))

####

library(rgl)

# WORKING

x <- cube3d()
b <- getBoundary3d(x)
open3d(); shade3d(b)


x <- cube3d()
x$ib <- x$ib[,-(1:2)]
b <- getBoundary3d(x)
open3d(); shade3d(x, alpha=0.2, col = "blue"); shade3d(b) 

x <- cube3d()
x$ib <- x$ib[,-(1:2)]
b <- getBoundary3d(x, sorted = TRUE, simplify = FALSE)
open3d(); shade3d(x, alpha=0.2, col = "blue"); shade3d(b) 

# Functions related to working in pkgdown

pkgdown_rdname <- function() 
	getOption("downlit.rdname", "")

in_pkgdown <- function() 
	requireNamespace("pkgdown", quietly = TRUE) && pkgdown::in_pkgdown()

in_pkgdown_example <- function() 
	nchar(pkgdown_rdname()) && 
	requireNamespace("downlit", quietly = TRUE) &&
	requireNamespace("pkgdown", quietly = TRUE)

fns <- local({
	plotnum <- 0
	
	pkgdown_print.rglId <- function(x, visible = TRUE) {
		
		if (inherits(x, "rglHighlevel"))
			plotnum <<- plotnum + 1
		
		if (visible) {
		  scene <- scene3d()
		  structure(list(plotnum = plotnum,
			  				     scene = scene),
				  			class = c("rglRecordedplot", "otherRecordedplot"))
		} else
			invisible()
	}
	
	pkgdown_print.rglOpen3d <- function(x, visible = TRUE) {
		plotnum <<- plotnum + 1
		invisible(x)
	}
	
	list(pkgdown_print.rglId = pkgdown_print.rglId,
			 pkgdown_print.rglOpen3d = pkgdown_print.rglOpen3d)
})

pkgdown_print.rglId <-     fns[["pkgdown_print.rglId"]]
pkgdown_print.rglOpen3d <- fns[["pkgdown_print.rglOpen3d"]]
rm(fns)

globalVariables("fig.asp")
pkgdown_dims <- function() {
	settings <- pkgdown::fig_settings()
	rgl <- settings$other.parameters$rgl
	
	settings[names(rgl)] <- rgl

	numparms <- length(intersect(names(rgl), c("fig.width", "fig.height", "fig.asp")))
	if (numparms > 0 && numparms < 3) {
		settings <- within(settings, {
	    if (is.null(rgl$fig.height))
		    fig.height <- fig.width * fig.asp
	    if (is.null(rgl$fig.width))
	    	fig.width <- fig.height / fig.asp
		})
	}
	
	width <- with(settings, dpi*fig.width)
	height <- with(settings, dpi*fig.height)
	list(width = width, height = height)
}

replay_html.rglRecordedplot <- local({
	rdname <- ""
	function(x, ...) {
		if (pkgdown_rdname() != rdname) 
			rdname <<- pkgdown_rdname()
		
		settings <- pkgdown_dims()
		rendered <- htmltools::renderTags(rglwidget(x$scene,
																								width = settings$width, height = settings$height))
		structure(rendered$html, dependencies = rendered$dependencies)
	}
})

pkgdown_info <- local({
	info <- NULL
	function() {
		if (!is.null(info))
			return(info)
    path <- "."
    repeat {
	    if (file.exists(file.path(path, "DESCRIPTION"))) {
	    	info <<- pkgdown::as_pkgdown(path)
	    	return(info)
	    }
	    newpath <- file.path(path, "..")
	    if (normalizePath(newpath) == normalizePath(path))
	    	return(list())
	    path <- newpath
    }
  }
})

register_pkgdown_methods <- local({
	registered <- FALSE
	function(register = in_pkgdown_example()) {
		if (!registered && register) {
			registerS3method("replay_html", "rglRecordedplot", 
											 replay_html.rglRecordedplot, 
											 envir = asNamespace("downlit"))
			registerS3method("is_low_change", "rglRecordedplot", 
											 is_low_change.rglRecordedplot,
											 envir = asNamespace("downlit"))
			registerS3method("pkgdown_print", "rglId", 
											 pkgdown_print.rglId, 
											 envir = asNamespace("pkgdown"))
			registerS3method("pkgdown_print", "rglOpen3d", 
											 pkgdown_print.rglOpen3d, 
											 envir = asNamespace("pkgdown"))
			registered <<- TRUE
		}
	}
})

latex3d <- function(x, y = NULL, z = NULL,
		       text, 
		       cex = par3d("cex"), adj = 0.5,
		       pos = NULL, offset = 0.5,
		       fixedSize = TRUE,
		       startsize = 480, initCex = 5, 
		       margin = "", floating = FALSE, tag = "",
		       polygon_offset = material3d("polygon_offset"),
		       verbose = FALSE,
		       ...) {
  if (!requireNamespace("xdvir"))
    stop("This function requires the `xdvir` package.")
  xyz <- xyz.coords(x, y, z, recycle = TRUE)
  n <- length(xyz$x)
  if (is.vector(text))
    text <- rep(text, length.out = n)
  cex <- rep(cex, length.out = n)
  if (!is.null(pos))
    pos <- rep_len(pos, n)
  adj <- c(adj, 0.5, 0.5, 0.5)[1:3]
  save3d <- par3d(skipRedraw = TRUE)
  save <- options(device.ask.default = FALSE, tinytex.verbose = verbose)
  
  # Github windows runners have trouble with the tempdir() path,
  # so we'll set it as the working directory.
  
  olddir <- setwd(tempdir())
  on.exit({options(save); par3d(save3d); setwd(olddir)}) # nolint
  result <- integer(n)
  if (verbose) {
    cat("TeX status:\n")
    xdvir::TeXstatus()
  }
  for (i in seq_len(n)) {
    # Open the device twice.  The first one is to measure the text...
    f <- tempfile(fileext = ".png")
    png(f, bg = "transparent", width = initCex*startsize, height = initCex*startsize, res = initCex*72)
    if (is.vector(text))
      thistext <- text[i]
    else
      thistext <- text
    texfile <- basename(tempfile(fileext = ".tex"))
    if (verbose) {
      doc <- xdvir::author(thistext, ...)
      cat("\nGenerated LaTeX:\n")
      print(doc)
      cat("\nWriting tex to ", texfile, " in ", getwd(), "\n")
      dvi <- xdvir::typeset(doc, texFile = texfile, 
                            ...)
      cat("\nGenerated DVI:\n")
      print(dvi)
      g <- xdvir::dviGrob(dvi, ...)
    } else 
      g <- xdvir::latexGrob(thistext, texFile = texfile,
                            ...)
    w_npc <- grid::convertWidth(grid::grobWidth(g), "npc", valueOnly = TRUE)
    h_npc <- grid::convertHeight(grid::grobHeight(g), "npc", valueOnly = TRUE)
    safe.dev.off()

    # Now make a smaller bitmap
    maxdim <- max(w_npc, h_npc)
    size <- round(initCex*startsize*maxdim)
    png(f, bg = "transparent", 
        width = size, height = size+1, 
        pointsize = 12, res = initCex*72)
    grid::grid.draw(g)
    
    safe.dev.off()
    # The 0.4 tries to match the text3d offset
    offseti <- 0.4*offset*h_npc/w_npc
    posi <- if (is.null(pos)) NULL else pos[i]
    result[i] <- with(xyz, sprites3d(x[i], y[i], z[i], texture = f, textype = "rgba", 
            col = "white", lit = FALSE, radius = cex[i]*size/initCex/20,
            adj = adj, pos = posi, offset = offseti,
            fixedSize = fixedSize,
            margin = margin, floating = floating, tag = tag,
            polygon_offset = polygon_offset))
  }
  lowlevel(result)
}

# shiny support functions

# Shiny objects if the widget sets elementId, so we
# need to detect it.  Thanks to Joe Cheng for suggesting this code.

inShiny <- function() isNamespaceLoaded("shiny") && !is.null(shiny::getDefaultReactiveDomain())

fns <- local({
	registered <- FALSE
	registerShinyHandlers <- function() {
		if (!registered) {
			if (requireNamespace("shiny")) {
			  shiny::registerInputHandler("shinyPar3d", convertShinyPar3d, force = TRUE)
			  shiny::registerInputHandler("shinyMouse3d", convertShinyMouse3d, force = TRUE)
			  registered <<- TRUE
			} else
				stop("Not in Shiny")
		}
	}
	unregisterShinyHandlers <- function() {
		if (registered) {
			shiny::removeInputHandler("shinyPar3d")
			shiny::removeInputHandler("shinyMouse3d")
		}
	}
	list(registerShinyHandlers = registerShinyHandlers,
			 unregisterShinyHandlers = unregisterShinyHandlers)
})
registerShinyHandlers <- fns$registerShinyHandlers
unregisterShinyHandlers <- fns$unregisterShinyHandlers
rm(fns)

# Widget output function for use in Shiny

rglwidgetOutput <- function(outputId, width = '512px', height = '512px') {
	registerShinyHandlers()
	shinyWidgetOutput(outputId, 'rglWebGL', width, height, package = 'rgl')
}

# Widget render function for use in Shiny

renderRglwidget <- function(expr, env = parent.frame(), quoted = FALSE, outputArgs = list()) {
	registerShinyHandlers()
	if (!quoted) expr <- substitute(expr)  # force quoted
	shiny::markRenderFunction(rglwidgetOutput,
														shinyRenderWidget(expr, rglwidgetOutput, env, quoted = TRUE),
														outputArgs = outputArgs)
}

shinySetPar3d <- function(..., session,
													subscene = currentSubscene3d(cur3d())) {
	registerShinyHandlers()
	args <- list(...)
	argnames <- names(args)
	if (length(args) == 1 && is.null(argnames)) {
		args <- args[[1]]
	}
	# We might have been passed modified shinyGetPar3d output;
	# clean it up.
	args[["subscene"]] <- NULL
	args[["tag"]] <- NULL
	argnames <- names(args)
	
	if (is.null(argnames) || any(argnames == ""))
		stop("Parameters must all be named")
	
	badargs <- argnames[!(argnames %in% rgl.par3d.names)]

	if (length(badargs))
		stop("Invalid parameter(s): ", badargs)
	
	for (arg in argnames) {
		session$sendCustomMessage("shinySetPar3d", 
															list(subscene = subscene,
																	 parameter = arg,
																	 value = args[[arg]]))
	}
}

shinyGetPar3d <- function(parameters, session,
													subscene = currentSubscene3d(cur3d()),
													tag = "") {
	registerShinyHandlers()
	badargs <- parameters[!(parameters %in% rgl.par3d.names)]
	if (length(badargs))
		stop("invalid parameter(s): ", badargs)
	session$sendCustomMessage("shinyGetPar3d",
														list(tag = tag, subscene = subscene, 
																 parameters = parameters))
}

convertShinyPar3d <- function(par3d, ...) {
	for (parm in c("modelMatrix", "projMatrix", "userMatrix", "userProjection"))
		if (!is.null(par3d[[parm]]))
			par3d[[parm]] <- matrix(unlist(par3d[[parm]]), 4, 4)
	for (parm in c("mouseMode", "observer", "scale", "viewport", "bbox", "windowRect"))
		if (!is.null(par3d[[parm]]))
			par3d[[parm]] <- unlist(par3d[[parm]])
	par3d
}

shinyResetBrush <- function(session, brush) {
	registerShinyHandlers()
	session$sendCustomMessage("resetBrush", brush)
}

convertShinyMouse3d <- function(mouse3d, ...) {
	for (parm in c("model", "proj")) 
		if (!is.null(mouse3d[[parm]]))
			mouse3d[[parm]] <- matrix(unlist(mouse3d[[parm]]), 4, 4)
	
	if (length(unlist(mouse3d$view)) == 4) 
		mouse3d$view <- structure(unlist(mouse3d$view),
															names = c("x", "y", "width", "height"))
	if (all(c("p1", "p2") %in% names(mouse3d$region)))
		mouse3d$region <- with(mouse3d$region,
													 c(x1 = (p1$x + 1)/2,
													 	y1 = (p1$y + 1)/2,
													 	x2 = (p2$x + 1)/2,
													 	y2 = (p2$y + 1)/2))
	structure(mouse3d, class = "rglMouseSelection")
}

## ----setup, echo=FALSE, results="asis"----------------------------------------
source("setup.R")
setupKnitr(autoprint = FALSE)
set.seed(123)

## -----------------------------------------------------------------------------
library(rgl)
plotids <- with(iris, plot3d(Sepal.Length, Sepal.Width, Petal.Length, 
                  type="s", col=as.numeric(Species)))
rglwidget(elementId = "plot3drgl")

## -----------------------------------------------------------------------------
toggleWidget(sceneId = "plot3drgl", ids = plotids["data"], label = "Data")

## -----------------------------------------------------------------------------
names(plotids)
unclass(plotids)

## -----------------------------------------------------------------------------
rglwidget() %>%
toggleWidget(ids = plotids["data"], label = "Data")

## ----eval=FALSE---------------------------------------------------------------
#  rglwidget() |>
#  toggleWidget(ids = plotids["data"], label = "Data")

## -----------------------------------------------------------------------------
toggleWidget(NA, ids = plotids["data"], label = "Data") %>%
rglwidget(controllers = .) 

## ----eval=FALSE---------------------------------------------------------------
#  toggleWidget(NA, ids = plotids["data"], label = "Data") |>
#    w => rglwidget(controllers = w)

## -----------------------------------------------------------------------------
clear3d() # Remove the earlier display

with(subset(iris, Species == "setosa"), 
     spheres3d(Sepal.Length, Sepal.Width, Petal.Length, 
                  col=as.numeric(Species),
                  radius = 0.211,
                  tag = "setosa"))
with(subset(iris, Species == "versicolor"), 
     spheres3d(Sepal.Length, Sepal.Width, Petal.Length, 
               col=as.numeric(Species),
     	       radius = 0.211, 
     	       tag = "versicolor"))
with(subset(iris, Species == "virginica"), 
     spheres3d(Sepal.Length, Sepal.Width, Petal.Length, 
               col=as.numeric(Species),
     	       radius = 0.211,
     	       tag = "virginica"))
aspect3d(1,1,1)
decorate3d(tag = "axes")
rglwidget() %>%
toggleWidget(tags = "setosa") %>%
toggleWidget(tags = "versicolor") %>%
toggleWidget(tags = "virginica") %>%
toggleWidget(tags = "axes") %>% 
asRow(last = 4)

## -----------------------------------------------------------------------------
rglwidget() %>%
playwidget(start = 0, stop = 3, interval = 1,
	   subsetControl(1, subsets = list(
	   			 Setosa = tagged3d("setosa"),
	   			 Versicolor = tagged3d("versicolor"),
	   			 Virginica = tagged3d("virginica"),
	   			 All = tagged3d(c("setosa", "versicolor", "virginica"))
	   			 )))

## -----------------------------------------------------------------------------
M <- r3dDefaults$userMatrix
fn <- par3dinterp(time = (0:2)*0.75, userMatrix = list(M,
                                      rotate3d(M, pi/2, 1, 0, 0),
                                      rotate3d(M, pi/2, 0, 1, 0)) )
rglwidget() %>%
playwidget(par3dinterpControl(fn, 0, 3, steps=15),
 	   step = 0.01, loop = TRUE, rate = 0.5)

## -----------------------------------------------------------------------------
setosavals <- subset(iris, Species == "setosa")
which <- which.min(setosavals$Sepal.Width)
init <- setosavals$Sepal.Length[which]
rglwidget() %>%
playwidget(
  vertexControl(values = matrix(c(init,   0,   0,   0, 
                                     8,   1,   1,   1), 
                                nrow = 2, byrow = TRUE),
                attributes = c("x", "red", "green", "blue"),
                vertices = which, tag = "setosa"),
	step = 0.01)

## -----------------------------------------------------------------------------
time <- 0:500
xyz <- cbind(cos(time/20), sin(time/10), time)
lineid <- plot3d(xyz, type="l", col = "black")["data"]
sphereid <- spheres3d(xyz[1, , drop=FALSE], radius = 8, col = "red")
rglwidget() %>%
playwidget(list(
  ageControl(births = time, ages = c(0, 0, 50),
		colors = c("gray", "red", "gray"), objids = lineid),
  ageControl(births = 0, ages = time,
		vertices = xyz, objids = sphereid)),
  start = 0, stop = max(time) + 20, rate = 50,
  components = c("Reverse", "Play", "Slower", "Faster",
                 "Reset", "Slider", "Label"),
  loop = TRUE)

## ----eval = requireNamespace("crosstalk", quietly=TRUE)-----------------------
# This example requires the crosstalk package
# We skip it if crosstalk is not available. 

ids <- with(iris, plot3d(Sepal.Length, Sepal.Width, Petal.Length, 
                  type="s", col=as.numeric(Species)))
par3d(mouseMode = "selecting")
rglwidget(shared = rglShared(ids["data"])) %>% 
rglMouse()

## ----eval=requireNamespace("crosstalk", quietly = TRUE)-----------------------
# This example requires the crosstalk package.  
# We skip it if crosstalk is not available. 

library(crosstalk)
sd <- SharedData$new(mtcars)
ids <- plot3d(sd$origData(), col = mtcars$cyl, type = "s")
# Copy the key and group from existing shared data
rglsd <- rglShared(ids["data"], key = sd$key(), group = sd$groupName())
rglwidget(shared = rglsd) %>%
asRow("Mouse mode: ", rglMouse(getWidgetId(.)), 
      "Subset: ", filter_checkbox("cylinderselector", 
		                "Cylinders", sd, ~ cyl, inline = TRUE),
      last = 4, colsize = c(1,2,1,2), height = 60)

## ----plot3d2------------------------------------------------------------------
plotids <- with(iris, plot3d(Sepal.Length, Sepal.Width, Petal.Length, 
                  type="s", col=as.numeric(Species)))
subid <- currentSubscene3d()
rglwidget(elementId="plot3drgl2")

## ----echo=FALSE, results="asis"-----------------------------------------------
writeIndex(cols = 5)

##
## Sweave device
##
##

rgl.Sweave <- function(name, width, height, options, ...) {
  if (length(hook <- getHook("on.rgl.close"))) {
    # test is for compatibility with R < 3.0.0
    if (is.list(hook)) hook <- hook[[1]]  
    dev <- environment(hook)$dev
    set3d(dev)
  } else {
    wr <- c(0, 0, width*options$resolution, height*options$resolution)
    open3d(windowRect=wr)
    if (is.null(delay <- options$delay)) delay <- 0.1
    Sys.sleep(as.numeric(delay))
    wrnew <- par3d("windowRect")
    if (wr[3] - wr[1] != wrnew[3] - wrnew[1] ||
        wr[4] - wr[2] != wrnew[4] - wrnew[2])
      stop("rgl window creation error; try reducing resolution, width or height")
    dev <- cur3d()
  } 
  
  snapshotDone <- FALSE
  
  # stayOpen is used below in rgl.Sweave.off
  stayOpen <- isTRUE(options$stayopen)
  
  type <- options$outputtype
  if (is.null(type)) type <- "png"
  
  setHook("on.rgl.close", action="replace", function(remove=TRUE) {
    prev.dev <- cur3d()
    on.exit(set3d(prev.dev))
    
    if (!snapshotDone) {
      set3d(dev)
      switch(type,
             png = snapshot3d(filename=paste(name, "png", sep=".")),
             pdf = rgl.postscript(filename=paste(name, "pdf", sep="."), fmt="pdf"),
             eps = rgl.postscript(filename=paste(name, "eps", sep="."), fmt="eps"),
             stop(gettextf("Unrecognized rgl outputtype: '%s'", type), domain = NA)
      )
      snapshotDone <<- TRUE
    }
    
    if (remove)
      setHook("on.rgl.close", action="replace", NULL)
  })
}

rgl.Sweave.off <- function() {
  if (length(hook <- getHook("on.rgl.close"))) {
    if (is.list(hook)) hook <- hook[[1]] # test is for R pre-3.0.0 compatibility
    stayOpen <- environment(hook)$stayOpen
    if (stayOpen) hook(FALSE)
    else close3d()
  }
}

##
## Sweave snapshot
##
##

Sweave.snapshot <- function() {
  if (length(hook <- getHook("on.rgl.close"))) {
    if (is.list(hook)) hook <- hook[[1]] # test is for R pre-3.0.0 compatibility
    hook(remove = FALSE)
  }
}

# The rgl font database is only used when rgl is configured for FreeType.  
# Since 0.105.13 this is always true on Windows, but since 0.106.2
# r3dDefaults sets useFreeType to FALSE, so the windowsFonts() are used instead.

# This code is closely modelled on the Quartz font database.

.rglEnv <- new.env()

assign(".rglFonts", list(), envir = .rglEnv)

# Check that the font has the correct structure and information
checkrglFont <- function(font) {
    if (!is.character(font) || length(font) != 4)
        stop("Invalid rgl font:  must be 4 filenames")
    font
}

setrglFonts <- function(fonts, fontNames) {
    fonts <- lapply(fonts, checkrglFont)
    fontDB <- get(".rglFonts", envir=.rglEnv)
    existingFonts <- fontNames %in% names(fontDB)
    if (sum(existingFonts) > 0)
        fontDB[fontNames[existingFonts]] <- fonts[existingFonts]
    if (sum(existingFonts) < length(fontNames))
        fontDB <- c(fontDB, fonts[!existingFonts])
    assign(".rglFonts", fontDB, envir=.rglEnv)
}

printFont <- function(font) {
    paste(font, "\n", sep="")
}

printFonts <- function(fonts) {
    cat(paste(names(fonts), ": ", unlist(lapply(fonts, printFont)),
              sep="", collapse=""))
}

# If no arguments spec'ed, return entire font database
# If no named arguments spec'ed, all args should be font names
# to get info on them from the database
# Else, must specify new fonts to enter into database (all
# of which must be valid filenames and
# all of which must be named args)
rglFonts <- function(...) {
  ndots <- length(fonts <- list(...))
  if (ndots==0) {
    get(".rglFonts", .rglEnv)
  } else {
    fontNames <- names(fonts)
    nnames <- length(fontNames)
    if (nnames == 0) {
      if (!all(sapply(fonts, is.character))) {
        stop("Invalid arguments in 'rglFonts' (must be font names)")
      } else {
        get(".rglFonts", .rglEnv)[unlist(fonts)]
      }
    } else {
      if (ndots != nnames) {
        stop("Invalid arguments in 'rglFonts' (need named args)")
      }
      setrglFonts(fonts, fontNames)
    }
  }
}

rglFont <- function(family) {
  checkrglFont(family)
}

expect_known_scene <- function(name, close = TRUE, file = paste0("testdata/", name, ".rds"), ...) {
	testthat::local_edition(2)
	result <- testthat::expect_known_value(object = scene3d(), file = file, ...)
  if (close)
  	close3d()
  result
}

# Support for objects from alphashape3d package

persp3d.ashape3d <- function(x, ..., add = FALSE) {
  plot3d(as.mesh3d(x, ...), add = add, ...)
}

plot3d.ashape3d <- function(x, ...) persp3d(x, ...)

reOrient <- function(vertices) {
  warned <- FALSE

  # Count how many other triangles touch each edge of this one, in order 2-3, 1-3, 1-2:
  edgeCounts <- function(index) {
    triangle <- vertices[,index]
    result <- integer(3)
    for (i in 1:3) {
      result[i] <- sum(apply(vertices, 2, function(col) all(triangle[-i] %in% col)))
    }
    result - 1
  }
  polys <- ncol(vertices)
  verts <- nrow(vertices)
  fixed <- 0L
  for (i in seq_len(polys - 1)) {
    fixed <- max(i, fixed)
    # Get all polygons touching polygon i
    thistriangle <- vertices[,i]
    if (any(is.na(thistriangle))) next
    touches <- which(matrix(vertices %in% thistriangle,nrow=3),  arr.ind = TRUE)
    if (!nrow(touches)) next
    counts <- table(touches[,2L])
    # Get col number of all polygons sharing an edge with i
    shared <- as.numeric(names(counts)[counts > 1L])
    shared <- shared[shared != i]
    if (!length(shared)) next
    otherverts <- vertices[, shared, drop=FALSE]
    
    # FIXME:  otherverts may include multiple triangles sharing the
    #         same edge, because ashape3d sometimes embeds 
    #         tetrahedrons (or larger polyhedra?) in the surfaces
    #         it produces.  It doesn't appear to be safe to just 
    #         delete these
    
    if (!warned && length(shared) > 1L && any( edgeCounts(i) > 1)) {
      warning("Surface may not be simple; smoothing may not be possible.")
      warned <- TRUE
    }
    
    shared <- shared[shared > fixed]
    if (!length(shared)) next
    
    otherverts <- vertices[, shared, drop=FALSE]
    
    for (m in seq_len(ncol(otherverts))) { # m is intersection number
      # For each vertex in i, see if it is in the shared one, and
      # if they have opposite orientations as we need
      for (j in seq_len(verts)) {
        # Where is j in the others?
        jother <- which(otherverts[,m] == vertices[j,i])
        if (!length(jother)) next
        k <- j %% verts + 1L               # k follows j
        kother <- jother  %% verts + 1L    # kother is entry following jother
        if (vertices[k, i] == otherverts[kother, m]) {
          otherverts[, m] <- rev(otherverts[, m])
          break
        }
      }
    }
    # Now move all of shared to the front
    unshared <- (fixed+1L):max(shared)
    unshared <- unshared[!(unshared %in% shared)]
    if (length(unshared))
      vertices[, (fixed+length(shared)+1L):max(shared)] <- vertices[,unshared]
    vertices[, fixed + seq_along(shared)] <- otherverts
    fixed <- fixed + length(shared)
  }
  vertices
}

as.mesh3d.ashape3d <- function(x, alpha = x$alpha[1], tri_to_keep = 2L,
                               col = "gray", smooth = FALSE,
                               normals = NULL, texcoords = NULL,
                               ...) {
  whichAlpha <- which(alpha == x$alpha)[1]
  if (!length(whichAlpha))
    stop("'alpha = ", alpha, "' not found in ", deparse(substitute(x)))
  triangles <- x$triang
  keep <- triangles[,8 + whichAlpha] %in% tri_to_keep
  triangs <- t(triangles[keep, 1:3])
  points <- t(x$x)
  if (!is.null(texcoords))
    texcoords <- texcoords[triangs, ]
  material <- .getMaterialArgs(...)
  material$color <- col
  result <- tmesh3d(points, triangs, homogeneous = FALSE,
                    normals = normals, texcoords = texcoords,
                    material = material)
  if (smooth) {
    if (is.null(normals)) {
      result$it <- reOrient(result$it)
      result <- addNormals(result)
    } else
      warning("smoothing ignored when 'normals' specified")
  }
  result 
}

#
# R3D rendering functions - rgl implementation
# 

# Node Management

getr3dDefaults <- function(class = NULL, value = NULL) {
  result <- r3dDefaults
  if (exists("r3dDefaults", envir = globalenv())) {
    user <- get("r3dDefaults", envir=.GlobalEnv)
    for (n in names(user)) {
      if (is.list(result[[n]]))
        result[[n]][names(user[[n]])] <- user[[n]]
      else
        result[[n]] <- user[[n]]
    }
  }
  if (!is.null(class))
    result <- result[[class]]
  if (!is.null(result) && !is.null(value))
    result <- result[[value]]
  result
}

clear3d     <- function(type = c("shapes", "bboxdeco", "material"), 
                        defaults=getr3dDefaults(),
                        subscene = 0) {
  .check3d()
  
  if (is.na(subscene)) 
    subscene <- currentSubscene3d()
  
  typeid0 <- rgl.enum.nodetype(type)
  
  userviewpoint <- 4 %in% typeid0
  material  <- 5 %in% typeid0
  modelviewpoint <- 8 %in% typeid0
  
  drop <- typeid0 %in% c(4:5, 8)
  typeid <- typeid0[!drop]
  type <- names(typeid)
  
  if (subscene == 0) {
    nobg <- setdiff(typeid, 6)
    idata <- as.integer(c(length(nobg), nobg))    	
    ret <- .C( rgl_clear, 
               success = FALSE,
               idata
    )$success
    if (6 %in% typeid)
      rgl.incrementID()  # for back compatibility
    
    if (! ret)
      stop("'rgl_clear' failed")
    
  } else {
    sceneids <- ids3d(type=type, subscene = 0)$id
    thisids <- ids3d(type=type, subscene = subscene)$id
    if (length(thisids)) {
      delFromSubscene3d(ids = thisids, subscene = subscene)
      gc3d(protect = setdiff(sceneids, thisids))
    }
  }
  
  if ( userviewpoint || modelviewpoint) 
    view3d(type = c("userviewpoint", "modelviewpoint")[c(userviewpoint, modelviewpoint)])
  
  if ( material ) 
    rgl.material0()
  
  if ( 4 %in% typeid0 ) { # userviewpoint
    do.call("par3d", defaults["FOV"])
  }
  if ( 8 %in% typeid0 ) { # modelviewpoint
    do.call("par3d", defaults["userMatrix"])
  }
  if ( 5 %in% typeid0 ) { # material
    if (length(defaults$material))
      do.call("material3d", defaults$material)
  }
  if ( 6 %in% typeid0 ) { # background
    do.call("bg3d", as.list(defaults$bg))
  }
  lowlevel()
}

# Environment

rgl.material.names <- c("color", "alpha", "lit", "ambient", "specular",
    "emission", "shininess", "smooth", "front", "back", "size", 
    "lwd", "fog", "point_antialias", "line_antialias",
    "texture", "textype", "texmode", "texmipmap",
    "texminfilter", "texmagfilter", "texenvmap",
    "depth_mask", "depth_test", "isTransparent",
    "polygon_offset", "margin", "floating", "tag",
    "blend")

rgl.material.readonly <- "isTransparent"

# Warn about putting a texture on a black surface, but only
# if the surface is black because that's the default.

warnBlackTexture <- function(...,
                             defaults = material3d(),
                             color = col, col = "missing",
                             texture = defaults$texture,
                             texmode = defaults$texmode) {
  if (!is.null(texture)) {
    if (length(color) == 1 &&
      !is.na(color) &&
      color == "missing" &&
      !is.na(texmode) && 
      texmode == "modulate" &&
      isTRUE(getOption("rgl.warnBlackTexture", TRUE)) &&
      length(defaults$color) &&
      !is.na(defaults$color[1]) &&
      defaults$color[1] %in% c("#000000", "black"))
      warning("Texture will be invisible on black surface", call. = FALSE) 
  }
}

# Attach the expression for the source of a texture if
# it is not already there

addTextureSource <- function(texture, ...) {
  if (!is.null(texture) && is.null(attr(texture, "src")))
    attr(texture, "src") <- substitute(texture)
  texture
}

# This function expands a list of arguments by putting
# all entries from Params (i.e. the current settings by default)
# in place for any entries that are not listed.  
# Unrecognized args are left in place.

.fixMaterialArgs <- function(..., Params = material3d(), col) {
   f <- function(...) list(...)
   dots <- list(...)
   if (!is.null(dots$texture)) {
     warnBlackTexture(...,  
       col = if (missing(col)) "missing" else col,
       defaults = Params)
     dots$texture <- addTextureSource(...)
   }
   if (!missing(col)) 
     Params$color <- col
   formals(f) <- c(Params, formals(f))
   names <- as.list(names(Params))
   names(names) <- names
   names <- lapply(names, as.name)
   b <- as.list(body(f))
   body(f) <- as.call(c(b[1], names, b[-1]))
   do.call(f, dots)
} 

# This one just expands the argument names to match the
# standard names
.fixMaterialArgs2 <- function(..., col) {
  call <- do.call(call, list("rgl.material0", ...))
  result <- as.list(match.call(rgl.material0, call))[-1]
  if (!missing(col) && is.null(result$color))
    result$color <- col
  result
}
     
# This one just gets the material args
# If warn is TRUE, give a warning instead of ignoring extras.

.getMaterialArgs <- function(..., material = list(), warn = FALSE, col = material[["col"]]) {
  fullyNamed <- as.list(match.call(rgl.material0, 
                           as.call(c(list(as.name("rgl.material0"),
                                        ...), material))))[-1]
  if (!is.null(col) && !("color" %in% names(fullyNamed)))
    fullyNamed$color <- col
  good <- names(fullyNamed) %in% rgl.material.names
  if (warn && !all(good))
    warning("Argument(s) ", paste(names(fullyNamed)[!good], collapse = ", "), " not matched.")
  fullyNamed[good]
}

material3d  <- function(..., id = NULL) {
  args <- list(...)
  argnames <- names(args)
  if (length(argnames) && !is.null(id))
    stop("Material properties cannot be set on existing objects.")
  if (length(id) > 1)
   stop("Material properties may only be queried for single objects.")
  argnames <- setdiff(argnames, rgl.material.readonly)
  if (!length(args))
    argnames <- rgl.material.names
  else {
    if (is.null(names(args)) && all(unlist(lapply(args, is.character)))) {
      argnames <- unlist(args)
      args <- NULL
    }
    
    if (length(args) == 1) {
      if (is.list(args[[1]]) | is.null(args[[1]])) {
        args <- args[[1]]
        argnames <- names(args)
      }
    }
  }
  value <- rgl.getmaterial(id = id)[argnames]
  if (length(args)) {
    save <- options(rgl.warnBlackTexture = FALSE)
    on.exit(options(save))
    args <- do.call(".fixMaterialArgs", args)
    do.call("rgl.material0", args)
    return(invisible(value))
  } else if (length(argnames) == 1) return(value[[1]])
  else return(value)
}

bg3d        <- function(color,
                        sphere=FALSE, 
                        back="lines",
                        fogtype="none", 
                        fogScale = 1, 
                        col, ... ) {
  .check3d(); save <- material3d(); on.exit(material3d(save))

  bgid <- ids3d("background")$id
  if (length(bgid) && nrow(flags <- rgl.attrib(bgid[1], "flags"))) {
    if (missing(sphere))
      sphere <- flags["sphere", 1]
    if (missing(fogtype))
      fogtype <- if (flags["linear_fog", 1]) "linear"
      else if (flags["exp_fog", 1]) "exp"
      else if (flags["exp2_fog", 1]) "exp2"
      else "none"
    if (currentSubscene3d() == rootSubscene())
      rgl.incrementID()  # for back compatibility
  }
  dots <- list(...)
  
  if (!missing(color)) {
    dots$color <- color
    if (!missing(col))
      warning("'color' specified, so 'col' is ignored.")
  } else if (!missing(col))
    dots$color <- col
  
  if ("fogtype" %in% names(dots))
    fogtype <- dots$fogtype
  if ("fogScale" %in% names(dots))
    fogScale <- dots$fogScale
  else
    fogScale <- 1
  
  if ("color" %in% names(dots))
    color <- dots$color
  else if ("col" %in% names(dots))
    color <- dots$col
  else
    color <- getr3dDefaults("bg", "color")
  if (is.null(color))
    color <- "white"
    
  new <- .fixMaterialArgs(sphere = sphere, 
                          fogtype = fogtype, 
                          fogScale = fogScale,
                          color = color, 
  			  back = back, lit = FALSE, Params = save)
  
  do.call("rgl.material0", .fixMaterialArgs(..., Params = new))
  fogtype <- rgl.enum.fogtype(fogtype)
  
  idata   <- as.integer(c(sphere,fogtype))
  
  fogScale <- as.numeric(fogScale)
  stopifnot(length(fogScale) == 1, fogScale > 0)
  
  ret <- .C( rgl_bg, 
             success = as.integer(FALSE),
             idata,
             fogScale
  )
  
  if (! ret$success)
    stop("'rgl_bg' failed")
  
  if (length(bgid))
    pop3d(type = "background", id = bgid)
  
  lowlevel(ret$success)
}

light3d     <- function(theta=0, phi=15,
                        x=NULL, y = NULL, z = NULL,
                        viewpoint.rel = TRUE, 
                        ambient = "#FFFFFF", 
                        diffuse = "#FFFFFF", 
                        specular = "#FFFFFF") {
  .check3d()

  ambient  <- rgl.color(ambient)
  diffuse  <- rgl.color(diffuse)
  specular <- rgl.color(specular)
  
  # if a complete set of x, y, z is given, the light source is assumed to be part of the scene, theta and phi are ignored
  # else the light source is infinitely far away and its direction is determined by theta, phi (default) 
  
  if ( !is.null(x) ) {
    if ( !missing(theta) || !missing(phi) )
      warning("'theta' and 'phi' ignored when 'x' is present")
    xyz <- xyz.coords(x,y,z, recycle = TRUE)
    x <- xyz$x
    y <- xyz$y
    z <- xyz$z
    if (length(x) > 1) stop("A light can only be in one place at a time")
    finite.pos <- TRUE
  }
  else {
    
    if ( !is.null(y) || !is.null(z) ) 
      warning("'y' and 'z' ignored, 'theta' and 'phi' used")
    finite.pos <- FALSE
    x <- 0
    y <- 0
    z <- 0
  }
  
  idata <- as.integer(c(viewpoint.rel, ambient, diffuse, specular, finite.pos))
  ddata <- as.numeric(c(theta, phi, x, y, z))
  
  ret <- .C( rgl_light,
             success = as.integer(FALSE),
             idata,
             ddata
  )
  
  if (! ret$success)
    stop("Error in 'rgl_light'.  Too many lights? maximum is 8 sources per scene")
  
  lowlevel(ret$success)
}

view3d      <- function(theta = 0.0, phi = 15.0, 
                        fov = 60.0, zoom = 1.0, scale = par3d("scale"),
                        interactive = TRUE, userMatrix, 
                        type = c("userviewpoint", "modelviewpoint")) {
  .check3d()
  
  zoom <- rgl.clamp(zoom,0,Inf)
  phi  <- rgl.clamp(phi,-90,90)
  fov  <- rgl.clamp(fov,0,179)
  
  type <- match.arg(type, several.ok = TRUE)
  
  polar <- missing(userMatrix)
  if (polar) userMatrix <- diag(4)
  
  idata <- as.integer(c(interactive,polar, "userviewpoint" %in% type, "modelviewpoint" %in% type))
  ddata <- as.numeric(c(theta,phi,fov,zoom,scale,userMatrix[1:16]))
  
  ret <- .C( rgl_viewpoint,
             success = FALSE,
             idata,
             ddata
  )
  
  if (! ret$success)
    stop("'rgl_viewpoint' failed")
  lowlevel()
}

bbox3d	    <- function(xat = NULL, 
                        yat = NULL, 
                        zat = NULL, 
                        xunit = "pretty",
                        yunit = "pretty",
                        zunit = "pretty",
                        expand = 1.03, draw_front = FALSE,
                        xlab=NULL, ylab=NULL, zlab=NULL,
                        xlen=5, ylen=5, zlen=5,
                        marklen=15.0, marklen.rel=TRUE, 
                       ...) {
  .check3d(); save <- material3d(); on.exit(material3d(save))
  # Force evaluation of args as in old bbox3d
  list(xat=xat, yat=yat, zat=zat, 
       xunit=xunit, yunit=yunit, zunit=zunit, expand=expand,
       draw_front=draw_front)
  
  do.call(rgl.material0, .fixMaterialArgs(..., Params = save))
  
  if (is.null(xat)) 
    xlab <- NULL
  else {
    xlen <- length(xat)
    if (is.null(xlab)) 
      xlab <- format(xat)
    else 
      xlab <- rep(xlab, length.out=xlen)
  }
  if (is.null(yat)) 
    ylab <- NULL
  else {
    ylen <- length(yat)
    if (is.null(ylab)) 
      ylab <- format(yat)
    else 
      ylab <- rep(ylab, length.out=ylen)
  }
  if (is.null(zat)) 
    zlab <- NULL
  else {
    zlen <- length(zat)
    if (is.null(zlab)) 
      zlab <- format(zat)
    else 
      zlab <- rep(zlab,length.out=length(zat))
  }
  xticks <- length(xat)
  yticks <- length(yat)
  zticks <- length(zat)
  
  if (identical(xunit, "pretty")) xunit <- -1
  if (identical(yunit, "pretty")) yunit <- -1
  if (identical(zunit, "pretty")) zunit <- -1
  
  length(xlen)        <- 1
  length(ylen)        <- 1
  length(zlen)        <- 1
  length(marklen.rel) <- 1
  length(draw_front)  <- 1
  length(xunit)       <- 1
  length(yunit)       <- 1
  length(zunit)       <- 1
  length(marklen)     <- 1
  length(expand)      <- 1
  
  idata <- as.integer(c(xticks,yticks,zticks, xlen, ylen, zlen, marklen.rel, draw_front))
  ddata <- as.numeric(c(xunit, yunit, zunit, marklen, expand))
  
  ret <- .C( rgl_bbox,
             success = as.integer(FALSE),
             idata,
             ddata,
             as.numeric(xat),
             as.character(xlab),
             as.numeric(yat),
             as.character(ylab),
             as.numeric(zat),
             as.character(zlab)
  )
  
  if (! ret$success)
    stop("'rgl_bbox' failed")
  
  lowlevel(ret$success)
}

dummyBbox <- function()
  bbox3d(xat = numeric(), yat = numeric(), zat = numeric(),
         front = "cull", back = "cull")

observer3d <- function(x, y=NULL, z=NULL, auto=FALSE) {
  if (missing(x))
    location <- c(NA, NA, NA)
  else {
    xyz <- xyz.coords(x,y,z, recycle = TRUE)
    location <- c(xyz$x, xyz$y, xyz$z)
    if (length(location) != 3) stop("A single point must be specified for the observer location") 
  }    
  prev <- .C(rgl_getObserver, success=integer(1), ddata=numeric(3), NAOK = TRUE)$ddata
  .C(rgl_setObserver, success=as.integer(auto), ddata=as.numeric(location), NAOK = TRUE)
  lowlevel(prev)
}

# Shapes

points3d    <- function(x,y=NULL,z=NULL,...) {
  .check3d(); save <- material3d(); on.exit(material3d(save))
  do.call("rgl.primitive0", c(list(type = "points", x=x,y=y,z=z), .fixMaterialArgs(..., Params = save)))
}

lines3d     <- function(x,y=NULL,z=NULL,...) {
  .check3d(); save <- material3d(); on.exit(material3d(save))
  do.call("rgl.primitive0", c(list(type = "linestrips", x=x,y=y,z=z), .fixMaterialArgs(..., Params = save)))
}

segments3d  <- function(x,y=NULL,z=NULL,...) {
  .check3d(); save <- material3d(); on.exit(material3d(save))
  do.call("rgl.primitive0", c(list(type = "lines", x=x,y=y,z=z), .fixMaterialArgs(..., Params = save)))
}

triangles3d <- function(x,y=NULL,z=NULL,...) {
  .check3d(); save <- material3d(); on.exit(material3d(save))
  do.call("rgl.primitive0", c(list(type = "triangles", x=x, y=y, z=z), 
                             .fixMaterialArgs(..., Params = save)))
}

quads3d     <- function(x,y=NULL,z=NULL,...) {
  .check3d(); save <- material3d(); on.exit(material3d(save))
  do.call("rgl.primitive0", c(list(type = "quadrangles", x=x,y=y,z=z), .fixMaterialArgs(..., Params = save)))
}

text3d      <- function(x, y = NULL, z = NULL,
			texts, adj = 0.5, pos = NULL, offset = 0.5,
			usePlotmath = is.language(texts),
			family = par3d("family"), font = par3d("font"), 
			cex = par3d("cex"), useFreeType = par3d("useFreeType"),
			...) {
  if (usePlotmath) 
    return(plotmath3d(x = x, y = y, z = z, text = texts, adj = adj, 
                      pos = pos, offset = offset, cex = cex, ...))
  .check3d(); save <- material3d(); on.exit(material3d(save))
  do.call("rgl.material0", .fixMaterialArgs(..., Params = save))
  # Force evaluation
  list(x = x, y = y, z = z, text = texts, 
                              adj = adj, pos = pos,
                              offset = offset)
  
  vertex  <- rgl.vertex(x,y,z)
  nvertex <- rgl.nvertex(vertex)
  
  if (!is.null(pos)) {
    npos <- length(pos)
    stopifnot(all(pos %in% 0:6))
    stopifnot(length(offset) == 1)
    adj <- offset
  } else {
    pos <- 0
    npos <- 1
  }
  if (length(adj) > 3) warning("Only the first three entries of 'adj' are used")
  adj <- c(adj, 0.5, 0.5, 0.5)[1:3]
  
  if (!length(texts)) {
    if (nvertex)
      warning("No text to plot")
    return(invisible(integer(0)))
  }
  if (nvertex) {
    texts    <- rep(texts, length.out=nvertex)
    
    idata <- as.integer(nvertex)
    
    nfonts <- max(length(family), length(font), length(cex)) 
    family <- rep(family, length.out = nfonts)
    font <- rep(font, length.out = nfonts)
    cex <- rep(cex, length.out = nfonts)  
    
    family[font == 5] <- "symbol"
    font <- ifelse( font < 0 | font > 4, 1, font)  
    
    ret <- .C( rgl_texts,
               success = as.integer(FALSE),
               idata,
               as.double(adj),
               as.character(texts),
               as.numeric(vertex),
               as.integer(nfonts),
               as.character(family), 
               as.integer(font),
               as.numeric(cex),
               as.integer(useFreeType),
               as.integer(npos),
               as.integer(pos),
               NAOK=TRUE
    )
    
    if (! ret$success)
      stop("'rgl_texts' failed")
    
    lowlevel(ret$success)
  }
  
}
texts3d	    <- text3d

spheres3d   <- function(x, y = NULL, z = NULL, radius = 1, fastTransparency = TRUE, ...) {
  .check3d(); save <- material3d(); on.exit(material3d(save))
  # Force evaluation of args
  list(x = x, y = y, z = z, radius = radius, fastTransparency = fastTransparency)
  do.call("rgl.material0", .fixMaterialArgs(..., Params = save))
  
  vertex  <- rgl.vertex(x,y,z)
  nvertex <- rgl.nvertex(vertex)
  radius  <- rgl.attr(radius, nvertex)
  nradius <- length(radius)
  if (nvertex && nradius) {
    
    idata <- as.integer( c( nvertex, nradius ) )
    
    ret <- .C( rgl_spheres,
               success = as.integer(FALSE),
               idata,
               as.numeric(vertex),    
               as.numeric(radius),
               as.integer(fastTransparency),
               NAOK=TRUE
    )
    
    if (! ret$success)
      stop("'rgl_spheres' failed")
    
    lowlevel(ret$success)
  }
}

planes3d   <- function(a,b=NULL,c=NULL,d=0,...) {
  .check3d(); save <- material3d(); on.exit(material3d(save))
  # Force evaluation of args
  list(a=a,b=b,c=c,d=d)
  do.call("rgl.material0", .fixMaterialArgs(..., Params = save))
  normals  <- rgl.vertex(a, b, c)
  nnormals <- rgl.nvertex(normals)
  noffsets <- length(d)
  
  if (nnormals && noffsets) {
    
    idata <- as.integer( c( nnormals, noffsets ) )
    
    ret <- .C( rgl_planes,
               success = as.integer(FALSE),
               idata,
               as.numeric(normals),    
               as.numeric(d),
               NAOK=TRUE
    )
    
    if (! ret$success)
      stop("'rgl_planes' failed")
    
    lowlevel(ret$success)
  }
}

clipplanes3d   <- function(a,b=NULL,c=NULL,d=0) {
  .check3d()
  normals  <- rgl.vertex(a, b, c)
  nnormals <- rgl.nvertex(normals)
  noffsets <- length(d)
  if (nnormals && noffsets) {
    idata <- as.integer( c( nnormals, noffsets ) )
    
    ret <- .C( rgl_clipplanes,
               success = as.integer(FALSE),
               idata,
               as.numeric(normals),    
               as.numeric(d),
               NAOK=TRUE
    )
    
    if (! ret$success)
      stop("'rgl_clipplanes' failed")
    
    lowlevel(ret$success)
  }
}

abclines3d   <- function(x,y=NULL,z=NULL,a,b=NULL,c=NULL,...) {
  .check3d(); save <- material3d(); on.exit(material3d(save))
  # Force evaluation of args
  list(x=x,y=y,z=z,a=a,b=b,c=c)
  do.call("rgl.material0", .fixMaterialArgs(..., Params = save))
  
  bases  <- rgl.vertex(x, y, z)
  nbases <- rgl.nvertex(bases)
  directions <- rgl.vertex(a, b, c)
  ndirs <-  rgl.nvertex(directions)
  if (nbases && ndirs) {
    idata <- as.integer( c( nbases, ndirs ) )
    
    ret <- .C( rgl_abclines,
               success = as.integer(FALSE),
               idata,
               as.numeric(bases),    
               as.numeric(directions),
               NAOK=TRUE
    )
    
    if (! ret$success)
      stop("'rgl_abclines' failed")
    
    lowlevel(ret$success)
  }
}

sprites3d   <- function(x, y = NULL, z = NULL, radius = 1, 
                        shapes = NULL, userMatrix, fixedSize = FALSE,  
                        adj = 0.5, pos = NULL, offset = 0.25,
                        rotating = FALSE,
												...) {
  .check3d(); save <- material3d(); on.exit(material3d(save))
  
  do.call("rgl.material0", .fixMaterialArgs(..., Params = save))
  
  if (missing(userMatrix)) {
    userMatrix <- getr3dDefaults()$userMatrix
    if (is.null(userMatrix)) userMatrix <- diag(4)
  }
  savepar <- par3d(skipRedraw=TRUE, ignoreExtent=TRUE)
  on.exit(par3d(savepar), add=TRUE)
  
  par3d(ignoreExtent=savepar$ignoreExtent)
  # Force evaluation of args
  list(x=x,y=y,z=z,radius=radius,shapes=shapes,
       userMatrix=userMatrix, fixedSize = fixedSize, 
       rotating = rotating,
       adj = adj, pos = pos, offset = offset)
  center  <- rgl.vertex(x,y,z)
  ncenter <- rgl.nvertex(center)
  radius  <- rgl.attr(radius, ncenter)
  nradius <- length(radius)
  
  if (is.list(shapes)) {
    nshapelens <- length(shapes)
    if (!nshapelens)
      stop("Must have at least one shape")
    
    shapelens <- vapply(shapes, length, 1L)
    shapes <- unlist(shapes)
  } else if (length(shapes)) {
    nshapelens <- 1
    shapelens <- length(shapes)
  } else {
    nshapelens <- 0
    shapelens <- NULL
  }
  
  if (any(is.na(shapes)))
    stop("shapes must not be NA")
  
  nshapes <- length(shapes)
    
  pos <- as.integer(pos)
  npos <- length(pos)
  if (npos) {
    pos <- rep(pos, length.out = ncenter)
    adj <- offset
  }
  adj <- c(adj, 0.5, 0.5, 0.5)[1:3]
  if (ncenter && nradius) {
    if (length(shapes) && length(userMatrix) != 16) stop("Invalid 'userMatrix'")
    if (length(fixedSize) != 1) stop("Invalid 'fixedSize'")

    idata   <- as.integer( c(ncenter,nradius,nshapes, fixedSize, npos, rotating, 
                             nshapelens, shapelens ) )
    ret <- .C( rgl_sprites,
               success = as.integer(FALSE),
               idata,
               as.numeric(center),
               as.numeric(radius),
               as.integer(shapes),
               as.numeric(t(userMatrix)),
               as.numeric(adj),
               pos,
               as.numeric(offset),
               NAOK=TRUE
    )
    
    if (! ret$success)
      stop("'rgl_sprites' failed")
    
    lowlevel(ret$success)
  }
}

surface3d   <- function(x, y = NULL, z = NULL, ...,
                        normal_x = NULL, normal_y = NULL, normal_z = NULL,
                        texture_s = NULL, texture_t=NULL,
                        flip = FALSE) {
  .check3d(); save <- material3d(); on.exit(material3d(save))
  # Evaluate args
  list(x=x,y=z,z=y,coords=c(1,3,2),
       normal_x=normal_x,normal_y=normal_z,normal_z=normal_y)
  do.call("rgl.material0", .fixMaterialArgs(..., Params = save))
  flags <- rep(FALSE, 4)
  nrows <- NA
  ncols <- NA
  if (is.matrix(x)) {
    nrows <- nrow(x)
    ncols <- ncol(x)
    flags[1] <- TRUE
  } else nrows <- length(x)
  
  if (is.matrix(y)) {
    if (is.na(ncols))
      ncols <- ncol(y)
    if ( any(dim(y) != c(nrows, ncols))) 
      stop(gettextf("Bad dimension for %s", "y"),
                                             domain = NA)
    flags[2] <- TRUE
  } else {
    if (is.na(ncols))
      ncols <- length(y)
    else if (ncols != length(y))
      stop(gettextf("Bad length for %s", "y"),
           domain = NA)
  }
  
  if (is.matrix(z)) {
    if (any(dim(z) != c(nrows, ncols)))
      stop(gettextf("Bad dimension for %s", "z"))
  } else if (!flags[1] && !flags[2] && length(z) == nrows*ncols) {
      z <- matrix(z, nrows, ncols)
  } else if (!flags[2]) {
    if (!flags[1])
      stop("At least one coordinate must be a matrix")
    if (nrows != length(z))
      stop(gettextf("Bad length for %s", "z"))
    z <- matrix(z, nrows, ncols)
  } else {
    if (ncols != length(z))
      stop(gettextf("Bad length for %s", "z"))
    z <- matrix(z, nrows, ncols, byrow = TRUE)
  }
  
  nz <- length(z)
  
  if ( nrows < 2 )
    stop("rows < 2")
  
  if ( ncols < 2 )   
    stop("cols < 2")
  
  coords <- c(1,3,2)
  nulls <- c(is.null(normal_x), is.null(normal_y), is.null(normal_z))
  if (!all( nulls ) ) {
    if (any( nulls )) stop("All normals must be supplied")
    if ( !identical(dim(z), dim(normal_x)) 
         || !identical(dim(z), dim(normal_y))
         || !identical(dim(z), dim(normal_z)) ) stop(gettextf("Bad dimension for %s", "normals"),
                                                     domain = NA)
    flags[3] <- TRUE
  }
  
  nulls <- c(is.null(texture_s), is.null(texture_t))
  if (!all( nulls ) ) {
    if (any( nulls )) stop("Both texture coordinates must be supplied")
    if ( !identical(dim(z), dim(texture_s))
         || !identical(dim(z), dim(texture_t)) ) stop(gettextf("Bad dimension for %s", "textures"),
                                                      domain = NA)
    flags[4] <- TRUE
  }
  
  idata <- as.integer( c( nrows, ncols ) )
  
  xdecreasing <- diff(x[!is.na(x)][1:2]) < 0
  ydecreasing <- diff(y[!is.na(y)][1:2]) < 0
  parity <- (perm_parity(coords) + xdecreasing + ydecreasing + flip) %% 2
  
  if (is.na(parity))
    parity <- 0
  
  ret <- .C( rgl_surface,
             success = as.integer(FALSE),
             idata,
             as.numeric(x),
             as.numeric(y),
             as.numeric(z),
             as.numeric(normal_x),
             as.numeric(normal_y),
             as.numeric(normal_z),
             as.numeric(texture_s),
             as.numeric(texture_t),
             as.integer(coords),
             as.integer(parity),
             as.integer(flags),
             NAOK=TRUE
  )
  
  if (! ret$success)
    stop("'rgl_surface' failed")
  
  lowlevel(ret$success)
}
terrain3d <- surface3d

# Interaction

select3d    <- function(button = c("left", "middle", "right"), 
                        dev = cur3d(), subscene = currentSubscene3d(dev)) {
  .check3d()
  rect <- rgl.select(button = button, dev = dev, subscene = subscene)
  if (is.null(rect)) return(NULL)
  proj <- rgl.projection(dev = dev, subscene = subscene)
  
  llx <- rect[1]
  lly <- rect[2]
  urx <- rect[3]
  ury <- rect[4]
  
  selectionFunction3d(proj, region = c(llx, lly, urx, ury))
}

# 3D Generic Object Rendering Attributes

dot3d <- function(x,...) UseMethod("dot3d")
wire3d  <- function(x,...) UseMethod("wire3d")
shade3d <- function(x,...) UseMethod("shade3d")

# 3D Generic transformation


translate3d <- function(obj,x,y,z,...) UseMethod("translate3d")
scale3d <- function(obj,x,y,z,...) UseMethod("scale3d")
rotate3d <- function(obj,angle,x,y,z,matrix,...) UseMethod("rotate3d")
transform3d <- function(obj,matrix,...) rotate3d(obj, matrix = matrix, angle = NA, x = NA, y = NA, z = NA)

subdivision3d <- function(x,...) UseMethod("subdivision3d")

# 3D Custom shapes

particles3d <- function(x,y=NULL,z=NULL,radius=1,...) sprites3d(
  x=x,y=y,z=z,radius=radius,
  lit=FALSE,alpha=0.2,
  textype="alpha",
  texture=system.file("textures/particle.png",package="rgl"),
  ...
)   

# r3d default settings for new windows

r3dDefaults <- list(userMatrix = rotationMatrix(290*pi/180, 1, 0, 0),
		  mouseMode = c("none", "trackball", "zoom", "fov", "pull"),
		  FOV = 30,
		  bg = list(color="white",fogtype = "none"),
		  family = "sans",
		  material = list(color="black", fog = TRUE))

if (.Platform$OS.type == "windows")
	r3dDefaults$useFreeType <- FALSE

open3d <- function(..., params = getr3dDefaults(), 
                   useNULL = rgl.useNULL(), silent = FALSE	) {
	
	  register_pkgdown_methods()
	
    args <- list(...)
    if (missing(useNULL) && !is.null(params$useNULL)) {
      useNULL <- params$useNULL
      params$useNULL <- NULL
    }
    if (missing(silent) && !is.null(params$silent)) {
      silent <- params$silent
      params$silent <- NULL
    }
    antialias <- -1L
    if (!is.null(args$antialias) 
        || !is.null(args$antialias <- r3dDefaults$antialias)) {
    	saveopt <- options(rgl.antialias = args$antialias)
    	antialias <- args$antialias
    	on.exit(options(saveopt))
    	args$antialias <- NULL
    }
    
    ret <- .C( rgl_dev_open, success=FALSE, useNULL=useNULL,
               antialias = as.integer(antialias))
    
    if (! ret$success)
      stop("open failed") 
    
    if (!is.null(args$material)) {
    	params$material <- do.call(.fixMaterialArgs, c(args$material, Params=list(params$material)))
    	args$material <- NULL
    }
    
    if (length(args) && (is.null(names(args)) 
                      || any(nchar(names(args)) == 0)))
      stop("open3d parameters must be named")
    
    params[names(args)] <- args
        
    clear3d("material", defaults = params)
    params$material <- NULL
    
    if (!is.null(params$bg)) {
      do.call("bg3d", params$bg)
      params$bg <- NULL
    }
 
    do.call("par3d", params)  
    result <- structure(cur3d(), class = "rglOpen3d")
    if (silent)
      invisible(result)
    else
      result
}

print.rglOpen3d <- function(x, ...) {
  print(unclass(x))
}

close3d <- function(dev = cur3d(), silent = TRUE) {
  for (d in dev[dev != 0]) {
    devname <- paste0("dev", d)
    rgl.callback.env[[devname]] <- NULL
    set3d(d, silent = silent)
    if (length(hook <- getHook("on.rgl.close"))) {
      if (is.list(hook)) hook <- hook[[1]]  # test is for compatibility with R < 3.0.0
      hook()
    }
    
    ret <- .C( rgl_dev_close, success=FALSE )
    
    if (! ret$success)
      stop("close failed")
    
    if (!silent)
      message("Closed device ", d)
  }
  invisible(cur3d())
}

cur3d <- function() .Call( rgl_dev_getcurrent )
  
rgl.cur <- cur3d

set3d <- function(dev, silent = FALSE) {
  prev <- cur3d()
  
  idata <- c( as.integer(dev), as.integer(silent) )
  
  ret <- .C( rgl_dev_setcurrent, 
             success=FALSE, 
             idata
  )
  
  if (! ret$success)
    stop(gettextf("No device opened with id %s", dev), domain = NA)
  
  prev
}

.check3d <- function() {
    if (result<-cur3d()) return(result)
    else return(open3d())
}

requireWebshot2 <- function() {
  suppressMessages(res <- requireNamespace("webshot2", quietly = TRUE))
  if (res) 
    res <- requireNamespace("chromote") &&
           !is.null(path <- chromote::find_chrome()) &&
           nchar(path) > 0 &&
           file.exists(path)
  res
}

snapshot3d <- function(filename = tempfile(fileext = ".png"), 
                       fmt = "png", top = TRUE, ..., scene, width = NULL, height = NULL,
                       webshot = as.logical(Sys.getenv("RGL_USE_WEBSHOT", 
                                                                                                                     "TRUE"))) {
  force(filename)
  
  if (webshot && !requireWebshot2()) {
    warning("webshot = TRUE requires the webshot2 package and Chrome browser; using rgl.snapshot() instead")
    webshot <- FALSE
  }
  saveopts <- options(rgl.useNULL = webshot)
  on.exit(options(saveopts))
  
    # The logic here is a little convoluted:
  # scene  resize  webshot getscene1 plot resize getscene2
  # no     no      no      
  # no     no      yes         X
  # no     yes     no                       X
  # no     yes     yes         X      X     X      X                 
  # yes    no      no                 X
  # yes    no      yes     
  # yes    yes     no                 X     X
  # yes    yes     yes                X     X      X
  
    
  resize <- !is.null(width) || !is.null(height)
  havescene <- !missing(scene)
  
  if (havescene) {
    if (inherits(scene, "rglWebGL")) {
      snapshot <- scene$x$snapshot
      if (!is.null(snapshot) && is.null(width) && is.null(height))
        return(saveURI(snapshot, filename))
      else
        scene <- attr(scene, "origScene")
    }
  }
  if (!havescene && webshot)
    scene <- scene3d()
  
  if ((!havescene && resize && webshot)
      || (havescene && (resize || !webshot))) {
    open3d()
    plot3d(scene)
    on.exit(close3d(), add = TRUE)
  }
  if (resize) {
    saverect <- rect <- par3d("windowRect")
    on.exit(par3d(windowRect = saverect), add = TRUE)
    if (!is.null(width))
      rect[3] <- rect[1] + width
    if (!is.null(height))
      rect[4] <- rect[2] + height
    par3d(windowRect = rect)
  }
  if (webshot) {
    if (resize)
      scene <- scene3d()
    rect <- par3d("windowRect")
    f1 <- tempfile(fileext = ".html")
    # Work around path issue in webshot2:
    f1 <- normalizePath(f1, winslash = "/", mustWork = FALSE)
    on.exit(unlink(f1), add = TRUE)
    width <- rect[3] - rect[1]
    height <- rect[4] - rect[2]
    saveWidget(rglwidget(scene,
                         elementId = "webshot",
                         width = width,
                         height = height,
                         webgl = TRUE),
               f1,
               selfcontained = FALSE)
    unlink(filename)
    res <- try(capture.output(webshot2::webshot(f1, file = filename, selector = "#webshot",
                        vwidth = width + 100, vheight = height, ...),
                   type = "message"))
    if (!inherits(res, "try-error") && file.exists(filename) && file.size(filename) > 0)
      return(invisible(filename))
    
    warning("webshot2::webshot() failed; trying rgl.snapshot()")
  }
  rgl.snapshot(filename, fmt, top)
}
     
rgl.incrementID <- function(n = 1L) {
  .C(rgl_incrementID, n = as.integer(n))$n
}

test_that("mergeVertices works", {
  # Pre 4.0.0, stringsAsFactors defaulted to TRUE
  options(stringsAsFactors = FALSE)
	open3d()
  col <- rainbow(14)
  # Older R versions added FF for default alpha=1
  col <- sub("^(#......)FF$", "\\1", col)
	(mesh1 <- cuboctahedron3d(col = col, meshColor = "face"))
	id <- shade3d(mesh1)
	(mesh2 <- as.mesh3d(id))
	shade3d(translate3d(mesh2, 3, 0, 0))
	(mesh3 <- mergeVertices(mesh2))
	shade3d(translate3d(mesh3, 6, 0, 0))
	expect_known_scene("mergeVertices")
})

library(rgl)

test_that("surface dimensions work", {
	x <- 1:3
	y <- 1:5
	z <- matrix(1:15, 3, 5)
	open3d()
	surface3d(x, y, z, col = "red")
	surface3d(x, z, y, col = "blue")
	surface3d(z, y, x, col = "green")
	surface3d(x, z, z, col = "purple")
	surface3d(z, y, z, col = "orange")
	surface3d(z, z, y, col = "brown")
	surface3d(z, z, z, col = "white")
	surface3d(x, y, as.numeric(z + 1), col = "black")
	expect_error( surface3d(y, x, z))
	expect_error( surface3d(y, z, x))
	expect_error( surface3d(z, x, y))
	expect_error( surface3d(y, z, z))
	expect_error( surface3d(z, z, x))
  expect_error( surface3d(z, x, z))
  expect_error( surface3d(as.numeric(z), y, x))
  expect_error( surface3d(x, as.numeric(z), y))
})

set.seed(123)
test_that("3D sprites work", {
  open3d()
  # Warning:  these tests depend on order of evaluation
  # of the x, y, z arguments in sprites3d. 
  particles3d( rnorm(100), rnorm(100), rnorm(100), color = rainbow(100) )
  # is the same as
  sprites3d( rnorm(100), rnorm(100), rnorm(100), color = rainbow(100),
             lit = FALSE, alpha = .2,
             textype = "alpha", texture = system.file("textures/particle.png", package = "rgl") )
  sprites3d( rnorm(10) + 6, rnorm(10), rnorm(10), shape = shade3d(tetrahedron3d(), col = "red") )
  expect_known_scene("r3d")
})

library(rgl)

test_that("getShaders works", {
	x <- 1:3
	y <- 1:5
	z <- matrix(1:15, 3, 5)
	open3d()
	id <- surface3d(x, y, z, col = "red")
	if (requireNamespace("V8"))
	  expect_no_condition(getShaders(id))
})

test_that("ensureMatrix works", {
  expect_equal(dim(ensureMatrix(1, nrow=1)), c(1,1))
})

test_that("mesh3d works", {
	expect_s3_class(mesh3d(1:3, 1:3, 4, triangles=1:3), "mesh3d")
})


test_that("tmesh3d works", {
	open3d()
	m <- tmesh3d(rbind(1, 2, 1:3), 1:3, homogeneous=FALSE)
	expect_s3_class(m, "mesh3d")
	expect_named(m, c("vb", "it", "material", 
										"normals", "texcoords", "meshColor"),
							 ignore.order = TRUE)
	shade3d(m)
	expect_known_scene("tmesh3d")
})

test_that("qmesh3d works", {
	open3d()
	m <- qmesh3d(rbind(1, 2, 1:4), 1:4, homogeneous=FALSE)
	expect_s3_class(m, "mesh3d")
	expect_named(m, c("vb", "ib", "material", 
										"normals", "texcoords", "meshColor"),
							 ignore.order = TRUE)
	shade3d(m)
	expect_known_scene("qmesh3d")
})

test_that("shade3d, wire3d and dot3d work", {
	open3d()
	mesh <- cuboctahedron3d(col = "red")
	shade3d(mesh)
	wire3d(translate3d(mesh, 1,1,1))
	dot3d(translate3d(mesh, 2,2,2))
	expect_known_scene("shade3detc")
})

test_that("transformations work", {
	open3d()
	mesh <- cuboctahedron3d(col = "red")
	shade3d(translate3d(mesh, 1,2,3))
  shade3d(rotate3d(mesh, 35, 1,2,3))
  shade3d(scale3d(mesh, 1,2,3))
	expect_known_scene("transformations")
})

getBoundary3d <- function(mesh, sorted = FALSE, simplify = TRUE, ...) {
  if (!inherits(mesh, "mesh3d"))
    stop(deparse(substitute(mesh)), " is not a mesh3d object.")
  edges <- NULL
  if (length(mesh$it))
    edges <- cbind(edges, mesh$it[1:2,],  mesh$it[2:3,], mesh$it[c(3,1),])
  if (length(mesh$ib))
    edges <- cbind(edges, mesh$ib[1:2,], mesh$ib[2:3,], mesh$ib[3:4,], mesh$ib[c(4,1),])
  if (ncol(edges)) {
    # undirect the edges
    uedges <- t(apply(edges, 2, sort))
    dup <- duplicated(uedges) | duplicated(uedges, fromLast = TRUE)
    edges <- edges[,!dup,drop=FALSE]
    if (sorted) {
      order <- rep(NA, ncol(edges))
      order[1] <- 1
      for (i in seq_len(length(order) - 1)) {
        j <- which(edges[1,] == edges[2, order[i]])
        j <- setdiff(j, order[1:i])
        if (length(j))
          order[i+1] <- j[1]
        else
          order[i+1] <- setdiff(1:length(order), order[1:i])
      }
      edges <- edges[, order, drop = FALSE]
    }
  }
  result <- mesh3d(vertices = mesh$vb, segments = edges,
                   material = .fixMaterialArgs2(...))
  if (simplify)
    result <- cleanMesh3d(result)
  result
}

# Get the attribute expanded by the indices, if any
expandAttrib <- function(id, attrib) {
  result <- rgl.attrib(id, attrib)
  if (length(result)) {
    indices <- rgl.attrib(id, "indices")
    if (length(indices))
      result <- result[indices,]
  }
  result
}

expandVertices <- function(id) 
  expandAttrib(id, "vertices")

expandColors <- function(id) {
  result <- rgl.attrib(id, "colors")
  if (nrow(result) > 1)
    result <- expandAttrib("colors")
  result
}

getExpandedNverts <- function(id) {
  result <- length(rgl.attrib(id, "indices"))
  if (!result)
    result <- NROW(rgl.attrib(id, "vertices"))
  result
}

getIndices <- function(id) {
  result <- rgl.attrib(id, "indices")
  if (!length(result))
    result <- seq_len(NROW(rgl.attrib(id, "vertices")))
  result
}

# Check for indices that differ from default
checkForIndices <- function(id, warn = TRUE) {
  has_indices <- length(indices <- rgl.attrib(id, "indices")) > 0 && 
                 !all(indices == seq_along(indices))
  if (has_indices && warn)
    warning("Indices not supported.  Skipping id", id, call. = FALSE)
  has_indices
}

##
## bring device to top
##
##

rgl.bringtotop <- function(stay = FALSE) {

  if ((.Platform$OS.type != "windows") && stay) warning("'stay' not implemented")
  ret <- .C( rgl_dev_bringtotop, success=FALSE, as.logical(stay) )

  if (! ret$success)
    stop("'rgl.bringtotop' failed")

}

subsetControl <- function(value = 1, subsets, subscenes = NULL,
                         fullset = Reduce(union, subsets),
                         accumulate = FALSE) {
  subsets <- lapply(subsets, as.integer)
  fullset <- as.integer(fullset)
  if (length(names(subsets)))
    labels <- names(subsets)
  else
    labels <- NULL
  structure(list(type = "subsetSetter",
       value = value - 1,
       subsets = unname(subsets),
       subscenes = subscenes,
       fullset = fullset,
       accumulate = accumulate,
       labels = labels),
      class = "rglControl")
}

propertyControl <- function(value = 0, entries, properties, objids = tagged3d(tags), tags, values = NULL,
                            param = seq_len(NROW(values)) - 1, interp = TRUE) {
  objids <- as.integer(objids)
  structure(list(type = "propertySetter",
       value = value,
       values = values,
       entries = entries,
       properties = properties,
       objids = objids,
       param = param,
       interp = interp),
      class = "rglControl")
}

clipplaneControl <- function(a=NULL, b=NULL, c=NULL, d=NULL,
                            plane = 1, clipplaneids = tagged3d(tag), tag,
                            ...) {
  values <- cbind(a = a, b = b, c = c, d = d)
  col <- which(colnames(values) == letters[1:4]) - 1
  propertyControl(values = values, entries = 4*(plane-1) + col,
                  properties = "vClipplane", objids = clipplaneids,
                  ...)
}

ageControl <- function(births, ages, objids = tagged3d(tags), tags, value = 0, colors = NULL, alpha = NULL,
                       radii = NULL, vertices = NULL, normals = NULL,
                       origins = NULL, texcoords = NULL,
                       x = NULL, y = NULL, z = NULL,
                       red = NULL, green = NULL, blue = NULL) {

  lengths <- c(colors = NROW(colors), alpha = length(alpha),
               radii = length(radii), vertices = NROW(vertices),
               normals = NROW(normals), origins = NROW(origins),
               texcoords = NROW(texcoords),
               x = length(x), y = length(y), z = length(z),
               red = length(red), green = length(green), blue = length(blue))
  lengths <- lengths[lengths > 0]
  n <- unique(lengths)
  stopifnot(length(n) == 1, n == length(ages), all(diff(ages) >= 0))

  ages <- c(-Inf, ages, Inf)
  rows <- c(1, 1:n, n)

  result <- list(type = "ageSetter",
                 objids = as.integer(objids),
                 value = value,
                 births = births,
                 ages = ages)

  if (!is.null(colors)) {
    colors <- col2rgb(colors)/255
    colors <- as.numeric(colors[,rows])
    result <- c(result, list(colors = colors))
  }

  if (!is.null(alpha))
    result <- c(result, list(alpha = alpha[rows]))

  if (!is.null(radii))
    result <- c(result, list(radii = radii[rows]))

  if (!is.null(vertices)) {
    stopifnot(ncol(vertices) == 3)
    result <- c(result, list(vertices = as.numeric(t(vertices[rows,]))))
  }

  if (!is.null(normals)) {
    stopifnot(ncol(normals) == 3)
    result <- c(result, list(normals = as.numeric(t(normals[rows,]))))
  }

  if (!is.null(origins)) {
    stopifnot(ncol(origins) == 2)
    result <- c(result, list(origins = as.numeric(t(origins[rows,]))))
  }

  if (!is.null(texcoords)) {
    stopifnot(ncol(texcoords) == 2)
    result <- c(result, list(texcoords = as.numeric(t(texcoords[rows,]))))
  }

  if (!is.null(x))
    result <- c(result, list(x = x[rows]))

  if (!is.null(y))
    result <- c(result, list(y = y[rows]))

  if (!is.null(z))
    result <- c(result, list(z = z[rows]))

  if (!is.null(red))
    result <- c(result, list(red = red[rows]))

  if (!is.null(green))
    result <- c(result, list(green = green[rows]))

  if (!is.null(blue))
    result <- c(result, list(blue = blue[rows]))

  structure(result, class = "rglControl")
}

vertexControl <- function(value = 0, values = NULL, vertices = 1, attributes, objid = tagged3d(tag), tag,
                          param = seq_len(NROW(values)) - 1, interp = TRUE) {
  attributes <- match.arg(attributes,
                          choices = c("x", "y", "z",
                                      "red", "green", "blue", "alpha",
                                      "radii",
                                      "nx", "ny", "nz",
                                      "ox", "oy", "oz",
                                      "ts", "tt",
                          	      "offset"),
                          several.ok = TRUE)
  if (!is.null(values)) {
    ncol <- max(length(vertices), length(attributes))
    if (is.matrix(values))
      stopifnot(ncol == ncol(values))
    else {
      stopifnot(ncol == 1)
      values <- matrix(values, ncol = 1)
    }
    # Repeat first and last values to make search simpler.
    param <- c(-Inf, param, Inf)
    values <- rbind(values[1,], values, values[nrow(values),])
  }

  structure(list(type = "vertexSetter",
       value = value,
       values = values,
       vertices = vertices - 1, # Javascript 0-based indexing
       attributes = attributes,
       objid = as.integer(objid),
       param = param,       # Javascript 0-based indexing
       interp = interp),
      class = "rglControl")
}

par3dinterpControl <- function(fn, from, to, steps, subscene = NULL,
			      omitConstant = TRUE, ...) {
  rename <- character()
  times <- seq(from, to, length.out = steps+1)
  fvals <- lapply(times, fn)
  f0 <- fvals[[1]]
  entries <- numeric(0)
  properties <- character(0)
  values <- NULL
	
  props <- c("FOV", "userMatrix", "scale", "zoom")
  for (i in seq_along(props)) {
    prop <- props[i]
    propname <- rename[prop]
    if (is.na(propname))
      propname <- prop
    if (!is.null(value <- f0[[prop]])) {
      newvals <- sapply(fvals, function(e) as.numeric(e[[prop]]))
      if (is.matrix(newvals)) newvals <- t(newvals)
      rows <- NROW(newvals)
      cols <- NCOL(newvals)
      stopifnot(rows == length(fvals))
      entries <- c(entries, seq_len(cols)-1)
      properties <- c(properties, rep(propname, cols))
      values <- cbind(values, newvals)
    }
  }
  if (omitConstant) keep <- apply(values, 2, var) > 0
  else keep <- TRUE
	
  if (is.null(subscene)) subscene <- f0$subscene
	
  propertyControl(values = c(t(values[,keep])), entries = entries[keep],
		 properties = properties[keep],
		 objids = subscene, param = times, ...)
}

# This is a bridge to the old system
# In the old system, the rglClass object was a global named
# <prefix>rgl, and controls install methods on it.  In the
# new system, the rglClass object is just a field of a <div>
# element.  The R code below creates an empty global for the
# controls to modify, then the Javascript code in oldBridge
# imports those into the real scene object.

elementId2Prefix <- function(elementId, prefix = elementId) {
  .Defunct("", msg = "This function is not needed if you use rglwidget().")
}

# This puts together a custom message for a more extensive change

sceneChange <- function(elementId, x = scene3d(minimal),
                        delete = NULL, add = NULL, replace = NULL,
                        material = FALSE, rootSubscene = FALSE,
                        delfromSubscenes = NULL, skipRedraw = FALSE,
                        minimal = TRUE) {
  allSubscenes <- function() {
    result <- numeric()
    for (obj in scene$objects)
      if (obj$type == "subscene")
        result <- c(result, obj$id)
    result
  }
  inSubscenes <- function(id, subs) {
    result <- numeric()
    for (sub in subs)
      if (id %in% sub$objects)
        result <- c(result, sub$id)
    result
  }
  delete <- unique(c(delete, replace))
  add <- unique(c(add, replace))

  scene <- convertScene(x, useBuffer = FALSE)
  allsubids <- allSubscenes()
  allsubs <- scene$objects[as.character(allsubids)]
  for (id in add)
    scene$objects[[as.character(id)]]$inSubscenes <- inSubscenes(id, allsubs)

  scene$elementId <- elementId
  allIds <- names(scene$objects)
  dontSend <- setdiff(allIds, as.character(add))
  scene$objects[dontSend] <- NULL
  if (!length(scene$objects))
    scene$objects <- NULL
  if (!material)
    scene$material <- NULL
  if (!rootSubscene)
    scene$rootSubscene <- NULL
  scene$delete <- delete
  if (is.null(delfromSubscenes))
    delfromSubscenes <- allsubids
  scene$delfromSubscenes <- as.numeric(delfromSubscenes)
  if (is.na(skipRedraw))
    scene$redrawScene <- FALSE
  else {
    scene$redrawScene <- !skipRedraw
    scene$skipRedraw <- skipRedraw
  }
  scene
}

registerSceneChange <- function() {
  tags$script(rglDependency, '
Shiny.addCustomMessageHandler("sceneChange",
  rglwidgetClass.prototype.sceneChangeHandler);
')
}

selectpoints3d <- function(objects = ids3d()$id, value = TRUE, closest = TRUE, 
                           multiple = FALSE, ...) {

  if (value) result <- cbind(x = numeric(0), y = numeric(0), z = numeric(0))
  else result <- cbind(id = integer(0), index = integer(0))

  first <- TRUE
  prevdist <- dist <- Inf
  
  while (first || is.function(multiple) || multiple) {
    f <- select3d(...)
    if (is.null(f)) break
    
    e <- environment(f)
    
    prev <- nrow(result) # Number to keep from previous selection
    
    for (id in objects) {
      verts <- expandVertices(id)
      hits <- f(verts)
      
      if (any(hits)) dist <- 0
      else if (closest && dist > 0 && nrow(verts)) {
        wincoords <- rgl.user2window(verts, projection = e$proj)
        wz <- wincoords[,3]
        keep <- (0 <= wz) & (wz <= 1)
        wincoords <- wincoords[keep,,drop=FALSE]
  
        if (!nrow(wincoords)) next 
  
        wx <- wincoords[,1]
        xdist <- ifelse(wx < e$llx, (wx-e$llx)^2, ifelse(wx < e$urx, 0, (wx-e$urx)^2))
   
        wy <- wincoords[,2]      
        ydist <- ifelse(wy < e$lly, (wy-e$lly)^2, ifelse(wy < e$ury, 0, (wy-e$ury)^2))  
  
        dists <- xdist + ydist
        hits <- (dists < dist) & (dists == min(dists))
        dist <- min(c(dist, dists))
      }
      
      if (!any(hits)) next
      
      if (prev && prevdist > dist) {
        result <- result[FALSE, , drop = FALSE]
        prev <- 0
      }
        
      if (value)
        result <- rbind(result, verts[hits,])
      else {
        indices <- getIndices(id)[which(hits)]
        result <- rbind(result, cbind(id, indices))
      }
      if (is.function(multiple) && nrow(result) > prev
          && !multiple(result[(prev+1):nrow(result),,drop=FALSE]))
        break  
        
      prevdist <- dist
      prev <- nrow(result)
      first <- FALSE
    }
    
    if (value)
      result <- unique(result)
  
  }
  result
}

ensureMatrix <- function(x, nrow) {
  if (!is.matrix(x))
    x <- matrix(x, nrow = nrow)
  x
}

mesh3d <- function( x, y = NULL, z = NULL, vertices, 
                    material = NULL,
                    normals = NULL, texcoords = NULL,
                    points = NULL, segments = NULL,
                    triangles = NULL, quads = NULL,
                    meshColor = c("vertices", "edges", "faces", "legacy")) {
  
  if (missing(vertices)) {
    xyz <- xyz.coords(x, y, z, recycle=TRUE)
    if (length(xyz$x) && length(xyz$y) && length(xyz$z))
      vertices <- rbind(xyz$x, xyz$y, xyz$z, 1) 
    else
      vertices <- matrix(numeric(), nrow = 4)
  } else vertices <- asHomogeneous2(vertices)
  
  # Remove dimnames
  dimnames(vertices) <- NULL

  if (missing(meshColor) 
      && !is.null(material) 
      && !is.null(material$meshColor)) {
    meshColor <- material$meshColor
    material$meshColor <- NULL
  }
  
  meshColor <- match.arg(meshColor)
  
  if (is.null(texcoords)
      && !is.null(material)
      && !is.null(material$texcoords)) {
    texcoords <- material$texcoords
    material$texcoords <- NULL
  }
  
  nvertex <- ncol(vertices)
  
  if ( !is.null(normals) ) {
    normals <- xyz.coords(normals, recycle=TRUE)
    x <- rep(normals$x, length.out = nvertex)
    y <- rep(normals$y, length.out = nvertex)
    z <- rep(normals$z, length.out = nvertex)
    normals <- rgl.vertex(x, y, z)
  }
  
  if ( !is.null(texcoords) ) {
    texcoords <- xy.coords(texcoords, recycle = TRUE)
    x <- rep(texcoords$x, length.out = nvertex)
    y <- rep(texcoords$y, length.out = nvertex)
    texcoords <- rbind(x, y)
  }
  
  object <- list(
    vb = vertices,
    material = .getMaterialArgs(material = material),
    normals = normals,
    texcoords = texcoords,
    meshColor = meshColor
  ) 
  
  if (!is.null(points))    object$ip <- ensureMatrix(points, 1)
  if (!is.null(segments))  object$is <- ensureMatrix(segments, 2)
  if (!is.null(triangles)) object$it <- ensureMatrix(triangles, 3)
  if (!is.null(quads))     object$ib <- ensureMatrix(quads, 4)

  register_compare_proxy()
  
  class(object) <- c("mesh3d", "shape3d")
  object
}

register_compare_proxy <- local({
  registered <- FALSE
  function() {
    if (!registered &&
        isNamespaceLoaded("waldo")) {
      if ("path" %in% names(formals(waldo::compare_proxy))) { # appeared after 0.2.5
        registerS3method("compare_proxy", "mesh3d", 
                         compare_proxy.mesh3d, 
                         envir = asNamespace("waldo"))
        registerS3method("compare_proxy", "rglscene",
                         compare_proxy.rglscene,
                         envir = asNamespace("waldo"))        
      } else {
        registerS3method("compare_proxy", "mesh3d", 
                         old_compare_proxy.mesh3d, 
                         envir = asNamespace("waldo"))
        registerS3method("compare_proxy", "rglscene",
                         old_compare_proxy.rglscene,
                         envir = asNamespace("waldo")) 
      }
      registered <<- TRUE
    }
  }
})

compare_proxy.mesh3d <- function(x, path = "x") {
  list(object = old_compare_proxy.mesh3d(x),
       path = paste0("compare_proxy(", path, ")"))
}

old_compare_proxy.mesh3d <- function(x) {
  for (n in names(x)) # Some elements are NULL, some are not there
    if (is.null(x[[n]])) x[[n]] <- NULL
  if (is.null(x$material) ||
      length(x$material$color) < 2)
    x$meshColor <- NULL  # It doesn't matter.
  for (n in c("ip", "is", "it", "ib"))
    if (!is.null(x[[n]]))
      x[[n]] <- as.numeric(x[[n]])
  x$primitivetype <- NULL # Not used since before 0.100.x
  x[sort(names(x))]
}

#
# triangle mesh object
#

tmesh3d <- function( vertices, indices, homogeneous=TRUE, material=NULL, normals=NULL,
                     texcoords = NULL,
                     meshColor = c("vertices", "edges", "faces", "legacy")) {
  
  if (missing(meshColor) && !is.null(material$meshColor))
    meshColor <- material$meshColor
  meshColor <- match.arg(meshColor)
  if (is.null(dim(vertices)))
    vertices <- matrix(vertices, nrow = if (homogeneous) 4 else 3)
  
  # For back-compatibility:
  colnames(indices) <- NULL   
  
  mesh3d(vertices = vertices, triangles = indices,
         material = material, normals = normals, texcoords = texcoords,
         meshColor = meshColor)
}

#
# R 3d object : quad mesh
#

qmesh3d <- function( vertices, indices, homogeneous=TRUE, material=NULL, normals=NULL,
                     texcoords=NULL,
                     meshColor = c("vertices", "edges", "faces", "legacy")) {
  
  if (missing(meshColor) && !is.null(material$meshColor))
    meshColor <- material$meshColor
  meshColor <- match.arg(meshColor)
  
  if (is.null(dim(vertices)))
    vertices <- matrix(vertices, nrow = if (homogeneous) 4 else 3)
  
  # For back-compatibility
  colnames(indices) <- NULL   
  
  mesh3d(vertices = vertices, quads = indices,
         material = material, normals = normals, texcoords = texcoords,
         meshColor = meshColor)
}

as.mesh3d <- function(x, ...) UseMethod("as.mesh3d")

as.mesh3d.deldir <- function(x, col = "gray", coords = c("x", "y", "z"), 
			     smooth = TRUE, normals = NULL, texcoords = NULL,
			     ...) {
  checkDeldir(error = TRUE)
  if (!identical(sort(coords), c("x", "y", "z")))
    stop(sQuote("coords"), " should be a permutation of c('x', 'y', 'z')")
  if (!all(coords %in% names(x$summary)))
    stop("The 'deldir' object needs x, y, and z coordinates.")
  if (smooth && !is.null(normals)) {
    warning("'smooth' ignored as 'normals' was specified.")
    smooth <- FALSE
  }
  pointnames <- as.numeric(rownames(x$summary))
  points <- matrix(NA, max(pointnames), 3)
  points[pointnames, ] <- as.matrix(x$summary[, coords])
  points <- t(points)
  triangs <- do.call(rbind, deldir::triang.list(x))

  if (!is.null(texcoords))
    texcoords <- texcoords[triangs$ptNum, ]
  material <- .getMaterialArgs(...)
  material$color <- col
  result <- mesh3d(vertices = asHomogeneous2(points), triangles = triangs$ptNum,
  	  normals = normals, texcoords = texcoords,
  	  material = material)
  if (smooth)
    result <- addNormals(result)
  result
}

as.mesh3d.triSht <-
as.mesh3d.tri <- function(x, z, col = "gray", 
                          coords = c("x", "y", "z"), 
                          smooth = TRUE, normals = NULL, texcoords = NULL,
                          ...) {
  if (inherits(x, "tri")) 
    triangles <- tripack::triangles
  else if (inherits(x, "triSht")) {
    triangles <- interp::triangles
  }
  if (!identical(sort(coords), c("x", "y", "z")))
    stop(sQuote("coords"), " should be a permutation of c('x', 'y', 'z')")
  if (!is.numeric(z) || length(z) != x$n)
    stop("z should be a numeric vector with one entry per node of x")
  if (smooth && !is.null(normals)) {
    warning("'smooth' ignored as 'normals' was specified.")
    smooth <- FALSE
  }
  points <- rbind(x$x, x$y, z)
  rownames(points) <- c("x", "y", "z")
  points <- points[coords,]
  
  triangs <- t(triangles(x)[, 1:3])
  if (inherits(x, "tri") && x$nc) {
    constraintIndex <- min(x$lc)
    keep <- apply(triangs, 2, function(col) any(col < constraintIndex))
    triangs <- triangs[, keep]
  }
  
  if (!is.null(texcoords))
    texcoords <- texcoords[triangs, ]
  material <- .getMaterialArgs(...)
  material$color <- col
  result <- mesh3d(vertices = asHomogeneous2(points), triangles = triangs,
                    normals = normals, texcoords = texcoords,
                    material = material)
  if (smooth)
    result <- addNormals(result)
  result  
}

# NO OP (issue #454)
as.mesh3d.mesh3d <- function(x, ...) x

# rendering support

dot3d.mesh3d <- function(x, ..., front = "points", back = "points") {
  if (missing(front) && !is.null(x$material$front))
    front <- x$material$front
  if (missing(back) && !is.null(x$material$back))
    back <- x$material$back
  shade3d.mesh3d(x, ..., front = front, back = back)
}

wire3d.mesh3d <- function(x, ..., front = "lines", back = "lines") {
  if (missing(front) && !is.null(x$material$front))
    front <- x$material$front
  if (missing(back) && !is.null(x$material$back))
    back <- x$material$back
  shade3d.mesh3d(x, ..., front = front, back = back)
}

allowedMeshColor <- function(meshColor, modes) {
  meshColor != "edges" || 
    (modes[1] == modes[2] && modes[1] == "lines")
}

shade3d.mesh3d <- function( x, override = TRUE, 
                             meshColor = c("vertices", "edges", "faces", "legacy"), 
                             texcoords = NULL, 
                             ...,
                             front = "filled", back = "filled") {
  if (missing(front) && !is.null(x$material$front))
    front <- x$material$front
  if (missing(back) && !is.null(x$material$back))
    back <- x$material$back
  argMaterial <- c(list(front = front, back = back), .getMaterialArgs(...))
  xHasColor <- !is.null(x$material) && !is.null(x$material$color)
  hasMeshColor <- !missing(meshColor)
  if ( override ) {
    material <- x$material
    if (is.null(material)) material <- list()    
    material[names(argMaterial)] <- argMaterial
    if (hasMeshColor || is.null(x$meshColor))
      meshColor <- match.arg(meshColor)
    else
      meshColor <- x$meshColor
    if (is.null(texcoords) && !is.null(x$texcoords))
      texcoords <- t(x$texcoords)
  } else {
    material <- argMaterial
    material[names(x$material)] <- x$material
    if (is.null(x$meshColor))
      meshColor <- match.arg(meshColor)
    else
      meshColor <- x$meshColor
    if (!is.null(x$texcoords))
      texcoords <- t(x$texcoords)
  }
  normals <- x$normals
  if (!is.null(normals))
    normals <- t(normals)
  
  modes <- c(front, back)
  if (!allowedMeshColor(meshColor, modes))
    stop("'meshColor = ", meshColor, "' not supported.")
  
  if (meshColor != "legacy") {
    if (is.null(material$color))
      material$color <- material3d("color")
    if (is.null(material$alpha))
      material$alpha <- material3d("alpha")
  }
  
  vertices <- t(asEuclidean2(x$vb))
  
  result <- integer(0)
  prev <- 0
  
  doVertices <- function(vals, inds, setPrev) {
    inds <- as.numeric(inds)
    if (length(unique(vals)) > 1) {
      vals <- rep_len(vals, max(inds))
      vals[inds]
    } else
      vals
  }
  doLegacy <- function(vals, inds, setPrev) {
    inds <- as.numeric(inds) + prev
    if (setPrev)
      prev <<- prev + length(inds)
    if (length(unique(vals)) > 1)
      vals <- rep_len(vals, length(inds))
    vals
  }
  doFaces <- function(vals, inds, setPrev) {
    if (!is.matrix(inds))
      inds <- matrix(inds, nrow = 1)
    if (prev) {
    	vals <- rep_len(vals, prev + ncol(inds))
    	vals <- vals[-seq_len(prev)]
    }
    if (setPrev)
      prev <<- prev + ncol(inds)
    vals <- rep_len(vals, ncol(inds))
    rep(vals, each = nrow(inds))
  }
  doEdges <- function(vals, inds, setPrev) {
    if (!is.matrix(inds)) 
      inds <- matrix(inds, nrow = 1)
    inds <- inds + prev
    if (setPrev)
      prev <<- prev + length(inds)
    newlen <- ncol(inds)
    if (nrow(inds) > 1)
      newlen <- newlen * nrow(inds) / 2
    vals <- rep_len(vals, newlen)
    if (nrow(inds) > 1)
      vals <- rep(vals, each = 2)
    vals
  }
  getArgs <- function(inds) {
    args <- c(list(x = vertices[as.numeric(inds),]), material)
    if (!is.null(texcoords))
      args$texcoords <- texcoords[as.numeric(inds),]
    if (!is.null(normals))
      args$normals <- normals[as.numeric(inds),]
    args$color <- repfn(args$color, inds, is.null(args$alpha))
    args$alpha <- repfn(args$alpha, inds, TRUE)
    args
  }
  
  repfn <- switch(meshColor, 
                  vertices = doVertices,
                  legacy = doLegacy,
                  faces = doFaces,
                  doLegacy)  
  if (length(x$ip)) {
    args <- getArgs(x$ip)
    result <- c(result, points = do.call(points3d, args = args ))
  }
  
  if (length(x$is)) {
    args <- getArgs(x$is)
    result <- c(result, segments = do.call(segments3d, args = args ))
  }
  
  if (length(x$it)) {
    if (meshColor == "edges") {
      x$it <- x$it[c(1,2,2,3,3,1),]
      fn <- segments3d
    } else
      fn <- triangles3d
    args <- getArgs(x$it)
    result <- c(result, triangles = do.call(fn, args = args ))
  }

  if (length(x$ib)) {
    if (meshColor == "edges") {
      x$ib <- x$ib[c(1,2,2,3,3,4,4,1),]
      fn <- segments3d
    } else
      fn <- quads3d
    args <- getArgs(x$ib)
    result <- c(result, quads = do.call(fn, args = args ))
  }
  
  lowlevel(result)
}

# transformation support

translate3d.mesh3d <- function( obj, x, y, z, ... ) {
  obj$vb <- t(translate3d(t(obj$vb), x, y, z))
  return(obj)                            
}  

rotate3d.mesh3d <- function( obj,angle,x,y,z,matrix, ... ) {
  obj$vb <- t(rotate3d(t(obj$vb), angle, x, y, z, matrix))
  if ( !is.null(obj$normals) ) {
    normals <- obj$normals  # 4xn matrix
    if ( missing(matrix) ) 
      normals <- rotate3d(t(normals), angle, x, y, z) # nx4
    else {
      if (nrow(matrix) == 4) matrix[4,1:3] <- 0
      if (ncol(matrix) == 4) matrix[1:3,4] <- 0
      normals <- rotate3d(t(normals), angle, x, y, z, t(solve(matrix))) # nx4
      # Reverse the normals if the transformation
      # changes the orientation of the mesh
      if (det(matrix) < 0)
        normals[, 1:3] <- -normals[, 1:3]  # nx4
    }
    normals <- asEuclidean(normals)  # nx3
    normals <- normals/sqrt(rowSums(normals^2))
    obj$normals <- rbind( t(normals), 1 ) # 4xn
  }
  return(obj)                            
}  

scale3d.mesh3d <- function( obj, x, y, z, ... ) {
  obj$vb <- t(scale3d(t(obj$vb), x, y, z))
  if ( !is.null(obj$normals) ) {
    obj$normals <- scale3d(t(obj$normals), 1/x, 1/y, 1/z)
    obj$normals <- t( obj$normals/sqrt(apply(obj$normals[,1:3]^2, 1, sum)) )
    if (nrow(obj$normals) == 4)
      obj$normals[4,] <- 1
  }
  return(obj)
}

# for debugging
showNormals <- function(obj, scale = 1) {
  v <- obj$vb
  n <- obj$normals
  if (is.null(n)) {
    warning("No normals found.")
    return()
  }
  save <- par3d(skipRedraw = TRUE)
  on.exit(par3d(save))
  for (i in seq_len(ncol(n))) {
    p0 <- v[1:3, i]/v[4, i]
    p1 <- p0 + n[1:3, i]*scale
    if (all(is.finite(c(p0, p1))))
      arrow3d(p0, p1, type = "lines")
  }
}

print.mesh3d <- function(x, prefix = "", ...) {
  cat(prefix, " mesh3d object with ", ncol(x$vb), " vertices, ", 
      paste(c(
        if (length(x$ip)) paste(length(x$ip), "points"),
        if (length(x$is)) paste(ncol(x$is), "segments"),
        if (length(x$it))  paste(ncol(x$it), "triangles"),
        if (length(x$ib))  paste(ncol(x$ib), "quads")), collapse = ", "),
      ".\n", sep = "")
}

match_names <- function(x, reference) {
  if (!is.null(x))
    structure(x[names(reference)], class = class(x))
}

# Compare old and new meshes
all.equal.mesh3d <- function(target, current, ...) {
  if (inherits(current, "mesh3d"))
    all.equal(compare_proxy.mesh3d(target), 
              compare_proxy.mesh3d(current), ...)
  else
    "'current' is not a mesh3d object"
}

rgl.par3d.names <- c("antialias", "FOV", "ignoreExtent", "listeners",
            "mouseMode", "observer", 
            "modelMatrix", "projMatrix", "skipRedraw", "userMatrix",
            "userProjection",
            "scale", "viewport", "zoom", "bbox", "windowRect",
            "family", "font", "cex", "useFreeType", "fontname",
            "maxClipPlanes", "glVersion", "activeSubscene"
)

rgl.par3d.readonly <- c( 
  "antialias", "observer",
  "modelMatrix", "projMatrix",
  "bbox", "fontname",
  "maxClipPlanes", "glVersion",
  "activeSubscene"
)

par3d <- function(..., no.readonly = FALSE, dev = cur3d(), subscene = currentSubscene3d(dev)) {
  single <- FALSE
  args <- list(...)
  if (!length(args))
    args <- as.list(if (no.readonly)
      rgl.par3d.names[-match(rgl.par3d.readonly, rgl.par3d.names)]
      else rgl.par3d.names)
  else {
    if (is.null(names(args)) && all(unlist(lapply(args, is.character))))
      args <- as.list(unlist(args))
    if (length(args) == 1) {
      if (is.list(args[[1]]) | is.null(args[[1]]))
        args <- args[[1]]
      else
        if(is.null(names(args)))
          single <- TRUE
    }
  }
  if ("dev" %in% names(args)) {
    if (!missing(dev) && dev != args[["dev"]]) stop("'dev' specified inconsistently")
    dev <- args[["dev"]]
    args[["dev"]] <- NULL
  }
  if (specifiedSubscene <- ("subscene" %in% names(args))) {
    if (!missing(subscene) && subscene != args[["subscene"]]) stop("'subscene' specified inconsistently")
    subscene <- args[["subscene"]]
    args[["subscene"]] <- NULL
  }
  dev <- as.integer(dev)
  if (!dev) dev <- open3d()
  subscene <- as.integer(subscene)
  
  if ("userMatrix" %in% names(args)) {
    m <- args$userMatrix
    svd <- svd(m[1:3, 1:3])
    m[1:3, 1:3] <- svd$u %*% t(svd$v)
    theta <- atan2(-m[1,3], m[1,1])
    m <-  m %*% rotationMatrix(theta, 0,1,0)
    svd <- svd(m[1:3, 1:3])
    m[1:3,1:3] <- svd$u %*% t(svd$v)	
    phi <- atan2(-m[2,3], m[3,3])
    args$.position <- c(theta, phi)*180/pi
  }
  if ("mouseMode" %in% names(args)) {
    m <- args$mouseMode
    if (is.null(names(m))) {
      if (length(m) < 5) 
        args$mouseMode <- c("none", m)
    } else {
      m0 <- par3d("mouseMode")
      m0[names(m)] <- m
      args$mouseMode <- m0
    }
  }
  if (forceViewport <- ("windowRect" %in% names(args) &&
      !("viewport" %in% names(args)))) {
    if (specifiedSubscene)
      warning("Viewport for subscene ", subscene,
              " will be adjusted; other viewports will not be.")
    oldviewport <- .Call(rgl_par3d, dev, subscene, list("viewport","windowRect"))
  }
  value <-
    if (single) .Call(rgl_par3d, dev, subscene, args)[[1]] 
  else .Call(rgl_par3d, dev, subscene, args)
  
  # The windowRect might be modified by the window manager (if
  # too large, for example), so we need to read it after the
  # change
  if (forceViewport) {
    oldsize <- oldviewport$windowRect[3:4] - oldviewport$windowRect[1:2]
    Sys.sleep(0.1)
    windowRect <- .Call(rgl_par3d, dev, subscene, list("windowRect"))$windowRect
    newsize <- windowRect[3:4] - windowRect[1:2]
    .Call(rgl_par3d, dev, subscene, 
          list(viewport = round(oldviewport$viewport*newsize/oldsize)))
  }
  
  if(!is.null(names(args))) invisible(value) else value
}

writePLY <- function(con, format=c("little_endian", "big_endian", "ascii"), 
                     pointRadius=0.005, 
                     pointShape = icosahedron3d(),
                     lineRadius = pointRadius,
                     lineSides = 20,
                     pointsAsEdges = FALSE,
                     linesAsEdges = pointsAsEdges,
                     withColors = TRUE,
                     withNormals = !(pointsAsEdges || linesAsEdges),
                     ids = tagged3d(tags), tags = NULL) {
 
  writeData <- function() {
    cat("ply\n", file=con)
    fmt <- switch(format,
      little_endian = "binary_little_endian",
      big_endian = "binary_big_endian",
      ascii = "ascii")
    
    cat("format", fmt, "1.0\n", file=con)
    cat("element vertex", nrow(Vertices), "\n", file=con)
    cat("property float x
property float y
property float z\n", file=con)
    if (withNormals)
      cat("property float nx
property float ny
property float nz\n", file=con)
    if (withColors)
      cat("property uchar red
property uchar green
property uchar blue
property uchar alpha\n", file=con)
    cat("element face", nrow(Triangles)+nrow(Quads), "\n", file=con)
    cat("property list uchar int vertex_indices\n", file=con)
    if (nrow(Edges) > 0) {
      cat("element edge", nrow(Edges), "\n", file=con)
      cat("property int vertex1
property int vertex2\n", file=con)
    }
    cat("end_header\n", file=con)
    
    if (format == "ascii") {
      for (i in seq_len(nrow(Vertices))) {
        cat(Vertices[i,], file=con)
        if (withColors)
          cat("", Colors[i,], file=con)
        cat("\n", file=con)
      }
      for (i in seq_len(nrow(Triangles))) 
        cat("3", Triangles[i,], "\n", file=con)
      for (i in seq_len(nrow(Quads)))
        cat("4", Quads[i,], "\n", file=con)
      for (i in seq_len(nrow(Edges)))
        cat(Edges[i,], "\n", file=con)
    } else {
      endian <- if (format == "little_endian") "little" else "big"
      if (nrow(Vertices)) {
        if (!withColors)
          writeBin(as.numeric(t(Vertices)), con, size = 4, endian=endian)
        else {
          for (i in seq_len(nrow(Vertices))) {
            writeBin(Vertices[i,], con, size = 4, endian = endian)
            writeBin(Colors[i,], con, size = 1)
          }
        }
      }
      if (nrow(Triangles)) {
        for (i in seq_len(nrow(Triangles))) {
          writeBin(3L, con, size = 1)
          writeBin(as.integer(Triangles[i,]), con, size = 4, 
                   endian = endian)
        }
      }
      if (nrow(Quads)) {
        for (i in seq_len(nrow(Quads))) {
          writeBin(4L, con, size = 1)
          writeBin(as.integer(Quads[i,]), con, size = 4, 
                   endian = endian)
        }
      }
      if (nrow(Edges))
        writeBin(as.integer(t(Edges)), con, size=4, endian=endian)
    }  
  }  
    
  Vertices <- matrix(0, 0, 3 + 3*withNormals)
  Colors <- matrix(0L, 0, 4)
  
  Triangles <- matrix(1L, 0, 3)
  Quads <- matrix(1L, 0, 4)
  Edges <- matrix(1L, 0, 2)

  getVertices <- function(id) {
    vertices <- rgl.attrib(id, "vertices")
    if (withNormals) {
      normals <- rgl.attrib(id, "normals")
      if (!nrow(normals))
      	normals <- 0*vertices
    }
    cbind(vertices, 
          if (withNormals) normals)
  }
  
  getColors <- function(id) {
    vertices <- rgl.attrib(id, "vertices")
    if (withColors) {
      colors <- rgl.attrib(id, "colors")
      if (nrow(colors) == 1)
        colors <- colors[rep(1, nrow(vertices)),, drop = FALSE]
      colors <- round(255*colors)
      mode(colors) <- "integer"
      colors
    } else NULL
  }
  
  writeTriangles <- function(id) {
    vertices <- getVertices(id)
    colors <- getColors(id)
    n <- nrow(vertices)      
    base <- nrow(Vertices)
    Vertices <<- rbind(Vertices, vertices)
    Colors <<- rbind(Colors, colors)
    Triangles <<- rbind(Triangles, matrix(base + getIndices(id) - 1, 
                        ncol=3, byrow=TRUE))
  }
  
  writeQuads <- function(id) {
    vertices <- getVertices(id)
    colors <- getColors(id)
    n <- nrow(vertices)
    base <- nrow(Vertices)
    Vertices <<- rbind(Vertices, vertices)
    Colors <<- rbind(Colors, colors)
    Quads <<- rbind(Quads, matrix(base + getIndices(id) - 1, 
                        ncol=4, byrow=TRUE))
  }
      
  writeSurface <- function(id) {
    vertices <- getVertices(id)
    colors <- getColors(id)
    dims <- rgl.attrib(id, "dim")
    nx <- dims[1]
    nz <- dims[2]
    base <- nrow(Vertices)
    Vertices <<- rbind(Vertices, vertices)
    Colors <<- rbind(Colors, colors)
    rows <- seq_len(nx)
    for (i in seq_len(nz)[-nz]) {
      indices <- getIndices(id)[(i-1)*nx + 
                                c(rows[-nx],rows[-nx],
                                  rows[-1]+nx,rows[-nx]+nx,
                                  rows[-1],rows[-1]+nx)]
      Triangles <<- rbind(Triangles, 
                        matrix(base + indices - 1,
                               ncol=3))
    }
  }
  
  writeMesh <- function(mesh, scale=1, offset=c(0,0,0)) {
    vertices <- asEuclidean(t(mesh$vb))*scale 
    vertices <- vertices + rep(offset, each=nrow(vertices))
    if (withColors) {
      colors <- mesh$material$col
      if (!length(colors)) colors <- material3d("color")
      colors <- rep(colors, length.out = nrow(vertices))
      colors <- t(col2rgb(colors, alpha=TRUE))
      Colors <<- rbind(Colors, colors)
    }
    if (withNormals) 
      normals <- asEuclidean(t(mesh$normals))
    base <- nrow(Vertices)
    Vertices <<- rbind(Vertices, cbind(vertices, 
                                       if (withNormals) normals))
    
    nt <- length(mesh$it)/3
    nq <- length(mesh$ib)/4
    if (nt) 
      Triangles <<- rbind(Triangles, t(mesh$it) - 1 + base)
    if (nq)
      Quads <<- rbind(Quads, t(mesh$ib) - 1 + base)
  }

  writeSpheres <- function(id) {
    vertices <- expandVertices(id)
    n <- nrow(vertices)    
    colors <- expandColors(id)
    if (nrow(colors) == 1)
      colors <- colors[rep(1, n),, drop = FALSE]
    radii <- expandAttrib(id, "radii")
    radii <- rep(radii, length.out=n)
    x <- subdivision3d(icosahedron3d(),3)
    r <- sqrt(x$vb[1,]^2 + x$vb[2,]^2 + x$vb[3,]^2)
    x$vb[4,] <- r
    x$normals <- x$vb
    for (i in seq_len(n)) {
      col <- colors[i,]
      x$material$col <- rgb(col[1], col[2], col[3], col[4], maxColorValue = 255)
      writeMesh(x, radii[i], vertices[i,])
    }
  }  
  
  avgScale <- function() {
    bbox <- par3d("bbox")
    ranges <- c(bbox[2]-bbox[1], bbox[4]-bbox[3], bbox[6]-bbox[5])
    if (prod(ranges) == 0) 1
    else exp(mean(log(ranges)))
  }  
   
  writePoints <- function(id) {
    vertices <- getVertices(id)
    colors <- getColors(id)
    n <- nrow(vertices)
    inds <- getIndices(id)    
    if (pointsAsEdges) {
      base <- nrow(Vertices)
      Vertices <<- rbind(Vertices, vertices)
      Colors <<- rbind(Colors, colors)
      Edges <<- rbind(Edges, base + cbind(inds, inds) - 1 )  
    } else {
      radius <- pointRadius*avgScale()
      if (withNormals && is.null(pointShape$normals))
        pointShape <- addNormals(pointShape)
      for (i in inds) {
        if (withColors) {
          col <- vertices[i,4:7]
          pointShape$material$col <- rgb(col[1], col[2], col[3], col[4], maxColorValue = 255)
        }
        writeMesh(pointShape, radius, vertices[i,1:3])
      }
    }
  }
  
  writeSegments <- function(id) {
    vertices <- getVertices(id)
    colors <- getColors(id)
    inds <- getIndices(id)
    n <- length(inds)/2
    if (linesAsEdges) {
      base <- nrow(Vertices)
      Vertices <<- rbind(Vertices, vertices)
      Colors <<- rbind(Colors, colors)
      Edges <<- rbind(Edges, base + matrix(inds, ncol = 2, byrow = TRUE) - 1)  
    } else {
      radius <- lineRadius*avgScale()
      for (i in seq_len(n)) {
        cyl <- cylinder3d( vertices[inds[(2*i-1):(2*i)],1:3],
     			   radius = radius,
     			   sides = lineSides, 
     			   closed = -2 )
       if (withColors) {
          col1 <- colors[inds[2*i-1],]
          col1 <- rgb(col1[1], col1[2], col1[3], col1[4], maxColorValue = 255)
          col2 <- colors[inds[2*i],]
          col2 <- rgb(col2[1], col2[2], col2[3], col2[4], maxColorValue = 255)
        
          cyl$material$col <- c(rep(col1, lineSides),
                                rep(col2, lineSides), col1, col2)
        }
        if (withNormals)
          cyl <- addNormals(cyl)
        writeMesh(cyl)
      }
    }
  }
  
  writeLines <- function(id) {
    vertices <- getVertices(id)
    colors <- getColors(id)
    if (linesAsEdges) {
      inds <- getIndices(id)
      n <- length(inds)   
      base <- nrow(Vertices)
      Vertices <<- rbind(Vertices, vertices)
      Colors <<- rbind(Colors, colors)
      Edges <<- rbind(Edges, base + cbind(inds[-n], inds[-1]) - 1)  
    } else {
      n <- nrow(vertices) - 1
      radius <- lineRadius*avgScale()
      colors <- vertices[, 4:7, drop = FALSE]
      for (i in seq_len(n)) {
        cyl <- cylinder3d( vertices[inds[i:(i+1)],1:3],
     			   radius = radius,
     			   sides = lineSides, 
     			   closed = -2 )
        if (withColors) {
          col1 <- colors[inds[i],]
          col1 <- rgb(col1[1], col1[2], col1[3], col1[4], maxColorValue = 255)
          col2 <- colors[inds[i+1],]
          col2 <- rgb(col2[1], col2[2], col2[3], col2[4], maxColorValue = 255)
        
          cyl$material$col <- c(rep(col1, lineSides),
                                rep(col2, lineSides), col1, col2)
        }
        if (withNormals) 
          cyl <- addNormals(cyl)
        writeMesh(cyl)
      }
    }
  }
  
  knowntypes <- c("triangles", "quads",
                  "surface", "spheres", "linestrip", "lines", "planes",
                  "points")
  
  #  Execution starts here!
  format <- match.arg(format)

  if (is.character(con)) {
    con <- file(con, if (format=="ascii") "w" else "wb")
    on.exit(close(con))
  }
  filename <- summary(con)$description
  
  if (NROW(bbox <- ids3d("bboxdeco")) && (is.null(ids) || bbox$id %in% ids)) {
    ids <- setdiff(ids, bbox$id)
    save <- par3d(skipRedraw = TRUE)
    bbox <- convertBBox(bbox$id)
    on.exit({ pop3d(id=bbox); par3d(save) }, add=TRUE) # nolint
    dobbox <- TRUE
  } else dobbox <- FALSE 
  
  if (is.null(ids)) {
    ids <- ids3d()
    types <- as.character(ids$type)
    ids <- ids$id
  } else {
    if (dobbox) ids <- c(ids, bbox)
    allids <- ids3d()
    ind <- match(ids, allids$id)
    keep <- !is.na(ind)
    if (any(!keep)) warning(gettextf("Object(s) with id %s not found", paste(ids[!keep], collapse=" ")), 
    			    domain = NA)
    ids <- ids[keep]
    types <- allids$type[ind[keep]]
  }  
    
  unknowntypes <- setdiff(types, knowntypes)
  if (length(unknowntypes))
    warning(gettextf("Object type(s) %s not handled", 
      paste("'", unknowntypes, "'", sep="", collapse=", ")), domain = NA)

  keep <- types %in% knowntypes
  ids <- ids[keep]
  types <- types[keep]

  for (i in seq_along(ids))
    switch(types[i],
      planes =,
      triangles = writeTriangles(ids[i]),
      quads = writeQuads(ids[i]),
      surface = writeSurface(ids[i]),
      spheres = writeSpheres(ids[i]),
      points = writePoints(ids[i]),
      lines = writeSegments(ids[i]),
      linestrip = writeLines(ids[i])
    )
  
  writeData()
  
  invisible(filename)
}

options(rgl.useNULL=FALSE)
suppressPackageStartupMessages(library(rgl))
options(rgl.useNULL=TRUE)
options(rgl.printRglwidget=FALSE)

if (!requireNamespace("rmarkdown", quietly = TRUE) || 
    !rmarkdown::pandoc_available("1.14")) {
  warning(call. = FALSE, "These vignettes assume rmarkdown and pandoc version 1.14.  These were not found. Older versions will not work.")
  knitr::knit_exit()
}

# knitr::opts_chunk$set(snapshot = TRUE)  # for snapshots instead of dynamic

documentedfns <- c()
deprecatedfns <- c()

backticked <- function(s) paste0("`", s, "`")

indexfns <- function(fns, text = backticked(fns), show = TRUE, pkg = "rgl") {
  documentedfns <<- c(documentedfns, fns)
  anchors <- paste0('<a name="', fns, '">',
                    if (show) linkfn(fns, text, pkg = pkg),
                    '</a>')
  paste(anchors, collapse=if (show) ", " else "")
}

deprecated <- function(fns, text = backticked(fns), show = TRUE, pkg = "rgl") {
  deprecatedfns <<- c(deprecatedfns, fns)
  if (show)
    paste(text, collapse = ", ")
}

indexclass <-
  indexproperties <- function(fns, text = backticked(fns), show = TRUE) {
    documentedfns <<- c(documentedfns, fns)
    anchors <- paste0('<a name="', fns, '">',
                      if (show) text,
                      '</a>')
    paste(anchors, collapse=if (show) ", " else "")
  }

indexmethods <- function(fns, text = backticked(paste0(fns, "()")), show = TRUE) {
  documentedfns <<- c(documentedfns, fns)
  anchors <- paste0('<a name="', fns, '">',
                    if (show) text,
                    '</a>')
  paste(anchors, collapse=if (show) ", " else "")
}

linkfn <- function(fn, text = backticked(fn), pkg = NA) {
  if (is.na(pkg))
    paste0('<a href="#', fn, '">', text, '</a>')
  else {
    if (requireNamespace("downlit", quietly = TRUE))
      url <- downlit::autolink_url(paste0(pkg, "::", fn))
    else
      url <- NA
    if (is.na(url))
      url <- paste0('../../', pkg, '/help/', fn)

    paste0('<a href="', url, '">', text, '</a>')
  }
}

# Write this once at the start of the document.

cat('<style>
    .nostripes tr.even {background-color: white;}
    table {border-style: none;}
    table th {border-style: none;}
    table td {border-style: none;}
    a[href^=".."] {text-decoration: underline;}
    </style>
    ')

writeIndex <- function(cols = 4) {
  if (!is.null(documentedfns)) {
    documentedfns <- sort(documentedfns)
    entries <- paste0('<a href="#', documentedfns, '">', documentedfns, '</a>&nbsp;&nbsp;')
    len <- length(entries)
    padding <- ((len + cols - 1) %/% cols) * cols - len
    if (padding)
      entries <- c(entries, rep("", length.out=padding))
    cat('\n<div class="nostripes">\n')
    print(knitr::kable(matrix(entries, ncol=cols), format="pandoc"))
    cat("</div>\n")
  }
}

# This displays the string code as `r code` when entered
# as `r rinline(code)`.  Due to Stephane Laurent
rinline <- function(code, script = FALSE){
  if (script)
    html <- "`r CODE`"
  else
    html <- '<code  class="r">``` `r CODE` ```</code>'
  sub("CODE", code, html)
}

rmarkdownOutput <- function() {
  if (requireNamespace("rmarkdown", quietly = TRUE)) {
    output <- rmarkdown::metadata$output
    if (length(output))
      if (is.character(output)) return(output[1])
      else if (is.list(output) && length(names(output))) return(names(output)[1]) 
  }
  NULL
}

rglShared <- function(id, key = NULL, group = NULL,
		      deselectedFade = 0.1, 
		      deselectedColor = NULL,
		      selectedColor = NULL,
		      selectedIgnoreNone = TRUE,
		      filteredFade = 0,
		      filteredColor = NULL) {
  if (!requireNamespace("crosstalk"))
    stop("This function requires crosstalk.")
  data <- as.data.frame(rgl.attrib(id, "vertices"))
  attr(data, "rglId") <- as.integer(id)
  attr(data, "rglOptions") <- list(deselectedFade = deselectedFade,
  				   deselectedColor = if (!is.null(deselectedColor)) as.numeric(col2rgb(deselectedColor, alpha = TRUE)/255),
  				   selectedColor =   if (!is.null(selectedColor)) as.numeric(col2rgb(selectedColor, alpha = TRUE)/255), 
  				   selectedIgnoreNone = selectedIgnoreNone,
  				   filteredFade = filteredFade,
  				   filteredColor =   if (!is.null(filteredColor)) as.numeric(col2rgb(filteredColor, alpha = TRUE)/255))
  n <- nrow(data)
  if (!n)
    stop("No vertices in object ", id)
  if (!is.null(key) && (n != length(key) || anyDuplicated(key)))
    stop("'key' must have exactly one unique value for each vertex")
  result <- if (is.null(group))
    crosstalk::SharedData$new(data, key)
  else
    crosstalk::SharedData$new(data, key, group)
  structure(result, class = c("rglShared", class(result)))
}

CSStoPixels <- function(x, DPI = 100) {
  if (is.null(x))
    return(x)
  num <- function(x)
    as.numeric(sub("[^[:digit:].]*$", "", x))
  units <- function(x)
    sub("^[[:digit:].]+", "", x)
  if (!is.numeric(x)) {
    units <- units(x)
    if (units == "auto")
      stop("Only fixed CSS sizes allowed")
    val <- num(x)
    if (units != "")
      val <- switch(units,
          "px" = val,
          "in" = val * DPI,
          "cm" = val * DPI / 2.54,
          "mm" = val * DPI / 254,
          "pt" = val * DPI / 72,
          "pc" = val * DPI / 6,
          stop("Only fixed CSS sizes allowed")
        )
  } else
    val <- x
  val
}

# These sizes are taken from htmlwidgets/R/sizing.R
DEFAULT_WIDTH <- 960
DEFAULT_HEIGHT <- 500
DEFAULT_PADDING <- 40
DEFAULT_WIDTH_VIEWER <- 450
DEFAULT_HEIGHT_VIEWER <- 350
DEFAULT_PADDING_VIEWER <- 15

# For widgets, we use the sizingPolicy to see how it would be
# displayed
resolveHeight <- function(x, inViewer = TRUE, default = 40) {
  if (inViewer)
    refsize <- DEFAULT_HEIGHT_VIEWER
  else
    refsize <- DEFAULT_HEIGHT
  result <- x$height
  if (is.null(result) && !is.null(policy <- x$sizingPolicy)) {
    if (inViewer) {
      viewer <- policy$viewer
      if (isTRUE(viewer$fill))
        result <- refsize
      else 
        result <- viewer$defaultHeight
    } else
      result <- NULL
    if (is.null(result) && isTRUE(policy$fill))
      result <- refsize
    if (is.null(result))
      result <- policy$defaultHeight
    if (is.null(result))
      result <- refsize
  }
  if (is.null(result))
    result <- default 

  CSStoPixels(result, refsize)  
}

getWidgetId <- function(widget) {
  if (inherits(widget, "htmlwidget"))
    widget$elementId
  else {
    NULL
  }
}

# Get information from previous objects being piped into 
# this one, and modify a copy of them as necessary

getHeights <- function(objects, defaultHeight = 40) {
  if (inherits(objects, "combineWidgets"))
    heights <- objects$params$rowsize
  else {
    if (inherits(objects, c("shiny.tag", "htmlwidget")) || 
        !is.list(objects))
      objects <- tagList(objects)
    heights <- rep(defaultHeight, length(objects))
    for (i in seq_along(objects)) {
      tag <- objects[[i]]
      if (inherits(tag, "rglWebGL") && is.null(tag$height))
        heights[i] <- tag$x$height
      else if (inherits(tag, "htmlwidget"))
        heights[i] <- resolveHeight(tag)
      else if (is.list(tag) && 
               !is.null(tag$height) && 
               !is.na(height <- suppressWarnings(as.numeric(tag$height))))
        heights[i] <- height
    }
  }
  heights
}

processUpstream <- function(upstream, elementId = NULL, playerId = NULL) {
  rowsizes <- getHeights(upstream)
  
  if (inherits(upstream, "combineWidgets")) 
    upstream <- upstream$widgets

  if (inherits(upstream, "knit_image_paths") && length(upstream))
    upstream <- img(src = image_uri(upstream[1]))
  
  if (inherits(upstream, c("shiny.tag", "htmlwidget")))
    upstream <- tagList(upstream)
       
  if (is.character(upstream) && !is.na(upstream[1]))
    return(list(prevRglWidget = upstream))
  
  if (is.list(upstream)) {
    # Objects upstream of the current one may need to know about an rgl widget,
    # or this object may need to know about an upstream rgl widget.  Stop when
    # you find one.
    lookForRglWidget <- function(upstream) {
      prevRglWidget <- NULL
      players <- character()
      for (i in rev(seq_along(upstream))) {
        tag <- upstream[[i]]
        if (inherits(tag, "rglWebGL")) {
          prevRglWidget <- tag$elementId
          if (is.null(prevRglWidget))
            prevRglWidget <- tag$elementId <- upstream[[i]]$elementId <- newElementId("rgl")
          if (!is.null(playerId) && !(playerId %in% tag$x$players))
            upstream[[i]]$x$players <- c(tag$x$players, playerId)
        } else if (inherits(tag, "rglPlayer") && is.null(tag$x$sceneId)) {
          players <- c(players, tag$elementId)
          if (!is.null(elementId))
            upstream[[i]]$x$sceneId <- elementId
        } else if (inherits(tag, "shiny.tag") && !tagHasAttribute(tag, "rglSceneId")) {
          upstream[[i]] <- tagAppendAttributes(tag, rglSceneId = elementId)
        } else if (inherits(tag, "combineWidgets")) {
          temp <- lookForRglWidget(tag$widgets)
          players <- c(players, temp$players)
          prevRglWidget <- temp$prevRglWidget
          upstream[[i]]$widgets <- temp$objects
        }
        if (!is.null(prevRglWidget))
          break
      }
      list(objects = upstream,
           players = players,
           prevRglWidget = prevRglWidget)
    }
    result <- lookForRglWidget(upstream)
    result$rowsizes <- rowsizes
  } else
    result <- list(objects = upstream,
                   players = if (is.character(upstream)) upstream else character(),
                   prevRglWidget = if (is.character(upstream)) upstream,
                   rowsizes = rowsizes)
        
  result     
}

asRow <- function(..., last = NA, height = NULL, colsize = 1) {
	if (!requireNamespace("manipulateWidget", quietly = TRUE)) {
		warning("asRow requires the 'manipulateWidget' package.", call. = FALSE)
		last <- NA
	}
  args <- list(...)
  if ((length(args) == 1 
      && inherits(args[[1]], "combineWidgets"))
  		|| !requireNamespace("manipulateWidget", quietly = TRUE)) {
    orig <- args[[1]]
  } else {
    orig <- do.call(manipulateWidget::combineWidgets, c(args, list(ncol = 1, rowsize = getHeights(args))))
  }
  origlen <- length(orig$widgets)
  for (i in seq_len(origlen))
    if (inherits(orig$widgets[[i]], "knit_image_paths"))
      orig$widgets[[i]] <- img(src = image_uri(orig$widgets[[i]]))
  if (is.na(last))
    last <- origlen
  else if (last > origlen)
    stop("'last' must be no more than the number of widgets")
  keep <- seq_len(origlen - last)
  inrow <- seq_len(last) + origlen - last
  origRowsizes <- rep_len(orig$params$rowsize, origlen)
  if (length(inrow)) {
    maxinrow <- max(origRowsizes[inrow])
    if (is.null(height))
      height <- maxinrow
  } else if (is.null(height))
    height <- 0
  
  orig$params$rowsize <- c(origRowsizes[keep], height)

  if (requireNamespace("manipulateWidget", quietly = TRUE)) {
    row <- do.call(manipulateWidget::combineWidgets, c(orig$widgets[inrow], list(nrow = 1, 
                                                             colsize = colsize)))
    orig$widgets <- c(orig$widgets[keep], list(row))
  }
  orig
}

newElementId <- function(prefix)
  paste0(prefix, p_sample(100000, 1))

isMarkdownHTMLformat <- function() {
  output <- rmarkdown::metadata$output
  
  if (is.list(output))
    output <- names(output)
  
  is.character(output) && 
    length(output) >= 1 &&
    output[1] == "markdown::html_format"
}

knitrNeedsSnapshot <- function(options = knitr::opts_current$get()) {
  if (!is.null(options$snapshot))
    return(options$snapshot)
  if (isFALSE(options$screenshot.force))
    return(FALSE)
  force <- isTRUE(options$screenshot.force)
  if (isMarkdownHTMLformat())
    return(FALSE)
  fmt <- pandoc_to()
  if (!length(fmt) || force)
    return(TRUE)
  html_format <- fmt %in% c("html", "html4", "html5", "revealjs", 
                            "s5", "slideous", "slidy")
  !html_format
}

rglwidget <- local({
  function(x = scene3d(minimal), width = figWidth(), height = figHeight(),
           controllers = NULL, 
           elementId = NULL,
           reuse = FALSE,
           webGLoptions = list(preserveDrawingBuffer = TRUE), 
  	       shared = NULL, 
           minimal = TRUE, 
           webgl,
           snapshot,
           shinyBrush = NULL, 
           altText = "3D plot", ...,
           oldConvertBBox = FALSE,
           fastTransparency = getOption("rgl.fastTransparency", TRUE)) {
    
  if (missing(snapshot)) {
    if (missing(webgl)) {
      if (in_knitr())
        snapshot <- knitrNeedsSnapshot()
      else
        snapshot <- FALSE
    } else 
      snapshot <- !webgl
  } else {
    if (!is.logical(snapshot)) {
      stop("snapshot must be TRUE or FALSE")
    }
  }
  if (missing(webgl)) webgl <- !snapshot
      
  if (webgl == snapshot || is.na(webgl) || is.na(snapshot))
    stop("Must specify either 'snapshot' or 'webgl' but not both")
  
  origScene <- x
  force(shared) # It might plot something...

  if (is.null(elementId) && 
      (!inShiny() || # If in Shiny, all of the classes below need the ID
       inherits(controllers, c("combineWidgets", "shiny.tag", "htmlwidget"))))
    elementId <- newElementId("rgl")

  if (!is.null(shinyBrush)) {
    if (!is.character(shinyBrush) || length(shinyBrush) != 1)
      stop("'shinyBrush' must be a single character value")
    if (!inShiny())
      warning("'shinySelectionInput' is only used in Shiny")
    else
      x$selectionInput <- shinyBrush
  }
  
  if (!inherits(x, "rglscene"))
    stop("First argument should be an rgl scene.")
  
  if (!is.null(shared) && !is.list(shared))
    shared <- list(shared)
  if (length(shared) && !requireNamespace("crosstalk", quietly = TRUE))
    stop("'shared' requires the crosstalk package.")
  dependencies <- list(rglDependency, CanvasMatrixDependency)
  if (length(shared) && isNamespaceLoaded("crosstalk")) {
    x$crosstalk <- list(key = vector("list", length(shared)),
    		        group = character(length(shared)),
    		        id = integer(length(shared)),
    		        options = vector("list", length(shared)))
    dependencies <- c(dependencies, crosstalk::crosstalkLibs())
  } else {
    x$crosstalk <- list(key = list(), 
    		        group = character(),
    		        id = integer(),
    		        options = list())
  }
  	
  for (i in seq_along(shared)) {
    s <- shared[[i]]
    if (crosstalk::is.SharedData(s) && inherits(s, "rglShared")) {
      x$crosstalk$key[[i]] <- s$key()
      x$crosstalk$group[i] <- s$groupName()
      x$crosstalk$id[i] <- attr(s$origData(), "rglId")
      x$crosstalk$options[[i]] <- attr(s$origData(), "rglOptions")
    } else if (!is.null(s))
      stop("'shared' must be an object produced by rglShared() or a list of these")
  }
  if (!is.null(width))
    width <- CSStoPixels(width)
  if (!is.null(height))
    height <- CSStoPixels(height)
  x <- convertScene(x, width, height,
                   elementId = elementId, 
                   webgl = webgl, snapshot = snapshot,
                   oldConvertBBox = oldConvertBBox)
  
  upstream <- processUpstream(controllers, elementId = elementId)
  
  if (webgl) {
    x$vertexShader <- paste(readLines(system.file("htmlwidgets/lib/rglClass/shaders/rgl_vertex.glsl", package = "rgl")), collapse = "\n")
    x$fragmentShader <- paste(readLines(system.file("htmlwidgets/lib/rglClass/shaders/rgl_fragment.glsl", package = "rgl")), collapse = "\n")
      
    x$players <- upstream$players

    x$webGLoptions <- webGLoptions
    
    x$fastTransparency <- fastTransparency

    if (inShiny())
      x$altText <- altText
    
    # create widget
    attr(x, "TOJSON_ARGS") <- list(na = "string")
    result <- structure(htmlwidgets::createWidget(
      name = 'rglWebGL',
      x = x,
      width = width,
      height = height,
      package = 'rgl',
      elementId = elementId,
      dependencies = dependencies,
      ...
    ), origScene = origScene)
    
    # We always emit aria-labelledby.  We need to 
    # choose here whether to write the label, or rely
    # on other code to write it.  We let other code write it
    # in new knitr and Shiny, and otherwise do it ourselves.
    
    if (!in_knitr_with_altText_support() && !inShiny())
      result <- htmlwidgets::prependContent(result, 
                  tags$p(altText, id = ariaLabelId(elementId),
                         hidden = NA))
  } else {
    if (is.list(upstream$objects)) {
      result <- img(src = image_uri(x), width = width, height = height)
    } else
      result <- x
  }
    
  if (in_positron())
    result$sizingPolicy$browser$fill <- TRUE
  
  if (is.list(upstream$objects)) {
  	if (requireNamespace("manipulateWidget", quietly = TRUE))
      result <- do.call(manipulateWidget::combineWidgets, c(upstream$objects, 
                              list(result, 
                                   rowsize = c(upstream$rowsizes, height), 
                                   ncol = 1)))
  	else 
  		warning("Combining widgets requires the 'manipulateWidget' package.", call. = FALSE)
  }
  result
}})

ariaLabelId <- function(id)
  paste0(id, "-aria")

widget_html.rglWebGL <- function(id, style, class, ...){
  result <- tags$div(id = id, style = style, class = class, role = "img",
           "aria-labelledby" = ariaLabelId(id))
  # In shiny, we need to write the alt text label.
  if (inShiny())
    result <- tags$div(tags$p("3D plot 1", id = ariaLabelId(id),
                                                 hidden = NA),
                       result)
  result
}

print.rglMouseSelection <- function(x, verbose = FALSE, ...) {
  if (!is.null(x$region)) {
    cat("Mouse selection:\n")
    if (verbose) {
      cat("p1=[", x$region[1], ",", x$region[2], 
        "] p2=[", x$region[3], ",", x$region[4],"]\n")
      cat("Projection data included: ", !is.null(x$model) && !is.null(x$proj) && !is.null(x$view), "\n")
    }
  } else
    cat("Inactive mouse selection.\n")
  invisible(x)
}

# Create the local dependencies

makeDependency <- function(name, src, script = NULL, package,
                           version = packageVersion(package),
                           minifile = paste0(basename(src), ".min.js"),
                           debugging = FALSE, ...) {
  javascript <- vapply(script, function(x) !is.list(x), FALSE)
  if (any(javascript) &&
      requireNamespace("js", quietly = TRUE) &&
      packageVersion("js") >= "1.2") {
    if (debugging) {
      for (f in script[javascript]) {
        hints <- js::jshint(readLines(file.path(system.file(src, package = package), f)),
                            undef = TRUE, bitwise = TRUE, eqeqeq = TRUE,
                            latedef = TRUE, browser = TRUE, devel = TRUE,
                            globals = list(CanvasMatrix4 = FALSE, 
                                           WebGLFloatArray = FALSE,
                                           rglwidgetClass = FALSE,
                                           rgltimerClass = FALSE,
                                           Shiny = FALSE
                                           ))
        for (i in seq_len(NROW(hints)))
          warning(f, "#", hints[i, "line"], ": ", hints[i, "reason"],
                  call. = FALSE, immediate. = TRUE)
      }
    }
    minified <- js::uglify_files(file.path(system.file(src, package = package), script[javascript]))
    writeLines(minified, file.path(system.file(src, package = package), minifile))
    if (!debugging)
      script <- c(minifile, script[!javascript])
  }
  htmlDependency(name = name, 
                      src = src,
                      package = package,
                      version = version,
                      script = script,
                      all_files = FALSE,
                      ...)
}

CanvasMatrixDependency <- makeDependency("CanvasMatrix4",
                                         src = "htmlwidgets/lib/CanvasMatrix",
                                         script = "CanvasMatrix.src.js",
                                         package = "rgl",
                                         debugging = isTRUE(as.logical(Sys.getenv("RGL_DEBUGGING", "FALSE"))))

local({
  shaders <- c('rglwidgetClass.rgl_vertex_shader = function() {',
               paste('return ',
                 paste0('"', readLines(system.file("htmlwidgets/lib/rglClass/shaders/rgl_vertex.glsl", package = "rgl")), '\\n"', collapse = "+\n"),
               ';};'),
               'rglwidgetClass.rgl_fragment_shader = function() {',
               paste('return ',
                 paste0('"', readLines(system.file("htmlwidgets/lib/rglClass/shaders/rgl_fragment.glsl", package = "rgl")), '\\n"', collapse = "+\n"),
               ';};'))
  writeLines(paste(shaders, collapse = "\n"), 
             file.path(system.file("htmlwidgets/lib/rglClass", package = "rgl"),
                       "shadersrc.src.js"))
})

rglDependency <- makeDependency("rglwidgetClass", 
                      src = "htmlwidgets/lib/rglClass",
                      script = c("rglClass.src.js",
                                 "utils.src.js",
                                 "buffer.src.js",
                                 "subscenes.src.js",
                                 "shaders.src.js",
                                 "shadersrc.src.js",
                                 "textures.src.js",
                                 "projection.src.js",
                                 "mouse.src.js",
                                 "init.src.js",
                                 "pieces.src.js",
                                 "draw.src.js",
                                 "controls.src.js",
                                 "selection.src.js",
                                 "rglTimer.src.js",
                                 "pretty.src.js",
                                 "axes.src.js",
                                 "animation.src.js"),
                      stylesheet = "rgl.css",
                      package = "rgl",
                      debugging = isTRUE(as.logical(Sys.getenv("RGL_DEBUGGING", "FALSE"))))

image_uri <- function(filename) {
  paste0("data:", mime::guess_type(filename),
         ";base64,", base64encode(filename))
}

in_positron <- function() nchar(Sys.getenv("POSITRON")) > 0

library(rgl)

## a function to plot good-looking spheres

sphere1.f <- function(x0 = 0, y0 = 0, z0 = 0, r = 1, n = 101, ...){
  f <- function(s,t){ 
    cbind(   r * cos(t)*cos(s) + x0,
             r *        sin(s) + y0,
             r * sin(t)*cos(s) + z0)
  }
  
  persp3d(f, slim = c(-pi/2,pi/2), tlim = c(0, 2*pi), n = n, add = T,  ...)
}

## a set of 3D coordinates for my spheres

agg <- as.data.frame(list(x = c(-0.308421860438279, -1.42503395393061), y = c(0.183223776337368, 1.69719822686475), z = c(-0.712687792799106, -0.0336746884947792)))

open3d()

##material and light effects for the spheres

clear3d(type = "lights")
light3d(theta = -30, phi = 60, viewpoint.rel = TRUE, ambient = "#FFFFFF", diffuse = "#FFFFFF", specular = "#FFFFFF", x = NULL, y = NULL, z = NULL)

## plot the spheres.
sphere <- sphere1.f(r=0.5)
sprites <- sprites3d(agg, radius = 0.5, shapes = sphere,
          rotating = TRUE)

rglwidget()

# This internal function returns a list with the following components:
# xlim, ylim, zlim:  the bounding box expanded so no coordinate has zero or negative extent
# strut:  a boolean indicating whether an expansion was done above
# x, y, z:  the box above expanded by a factor of expand

.getRanges <- function(expand = 1.03, ranges=par3d('bbox')) {
  ranges <- list(xlim=ranges[1:2], ylim=ranges[3:4], zlim=ranges[5:6])
  
  strut <- FALSE
  
  ranges <- lapply(ranges, function(r) {
    d <- diff(r)
    if (d > 0) return(r)
    strut <<- TRUE
    if (d < 0) return(c(0,1))
    else if (r[1] == 0) return(c(-1, 1))
    else return(r[1] + 0.4*abs(r[1])*c(-1,1))
  })
  
  ranges$strut <- strut
  
  ranges$x <- (ranges$xlim - mean(ranges$xlim))*expand + mean(ranges$xlim)
  ranges$y <- (ranges$ylim - mean(ranges$ylim))*expand + mean(ranges$ylim)
  ranges$z <- (ranges$zlim - mean(ranges$zlim))*expand + mean(ranges$zlim)
  
  ranges
}

axis3d <- function(edge, at = NULL, labels = TRUE, tick = TRUE, line = 0,
                   pos = NULL, nticks = 5, ...) {
  save <- par3d(skipRedraw = TRUE, ignoreExtent = TRUE)
  on.exit(par3d(save))
  
  ranges <- .getRanges()
  
  edge <- c(strsplit(edge, '')[[1]], '-', '-')[1:3]
  coord <- match(toupper(edge[1]), c('X', 'Y', 'Z')) 
  
  # Put the sign in the appropriate entry of edge
  if (coord == 2) edge[1] <- edge[2]
  else if (coord == 3) edge[1:2] <- edge[2:3]
  
  range <- ranges[[coord]]
  
  if (is.null(at)) {
    at <- pretty(range, nticks)
    at <- at[at >= range[1] & at <= range[2]]
  }
  if (length(at)) {
    if (is.logical(labels)) {
      if (labels) labels <- format(at)
      else labels <- NA
    }
    
    mpos <- matrix(NA,3,length(at))
    if (edge[1] == '+') mpos[1,] <- ranges$x[2]
    else mpos[1,] <- ranges$x[1]
    if (edge[2] == '+') mpos[2,] <- ranges$y[2]
    else mpos[2,] <- ranges$y[1]
    if (edge[3] == '+') mpos[3,] <- ranges$z[2]
    else mpos[3,] <- ranges$z[1]
    
    ticksize <- 0.05*(mpos[,1]-c(mean(ranges$x),mean(ranges$y),mean(ranges$z)))
    ticksize[coord] <- 0
    
    if (!is.null(pos)) mpos <- matrix(pos,3,length(at))
    mpos[coord,] <- at
    
    x <- c(mpos[1,1],mpos[1,length(at)])
    y <- c(mpos[2,1],mpos[2,length(at)])
    z <- c(mpos[3,1],mpos[3,length(at)])
    if (tick) {
      x <- c(x,as.double(rbind(mpos[1,],mpos[1,]+ticksize[1])))
      y <- c(y,as.double(rbind(mpos[2,],mpos[2,]+ticksize[2])))
      z <- c(z,as.double(rbind(mpos[3,],mpos[3,]+ticksize[3])))
    }
    result <- c(ticks=segments3d(x,y,z,...))
    
    if (!all(is.na(labels)))
      result <- c(result, labels=text3d(mpos[1,]+3*ticksize[1],
                                        mpos[2,]+3*ticksize[2],
                                        mpos[3,]+3*ticksize[3],
                                        labels, ...))
    lowlevel(result)
  }
}

axes3d <- function(edges='bbox', labels=TRUE,
                   tick=TRUE, nticks = 5, box = FALSE, expand = 1.03, ...) {
  save <- par3d(skipRedraw = TRUE, ignoreExtent = TRUE)
  on.exit(par3d(save))
  if (identical(edges, 'bbox')) {
    bboxargs <- names(formals(rgl.bbox))
    bboxargs <- intersect(bboxargs, names(list(...)))
    otherargs <- setdiff(names(list(...)), bboxargs)
    bboxargs <- list(...)[bboxargs]
    otherargs <- list(...)[otherargs]
    result <- do.call('bbox3d', c(list(draw_front=box, expand=expand), bboxargs, 
                                  do.call('.fixMaterialArgs', c(otherargs, list(Params = list(front='lines', back='lines'))))))
  } else {
    result <- numeric(0)
    for (e in edges)
      result <- c(result, axis3d(e,labels=labels,tick=tick,nticks=nticks, ...))
    names(result) <- e
  }	
  lowlevel(result)  
}

box3d <- function(...) {
  save <- par3d(ignoreExtent = TRUE)        
  on.exit(par3d(save))
  
  result <- numeric(0)
  
  ranges <- .getRanges()          
  if (ranges$strut) {
    par3d(ignoreExtent = FALSE)
    result <- c(result, strut=segments3d(rep(ranges$xlim, c(2,2)),
                                         rep(ranges$ylim, c(2,2)),
                                         rep(ranges$zlim, c(2,2))))
    par3d(ignoreExtent = TRUE)
  }
  
  x <- c(rep(ranges$x[1],8),rep(ranges$x,4),rep(ranges$x[2],8))
  y <- c(rep(ranges$y,2),rep(ranges$y,c(2,2)),rep(ranges$y,c(4,4)),
         rep(ranges$y,2),rep(ranges$y,c(2,2)))
  z <- c(rep(ranges$z,c(2,2)),rep(ranges$z,2),rep(rep(ranges$z,c(2,2)),2),
         rep(ranges$z,c(2,2)),rep(ranges$z,2))
  lowlevel(c(result, lines=segments3d(x,y,z,...)))
}

# Convert edge code into values:
# coord = 1:3 for coordinate
# edge = 3 of "+" or "-" for other coords (ignore the one
# that matches coord)
# floating = TRUE if edge should move as bboxdeco moves
# labels

parseMargin <- function(code, floating = FALSE) {
  stopifnot(length(code) == 1)
  if (!nchar(code)) 
    list(coord = 0, edge = c(0,0,0), floating = floating)
  else {
    default <- if (isTRUE(floating)) "+" else "-"
    edge <- c(strsplit(code, '')[[1]], default, default)[1:3]
    coord <- match(toupper(edge[1]), c('X', 'Y', 'Z')) 
    edge <- match(edge, c("-", "+"))*2 - 3 # +/- 1
    edge[is.na(edge)] <- 0
    # Put the sign in the appropriate entry of edge
    if (coord == 2) edge[1] <- edge[2]
    else if (coord == 3) edge[1:2] <- edge[2:3]
    list(coord = coord, edge = edge, floating = floating)
  }
}

deparseMargin <- function(margin) {
  if (margin$coord %in% 1:3) {
    coord <- margin$coord
    edge <- margin$edge[-coord]
    edge <- c("-", ".", "+")[edge + 2]
    paste(c("x", "y", "z")[coord], 
          edge[1],
          edge[2], sep = "")
  } else
    ""
}

mtext3d <- function(text, edge, at = NULL, line = 0,
                    level = 0, 
                    floating = FALSE, 
                    pos = NA, ...) {
  save <- par3d(ignoreExtent = TRUE)
  on.exit(par3d(save))
  
  bboxdeco <- ids3d("bboxdeco")
  if (!is.na(floating) && !nrow(bboxdeco)) {
    dummyBbox()
    bboxdeco <- ids3d("bboxdeco")
  }
    
  if (!nrow(bboxdeco) || (missing(floating) && !missing(pos)))
    floating <- NA
  
  ranges <- .getRanges()
  
  if (is.language(text))
    text <- as.expression(text)
  
  newlen <- max(length(text),length(line),length(at))
  text <- rep(text, length.out = newlen)
  line <- rep(line, length.out = newlen)
  
  if (is.na(floating)) {
    margin <- parseMargin(edge, floating)
    coord <- margin$coord
    if (!(coord %in% 1:3))
      stop("Bad edge spec.")
    range <- ranges[[coord]]
    if (is.null(at)) at <- mean(range)
    at <- rep(at, length.out = newlen)
    if (all(is.na(pos))) {
      edge <- margin$edge
      pos <- matrix(NA,3,length(at))
      if (edge[1] == +1) pos[1,] <- ranges$x[2]
      else pos[1,] <- ranges$x[1]
      if (edge[2] == +1) pos[2,] <- ranges$y[2]
      else pos[2,] <- ranges$y[1]
      if (edge[3] == +1) pos[3,] <- ranges$z[2]
      else pos[3,] <- ranges$z[1]
    }
    else pos <- matrix(pos,3,length(at))
    pos[coord,] <- at
    ticksize <- 0.05*(pos[,1]-c(mean(ranges$x),mean(ranges$y),mean(ranges$z)))
    ticksize[coord] <- 0
    
    text3d(pos[1,]+3*ticksize[1]*line,
           pos[2,]+3*ticksize[2]*line,
           pos[3,]+3*ticksize[3]*line,
           text, ...)
  } else {
    if (!missing(pos))
      stop("'pos' is only used with floating = NA")
    if (is.null(at))
      at <- NA
    text3d(x = cbind(at, line, level),
           texts = text, margin = edge,
           floating = floating, 
           ...)
  }
}   

title3d <- function(main = NULL, sub = NULL, xlab = NULL, ylab = NULL, 
                    zlab = NULL, line = NA, level = NA, floating = NULL, ...) {
  save <- par3d(skipRedraw = TRUE, ignoreExtent = TRUE)
  on.exit(par3d(save))
  
  if (length(floating)) {
    floatingtitles <- floating
    floatinglabels <- floating
  } else {
    floatingtitles <- FALSE
    floatinglabels <- TRUE
  }
  result <- numeric(0)
  if (!is.null(main)) {
    aline <- ifelse(is.na(line), 2, line)
    alevel <- ifelse(is.na(level), 2, level)
    result <- c(result, 
                main=mtext3d(main, 'x++', line = aline, level = alevel, floating = floatingtitles, ...))
  }
  if (!is.null(sub)) {
    aline <- ifelse(is.na(line), 2, line)
    alevel <- ifelse(is.na(level), 2, level)
    result <- c(result, sub=mtext3d(sub, 'x--', line = aline, floating = floatingtitles, ...))
  }
  if (!is.null(xlab)) {
    aline <- ifelse(is.na(line), 4, line)
    alevel <- ifelse(is.na(level), 1, level)
    result <- c(result, xlab=mtext3d(xlab, 'x', line = aline, level = alevel, floating = floatinglabels, ...))
  }
  if (!is.null(ylab)) {
    aline <- ifelse(is.na(line), 4, line)
    alevel <- ifelse(is.na(level), 1, level)
    result <- c(result, ylab=mtext3d(ylab, 'y', line = aline, level = alevel, floating = floatinglabels, ...))
  }
  if (!is.null(zlab)) {
    aline <- ifelse(is.na(line), 4, line)
    alevel <- ifelse(is.na(level), 1, level)
    result <- c(result, zlab=mtext3d(zlab, 'z', line = aline, level = alevel, floating = floatinglabels, ...))
  }                  
  lowlevel(result)
}

# Project the shadow of one mesh object onto another

shadow3d <- function(obj, mesh, plot = TRUE, 
										 up = c(0, 0, 1), P = projectDown(up),
										 outside = FALSE, ...) {
	triangles <- as.triangles3d(mesh)
	ntri <- nrow(triangles) / 3
	
	P <- t(P)   # The convention in rgl is row vectors on the left
	# but we'll be using column vectors on the right
	
	projected <- P %*% rbind(t(triangles), 1)
	projected <- projected[1:2,]/rep(projected[4,], each = 2)
	
	fn <- function(xyz) {
		result <- rep(-Inf, nrow(xyz))
		r <- P %*% rbind(t(xyz), 1)
		r <- rbind(r[1:2,]/rep(r[4,], each = 2), 1)
		for (i in 1:ntri) {
			# For each triangle, find the barycentric parameters
			# of each point in xyz using relation R lambda = r = (x,y,1)'
			R <- rbind(projected[,(0:2) + 3*i - 2], 1)
			lambda <- tryCatch(solve(R, r),
												 error = function(e) matrix(-Inf, 3, ncol(r)))
			# If the smallest barycentric coord is positive, the point
			# is in the triangle
			minlambda <- apply(lambda, 2, min)
			result <- pmax(result, minlambda)
		}
		result
	}
	
	if (outside)
		levels <- c(Inf, 0, -Inf)
	else
	  levels <- c(0, Inf)

	filledContour3d(obj, fn, levels = levels, 
								  plot = plot, ...)
}

projectDown <- function(up) {
	if (length(up) == 4)
		up <- up[1:3]/up[4]
	else if (length(up) != 3)
		stop("'up' vector should be length 3.")
	P <- GramSchmidt(up, c(1, 0, 0), c(0, 1, 0))
	if (det(P) < 0)
		P[3,] <- -P[3,]
	cbind(rbind(t(P[c(2,3,1),]), 0), c(0, 0, 0, 1))
}

facing3d <- function(obj, up = c(0, 0, 1),
										 P = projectDown(up), 
										 front = TRUE, strict = TRUE) {
	obj <- as.tmesh3d(obj)
	P <- t(P)   # The convention in rgl is row vectors on the left
	# but we'll be using column vectors on the right
	r <- P %*% obj$vb
	r <- r[1:2,]/rep(r[4,], each = 2)
	area <- function(i) {
		x <- r[1,obj$it[,i]]
		y <- r[2,obj$it[,i]]
		area <- 0
		for (j in 1:3)
			area <- area + x[j]*y[j %% 3 + 1] - x[j %% 3 + 1]*y[j]
		area
	}
	areas <- vapply(seq_len(ncol(obj$it)), area, 0)
	if (isTRUE(front))
		keep <- areas > 0
	else
		keep <- areas < 0
	if (!strict)
		keep <- keep | areas == 0
	obj$it <- obj$it[, keep]
	cleanMesh3d(obj)
}

writeASY <- function(scene = scene3d(),
                     title = "scene",
                     outtype = c("pdf", "eps", "asy", "latex", "pdflatex"),
                     prc = TRUE,
                     runAsy = "asy %filename%",
                     defaultFontsize = 12,
                     width = 7,
                     height = 7,
                     ppi = 100,
                     ids = tagged3d(tags),
                     tags = NULL,
                     version = "2.65") {
  withColors <- TRUE
  withNormals <- FALSE

  outtype <- match.arg(outtype)
  version <- numeric_version(version)
  
  writeHeader <- function() {
    outformat <- c(pdf = "pdf", eps = "eps", asy = "",
    	           latex = "eps", pdflatex = "pdf")[outtype]
    prc <- if (prc) "true" else "false"
    userMatrix <- get.par3d("userMatrix")
    defaultObserver <- (c(get.par3d("observer"), 1) %*% userMatrix)[1:3]
    up <- (c(0, 1, 0, 1) %*% userMatrix)[1:3]
    FOV <- get.par3d("FOV")*pi/180
    if (FOV > 0) {
      projection <- "perspective"
      dist <- 0.8/tan(FOV/2)
      defaultObserver <- defaultObserver*dist/sqrt(sum(defaultObserver^2))
    } else 
      projection <- "orthographic"
    result <<- c(result, subst(
'// %title% produced by rgl', title),
      if (outtype %in% c("pdf", "eps")) subst(
'settings.outformat = "%outformat%";', outformat),
      subst(
'settings.prc = %prc%;
size(%width%inches, %height%inches);
import graph3;
currentprojection = %projection%(%x%, %y%, %z%, up = (%ux%, %uy%, %uz%));
defaultpen(fontsize(%defaultFontsize%));
ticklabel RGLstrings(real[] at, string[] label)
{
  return new string(real x) {
    int i = search(at, x);
    if (i < 0) return "";
    else return label[i];
  };
}

ticklabel RGLScale(real s)
{
  return new string(real x) {return format(s*x);};
}', 
  prc, width, height,
  x=defaultObserver[1], y=defaultObserver[2], z=defaultObserver[3],
  ux = up[1], uy = up[2], uz = up[3],
  projection,
  defaultFontsize))
  }
  
  # simulate rgl.attrib
  get.attrib <- function(id, attrib) {
    obj <- scene$objects[[as.character(id)]]
    obj[[attrib]]
  }
  
  # simulate ids3d
  get.ids <- function(type = "shapes") {
    ids <- names(scene$objects)
    types <- vapply(ids, function(x) scene$objects[[x]]$type, "")
    if (length(s <- which(type %in% "shapes"))) {
      type <- c(type[-s], "points", "linestrip", "lines",
                "text", "triangles", "quads", "surface",
                "spheres", "planes", "abclines",
                "clipplanes", "sprites")
    }
    keep <- types %in% type
    data.frame(id = as.numeric(ids[keep]), type = types[keep])
  }
  
  getmaterial <- function(id) {
    result <- scene$material
    this <- scene$objects[[as.character(id)]]$material
    result[names(this)] <- this
    result
  }
  
  getScaling <- function() {
    scale <- get.par3d("scale")
    scale/avgScale()  
  }
  
  getCentre <- function() {
    bbox <- get.par3d("bbox")
    c(mean(bbox[1:2]), mean(bbox[3:4]), mean(bbox[5:6]))
  }
  
  getVertices <- function(id) {
    scale <- getScaling()
    vertices <- scale(get.attrib(id, "vertices"), center=FALSE, scale=1/scale)
    if (withColors) {
      colors <- get.attrib(id, "colors")
      if (nrow(colors) == 1)
        colors <- colors[rep(1, nrow(vertices)),,drop = FALSE]
    }
    if (withNormals) {
      normals <- get.attrib(id, "normals")
      if (!nrow(normals))
      	normals <- 0*vertices
    }
    cbind(vertices, 
          if (withColors) colors,
          if (withNormals) normals)
  }
  
  get.par3d <- function(attr = NULL) {
    par3d <- scene$rootSubscene$par3d
    if (!is.null(attr))
      par3d <- par3d[[attr]]
    par3d
  }
  rgba <- c("r", "g", "b", "a")
  lastCol <- c(0,0,0,1)  
  lastSize <- 0.5
  setPen <- function(col = lastCol, size = lastSize) {
    if (any(col[1:3] != lastCol[1:3])) {
      result <<- c(result, subst(
  	'currentpen = colorless(currentpen) + rgb(%r%, %g%, %b%);', 
  	r = col[1], g = col[2], b = col[3]))
      lastCol[1:3] <<- col[1:3]
    }
    if (col[4] != lastCol[4]) {
      result <<- c(result, subst(
    	  'currentpen += opacity(%a%);', a = col[4]))
      lastCol[4] <<- col[4]
    }
    if (size != lastSize) {
      result <<- c(result, subst(
    	'currentpen += linewidth(%size%);', size))
      lastSize <<- size
    }
  }
  	
  writePoly <- function(vertices) {
    if (any(!is.finite(vertices)))
      return()
    setPen(apply(vertices[, rgba], 2, mean))
    v <- vertices[1, 1:3]
    result <<- c(result, subst('draw(surface((%x%, %y%, %z%)', x=v[1], y=v[2], z=v[3]))
    for (j in seq_len(nrow(vertices))[-1]) {
      v <- vertices[j, 1:3]
      result <<- c(result, subst('--(%x%, %y%, %z%)', x=v[1], y=v[2], z=v[3]))
    }
    result <<- c(result, '--cycle), light=currentlight);')
  }
  
  writeTriangles <- function(id) {
    vertices <- getVertices(id)
    n <- nrow(vertices) %/% 3
    for (i in seq_len(n)) 
      writePoly(vertices[3*i-2:0,])
  }
  
  writeQuads <- function(id) {
    vertices <- getVertices(id)
    n <- nrow(vertices) %/% 4
    for (i in seq_len(n)) 
      writePoly(vertices[4*i-3:0,])
  }
  
  writeSurface <- function(id) {
    vertices <- getVertices(id)
    dims <- get.attrib(id, "dim")
    nx <- dims[1]
    nz <- dims[2]
    for (i in seq_len(nz)[-nz]) 
      for (j in seq_len(nx)[-nx]) 
      	writePoly(vertices[c((i-1)*nx + j, i*nx + j, i*nx + j + 1, (i-1)*nx + j + 1),])
  }
  
  writeSpheres <- function(id) {
    vertices <- getVertices(id)
    n <- nrow(vertices)    
    radii <- get.attrib(id, "radii")/4
    radii <- rep(radii, length.out=n)
    for (i in seq_len(n)) {
      setPen(vertices[i, rgba])
      v <- vertices[i, 1:3]
      result <<- c(result, subst('draw(shift((%x%, %y%, %z%))*scale3(%r%)*unitsphere);',
      	x = v[1], y = v[2], z = v[3], r = radii[i]))  
    }
  }  
  
  avgScale <- function() {
    bbox <- get.par3d("bbox")
    ranges <- c(bbox[2]-bbox[1], bbox[4]-bbox[3], bbox[6]-bbox[5])
    if (prod(ranges) == 0) 1
    else exp(mean(log(ranges)))
  }  
   
  writePoints <- function(id) {
    setPen(size = getmaterial(id)$size)
    vertices <- getVertices(id)
    n <- nrow(vertices)
    for (i in seq_len(n)) {
      setPen(vertices[i, rgba])
      result <<- c(result, subst('draw((%x%, %y%, %z%));', 
        x = vertices[i, 1], y = vertices[i, 2], z = vertices[i, 3]))
    }
  }
  
  writeText <- function(id) {
    vertices <- getVertices(id)
    n <- nrow(vertices)
    texts <- get.attrib(id, "texts")
    texts <- rep(texts, length.out = n)
    adj <- get.attrib(id, "adj")
    adj <- adj[rep(seq_len(nrow(adj)), length.out = n),, drop = FALSE]
    for (i in seq_len(n)) {
      setPen(vertices[i, rgba])	
      if (all(!is.na(vertices[i, 1:3]))) 
      	result <<- c(result, subst('label("%text%", position = (%x%, %y%, %z%), align = (%ax%,%ay%));', 
          x = vertices[i, 1], y = vertices[i, 2], z = vertices[i, 3],
          text = texts[i], ax = 1-2*adj[i, 1], ay = 1-2*adj[i, 2]))
    }
  }
  
  writeSegments <- function(id) {
    setPen(size = getmaterial(id)$lwd)
    vertices <- getVertices(id)
    n <- nrow(vertices) %/% 2    
    for (i in seq_len(n)) {
      i1 <- 2*i - 1
      i2 <- i1 + 1
      if (all(!is.na(vertices[c(i1, i2), 1:3]))) {
      	setPen((vertices[i1, rgba] + vertices[i2, rgba])/2)
  	result <<- c(result, subst('draw((%x1%, %y1%, %z1%)--(%x2%, %y2%, %z2%));', 
          x1 = vertices[i1, 1], y1 = vertices[i1, 2], z1 = vertices[i1, 3],
          x2 = vertices[i2, 1], y2 = vertices[i2, 2], z2 = vertices[i2, 3]))
      }
    }
  }
  
  writeLines <- function(id) {
    setPen(size = getmaterial(id)$lwd)          
    vertices <- getVertices(id)
    n <- nrow(vertices)    
    open <- FALSE
    for (i in seq_len(n)) {
      if (open) {
      	if (any(is.na(vertices[i, 1:3]))) {
      	  result <<- c(result, ");")
      	  open <- FALSE
      	} else 
      	  result <<- c(result, subst('--(%x%, %y%, %z%)',
      	    x = vertices[i, 1], y = vertices[i, 2], z = vertices[i, 3]))
      } else 
      	if (all(!is.na(vertices[i, 1:3]))) {
      	  setPen(vertices[i, rgba])
      	  result <<- c(result, subst('draw((%x%, %y%, %z%)',
      	    x = vertices[i, 1], y = vertices[i, 2], z = vertices[i, 3]))
      	  open <- TRUE
      	}
    }
    if (open)
      result <<- c(result, ');')
  }
  
  writeBackground <- function(id) {
    col <- get.attrib(id, "colors")
    result <<- c(result, subst(
      'currentlight.background = rgb(%r%, %g%, %b%);',
      r = col[1], g=col[2], b=col[3]
    ))
  }
  
  writeBBox <- function(id) {
    setPen(size = getmaterial(id)$lwd)
    bbox <- get.par3d("bbox")
    scale <- getScaling()
    vertices <- getVertices(id)
    ticks <- vertices[,1]
    ticks <- ticks[!is.na(ticks)]
    if (length(ticks))
    	xticks <- subst('Ticks3(1, Ticks = new real[] {%ticks%},
    			          ticklabel = RGLScale(%scale%))',
    			scale = 1/scale[1], ticks = paste(ticks, collapse=","))
    else
    	xticks <- subst('Ticks3(1, ticklabel = RGLScale(%scale%))',
    			scale = 1/scale[1])
    ticks <- vertices[,2]
    ticks <- ticks[!is.na(ticks)]
    if (length(ticks))
    	yticks <- subst('Ticks3(1, Ticks = new real[] {%ticks%},
    			ticklabel = RGLScale(%scale%))',
    			scale = 1/scale[2], ticks = paste(ticks, collapse=","))
    else
    	yticks <- subst('Ticks3(1, ticklabel = RGLScale(%scale%))',
    			scale = 1/scale[2])
    ticks <- vertices[,3]
    ticks <- ticks[!is.na(ticks)]
    if (length(ticks))
    	zticks <- subst('Ticks3(1, Ticks = new real[] {%ticks%},
    			ticklabel = RGLScale(%scale%))',
    			scale = 1/scale[3], ticks = paste(ticks, collapse=","))
    else
    	xticks <- subst('Ticks3(1, ticklabel = RGLScale(%scale%))',
    			scale = 1/scale[3])
    bbox <- bbox * rep(scale, each = 2)
    result <<- c(result, subst(
      'xaxis3(axis=YZEquals(y=%ymin%, z=%zmin%),
              xmin=%xmin%, xmax=%xmax%,
              ticks = %xticks%);
       xaxis3(axis=YZEquals(y=%ymin%, z=%zmax%),
              xmin=%xmin%, xmax=%xmax%);
       xaxis3(axis=YZEquals(y=%ymax%, z=%zmin%),
              xmin=%xmin%, xmax=%xmax%);
       xaxis3(axis=YZEquals(y=%ymax%, z=%zmax%),
              xmin=%xmin%, xmax=%xmax%);
       yaxis3(axis=XZEquals(x=%xmin%, z=%zmin%),
              ymin=%ymin%, ymax=%ymax%,
              ticks = %yticks%);
       yaxis3(axis=XZEquals(x=%xmin%, z=%zmax%),
              ymin=%ymin%, ymax=%ymax%);
       yaxis3(axis=XZEquals(x=%xmax%, z=%zmin%),
              ymin=%ymin%, ymax=%ymax%);
       yaxis3(axis=XZEquals(x=%xmax%, z=%zmax%),
              ymin=%ymin%, ymax=%ymax%);
       zaxis3(axis=XYEquals(x=%xmin%, y=%ymin%),
              zmin=%zmin%, zmax=%zmax%,
              ticks = %zticks%);
       zaxis3(axis=XYEquals(x=%xmin%, y=%ymax%),
              zmin=%zmin%, zmax=%zmax%);
       zaxis3(axis=XYEquals(x=%xmax%, y=%ymin%),
              zmin=%zmin%, zmax=%zmax%);
       zaxis3(axis=XYEquals(x=%xmax%, y=%ymax%),
              zmin=%zmin%, zmax=%zmax%);',
              xmin = bbox[1], xmax = bbox[2], ymin=bbox[3], ymax=bbox[4],
              zmin=bbox[5], zmax=bbox[6], 
              xticks, yticks, zticks))
  }
  
  writeLights <- function(ids) {
    if (!length(ids))
      result <<- c(result, 'currentlight = nolight;')
    else {
      col <- array(NA, c(length(ids), 3, 3))
      pos <- matrix(NA, nrow=length(ids), ncol=3)
      for (i in seq_along(ids)) {
      	col[i,,] <- get.attrib(ids[i], "colors")[1:3, 1:3]
      	pos[i,] <- get.attrib(ids[i], "vertices")[1,]
      }
      result <<- c(result, 'currentlight = light(')
      if (version < "2.50") {
        cols <-
          paste(paste0("rgb(", col[, 1, 1], ",", col[, 1, 2], ",", col[, 1, 3], ")"),
                collapse = ",")
        result <<-
          c(result,
            subst(
              'ambient=new pen[] {%cols%},',
              cols
            ))
      }
      cols <-
        paste(paste0("rgb(", col[, 2, 1], ",", col[, 2, 2], ",", col[, 2, 3], ")"),
              collapse = ",")
      result <<-
        c(result, subst('diffuse = new pen[] {%cols%},', cols))
      cols <-
        paste(paste0("rgb(", col[, 3, 1], ",", col[, 3, 2], ",", col[, 3, 3], ")"),
              collapse = ",")
      result <<-
        c(result, subst('specular = new pen[] {%cols%},', cols))
      pos <-
        paste(paste0("(", pos[, 1], ",", pos[, 2], ",", pos[, 3], ")"), collapse = ",")
      result <<-
        c(result, subst('position = new triple[] {%pos%}', pos))
      if (version < "2.47")
        result <<-
        c(
          result,
          subst(
            ', viewport = %viewpoint%',
            pos,
            viewpoint = if (get.attrib(ids[1], "viewpoint"))
              "true"
            else
              "false"
          )
        )
      result <<- c(result, ');')
    }
  }    
  knowntypes <- c("points", "linestrip", "lines",
  		  "text", "triangles", "quads", "surface",
  		  "spheres", "planes", "abclines",
  		  "background", "bboxdeco", "light")
  
  #  Execution starts here!
  
  allids <- get.ids(c("shapes", "background", "bboxdeco", "light"))
  if (is.null(ids)) {
    ids <- allids
    types <- as.character(ids$type)
    ids <- ids$id
  } else {
    ind <- match(ids, allids$id)
    keep <- !is.na(ind)
    if (any(!keep)) warning(gettextf("Object(s) with id %s not found", paste(ids[!keep], collapse=" ")), 
    			    domain = NA)
    ids <- ids[keep]
    types <- allids$type[ind[keep]]
  }
    
  unknowntypes <- setdiff(types, knowntypes)
  if (length(unknowntypes))
    warning(gettextf("Object type(s) %s not handled", 
      paste("'", unknowntypes, "'", sep="", collapse=", ")), domain = NA)

  keep <- types %in% knowntypes
  ids <- ids[keep]
  types <- types[keep]
  
  result <- NULL
  writeHeader()
  
  # Lights are done first.
  writeLights(ids[types == "light"])

  for (i in seq_along(ids)) {
    result <<- c(result, subst('// %type% object %id%', type = types[i], id = ids[i]))
    switch(types[i],
      planes =,
      triangles = writeTriangles(ids[i]),
      quads = writeQuads(ids[i]),
      surface = writeSurface(ids[i]),
      spheres = writeSpheres(ids[i]),
      points = writePoints(ids[i]),
      abclines =,
      lines = writeSegments(ids[i]),
      linestrip = writeLines(ids[i]),
      text = writeText(ids[i]),
      background = writeBackground(ids[i]),
      bboxdeco = writeBBox(ids[i]),
      light = {} # nolint
    )
  }
  if (outtype %in% c("latex", "pdflatex")) {
    filename <- paste0(title, ".tex")
    result <- c("\\begin{asy}", result, "\\end{asy}")
  } else
    filename <- paste0(title, ".asy")
  base::writeLines(result, filename) 
  if (outtype %in% c("pdf", "eps")) {
    system(subst(runAsy, filename))
    filename <- paste0(title, ".", outtype)
  }
  invisible(filename)
}

rgl.setMouseCallbacks <- function(button, begin=NULL, update=NULL, end=NULL, dev = cur3d(), subscene = currentSubscene3d(dev)) {
    invisible(.Call(rgl_setMouseCallbacks, as.integer(button), begin, update, end, 
                    as.integer(dev), as.integer(subscene)))
}

rgl.getMouseCallbacks <- function(button, dev = cur3d(), subscene = currentSubscene3d(dev)) 
    .Call(rgl_getMouseCallbacks, as.integer(button), as.integer(dev), as.integer(subscene))

rgl.setWheelCallback <- function(rotate=NULL, dev = cur3d(), subscene = currentSubscene3d(dev)) {
    invisible(.Call(rgl_setWheelCallback, rotate, as.integer(dev), as.integer(subscene)))
}

rgl.getWheelCallback <- function(dev = cur3d(), subscene = currentSubscene3d(dev)) 
    .Call(rgl_getWheelCallback, as.integer(dev), as.integer(subscene))

rgl.setAxisCallback <- function(axis, draw = NULL, dev = cur3d(), subscene = currentSubscene3d(dev)) {
  
  .Defunct("setAxisCallbacks")
  
}

rgl.getAxisCallback <- function(axis, dev = cur3d(), subscene = currentSubscene3d(dev)) {
  .Call(rgl_getAxisCallback, as.integer(dev), as.integer(subscene), as.integer(axis - 1))
}

# Configuration for Windows only

lines <- readLines("src/Makevars.win.in")
lines <- sub("@HIDE_IF_R42PLUS@", if (getRversion() < '4.2.0') "" else "#", lines, fixed = TRUE)
writeLines(lines, "src/Makevars.win")

cone3d <- function(base=c(0,0,0),tip=c(0,0,1),rad=1,n=30,draw.base=TRUE,qmesh=FALSE,
                   trans = par3d("userMatrix"), ...) {
  ax <- tip-base
  if (missing(trans) && !cur3d()) trans <- diag(4)
  ### is there a better way?
  if (ax[1]!=0) {
    p1 <- c(-ax[2]/ax[1],1,0)
    p1 <- p1/sqrt(sum(p1^2))
    if (p1[1]!=0) {
      p2 <- c(-p1[2]/p1[1],1,0)
      p2[3] <- -sum(p2*ax)
      p2 <- p2/sqrt(sum(p2^2))
    } else {
      p2 <- c(0,0,1)
    }
  } else if (ax[2]!=0) {
    p1 <- c(0,-ax[3]/ax[2],1)
    p1 <- p1/sqrt(sum(p1^2))
    if (p1[1]!=0) {
      p2 <- c(0,-p1[3]/p1[2],1)
      p2[3] <- -sum(p2*ax)
      p2 <- p2/sqrt(sum(p2^2))
    } else {
      p2 <- c(1,0,0)
    }
  } else {
    p1 <- c(0,1,0); p2 <- c(1,0,0)
  }
  degvec <- seq(0,2*pi,length=n+1)[-1]
  ecoord2 <- function(theta) {
    base+rad*(cos(theta)*p1+sin(theta)*p2)
  }
  i <- rbind(1:n,c(2:n,1),rep(n+1,n))
  v <- cbind(sapply(degvec,ecoord2),tip)
  if (qmesh) 
    ## minor kluge for quads -- draw tip twice
    i <- rbind(i,rep(n+1,n))
  if (draw.base) {
    v <- cbind(v,base)
    i.x <- rbind(c(2:n,1),1:n,rep(n+2,n))
    if (qmesh)  ## add base twice
      i.x <-  rbind(i.x,rep(n+2,n))
    i <- cbind(i,i.x)
  }
  if (qmesh) v <- rbind(v,rep(1,ncol(v))) ## homogeneous
  if (!qmesh)
    triangles3d(v[1,i],v[2,i],v[3,i],...)
  else
    return(rotate3d(qmesh3d(v,i,material=list(...)), matrix=trans))
}     


ellipsoid3d <- function(rx=1,ry=1,rz=1,n=30,ctr=c(0,0,0),
                        qmesh=FALSE,
                        trans = par3d("userMatrix"),...) {
  if (missing(trans) && !cur3d()) trans <- diag(4)
  degvec <- seq(0,pi,length=n)
  ecoord2 <- function(p)
    c(rx*cos(p[1])*sin(p[2]),ry*sin(p[1])*sin(p[2]),rz*cos(p[2]))
  v <- apply(expand.grid(2*degvec,degvec),1,ecoord2)
  if (qmesh) v <- rbind(v,rep(1,ncol(v))) ## homogeneous
  e <- expand.grid(1:(n-1),1:n)
  i1 <- apply(e,1,function(z)z[1]+n*(z[2]-1))
  i2 <- i1+1
  i3 <- (i1+n-1) %% n^2 + 1
  i4 <- (i2+n-1) %% n^2 + 1
  i <- rbind(i1,i2,i4,i3)
  if (!qmesh)
    quads3d(v[1,i],v[2,i],v[3,i],...)
  else return(rotate3d(qmesh3d(v,i,material=list(...)),matrix=trans))
}

############
open3d()
ellipsoid3d(ctr=c(2,2,2),rx=3,ry=2,col="red",alpha=0.4)
cone3d(base=c(-2,-2,-2),rad=0.5,tip=c(-3,0,-4),col="blue",front="lines",back="lines")
shade3d(translate3d(cube3d(),3,-2,3,col="purple"))

### now with qmesh()
open3d()
q1 <- cone3d(qmesh=TRUE,trans=diag(4))  ## the "unit cone";
                                        ## height=1,radius=1, base at (0,0,0)
shade3d(q1)
## various transformations and rotations
wire3d(translate3d(q1,3,0,0),col="green")
wire3d(translate3d(scale3d(q1,1,1,2),6,0,0),col="green")
dot3d(translate3d(q1,0,3,0),col="green")
dot3d(translate3d(scale3d(q1,2,1,1),0,6,0),col="green")
shade3d(translate3d(q1,0,0,3),col="red")
shade3d(translate3d(rotate3d(scale3d(q1,1,1,2),pi/4,0,1,0),0,0,6),col="red")

axes3d()
open3d()

s1 <- ellipsoid3d(qmesh=TRUE,trans=diag(4))  ## the "unit sphere";
                                        ## radius=1, ctr at (0,0,0)
shade3d(s1)
## various transformations and rotations
wire3d(translate3d(s1,3,0,0),col="green")
wire3d(translate3d(scale3d(s1,1,1,2),6,0,0),col="green")
dot3d(translate3d(s1,0,3,0),col="green")
dot3d(translate3d(scale3d(s1,2,1,1),0,6,0),col="green")
shade3d(translate3d(s1,0,0,3),col="red")
shade3d(translate3d(rotate3d(scale3d(s1,1,1,2),pi/4,0,1,0),0,0,6),col="red")

axes3d()

hover3d <- function(x, y = NULL, z = NULL, 
                    labeller = NULL,
                    tolerance = 20, 
                    persist = c("no", "one", "yes"),
                    labels = seq_along(x),
                    adj = c(-0.2, 0.5),
                    scene = scene3d(minimal = FALSE),
                    applyToScene = TRUE,
                    ...) {
  
	save <- par3d(skipRedraw = TRUE)
	on.exit(par3d(save))
	
  labelIndex <- function(sel, ...)
    text3d(x[sel], y[sel], z[sel], texts=labels[sel], adj=adj,
           ...)

  custom_labeller <- FALSE
  if (is.null(labeller)) 
    labeller <- labelIndex
  else
    custom_labeller <- TRUE
  
  stopifnot(is.function(labeller))
  
  persist <- match.arg(persist)
  
  opar <- par3d("mouseMode")
  odev <- cur3d()
  
  if (inherits(x, "rglId") ||
      (length(x) == 1 && is.null(y) && is.null(z))) {
    if (length(x) != 1)
      stop("Exactly one object may be specified as x.")
    idverts <- x
    x <- expandVertices(idverts)
  } else
    idverts <- numeric()
  
  xyz <- xyz.coords(x, y, z)
  x <- xyz$x
  y <- xyz$y
  z <- xyz$z
  if (length(x)==0) 
    stop("No points to identify.")
  
  force(labels)
  force(adj)
  
  selected <- list()
  
  hoverSelect <- function(mousex, mousey) {
    disp <- cur3d()
    save2 <- par3d(skipRedraw = TRUE)
    on.exit({par3d(save2); set3d(disp)})
    
    set3d(odev)

    viewport <- par3d("viewport")
    winxyz <- rgl.user2window(xyz)
    winxyz[,1] <- winxyz[,1]*viewport[3]
    winxyz[,2] <- (1-winxyz[,2])*viewport[4]

    change <- FALSE
    
    dist <- sqrt( (mousex-winxyz[,1])^2 + (mousey - winxyz[,2])^2 )
    dist[winxyz[,3] < 0 | winxyz[,3] > 1] <- Inf
    sel <- which.min(dist)
    if (dist[sel] <= tolerance) {
      if (persist != "yes")
        selected <<- Filter(function(s) 
                             if (s$sel != sel) {
                               pop3d(id = s$ids)
                             	 change <- TRUE
                               FALSE
                             } else TRUE, selected)
      prev <- Find(function(s) 
                     s$sel == sel, selected)
      if (is.null(prev)) {
        this <- list(sel = sel, ids = labeller(sel, ...))
        change <- TRUE
        selected <<- c(selected, list(this))
      }
    } else if (persist == "no" && length(selected)) {
      lapply(selected, function(s) pop3d(id = s$ids))
    	change <- TRUE
      selected <<- list()
    }
    if (change) {
    	par3d(skipRedraw = FALSE)
    }
  }
  
  js <- NULL
  if (applyToScene) {
    # Define the Javascript functions with the same names to use in WebGL
    js <-
    'window.hoverSelect = function(x, y) { 
       var   obj = this.getObj(%idverts%), i, newdist, 
             dist = Infinity, pt, newclosest, text,
             customlabeller = %customlabeller%[0],
             idtexts = %idtexts%, change = false,
             selected = window.hoverSelect.selected;
       for (i = 0; i < obj.vertices.length; i++) {
         pt = obj.vertices[i].concat(1);
         pt = this.user2window(pt, %subid%);
         pt[0] = x - pt[0]*this.canvas.width;
         pt[1] = y - pt[1]*this.canvas.height;
         pt[2] = 0;
         newdist = rglwidgetClass.vlen(pt);
         if (newdist < dist) {
           dist = newdist;
           newclosest = i;
         }
       }
       if (dist <= %tolerance%) {
         if ("%persist%" !== "yes" && 
             selected.length > 0 &&
             selected[0] !== newclosest) {
               if (customlabeller) {
                 idtexts[selected[0]].forEach(id => this.delFromSubscene(id, %subid%));
               } else {
                 text = this.getObj(idtexts[0]);
                 text.colors[selected[0]][3] = 0; // invisible!
               }
               selected = [];
               change = true;
             }
         if (!selected.includes(newclosest)) {
           selected.push(newclosest);
           if (customlabeller) {
             idtexts[newclosest].forEach(id => this.addToSubscene(id, %subid%));
           } else {
             text = this.getObj(idtexts[0]);
             text.colors[newclosest][3] = 1; // alpha is here!
           }
           change = true;
         }
       } else if ("%persist%" === "no" && selected.length > 0) {
         if (customlabeller) {
           idtexts[selected[0]].forEach(id => this.delFromSubscene(id, %subid%));
         } else {
           text = this.getObj(%idtexts%);
           text.colors[selected[0]][3] = 0;
         }
         selected = [];
         change = true;
       }
       if (change) {
         if (!customlabeller)
           text.initialized = false;
         window.hoverSelect.selected = selected;
         this.drawScene();
       }
     };
     window.hoverSelect.selected = [];
    '
    if (!length(idverts)) {
      idverts <- points3d(xyz)
      delFromSubscene3d(idverts)
    }
    
    if (custom_labeller) {
      idtexts <- vector("list", length(x))
      for (i in seq_along(x)) {
        idtexts[[i]] <- labeller(i, ...)
      }
      delFromSubscene3d(unlist(idtexts))
    } else {
      idtexts <- text3d(xyz, texts = labels, adj = adj, 
                      alpha = rep(0, length(labels)), ...)
      delFromSubscene3d(idtexts)
    }
    js <- gsub("%idverts%", idverts, js)  
    js <- gsub("%subid%", subsceneInfo()$id, js)
    js <- gsub("%idtexts%", toJSON(idtexts), js)
    js <- gsub("%tolerance%", tolerance, js)
    js <- gsub("%persist%", persist, js)
    js <- gsub("%customlabeller%", toJSON(custom_labeller), js)
  } else
    idtexts <- numeric()
  
  setUserCallbacks(0, update="hoverSelect",
                   scene = scene,
                   applyToScene = applyToScene,
                   applyToDev = TRUE,
                   javascript = js)
  
  structure(lowlevel(c(idverts, unlist(idtexts))),                         oldPar = opar, oldDev = odev)
}

# rgl demo: rgl-bivar.r
# author: Daniel Adler

rgl.demo.bivar <- function() {
  if (!requireNamespace("MASS"))
    stop("This demo requires MASS")
  
  # parameters:
  n<-50; ngrid<-40
  
  # generate samples:
  set.seed(31415)
  x<-rnorm(n); y<-rnorm(n)
  
  # estimate non-parameteric density surface via kernel smoothing
  denobj <- MASS::kde2d(x, y, n=ngrid)
  den.z <-denobj$z
  
  # generate parametric density surface of a bivariate normal distribution
  xgrid <- denobj$x
  ygrid <- denobj$y
  bi.z <- dnorm(xgrid)%*%t(dnorm(ygrid))
  
  # visualize:
  zscale<-20
  
  # New window
  open3d()
  
  # clear scene:
  clear3d("all")
  
  # setup env:
  bg3d(color="#887777")
  light3d()
  
  # Draws the simulated data as spheres on the baseline
  spheres3d(x,y,rep(0,n),radius=0.1,color="#CCCCFF")
  
  # Draws non-parametric density
  surface3d(xgrid,ygrid,den.z*zscale,color="#FF2222",alpha=0.5)
  
  # Draws parametric density
  surface3d(xgrid,ygrid,bi.z*zscale,color="#CCCCFF",front="lines") 
}

rgl.demo.bivar()

# Widget output function for use in Shiny

playwidgetOutput <- function(outputId, width = '0px', height = '0px') {
	registerShinyHandlers()
  shinyWidgetOutput(outputId, 'rglPlayer', width, height, package = 'rgl')
}

# Widget render function for use in Shiny

renderPlaywidget <- function(expr, env = parent.frame(), quoted = FALSE, outputArgs = list()) {
	registerShinyHandlers()
  if (!quoted) expr <- substitute(expr)  # force quoted
  shiny::markRenderFunction(playwidgetOutput,
                     shinyRenderWidget(expr, playwidgetOutput, env, quoted = TRUE),
  		     outputArgs = outputArgs)
}

playwidget <- function(sceneId, controls, start = 0, stop = Inf, interval = 0.05,  rate = 1,
                       components = c("Reverse", "Play", "Slower", "Faster", "Reset", "Slider", "Label"),
                       loop = TRUE,
                       step = 1, labels = NULL,
                       precision = 3,
                       elementId = NULL, respondTo = NULL,
                       reinit = NULL,
		       buttonLabels = components,
		       pause = "Pause",
		       height = 40,
                       ...) {

  if (is.null(elementId) && !inShiny())
    elementId <- newElementId("rgl-play")

  # sceneId = NA turns into prevRglWidget = NULL      
  upstream <- processUpstream(sceneId, playerId = elementId)
  
  if (!is.null(respondTo))
    components <- buttonLabels <- NULL

  if (length(stop) != 1 || !is.finite(stop)) stop <- NULL

  if (identical(controls, NA)) 
  	stop(dQuote("controls"), " should not be NA.")
  	
  stopifnot(is.list(controls))

  if (inherits(controls, "rglControl"))
    controls <- list(controls)

  okay <- vapply(controls, inherits, TRUE, "rglControl")
  if (any(bad <- !okay)) {
    bad <- which(bad)[1]
    stop("Controls should inherit from 'rglControl', control ", bad, " is ", class(controls[[bad]]))
  }
  names(controls) <- NULL

  if (length(reinit)) {
    bad <- vapply(controls, function(x) x$type == "vertexSetter" && length(intersect(reinit, x$objid)), FALSE)
    if (any(bad))
      warning("'vertexControl' is incompatible with re-initialization")
  }

  if (!length(components))
    components <- character()
  else
    components <- match.arg(components, 
    			c("Reverse", "Play", "Slower", "Faster", "Reset", "Slider", "Label", "Step"),
    			several.ok = TRUE)

  buttonLabels <- as.character(buttonLabels)
  pause <- as.character(pause)
  stopifnot(length(buttonLabels) == length(components),
  	    length(pause) == 1)
  
  dependencies <- list(rglDependency)
  
  for (i in seq_along(controls)) {
    control <- controls[[i]]
    if (is.null(labels)) 
      labels <- control$labels
    if (!is.null(control$param)) {
      start <- min(start, control$param[is.finite(control$param)])
      stop <- max(stop, control$param[is.finite(control$param)])
    }
    if (!is.null(control$dependencies))
      dependencies <- c(dependencies, control$dependencies)
  }

  if (is.null(stop) && length(labels)) {
    stop <- start + (length(labels) - 1)*step
  }
  
  if (is.null(stop)) {
    if ("Slider" %in% components) {
      warning("Cannot have slider with non-finite limits")
      components <- setdiff(components, "Slider")
    }
    labels <- NULL
  } else {
    if (stop == start)
      warning("'stop' and 'start' are both ", start)
  }

  control <- list(
       actions = controls,
       start = start,
       stop = stop,
       value = start,
       interval = interval,
       rate = rate,
       components = components,
       buttonLabels = buttonLabels,
       pause = pause,
       loop = loop,
       step = step,
       labels = labels,
       precision = precision,
       reinit = reinit,
       sceneId = upstream$prevRglWidget,
       respondTo = respondTo)

  result <- createWidget(
    name = 'rglPlayer',
    x = control,
    elementId = elementId,
    package = 'rgl',
    height = height,
    sizingPolicy = sizingPolicy(defaultWidth = "auto",
                                defaultHeight = "auto"),
    dependencies = dependencies,
    ...
  )
  if (is.list(upstream$objects)) {
  	if (requireNamespace("manipulateWidget", quietly = TRUE))
  		result <- do.call(manipulateWidget::combineWidgets, 
                        c(upstream$objects, 
                         list(manipulateWidget::combineWidgets(result, nrow = 1), 
                         rowsize = c(upstream$rowsizes, height), 
                        ncol = 1)))
  	else
  		warning("Combining widgets requires the 'manipulateWidget' package.", call. = FALSE)
  }
  result
}

toggleWidget <- function(sceneId, ids = tagged3d(tags), tags = NULL, 
                         hidden = integer(),
                         subscenes = NULL, 
                         label, 
                         ...) {
  if (missing(label))
    if (missing(ids)) 
       label <- paste(tags, collapse = ", ")
    else 
       label <- deparse(substitute(ids)) 
  playwidget(sceneId, 
             subsetControl(subsets = list(ids, hidden), subscenes = subscenes),
             start = 0, stop = 1,
             components = "Step",
             buttonLabels = label,
             interval = 1,
             ...)
}

writeSTL <- function(con, ascii=FALSE, pointRadius=0.005, 
                     pointShape = icosahedron3d(),
                     lineRadius = pointRadius,
                     lineSides = 20,
                     ids = tagged3d(tags),
                     tags = NULL) {
 
  writeHeader <- function() {
    ident <- paste(filename, " produced by RGL\n")
    if (ascii) 
      cat("solid ", ident, file=con)
    else {
      padding <- paste(rep(" ", 80), collapse="")
      ident <- substr( paste("binary", ident, padding), 1, 80)
      writeChar(ident, con, nchars=80, useBytes=TRUE, eos=NULL)
      writeBin(0L, con, size=4, endian="little")
    }
  }
  
  finish  <- function()
  {  if (!ascii) {
      seek(con, 80)
      writeBin(as.integer(triangles), con, size=4, endian="little")
     } else
       cat("endsolid\n", file = con)
  }
      
  triangles <- 0
  
  writeTriangles <- function(vertices) {
    if (nrow(vertices) %% 3 != 0) stop("Need 3N vertices")
    n <- nrow(vertices) / 3
    for (i in seq_len(n)) {
      vec0 <- vertices[3*i - 2,]
      vec1 <- vertices[3*i - 1,]
      vec2 <- vertices[3*i,]
      normal <- normalize(xprod(vec2-vec0, vec1-vec0))
      if (ascii) {
        cat("facet normal ", normal, "\n", file=con)
        cat("outer loop\n", file=con)
        cat("vertex ", vec0, "\n", file=con)
        cat("vertex ", vec1, "\n", file=con)
        cat("vertex ", vec2, "\n", file=con)
        cat("endloop\n", file=con)
        cat("endfacet\n", file=con)
      } else {
        writeBin(c(normal, vec0, vec1, vec2), con, size=4, endian="little")
        writeBin(0L, con, size=2, endian="little")
      }
    }
    triangles <<- triangles + n
  }
  
  writeQuads <- function(vertices) {
    if (nrow(vertices) %% 4 != 0) stop("Need 4N vertices")
    n <- nrow(vertices) / 4
    indices <- rep(seq_len(n)*4, each=6) - rep(c(3, 2, 1, 3, 1, 0), n)
    writeTriangles( vertices[indices,] )
  }
      
  writeSurface <- function(id) {
    vertices <- expandVertices(id)
    dims <- rgl.attrib(id, "dim")
    nx <- dims[1]
    nz <- dims[2]
    indices <- integer(0)
    for (j in seq_len(nx-1) - 1) {
      v1 <- j + nx*(seq_len(nz) - 1) + 1
      indices <- c(indices, rbind(v1[-nz], v1[-nz]+1, v1[-1]+1, v1[-1]))
    }
    writeQuads(vertices[indices,])
  }
  
  writeMesh <- function(mesh, scale=1, offset=c(0,0,0)) {
    vertices <- asEuclidean(t(mesh$vb))*scale 
    vertices <- vertices + rep(offset, each=nrow(vertices))
    nt <- length(mesh$it)/3
    nq <- length(mesh$ib)/4
    newverts <- matrix(NA, 3*nt+6*nq, 3)
    if (nt) 
      newverts[1:(3*nt),] <- vertices[c(mesh$it),]
    if (nq)
      newverts[3*nt+1:(6*nq),] <- vertices[c(mesh$ib[1:3,], 
      				    mesh$ib[c(1,3,4),]),]
    writeTriangles(newverts)
  }

  writeSpheres <- function(id) {
    vertices <- expandVertices(id)
    radii <- rgl.attrib(id, "radii") # FIXME to handle indices?
    n <- nrow(vertices)
    radii <- rep(radii, length.out=n)
    x <- subdivision3d(icosahedron3d(),3)
    r <- sqrt(x$vb[1,]^2 + x$vb[2,]^2 + x$vb[3,]^2)
    x$vb[4,] <- r
    for (i in seq_along(radii)) 
      writeMesh(x, radii[i], vertices[i,])
  }  
  
  avgScale <- function() {
    bbox <- par3d("bbox")
    ranges <- c(bbox[2]-bbox[1], bbox[4]-bbox[3], bbox[6]-bbox[5])
    if (prod(ranges) == 0) 1
    else exp(mean(log(ranges)))
  }  
   
  writePoints <- function(id) {
    vertices <- expandVertices(id)
    n <- nrow(vertices)
    radius <- pointRadius*avgScale()
    for (i in seq_len(n))
      writeMesh(pointShape, radius, vertices[i,])
  }
  
  writeSegments <- function(id) {
    vertices <- expandVertices(id)
    n <- nrow(vertices)/2
    radius <- lineRadius*avgScale()
    for (i in seq_len(n)) {
      cyl <- cylinder3d(vertices[(2*i-1):(2*i),], 
      			radius = radius,
      			sides = lineSides,
      			closed = -2)
      writeMesh(cyl)
    }
  }
  
  writeLines <- function(id) {
    vertices <- expandVertices(id)
    n <- nrow(vertices) - 1
    radius <- lineRadius*avgScale()
    for (i in seq_len(n)) 
      writeMesh( cylinder3d( vertices[i:(i+1),],
     			   radius = radius,
     			   sides = lineSides, 
     			   closed = -2) )
  }
  
  knowntypes <- c("triangles", "quads",
                  "surface", "planes", "spheres",
                  "points", "lines", "linestrip")
  
  #  Execution starts here!

  if (is.character(con)) {
    con <- file(con, if (ascii) "w" else "wb")
    on.exit(close(con))
  }
  filename <- summary(con)$description
  
  if (NROW(bbox <- ids3d("bboxdeco")) && (is.null(ids) || bbox$id %in% ids)) {
    ids <- setdiff(ids, bbox$id)
    save <- par3d(skipRedraw = TRUE)
    bbox <- convertBBox(bbox$id)
    on.exit({ pop3d(id=bbox); par3d(save) }, add=TRUE) # nolint
    dobbox <- TRUE
  } else dobbox <- FALSE 
  
  if (is.null(ids)) {
    ids <- ids3d()
    types <- as.character(ids$type)
    ids <- ids$id
  } else {
    if (dobbox) ids <- c(ids, bbox)
    allids <- ids3d()
    ind <- match(ids, allids$id)
    keep <- !is.na(ind)
    if (any(!keep)) warning(gettextf("Object(s) with id %s not found", paste(ids[!keep], collapse=" ")),
    			    domain = NA)
    ids <- ids[keep]
    types <- allids$type[ind[keep]]
  }  
    
  unknowntypes <- setdiff(types, knowntypes)
  if (length(unknowntypes))
    warning(gettextf("Object type(s) %s not handled", 
      paste("'", unknowntypes, "'", sep="", collapse=", ")), domain = NA)

  keep <- types %in% knowntypes
  ids <- ids[keep]
  types <- types[keep]

  writeHeader()
  
  for (i in seq_along(ids)) 
    switch(types[i],
      planes =,
      triangles = writeTriangles(expandVertices(ids[i])),
      quads = writeQuads(expandVertices(ids[i])),
      surface = writeSurface(ids[i]),
      spheres = writeSpheres(ids[i]),
      points = writePoints(ids[i]),
      lines = writeSegments(ids[i]),
      linestrip = writeLines(ids[i])
    )
  
  finish()
    
  invisible(filename)
}

readSTL <- function(con, ascii=NA, plot=TRUE, ...) {
 
  # Utility functions and constants defined first; execution starts way down there...
  
  triangles <- NULL
  
  readHeader <- function() {
    if (is.na(ascii)) {
      summ <- summary(con)
      if (summ$class != "file" || readChar(con, 5) != "solid")
        ascii <<- FALSE
      else {
        ascii <<- TRUE
        filename <- summ$description
        close(con)
        con <<- file(filename, "rt")
      }
    }
    if (ascii) {
      ident <- readLines(con, 1)
      if (!grepl("^solid ", ident)) stop("This does not appear to be an ASCII STL file")
    }
    if (!ascii) {
      seek(con, 80)
      triangles <<- readBin(con, "integer", n=1, size=4, endian="little")
    }
  }
  
  readTriangles <- function(n) {
    
    if (ascii) {
      lines <- readLines(con)
      facet <- c(grep("^ *facet", lines),
                 length(lines) + 1) # sentinel
      n <- length(facet) - 1
      vertex <- grep("^ *vertex ", lines)
    }
   
    vertices <- matrix(NA, 3*n, 3)
    
    if (ascii) {
      for (i in seq_len(n)) {
        j <- vertex[facet[i] < vertex & vertex < facet[i+1]]

        if (any(is.na(j)) || length(j) != 3) {
          warning("problem on triangle ", i, " j = ", j)
          j <- j[1:3]
        }
        v <- read.table(text = lines[j],
                        colClasses = c("character", "numeric", "numeric", "numeric"))
        vertices[3*i + c(-2, -1, 0),] <- as.matrix(v[, 2:4])
      }
    } else {
      for (i in seq_len(n)) {
        m <- matrix(readBin(con, "double", n=12, size=4, endian="little"), 4, 3, byrow=TRUE)
        vertices[3*i + c(-2,-1,0),] <- m[2:4,]
        count <- readBin(con, "integer", n=1, size=2, endian="little") # nolint
      }
    }
    vertices
  }
      
  #  Execution starts here!

  if (is.character(con)) {
    con <- file(con, if (isTRUE(ascii)) "rt" else "rb")
    on.exit(close(con))
  }
  
  readHeader()
  vertices <- readTriangles(triangles)
  if (plot)
    triangles3d(vertices, ...)
  else
    vertices
}

##
## R source file
## This file is part of rgl
##
##

##
## ===[ SECTION: internal ]===================================================
##

#
# rgl.clamp
#
# clamp value if lower than low or higher than high
#

rgl.clamp <- function(value, low, high) {
  if (value < low) {
    warning( gettextf("Value clamped to %s",low), domain = NA )
    result <- low
  }
  else if (value > high) {
    warning( gettextf("Value clamped to %s",high), domain = NA )
    result <- high
  }
  else {
    result <- value
  }

  return(result)
}

##
## types verification
##


#
# single field bool
#

rgl.bool <- function( x ) {
  if (length(x) > 1)
    stop( gettextf("'%s' must be a single boolean value", deparse(substitute(x))),
          domain = NA)
}


#
# single field numeric
#

rgl.numeric <- function( x ) {
  if (length(x) > 1)
    stop( gettextf("'%s' must be a single numeric value", deparse(substitute(x))),
          domain = NA)
}

#
# single string
#

rgl.string <- function( x ) {
  if (length(x) != 1)
    stop( gettextf("'%s' must be a single character value",
                   deparse(substitute(x))), domain = NA)
}


#
# vertex data object
#

rgl.vertex <- function(x, y = NULL, z = NULL) {
  xyz <- xyz.coords(x, y, z, recycle=TRUE)
  # This is not the same as just rbind(xyz$x,xyz$y,xyz$z)!
  # xyz.coords puts all of named vector x into xyz$x
  return( matrix( rbind(xyz$x,xyz$y,xyz$z), nrow=3, dimnames=list( c("x","y","z"), NULL ) ) )
}

#
# texture coordinate data object
#

rgl.texcoords <- function(s,t=NULL) {
  xy <- xy.coords(s, t, recycle=TRUE)
  return( matrix( rbind(xy$x, xy$y), nrow=2, dimnames=list( c("s", "t"), NULL ) ) )
}

#
# obtain number of vertices
#

rgl.nvertex <- function(vertex) {
  return( ncol(vertex) )
}


#
# rgl.color - single field color
#

rgl.color <- function( color ) {
  if (length(color) > 1)
    stop( gettextf("'%s' must be a single color character string", deparse(substitute(color))), 
          domain = NA)
  else
    return(col2rgb(color))
}


#
# rgl.mcolor - multiple field colors
#

rgl.mcolor <- function( colors ) {
  return( col2rgb(colors) )
}


#
# if vattr > 1, recycle data
#

rgl.attr <- function(vattr, nvertex) {
  nvattr <- length(vattr)

  if ((nvattr > 1) && (nvattr != nvertex))
    vattr  <- rep(vattr,length.out=nvertex)
  
  return(vattr)
}

# This displays an HTML5 input widget to show a subset of objects.  It assigns a random id
# and returns that invisibly.

subsetSlider <- function(subsets, labels = names(subsets),
                         fullset = Reduce(union, subsets), 
                         subscenes = currentSubscene3d(), prefixes = "", 
                         accumulate = FALSE, ...) {
  .Defunct("subsetControl")
}

subsetSetter <- function(subsets, subscenes = currentSubscene3d(), prefixes = "",
			 fullset = Reduce(union, subsets),
			 accumulate = FALSE) {
  .Defunct("subsetControl")
}

toggleButton <- function(subset, subscenes = currentSubscene3d(), prefixes = "",
			 label = deparse(substitute(subset)),
			 id = paste0(basename(tempfile("input"))), name = id) {
  .Defunct("toggleWidget")
}

clipplaneSlider <- function(a=NULL, b=NULL, c=NULL, d=NULL,
			    plane = 1, clipplaneids, prefixes = "",
			    labels,
			      ...) {
  .Defunct("clipplaneControl")
}

propertySlider <- function(setter = propertySetter,
                           minS = NULL, maxS = NULL, step = 1, init = NULL,
                           labels,
                           id = basename(tempfile("input")), name = id,
			   outputid = paste0(id, "text"),
			   index = NULL,
                           ...)  {
  .Defunct("propertyControl")
}

propertySetter <- function(values = NULL, entries, properties, objids, prefixes = "",
                           param = seq_len(NROW(values)), interp = TRUE,
			   digits = 7)  {
  .Defunct("propertyControl")
}

vertexSetter <- function(values = NULL, vertices = 1, attributes, objid, prefix = "",
			 param = seq_len(NROW(values)), interp = TRUE,
			 digits = 7)  {
  .Defunct("vertexControl")
}


par3dinterpSetter <- function(fn, from, to, steps, subscene = NULL,
			      omitConstant = TRUE, rename = character(), ...) {
  .Defunct("par3dinterpControl")
}

matrixSetter <- function(fns, from, to, steps, subscene = currentSubscene3d(), matrix = "userMatrix",
			omitConstant = TRUE, prefix = "", ...) {
  .Defunct("par3dinterpControl")
}

ageSetter <- function(births, ages, colors = NULL, alpha = NULL,
		      radii = NULL, vertices = NULL, normals = NULL,
		      origins = NULL, texcoords = NULL, objids, prefixes = "",
		      digits = 7, param = seq(floor(min(births)), ceiling(max(births))))  {
  .Defunct("ageControl")
}

as.tmesh3d <- function(x, ...)
  UseMethod("as.tmesh3d")

as.tmesh3d.default <- function(x, drop = FALSE, ...) {
  as.tmesh3d(as.mesh3d(x, ...), drop = drop)  
}

as.tmesh3d.mesh3d <- function(x, drop = FALSE, keepTags = FALSE,
															...) {
  mesh <- x
  ib <- mesh$ib
  it <- mesh$it
  np <- length(mesh$ip)
  ns <- length(mesh$is)/2
  nt <- length(it)/3
  nq <- length(ib)/4
  
  hasColors <- FALSE
  hasAlpha <- FALSE
  if (!is.null(mesh$material) &&
  	  !is.null(mesh$meshColor) &&
  	  mesh$meshColor == "faces") {
  	hasColors <- length(mesh$material$color) > 1
  	hasAlpha  <- length(mesh$material$alpha) > 1
  }
  
  tags <- getTags(mesh)
  
  if (nq) {
    mesh$it <- cbind(it, 
                     matrix(ib[rep(4*(seq_len(nq) - 1), each = 6) + 
                                  rep(c(1, 2, 3, 1, 3, 4), nq)], nrow = 3))
    mesh$ib <- NULL
    ofs <- np + ns + nt
    
    if (hasColors) {
    	color <- rep_len(mesh$material$color, ofs + nq)
    	mesh$material$color[ofs + seq_len(2*nq)] <- rep(color[ofs + seq_len(nq)], each = 2)
    }
    
    if (hasAlpha) {
    	alpha <- rep_len(mesh$material$alpha, ofs + nq)
    	mesh$material$alpha[ofs + seq_len(2*nq)] <- rep(alpha[ofs + seq_len(nq)], each = 2)
    }
    
    tags[ofs + seq_len(2*nq)] <- rep(tags[ofs + seq_len(nq)], each = 2)

  }
  if (drop) {
    mesh$is <- NULL
    mesh$ip <- NULL
    if (ns + np)
    	tags <- tags[-seq_len(ns + np)]
  }
  if (keepTags)
  	mesh$tags <- tags
  else
  	mesh$tags <- NULL
  mesh
}

.getVertexFn <- function(fn, envir) {
  if (is.character(fn))
    fn <- switch(fn, 
                 x = function(xyz) xyz[,1],
                 y = function(xyz) xyz[,2],
                 z = function(xyz) xyz[,3],
                 get(fn, envir = envir, mode = "function"))
  # Accept functions that take x, y, z as parameters
  if (all(c("x", "y", "z") %in% names(formals(fn)))) {
    oldfn <- fn
    fn <- function(xyz) oldfn(x = xyz[,1], y = xyz[,2], z = xyz[,3])
  }
  fn
}

clipMesh3d <- function(mesh, fn = "z", bound = 0, greater = TRUE,
                       minVertices = 0, plot = FALSE, keepValues = FALSE, keepTags = FALSE) {
  stopifnot(inherits(mesh, "mesh3d"))
  # First, convert quads to triangles
  mesh <- as.tmesh3d(mesh, keepTags = TRUE)
  nverts <- ncol(mesh$vb)
  oldnverts <- nverts - 1
  while (nverts < minVertices && oldnverts < nverts && !is.numeric(fn)) {
    oldnverts <- nverts
    mesh <- subdivision3d(mesh, deform = FALSE, 
    											normalize = TRUE, keepTags = TRUE)
    nverts <- ncol(mesh$vb)
  }
  if (is.null(fn))
    fn <- mesh$values
  if (is.null(fn))
    stop("'fn' can only be NULL if 'mesh' contains values")
  if (is.numeric(fn)) 
    values <- fn
  else {
    fn <- .getVertexFn(fn, parent.frame())
    verts <- asEuclidean(t(mesh$vb))
    values <- fn(verts)
  }
  # The bound might be infinite, which messes up the arithmetic.
  # Force it to a finite value
  r <- range(values)
  delta <- max(abs(r))
  if (bound < r[1])
    bound <- r[1] - delta
  else if (bound > r[2])
    bound <- r[2] + delta
  
  values <- values - bound
  if (!greater)
    values <- -values
  
  if (length(values) != ncol(mesh$vb))
    stop("'fn' should give one value per vertex")
  
  if (anyNA(values))
    stop("'fn' should not include NA values")

  # Now, set all w values to 1
  mesh$vb <- t( cbind(t(mesh$vb[1:3,])/mesh$vb[4,], 1))
  if (!is.null(mesh$normals) && nrow(mesh$normals) == 4)
    mesh$normals <- t(t(mesh$normals[1:3,])/mesh$normals[4,])
  
  tags <- getTags(mesh)
  np <- length(mesh$ip)
  ns <- length(mesh$is)/2
  nt <- length(mesh$it)/3
  # nq is zero because of as.tmesh3d
  
  ofs <- np + ns
  tags0 <- tags[ofs + seq_len(nt)]
  
  newVertices <- integer()
  getNewVertex <- function(good, bad) {
    names <- paste0(good, "_", bad)
    new <- which(!(names %in% names(newVertices)))
    if (length(new)) {
      goodvals <- values[good[new]]
      badvals <- values[bad[new]]
      alphas <- goodvals/(goodvals - badvals)
      newverts <- t((1 - alphas)*t(mesh$vb[, good[new]]) + alphas*t(mesh$vb[, bad[new]]))
      newvertnums <- seq_len(ncol(newverts)) + ncol(mesh$vb)
      if (!is.null(mesh$normals)) 
        mesh$normals <<- cbind(mesh$normals, t((1 - alphas)*t(mesh$normals[, good[new]]) +
                                             alphas*t(mesh$normals[, bad[new]])))
      if (!is.null(mesh$texcoords))
        mesh$texcoords <<- cbind(mesh$texcoords, t((1 - alphas)*t(mesh$texcoords[, good[new]]) +
                                             alphas*t(mesh$texcoords[, bad[new]])))
      if (!is.null(mesh$material)) {
        if (!is.null(mesh$material$color) && length(mesh$material$color) == ncol(mesh$vb)) {
          rgb <- col2rgb(mesh$material$color)
          newrgb <- (1 - alphas)*rgb[, good[new],drop = FALSE] + alphas*rgb[, bad[new], drop = FALSE]
          mesh$material$color <<- c(mesh$material$color, 
                                   rgb(newrgb["red",], newrgb["green",], newrgb["blue",], maxColorValue = 255))
        }
        if (!is.null(mesh$material$alpha) && length(mesh$material$alpha) == ncol(mesh$vb))
          mesh$material$alpha <<- c(mesh$material$alpha,
                                    (1-alphas)*mesh$material$alpha[good[new]] +
                                        alphas*mesh$material$alpha[bad[new]])
        
      }
      values <<- c(values, rep(0, ncol(newverts)))
      mesh$vb <<- cbind(mesh$vb, newverts)
      newVertices[names[new]] <<- newvertnums
    }
    newVertices[names]
  }
  keep <- values >= 0
  keept <- matrix(keep[mesh$it], nrow = 3)
  counts <- colSums(keept)  # Number of vertices to keep for each triangle
  singles <- which(counts == 1)
  if (length(singles)) {
    theseTriangles <- mesh$it[, singles, drop = FALSE]
    goodRow <- apply(keept[, singles, drop = FALSE], 2, function(col) which(col))
    allcols <- seq_len(ncol(theseTriangles))
    goodVertex <- theseTriangles[cbind(goodRow, allcols)]
    badVertex1 <- theseTriangles[cbind(goodRow %% 3 + 1, allcols)]
    badVertex2 <- theseTriangles[cbind((goodRow + 1) %% 3 + 1, allcols)]
    mesh$it[cbind(goodRow %% 3 + 1, singles)] <- 
      getNewVertex(goodVertex, badVertex1)
    mesh$it[cbind((goodRow + 1) %% 3 + 1, singles)] <-
      getNewVertex(goodVertex, badVertex2)
  }
  doubles <- which(counts == 2)
  if (length(doubles)) {
    theseTriangles <- mesh$it[, doubles, drop = FALSE]
    badRow <- apply(keept[, doubles, drop = FALSE], 2, function(col) which(!col))
    allcols <- seq_len(ncol(theseTriangles))
    badVertex <- theseTriangles[cbind(badRow, allcols)]
    goodVertex1 <- theseTriangles[cbind(badRow %% 3 + 1, allcols)]
    goodVertex2 <- theseTriangles[cbind((badRow + 1) %% 3 + 1, allcols)]
    newVertex1 <- getNewVertex(goodVertex1, badVertex)
    newVertex2 <- getNewVertex(goodVertex2, badVertex)
    mesh$it[cbind(badRow, doubles)] <- newVertex1
    mesh$it <- cbind(mesh$it, rbind(newVertex1, goodVertex2, newVertex2))
    tags0 <- c(tags0, tags0[doubles])
  }
  zeros <- which(counts == 0)
  if (length(zeros)) {
    mesh$it <- mesh$it[, -zeros]
    tags0 <- tags0[-zeros]
  }
  
  tags <- c(tags[seq_len(ofs)], tags0)
  mesh$tags <- tags
  
  if (plot)
    shade3d(mesh)
  else {
    if (keepValues) {
      if (!greater) values <- -values
      mesh$values <- values + bound
    }
    cleanMesh3d(mesh, keepTags = keepTags)
  }
}

cleanMesh3d <- function(mesh, onlyFinite = TRUE, allUsed = TRUE, rejoin = FALSE, keepTags = TRUE) {
  if (rejoin) {
    ntriangs <- ncol(mesh$it)
    oldntriangs <- ntriangs + 1
    while (ntriangs < oldntriangs) {
      oldntriangs <- ntriangs
      mesh <- rejoinMesh3d(mesh)
      ntriangs <- ncol(mesh$it)
    }    
  }
	tags <- getTags(mesh)
  nold <- ncol(mesh$vb)
  keep <- TRUE
  if (onlyFinite)
    keep <- keep & apply(mesh$vb, 2, function(col) all(is.finite(col)))
  if (allUsed)
    keep <- keep & (seq_len(nold) %in% c(mesh$ip, mesh$is, mesh$it, mesh$ib))
  if (!all(keep)) {
    oldnums <- which(keep)
    newnums <- rep(NA, nold)
    nnew <- sum(keep)
    newnums[oldnums] <- seq_len(nnew)
    mesh$vb <- mesh$vb[,oldnums]
    np <- length(mesh$ip)
    ns <- length(mesh$is)/2
    nt <- length(mesh$it)/3
    nq <- length(mesh$ib)/4
    if (!is.null(mesh$ip)) {
      newcols <- newnums[mesh$ip]
      dim(newcols) <- dim(mesh$ip)
      keep <- apply(newcols, 2, function(col) !is.na(col))
      mesh$ip <- newcols[,keep, drop = FALSE]
      tags0 <- tags[seq_len(np)]
      tags0 <- tags0[keep]
      tags <- c(tags0, tags[np + seq_len(np + nt + nq)])
      np <- length(tags0)
    }
    if (!is.null(mesh$is)) {
      newcols <- newnums[mesh$is]
      dim(newcols) <- dim(mesh$is)
      keep <- apply(newcols, 2, function(col) all(!is.na(col)))
      mesh$is <- newcols[,keep, drop = FALSE]
      tags0 <- tags[np + seq_len(ns)]
      tags0 <- tags0[keep]
      tags <- c(tags[seq_len(np)], tags0, tags[np + ns + seq_len(nt + nq)])
      ns <- length(tags0)
    }
    if (!is.null(mesh$it)) {
      newcols <- newnums[mesh$it]
      dim(newcols) <- dim(mesh$it)
      keep <- apply(newcols, 2, function(col) all(!is.na(col)))
      mesh$it <- newcols[,keep, drop = FALSE]
      tags0 <- tags[np + ns + seq_len(nt)]
      tags0 <- tags0[keep]
      tags <- c(tags[seq_len(np + ns)], tags0, tags[np + ns + nt + seq_len(nq)])
      nt <- length(tags0)
    }
    if (!is.null(mesh$ib)) {
      newcols <- newnums[mesh$ib]
      dim(newcols) <- dim(mesh$ib)
      keep <- apply(newcols, 2, function(col) all(!is.na(col)))
      mesh$ib <- newcols[,keep, drop = FALSE]
      tags0 <- tags[np + ns + nt + seq_len(nq)]
      tags0 <- tags0[keep]
      tags <- c(tags[seq_len(np + ns + nt)], tags0)
      nq <- length(tags0)
    }
    if (!is.null(mesh$normals)) 
      mesh$normals <- mesh$normals[, oldnums]
    if (!is.null(mesh$texcoords)) 
      mesh$texcoords <- mesh$texcoords[, oldnums]
    if (!is.null(mesh$meshColor) && mesh$meshColor == "vertices" 
        && !is.null(mesh$material) 
        && length(mesh$material$color) == nold)
      mesh$material$color <- mesh$material$color[oldnums]
    if (!is.null(mesh$values))
      mesh$values <- mesh$values[oldnums]
  }
  if (keepTags)
  	mesh$tags <- tags
  else
  	mesh$tags <- NULL
  mesh
}

subdivideLines <- function(x) {
  nverts <- nrow(x$vertices)
  newverts <- nverts
  indices <- rbind(seq_len(nverts), NA, NA)
  finite <- apply(x$vertices, 1, function(row) all(is.finite(row)))
  type <- x$type
  if (type == "lines") {
    for (j in 2*seq_len(nverts %/% 2)) {
      i <- j - 1
      if (finite[i] && finite[j]) {
        for (attr in c("vertices", "normals", "colors")) {
          if (!is.null(x[[attr]]) && nrow(x[[attr]]) > 1) {
            new <- (x[[attr]][i,] + x[[attr]][j,])/2
            while (nrow(x[[attr]]) < newverts + 2) 
              x[[attr]] <- rbind(x[[attr]], x[[attr]])
            x[[attr]][newverts + 1,] <- new
            x[[attr]][newverts + 2,] <- new
          }
        }
        indices[2, i] <- newverts + 1
        indices[3, i] <- newverts <- newverts + 2
      }
    }
  } else if (x$type == "linestrip") {
    for (j in seq_len(nverts)[-1]) {
      i <- j - 1
      if (finite[i] && finite[j]) {
        for (attr in c("vertices", "normals", "colors")) {
          if (!is.null(x[[attr]]) && nrow(x[[attr]]) > 1) {
            new <- (x[[attr]][i,] + x[[attr]][j,])/2
            while (nrow(x[[attr]]) < newverts + 1) 
              x[[attr]] <- rbind(x[[attr]], x[[attr]])
            x[[attr]][newverts + 1,] <- new
          }
        }
        indices[2, i] <- newverts <- newverts + 1
      }
    }
  }
  if (nverts < newverts) {
    indices <- c(indices)
    indices <- indices[!is.na(indices)]
    for (attr in c("vertices", "normals", "colors")) 
      if (!is.null(x[[attr]]) && nrow(x[[attr]]) > 1) 
        x[[attr]] <- x[[attr]][indices,,drop = FALSE]
  }
  x
}

# Undo subdivision for triangles
rejoinMesh3d <- function(x, tol = 1.e-6) {
  ntriangs <- length(x$it)/3
  if (ntriangs < 4)
    return(x)
  np <- length(x$ip)
  ns <- length(x$is)/2
  nq <- length(x$ib)/4
  ofs <- np + ns
  tags <- getTags(x)
  tags0 <- tags[ofs + seq_len(ntriangs)]
  vals <- if (nrow(x$vb) == 4) asEuclidean(t(x$vb)) else t(x$vb)
  if (!is.null(x$normals))
    vals <- cbind(vals, if(nrow(x$normals) == 4) asEuclidean(t(x$normals)) else t(x$normals))
  if (!is.null(x$texcoords))
    vals <- cbind(vals, t(x$texcoords))
  indices <- seq_len(ntriangs)
  for (j in seq_len(ntriangs)[-(1:3)]) {
    i <- j - (3:0)
    verts <- c(x$it[,i])
    thistags <- tags0[i]
    if ( all(thistags == thistags[1])
    	&& !any(is.na(indices[i]))
      && verts[2] == verts[6]
      && verts[3] == verts[8]
      && verts[5] == verts[9]
      && verts[6] == verts[11]
      && verts[8] == verts[10]
      && verts[9] == verts[12] 
      && !(verts[1] %in% verts[-1]) 
      && !(verts[4] %in% verts[-4])
      && !(verts[7] %in% verts[-7])) {
        v1 <- verts[c(1,4,7,2,3,5)]
        diffs <- c(vals[v1[1],] - vals[v1[2],],
                   vals[v1[1],] - vals[v1[3],],
                   vals[v1[2],] - vals[v1[3],])
        use <- !is.na(diffs) & diffs > tol
        if (any(use)) {
          p <- c(vals[v1[1],] - vals[v1[4],],
                 vals[v1[1],] - vals[v1[5],],
                 vals[v1[2],] - vals[v1[6],])[use]/diffs[use]
          if (max(abs(p - 0.5)) < tol) {
            # Found one!
            indices[i[-1]] <- NA
            x$it[,i[1]] <- c(v1[1], v1[2], v1[3])
          }
        }
    }
  }
  indices <- indices[!is.na(indices)]
  x$it <- x$it[,indices]
  x$tags <- c(tags[seq_len(ofs)], 
  						tags0[indices], 
  						tags[ofs + ntriangs + seq_len(nq)])
  x
}

rejoinLines3d <- function(x, tol = 1.e-6) {
  nverts <- nrow(x$vertices)
  indices <- seq_len(nverts)
  finite <- apply(x$vertices, 1, function(row) all(is.finite(row)))
  type <- x$type
  hasattrs <- character()
  for (attr in c("vertices", "normals", "colors")) 
    if (!is.null(x[[attr]]) && nrow(x[[attr]]) > 1) 
      hasattrs <- c(hasattrs, attr)
  if (type == "lines") {
    for (j in 2*seq_len(nverts %/% 2)[-1]) {
      i <- j - (3:0)
      if (all(finite[i])) {
        attrs <- matrix(numeric(), nrow = 4, ncol = 0)
        for (attr in hasattrs) 
          attrs <- cbind(attrs, x[[attr]][i,])
        diff <- attrs[4,] - attrs[1,]
        use <- !is.na(diff) & abs(diff) > tol
        p <- c((attrs[2,use] - attrs[1,use])/diff[use],
               (attrs[3,use] - attrs[1,use])/diff[use])
        if (all(!is.na(p) & p >= 0 & p <= 1)
            && diff(range(p)) < tol) {
          # pts 2 & 3 are equal and are exactly between
          # pts 1 & 4.  Merge them into 3,4.
          for (attr in hasattrs) 
            x[[attr]][i[3],] <- x[[attr]][i[1],]
          indices[i[1:2]] <- NA  
        }
      }
    }
  } else if (type == "linestrip") {
    for (j in seq_len(nverts)[-(1:2)]) {
      i <- j - (2:0)
      if (all(finite[i])) {
        attrs <- matrix(numeric(), nrow = 3, ncol = 0)
        for (attr in hasattrs) 
          attrs <- cbind(attrs, x[[attr]][i,])
        diff <- attrs[3,] - attrs[1,]
        use <- !is.na(diff) & abs(diff) > tol
        p <- (attrs[2,use] - attrs[1,use])/diff[use]
        if (all(!is.na(p) & p >= 0 & p <= 1)
            && diff(range(p)) < tol) {
          # pt 2 is exactly between
          # pts 1 & 3.  Move 1 there.
          for (attr in hasattrs) 
            x[[attr]][i[2],] <- x[[attr]][i[1],]
          indices[i[1]] <- NA  
        }
      }
    }
  }
  if (anyNA(indices)) {
    indices <- indices[!is.na(indices)]
    for (attr in hasattrs)
      x[[attr]] <- x[[attr]][indices,]
  }
  x
}

clipObj3d <- function(ids = tagged3d(tags), fn, bound = 0, greater = TRUE,
                      minVertices = 0,
                      replace = TRUE, tags) {
  getValues <- function(obj) {
    verts <- obj$vertices
    nverts <- nrow(verts)
    values <- rep(NA_real_, nverts)
    finite <- apply(verts, 1, function(row) all(is.finite(row)))
    values[finite] <- fn(verts[finite,]) - bound
    if (!greater) values <- -values
    values
  }
  getKeep <- function(values) {
    !is.na(values) & values > 0
  }
  applyKeep <- function() {
    for (attr in c("vertices", "normals", "colors",
                   "texcoords", "centers", "adj"))
      if (!is.null(obj[[attr]]) &&
          nrow(obj[[attr]]) > 1) 
        obj[[attr]] <- obj[[attr]][keep,,drop = FALSE]
    for (attr in c("texts", "cex", "adj", 
                   "radii", "ids", "types", 
                   "flags", "offsets", "pos"))
      if (length(obj[[attr]]) > 1)
        obj[[attr]] <- obj[[attr]][keep] 
    obj
  }
  scene <- scene3d()
  if (missing(ids))
    ids <- names(scene$objects)
  names <- names(ids)
  ids <- as.character(ids)
  result <- integer()
  minVertices <- rep(minVertices, length.out = length(ids))
  names(minVertices) <- ids
  fn <- .getVertexFn(fn, parent.frame())
  for (id in ids) {
    obj <- scene$objects[[id]]
    type <- obj$type
    nverts <- nrow(obj$vertices)
    newid <- NA
    switch(type,
           triangles=,
           quads=,
           planes=,
           surface= {
             mesh <- as.mesh3d(obj)
             mesh <- cleanMesh3d(mesh)
             clipped <- clipMesh3d(mesh, fn, bound, greater,
                                   minVertices[id])
             clipped <- cleanMesh3d(clipped, rejoin = TRUE)
             newid <- shade3d(clipped, override = FALSE)
           }, 
           points =,
           text =,
           spheres =,
           sprites = {
             keep <- getKeep(getValues(obj))
             if (!all(keep)) {
               obj <- applyKeep()
               newid <- plot3d(obj)
             }
           },
           lines = {
             oldnverts <- nverts - 1
             while (nverts < minVertices[id] && nverts > oldnverts) {
               oldnverts <- nverts
               obj <- subdivideLines(obj)
               nverts <- nrow(obj$vertices)
             }
             values <- getValues(obj)
             keep <- getKeep(values)
             if (!all(keep)) {
               for (j in 2*seq_len(nverts %/% 2)) {
                 i <- j - 1
                 if (is.na(values[i]) || is.na(values[j])) {
                   keep[i] <- keep[j] <- FALSE
                 } else if (!keep[i] && !keep[j]) {# no change
                 } else if (!keep[i]) {
                   p <- 1 - abs(values[i])/(abs(values[i]) + values[j])
                   obj$vertices[i,] <- p*obj$vertices[i,] + (1-p)*obj$vertices[j,]
                   keep[i] <- TRUE
                 } else if (!keep[j]) {
                   p <- 1 - abs(values[j])/(abs(values[j]) + values[i])
                   obj$vertices[j,] <- p*obj$vertices[j,] + (1-p)*obj$vertices[i,]
                   keep[j] <- TRUE
                 }
               }
               obj <- applyKeep()
               obj <- rejoinLines3d(obj)
               newid <- plot3d(obj)
             }
           },
           linestrip = {
             oldnverts <- nverts - 1
             while (nverts < minVertices[id] && nverts > oldnverts) {
               oldnverts <- nverts
               obj <- subdivideLines(obj)
               nverts <- nrow(obj$vertices)
             }
             newverts <- nverts
             values <- getValues(obj)
             if (!all(values >= 0, na.rm = TRUE)) {
               hasattrs <- character()
               for (attr in c("vertices", "normals", "colors"))
                 if (!is.null(obj[[attr]]) && nrow(obj[[attr]]) > 1)
                   hasattrs <- c(hasattrs, attr)
                 indices <- rbind(seq_len(nverts), NA, NA)
                 for (j in seq_len(nverts)[-1]) {
                   i <- j - 1
                   # There are lots of cases to consider here.
                   # We could have a point that was already ignored
                   # to create a gap in the linestrip, a point
                   # that should be deleted because it is outside the
                   # range, or a point that should be kept.  These
                   # have value NA, negative, or non-negative respectively.
                   # The 9 cases are handled as follows (assuming we
                   # go through (i, j=i+1) pairs in sequence):
                   # i  j  disposition
                   # NA NA drop i, keep j
                   # NA  - drop i, keep j
                   # NA  + keep both
                   # -  NA drop i, keep j
                   # -   - drop i, keep j unless it is last
                   # -   + change i to NA (unless it's first, then drop it), insert interpolant, keep j
                   # +  NA keep both
                   # +   - keep i, insert interpolant then NA, move to j
                   # +   + keep both
                   if (is.na(values[i])) {
                     if (is.na(values[j]) || values[j] < 0) 
                       indices[1,i] <- NA
                   } else if (values[i] < 0) {
                     if (is.na(values[j]) || values[j] < 0)
                       indices[1,i] <- NA
                     else {
                       indices[2,i] <- newverts + 1
                       p <- 1 - abs(values[i])/(abs(values[i]) + values[j])
                       for (attr in hasattrs) {
                         while (nrow(obj[[attr]]) < newverts + 1)
                           obj[[attr]] <- rbind(obj[[attr]], obj[[attr]])
                         new <- p*obj[[attr]][i,] + (1-p)*obj[[attr]][j,]
                         obj[[attr]][newverts + 1,] <- new
                       }
                       newverts <- newverts + 1
                       obj$vertices[i,] <- NA
                       if (j == 2)
                         indices[1,i] <- NA
                     }
                   } else {
                     if (!is.na(values[j]) && values[j] < 0) {
                       indices[2,i] <- newverts + 1
                       indices[3,i] <- newverts + 2
                       p <- 1 - abs(values[j])/(abs(values[j]) + values[i])
                       obj$vertices <- rbind(obj$vertices, 
                                             p*obj$vertices[j,] + (1-p)*obj$vertices[i,],
                                             c(NA, NA, NA))
                       for (attr in hasattrs) {
                         while (nrow(obj[[attr]]) < newverts + 2)
                           obj[[attr]] <- rbind(obj[[attr]], obj[[attr]])
                         new <- p*obj[[attr]][j,] + (1-p)*obj[[attr]][i,]
                         obj[[attr]][newverts+1,] <- new
                         obj[[attr]][newverts+2,] <- NA
                       }
                       newverts <- newverts + 2
                       if (j == nverts)
                         indices[1,j] <- NA                   
                     }
                   }
                 }
                 indices <- c(indices)
                 indices <- indices[!is.na(indices)]
                 for (attr in hasattrs) 
                   obj[[attr]] <- obj[[attr]][indices,,drop = FALSE]
             }
             obj <- rejoinLines3d(obj)
             newid <- plot3d(obj)
           }
    )
    if (!is.na(newid)) {
      result[id] <- newid
      if (replace)
        pop3d(id = id)
    } else
      result[id] <- as.integer(id)
  }
  if (!is.null(names))
    names(result) <- names
  rglId(result)
}

contourLines3d <- function(obj, ...) 
  UseMethod("contourLines3d")

contourLines3d.rglId <- function(obj, ...) 
  contourLines3d(as.mesh3d(obj), ...)
  
contourLines3d.mesh3d <- function(obj, fn = "z", nlevels = 10, levels = NULL, 
                                  minVertices = 0, plot = TRUE, ... ) {
  obj <- as.tmesh3d(obj)
  nverts <- ncol(obj$vb)
  oldnverts <- nverts - 1
  while (nverts < minVertices && oldnverts < nverts) {
    oldnverts <- nverts
    obj <- subdivision3d(obj, deform = FALSE, normalize = TRUE)
    nverts <- ncol(obj$vb)
  }
  verts <- asEuclidean(t(obj$vb))
  if (is.null(fn))
    fn <- obj$values
  if (is.null(fn))
    stop("'fn' can only be NULL if 'obj' contains values")
  if (is.character(fn))
    fn <- structure(as.list(fn), names = fn)
  else if (is.function(fn) || is.numeric(fn)) 
    fn <- list(fn)
  funnames <- names(fn)
  if (is.null(funnames))
    funnames <- seq_along(fn)
    
  result <- data.frame(x = numeric(), y = numeric(), z = numeric(), fn = funnames[0], level = numeric())
  
  for (i in seq_along(fn)) {
    if (is.numeric(fn[[i]]))
      values <- fn[[i]]
    else {
      fun <- .getVertexFn(fn[[i]], parent.env())
      values <- fun(verts)
    }
    if (is.null(levels))
      levs <- pretty(range(values, na.rm = TRUE), nlevels)
    else
      levs <- levels
    for (lev in levs) {
      greater <- matrix(values[as.numeric(obj$it)] > lev, nrow = 3)
      counts <- colSums(greater)
      crossings <- which(counts %in% 1:2)
      if (length(crossings)) {
        greater <- greater[,crossings, drop = FALSE]
        counts <- counts[crossings]
        # Find the single vertex on one side of the contour
        # line (v1), and the other two on the other side
        # (v2, v3)
        r1 <- ifelse(counts == 1, 
                     apply(greater, 2, which.max),
                     apply(greater, 2, which.min))
        v1 <- obj$it[cbind(r1, crossings)]
        r2 <- r1 %% 3 + 1
        v2 <- obj$it[cbind(r2, crossings)]
        r3 <- (r1 + 1) %% 3 + 1
        v3 <- obj$it[cbind(r3, crossings)]
        p1 <- (lev - values[v2])/(values[v1] - values[v2])
        i1 <- p1*verts[v1,] + (1 - p1)*verts[v2,]
        p2 <- (lev - values[v3])/(values[v1] - values[v3])
        i2 <- p2*verts[v1,] + (1 - p2)*verts[v3,] 
        xyz <- matrix(t(cbind(i1, i2)), ncol = 3, byrow = TRUE)
        result <- rbind(result, data.frame(x = xyz[,1], y = xyz[,2], z = xyz[,3], fn = funnames[i], level = lev))
      }
    }
  }
  if (plot)
    segments3d(result, ...)
  else
    result
}

rglToLattice <- function(rotm = par3d("userMatrix")) {
  if (!requireNamespace("orientlib", quietly = TRUE))
    stop("The orientlib package is needed for this function")
  e <- -orientlib::eulerzyx(orientlib::rotmatrix(rotm[1:3, 1:3]))@x*180/pi
  list(z = e[1], y = e[2], x = e[3])
}

rglToBase <- function(rotm = par3d("userMatrix")) {
  if (!requireNamespace("orientlib", quietly = TRUE))
    stop("The orientlib package is needed for this function")
  e <- (orientlib::eulerzyx(orientlib::rotmatrix((rotm[1:3,1:3]))))@x*180/pi
  list(theta = e[1], phi = 90 - e[3])
}

scene3d <- function(minimal = TRUE) {
  
  saveSubscene <- currentSubscene3d()
  on.exit(useSubscene3d(saveSubscene))
  
  defaultmaterial <- material3d()
  
  matdiff <- function(mat) {
    for (m in names(mat)) {
      if (identical(mat[[m]], defaultmaterial[[m]]))
        mat[[m]] <- NULL
    }
    mat
  }
  
  getObject <- function(id, type) {
    result <- list(id=id, type=type)
    
    if (!(type %in% c("light", "clipplanes"))) {
      mat <- rgl.getmaterial(id=id)
      lit <- mat$lit
      result$material <- matdiff(mat)
    } else
      lit <- FALSE
    
    attribs <- c("vertices", "colors", "texcoords", "dim",
          "texts", "cex", "adj", "radii", "ids",
          "usermatrix", "types", "offsets", "centers",
          "family", "font", "pos", "axes", "indices", "normals", "shapenum")
    for (a in attribs) 
      if (rgl.attrib.count(id, a))
        result[[a]] <- rgl.attrib(id, a)
    flags <- rgl.attrib(id, "flags")
    if (length(flags)) {
    	if ("ignoreExtent" %in% rownames(flags)) 
          result$ignoreExtent <- flags["ignoreExtent", 1]
    	if ("fixedSize" %in% rownames(flags))
    	  result$fixedSize <- flags["fixedSize", 1]
    	if ("rotating" %in% rownames(flags))
    	  result$rotating <- flags["rotating", 1]
    	if ("fastTransparency" %in% rownames(flags))
    	  result$fastTransparency <- flags["fastTransparency", 1]
    	if ("flipped" %in% rownames(flags))
    	  result$flipped <- flags["flipped", 1]
    }
    if (!is.null(result$ids)) {
      objlist <- vector("list", nrow(result$ids))
      for (i in seq_len(nrow(result$ids)))
        objlist[[i]] <- getObject(result$ids[i,1], result$types[i,1])
      result$objects <- objlist
    }
    if (type == "background") {
      flags <- rgl.attrib(id, "flags")
      result$sphere <- flags["sphere", 1]
      result$fogtype <- if (flags["linear_fog", 1]) "linear"
                        else if (flags["exp_fog", 1]) "exp"
			else if (flags["exp2_fog", 1]) "exp2"
			else "none"
      result$fogscale <- as.numeric(rgl.attrib(id, "fogscale"))
    } else if (type == "bboxdeco") {
      flags <- rgl.attrib(id, "flags")
      result$draw_front <- flags["draw_front", 1]
    } else if (type == "light") {
      flags <- rgl.attrib(id, "flags")
      result$viewpoint <- flags["viewpoint", 1]
      result$finite    <- flags["finite", 1]
    }
    class(result) <- c(paste0("rgl", type), "rglobject")
    result
  }

  getSubscene <- function(id) {
    useSubscene3d(id)
      
    result <- list(id = id, type = "subscene", par3d = par3d())

    result$embeddings <- subsceneInfo()$embeddings
    
    objs <- ids3d(c("background", "bboxdeco", "shapes", "lights"))
    result$objects <- objs$id
  
    if (nrow(obj <- ids3d("subscene", subscene = id))) {
      subscenes <- vector("list", nrow(obj))
      ids <- obj$id
      for (i in seq_len(nrow(obj)))
        subscenes[[i]] <- getSubscene(ids[i])
      result$subscenes <- subscenes
    }
    class(result) <- c("rglsubscene", "rglobject")
    result
  }
  
  result <- list()
  result$material <- defaultmaterial
  
  result$rootSubscene <- getSubscene(rootSubscene())
  
  objs <- ids3d(c("shapes", "lights", "background", "bboxdeco"), subscene=0)  
  objlist <- vector("list", nrow(objs))
  ids <- objs$id
  types <- as.character(objs$type)
  for (i in seq_len(nrow(objs))) {
    objlist[[i]] <- getObject(ids[i], types[i])
    names(objlist)[i] <- as.character(ids[i])
  }
  result$objects <- objlist
  
  # If there are user callbacks for the mouse, they'll
  # be recorded in rgl.callback.env
  
  devname <- paste0("dev", cur3d())
  callbacks <- rgl.callback.env[[devname]]
  if (!is.null(callbacks)) {
    for (name in names(callbacks)) {
      subscene <- sub("^sub", "", name)
      sub <- findSubscene(result$rootSubscene, subscene)
      if (!is.null(sub)) {
        sub$callbacks <- callbacks[[name]]
        result$rootSubscene <- replaceSubscene(result$rootSubscene, subscene, sub)
      }
      bboxid <- sub("^bbox", "", name)
      bbox <- result$objects[[bboxid]]
      if (!is.null(bbox)) {
        bbox$callbacks <- callbacks[[name]]
        result$objects[[bboxid]] <- bbox
      }
    }
    result$javascript <- callbacks$javascript
  }
    
  class(result) <- "rglscene"
  result
}

print.rglscene <- function(x, ...) {
  cat(gettext("RGL scene containing:\n"))
  if (!is.null(x$par3d))
    cat(gettext("  par3d:\tscene information\n"))
  if (!is.null(x$material))
    cat(gettext("  material:\tdefault material properties\n"))
  if (!is.null(x$objects)) {
    cat(gettextf("  objects:\tlist of %d object(s):\n", length(x$objects)))
    cat("          \t", sapply(x$objects, function(obj) obj$type), "\n")
  }
  if (!is.null(x$root)) 
    cat(gettext("  root:\ta root subscene\n"))
  invisible(x)
}

summary.rglscene <- function(object, ...) {
  result <- list()
  nobjs <- length(object$objects)
  if (nobjs) result$objects <- data.frame(type=sapply(object$objects, function(obj) obj$type))
  if (!is.null(object$rootSubscene))
    result$subscenes <- summary(object$rootSubscene)
  result
}

summary.rglsubscene <- function(object, ...) {
  result <- data.frame(id = object$id, parent = NA, objects = 0)
  result$objects <- list(object$objects)
  if (length(object$subscenes)) {
    children <- do.call(rbind, lapply(object$subscenes, summary))
    children[is.na(children$parent),"parent"] <- object$id
    result <- rbind(result, children)
  }
  result
}

print.rglobject <- function(x, ...) {
  cat(gettextf("RGL object of type %s containing components\n", x$type))
  cat("  ")
  cat(names(x), sep=", ")
  cat("\n")
}

print.rglsubscene <- function(x, ...) {
  cat(gettext("RGL subscene containing components\n"))
  cat("  ")
  cat(names(x), sep=", ")
  cat("\n")
}
  

plot3d.rglscene <- function(x, add=FALSE, open3dParams = getr3dDefaults(), ...) {
  root <- x$rootSubscene
  if (is.null(root)) root <- x  # Should work with pre-subscene objects
  if (!add) {
    args <- list(...)
    params <- open3dParams
    params[names(args)] <- args
    if (!is.null(x$material)) {
      if (is.null(params$material)) params$material <- list()
      params$material[names(x$material)] <- x$material
    }
    if (!is.null(root$bg)) {
      if (is.null(params$bg)) params$bg <- list()
      params$bg[names(params$material)] <- params$material
      params$bg[names(x$bg$material)] <- x$bg$material
      x$bg$material <- x$bg$id <- x$bg$type <- NULL
      params$bg[names(x$bg)] <- x$bg
    }
    if (!is.null(root$par3d)) {
      ind <- !(names(root$par3d) %in% rgl.par3d.readonly)
      params[names(root$par3d)[ind]] <- root$par3d[ind]
    }
    open3d(params = params)
    
    # Some older scenes might not have a light in them, so only clear if one is there
    for (i in seq_along(x$objects)) {
      obj <- x$objects[[i]]
      if (obj$type == "light") {
        clear3d("lights")
        break
      }
    }
  } 
  save <- par3d(skipRedraw = TRUE)
  on.exit(par3d(save))
  
  if (is.null(x$rootSubscene)) {
    results <- NULL
    for (i in seq_along(x$objects)) {
      obj <- x$objects[[i]]
      results <- c(results, plot3d(obj))
    }
    if (!is.null(obj <- x$bbox)) 
      plot3d(obj)  
  } else 
    results <- plot3d(root, x$objects, root = TRUE, ...)
  
  highlevel(results)
}   

plot3d.rglsubscene <- function(x, objects, root = TRUE, ...) {
  if (root) {
    if (!is.null(x$embeddings)) {
      info <- subsceneInfo(embeddings = x$embeddings)
      subscene <- info$id
    } else
      subscene <- currentSubscene3d()
    if (!is.null(x$par3d$viewport))
      par3d(viewport = x$par3d$viewport)
  } else
    subscene <- newSubscene3d(viewport = x$embeddings["viewport"],
			    projection = x$embeddings["projection"],
			    model = x$embeddings["model"],
			    newviewport = x$par3d$viewport,
			    copyLights = FALSE)
			   
  if (!is.null(scale <- x$par3d$scale))
    par3d(scale = scale)
  if (!is.null(userMatrix <- x$par3d$userMatrix))
    par3d(userMatrix = userMatrix)
  listeners <- list(x$par3d$listeners) # list contains old ids
  names(listeners) <- subscene         # names are new ids
    
  results <- subscene
  names(results) <- paste0("subscene", as.character(x$id))
  
  objs <- x$objects
  for (id in as.character(objs)) {
    obj <- objects[[id]]
    if (is.null(obj$newid)) 
      results <- c(results, objects[[id]]$newid <- plot3d(obj, ...))
    else
      addToSubscene3d(obj$newid)
  }
  for (i in seq_along(x$subscenes)) {
    useSubscene3d(subscene)
    res <- plot3d(x$subscenes[[i]], objects, root=FALSE, ...)
    results <- c(results, res$results)
    listeners <- c(listeners, res$listeners)
    objects <- res$objects
  }
  if (root) {
    # Translate all the listener values
    dotranslations <- function(id) {
      info <- subsceneInfo(id = id)
      oldlisteners <- listeners[[as.character(id)]]
      par3d(listeners = results[paste0("subscene", oldlisteners)], subscene = id)
      for (child in info$children)
        dotranslations(child)
    }
    dotranslations(subscene)
    useSubscene3d(subscene)
    return(results)
  } else
    return(list(results=results, objects=objects, listeners=listeners))
}

plot3d.rglobject <- function(x, ...) {
  type <- x$type
  fn <- switch(type,
    points = points3d,
    lines = segments3d,
    linestrip = lines3d,
    triangles = triangles3d,
    quads = quads3d,
    text = texts3d,
    spheres = spheres3d,
    abclines = abclines3d,
    planes = planes3d,
    surface = surface3d,
    sprites = sprites3d,
    light = light3d,
    clipplanes = clipplanes3d,
    NULL)
  if (is.null(fn)) {
    warning(gettextf("Object type '%s' not handled.", type), 
    	    domain = NA)
    return()
  }
  if (!is.null(x$ignoreExtent)) {
    save <- par3d(ignoreExtent = x$ignoreExtent)
    on.exit(par3d(save))
  }
  args <- list()
  args$x <- x$vertices
  args$normals <- x$normals
  args$texcoords <- x$texcoords
  args$texts <- x$texts
  args$cex <- x$cex
  args$adj <- x$adj
  args$radius <- x$radii
  args$d <- x$offsets
  args$indices <- x$indices
  
  switch(type, 
    abclines = {
      odd <- seq_len(nrow(args$x)) %% 2 == 1
      ends <- args$x[odd,,drop=FALSE]
      args$a <- args$x[!odd,,drop=FALSE] - ends
      args$x <- ends
    },
    planes =,
    clipplanes = {
      args$a <- args$normals
      args$x <- NULL
      args$normals <- NULL
    },  
    surface = {
      dim <- x$dim
      args$y <- matrix(args$x[,2], dim[1], dim[2])
      args$z <- matrix(args$x[,3], dim[1], dim[2])
      args$x <- matrix(args$x[,1], dim[1], dim[2])
      if (!is.null(args$normals)) {
        args$normal_x <- matrix(args$normals[,1], dim[1], dim[2])
	args$normal_y <- matrix(args$normals[,2], dim[1], dim[2])
	args$normal_z <- matrix(args$normals[,3], dim[1], dim[2])
        args$normals <- NULL
      }
      if (!is.null(args$texcoords)) {
        args$texture_s <- matrix(args$texcoords[,1], dim[1], dim[2])
        args$texture_t <- matrix(args$texcoords[,2], dim[1], dim[2])
        args$texcoords <- NULL
      }
    },
    sprites = {
      save2 <- par3d(skipRedraw = TRUE)
      on.exit(par3d(save2), add=TRUE)
  
      if (!is.null(x$objects)) {
        ids <- numeric(length(x$objects))
        for (i in seq_along(ids)) 
          ids[i] <- plot3d(x$objects[[i]])
        args$shapes <- ids
      }
      args$userMatrix <- x$usermatrix
    })
      
  mat <- x$material
  if (is.null(mat)) mat <- list()
  if (!is.null(col <- x$colors)) {
    mat$color <- rgb(col[,1], col[,2], col[,3])
    mat$alpha <- col[,4]
  }

  if (type == "light") {
    if (!x$finite) {
      args$x <- NULL
      vx <- x$vertices[1]
      vy <- x$vertices[2]
      vz <- x$vertices[3]
      args$phi <- atan2(vy, sqrt(vx^2 + vz^2))*180/pi
      args$theta <- atan2(vx, vz)*180/pi
    }
    args$viewpoint.rel <- x$viewpoint
    args$ambient <- mat$color[1]
    args$diffuse <- mat$color[2]
    args$specular <- mat$color[3]
  } else 
    args <- c(args, mat)
  
  do.call(fn, args)
}

plot3d.rglbboxdeco <- function(x, ...) {
  args <- list()     
  v <- x$vertices
  t <- x$texts
  m <- x$axes$mode
  if (m[1] == "none")
    args$xat <- numeric()
  else if (m[1] != "pretty") {
    ind <- is.na(v[,2]) & is.na(v[,3])
    if (any(ind)) {
      args$xat <- v[ind,1]
      if (!is.null(t))
        args$xlab <- t[ind]
      else
        args$xlab <- signif(args$xat, 4)
    }
  }
  if (m[2] == "none")
    args$yat <- numeric()
  else if (m[2] != "pretty") {
    ind <- is.na(v[,1]) & is.na(v[,3])
    if (any(ind)) {
      args$yat <- v[ind,2]
      if (!is.null(t))
        args$ylab <- t[ind]
      else
        args$ylab <- signif(args$yat, 4)
    }
  }
  if (m[3] == "none")
    args$zat <- numeric()
  else if (m[3] != "pretty") {
    ind <- is.na(v[,1]) & is.na(v[,2])
    if (any(ind)) {
      args$zat <- v[ind,3]
      if (!is.null(t))
        args$zlab <- t[ind]
      else
        args$zlab <- signif(args$zat, 4)
    }
  }
  args$draw_front <- x$draw_front
  args <- c(args, x$material)
  
  do.call("bbox3d", args)
}

plot3d.rglbackground <- function(x, ...) {
  mat <- x$material 
  if (is.null(mat)) mat <- list()
  if (!is.null(col <- x$colors)) {
    mat$color <- rgb(col[,1], col[,2], col[,3])
    mat$alpha <- col[,4]
  }
  args <- c(list(sphere = x$sphere, fogtype = x$fogtype), mat)
  do.call("bg3d", args)
}

plot3d.rglWebGL <- function(x, ...) {
  plot3d(attr(x, "origScene"), ...)
}

compare_proxy.rglscene <- function(x, path = "x") {
  list(object = old_compare_proxy.rglscene(x),
       path = paste0("compare_proxy(", path, ")"))
}

old_compare_proxy.rglscene <- function(x) {

	doSubscene <- function(obj) {
		if (!is.null(obj$par3d)) {
		  rect <- obj$par3d$windowRect
		  if (!is.null(rect)) {
		    rect <- rect - rect[1:2]
		    obj$par3d$windowRect <- rect
		  }
		}
		ids <<- c(ids, obj$id)
		if (is.list(obj$subscenes))
			obj$subscenes <- lapply(obj$subscenes, doSubscene)
	}

	ids <- c(x$rootSubscene$id,
					 sapply(x$objects, function(obj) obj$id))
	
	x$rootSubscene <- doSubscene(x$rootSubscene)
	
	ids <- unique(ids)
	newids <- ids - min(ids) + 1
	names(newids) <- ids
	
	newid <- function(id) {
		result <- if (is.null(id)) NULL
	            else unname(newids[as.character(id)])
		if (any(is.na(result)))
			stop("id gives NA")
		result
	}
	newidc <- function(id) as.character(newid(id))
	fixvec <- function(vec) {
		if (is.list(vec))
			lapply(vec, fixobj)
	  else
	  	newid(vec)
  }
	fixobj <- function(obj) {
		obj$id <- newid(obj$id)
		obj$objects <- fixvec(obj$objects)
		obj$ids <- fixvec(obj$ids)
		if (!is.null(obj$par3d)) {
			obj$par3d$listeners <- newid(obj$par3d$listeners)
			obj$par3d$fontname <- NULL
			obj$par3d$maxClipPlanes <- NULL
			obj$par3d$glVersion <- NULL
		}
		if (!is.null(obj$material)) {
			obj$material$texture <- NULL
		}
		obj$subscenes <- fixvec(obj$subscenes)
		obj
	}
	x$rootSubscene <- fixobj(x$rootSubscene)

	x$objects <- lapply(x$objects, fixobj)
	names(x$objects) <- newidc(names(x$objects))
	class(x) <- NULL
	x
}

# Compare old and new scenes
all.equal.rglscene <- function(target, current, ...) {
	if (inherits(current, "rglscene")) {
		target <- compare_proxy.rglscene(target)
		current <- compare_proxy.rglscene(current)
		result <- all.equal(target, 
							          current, ...)
		
		if (!isTRUE(result) && isNamespaceLoaded("waldo")) 
			result <- waldo::compare(target, current)
		
		result
	} else
		"'current' is not an rglscene object"
}

as.rglscene <- function(x, ...) {
  UseMethod("as.rglscene")
}

identify3d <- function(x, y = NULL, z = NULL, labels = seq_along(x),
              n = length(x), plot = TRUE, adj = c(-0.1, 0.5),
              tolerance = 20, buttons = c("right", "middle")) {
              
    cat <- function(...) {
      base::cat(...)
      flush.console()
    }
    
    opar <- par3d("mouseMode")
    odisp <- cur3d()
    
    on.exit( {
      disp <- cur3d()
      if (odisp != disp) 
        try(set3d(odisp), silent=TRUE)
      if (cur3d() == odisp) 
        par3d(mouseMode = opar) 
      try(set3d(disp), silent=TRUE)
    } )     
    
    xyz <- xyz.coords(x, y, z)
    x <- xyz$x
    y <- xyz$y
    z <- xyz$z
    if (length(x)==0) 
        return(numeric())
        
    force(labels)
    force(adj)
        
    buttons <- match.arg(buttons, c("left", "right", "middle"), several.ok = TRUE)
    
    if (length(buttons > 1))
      cat(gettextf("Use the %s button to select, the %s button to quit\n", 
      	           buttons[1], buttons[2]))
    else
      cat(gettextf("Use the %s button to select\n"), buttons[1])

    buttons <- c(left=1, right=2, middle=3)[buttons]
    
    selected <- NULL
    
    select <- function(mousex, mousey) {
       disp <- cur3d()
       if (disp != odisp) {
         set3d(odisp)
         on.exit(set3d(disp))
       }
       viewport <- par3d("viewport")
       winxyz <- rgl.user2window(xyz)
       winxyz[,1] <- winxyz[,1]*viewport[3]
       winxyz[,2] <- (1-winxyz[,2])*viewport[4]
      
       dist <- sqrt( (mousex-winxyz[,1])^2 + (mousey - winxyz[,2])^2 )
       dist[winxyz[,3] < 0 | winxyz[,3] > 1] <- Inf
       sel <- which.min(dist)
       if (dist[sel] > tolerance)
       	  cat(gettext("Warning:  no point within tolerance\n"))
       else if (sel %in% selected)
          cat(gettext("Warning:  nearest point already identified\n"))
       else {
          selected <<- c(selected, sel)
          if (plot) 
              text3d(x[sel], y[sel], z[sel], texts=labels[sel], adj=adj)
       }
    }
    
    doquit <- FALSE
    
    quit <- function(mousex, mousey) {
       doquit <<- TRUE
    }
    
    rgl.setMouseCallbacks(buttons[1], begin=select)
    if (length(buttons) > 1)
    	rgl.setMouseCallbacks(buttons[2], begin=quit)
    
    while(!doquit && length(selected) < n) Sys.sleep(0.2)
    selected
}

persp3d <- function(x, ...) UseMethod("persp3d")


persp3d.default <-
function(x = seq(0, 1, length.out = nrow(z)), y = seq(0, 1, length.out = ncol(z)),
    z, xlim = NULL, ylim = NULL, zlim = NULL,
    xlab = NULL, ylab = NULL, zlab = NULL, add = FALSE, aspect = !add, 
    forceClipregion = FALSE, ...) {
    if (!add) next3d()
    skip <- par3d(skipRedraw=TRUE)
    on.exit(par3d(skip))
    
    if (is.null(xlab))
        xlab <- if (!missing(x)) deparse(substitute(x)) else "X"
    if (is.null(ylab))
        ylab <- if (!missing(y)) deparse(substitute(y)) else "Y"
    if (is.null(zlab))
        zlab <- if (!missing(z)) deparse(substitute(z)) else "Z"
    ## labcex is disregarded since we do NOT yet put  ANY labels...
    if (missing(z)) {
        if (!missing(x)) {
            if (is.list(x)) {
                z <- x$z
                y <- x$y
                x <- x$x
            }
            else {
                z <- x
                x <- seq(0, 1, length.out = nrow(z))
            }
        }
        else stop("No 'z' matrix specified")
    }
    else if (is.list(x)) {
        y <- x$y
        x <- x$x
    }
    if ( (!is.matrix(x) && any(diff(x) <= 0)) 
      || (!is.matrix(y) && any(diff(y) <= 0)))
        stop("Increasing 'x' and 'y' values expected")
        
    savesubscene <- currentSubscene3d()
    result <- setClipregion(xlim, ylim, zlim, forceClipregion)
    result <- c(result, surface=surface3d(x,y,z,...))
    useSubscene3d(savesubscene)
    if (!add) {
      result <- c(result, decorate3d(xlim = xlim, ylim = ylim, zlim = zlim, 
                          xlab = xlab, ylab = ylab, zlab = zlab, aspect = aspect, ...))
      highlevel(result)
    } else
      lowlevel(result)
}

setClipregion <- function(xlim = NULL, ylim = NULL, zlim = NULL, force = FALSE) {
  if (force || length(c(xlim, ylim, zlim))) {
    listeners <- par3d("listeners")
    result <- c(clipregion = newSubscene3d("inherit", "inherit", "inherit"))
    par3d(listeners = listeners)
    normals <- matrix(nrow = 0, ncol = 3)
    offsets <- numeric()
    if (length(xlim)) {
      normals <- rbind(normals, matrix(c(1, 0, 0,
                                         -1, 0, 0), nrow = 2, byrow = TRUE))
      offsets <- c(offsets, -xlim[1], xlim[2])
    }
    if (length(ylim)) {
      normals <- rbind(normals, matrix(c(0, 1, 0,
                                         0, -1, 0), nrow = 2, byrow = TRUE))
      offsets <- c(offsets, -ylim[1], ylim[2])
    }
    if (length(zlim)) {
      normals <- rbind(normals, matrix(c(0, 0, 1, 
                                         0, 0, -1), nrow = 2, byrow = TRUE))
      offsets <- c(offsets, -zlim[1], zlim[2])
    }
    keep <- is.finite(offsets)
    if (length(offsets[keep]))
      result <- c(result, clipplanes = clipplanes3d(normals[keep,], d = offsets[keep]))
  } else 
    result <- integer()
  lowlevel(result)
}

persp3d.function <- function(x, xlim = c(0,1), ylim = c(0,1), 
                             slim = NULL, tlim = NULL, n = 101,
                             xvals = seq.int(min(xlim), max(xlim), length.out = n[1]), 
                             yvals = seq.int(min(ylim), max(ylim), length.out = n[2]), 
                             svals = seq.int(min(slim), max(slim), length.out = n[1]), 
                             tvals = seq.int(min(tlim), max(tlim), length.out = n[2]),
                             xlab = NULL, ylab = NULL, zlab = NULL, 
                             col = "gray",
                             otherargs = list(), 
                             normal = NULL, texcoords = NULL, ...) {
  f <- x
  n <- rep(n, length.out = 2)
  parametric <- !is.null(slim) || !is.null(tlim)
  if (!parametric) {
    n1 <- length(xvals)
    n2 <- length(yvals)
    xvals <- matrix(xvals, n1, n2)
    yvals <- matrix(yvals, n1, n2, byrow = TRUE)
    args <- c(list(c(xvals), c(yvals)), otherargs)
    zvals <- do.call(f, args)
    dim(zvals) <- dim(xvals)
    argnames <- names(as.list(f))
    if (is.null(xlab)) xlab <- argnames[1]
    if (is.null(ylab)) ylab <- argnames[2]
    if (is.null(zlab)) zlab <- deparse(substitute(x))
  } else {
    if (is.null(slim)) slim <- c(0,1)
    if (is.null(tlim)) tlim <- c(0,1)
    n1 <- length(svals)
    n2 <- length(tvals)
    svals <- matrix(svals, n1, n2)
    tvals <- matrix(tvals, n1, n2, byrow = TRUE)
    
    args <- c(list(as.numeric(svals), as.numeric(tvals)), otherargs)
    allvals <- do.call(f, args)
    xvals <- matrix(allvals[,1], n1, n2)
    yvals <- matrix(allvals[,2], n1, n2)
    zvals <- matrix(allvals[,3], n1, n2)
    if (!is.null(colnames <- colnames(allvals))) {
      if (is.null(xlab)) xlab <- colnames[1]
      if (is.null(ylab)) ylab <- colnames[2]
      if (is.null(zlab)) zlab <- colnames[3]
    }
  }
  
  if (is.function(col)) {
    zmin <- min(zvals, na.rm = TRUE)
    zscale <- 1/(max(zvals, na.rm = TRUE) - zmin)
    colfn <- colorRamp(col(100))
    colrgba <- colfn(c((zvals - zmin)*zscale))
    colrgba[is.na(colrgba)] <- 0
    col <- rgb(colrgba, maxColorValue = 255)
    dim(col) <- dim(zvals)
  }
  if (is.function(normal)) 
    normal <- do.call(normal, args)	
  if (is.function(texcoords))
    texcoords <- do.call(texcoords, args)
  
  args <- list(xvals, yvals, zvals, col = col, 
               xlab = xlab, ylab = ylab, zlab = zlab, ...)	
  if (is.matrix(normal)) 
    args <- c(args, list(normal_x = matrix(normal[,1], n1, n2),
                         normal_y = matrix(normal[,2], n1, n2),
                         normal_z = matrix(normal[,3], n1, n2)))	
  if (is.matrix(texcoords))
    args <- c(args, list(texture_s = matrix(texcoords[,1], n1, n2),
                         texture_t = matrix(texcoords[,2], n1, n2)))
  if (parametric)
    args <- c(args, list(xlim = if (!missing(xlim)) xlim, 
                         ylim = if (!missing(ylim)) ylim))
  
  do.call(persp3d, args)
}

persp3d.deldir <- function(x, ..., add = FALSE) {
  plot3d(as.mesh3d(x, ...), add = add, ...)
}

persp3d.triSht <- function(x, z, ..., add = FALSE) {
  plot3d(as.mesh3d(x, z, ...), add = add, ...)
}

persp3d.formula <- function(x, data=NULL, xlab = xyz$xlab, ylab = xyz$ylab, zlab = xyz$zlab, ...) {
  checkDeldir(error = TRUE)
  if (!is.null(data))
    environment(x) <- list2env(data, envir = environment(x))
  xyz <- xyz.coords(x, recycle = TRUE)
  dxyz <- with(xyz, deldir::deldir(x, y, z = z, suppressMsge = TRUE)) # nolint
  persp3d(dxyz, xlab = xlab, ylab = ylab, zlab = zlab, ...)
}

library(rgl)
tet <- tetrahedron3d()

open3d()
segments3d(t(tet$vb[1:3,]), indices = c(1,2,1,3,1,4,2,3,2,4,3,4))

open3d()
text3d(t(tet$vb[1:3,]), text=1:4)
triangles3d(t(tet$vb[1:3,]), indices = c(1,2,3,1,4,2,1,3,4,2,4,3),
            col = "red")

# This displayed a triangle for the red quad (issue #154)

quad <- cbind(x = c(-1, 1, 1, -1),
              y = c( 0, 0, 0, 0),
              z = c(-1, -1, 1, 1))/2
open3d()
triangles3d(quad, alpha = 0.5, col = "red", indices=c(1,2,3,1,3,4))
id <- triangles3d(quad+1, col = "blue", indices=c(1,2,3,1,3,4))

# These signalled errors or warnings

tri <- as.triangles3d(id)
m <- as.mesh3d(id)

filename <- tempfile(fileext = ".ply")
writePLY(filename)
unlink(filename)

filename <- tempfile(fileext = ".obj")
writeOBJ(filename, separateObjects = FALSE)
readOBJ(filename)
unlink(filename)

filename <- tempfile(fileext = ".stl")
writeSTL(filename)
readSTL(filename)
unlink(filename)

in_knitr <- function()
  isTRUE(getOption("knitr.in.progress"))

##
## knitr hook functions
##
##

fns <- local({
  newwindowdone <- FALSE
  closewindowsdone <- FALSE
  do_newwindow <- function(options) {
    if (!newwindowdone) {
      newwindow <- options$rgl.newwindow
      if (!is.null(newwindow) && newwindow)
        open3d()
      if (!is.null(options$rgl.keepopen))
        warning("rgl.keepopen has been replaced by rgl.newwindow")
      newwindowdone <<- TRUE
      closewindowsdone <<- FALSE
    }
  }
  do_closewindows <- function(options) {
    if (!closewindowsdone) {
      closewindows <- options$rgl.closewindows
      if (!is.null(closewindows) && closewindows)
        while (cur3d())
          close3d()
      newwindowdone <<- FALSE
      closewindowsdone <<- TRUE
    }
  }
  list(do_newwindow = do_newwindow,
       do_closewindows = do_closewindows)
})

do_newwindow <- fns[["do_newwindow"]]
do_closewindows <- fns[["do_closewindows"]]
rm(fns)

hook_webgl <- function(before, options, envir) {
  if (before) {
    do_newwindow(options)
    return()
  }
  res <- if (cur3d() != 0) {
    out_type <- opts_knit$get("out.format")
    if (!length(intersect(out_type, c("markdown", "html"))))
      stop("'hook_webgl' is for HTML only. ",
           "Use 'hook_rgl' instead, or 'setupKnitr(autoprint = TRUE)'.")
    knit_print(rglwidget(), options = options)
  }
  do_closewindows(options)
  m <- attr(res, 'knit_meta')
  knit_meta_add(m, if (missing(options)) '' else options$label)
  res
}

hook_rgl <- function(before, options, envir) {
  if (before) {
    do_newwindow(options)
    return()
  }
  res <- if (cur3d() != 0) {
    name <- fig_path("", options)
    margin <- options$rgl.margin
    if (is.null(margin)) margin <- 100
    par3d(windowRect = margin + options$dpi * 
                                c(0, 0, options$fig.width, options$fig.height))
    Sys.sleep(.05) # need time to respond to window size change

    dir <- knitr::opts_knit$get("base_dir")
    if (is.character(dir)) {
      if (!file_test("-d", dir)) dir.create(dir, recursive = TRUE)
      owd <- setwd(dir)
      on.exit(setwd(owd))
    }
    save_rgl(name, options$dev)
    options$fig.num <- 1L  # only one figure in total
    options$dev <- "png"
    hook_plot_custom(before, options, envir)
  }
  do_closewindows(options)
  res
}

hook_rglchunk <- function(before, options, envir) {
  if (before)
    do_newwindow(options)
  else if (is.null(options$webgl) && is.null(options$rgl))
    do_closewindows(options)
}

save_rgl <- function(name, devices) {
  if (!file_test("-d", dirname(name))) 
    dir.create(dirname(name), recursive = TRUE)
  # support 3 formats: eps, pdf and png (default)
  for (dev in devices) switch(
    dev,
    eps = ,
    postscript = rgl.postscript(paste0(name, ".eps"), fmt = "eps"),
    pdf = rgl.postscript(paste0(name, '.pdf'), fmt = "pdf"),
    snapshot3d(paste0(name, ".png"))
  )
}


# This code is mostly taken from Shiny, which has lots of authors:  see
# https://github.com/rstudio/shiny/blob/master/R/utils.R

# Evaluate an expression using our own private stream of
# randomness (not affected by set.seed).
withPrivateSeed <- local({

  ownSeed <- NULL

  function(expr) {
    # Save the old seed if present.
    if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
      hasOrigSeed <- TRUE
      origSeed <- .GlobalEnv$.Random.seed
    } else {
      hasOrigSeed <- FALSE
    }

    # Swap in the private seed.
    if (is.null(ownSeed)) {
      if (hasOrigSeed) {
        # Move old seed out of the way if present.
        rm(.Random.seed, envir = .GlobalEnv, inherits = FALSE)
      }
    } else {
      .GlobalEnv$.Random.seed <- ownSeed
    }

    # On exit, save the modified private seed, and put the old seed back.
    on.exit({
      ownSeed <<- .GlobalEnv$.Random.seed

      if (hasOrigSeed) {
        .GlobalEnv$.Random.seed <- origSeed
      } else {
        rm(.Random.seed, envir = .GlobalEnv, inherits = FALSE)
      }
      # Shiny had this workaround.  I think we don't need it, and have
      # commented it out.
      # Need to call this to make sure that the value of .Random.seed gets put
      # into R's internal RNG state. (Issue #1763)

      # httpuv::getRNGState() # nolint
    })
    
    expr
  }
})

# Version of sample that runs with private seed
p_sample <- function(...) {
  withPrivateSeed(sample(...))
}

fns <- local({
  saveopts <- NULL
  plotnum <- 0
  setupDone <- NULL
  latex <- FALSE
  
  counter <- 0L
  rgl_counter <- function() {
    counter <<- counter + 1L
    counter
  }
  
  setupKnitr <- function(autoprint = FALSE, 
                         rgl.newwindow = autoprint,
                         rgl.closewindows = autoprint) {
    if (!is.null(setupDone)) {
      if (setupDone$autoprint) {
        message("undoing setup")
        options(saveopts)
        saveopts <<- NULL
        setupDone <<- NULL
      }
    }

      if (!is.null(setupDone)) {
        if (setupDone$autoprint != autoprint ||
            setupDone$rgl.newwindow != rgl.newwindow ||
            setupDone$rgl.closewindows != rgl.closewindows) {
          warning("Already set autoprint = ", setupDone$autoprint,
                  ", rgl.newwindow = ", setupDone$rgl.newwindow,
                  ", rgl.closewindows = ", setupDone$rgl.closewindows)
        }
        return()
      }
      counter <<- 0L

    setupDone <<- list(autoprint = autoprint,
                       rgl.newwindow = rgl.newwindow,
                       rgl.closewindows = rgl.closewindows)
    
    # R produces multiple vignettes in the same session.
    knitr::opts_chunk$set(rgl.newwindow = rgl.newwindow, 
                          rgl.closewindows = rgl.closewindows,
                          rgl.chunk = TRUE)
    
    knit_hooks$set(webgl = hook_webgl)
    knit_hooks$set(webGL = hook_webgl)
    knit_hooks$set(rgl = hook_rgl)
    knit_hooks$set(rgl.chunk = hook_rglchunk)
    latex <<- identical(opts_knit$get("out.format"), "latex") || identical(opts_knit$get("rmarkdown.pandoc.to"), "latex")
    if (autoprint)
      saveopts <<- options(rgl.printRglwidget = TRUE)
  }  
  
  knit_print.rglOpen3d <- function(x, options, ...) {
    print(x, ...)
    if (getOption("rgl.printRglwidget", FALSE)) {
      plotnum <<- plotnum + 1
    }
    invisible(x)
  }
  
  knit_print.rglId <- function(x, options, ...) {
    if (getOption("rgl.printRglwidget", FALSE) && !par3d("skipRedraw"))	{
      scene <- scene3d()
      args <- list(...)
      if (inherits(x, "rglHighlevel"))
        plotnum <<- plotnum + 1
      structure(list(plotnum = plotnum,
                     scene = scene,
                     width = figWidth(),
                     height = figHeight(),
                     options = options, args = args),
                class = c("rglRecordedplot", "knit_other_plot"))
    } else
      invisible(x)
  }
    
    sew.rglRecordedplot <- function(x, options = list(), ...)  {
      latex <- identical(opts_knit$get("out.format"), "latex") || identical(opts_knit$get("rmarkdown.pandoc.to"), "latex")
      scene <- x$scene
      doSnapshot <- knitrNeedsSnapshot(options)
      content <- rglwidget(scene,
                           width = x$width,
                           height = x$height,
                           webgl = !doSnapshot)
      if (inherits(content, "knit_image_paths")) {
        # # We've done a snapshot, put it in the right place.
        name <- fig_path("-rgl.png", options, rgl_counter())
        if (!file_test("-d", dirname(name)))
          dir.create(dirname(name), recursive = TRUE)
        file.copy(content, name, overwrite = TRUE)
        unlink(content)
        content <- structure(list(file = name,
                                  extension = "png"),
                             class = "html_screenshot")
      }
      fig.align <- options$fig.align
      if (length(fig.align) ==  1 && fig.align != "default")
        content <- prependContent(content,
                                  tags$style(sprintf(
                                    "#%s {%s}",
                                    content$elementId,
                                    switch(fig.align,
                                           center = "margin:auto;",
                                           left   = "margin-left:0;margin-right:auto;",
                                           right  = "margin-left:auto;margin-right:0;",
                                           ""))))
      result <- do.call("knit_print", c(list(content, options), x$args))
      if (!latex)
        class(result) <- c(class(result), "knit_asis_htmlwidget")
      
      sew(result, options)
    }
    
    list(
      setupKnitr = setupKnitr,
      knit_print.rglOpen3d = knit_print.rglOpen3d,
      knit_print.rglId = knit_print.rglId,
      sew.rglRecordedplot = sew.rglRecordedplot)
})

setupKnitr <- fns[["setupKnitr"]]
knit_print.rglId <- fns[["knit_print.rglId"]]
knit_print.rglOpen3d <- fns[["knit_print.rglOpen3d"]]
sew.rglRecordedplot <- fns[["sew.rglRecordedplot"]]

rm(fns)

is_low_change.rglRecordedplot <- function(p1, p2) {
  inherits(p2, "rglRecordedplot") && p1$plotnum == p2$plotnum
}

figWidth <- function() {
  if (in_pkgdown_example())
    return(pkgdown_dims()$width)
  else if (in_knitr())
    opts <- opts_current$get(c("fig.width", "dpi", "fig.retina"))
  else
    opts <- NULL
  if (length(opts)) {
    result <- with(opts, fig.width*dpi/fig.retina)
    result[1] 
  } else NULL
}

figHeight <- function() {
  if (in_pkgdown_example())
    return(pkgdown_dims()$height)
  else if (in_knitr())
    opts <- opts_current$get(c("fig.height", "dpi", "fig.retina"))
  else
    opts <- NULL
  if (length(opts)) {
    result <- with(opts, fig.height*dpi/fig.retina)
    result[1] 
  } else NULL
}

# enumerations

rgl.enum <- function( name, ..., multi = FALSE) {
  choices <- list( ... )
  names   <- attr(choices,"names")

  if (multi) pos <- pmatch( name, c(names, "all") )
  else pos <- pmatch( name, names )
 
  max <- length(names)

  if ( any( is.na(pos) ) )
    stop(gettextf("Symbolic value must be chosen from: %s", list(names)), domain = NA)
  else if ( (max+1) %in% pos )
    pos <- seq_along(names)
    
  id  <- unlist(choices[pos])
  
  if ( length(id) > 1 && !multi )
    stop("Multiple choices not allowed")
 
  return( id )
}

rgl.enum.nodetype <- function(type) 
rgl.enum( type, shapes=1, lights=2, bboxdeco=3, userviewpoint=4, material=5, background=6, subscene=7, modelviewpoint=8, 
          multi = TRUE )

rgl.enum.attribtype <- function(attrib)
rgl.enum( attrib, vertices=1, normals=2, colors=3, texcoords=4, dim=5, 
          texts=6, cex=7, adj=8, radii=9, centers=10, ids=11, 
          usermatrix=12, types=13, flags=14, offsets=15,
	        family=16, font=17, pos=18, fogscale=19, axes=20,
          indices=21, shapenum=22)

rgl.enum.pixfmt <- function(fmt)
rgl.enum( fmt, png=0 )

rgl.enum.polymode <- function(mode)
rgl.enum( mode, filled=1, lines=2, points=3, culled=4)

rgl.enum.textype <- function(textype)
rgl.enum( textype, alpha=1, luminance=2, luminance.alpha=3, rgb=4, rgba=5 )

rgl.enum.fogtype <- function(fogtype)
rgl.enum(fogtype, none=1, linear=2, exp=3, exp2=4)

rgl.enum.primtype <- function(primtype)
rgl.enum( primtype, points=1, lines=2, triangles=3, quadrangles=4, linestrips=5 )
  
rgl.enum.texminfilter <- function(minfiltertype)
rgl.enum(minfiltertype, nearest=0, linear=1, nearest.mipmap.nearest=2, nearest.mipmap.linear=3, linear.mipmap.nearest=4, linear.mipmap.linear=5)
  
rgl.enum.texmagfilter <- function(magfiltertype)
rgl.enum(magfiltertype, nearest=0, linear=1)

rgl.enum.gl2ps <- function(postscripttype)
rgl.enum(postscripttype, ps=0, eps=1, tex=2, pdf=3, svg=4, pgf=5)

rgl.enum.pixelcomponent <- function(component)
rgl.enum(component, red=0, green=1, blue=2, alpha=3, depth=4, luminance=5)

rgl.enum.depthtest <- function(depthtest)
rgl.enum(depthtest, never=0, less=1, equal=2, lequal=3, greater=4, 
                    notequal=5, gequal=6, always= 7)

rgl.enum.blend <- function(blend)
rgl.enum(blend, zero=0, one=1, 
         src_color=2, one_minus_src_color=3, 
         dst_color=4, one_minus_dst_color=5,
         src_alpha=6, one_minus_src_alpha=7,
         dst_alpha=8, one_minus_dst_alpha=9,
         constant_color=10, one_minus_constant_color=11,
         constant_alpha=12, one_minus_constant_alpha=13,
         src_alpha_saturate=14)

rgl.enum.texmode <- function(texmode)
  rgl.enum(texmode, replace = 0, modulate = 1, 
           decal = 2, blend = 3, add = 4)

pointInPoly <- function(poly, pt) {
  # polygon is 2 x n, columns are vertices
  # point is 2 vector
  n <- ncol(poly)
  i1 <- seq_len(n)
  i2 <- i1 %% n + 1
  x <- poly[1,i1] + (poly[1,i2] - poly[1,i1])*(pt[2] - poly[2,i1])/(poly[2,i2] - poly[2,i1])
  crossings <- ((poly[2,i1] < pt[2]) & (pt[2] <= poly[2,i2]) 
              | (poly[2,i2] < pt[2]) & (pt[2] <= poly[2,i1])) & pt[1] < x
  sum(crossings) %% 2 == 1
}
	
intersectSegSeg <- function(seg1,seg2) {
  # do segments intersect?
  # both segments have endpoints as columns
  coeffs <- try(solve(cbind(seg1[,2]-seg1[,1], seg2[,1]-seg2[,2]), seg2[,1]-seg1[,1]), silent=TRUE)
  if (inherits(coeffs, "try-error")) return(FALSE)
  all(zapsmall(coeffs) >= 0) && all(zapsmall(1-coeffs) >= 0)
}
  
intersectTriSeg <- function(tri, seg) {
  # intersect a triangle with a segment
  # tri is 2 x 3, columns are vertices
  # seg is 2 x 2, columns are endpoints
  coeffs <- try(solve(rbind(tri,1), rbind(seg,1)), silent=TRUE)
  if (inherits(coeffs, "try-error")) return(TRUE)
  coeffs <- zapsmall(coeffs)  
  if (any(apply(coeffs <= 0, 1, all))) return(FALSE)
  if (any(apply(coeffs > 0, 2, all))) return(TRUE)
  up <- coeffs[,1] < 0
  dn <- coeffs[,2] < 0  
  lb <- max( -coeffs[up,1]/(coeffs[up,2]-coeffs[up,1]) )
  ub <- 1 - max( -coeffs[dn,2]/(coeffs[dn,1] - coeffs[dn,2]) )
  lb <= ub
}

triangulate <- function(x, y = NULL, z = NULL, random = TRUE, plot = FALSE, partial = NA) {
  xyz <- xyz.coords(x, y, z)
  if (xyz$xlab == "Index" && is.null(z) && (is.null(ncol(x)) || ncol(x) == 2L)) {
    x <- xyz$y
    y <- xyz$z
  } else {
    x <- xyz$x
    y <- xyz$y
    if (!diff(range(x, na.rm = TRUE))) 
      x <- xyz$z
    else if (!diff(range(y, na.rm = TRUE))) 
      y <- xyz$z
  } 
  
  nesting <- nestPolys(x, y)
  verts <- nesting$verts
  
  processInside <- function(v) {
    result <- matrix(NA, ncol = 0, nrow = 3)
    indices <- verts[[v]]
    for (i in nesting$nesting[[v]]) {
      result <- cbind(result, processOutside(i))
      indices <- c(indices, NA, verts[[i]])
      
    }
    res0 <- .Call(rgl_earcut, x[indices], y[indices])
    result <- cbind(result, 
                matrix(indices[res0+1], nrow = 3))
  }
  
  processOutside <- function(fwd) {
    result <- matrix(NA, ncol = 0, nrow = 3)
    for (i in nesting$nesting[[fwd]])
      result <- cbind(result, processInside(i))
    
    result
  }
  
  # Done all polys, now combine
  res <- matrix(nrow=3, ncol=0)
  for (i in nesting$toplevel)
    res <- cbind(res, processInside(i))
  
  # Get vertex order
  nextvert <- rep(NA, length(x))
  for (i in seq_along(verts)) {
    poly <- verts[[i]]
    first <- poly[1]
    second <- poly[2]

    # Find first triangle holding first 
    # and second
    tri <- intersect(col(res)[res == first],
                     col(res)[res == second])
    if (!length(tri))
      warning("edge not found:", first, " ", second)
    else {
      tri <- tri[1]
      counter <- (which(res[,tri] == first) - which(res[,tri] == second) + 3) %% 3 == 2
      if (counter) {
        nextvert[poly[-length(poly)]] <- poly[-1]
        nextvert[poly[length(poly)]] <- poly[1]
      } else {
        nextvert[poly[-1]] <- poly[-length(poly)]
        nextvert[poly[1]] <- poly[length(poly)]
      }
    }
  }
  if (plot) {
    for (i in seq_len(ncol(res)))
      polygon(x[res[,i]], y[res[,i]], col = i)
  }
  attr(res, "nextvert") <- nextvert
  res
}

# Rewrite a complex polygon as a list of the individual parts, oriented correctly,
# with attribute showing nesting

nestPolys <- function(x,y = NULL) {
  xy <- xy.coords(x, y)
  x <- xy$x
  y <- xy$y
  n <- length(x)
  nas <- c(which(is.na(x) | is.na(y)), n + 1L)
  prev <- 0L
  verts <- list()
  for (i in seq_along(nas)) {
    verts[[i]] <- (prev + 1L):(nas[i] - 1L)
    prev <- nas[i]
  }
  # nesting is a list of vectors
  # of poly numbers that are directly nested within the corresponding element of verts
  # The last one at length(verts)+1 lists polys not nested anywhere
  
  nesting <- rep(list(integer()), length(verts)+1)
  place <- function(new, toplevel) {
    placed <- FALSE
    contains <- integer()
    if (length(nesting[[toplevel]])) {
      newverts <- rbind(x[verts[[new]]], y[verts[[new]]])
      
      for (j in nesting[[toplevel]]) {
        prev <- rbind(x[verts[[j]]], y[verts[[j]]])    
        if (pointInPoly(prev, newverts[,1])) {
          place(new, j)
          placed <- TRUE
          break
        }
        if (pointInPoly(newverts, prev[,1]))
          contains <- c(contains, j)
      }
    }
    if (!placed) {
      nesting[[toplevel]] <<- c(setdiff(nesting[[toplevel]], contains), new)
      nesting[[new]] <<- contains
    }
  }
  
  for (i in seq_along(verts)) {
    place(i, length(verts)+1)
  }
  
  list(verts=verts, nesting=nesting[-length(nesting)], 
       toplevel=nesting[length(nesting)])
} 

extrude3d <- function(x,y = NULL, thickness=1, smooth=FALSE, ...) {
  xy <- xy.coords(x, y)
  x <- xy$x
  y <- xy$y
  it <- triangulate(x, y)
  nextvert <- attr(it, "nextvert")
  n <- length(x)
  res <- tmesh3d(rbind(c(x,x), c(y,y), c(rep(thickness,n), rep(0,n)), 1),
                 cbind(it, it[c(1,3,2),]+n), ...)
  i1 <- seq_len(n)
  i2 <- nextvert
  i3 <- i2 + n
  i4 <- i1 + n
  keep <- !is.na(nextvert)
  res$ib <- rbind(i4,i3,i2,i1)[,keep]
  if (smooth) {
    res$ib <- res$ib + ncol(res$vb)
    res$vb <- cbind(res$vb, res$vb)
    i3 <- nextvert[nextvert]
    diff <- cbind(x[i3] - x[i1],  y[i3] - y[i1])
    len <- sqrt(apply(diff^2, 1, sum))
    diff <- diff/len
    
    res$normals <- cbind( rbind(0,0,c(rep(1, n), rep(-1, n))) )
    res$normals <- cbind(res$normals, res$normals)
    i2 <- c(i2 + 2*n, i2 + 3*n)
    keep <- !is.na(i2)
    res$normals[,i2[keep]] <- rbind(rep(diff[,2], 2), -rep(diff[,1], 2), 0)[,keep]   
  }
  res
}

polygon3d <- function(x, y = NULL, z = NULL, fill = TRUE, plot = TRUE, 
                      coords, random = TRUE, ...) {
  xyz <- xyz.coords(x,y,z, recycle = TRUE)
  if (!fill) {
    n <- length(xyz$x)
    nas <- with(xyz, c(which(is.na(x) | is.na(y) | is.na(z)), n + 1L))
    prev <- 0L
    loop <- integer()
    for (i in seq_along(nas)) {
      loop <- c(loop, if (i > 1) NA, (prev + 1L):(nas[i] - 1L), prev + 1L)
      prev <- nas[i]
    }
    res <- cbind(xyz$x[loop], xyz$y[loop], xyz$z[loop])
    if (plot)
      lines3d(res, ...)
    else
      res
  } else {
    if (missing(coords)) 
      tri <- triangulate(xyz)
    else {
      cnames <- c("x", "y", "z")
      x <- xyz[[cnames[coords[1]]]]
      y <- xyz[[cnames[coords[2]]]]
      tri <- triangulate(x, y)
    }
    shape <- tmesh3d(rbind(xyz$x, xyz$y, xyz$z, 1), indices = tri)
    if (plot)
      shade3d(shape, ...)
    else
      shape
  }
}  

thigmophobe3d <- function(x, y = NULL, z = NULL, 
                          P = par3d("projMatrix"),
                          M = par3d("modelMatrix"),
                          windowRect = par3d("windowRect")) {
  if (!requireNamespace("plotrix", quietly = TRUE) || 
      packageVersion("plotrix") < "3.7-3") 
    stop("This function requires the plotrix package, version 3.7-3 or higher.")
  
  xyz <- xyz.coords(x, y, z)
  
  pts3d <- rbind(xyz$x, xyz$y, xyz$z, 1)
  pts2d <- asEuclidean(t(P %*% M %*% pts3d))
  w <- diff(windowRect[c(1,3)])
  h <- diff(windowRect[c(2,4)])
  pts2d <- cbind(w*pts2d[,1], h*pts2d[,2])
  
  if (packageVersion("plotrix") < "3.7.6")
    plotrix::thigmophobe(pts2d, plot.span = c(-w, w, -h, h),
                         xlog = FALSE, ylog = FALSE)  
  else
    plotrix::thigmophobe(pts2d, usr = c(-w, w, -h, h),
                     xlog = FALSE, ylog = FALSE,
                     pin = c(w, h)/50) # This doesn't seem to matter...
}

shapelist3d <- function(shapes,x=0,y=NULL,z=NULL,size=1,matrix=NULL,override=TRUE, ..., plot=TRUE) {
  # This function gets an element with recycling
  e <- function(x, i) x[[ (i-1) %% length(x) + 1 ]]
  
  xyz <- xyz.coords(x, y, z, recycle = TRUE)
  x <- xyz$x
  y <- xyz$y
  if (length(y) == 0) y <- 0
  z <- xyz$z
  if (length(z) == 0) z <- 0
  
  if (inherits(shapes, "shape3d")) shapes <- list(shapes)
  
  material <- list(...)
  
  if (!is.null(matrix)) {
    if (!is.list(matrix)) matrix <- list(matrix)
    len <- length(matrix)
  } else len <- 0
  
  len <- max(len, length(x), length(shapes), length(size), length(override))
  if (length(material)) len <- max(len, sapply(material, length))
  
  result <- vector("list", len)
  class(result) <- c("shapelist3d", "shape3d")
  
  for (i in seq_len(len)) {
    if (is.null(matrix)) this <- e(shapes, i)
    else this <- rotate3d(e(shapes,i), matrix=e(matrix,i))
    thissize <- e(size, i)
    this <- translate3d(scale3d(this, thissize, thissize, thissize), e(x,i), e(y,i), e(z,i))
    thismaterial <- lapply(material, function(item) e(item,i))
    if (!e(override,i))
      thismaterial[names(this$material)] <- this$material
    this$material[names(thismaterial)] <- thismaterial
    result[[i]] <- this
  }
  if (plot) {
    shade3d(result)
    lowlevel(result)
  } else
    invisible(result)
}

dot3d.shapelist3d <- function(x, override = TRUE, ...) {
  .check3d()
  save <- par3d(skipRedraw = TRUE)
  on.exit(par3d(save))
  
  lowlevel(unlist(sapply( x, function(item) dot3d(item, override=override, ...) ) ) )
}

wire3d.shapelist3d <- function(x, override = TRUE, ...) {
  .check3d()
  save <- par3d(skipRedraw = TRUE)
  on.exit(par3d(save))
  
  lowlevel(unlist(sapply( x, function(item) wire3d(item, override=override, ...) ) ) )
}


shade3d.shapelist3d <- function(x, override = TRUE, ...) {
  .check3d()
  save <- par3d(skipRedraw = TRUE)
  on.exit(par3d(save))
  
  lowlevel(unlist(sapply( x, function(item) shade3d(item, override=override, ...) ) ) )
}

translate3d.shapelist3d <- function( obj, x, y, z, ... ) {
  structure(lapply( obj, function(item) translate3d(item, x, y, z, ...) ),
            class = class(obj))
}  

rotate3d.shapelist3d <- function( obj,angle,x,y,z,matrix, ... ) {
  structure(lapply( obj, function(item) rotate3d(item,angle, x,y,z,matrix,...) ),
            class = class(obj))
}  

scale3d.shapelist3d <- function( obj, x, y, z, ... ) {
  structure(lapply( obj, function(item) scale3d(item, x,y,z,...) ),
            class = class(obj))
}

addNormals.shapelist3d <- function( x, ... ) {
  structure(lapply( x, function(item) addNormals(item, ...) ),
            class = class(x))
}

print.shapelist3d <- function(x, prefix = "", ...) {
  cat(prefix, " shapelist3d object with ", length(x), " items:\n", sep = "")
  for (i in seq_along(x))
    print(x[[i]], prefix = paste0(prefix, "[[", i, "]]"))
}

tkpar3dsave <- function(params = c("userMatrix", "scale", "zoom", "FOV"), 
                        times = FALSE, 
                        dev = cur3d(),
                        ...) {
  
  if (!requireNamespace("tcltk", quietly = TRUE))
    stop("This function requires 'tcltk'")
  
  results <- list()
  for (n in params) results[[n]] <- list()
  if (times) {
    start <- proc.time()[3]
    results$times <- numeric(0)
  }
  
  RecordParms <- function() {
    values <- par3d(params)
    if (length(params) == 1) {
      values <- list(values)
      names(values) <- params
    }
    for (n in params) results[[n]] <<- c(results[[n]], list(values[[n]]))
    if (times) results$times <<- c(results$times, proc.time()[3] - start)
  }	
  base <- tcltk::tktoplevel(...)
  tcltk::tkwm.title(base, "par3d")
  
  text <- tcltk::tklabel(base, text="Click on Record to save par3d parameters.",
                         justify="left",
                         wraplength="2i")
  frame <- tcltk::tkframe(base)	
  save <- tcltk::tkbutton(frame, text="Record", command=RecordParms)		
  
  quit <- tcltk::tkbutton(frame,text="Quit", command=function()tcltk::tkdestroy(base))
  
  tcltk::tkpack(save, quit, side="left")
  tcltk::tkpack(text, frame)
  
  tcltk::tkwait.window(base)
  
  results
}

library(rgl)
mfrow3d(2,2, sharedMouse = TRUE)
sprites3d( rnorm(10) + 6, rnorm(10), rnorm(10), shape = shade3d(tetrahedron3d(), col = "red"), rotating = FALSE, fixedSize = FALSE)
next3d()
sprites3d( rnorm(10) + 6, rnorm(10), rnorm(10), shape = shade3d(tetrahedron3d(), col = "red"), rotating = FALSE, fixedSize = TRUE)
next3d()
sprites3d( rnorm(10) + 6, rnorm(10), rnorm(10), shape = shade3d(tetrahedron3d(), col = "red"), rotating = TRUE, fixedSize = FALSE)
next3d()
sprites3d( rnorm(10) + 6, rnorm(10), rnorm(10), shape = shade3d(tetrahedron3d(), col = "red"), rotating = TRUE, fixedSize = TRUE)
rglwidget()

setUserShaders <- function(ids, vertexShader = NULL, fragmentShader = NULL,
			   attributes = NULL, uniforms = NULL, textures = NULL,
			   scene = scene3d(minimal),
			   minimal = TRUE) {
  stopifnot(inherits(scene, "rglscene"))
  for (i in ids) {
    id <- as.character(i)
    obj <- scene$objects[[id]]
    if (!is.null(vertexShader))
      obj$userVertexShader <- paste(vertexShader, collapse = "\n")
    if (!is.null(fragmentShader))
      obj$userFragmentShader <- paste(fragmentShader, collapse = "\n")
    obj$userAttributes <- attributes
    obj$userUniforms <- uniforms
    obj$userTextures <- textures
    alldata <- c(obj$userAttributes, obj$userUniforms, obj$userTextures)
    allnames <- names(alldata)
    if (length(allnames) != length(alldata) ||
        any(nchar(allnames) == 0) ||
        any(duplicated(allnames)))
      stop("attributes, uniforms and textures should be named")
    scene$objects[[id]] <- obj
  }
  scene
}

getShaders <- function(id, scene = scene3d(minimal), minimal = TRUE) {
  obj <- scene$objects[[as.character(id)]]
  vertexShader <- obj$userVertexShader
  fragmentShader <- obj$userFragmentShader
  if (is.null(vertexShader) || is.null(fragmentShader)) {
    if (!requireNamespace("V8"))
      stop("This function requires the V8 package.")
    scene <- convertScene(scene)
    obj <- scene$objects[[as.character(id)]]
    ctx <- V8::v8()
    ctx$source(system.file("htmlwidgets/lib/rglClass/rglClass.src.js", package="rgl"))
    ctx$source(system.file("htmlwidgets/lib/rglClass/utils.src.js", package="rgl"))
    ctx$source(system.file("htmlwidgets/lib/rglClass/shaders.src.js", package="rgl"))
  
    pointSize <- obj$material$size
    if (is.null(pointSize))
      pointSize <- scene$material$size
  
    antialias <- obj$material$point_antialias
    if (is.null(antialias))
      antialias <- scene$material$point_antialias
  
    textype <- obj$material$textype
    if (is.null(textype))
      textype <- scene$material$textype
  
    nclipplanes <- 0
    nlights <- 0
    for (i in seq_along(scene$objects)) {
      nclipplanes <- nclipplanes + (scene$objects[[i]]$type == "clipplanes")
      nlights <- nlights + (scene$objects[[i]]$type == "light")
    }
    if (is.null(vertexShader))
      vertexShader <- readLines(system.file("htmlwidgets/lib/rglClass/shaders/rgl_vertex.glsl", package = "rgl"))
    
    if (is.null(fragmentShader))
      fragmentShader <- readLines(system.file("htmlwidgets/lib/rglClass/shaders/rgl_fragment.glsl", package = "rgl"))
  }
  
  defines <- ctx$eval(subst(
    'rglwidgetClass.getDefines(%id%, "%type%", %flags%, %nclipplanes%, %nlights%, %normals%, %pointSize%, "%textype%", %antialias%)', 
    id = id, 
    type = obj$type, 
    flags = obj$flags, 
    nclipplanes = nclipplanes, 
    nlights = nlights, 
    normals = if (is.null(obj$normals)) "undefined" else 1,
    pointSize = pointSize,
    textype = textype, 
    antialias = tolower(antialias)
  ))
  
  structure(list(vertexShader = vertexShader,
            fragmentShader = fragmentShader,
            defines = defines), class = "rglshaders")
}

print.rglshaders <- function(x, ...) {
  cat(x$defines, sep = "\n")
  cat(x$vertexShader, sep = "\n")
  cat("\n")
  cat(x$defines, sep = "\n" )
  cat(x$fragmentShader, sep = "\n")
  invisible(x)
}

library(rgl)
open3d()
set.seed(3)
 x <- cube3d(col="white", texture = system.file("textures/rgl2.png",
                                                package = "rgl"),
             texcoords = matrix(runif(16), ncol=2))
id <- shade3d(x)
# rglwidget()



vShader <- "
/* ****** quads object 13 vertex shader ****** */
  #ifdef GL_ES
  #ifdef GL_FRAGMENT_PRECISION_HIGH
  precision highp float;
#else
precision mediump float;
#endif
#endif
attribute vec3 aPos;
attribute vec4 aCol;
uniform mat4 mvMatrix;
uniform mat4 prMatrix;
varying vec4 vCol;
varying vec4 vPosition;
attribute vec3 aNorm;
uniform mat4 normMatrix;
varying vec4 vNormal;
attribute vec2 aTexcoord;
varying vec2 vTexcoord;
void main(void) {
  vPosition = mvMatrix * vec4(aPos, 1.);
  gl_Position = prMatrix * vPosition;
  vCol = aCol;
  vNormal = normMatrix * vec4(-aNorm, dot(aNorm, aPos));
  vNormal = vec4(normalize(vNormal.xyz/vNormal.w), 1);
  vTexcoord = aTexcoord;
}
"
fShader <- "
/* ****** quads object 13 fragment shader ****** */
  #ifdef GL_ES
  #ifdef GL_FRAGMENT_PRECISION_HIGH
  precision highp float;
#else
precision mediump float;
#endif
#endif
varying vec4 vCol; // carries alpha
varying vec4 vPosition;
varying vec2 vTexcoord;
uniform sampler2D uSampler;
uniform sampler2D userSampler;
uniform int uFogMode;
uniform vec3 uFogColor;
uniform vec4 uFogParms;
varying vec4 vNormal;
uniform mat4 mvMatrix;
uniform vec3 emission;
uniform float shininess;
uniform vec3 ambient0;
uniform vec3 specular0; // light*material
uniform vec3 diffuse0;
uniform vec3 lightDir0;
uniform bool viewpoint0;
uniform bool finite0;
void main(void) {
  vec4 fragColor;
  vec3 n = normalize(vNormal.xyz + texture2D(userSampler, vTexcoord).rgb);
  vec3 eye = normalize(-vPosition.xyz/vPosition.w);
  vec3 lightdir;
  vec4 colDiff;
  vec3 halfVec;
  vec4 lighteffect = vec4(emission, 0.);
  vec3 col;
  float nDotL;
  n = -faceforward(n, n, eye);
  colDiff = vec4(vCol.rgb * diffuse0, vCol.a);
  lightdir = lightDir0;
  if (!viewpoint0)
    lightdir = (mvMatrix * vec4(lightdir, 1.)).xyz;
  if (!finite0) {
    halfVec = normalize(lightdir + eye);
  } else {
    lightdir = normalize(lightdir - vPosition.xyz/vPosition.w);
    halfVec = normalize(lightdir + eye);
  }
  col = ambient0;
  nDotL = dot(n, lightdir);
  col = col + max(nDotL, 0.) * colDiff.rgb;
  col = col + pow(max(dot(halfVec, n), 0.), shininess) * specular0;
  lighteffect = lighteffect + vec4(col, colDiff.a);
  vec4 textureColor = lighteffect*vec4(1.,1.,1.,1.);
  fragColor = textureColor;
  float fogF;
  if (uFogMode > 0) {
    fogF = (uFogParms.y - vPosition.z/vPosition.w)/(uFogParms.y - uFogParms.x);
    if (uFogMode > 1)
      fogF = mix(uFogParms.w, 1.0, fogF);
    fogF = fogF*uFogParms.z;
    if (uFogMode == 2)
      fogF = 1.0 - exp(-fogF);
    else if (uFogMode == 3)
      fogF = 1.0 - exp(-fogF*fogF);
    fogF = clamp(fogF, 0.0, 1.0);
    gl_FragColor = vec4(mix(fragColor.rgb, uFogColor, fogF), fragColor.a);
  } else gl_FragColor = fragColor;
}
"

s <- setUserShaders(id, vertexShader = vShader,
                        fragmentShader = fShader,
               textures = c(userSampler = system.file("textures/rgl2.png",
                                      package = "rgl"),
                            unusedSampler = system.file("textures/rgl2.png",
                                                        package = "rgl")),
               uniforms = list(unusedUniform = 3),
               attributes = list(unusedAttribute = 1:10))
rglwidget(s)

##
## R source file
## This file is part of rgl
##
##

##
## ===[ SECTION: generic appearance function ]================================
##

rgl.material0 <- function(
  color        = "white",
  alpha        = 1.0,
  lit          = TRUE, 
  ambient      = "black",
  specular     = "white", 
  emission     = "black", 
  shininess    = 50.0, 
  smooth       = TRUE,
  texture      = NULL, 
  textype      = "rgb",
  texmode      = "modulate",
  texmipmap    = FALSE, 
  texminfilter = "linear", 
  texmagfilter = "linear",
  texenvmap    = FALSE,
  front        = "filled", 
  back         = "filled",
  size         = 3.0,
  lwd          = 1.0, 
  fog          = TRUE,
  point_antialias = FALSE,
  line_antialias = FALSE,
  depth_mask   = TRUE,
  depth_test   = "less",
  polygon_offset = c(0.0, 0.0),
  margin = "",
  floating = FALSE,
  tag = "",
  blend = c("src_alpha", "one_minus_src_alpha"),
  col,
  ...
) {
  # Allow compatibility with base graphics without relying
  # on abbreviated arguments
  
  if (missing(color) && !missing(col))
    color <- col
  
  # solid or diffuse component
  color     <- rgl.mcolor(color)
  if (length(color) < 1)
    stop("There must be at least one color")

  # light properties

  ambient   <- rgl.color(ambient)
  specular  <- rgl.color(specular)
  emission  <- rgl.color(emission)

  # others

  rgl.bool(lit)
  rgl.bool(fog)
  rgl.bool(smooth)
  rgl.bool(point_antialias)
  rgl.bool(line_antialias)
  rgl.bool(depth_mask)
  rgl.clamp(shininess,0,128)
  rgl.numeric(size)
  rgl.numeric(lwd)
  depth_test <- rgl.enum.depthtest(depth_test)
  
  # side-dependant rendering

  front <- rgl.enum.polymode(front)
  back  <- rgl.enum.polymode(back)

  # texture mapping

  rgl.bool(texmipmap)

  texture <- prepareTexture(texture)

  textype <- rgl.enum.textype( textype )
  texmode <- rgl.enum.texmode( texmode )
  texminfilter <- rgl.enum.texminfilter( texminfilter )
  texmagfilter <- rgl.enum.texmagfilter( texmagfilter )
  rgl.bool(texenvmap)
  texdelete <- !is.null(attr(texture, "rgl_source"))
  
  # polygon offset
  
  stopifnot(is.numeric(polygon_offset), 
            length(polygon_offset) <= 2, 
            length(polygon_offset) >= 1)
  if (length(polygon_offset) == 1)
    polygon_offset <- c(polygon_offset, polygon_offset)

  # blending
  
  stopifnot(length(blend) == 2)
  blend <- c(rgl.enum.blend(blend[1]), rgl.enum.blend(blend[2]))
  
  # vector length

  ncolor <- dim(color)[2]
  nalpha <- length(alpha)

  margin <- parseMargin(margin, floating = floating)
  
  # user data
  
  tag <- as.character(tag)
  rgl.string(tag)

  # pack data

  idata <- as.integer( c( ncolor, lit, smooth, front, back, fog, 
                          textype, texmipmap, texminfilter, texmagfilter, 
                          nalpha, ambient, specular, emission, texenvmap, 
                          point_antialias, line_antialias, 
                          depth_mask, depth_test, 
                          margin$coord - 1, margin$edge, floating,

                          blend, texmode, texdelete, color) )
  cdata <- as.character(c( tag, texture ))
  ddata <- as.numeric(c( shininess, size, lwd, polygon_offset, alpha ))

  ret <- .C( rgl_material,
    success = FALSE,
    idata,
    cdata,
    ddata
  )
}

rgl.material <- function(...) {
  .Deprecated("material3d")
  rgl.material0(...)
}

rgl.getcolorcount <- function() .C( rgl_getcolorcount, count=integer(1) )$count
  
rgl.getmaterial <- function(ncolors, id = NULL) {

  if (!length(id)) id <- 0L
  if (missing(ncolors))
    ncolors <- if (id) rgl.attrib.count(id, "colors") else rgl.getcolorcount()
  
  idata <- rep(-1, 34+3*ncolors)
  idata[1] <- ncolors
  idata[11] <- ncolors
  
  cdata <- rep(paste(rep(" ", 512), collapse=""), 2)
  ddata <- rep(0, 5+ncolors)
  
  ret <- .C( rgl_getmaterial,
    success = FALSE,
    id = as.integer(id),
    idata = as.integer(idata),
    cdata = cdata,
    ddata = as.numeric(ddata)
  )
  
  if (!ret$success) stop('rgl.getmaterial failed')
  
  polymodes <- c("filled", "lines", "points", "culled")
  textypes <- c("alpha", "luminance", "luminance.alpha", "rgb", "rgba")
  texmodes <- c("replace", "modulate", "decal", "blend", "add")
  minfilters <- c("nearest", "linear", "nearest.mipmap.nearest", "nearest.mipmap.linear", 
                  "linear.mipmap.nearest", "linear.mipmap.linear")
  magfilters <- c("nearest", "linear")
  depthtests <- c("never", "less", "equal", "lequal", "greater", 
                  "notequal", "gequal", "always")
  blendmodes <- c("zero", "one", 
                  "src_color", "one_minus_src_color", 
                  "dst_color", "one_minus_dst_color",
                  "src_alpha", "one_minus_src_alpha",
                  "dst_alpha", "one_minus_dst_alpha",
                  "constant_color", "one_minus_constant_color",
                  "constant_alpha", "one_minus_constant_alpha",
                  "src_alpha_saturate")
  idata <- ret$idata
  ddata <- ret$ddata
  cdata <- ret$cdata
  
  list(color = rgb(idata[32 + 3*(seq_len(idata[1]))], 
                   idata[33 + 3*(seq_len(idata[1]))], 
                   idata[34 + 3*(seq_len(idata[1]))], maxColorValue = 255),
       alpha = if (idata[11]) ddata[seq(from=6, length.out = idata[11])] else 1,
       lit = idata[2] > 0,
       ambient = rgb(idata[12], idata[13], idata[14], maxColorValue = 255),
       specular = rgb(idata[15], idata[16], idata[17], maxColorValue = 255),
       emission = rgb(idata[18], idata[19], idata[20], maxColorValue = 255),
       shininess = ddata[1],
       smooth = idata[3] > 0,
       texture = if (cdata[2] == "") NULL else cdata[2],
       textype      = textypes[idata[7]], 
       texmipmap    = idata[8] == 1,
       texminfilter = minfilters[idata[9] + 1],
       texmagfilter = magfilters[idata[10] + 1],
       texenvmap    = idata[21] == 1,
       front = polymodes[idata[4]],
       back = polymodes[idata[5]],
       size = ddata[2],
       lwd  = ddata[3],
       fog = idata[6] > 0,
       point_antialias = idata[22] == 1,
       line_antialias = idata[23] == 1,
       depth_mask = idata[24] == 1,
       depth_test = depthtests[idata[25] + 1],
       isTransparent = idata[26] == 1,
       polygon_offset = ddata[4:5],
       margin = deparseMargin(list(coord = idata[27] + 1, edge = idata[28:30])),
       floating = idata[31] == 1,
       blend = blendmodes[idata[32:33] + 1],
       texmode = texmodes[idata[34] + 1],
       tag = cdata[1]
       )
                   
}

prepareTexture <- function(texture) {
  arr <- NULL
  src <- NULL
  if (is.null(texture))
    result <- ""
  else if (is.character(texture) && length(texture) == 1) {
    # Assume it's a filename
    ext <- tolower(file_ext(texture))
    if (ext %in% c("jpg", "jpeg")) {
      if (requireNamespace("jpeg"))
        arr <- jpeg::readJPEG(texture)
      else
        stop("JPEG textures require the 'jpeg' package")
    } else
      result <- normalizePath(texture)
  } else {
    raster <- as.raster(texture)
    arr <- t(col2rgb(raster))/255
    dim(arr) <- c(rev(dim(raster)), 3)
    arr <- aperm(arr, c(2,1,3))
  }
  
  if (!is.null(arr)) {
    if (!requireNamespace("png"))
      stop("non-PNG textures require the 'png' package")
    result <- tempfile(fileext = ".png", tmpdir = .rglEnv$textureDir)
    src <- attr(texture, "src")
    png::writePNG(arr, target = result, 
                  text = c(rgl_source = deparse(src)))
  }
  
  structure(result, rgl_source = src)
}

textureSource <- function(texture) {
  if (missing(texture))
    return(.rglEnv$textureDir)
  if (requireNamespace("png")) {
    png <- png::readPNG(texture, info = TRUE)
    if (!is.null(info <- attr(png, "info")) && 
        !is.null(info$text) &&
        !is.na(src <- info$text["rgl_source"]))
      return(parse(text = src)[[1]])
  }
  texture
}

ellipse3d <- function(x, ...) 
  UseMethod("ellipse3d")

ellipse3d.default <-
  function(x, scale = c(1, 1, 1), centre = c(0, 0, 0), level = 0.95, 
            t = sqrt(qchisq(level, 3)), which = 1:3, subdivide = 3, smooth = TRUE, ...) {
  stopifnot(is.matrix(x))
  
  cov <- x[which, which]
  chol <- chol(cov)

  sphere <- subdivision3d(cube3d(...), subdivide) 
  norm <- sqrt( sphere$vb[1,]^2 + sphere$vb[2,]^2 + sphere$vb[3,]^2 )
  for (i in 1:3) sphere$vb[i,] <- sphere$vb[i,]/norm
  sphere$vb[4,] <- 1
  if (smooth)
    sphere$normals <- sphere$vb
  
  result <-scale3d(transform3d( sphere, chol), t,t,t)
  
  if (!missing(scale)) 
    result <- scale3d(result, scale[1], scale[2], scale[3])
  if (!missing(centre))
    result <- translate3d(result, centre[1], centre[2], centre[3])
  return(result)
}

ellipse3d.lm <-
  function(x, which = 1:3, level = 0.95, t = sqrt(3 * qf(level, 
                                                3, x$df.residual)), ...) {
  s <- summary(x)  
  names <- names(x$coefficients[which])
  structure(c(ellipse3d.default(s$sigma^2 * s$cov.unscaled[which, which], 
                              centre = x$coefficients[which], t = t, ...),
              xlab=names[1], ylab=names[2], zlab=names[3]), class="mesh3d")
}

ellipse3d.glm <- function(x, which = 1:3, level = 0.95, t, dispersion, ...) {
  s <- summary(x)
  est.disp <- missing(dispersion) & !(x$family$family %in% c('poisson','binomial'))
  if (missing(dispersion)) dispersion <- s$dispersion
  if (missing(t)) t <- ifelse(est.disp,sqrt(3 * qf(level, 3, s$df[2])),
			                           sqrt(qchisq(level, 3)))
			                           
  names <- names(x$coefficients[which])			                           
  structure(c(ellipse3d.default(dispersion * s$cov.unscaled[which, which], 
                  centre = x$coefficients[which], t = t, ...),
              xlab=names[1], ylab=names[2], zlab=names[3]), class="mesh3d")
}

ellipse3d.nls <- function(x, which = 1:3, level = 0.95, t = sqrt(3 * qf(level, 
                                                3, s$df[2])), ...) {
  s <- summary(x)  
  names <- names(x$m$getPars()[which])
  structure(c(ellipse3d.default(s$sigma^2 * s$cov.unscaled[which, which], 
                  centre = x$m$getPars()[which], t = t, ...),
	      xlab=names[1], ylab=names[2], zlab=names[3]), class="mesh3d")                  
}

convertScene <- function(x = scene3d(minimal), width = NULL, height = NULL,
                         elementId = NULL,
                         minimal = TRUE, webgl = TRUE,
                         snapshot = FALSE,
                         oldConvertBBox = FALSE,
                         useBuffer = TRUE) {
  
  # Lots of utility functions and constants defined first; execution starts way down there...
  
  getObj <- function(id) {
    result$objects[[as.character(id)]]
  }
  
  setObj <- function(id, newval) {
    result$objects[[as.character(id)]] <<- newval
  }
  
  getIdsByType <- function(type, subscene = NULL) {
    if (is.null(subscene))
      ids <- vapply(result$objects, function(x)
        if (x$type == type) x$id else NA, numeric(1))
    else {
      ids <- vapply(getObj(subscene)$objects, function(x) {
        obj <- getObj(x)
        if (obj$type == type) obj$id
        else NA
      }, numeric(1))
    }
    ids[!is.na(ids)]
  }
  
  getMaterial <- function(id) {
    default <- result$material
    obj <- getObj(id)
    mat <- obj$material
    missing <- setdiff(names(default), names(mat))
    mat[missing] <- default[missing]
    mat
  }
  
  # This counts how many clipping planes might affect a particular object
  countClipplanes <- function(id, minValue = getOption("rgl.minClipplanes", 0)) {
    recurse <- function(subscene) {
      count <- 0
      ids <- getIdsByType("clipplanes", subscene)
      for (clipid in ids)
        count <- count + nrow(getObj(clipid)$normals)
      
      subscenes <- getIdsByType("subscene", subscene)
      for (sub in subscenes) {
        if (count >= bound)
          break
        count <- max(count, recurse(sub))
      }
      count
    }
    ids <- getIdsByType("clipplanes")
    bound <- 0
    for (i in seq_along(ids))
      bound <- bound + nrow(getObj(ids[i])$normals)
    if (bound < minValue) return(minValue)
    max(minValue, recurse(result$rootSubscene))
  }
  
  makeList <- function(x) {
    if (is.list(x)) x <- lapply(x, makeList)
    if (length(names(x))) x <- as.list(x)
    x
  }
  
  initResult <- function() {
    result <<- makeList(x)
    
    recurse <- function(subscene) {
      subscenes <- subscene$subscenes
      for (i in seq_along(subscenes)) {
        subscenes[[i]]$parent <- subscene$id
        subscenes[[i]] <- recurse(subscenes[[i]])
      }
      if (length(subscenes)) {
        subscene$subscenes <- unlist(subscenes)
        subscene$objects <- c(subscene$objects, subscene$subscenes)
      } else
        subscene$subscenes <- numeric(0)
      setObj(subscene$id, subscene)
      subscene$id
    }
    result$rootSubscene <<- recurse(result$rootSubscene)
    
    if (snapshot)
      result$snapshot <- getSnapshot()
  }
  
  flagnames <- c("is_lit", "is_smooth", "has_texture",
           "depth_sort", "fixed_quads", "is_transparent",
           "is_lines", "sprites_3d", 
           "is_subscene", "is_clipplanes",
           "fixed_size", "is_points", "is_twosided",
           "fat_lines", "is_brush", "has_fog",
           "rotating")
  
  getFlags <- function(id) {
    
    obj <- getObj(id)
    if (is.null(obj)) {
      warning("object", id, " not found.")
      return(structure(rep(FALSE, length(flagnames)), names = flagnames))
    }
    type <- obj$type
    
    if (type == "subscene")
      return(getSubsceneFlags(id))
    
    result <- structure(rep(FALSE, length(flagnames)), names = flagnames)
    if (type == "clipplanes") {
      result["is_clipplanes"] <- TRUE
      return(result)
    }
    
    if (type == "light")
      return(result)
    
    result["is_transparent"] <- any(obj$colors[,"a"] < 1); # More later...
    
    mat <- getMaterial(id)
    result["is_lit"] <- mat$lit && type %in% c("triangles", "quads", "surface", "planes",
                 "spheres", "sprites", "bboxdeco")
    
    result["is_smooth"] <- mat$smooth && type %in% c("triangles", "quads", "surface", "planes",
                 "spheres")

    result["sprites_3d"] <- sprites_3d <- type == "sprites" && length(obj$ids)
    
    result["has_texture"] <- has_texture <- !is.null(mat$texture) &&
                                            (!is.null(obj$texcoords) 
                                             || (type == "sprites" && !sprites_3d)
                                             || (type == "background" && obj$sphere)
                                             || (type == "spheres"))
    
    result["is_transparent"] <- is_transparent <- (has_texture && mat$isTransparent) || result["is_transparent"]
    
    result["depth_sort"] <- depth_sort <- is_transparent && type %in% c("triangles", "quads", "surface",
                        "spheres", "sprites", "text")
    
    result["fixed_quads"] <- type %in% c("text", "sprites") && !sprites_3d
    result["is_lines"]    <- type %in% c("lines", "linestrip", "abclines")
    result["is_points"]   <- type == "points" || "points" %in% c(mat$front, mat$back)
    result["is_twosided"] <- (type %in% c("quads", "surface", "triangles", "spheres", "bboxdeco") && 
      length(unique(c(mat$front, mat$back))) > 1) ||
      (type == "background" && obj$sphere)
    result["fixed_size"]  <- type == "text" || isTRUE(obj$fixedSize)
    result["rotating"] <- isTRUE(obj$rotating)
    result["fat_lines"]   <- mat$lwd != 1 && (result["is_lines"] || 
                  "lines" %in% unlist(mat[c("front", "back")]))
    result["is_brush"] <- !is.na(brushId) && id == brushId
    result["has_fog"] <- mat$fog
    result
  }
  
  getSubsceneFlags <- function(id) {
    result <- structure(rep(FALSE, length(flagnames)), names = flagnames)
    result["is_subscene"] <- TRUE
    objs <- getObj(id)$objects
    for (i in seq_along(objs))
      result <- result | getFlags(objs[i])
    return(result)
  }
  
  numericFlags <- function(flags) {
    if (is.matrix(flags))
      n <- ncol(flags)
    else
      n <- length(flags)
    unname(flags %*% 2^(seq_len(n)-1))
  }
  
  expandFlags <- function(numericflags) {
    result <- matrix(FALSE, nrow = length(numericflags),
         ncol = length(flagnames),
         dimnames = list(names(numericflags), flagnames))
    for (i in seq_along(flagnames)) {
      result[,i] <- numericflags %% 2 == 1
      numericflags <- numericflags %/% 2
    }
    result
  }
  
  plotClipplanes <- function(subscene) {
    for (id in subscene$objects) {
      obj <- getObj(id)
      if (obj$type == "clipplanes") {
        class(obj) <- "rglobject"
        plot3d(obj)
      } else if (obj$type == "subscene")
        plotClipplanes(getObj(id))
    }
  }
  
  convertBBox <- function(id, subscene) {
    obj <- getObj(id)
    verts <- obj$vertices
    text <- obj$texts
    if (!length(text))
      text <- rep("", NROW(verts))
    else
      text <- text[,"text"]
    mat <- getMaterial(id)
    if (length(mat$color) > 1)
      mat$color <- mat$color[2] # We ignore the "box" colour
    
    if(any(missing <- text == ""))
      text[missing] <- apply(verts[missing,], 1, function(row) format(row[!is.na(row)]))
    
    res <- numeric(0)
    bbox <- subscene$par3d$bbox
    repeat { # Need to make sure the ids here don't clash with those in the scene
      tempID <- points3d(bbox[1:2],
             bbox[3:4],
             bbox[5:6])
      if (tempID > lastID)
        break
      else
        delFromSubscene3d(tempID)
    }
    
    lastID <<- tempID
    
    intersect <- function(limits, points)
      which(limits[1] <= points & points <= limits[2])
    
    # plot the clipping planes as they affect the bounding box
    plotClipplanes(subscene)
    
    mat$front <- mat$back <- "filled"
    
    if (any(inds <- is.na(verts[,2]) & is.na(verts[,3])) && length(keep <- intersect(bbox[1:2], verts[inds, 1])))
      res <- c(res, do.call(axis3d, c(list(edge = "x", at = verts[inds, 1][keep], labels = text[inds][keep]), mat)))
    if (any(inds <- is.na(verts[,1]) & is.na(verts[,3])) && length(keep <- intersect(bbox[3:4], verts[inds, 2])))
      res <- c(res, do.call(axis3d, c(list(edge = "y", at = verts[inds, 2][keep], labels = text[inds][keep]), mat)))
    if (any(inds <- is.na(verts[,1]) & is.na(verts[,2])) && length(keep <- intersect(bbox[5:6], verts[inds, 3])))
      res <- c(res, do.call(axis3d, c(list(edge = "z", at = verts[inds, 3][keep], labels = text[inds][keep]), mat)))
    res <- c(res, do.call(box3d, mat))
    delFromSubscene3d(c(res, tempID))
    res
  }
  
  convertBBoxes <- function(id) {
    if (!oldConvertBBox)
      return(NULL)
    ids <- origIds <- NULL
    id <- as.character(id)
    sub <- getObj(id)
    types <- vapply(sub$objects,
        function(x) getObj(x)$type,
        character(1))
    names(types) <- as.character(sub$objects)
    if (length(bboxes <- names(types)[types == "bboxdeco"])) {
      for (i in bboxes) {
        newids <- convertBBox(i, sub)
        sub$objects <- c(sub$objects, as.numeric(newids))
        setObj(id, sub)
        ids <- c(ids, newids)
        origIds <- c(origIds, rep(i, length(newids)))
      }
    }
    children <- sub$subscenes
    for (i in children) {
      childids <- convertBBoxes(i)
      ids <- c(ids, childids)
      origIds <- c(origIds, attr(childids, "origIds"))
    }
    if (length(origIds))
      names(origIds) <- as.character(ids)
    if (is.null(ids)) ids
    else structure(ids, origIds = origIds)
  }
  
  createBrush <- function() {
    repeat { # Need to make sure the id doesn't clash with those in the scene
      tempID <- lines3d(x = c(0, 1, 1, 0, 0),
      		        y = c(0, 0, 1, 1, 0), 
      		        z = rep(-.999, 5), 
      		        depth_test = "always",
      		        lit = FALSE,
      		        alpha = 0.5)
      if (tempID > lastID)
        break
      else
        delFromSubscene3d(tempID)
    } 
    lastID <<- tempID
  }
  
  getSnapshot <- function() 
    knitr::include_graphics(snapshot3d(scene = x, width = width, height = height))
  
  knowntypes <- c("points", "linestrip", "lines", "triangles", "quads",
      "surface", "text", "abclines", "planes", "spheres",
      "sprites", "clipplanes", "light", "background", "bboxdeco",
      "subscene")
  
  #  Execution starts here!

  # Do a few checks first

  if (!webgl)
    return(getSnapshot())
  
  if (is.null(elementId))
    elementId <- ""
  
  if (is.list(x$rootSubscene))
    rect <- x$rootSubscene$par3d$windowRect
  else
    rect <- x$objects[[x$rootSubscene]]$par3d$windowRect
  
  rwidth <- rect[3] - rect[1] + 1
  rheight <- rect[4] - rect[2] + 1
  if (!length(width)) {
    if (!length(height)) {
      wfactor <- hfactor <- 1  # width = wfactor*rwidth, height = hfactor*rheight
    } else
      wfactor <- hfactor <- height/rheight
  } else {
    if (!length(height)) {
      wfactor <- hfactor <- width/rwidth
    } else {
      wfactor <- width/rwidth
      hfactor <- height/rheight
    }
  }
  width <- wfactor*rwidth
  height <- hfactor*rheight
  
  shared <- x$crosstalk$id
  
  result <- NULL
  
  initResult()
  
  result$width <- width
  result$height <- height
  
  types <- vapply(result$objects, function(x) x$type, character(1))
  lastID <- max(vapply(result$objects, function(x) x$id, numeric(1)))
  
  if (any(types == "bboxdeco")) {
    saveNULL <- options(rgl.useNULL = TRUE)
    dev <- cur3d()
    open3d()
    ids <- convertBBoxes(result$rootSubscene)
    origIds <- attr(ids, "origIds")
    scene <- scene3d(minimal)
    temp <- lapply(as.character(ids),
             function(id) {
               x <- scene$objects[[id]]
               x$origId <- as.numeric(origIds[id])
               x
             })
    result$objects[as.character(ids)] <- temp
    for (id in unique(origIds))
      result$objects[[as.character(id)]]$newIds <- as.numeric(ids[origIds == id])
    types <- vapply(result$objects, function(x) x$type, character(1))
    close3d()
    if (dev)
      set3d(dev)
    options(saveNULL)
  }
  
  if (length(shared)) {
    saveNULL <- options(rgl.useNULL = TRUE)
    dev <- cur3d()
    open3d()
    result$brushId <- brushId <- createBrush()
    brush <- as.character(result$brushId)
    scene <- scene3d(minimal)
    result$objects[[brush]] <- scene$objects[[brush]]
    close3d()
    if (dev)
      set3d(dev)
    options(saveNULL)
  } else
    brushId <- NA
  
  ids <- vapply(result$objects, function(x) x$id, numeric(1))
  flags <- vapply(result$objects, function(obj) numericFlags(getFlags(obj$id)),
      numeric(1), USE.NAMES = FALSE)
  
  unknowntypes <- setdiff(types, knowntypes)
  if (length(unknowntypes))
    warning(gettextf("Object type(s) %s not handled",
         paste("'", unknowntypes, "'", sep="", collapse=", ")), domain = NA)
  
  keep <- types %in% setdiff(knowntypes, c("light"))
  ids <- ids[keep]
  cids <- as.character(ids)
  nflags <- flags[keep]
  types <- types[keep]
  flags <- expandFlags(nflags)
  rownames(flags) <- cids
  fullviewport <- getObj(result$rootSubscene)$par3d$viewport
  
  for (i in seq_along(ids)) {
    obj <- getObj(cids[i])
    obj$flags <- nflags[i]
    if (obj$type != "subscene") {
      texturefile <- ""
      if (!is.null(obj$material) && "texture" %in% names(obj$material))
        texture <- obj$material$texture
      else
        texture <- result$material$texture
      if (!is.null(texture) && nchar(texture)) {
        texturefile <- texture
        obj$material$uri <- image_uri(texturefile)
        obj$material$texture <- NULL
      }
      if (!is.null(obj$material)) # Never use material$color
        obj$material$color <- NULL
      
    } else if (obj$type == "subscene") {
      obj$par3d$viewport$x <- obj$par3d$viewport$x/fullviewport$width
      obj$par3d$viewport$width <- obj$par3d$viewport$width/fullviewport$width
      obj$par3d$viewport$y <- obj$par3d$viewport$y/fullviewport$height
      obj$par3d$viewport$height <- obj$par3d$viewport$height/fullviewport$height
    }
    if (obj$type == "planes" && nrow(obj$vertices) > 3) {
      obj$vertices <- obj$vertices[1:3,] # These will be redone
      # in Javascript
    } else if (obj$type == "spheres")
      obj$centers <- obj$vertices
    if (!is.null(obj$material$margin)) {
      margin <- parseMargin(obj$material$margin, obj$material$floating)
      obj$material$margin <- margin$coord - 1
      obj$material$floating <- margin$floating
      obj$material$edge <- margin$edge
    }
    if (is.list(obj$userTextures) && 
        length(obj$userTextures) &&
        !is.list(obj$userTextures[[1]])) {
      textureNames <- names(obj$userTextures)
      userTextures <- as.character(obj$userTextures)
      obj$userTextures <- list()
      for (j in seq_along(userTextures)) {
        texturefile <- userTextures[j]
        obj$userTextures[[j]] <- list(file = texturefile,
                                      uri = image_uri(texturefile))
      }
      names(obj$userTextures) <- textureNames
    }
    setObj(cids[i], obj)
  }
  if (useBuffer) {
    # Put the data into the buffer
    buffer <- Buffer$new()
    for (i in seq_along(ids)) {
      obj <- getObj(cids[i])
      # This list needs to match the one in buffer.src.js
      for (n in c("vertices", "normals", "indices", 
                  "texcoords", "colors", "centers")) {
        if (!is.null(obj[[n]])) {
          normalized <- FALSE
          normalization <- ""
          if (length(obj[[n]]) > 6 &&  # Don't bother for short ones
              n %in% c("colors","texcoords") &&
              0 <= (objrange <- range(obj[[n]]))[1] &&
              objrange[2] <= 1) {
            scaled <- 255*obj[[n]]
            rounded <- round(scaled)
            normalized <- max(abs(scaled - rounded))/255 < 1.e-7
            if (normalized)
              normalization <- "ubyte"
            else {
              scaled <- 65535*obj[[n]]
              rounded <- round(scaled)
              normalized <- max(abs(scaled - rounded))/65535 < 1.e-7
              if (normalized)
                normalization <- "ushort"
            }
          }
          if (normalized)
            obj[[n]] <- as.character(buffer$addAccessor(t(rounded), types = normalization, normalized = TRUE))
          else
            obj[[n]] <- as.character(buffer$addAccessor(t(obj[[n]])))
        }
      }
      setObj(cids[i], obj)
    }
    buffer$closeBuffers()
    buf <- buffer$as.list()
    
    result$buffer <- buf
  }

  result$context <- list(shiny = inShiny(), rmarkdown = rmarkdownOutput())

  result
}

# This file supports auto-printing of RGL scenes in
# RStudio

# Called just after a low level function has been
# called, likely changing an existing display
# Returns the ids that were created by the function,
# which should be passed as the ids arg.

lowlevel <- function(ids = integer()) {
  structure(ids, class = c("rglLowlevel", "rglId", "numeric"))
}

# Called just after a high level function (plot3d
# or persp3d) has been called, if it wasn't
# called with add = TRUE (in which case it would be
# treated as low level).
# Returns the ids that were created by the function,
# which should be passed as the ids arg.

highlevel <- function(ids = integer()) {
  structure(ids, class = c("rglHighlevel", "rglId", "numeric"))
}

rglId <- function(ids = integer()) {
  structure(ids, class = "rglId")
}

print.rglId <- function(x, rglwidget = getOption("rgl.printRglwidget", FALSE),
			...) {
  if (!par3d("skipRedraw")) {
    if (rglwidget)
      # FIXME:  For lowlevel, this should replace the scene, not update the history
      print(rglwidget(...))
    else if (in_pkgdown_example())
      pkgdown::pkgdown_print(x)
  }
  invisible(x)
}

plotmath3d <- function(x, y = NULL, z = NULL,
		       text, 
		       cex = par("cex"), adj = 0.5,
		       pos = NULL, offset = 0.5,
		       fixedSize = TRUE,
		       startsize = 480, initCex = 5, 
		       margin = "", floating = FALSE, tag = "",
		       ...) {
  xyz <- xyz.coords(x, y, z, recycle = TRUE)
  n <- length(xyz$x)
  if (is.vector(text))
    text <- rep(text, length.out = n)
  cex <- rep(cex, length.out = n)
  if (!is.null(pos))
    pos <- rep_len(pos, n)
  adj <- c(adj, 0.5, 0.5, 0.5)[1:3]
  save3d <- par3d(skipRedraw = TRUE)
  save <- options(device.ask.default = FALSE)
  on.exit({options(save); par3d(save3d)}) # nolint
  result <- integer(n)
  for (i in seq_len(n)) {
    # Open the device twice.  The first one is to measure the text...
    f <- tempfile(fileext = ".png")
    png(f, bg = "transparent", width = startsize, height = startsize)
    par(mar = c(0, 0, 0, 0), xaxs = "i", xaxt = "n",  
        yaxs = "i", yaxt = "n",
        usr = c(0, 1, 0, 1))
    plot.new()
    if (is.vector(text))
      thistext <- text[i]
    else
      thistext <- text
    w <- strwidth(thistext, cex = initCex, ...)
    w1 <- strwidth("m", cex = initCex, ...)
    h <- strheight(thistext, cex = initCex, ...)
    safe.dev.off()

    # Now make a smaller bitmap
    expand <- 1.5
    size <- round(expand*startsize*max(c(w, h)))
    png(f, bg = "transparent", width = size, height = size)
    par(mar = c(0, 0, 0, 0), xaxs = "i", xaxt = "n", 
        yaxs = "i", yaxt = "n",
        usr = c(0, 1, 0, 1))
    plot.new()
    text(0.5, 0.5, thistext, adj = c(0.5,0.5), cex = initCex, ...)
    safe.dev.off()
    # The 0.4 tries to match the text3d offset
    offseti <- 0.4*offset*h/w
    posi <- if (is.null(pos)) NULL else pos[i]
    result[i] <- with(xyz, sprites3d(x[i], y[i], z[i], texture = f, textype = "rgba", 
            col = "white", lit = FALSE, radius = cex[i]*size/initCex/20,
            adj = adj, pos = posi, offset = offseti,
            fixedSize = fixedSize,
            margin = margin, floating = floating, tag = tag))
  }
  lowlevel(result)
}

aspect3d <- function(x, y = NULL, z = NULL) {
    if (is.character(x) && pmatch(x, "iso") == 1) {
    	scale <- c(1,1,1)
    } else {
	if (is.null(y)) {
	    x <- rep(x, length.out = 3)
	    z <- x[3]
	    y <- x[2]
	    x <- x[1]
	}
        for (i in 1:5) { # To handle spheres, repeat this 
  	    bbox <- .getRanges()
	    scale <- c(diff(bbox$xlim), diff(bbox$ylim), diff(bbox$zlim))
	    scale <- ifelse(scale <= 0, 1, scale)

	    avgscale <- sqrt(sum(scale^2)/3)
            scale <- c(x,y,z)*avgscale/scale
            oldscale <- par3d(scale = scale)$scale
            if (isTRUE(all.equal(scale, oldscale)))
            	break
        }
    }
    par3d(scale = scale)
}

# Suggested package checks

checkDeldir <- function(error = FALSE) {
  result <- requireNamespace("deldir", quietly = TRUE) &&
    (packageVersion("deldir") < "1.0.2" ||
       packageVersion("deldir") >= "1.0.4")
  if (error && !result)
    stop("This function requires the 'deldir' package (but is not compatible with version 1.0-2).", call. = FALSE)
  result
}

grid3d <- function(side, at = NULL, col="gray",
                    lwd = 1, lty = 1, n = 5) {

  save <- par3d(skipRedraw = TRUE, ignoreExtent = TRUE)
  on.exit(par3d(save))

  if (!missing(side) && length(side)>1)
    return(lowlevel(sapply(side,grid3d,at=at,col=col,lwd=lwd,lty=lty,n=n)))
  
  ranges <- .getRanges()

  side <- c(strsplit(side, '')[[1]], '-')[1:2]
  coord <- match(toupper(side[1]), c('X', 'Y', 'Z'))
  spos <- match(side[2],c('-','+'))

  sidenames <- c('x', 'y', 'z')
  
  sides <- 1:3
  sides <- sides[-coord]

  if (is.null(at)) at <- list()
  else if (is.numeric(at)) {
    at <- list(x=at, y=at, z=at)
    at[[coord]] <- NULL
  }
  
  result <- integer()
  
  for (cside in sides) {
    range <- ranges[[cside]]
    if (is.null(at1 <- at[[sidenames[cside]]]))
      at1 <- pretty(range, n)
    at1 <- at1[at1 >= range[1] & at1 <= range[2]]
    mpos1 <- matrix(rep(c(ranges$x[1],ranges$y[1],ranges$z[1]),
                        each=length(at1)),
                    ncol=3)
    mpos2 <- matrix(rep(c(ranges$x[2],ranges$y[2],ranges$z[2]),
                        each=length(at1)),
                  ncol=3)
    mpos1[,cside] <- mpos2[,cside] <- at1
    if (is.null(at[[sidenames[coord]]]))
      mpos1[,coord] <- mpos2[,coord] <- ranges[c("x","y","z")][[coord]][spos]
    else {
      # may need to duplicate 
      temp1 <- temp2 <- matrix(NA, nrow=0, ncol=3)
      planes <- at[[sidenames[coord]]]
      planes <- planes[planes >= ranges[[coord]][1] & planes <= ranges[[coord]][2]]
      for (at2 in planes) {
      	mpos1[,coord] <- mpos2[,coord] <- at2
      	temp1 <- rbind(temp1, mpos1)
      	temp2 <- rbind(temp2, mpos2)
      }
      mpos1 <- temp1
      mpos2 <- temp2
    }
    if (nrow(mpos1) + nrow(mpos2) > 0)
      result[sidenames[cside]] <- segments3d(x=c(rbind(mpos1[,1],mpos2[,1])),
               y=c(rbind(mpos1[,2],mpos2[,2])),
               z=c(rbind(mpos1[,3],mpos2[,3])),
               lwd=lwd,color=col)
  }
  lowlevel(result)
}

---
title: "rgl Overview"
author: "Duncan Murdoch"
date: "`r format(Sys.time(), '%B %d, %Y')`"
output:
  rmarkdown::html_vignette:
    fig_height: 5
    fig_width: 5
    toc: true
  html_document:
    df_print: paged
    toc: true
  pdf_document:
    fig_height: 5
    fig_width: 5
    toc: true
vignette: >
  %\VignetteIndexEntry{rgl Overview}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r setup, echo=FALSE, results="asis"}
source("setup.R")
setupKnitr(autoprint = TRUE)
set.seed(123)
```



## Introduction

The `rgl` package is used to produce interactive 3-D plots.  It contains
high-level graphics commands modelled loosely after classic R graphics,
but working in three dimensions.  It also contains low level structure
inspired by (but incompatible with) the `grid` package.

This document gives an overview.  See the help pages for details.

For installation instructions, see the `README` file in the top
level directory of the source tarball 
`r paste0("rgl_", packageVersion("rgl"), ".tar.gz")` (or a later
version).


### About this document

This document was written in R Markdown, using the `knitr` package
for production.  It corresponds to `rgl` version 
`r packageVersion("rgl")`.

Most of the highlighted function names are HTML links.
The internal links should work in any browser; the links to
help topics should work if you view the vignette from
within the R help system.

The document includes WebGL figures.  To view these, you must have
Javascript and WebGL enabled in your browser.  Some older browsers
may not support this -- see https://get.webgl.org for tests
and links to a discussion.


## Basics and High Level Functions

The `r indexfns("plot3d")` function
plots points within an RGL window.  It is similar to the classic 
`r linkfn("plot", pkg="graphics")` function,
but works in 3 dimensions.

For example
```{r "plot3d()"}
with(iris, plot3d(Sepal.Length, Sepal.Width, Petal.Length, 
                  type="s", col=as.numeric(Species)))
```

can be used to plot three columns of the `iris` data.
Allowed plot types include `"p", "l", "h", "s"`,
meaning points, lines, segments from z=0, and spheres.  There's
a lot of flexibility in specifying the coordinates; the 
`r linkfn("xyz.coords", pkg = "grDevices")` function
from the `grDevices` package is used for this.

You can use your mouse to manipulate the plot.  The default is that
if you click and hold with the left mouse button, you can rotate 
the plot by dragging it.  The right mouse button is used to resize
it, and the middle button changes the perspective in the point of view.

If you call `r linkfn("plot3d")` again, it will overwrite the current plot.
To open a new graphics window, use `r linkfn("open3d")`.
 
The other high level function is `r indexfns("persp3d")` to draw surfaces.
It is
similar to the classic `r linkfn("persp", pkg = "graphics")`
function, but with greater flexibility.
First, any of `x`, `y` or `z`
can be specified using matrices, not just `z`.  This allows parametric
surfaces to be plotted. 
An even simpler specification is possible:  `x` may be a function,
in which case `persp3d` will work out the grid itself.  See 
`r linkfn("persp3d.function", text="?persp3d.function", pkg="rgl")`
for details.  For example, the `MASS` package estimates
Gamma parameters using maximum likelihood in a 
`r linkfn("fitdistr", text="?MASS::fitdistr", pkg="MASS")` example.
Here we show the log likelihood surface. 
```{r "persp3d()", fig.height=3, fig.width=6, fig.keep="last", eval=requireNamespace("MASS",quietly=TRUE)}
# This example requires the MASS package
library(MASS)
# from the fitdistr example
set.seed(123)
x <- rgamma(100, shape = 5, rate = 0.1)
fit <- fitdistr(x, dgamma, list(shape = 1, rate = 0.1), lower = 0.001)
loglik <- function(shape, rate) sum(dgamma(x, shape=shape, rate=rate, 
                                           log=TRUE))
loglik <- Vectorize(loglik)
xlim <- fit$estimate[1]+4*fit$sd[1]*c(-1,1)
ylim <- fit$estimate[2]+4*fit$sd[2]*c(-1,1)

mfrow3d(1, 2, sharedMouse = TRUE)
persp3d(loglik, 
        xlim = xlim, ylim = ylim,
        n = 30)
zlim <- fit$loglik + c(-qchisq(0.99, 2)/2, 0)
next3d()
persp3d(loglik, 
        xlim = xlim, ylim = ylim, zlim = zlim,
        n = 30)
```
 On the left, the whole surface over a range of the parameters; on the right, only the 
parts of the surface with log likelihood values near the maximum.

Note:  this example used the `knitr` hook functions 
(see `r linkfn("setupKnitr")`) to insert the
scene into this vignette; the previous example used the `rglwidget`
function.  We generally
recommend the newer `r linkfn("rglwidget")` approach.

Note that both `plot3d` and `persp3d` are generic functions,
with the following methods defined:
```{r}
methods(plot3d)
methods(persp3d)
```


## Adding Graphical Elements

### Primitive shapes

Just as we have `r linkfn("points", pkg="graphics")` and 
`r linkfn("lines", pkg="graphics")` in classic graphics, there are a number
of low level functions in `rgl` to add graphical elements to the
currently active plot.  The "primitive" shapes are those that are 
native to OpenGL:

Function                      | Description
----------------------------- | -----------
`r indexfns("points3d")`:     | adds points
`r indexfns("lines3d")`:      | adds lines
`r indexfns("segments3d")`:   | adds line segments
`r indexfns("triangles3d")`:  | adds triangles
`r indexfns("quads3d")`:      | adds quadrilaterals

Each of the above functions takes arguments `x`, `y` and `z`, again
using `r linkfn("xyz.coords", pkg="grDevices")` for flexibility.
They group successive entries
as necessary.  For example, the `r linkfn("triangles3d")` function takes each
successive triple of points as the vertices of a triangle.

You can use these functions to annotate the current graph, or to 
construct a figure from scratch.

### Constructed shapes

`rgl` also has a number of objects which it constructs
from the primitives.

Function                      | Description
----------------------------- | -----------
`r indexfns(c("text3d", "texts3d"))`: | adds text
`r indexfns("abclines3d")`:   | adds straight lines to plot (like `abline`)
`r indexfns("arc3d")`:        | adds spherical arcs or spirals to plot
`r indexfns("planes3d")`:     | adds planes to plot
`r indexfns("clipplanes3d")`: | add clipping planes to plot
`r indexfns(c("sprites3d", "particles3d"))`: | add sprites (fixed shapes or images) to plot
`r indexfns("spheres3d")`:    | adds spheres
`r indexfns(c("surface3d", "terrain3d"))`:    | a surface (as used in `r linkfn("persp3d")`)
`r indexfns("drape3d")`:      | drapes lines on a surface or object(s)
`r indexfns("shadow3d")`:     | projects mesh onto a surface
`r indexfns("arrow3d")`:      | add an arrow to a scene
`r indexfns("pch3d")`:	      | draw base-style plotting symbols
`r indexfns("plotmath3d")`:   | used by `r linkfn("text3d")` for math text
`r indexfns("latex3d")`:      | render LaTeX fragments

### Axes and other "decorations"

The following low-level functions control the look of the graph:

Function                             | Description
------------------------------------ | -----------
`r indexfns(c("axes3d", "axis3d"))`: | add axes to plot
`r indexfns(c("box3d", "bbox3d"))`:               | add box around plot
`r indexfns("title3d")`:             | add title to plot
`r indexfns("mtext3d")`:             | add marginal text to plot
`r indexfns("decorate3d")`:          | add multiple "decorations" (scales, etc.) to plot
`r indexfns("aspect3d")`:            | set the aspect ratios for the plot
`r indexfns(c("bg3d", "bgplot3d"))`: | set the background of the scene
`r indexfns("show2d")`:              | show a 2D plot or image in a 3D scene
`r indexfns("legend3d")`:            | set a legend for the scene
`r indexfns("grid3d")`:              | add a reference grid to a graph
`r indexfns("thigmophobe3d")`:       | choose label positions to avoid overlap
`r indexfns("setAxisCallbacks")`:    | set user-defined axis annotations

For example, to plot three random triangles, one could use
```{r "triangles3d()", fig.width=3, fig.height=3}
triangles3d(cbind(x=rnorm(9), y=rnorm(9), z=rnorm(9)), col = "green")
decorate3d()
bg3d("lightgray")
aspect3d(1,1,1)
```

Besides the `*3d` functions mentioned above, there are 
deprecated
functions 
`r deprecated(c("rgl.abclines", "rgl.bbox", "rgl.bg",
"rgl.clipplanes", "rgl.lines", 
"rgl.linestrips", "rgl.planes", "rgl.points",
"rgl.primitive", 
"rgl.quads", "rgl.setAxisCallback",
"rgl.spheres", "rgl.sprites", "rgl.surface",
"rgl.texts", 
"rgl.triangles"))`. 
You should avoid using all of these functions, which do not
work properly with the `*3d` functions and will
soon be removed from `rgl`.  See the 
`r linkfn("r3d", text="?r3d", pkg="rgl")` help
topic for details.  

The function
`r indexfns(c("rgl.getAxisCallback"))`
provides low-level support for `r linkfn("setAxisCallbacks")`.

## Controlling the Look of the Scene

### Camera angle

By default when you open a new plot with `r linkfn("open3d")`:

* The x axis goes from left to right.
* The y axis goes from near the camera to far away.
* The z axis goes from down to up.

You can change the camera angle simply by dragging the picture with
the mouse.

To set the camera angle programmatically, use `r indexfns("view3d")`. 
This uses polar coordinates:

* `theta` sets the rotation around the vertical axis, in degrees. 
* `phi` sets the "vertical" rotation around the horizontal axis, 
  from -90 to 90 degrees.
  
The default angle is roughly `theta = 0, phi = -70`. Starting from
this position:

* As you increase `theta`, the graph will spin anticlockwise from your
  point of view.
* As you increase `phi` to 0, you start to look down at the scene
  from the top. If you increase `phi` above 0, you continue over and 
  start to see the graph from the "back" (and upside down).

You can also use `r indexfns("observer3d")` to change the camera 
location using `x,y,z` coordinates. In particular, increasing
the `z` coordinate lets you zoom out, and decreasing it zooms you in.

One useful approach is to use the mouse to find a nice viewing angle.
You can then save it using `par3d("userMatrix")` and restore
the same view later:

```{r, eval = FALSE}
myview <- par3d("userMatrix")
# ... later ...
par3d(userMatrix = myview)
```


### Lighting

In most scenes, objects are "lit", meaning that their appearance
depends on their position and orientation relative to lights
in the scene.  The lights themselves don't normally show up,
but their effect on the objects does.

Use the `r indexfns("light3d")` function to 
specify the position and characteristics of a light.
Lights may be infinitely distant, or may be embedded
within the scene.  Their characteristics include 
`ambient`, `diffuse`, and `specular` components, all
defaulting to white.  The `ambient` component appears
the same from any direction.  The `diffuse` component 
depends on the angle between the surface and the light, 
while the `specular` component also takes the viewer's
position into account.

The deprecated `r deprecated("rgl.light")` function should
no longer be used; use `r linkfn("light3d")` instead.

### Materials

The mental model used in `rgl` is that the objects being shown
in scenes are physical objects in space, with material properties
that affect how light reflects from them (or is emitted by them).
These are mainly controlled by the `r indexfns("material3d")` function,
or by arguments to other functions that are passed to it.

The material properties that are recognized in calls to `material3d` are
described in detail in the 
`r linkfn("material3d", text="?material3d", pkg="rgl")` help page, and listed in the `r indexfns("rgl.material.names")`
variable.  All of them can be set except the ones in 
`r indexfns("rgl.material.readonly")`.
Here we give an overview.

Property  | Default | Meaning 
--------- | ------- | ------------------
color     | white   | vector of surface colors to apply to successive vertices for diffuse light
alpha     | 1       | transparency:  0 is invisible, 1 is opaque
lit       | TRUE    | whether lighting calculations should be done
ambient   | black   | color in ambient  light
specular  | white   | color in specular light
emission  | black   | color emitted by the surface
shininess | 50      | controls the specular lighting:  high values look shiny
smooth    | TRUE    | whether shading should be interpolated between vertices
texture   | NULL    | optional path to a "texture" bitmap to be displayed on the surface
front, back | fill  | should polygons be filled, or outlined? 
size      | 3       | size of points in pixels
lwd       | 1       | width of lines in pixels

Other properties include 
"texmipmap", "texmagfilter", "texminfilter", "texenvmap", "fog", 
"point\_antialias", "line\_antialias", "depth\_mask", "depth\_test", "polygon_offset", "margin", "floating", "tag" and "blend"; 
see `r linkfn("material3d", "the help page", pkg = "rgl")` for details.

There is also a deprecated `r deprecated("rgl.material")` function that works
at a lower level; users should avoid it.

### Textures

As described in the previous section, one of the material
properties is `texture`, the name of a bitmap file 
containing an image to be displayed on the surface. 
This section gives more details about textures.

Internally, `rgl` always uses PNG files for textures.  JPEG
files or raster objects may be specified, in which case `rgl`
will convert
them to PNG.  The `r indexfns("textureSource")` function can
be used to retrieve the original expression that provided
the bitmap.

In `OpenGL`, each vertex in a polygon may be associated with a
particular location in the bitmap.  The interior of the
polygon interpolates within the bitmap.  There are two 
conventions in `rgl` functions for specifying these coordinates.

Functions which specify primitives (`r linkfn("triangles3d")`, etc.) accept an optional matrix argument
`texcoords` which gives `s` (horizontal) and `t` (vertical)
locations within the bitmap in columns with one row per vertex.
The coordinates are `(0,0)` for the lower left, and `(1,1)`
for the upper right.  If values outside this range are given,
the image repeats, i.e. `(1.1, 1.2)` would specify the same
point in the image as `(0.1, 0.2)`.

Other functions such as `r linkfn("surface3d")` that take
matrices for each vertex coordinate accept texture coordinates
as matrices as well, in arguments `texture_s` and `texture_t`.

For example, the 
following code displays four copies of a 2D plot on a quad, 
because the texture coordinates run from 0 to 2 in both
`s` and `t`:

```{r Texture}
filename <- tempfile(fileext = ".png")
png(filename = filename)
plot(rnorm(1000), rnorm(1000))
safe.dev.off()
open3d()
xyz <- cbind(c(0,1,1,0), 0, c(0,0,1,1))
quads3d(xyz, texture = filename, texcoords = xyz[,c(1, 3)]*2, 
        col = "white", specular = "black")
```

Some other notes:

 - The color in `quads3d()` above was specified to be white.  By default,
the colors in the bitmap will modify the color of the
surface.  If `col` is black (a common default), you won't see anything,
so a warning may be issued.  
 - You usually don't want specular 
 reflections (which show up as glare).  Setting `specular`
 to black prevents those.
 - The material property `"texmode"` allows texture colors to
be used differently.  The default is `"modulate"`, where the texture
values combine multiplicatively with the underlying values.
 - Another aspect of how the bitmap is handled is controlled
 by the material
 property `"textype"`.  The default is `"rgb"`, which 
 takes the red-green-blue colors from the bitmap and
 uses them to modify the corresponding colors in the polygon.  
 - Other possibilities for `"textype"` and `"texmode"` are described
 in `r linkfn("material3d", "the material3d help page", pkg = "rgl")`.
 - The other `"tex*"` material properties control how 
 the interpolation within the image is done.
 - Modern `OpenGL` supports 1- and 3-dimensional textures;
 these are not currently supported in `rgl`.
 
### Fonts

`rgl` uses the same ideas as base graphics for drawing
text:  there are font families named `"sans"`,
`"serif"`, and `"mono"` for drawing text of those
types.  In `rgl`, the `"symbol"` family is not supported.

New font families can be defined using the low-level
function `r indexfns("rglFonts")`, or more simply
using the higher level function 
`r indexfns("rglExtrafonts")`.  The latter function
requires the `extrafont` package to be installed.

### par3d:  Miscellaneous graphical parameters

The `r indexfns("par3d")` function, modelled after the classic
graphics `r linkfn("par", pkg="graphics")` function, sets or reads 
a variety of different `rgl` internal
parameters, listed in the `r indexfns("rgl.par3d.names")`
variable.  All of them can be set except the ones in 
`r indexfns("rgl.par3d.readonly")`.  Some parameters are completely read-only; others are
fixed at the time the window is opened, and others may be changed 
at any time.

Name        |  Changeable?     | Description
----------- | ----- | -----------
antialias   | fixed | Amount of hardware antialiasing 
cex         |       | Default size for text
family      |       | Device-independent font family name; see `r linkfn("text3d", text="?text3d", pkg="rgl")`
font        |       | Integer font number
useFreeType |       | Should FreeType fonts be used if available?
fontname    | read-only | System-dependent font name set by `r linkfn("rglFonts")`
FOV         |       | Field of view, in degrees.  Zero means isometric perspective
ignoreExtent |      | Should `rgl` ignore the size of new objects when computing the bounding box?
skipRedraw   |      | Should `rgl` suppress updates to the display? 
maxClipPlanes | read-only | How many clip planes can be defined?
modelMatrix | read-only | The OpenGL ModelView matrix; partly set by `r indexfns("view3d")`
projMatrix  | read-only | The OpenGL Projection matrix
bbox        | read-only | Current bounding-box of the scene
viewport    |       | Dimensions in pixels of the scene within the window
windowRect  |           | Dimensions in pixels of the window on the whole screen
listeners   |           | Which subscenes respond to mouse actions in the current one
mouseMode   |       | What the mouse buttons do.  See [User interaction](#user-interaction)
observer    | read-only | The position of the observer; set by `r indexfns("observer3d")`
scale       |           | Rescaling for each coordinate; see `r linkfn("aspect3d")`
zoom        |           | Magnification of the scene

The deprecated `r deprecated("rgl.viewpoint")` function
should not be used.

### Default settings

The `r indexfns("r3dDefaults")` list and the `r indexfns("getr3dDefaults")` 
function control defaults in new windows opened by `r linkfn("open3d")`.  
The function looks for the variable in the user's global environment,
and if not found there, finds the one in the `rgl` namespace.  This
allows the user to override the default settings for new windows.

Once found, the `r3dDefaults` list provides initial values for `r linkfn("par3d")` parameters, as well as defaults for `r linkfn("material3d")` and `r linkfn("bg3d")` in components `"material"`
and `"bg"` respectively.


## Meshes:  Constructing Shapes 

`rgl` includes a number of functions to construct and display 
various solid shapes.  These generate objects of class `"shape3d"`,
`"mesh3d"` or `"shapelist3d"`.  The details of the classes are 
described below.  We start with functions to generate them.

### Specific solids

These functions generate specific shapes.  Optional arguments allow
attributes such as color or transformations to be specified.

Function                             | Description
------------------------------------ | -----------
`r indexfns(c("tetrahedron3d", "cube3d", "octahedron3d", "dodecahedron3d", "icosahedron3d"))`: | Platonic solids
`r indexfns(c("cuboctahedron3d", "oh3d"))`:             | other solids

```{r Solids}
cols <- rainbow(7)
layout3d(matrix(1:16, 4,4), heights=c(1,3,1,3))
text3d(0,0,0,"tetrahedron3d"); next3d()
shade3d(tetrahedron3d(col=cols[1])); next3d()
text3d(0,0,0,"cube3d"); next3d()
shade3d(cube3d(col=cols[2])); next3d()
text3d(0,0,0,"octahedron3d"); next3d()
shade3d(octahedron3d(col=cols[3])); next3d()
text3d(0,0,0,"dodecahedron3d"); next3d()
shade3d(dodecahedron3d(col=cols[4])); next3d()
text3d(0,0,0,"icosahedron3d"); next3d()
shade3d(icosahedron3d(col=cols[5])); next3d()
text3d(0,0,0,"cuboctahedron3d"); next3d()
shade3d(cuboctahedron3d(col=cols[6])); next3d()
text3d(0,0,0,"oh3d"); next3d()
shade3d(oh3d(col=cols[7]))
```

A very large collection of polyhedra is contained in the `Rpolyhedra` package.
This package was formerly on CRAN, and is perhaps still available on Github if
it has not returned to CRAN.

### Generating new shapes 

These functions generate new shapes:

Function                             | Description
------------------------------------ | -----------
`r indexfns("cylinder3d")`: | generate a tube or cylinder
`r indexfns("polygon3d")`:  | generate a flat polygon by triangulation
`r indexfns("extrude3d")`:  | generate an "extrusion" of a polygon
`r indexfns("turn3d")`:     | generate a solid of rotation
`r indexfns("ellipse3d")`:  | generate an ellipsoid in various ways
`r indexfns("mesh3d")`:      | generate a shape from indexed vertices
`r indexfns("shapelist3d")`: | generate a shape by combining other shapes
`r linkfn("as.mesh3d")`:   | a generic function; see below

A related function is `r indexfns("triangulate")`, which takes a
two dimensional polygon and divides it up into triangles using the 
"ear-clipping" algorithm.

The generic function `r indexfns("as.mesh3d")`
is provided to allow data structures produced by
other code to be converted to mesh
structures.  Currently the following classes
are supported:

Class |  Package | Description
----- |  ------- | -----------
`r indexfns("deldir", pkg = "deldir")` | `deldir` | Delaunay triangulations of irregular point clouds 
`r indexfns("triSht", pkg = "interp")` | `interp` | Also Delaunay triangulations
`r indexfns("ashape3d", pkg = "alphashape3d")` | `alphashape3d` | Alpha-shapes
`r linkfn("rglId")` | `rgl` | `rgl` object identifiers

The `r indexfns("checkDeldir")` function checks that a compatible version
of the `deldir` package is installed.

The default `r indexfns("as.mesh3d.default")` method is a simple way
to construct a mesh from a matrix of vertices; it can use
`r indexfns("mergeVertices")` (which can also be used on its own)
to merge repeated vertices within the matrix, allowing 
`r linkfn("addNormals")` to be used to give a smooth appearance.

The `r indexfns("as.tmesh3d")` generic is a variation that
guarantees the resulting object will have no quad entries.

Functions `r indexfns(c("tmesh3d","qmesh3d"))` are now obsolete;
use `r linkfn("mesh3d")` instead.

### The underlying class structure for shapes

`"shape3d"` is the basic abstract type.  Objects of this class can be
displayed by `r indexfns("shade3d")` (which shades faces), 
`r indexfns("wire3d")` (which draws edges), or `r indexfns("dot3d")`
(which draws points at each vertex.)  

`"mesh3d"` is a descendant type.  Objects of this type contain the following 
fields:

Field        | Meaning
------------ | ---------------
vb           | A 4 by n matrix of vertices in homogeneous coordinates.  Each column is a point.
ip           | (optional) A vector of vertex indices for points.
is           | (optional) A 2 by s matrix of vertex indices. Each column is a line segment.
it           | (optional) A 3 by t matrix of vertex indices.  Each column is a triangle.
ib           | (optional) A 4 by q matrix of vertex indices.  Each column is a quadrilateral.
material     | (optional) A list of material properties.
normals      | (optional) A matrix of the same shape as vb, containing normal vectors at each vertex.
texcoords    | (optional) A 2 by n matrix of texture coordinates corresponding to each vertex.
values       | (optional) A vector of length n holding values at each vertex
meshColor    | (optional) A text value indicating how colors and texture coordinates should be interpreted.
tags         | (optional) A vector added by some functions (e.g. `r linkfn("clipMesh3d")`) to relate output parts to input parts.


### Contouring and clipping shapes

These functions compute and plot contours of functions
on surfaces, or clip objects along a contour of a function.

Function                         | Description
-------------------------------- | -----------
`r indexfns("contourLines3d")`:  | draw contour lines on surface
`r indexfns("filledContour3d")`: | fill between contours on surface
`r indexfns("clipMesh3d")`:      | clip mesh object using curved boundary
`r indexfns("clipObj3d")`:       | clip general object using curved boundary

### Manipulating shapes

These functions manipulate and modify mesh objects:

Function                             | Description
------------------------------------ | -----------
`r indexfns("addNormals")`: | add normal vectors to make a shape look smooth
`r indexfns("subdivision3d")`: | add extra vertices to make it look even smoother
`r indexfns("merge.mesh3d", backticked("merge"))`: | merge mesh objects
`r indexfns("facing3d")`:   | subset of mesh facing "up"
`r indexfns("getBoundary3d")`: | get the boundary of a mesh object

The individual steps in `r linkfn("subdivision3d")` are also available:
`r indexfns(c("deform.mesh3d", "divide.mesh3d", "normalize.mesh3d"))`. These
are mainly intended for internal use.

## Multi-figure Layouts

`rgl` has several functions to support displaying multiple different
"subscenes" in the same window.  The high level functions are

Function                  | Description
------------------------- | -----------
`r indexfns("mfrow3d")`:  | Multiple figures (like `r linkfn("par", text = 'par("mfrow")', pkg="graphics")`
`r indexfns("layout3d")`: | Multiple figures (like `r linkfn("layout", pkg="graphics")`)
`r indexfns("next3d")`:   | Move to the next figure (like `r linkfn("plot.new", pkg="graphics")` or `r linkfn("frame", pkg="graphics")`)
`r indexfns("subsceneList")`: | List all the subscenes in the current layout
`r indexfns("clearSubsceneList")`: | Clear the current list and revert to the previous one

There are also lower level functions.

Function                           | Description
---------------------------------- | -----------
`r indexfns("newSubscene3d")`:     | Create a new subscene, with fine control over what is inherited from the parent
`r indexfns("currentSubscene3d")`: | Report on the active subscene
`r indexfns("subsceneInfo")`:      | Get information on current subscene
`r indexfns("useSubscene3d")`:     | Make a different subscene active
`r indexfns(c("addToSubscene3d", "delFromSubscene3d"))`: | Add objects to a subscene, or delete them
`r indexfns("gc3d")`:              | Do "garbage collection":  delete objects that are not displayed in any subscene

## Documents with `rgl` Scenes

The `rgl` package can produce output that can be embedded
in other documents.  The recommended way to do this has changed several
times over the years.  We will start with the current
recommendation, then list older methods.

### The recommended method

Currently the best way to embed an `rgl` scene in a document
is to produce the document in HTML using R Markdown.  Early in
the document, you should have code like this in one of the
setup code chunks:

````markdown
`r ''````{r echo=FALSE, include=FALSE}
library(rgl)
setupKnitr(autoprint = TRUE)
```
````

The call to `setupKnitr()` will install a number of hooks
and set options in `knitr` so that `rgl` code is handled properly.
The `autoprint = TRUE` argument makes `rgl` act in the document
almost the same way it would act in the console, or the way 
base graphics are handled by `knitr`:  If you print the value
of high level `rgl` functions, a plot will be inserted into the
output, but maybe only after low level modifications to it are
complete.  For example, this code block prints both
triangles and spheres in a single plot at the end:

```{r Autoprint}
xyz <- matrix(rnorm(27), ncol = 3)
triangles3d(xyz, col = rainbow(9))
spheres3d(xyz, col = rainbow(9), radius = 0.1)
```

There are a few differences if you have a complicated situation:

  -  The mechanism depends on the result of the `rgl` function
  calls being automatically printed.
  If the calls are in a loop or other code block where automatic
  printing doesn't happen, you'll need some trickery to get 
  things to print.  For example, this will print three plots:
  
```{r eval = FALSE}
plots <- NULL
for (i in 1:3) {
  plot3d(rnorm(10), rnorm(10), rnorm(10))
  plots <- htmltools::tagList(plots, rglwidget())
  close3d()
}
plots
```

  - It also depends on the fact that `rgl` functions return
  results using `lowlevel()` or `highlevel()` to mark which
  kind of plot they are.  If you are using a function from
  another package to produce the plot, you may need to insert
  an explicit call to one of those to get it to print.  Use `lowlevel()` if the function just modifies an existing plot, 
  `highlevel()` if it starts a new one.  For example,
  
```{r eval = FALSE}
foreignHigh()   # Produces a high level plot, but doesn't return
                # an appropriate value
highlevel()
foreignLow()    # Modifies the previous plot
lowlevel()
```  

This should display the output at the end of the code chunk, when
modifications are assumed complete.

### Producing PDF output

While some PDF previewers support interactive 3D graphics,
most don't.  To produce a screenshot of an `rgl` scene in an R Markdown
document with PDF output, simply follow the directions given
above.  The auto-printing will detect PDF output and use
`snapshot3d` to produce a PNG file to insert.  (See below
if you want to insert a different format of graphic.)

If you really need interactive output, see the 
`r linkfn("writeASY")` function.

### Manual insertion of plots

You may not want to use the `setupKnitr(autoprint = TRUE)`
method described above.  It is very new, and may still have
bugs; you may have an older document and not want to edit it
to work that way.

In this case, you can insert plots manually.  Use setup
code

````markdown
`r ''````{r echo=FALSE, include=FALSE}
library(rgl)
setupKnitr()
```
````

and call `rglwidget()` at top level whenever you want to 
insert a plot.

There are a couple of other differences in default behaviour if
you are not using `autoprint`:

 - By default, each code chunk continues the `rgl` scene from
earlier chunks.  You'll need an explicit `r linkfn("open3d")`
call to get a clean window.

 - Also by default, the `rgl` window is not closed at the end of
the chunk.  This probably doesn't matter, but you may find
you run out of memory if your scenes are really big.

### Older methods

The original way to insert an `rgl` scene in a document was to use
the deprecated `r deprecated("writeWebGL")` function to write HTML code to insert in a document. 
Later, `Sweave` and `knitr` hooks were added.  These are no longer supported, and you should update old documents to use the newer methods.  If you are
reading documents that suggest using those methods, let the
author know they need updating!
  
## Utility Functions


### User interaction

By default, `rgl` detects and handles mouse clicks within your scene,
and uses these to control its appearance.  You can find out the current
handlers using the following code:
```{r}
par3d("mouseMode")
```
The labels `c("left", "right", "middle")` refer to the buttons on
a three button mouse, or simulations of them on other mice.  `"wheel"` 
refers to the mouse wheel, and `"none"` refers to
actions that take place when the mouse is moved
without pressing any button.

The button actions generally correspond to click and drag operations.
Possible values for `"mouseMode"` for the mouse pointer or wheel are as follows:

Mode           | Description
-------------- | ---------
`"none"`       | No action
`"trackball"`  | The mouse acts as a virtual trackball. Clicking and dragging rotates the scene
`"xAxis"`, `"yAxis"`, `"zAxis"`      | Like `"trackball"`, but restricted to rotation about one axis
`"polar"`      | The mouse affects rotations by controlling polar coordinates directly
`"selecting"`  | The mouse is being used by the `r linkfn("select3d")` function
`"zoom"`       | The mouse zooms the display
`"fov"`        | The mouse affects perspective by changing the field of view
`"pull"`       | Rotating the mouse wheel towards the user "pulls the scene closer"
`"push"`       | The same rotation "pushes the scene away"
`"user"`       | A user action set by `r indexfns(c("setUserCallbacks", "rgl.setMouseCallbacks", "rgl.setWheelCallback"))`.  Use `r indexfns("rgl.getMouseCallbacks")` and `r indexfns("rgl.getWheelCallback")` to retrieve.

The following functions make use of the mouse for selection within a 
scene.

Function                     | Description
---------------------------- | -----------
`r indexfns("identify3d")`:   | like the classic graphics `r linkfn("identify", pkg="graphics")` function
`r indexfns("select3d")`:     | returns a function that tests whether a coordinate was selected
`r indexfns("selectpoints3d")`: | selects from specific objects
`r indexfns("hover3d")`:      | displays "hover" info about points

`r indexfns("selectionFunction3d")` produces the selection
function from information about the projection and mouse
selection region; it is used internally in the 
functions above.

The deprecated `r deprecated("rgl.select3d")` function is an obsolete version of
`select3d`, and `r indexfns("rgl.select")` is a low-level support
function.


### Animations

`rgl` has several functions that can be used to construct
animations.  These are based on functions that update the
scene according to the current real-world time, and repeated
calls to those.  The functions are:

Function               | Description
---------------------- | -------------
`r indexfns("play3d")`: | Repeatedly call the update function 
`r indexfns("spin3d")`: | Update the display by rotating at a constant rate
`r indexfns("par3dinterp")`: | Compute new values of some `r linkfn("par3d")` parameters by interpolation over time

See the `r linkfn("movie3d")` function for a way to output an animation
to a file on disk.  
Animations are not currently supported 
in the HTML written by `r linkfn("rglwidget")`, though the
`playwidget` function provides equivalent functionality.

### Integration with TCL/TK

There are three functions in `rgl` that support control of an `rgl`
scene using the TCL/TK framework.

Function               | Description
---------------------- | -------------
`r indexfns("tkspin3d")`: | Set up buttons in a window to control a scene 
`r indexfns("tkspinControl")`: | Embed the control buttons in a separate TCL/TK frame
`r indexfns("tkpar3dsave")`: | Create a dialog to interactively save mouse actions

These functions were formerly contained (without the `tk` prefixes on 
their names) in the `tkrgl` package.  That package is now deprecated.

### Exporting and importing scenes

`rgl` contains several functions to write scenes to disk for use
by other software, or to read them in.

In order from highest fidelity to lowest, the functions are:

Function    | Description
----------- | -------------
`r indexfns("scene3d")`:    | Save a scene to an R variable, which can be saved and reloaded
`r indexfns("rglwidget")`: | Prints as HTML and Javascript to display a scene in a web browser.  (See also [User Interaction in WebGL](WebGL.html).)
`r indexfns("writeASY")`:   | Write files for Asymptote
`r indexfns("writePLY")`:   | Write PLY files (commonly used in 3D printing)
`r indexfns(c("readOBJ", "writeOBJ"))`:   | Read or write OBJ files (commonly used in 3D graphics)
`r indexfns(c("readSTL", "writeSTL"))`:   | Read or write STL files (also common in 3D printing)
`r indexfns("as.rglscene")`: | Generic function, no methods in `rgl`

The [`rgl2gltf`](https://dmurdoch.github.io/rgl2gltf/dev/)
package can read or write GLTF and GLB files.  It includes an
`as.rglscene` method to convert GLTF objects to `rgl` scenes.  The
code in `rgl`'s `r indexfns("Buffer")` R6 class is based on the GLTF
format.  It is used by `r linkfn("rglwidget")` to make output
webpages somewhat smaller than they were previously.

There are also functions to save snapshots or other recordings of a scene,
without any 3D information being saved:

Function     | Description
------------ | -------------
`r indexfns("snapshot3d")`: | Save a PNG file bitmap of the scene
`r indexfns("rgl.postscript")`: | Save a Postscript, LaTeX, PDF, SVG or PGF vector rendering of the scene
`r indexfns("movie3d")`:    | Save a series of bitmaps to be assembled into a movie
`r indexfns("rgl.pixels")`: | Obtain pixel-level information about the scene in an R variable
`r indexfns("rgl.Sweave")`: | Driver function for inserting a snapshot into a Sweave document.
`r indexfns(c("hook_rgl", "hook_webgl"))`: | `knitr` hook functions for inserting images into a document.
`r indexfns("setupKnitr")`: | Function to set up `knitr` hooks

The `r indexfns("rgl.snapshot")` function is a low level
version of `snapshot3d()`; it requires that the `rgl` display
be onscreen and copies from there.  `snapshot3d()` tries to 
use the `webshot2` package so it will work even with no display.
The functions `r indexfns(c("rgl.Sweave.off", "Sweave.snapshot"))` are 
involved in Sweave processing and not normally called by users.


### Default display

There are two ways in which `rgl` scenes are normally
displayed within R.  The older one is in a dedicated
window.  In Unix-alikes this is an X11 window;  it is
a native window in Microsoft Windows.  On macOS, the
XQuartz system (see https://www.xquartz.org) needs to be
installed to support this.

To suppress this display, set `options(rgl.useNULL = TRUE)` before opening a new `rgl` window.  See
the help page for the `r indexfns("rgl.useNULL")` function
for how to set this before starting R.

The newer way to display a scene is by using WebGL
in a browser window or in the Viewer pane in RStudio.
To select this, set `options(rgl.printRglwidget = TRUE)`.
Each operation that would change the scene will 
return a value which triggers a new WebGL display
when printed.

### Working with WebGL scenes

You should use the following scheme for exporting a scene to a web page.  There's also an older scheme, which is no longer 
supported.

The recommended approach works with the `htmlwidgets` framework
(see http://www.htmlwidgets.org/).  In an R Markdown document
in `knitr`, use the 
`r linkfn("rglwidget")` function.  (You can also use
chunk option `webgl=TRUE`; we recommend the explicit use of
`rglwidget`.)  This approach also
allows display of `rgl` scenes in
[RStudio](https://posit.co/).  Besides 
`rgl` scenes, various controls for them can
be displayed, and there are a few utility functions
that can be useful:

Function                             | Description
------------------------------------ | -------------
`r indexfns("propertyControl")`:       | set individual properties
`r indexfns("clipplaneControl")`:       |  control a clipping plane
`r indexfns("subsetControl")`:       | control which objects are displayed
`r indexfns("ageControl")`:       | "age" vertices of an object
`r indexfns("vertexControl")`:       | control properties of vertices
`r indexfns("par3dinterpControl")`:  | WebGL control like `r linkfn("par3dinterp")`
`r indexfns("playwidget")`:       | display and automate controls
`r indexfns("toggleWidget")`:     | display a button to toggle some items
`r documentedfns <- c(documentedfns, "%>%");indexfns("pipe", text="%>%")`:    | `magrittr` pipe
`r indexfns(c("figHeight", "figWidth"))`:     | Dimensions of figures in R Markdown document
`r indexfns("rglShared")`:        | share data using `crosstalk` package
`r indexfns("rglMouse")`:         | change mouse mode in RGL scene
`r indexfns("asRow")`:            | arrange multiple objects in an HTML display
`r indexfns("getWidgetId")`:      | get the `elementId` from a widget

These functions work with the above scheme in Shiny apps:

Function                             | Description
------------------------------------ | -------------
`r indexfns("sceneChange")`:      | used in `Shiny` for large scene changes
`r indexfns(c("shinyGetPar3d", "shinySetPar3d"))`: | get or set `r linkfn("par3d")` values from Shiny
`r indexfns("shinyResetBrush")`:  | reset the mouse brush in Shiny

The `r linkfn("selectionFunction3d")` function is also likely
to be involved in mouse interactions when using Shiny.

Some functions are mainly for internal use:
`r indexfns(c("elementId2Prefix", "playwidgetOutput", "renderPlaywidget", "rglwidgetOutput", "renderRglwidget", "registerSceneChange"))`. 
More details are given in the vignette 
[User Interaction in WebGL](WebGL.html).
The functions `r indexfns(c("lowlevel", "highlevel", "rglId", "rgl.incrementID"))` are also for internal use,
marking function results for automatic
printing.  Finally, the function
`r indexfns("setUserShaders")` allows you to
use hand-written shaders in WebGL, and 
`r indexfns("getShaders")` allows you to see what
shader would be used.

### Working with the scene

`rgl` maintains internal structures for all the scenes it displays.
The following functions allow users to find information about them
and manipulate them.  In cases where there are both `*3d`
and `rgl.*` versions of functions, most users should use
the `*3d` version:  the `rgl.*` functions are more primitive
and are mainly intended for internal use.

Function                             | Description
------------------------------------ | -----------
`r indexfns("open3d")`:               | open a new window
`r indexfns("close3d")`:              | close the current window
`r indexfns("cur3d")`:                | id of the active device
`r indexfns("set3d")`:                | set a particular device to be active
`r indexfns("pop3d")`:     | delete objects from the scene
`r indexfns(c("clear3d"))`:            | delete all objects of certain classes
`r indexfns("ids3d")`:              | ids, types and tags of current objects
`r indexfns("tagged3d")`:                         | find tags or objects with tags

Some of these functions have alternate names for
back compatibility:  
`r deprecated(c("rgl.cur", "rgl.ids", "rgl.pop"))`. 
Either name will work,
but the `*3d` version is recommended for new code.
Some have deprecated versions:
`r deprecated(c("rgl.clear", "rgl.set"))`.  Those should not be called.

These functions are mainly intended for programming, and 
have no corresponding `*3d` counterparts:

Function | Description
------------- | -----------
`r indexfns("rgl.bringtotop")`: | bring the current window to the top
`r indexfns("rgl.dev.list")`:   | ids of all active devices
`r indexfns(c("rgl.attrib", "rgl.attrib.info", "rgl.attrib.count"))`: | attributes of objects in the scene
`r indexfns("rgl.projection")`: | return information about the current projection
`r indexfns(c("rgl.user2window", "rgl.window2user"))`: | convert between coordinates in the current projection

The `r indexfns("as.triangles3d")` generic function is intended to 
extract coordinates in a form suitable for passing to 
`r linkfn("triangles3d")`.  Currently a method is provided for
`r linkfn("rglId")` objects.

In addition to these, there are some deprecated functions which
should not be called: `r deprecated(c("rgl.init", "rgl.open", "rgl.close", "rgl.quit"))`.  


### Working with 3-D vectors

Most `rgl` functions work internally with "homogeneous" coordinates.
In this system, 3-D points are represented with 4 coordinates, generally
called (x, y, z, w).  The corresponding Euclidean point is
(x/w, y/w, z/w), if w is nonzero; zero values of w correspond to 
"points at infinity".  The advantage of this system is that
affine transformations including translations and perspective shifts
become linear transformations, with multiplication by a 4 by 4 matrix.

`rgl` has the following functions to work with homogeneous coordinates:

Function                             | Description
------------------------------------ | -----------
`r indexfns(c("asEuclidean", "asHomogeneous"))`: | convert between homogeneous and Euclidean coordinates when x, y and z are columns
`r indexfns(c("asEuclidean2", "asHomogeneous2"))`: | convert when x, y and z are rows
`r indexfns(c("rotate3d", "scale3d", "translate3d"))`:  | apply a transformation
`r indexfns("transform3d")`:  | apply a general transformation
`r indexfns(c("rotationMatrix", "scaleMatrix", "translationMatrix"))`:  | compute the transformation matrix
`r indexfns("identityMatrix")`: | return a 4 x 4 identity matrix
`r indexfns("projectDown")`: | a 3D to 2D projection down a vector

There is also a function `r indexfns("GramSchmidt")`, mainly for internal
use:  it does a Gram-Schmidt orthogonalization of a 3x3 matrix,
with some specializations for its use in `r linkfn("cylinder3d")`.

### Working with other packages

Sometimes it may be convenient to interactively
rotate a scene to a particular view, then 
display it in `lattice` or base graphics.
The `r indexfns("rglToLattice")` and `r indexfns("rglToBase")` functions support this.

For example, we first display the volcano data in `rgl`:
```{r fig.alt="Volcano in rgl", echo = 2:3}
close3d()
persp3d(volcano, col = "green")
```
This display is interactive, but we can reproduce the initial view using the
`lattice` `r linkfn("wireframe", pkg = "lattice")` or base graphics `r linkfn("persp", pkg = "graphics")` functions:
```{r eval=requireNamespace("orientlib", quietly = TRUE) && requireNamespace("lattice", quietly = TRUE), fig.alt=paste0("Volcano in ", c("lattice", "base"), "graphics.")}
# Only evaluated if the lattice & orientlib packages are installed
lattice::wireframe(volcano, col = "green", 
		   screen = rglToLattice())
angles <- rglToBase()
persp(volcano, col = "green", shade = TRUE,
      theta = angles$theta, phi = angles$phi)
```

Note that the `orientlib` package must be 
available for these functions to work.

### Creating `pkgdown` websites

The ["Using RGL in pkgdown web sites"](pkgdown.html) vignette
describes how to use `rgl` in a `pkgdown` web site.  The
utility functions `r indexfns(c("in_pkgdown", "in_pkgdown_example"))` can be used
to detect that `pkgdown` is being used.

### Working with `testthat`

The `testthat` package is widely used for unit tests
in packages.  Such tests are hard to write with `rgl`,
because the output is visual and interactive rather than
a simple value.  The `r indexfns("expect_known_scene")`,
`r indexfns("compare_proxy.mesh3d")` and 
`r indexfns("all.equal.mesh3d")` functions help with
this by removing system-dependent features of `rgl`
output.


### Working with Javascript

The WebGL displays created using `r linkfn("rglwidget")`
rely on a large body of Javascript code included in this
package.  To help in development of this code, the
`r indexfns("makeDependency")` function was written.
It may be useful in other packages that include Javascript.

### Other functions and objects

This section is for miscellaneous functions and objects 
that don't fall in any of the other categories in this document.

The `r indexfns("setGraphicsDelay")` function is designed
to work around what appears to be a bug on macOS:  if
a standard plot window is opened too quickly after an `rgl`
window, R can crash.  This function inserts a one second 
delay when it appears to be needed.

The `r indexfns("gltfTypes")` vector contains
constants used in OpenGL and glTF.

The `r indexfns("safe.dev.off")` function closes a base
graphics display and sets the previous display as current.

## Warning:  Work in Progress!

This vignette is always a work in progress.  Some aspects of the `rgl`
package are not described, or do not have examples.  There may
even be functions that are missed completely, if the following
list is not empty:
```{r echo=FALSE}
setdiff(ls("package:rgl"), c(documentedfns, deprecatedfns))
```

## Index of Functions

The following functions and constants are described in this document:<br>

```{r echo=FALSE, results="asis"}
writeIndex(cols = if (knitr::is_html_output()) 5 else 4)
```

---
title: "Using RGL in pkgdown web sites"
author: "Duncan Murdoch"
output:
  rmarkdown::html_vignette:
    fig_height: 5
    fig_width: 5
    toc: true
  pdf_document:
    fig_height: 5
    fig_width: 5
    toc: true
  html_document:
    default
vignette: >
  %\VignetteIndexEntry{Using RGL in pkgdown web sites}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r setup, include=FALSE}
if (!requireNamespace("rmarkdown", quietly = TRUE) ||
    !rmarkdown::pandoc_available("1.14")) {
  warning(call. = FALSE, "These vignettes assume rmarkdown and Pandoc
          version 1.14.  These were not found. Older versions will not work.")
  knitr::knit_exit()
}
knitr::opts_chunk$set(echo = TRUE, snapshot = FALSE, screenshot.force = FALSE)
library(rgl)
options(rgl.useNULL = TRUE)
setupKnitr(autoprint = TRUE)
```


## What is the problem?

[pkgdown](https://pkgdown.r-lib.org/) is an R package that
makes it easy to build a web site for your package.  However, the previous version 1.6.1 on CRAN didn't work so well for
packages whose examples use RGL or other packages like [leaflet](http://rstudio.github.io/leaflet/) that use
[htmlwidgets](http://www.htmlwidgets.org).  This document
described changes that support both of these.  The changes
are now incorporated into `pkgdown` 2.0.0, so everything but
user instructions has been removed.

## Using `RGL in examples.

If you use RGL in example code, they should
appear automatically in examples.
The RGL output is set up to mimic what would happen in a
`knitr` document that uses 
```{r eval=FALSE}
setupKnitr(autoprint = TRUE)
```
i.e. the output RGL commands will automatically be 
included in the display.  Low-level changes will be collected
into a single display:
```{r}
# Show regression plane with z as dependent variable
library(rgl)
open3d()
x <- rnorm(100)
y <- rnorm(100)
z <- 0.2*x - 0.3*y + rnorm(100, sd = 0.3)
fit <- lm(z ~ x + y)

plot3d(x, y, z, type = "s", col = "red", size = 1)

# No plot here, because of the planes3d() call below

coefs <- coef(fit)
a <- coefs["x"]
b <- coefs["y"]
c <- -1
d <- coefs["(Intercept)"]
planes3d(a, b, c, d, alpha = 0.5)
```

## Specifying the size of figures

By default, pkgdown generates standard plots which are wider
than they are high (according to the golden ratio).  Often
RGL plots look better with equal width and height, since
the contents may be rotated by the user.

To specify the size of plots in pkgdown, you use the `figures`
entry in `_pkgdown.yml`.  The defaults are similar to
```
figures:
  dev: ragg::agg_png
  dpi: 96
  dev.args: []
  fig.ext: png
  fig.width: 7
  fig.height: ~
  fig.retina: 2
  fig.asp: 1.618
  bg: NA
  other.parameters: []
```

By default RGL uses these parameters as well, but allows you 
to override any of `fig.width`, `fig.height` and `fig.asp` by
specifying an `rgl` entry in `other.parameters`.  For example:
```
figures:
  fig.width: 5
  other.parameters:
    rgl:
      fig.asp: 1
```
This will make all plots have a width of 5 inches
and will make RGL plots square.

---
title: "Deprecating the `rgl.*` interface"
author: "Duncan Murdoch"
date: "`r Sys.Date()`"
output: 
  rmarkdown::html_vignette:
    toc: true
vignette: >
  %\VignetteIndexEntry{Deprecating the `rgl.*` interface}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Introduction

Since at least 2004, `rgl` has had two interfaces for many 
of the primitive functions:  `rgl.*` and `*3d`.  For example,
to draw points you could use `rgl.points()` or `points3d()`.
With the upcoming version 1.0.0 release of `rgl`, most of
the duplication will be removed.  The first step will be
to deprecate a large number of `rgl.*` functions so they give
warnings when they are called, and a few months later 
they will be removed from the package exports.

This document describes the differences and changes needed 
by users of the `rgl.*` interface.

# Differences between the interfaces

## Opening a window

The `rgl.open()` function has a single argument, `useNULL`.
If set to `TRUE`, the NULL `rgl` device will  be used.  The
`par3d()` settings will be set to their defaults.

The `open3d()` function has arguments
```r
function(..., 
         params = getr3dDefaults(), 
         useNULL = rgl.useNULL(), 
         silent = FALSE	)
```
and allows `par3d()` values to be specified,
and uses the `r3dDefaults` variable to set default values
for `par3d()`, `material3d()`, and `bg3d()`.  Initially
`r3dDefaults` is defined as
```r
list(userMatrix = rotationMatrix(290*pi/180, 1, 0, 0),
     mouseMode = c("none", "trackball", "zoom", "fov", "pull"),
     FOV = 30,
     family = "sans",
     bg = list(color="white",
               fogtype = "none"),
     material = list(color="black", fog = TRUE)
)
```
Users can create their own default lists; e.g. to get the
same result as `rgl.open()` would give, use
```r
open3d(params = list())
```
or
```r
r3dDefaults <- list()
open3d()
```

## Material properties

The `rgl.material()` function has a large number of parameters.  The pre-deprecation arguments were:
```r
function(
  color        = "white",
  alpha        = 1.0,
  lit          = TRUE, 
  ambient      = "black",
  specular     = "white", 
  emission     = "black", 
  shininess    = 50.0, 
  smooth       = TRUE,
  texture      = NULL, 
  textype      = "rgb", 
  texmipmap    = FALSE, 
  texminfilter = "linear", 
  texmagfilter = "linear",
  texenvmap    = FALSE,
  front        = "filled", 
  back         = "filled",
  size         = 3.0,
  lwd          = 1.0, 
  fog          = TRUE,
  point_antialias = FALSE,
  line_antialias = FALSE,
  depth_mask   = TRUE,
  depth_test   = "less",
  polygon_offset = c(0.0, 0.0),
  margin = "",
  floating = FALSE,
  tag = "",
  blend = c("src_alpha", "one_minus_src_alpha"),
  col,
  ...
)
```

Thus a call like `rgl.material(color = "black")` will set
the color to black, and will also set all of the other
parameters to the default values listed above.

On the other hand, the arguments to `material3d()` are
```r
function (..., id = NULL)  
```
Calling `material3d(color = "black")` will set
the color to black and leave all other parameters unchanged.

## Primitive shapes

The primitive shapes (points etc.) can be set using calls
like `rgl.points(x, y, z, color = "black")` or 
`points3d(x, y, z, color = "black")`.  

The first difference is
that `rgl.*` primitives will call `rgl.material()` to set
the material properties:  in this example `color` will be set
to `black`, and all other parameters will be set to their 
defaults.  The `*3d` versions of the primitives use
`material3d()` to set material properties, so only those that
were specified will be changed, and the original values
will be restored afterwards.

The second difference is what happens if there is no
window already open.  The `rgl.*` functions will call
`rgl.open()` (ignoring `r3dDefaults`), whereas the `*3d`
functions will call `open3d()`.

# Why deprecate `rgl.*`?

Both of the systems worked, but they do not work together.
For example, calling `rgl.points()` will have carry-on effects
on later `points3d()` calls, whereas each `points3d()` call
will just draw the points, it won't affect future calls.

Users have found this confusing, and it makes their code hard
to debug, and the `rgl` package hard to maintain.  The `*3d` 
interface is more flexible, and more similar to the base graphics
interface in R, so I've decided it will be the only one 
available going forward.

## Some `rgl.*` functions are not deprecated

There will still be some `rgl.*` functions in the package.
These are functions that are mainly intended for 
programming, such as `rgl.attrib()` and `rgl.user2window()`, and
a few legacy functions like `rgl.Sweave()` supporting
older approaches of using `rgl`.  

In a few cases
both function versions are identical
(`rgl.cur`, `rgl.ids`, and `rgl.pop` are identical
to `cur3d`, `ids3d` and `pop3d` respectively),
and for those the `rgl.*` versions will be kept,
but the documentation will concentrate on the `*3d` functions.


# My package uses `rgl.*`.  What do I need to do?

If your package is using `rgl.*` functions, the first
step is to just make the substitutions suggested by the
deprecation warning message.  For example, if you use `rgl.points(rnorm(10), rnorm(10), rnorm(10))` try using
`points3d(rnorm(10), rnorm(10), rnorm(10))` instead.
In most cases this will give you what you want.  In
some cases more changes will be needed.

## `rgl.open` and `rgl.material`

See above if you were using these.

## Textures

The default color after `rgl.open()` was white, whereas
with `open3d()` the default color is black, with a white
background.  Textures multiplicatively 
modify the color of the object,
so after `open3d()`, a texture on an object will still appear 
black.  Explicitly specifying `color = "white"` when a texture
is used will fix this.

## `rgl.surface()`

The arguments to `rgl.surface()` and `surface3d()` functions
are different.  The argument lists are
```r
rgl.surface( x, z, y, 
             coords = 1:3,  ..., 
             normal_x = NULL, normal_y = NULL, normal_z = NULL,
             texture_s = NULL, texture_t = NULL)
surface3d(x, y = NULL, z = NULL,
          ...,
          normal_x = NULL, normal_y = NULL,
          normal_z = NULL,
          texture_s = NULL, texture_t=NULL)
```
Notice that the arguments are in a different order.
Another difference is that `rgl.surface()` expects the surface
to be defined in the `y` coordinate and viewed
in the orientation produced by `rgl.open()`, not the one produced by
`open3d()`.  Up until very recently, `surface3d()` didn't allow both `x` and `z` to be vectors.  

The excellent [`rayshader` package](https://www.rayshader.com) 
used the convention
that the `y` argument held the surface, so the `y` direction 
should point up.  Using `view3d(theta = 45, phi = 45)`
(which it was already doing) gives a reasonable view.

## Lists of material names and par3d properties

Many functions in `rgl` and other packages use `...` to set
material or `par3d` properties in a call, and for some,
`...` will
contain other optional arguments.  Some packages used the
argument list of `rgl.material()` to identify the 
material property names.  Going forward, packages should 
use the
variables 
```r
rgl.material.names
rgl.material.readonly
```
These are character variables holding
all the material property names.  (`rgl.material.names`
contains all names; `rgl.material.readonly` is the
read-only subset of that list.)  There are also variables
```r
rgl.par3d.names
rgl.par3d.readonly
```
which give the same information for `par3d`.  

Since these variables are recently added, you will need to add
a dependence on `rgl (>= 0.111.5)` if you use them. 



## Others

If you have particular problems adapting other `rgl.*` to the
`*3d` interface, please post them as issues on https://github.com/dmurdoch/rgl/issues .  I'll
explain how to get what you want or fix things in `rgl` so 
you can do it.

---
title: "User Interaction in WebGL"
author: "Duncan Murdoch"
date: "`r format(Sys.time(), '%B %d, %Y')`"
output:
  rmarkdown::html_vignette:
    toc: true
    fig_width: 5
    fig_height: 5
vignette: >
  %\VignetteIndexEntry{User Interaction in WebGL} 
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteDepends{crosstalk}
---


```{r setup, echo=FALSE, results="asis"}
source("setup.R")
setupKnitr(autoprint = FALSE)
set.seed(123)
```

## Introduction

This document describes how to embed `rgl` scenes in HTML documents
and use embedded Javascript to 
control a WebGL display in an HTML document.  For more 
general information about `rgl`, see [rgl Overview](rgl.html).

We assume that the HTML document is produced from R markdown
source using `knitr` or `rmarkdown`.  This format mixes
text with Markdown markup with chunks of R code.  There
is a limited amount of discussion of other methods.

There are two ways to embed an `rgl` scene in the document. 
The newest one is recommended:  call `r linkfn("setupKnitr")`
with argument `autoprint = TRUE`
early in the document.  This will set things up to be
quite similar to the way standard 2D graphics are included
by `knitr`, i.e. it will detect the fact that you've 
drawn something, and just include it automatically.

If `autoprint = FALSE` is used or no call is made to
`setupKnitr()`, an explicit call to `r linkfn("rglwidget")` will
produce a "widget" which can
be embedded into your document by printing it.  This document
uses that method.

Older methods (e.g. `writeWebGL` or various hooks) that were
used before `rgl` version
0.102.0 are no longer supported.


## Browser support

Most browsers now support WebGL, but in some browsers it may be disabled by
default.  See https://get.webgl.org for help on a number of
different browsers.  

## Examples

We start with a simple plot of the iris data.  We
insert a code chunk and call the `r linkfn("rglwidget")` 
function with optional argument `elementId`.  This allows later
Javascript code to refer to the image.  We also 
save the object ids from the plot, so that they
can be manipulated later.  (The first example in [Controls](#controls) uses tags instead of saving the ids.)
```{r elementId}
library(rgl)
plotids <- with(iris, plot3d(Sepal.Length, Sepal.Width, Petal.Length, 
                  type="s", col=as.numeric(Species)))
rglwidget(elementId = "plot3drgl")
```

Next we insert a button to toggle the display of the data.
```{r}
toggleWidget(sceneId = "plot3drgl", ids = plotids["data"], label = "Data")
```

The `sceneId` is the same as the `elementId` we used
in `rglwidget()`, the `ids` are the object ids of the
objects that we'd like to toggle, and the `label` is the
label shown on the button.  To find the names in the 
`plotids` variable, apply `names()` or `unclass()`:
```{r}
names(plotids)
unclass(plotids)
```

## Using `magrittr` or base pipes

It can be error-prone to set the `elementId` in the
`rglwidget()` to match the `sceneId` in the `toggleWidget()` (or `playwidget()`, described below).
In the usual case where both are intended to appear
together, [`magrittr`](https://CRAN.R-project.org/package=magrittr)-style pipes can be used quite flexibly:
the first argument of the control widget accepts 
the result of `rglwidget()` (or other control widgets),
and the `controllers` argument of `rglwidget()` accepts
control widgets.  In R 4.1.0, the new base pipe
operator `|>` should be usable in the same way.

For example,

```{r Pipes}
rglwidget() %>%
toggleWidget(ids = plotids["data"], label = "Data")
```

If you have R 4.1.0 or greater, this should do the same:

```{r eval=FALSE}
rglwidget() |>
toggleWidget(ids = plotids["data"], label = "Data")
```

You can swap the order of button and scene; use the `magrittr` dot (or the `=>` syntax in base pipes)
to pass the `toggleWidget` to `rglwidget` in the `controllers` argument:

```{r "Control before widget"}
toggleWidget(NA, ids = plotids["data"], label = "Data") %>%
rglwidget(controllers = .) 
```

or using R 4.1.0 or later,

```{r eval=FALSE}
toggleWidget(NA, ids = plotids["data"], label = "Data") |> 
  w => rglwidget(controllers = w) 
```

## Controls

We have seen how to change the contents of the plot using `r indexfns("toggleWidget")`.  We can do
more elaborate displays. 
For example, we can redo the previous plot, but with the
three species as separate "spheres" objects and buttons to
toggle them:
```{r "Toggle subsets"}
clear3d() # Remove the earlier display

with(subset(iris, Species == "setosa"), 
     spheres3d(Sepal.Length, Sepal.Width, Petal.Length, 
                  col=as.numeric(Species),
                  radius = 0.211,
                  tag = "setosa"))
with(subset(iris, Species == "versicolor"), 
     spheres3d(Sepal.Length, Sepal.Width, Petal.Length, 
               col=as.numeric(Species),
     	       radius = 0.211, 
     	       tag = "versicolor"))
with(subset(iris, Species == "virginica"), 
     spheres3d(Sepal.Length, Sepal.Width, Petal.Length, 
               col=as.numeric(Species),
     	       radius = 0.211,
     	       tag = "virginica"))
aspect3d(1,1,1)
decorate3d(tag = "axes")
rglwidget() %>%
toggleWidget(tags = "setosa") %>%
toggleWidget(tags = "versicolor") %>%
toggleWidget(tags = "virginica") %>%
toggleWidget(tags = "axes") %>% 
asRow(last = 4)
```

Since we skipped the `label` argument, the buttons are
labelled with the values of the tags.  The `asRow` function is discussed `r linkfn("asRow", "below")`.

`toggleWidget()` is actually a convenient wrapper for
two functions:  `r indexfns("playwidget")` and `r indexfns("subsetControl")`.  `playwidget()` adds
the button to the web page (and can also add sliders,
do animations, etc.), while `subsetControl()` chooses
a subset of objects to display.

### `subsetControl`

For a more general example, we could use a slider to
select several subsets of the data in the iris display.
For example,

```{r Slider}
rglwidget() %>%
playwidget(start = 0, stop = 3, interval = 1,
	   subsetControl(1, subsets = list(
	   			 Setosa = tagged3d("setosa"),
	   			 Versicolor = tagged3d("versicolor"),
	   			 Virginica = tagged3d("virginica"),
	   			 All = tagged3d(c("setosa", "versicolor", "virginica"))
	   			 )))
```

There are several other "control" functions.  

### `par3dinterpControl`

`r indexfns("par3dinterpControl")` approximates the result of `r linkfn("par3dinterp")`.

For example, the following code (similar to the `r linkfn("play3d")`
example) rotates the scene in a complex way.

```{r "par3dinterpControl()"}
M <- r3dDefaults$userMatrix
fn <- par3dinterp(time = (0:2)*0.75, userMatrix = list(M,
                                      rotate3d(M, pi/2, 1, 0, 0),
                                      rotate3d(M, pi/2, 0, 1, 0)) )
rglwidget() %>%
playwidget(par3dinterpControl(fn, 0, 3, steps=15),
 	   step = 0.01, loop = TRUE, rate = 0.5)
```

Some things to note:  The generated Javascript slider has 300 increments,
so that motion appears smooth.  However, storing 300 `userMatrix` values
would take up a lot of space, so we use interpolation
in the Javascript code.  However, the Javascript code can only do 
linear interpolation, not the more complex spline-based SO(3)
interpolation done by `r linkfn("par3dinterp")`.  Because of this,
we need to output 15 steps from `r linkfn("par3dinterpControl")`
so that the distortions of linear interpolation are not visible.  

### `propertyControl`

`r indexfns("propertyControl")` is a more general function to set
the value of properties of the scene. Currently most
properties are supported, but use does require knowledge
of the internal implementation.

### `clipplaneControl`

`r indexfns("clipplaneControl")` allows the user to control
the location of a clipping plane by moving a slider.  

### `vertexControl`

Less general than `r linkfn("propertyControl")` is
`r indexfns("vertexControl")`.  This function sets attributes
of individual vertices in a scene.  For example, to set the
x-coordinate of the closest point in the setosa group, and modify
its colour from black to white,

```{r "vertexControl()"}
setosavals <- subset(iris, Species == "setosa")
which <- which.min(setosavals$Sepal.Width)
init <- setosavals$Sepal.Length[which]
rglwidget() %>%
playwidget(
  vertexControl(values = matrix(c(init,   0,   0,   0, 
                                     8,   1,   1,   1), 
                                nrow = 2, byrow = TRUE),
                attributes = c("x", "red", "green", "blue"),
                vertices = which, tag = "setosa"),
	step = 0.01)
```

### `ageControl`	     

A related function is `r indexfns("ageControl")`, though it uses
a very different specification of the attributes.
It is used when the slider controls the "age" of the scene, 
and attributes of vertices change with their age. 

To illustrate we will
show a point moving along a curve. We
give two `ageControl` calls in a list; the first
one controls the colour of the trail, the second controls
the position of the point:

```{r "ageControl()"}
time <- 0:500
xyz <- cbind(cos(time/20), sin(time/10), time)
lineid <- plot3d(xyz, type="l", col = "black")["data"]
sphereid <- spheres3d(xyz[1, , drop=FALSE], radius = 8, col = "red")
rglwidget() %>%
playwidget(list(
  ageControl(births = time, ages = c(0, 0, 50),
		colors = c("gray", "red", "gray"), objids = lineid),
  ageControl(births = 0, ages = time,
		vertices = xyz, objids = sphereid)),
  start = 0, stop = max(time) + 20, rate = 50,
  components = c("Reverse", "Play", "Slower", "Faster",
                 "Reset", "Slider", "Label"),
  loop = TRUE)
```

### `rglMouse`

While not exactly a control in the sense of the other
functions in this section, the `r indexfns("rglMouse")`
function is used to add an HTML control to a display
to allow the user to select the mouse mode.  

For example, the display below initially allows selection
of particular points, but the mouse mode may be changed
to let the user rotate the display for a another 
view of the scene.

```{r crosstalk,eval = requireNamespace("crosstalk", quietly=TRUE)}
# This example requires the crosstalk package
# We skip it if crosstalk is not available. 

ids <- with(iris, plot3d(Sepal.Length, Sepal.Width, Petal.Length, 
                  type="s", col=as.numeric(Species)))
par3d(mouseMode = "selecting")
rglwidget(shared = rglShared(ids["data"])) %>% 
rglMouse()
```

The `rglShared()` call used here is described `r linkfn("rglShared", "below")`.

## Layout of the display

Many `rgl` displays will contain several elements:  one or more
`rgl` scenes and controls.  Internally `rgl` uses
the `combineWidgets` function from the
[`manipulateWidget`](https://github.com/rte-antares-rpackage/manipulateWidget)
package.  

The `rgl` package provides 3 convenience functions for arranging
displays.  We have already met the first:  the `magrittr` pipe, `%>%`.
When the display is constructed as a
single object using pipes, the objects in the pipeline
will be arranged in a single column.   

The second convenience function is  `r indexfns("asRow")`.  This takes
as input a list of objects or a `combineWidgets` object (perhaps 
the result of a pipe), and rearranges (some of) them into a 
horizontal row.  As in the `r linkfn("toggleWidget", "toggleWidget example")`, 
the `last` argument can be used to limit the actions of `asRow` to the
specified number of components.  (If `last = 0`, all objects are stacked:  this can be useful if some of them are not from the `rgl`
package, so piping doesn't work for them.)

Finally, `r indexfns("getWidgetId")` can be used to extract the 
HTML element ID from an HTML widget.  This is useful when combining
widgets that are not all elements of the same pipe, as in the
`crosstalk` example below.

If these convenience functions are not sufficient, you can call
`r linkfn("combineWidgets", text = "manipulateWidget::combineWidgets", pkg = "manipulateWidget")` or
other functions from `manipulateWidget` for more flexibility in
the display arrangements.


## Integration with `crosstalk`

The [`crosstalk`](https://rstudio.github.io/crosstalk/) package allows
widgets to communicate with each other.  Currently it supports selection
and filtering of observations.

`rgl` can send, receive and display these messages.  An `rgl` display
may have several subscenes, each displaying different datasets.  Each object in the 
scene is potentially a shared dataset in the `crosstalk` sense.

The linking depends on the `r indexfns("rglShared")` function.  Calling `rglShared(id)`,
where `id` is the `rgl` id value for an object in the current scene,
creates a shared data object containing the coordinates of the vertices of 
the `rgl` object.  This object is passed to `r linkfn("rglwidget")` in the `shared` 
argument.  It can also be passed to other widgets that accept shared data,
linking the two displays.

If a shared data object has been created in some other way, it can be linked to
a particular `rgl` `id` value by copying its `key` and `group` properties
as shown in the example below.

```{r "rglShared()",eval=requireNamespace("crosstalk", quietly = TRUE)}
# This example requires the crosstalk package.  
# We skip it if crosstalk is not available. 

library(crosstalk)
sd <- SharedData$new(mtcars)
ids <- plot3d(sd$origData(), col = mtcars$cyl, type = "s")
# Copy the key and group from existing shared data
rglsd <- rglShared(ids["data"], key = sd$key(), group = sd$groupName())
rglwidget(shared = rglsd) %>%
asRow("Mouse mode: ", rglMouse(getWidgetId(.)), 
      "Subset: ", filter_checkbox("cylinderselector", 
		                "Cylinders", sd, ~ cyl, inline = TRUE),
      last = 4, colsize = c(1,2,1,2), height = 60)
```

If multiple objects in the `rgl` scene need to be considered
as shared data, you can pass the results of several `rglShared()`
calls in a list, i.e.  `rglwidget(shared = <list>)`.  The key
values will be assumed to be shared across datasets; if this is
not wanted, use a prefix or some other means to make sure they
differ between objects.

If the same `rgl` id is used in more than one `rglShared()` object,
it will respond to messages from all of them.  This may lead to
undesirable behaviour as one message cancels the previous one.

## Low level controls

We repeat the initial plot from this document:

```{r plot3d2}
plotids <- with(iris, plot3d(Sepal.Length, Sepal.Width, Petal.Length, 
                  type="s", col=as.numeric(Species)))
subid <- currentSubscene3d()
rglwidget(elementId="plot3drgl2")
```

We might like a button on the web page to cause a change to the
display, e.g. a rotation of the plot.  First we add buttons, with
the "onclick" event set to a function described below:

    <button type="button" onclick="rotate(10)">Forward</button>
    <button type="button" onclick="rotate(-10)">Backward</button>

which produces these buttons: 
<button type="button" onclick="rotate(10)">Forward</button>
<button type="button" onclick="rotate(-10)">Backward</button>

We stored the subscene number that is currently active in
`subid` in the code chunk above, and use it as `r rinline("subid")`
in the script below.  `knitr` substitutes the value
when it processes the document.

The `rotate()` function uses the Javascript function `document.getElementById` to retrieve the `<div>` component
of the web page containing the scene.  It will have a
component named `rglinstance` which contains information about the scene that we can modify:

    <script type="text/javascript">
    var rotate = function(angle) {
      var rgl = document.getElementById("plot3drgl2").rglinstance;
      rgl.getObj(`r rinline("subid",
                           script=TRUE)`).par3d.userMatrix.rotate(angle, 0,1,0);
      rgl.drawScene();
    };
    </script>
    
<script type="text/javascript">
var rotate = function(angle) {
  var rgl = document.getElementById("plot3drgl2").rglinstance;
  rgl.getObj(`r subid`).par3d.userMatrix.rotate(angle, 0,1,0);
  rgl.drawScene();
};
</script>

If we had used `webGL=TRUE` in the chunk header,
the `knitr` WebGL support would create a global object with a name of the form `<chunkname>rgl`.  For example,  if the code chunk
was named `plot3d2`, the object
would be called `plot3d2rgl`, and this code would work:

    <script type="text/javascript">
    var rotate = function(angle) {
      plot3d2rgl.getObj(`r rinline("subid",
                           script=TRUE)`).par3d.userMatrix.rotate(angle, 0,1,0);
      plot3d2rgl.drawScene();
    };
    </script>

## Index 

The following functions are described in this document:<br>

```{r echo=FALSE, results="asis"}
writeIndex(cols = 5)
```

---
title: "A Note on Transparency"
author: "Duncan Murdoch"
date: "24/10/2020"
output:
  rmarkdown::html_vignette:
    fig_height: 5
    fig_width: 5
    toc: true
  pdf_document:
    fig_height: 5
    fig_width: 5
    toc: true
  html_document:
    default
vignette: >
  %\VignetteIndexEntry{A Note on Transparency}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r setup, include=FALSE}
if (!requireNamespace("rmarkdown", quietly = TRUE) ||
    !rmarkdown::pandoc_available("1.14")) {
  warning(call. = FALSE, "These vignettes assume rmarkdown and Pandoc
          version 1.14.  These were not found. Older versions will not work.")
  knitr::knit_exit()
}
knitr::opts_chunk$set(echo = TRUE, snapshot = FALSE, screenshot.force = FALSE)
library(rgl)
options(rgl.useNULL = TRUE)
setupKnitr(autoprint = TRUE)

M <- structure(c(0.997410774230957, 0.0707177817821503, -0.0130676832050085, 
0, -0.0703366547822952, 0.99714070558548, 0.02762770652771, 0, 
0.0149840852245688, -0.0266370177268982, 0.999532878398895, 0, 
0, 0, 0, 1), .Dim = c(4L, 4L))
```

## Introduction

When drawing transparent surfaces, `rgl` tries to sort objects
from back to front to get better rendering of transparency.
However, it doesn't sort each pixel separately, so some pixels
end up drawn in the incorrect order.  This note
describes the consequences of that error, and 
suggests remedies.

## Correct Drawing

We'll assume that the standard `glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)` blending is used.  That is,
when drawing with transparency $\alpha$, the new colour is
mixed at proportion $\alpha$ with the colour previously
drawn (which gets weight $1-\alpha$.)

This note is concerned with what happens when two transparent
objects are drawn at the same location.  We suppose the further
one has transparency $\alpha_1$, and the closer one 
has transparency $\alpha_2$.  If they are drawn in the correct
order (further first), the three colours (background, further
object, closer object) should end up mixed in proportions
$C = [(1-\alpha_1)(1-\alpha_2), \alpha_1(1-\alpha_2), \alpha_2]$
respectively.

## Incorrect sorting

If the objects are drawn in the wrong order, the actual 
proportions of each colour will be 
$N = [(1-\alpha_1)(1-\alpha_2),\alpha_1, \alpha_2(1-\alpha_1)]$ 
if no masking occurs. 

`rgl` currently defaults to depth masking using `glDepthMask(GL_TRUE)`.  This means that depths are saved 
when the objects are drawn, and if an attempt is made to 
draw the further object after the closer one (i.e. as here),
the further object will be culled, and the proportions will be
$M = [(1-\alpha_2), 0, \alpha_2]$.

## Mask or Not?

The question is:  which is better, `glDepthMask(GL_TRUE)` or
`glDepthMask(GL_FALSE)`?  One way to measure this is to 
measure the distance between $C$ and the incorrect proportions.
(This is unlikely to match perceptual distance, which will
depend on the colours as well, but we need something. Some
qualitative comments below.)

So we have 

$$|C-N|^2 = 2\alpha_1^2\alpha_2^2,$$

and 

$$|C-M|^2 = 2\alpha_1^2(1-\alpha_2)^2.$$

Thus the error is larger with $N$ when $\alpha_2 > 1 / 2$, and 
larger with $M$ when $\alpha_2 < 1 / 2$.  The value of $\alpha_1$
doesn't affect the preference, though small values of $\alpha_1$ will be associated with smaller errors.  

Depending on the colours of the background and the two
objects, this recommendation could be modified.  For example,
if the two objects are the same colour (or very close),
it doesn't really matter how the 2nd and 3rd proportions
are divided up, and $N$ will be best because it gets the
background proportion exactly right.


## Recommendation

Typically in `rgl` we don't know which object will be closer 
and which one will be further, so we can't base our choice on
a single $\alpha_i$.  The recommendation would be to use
all small levels of `alpha` and disable masking, or use all
large values of `alpha` and retain masking.

## Example

The classic example of an impossible to sort scene involves
three triangles arranged cyclicly so each one is behind one and in front
of one of the others (based on https://paroj.github.io/gltut/Positioning/Tut05%20Overlap%20and%20Depth%20Buffering.html).

```{r}
theta <- 2*pi*c(0:2, 4:6, 8:10)/12
x <- cos(theta)
y <- sin(theta)
z <- rep(c(0,0,1), 3)
xyz <- cbind(x, y, z)
xyz <- xyz[c(1,2,6, 4,5,9, 7,8,3),]
open3d()
par3d(userMatrix = M)
triangles3d(xyz, col = rep(c("red", "green", "blue"), each = 3))
```

To see the effect of the calculations above, consider the following four displays.

```{r fig.width=8}
open3d()
par3d(userMatrix = M)
layout3d(matrix(1:9, ncol = 3, byrow=TRUE),
         widths = c(1,2,2), heights = c(1, 3,3), 
         sharedMouse = TRUE)
text3d(0,0,0, " ")
next3d()
text3d(0,0,0, "depth_mask = TRUE")
next3d()
text3d(0,0,0, "depth_mask = FALSE")
next3d()
text3d(0,0,0, "alpha = 0.7")
next3d()
triangles3d(xyz, col = rep(c("red", "green", "blue"), each = 3), alpha = 0.7, depth_mask = TRUE)
next3d()
triangles3d(xyz, col = rep(c("red", "green", "blue"), each = 3), alpha = 0.7, depth_mask = FALSE)
next3d()
text3d(0,0,0, "alpha = 0.3")
next3d()
triangles3d(xyz, col = rep(c("red", "green", "blue"), each = 3), alpha = 0.3, depth_mask = TRUE)
next3d()
triangles3d(xyz, col = rep(c("red", "green", "blue"), each = 3), alpha = 0.3, depth_mask = FALSE)
```

As you rotate the figures, you can see imperfections in
rendering.  On the right, the last drawn appears to be on top,
while on the left, the first drawn appears more opaque than
it should.  

In the figure below, the three triangles each have different transparency, and each use the recommended setting:

```{r}
open3d()
par3d(userMatrix = M)
triangles3d(xyz[1:3,], col = "red", alpha = 0.3, depth_mask = FALSE)
triangles3d(xyz[4:6,], col = "green", alpha = 0.7, depth_mask = TRUE)
triangles3d(xyz[7:9,], col = "blue", depth_mask = TRUE)
```

In this figure, all three triangles are the same colour,
only lighting affects the display:
```{r fig.width=8}
open3d()
par3d(userMatrix = M)
layout3d(matrix(1:9, ncol = 3, byrow=TRUE),
         widths = c(1,2,2), heights = c(1, 3,3), 
         sharedMouse = TRUE)
text3d(0,0,0, " ")
next3d()
text3d(0,0,0, "depth_mask = TRUE")
next3d()
text3d(0,0,0, "depth_mask = FALSE")
next3d()
text3d(0,0,0, "alpha = 0.7")
next3d()
triangles3d(xyz, col = "red", alpha = 0.7, depth_mask = TRUE)
next3d()
triangles3d(xyz, col = "red", alpha = 0.7, depth_mask = FALSE)
next3d()
text3d(0,0,0, "alpha = 0.3")
next3d()
triangles3d(xyz, col = "red", alpha = 0.3, depth_mask = TRUE)
next3d()
triangles3d(xyz, col = "red", alpha = 0.3, depth_mask = FALSE)
```

Here `depth_mask = FALSE` seems to be the right choice in both cases.

---
title: "User Interaction in WebGL (updated)"
author: "Duncan Murdoch"
date: "`r format(Sys.time(), '%B %d, %Y')`"
output:
  markdown::html_format:
    options:
      toc: true
  rmarkdown::html_vignette:
    toc: true
    fig_width: 5
    fig_height: 5
vignette: >
  %\VignetteIndexEntry{User Interaction in WebGL (updated)} 
  %\VignetteEngine{knitr::knitr}
  %\VignetteDepends{crosstalk}
---


```{r setup, echo=FALSE, results="asis"}
source("setup.R")
setupKnitr(autoprint = FALSE)
set.seed(123)
```

## Introduction

This document describes how to embed `rgl` scenes in HTML documents
and use embedded Javascript to 
control a WebGL display in an HTML document.  For more 
general information about `rgl`, see [rgl Overview](rgl.html).

We assume that the HTML document is produced from R markdown
source using `knitr` or `rmarkdown`.  This format mixes
text with Markdown markup with chunks of R code.  There
is a limited amount of discussion of other methods.

There are two ways to embed an `rgl` scene in the document. 
The newest one is recommended:  call `r linkfn("setupKnitr")`
with argument `autoprint = TRUE`
early in the document.  This will set things up to be
quite similar to the way standard 2D graphics are included
by `knitr`, i.e. it will detect the fact that you've 
drawn something, and just include it automatically.

If `autoprint = FALSE` is used or no call is made to
`setupKnitr()`, an explicit call to `r linkfn("rglwidget")` will
produce a "widget" which can
be embedded into your document by printing it.  This document
uses that method.

Older methods (e.g. `writeWebGL` or various hooks) that were
used before `rgl` version
0.102.0 are no longer supported.


## Browser support

Most browsers now support WebGL, but in some browsers it may be disabled by
default.  See https://get.webgl.org for help on a number of
different browsers.  

## Examples

We start with a simple plot of the iris data.  We
insert a code chunk and call the `r linkfn("rglwidget")` 
function with optional argument `elementId`.  This allows later
Javascript code to refer to the image.  We also 
save the object ids from the plot, so that they
can be manipulated later.  (The first example in [Controls](#controls) uses tags instead of saving the ids.)
```{r elementId}
library(rgl)
plotids <- with(iris, plot3d(Sepal.Length, Sepal.Width, Petal.Length, 
                  type="s", col=as.numeric(Species)))
rglwidget(elementId = "plot3drgl")
```

Next we insert a button to toggle the display of the data.
```{r}
toggleWidget(sceneId = "plot3drgl", ids = plotids["data"], label = "Data")
```

The `sceneId` is the same as the `elementId` we used
in `rglwidget()`, the `ids` are the object ids of the
objects that we'd like to toggle, and the `label` is the
label shown on the button.  To find the names in the 
`plotids` variable, apply `names()` or `unclass()`:
```{r}
names(plotids)
unclass(plotids)
```

## Using `magrittr` or base pipes

It can be error-prone to set the `elementId` in the
`rglwidget()` to match the `sceneId` in the `toggleWidget()` (or `playwidget()`, described below).
In the usual case where both are intended to appear
together, [`magrittr`](https://CRAN.R-project.org/package=magrittr)-style pipes can be used quite flexibly:
the first argument of the control widget accepts 
the result of `rglwidget()` (or other control widgets),
and the `controllers` argument of `rglwidget()` accepts
control widgets.  In R 4.1.0, the new base pipe
operator `|>` should be usable in the same way.

For example,

```{r Pipes}
rglwidget() %>%
toggleWidget(ids = plotids["data"], label = "Data")
```

If you have R 4.1.0 or greater, this should do the same:

```{r eval=FALSE}
rglwidget() |>
toggleWidget(ids = plotids["data"], label = "Data")
```

You can swap the order of button and scene; use the `magrittr` dot (or the `=>` syntax in base pipes)
to pass the `toggleWidget` to `rglwidget` in the `controllers` argument:

```{r "Control before widget"}
toggleWidget(NA, ids = plotids["data"], label = "Data") %>%
rglwidget(controllers = .) 
```

or using R 4.1.0 or later,

```{r eval=FALSE}
toggleWidget(NA, ids = plotids["data"], label = "Data") |> 
  w => rglwidget(controllers = w) 
```

## Controls

We have seen how to change the contents of the plot using `r indexfns("toggleWidget")`.  We can do
more elaborate displays. 
For example, we can redo the previous plot, but with the
three species as separate "spheres" objects and buttons to
toggle them:
```{r "Toggle subsets"}
clear3d() # Remove the earlier display

with(subset(iris, Species == "setosa"), 
     spheres3d(Sepal.Length, Sepal.Width, Petal.Length, 
                  col=as.numeric(Species),
                  radius = 0.211,
                  tag = "setosa"))
with(subset(iris, Species == "versicolor"), 
     spheres3d(Sepal.Length, Sepal.Width, Petal.Length, 
               col=as.numeric(Species),
     	       radius = 0.211, 
     	       tag = "versicolor"))
with(subset(iris, Species == "virginica"), 
     spheres3d(Sepal.Length, Sepal.Width, Petal.Length, 
               col=as.numeric(Species),
     	       radius = 0.211,
     	       tag = "virginica"))
aspect3d(1,1,1)
decorate3d(tag = "axes")
rglwidget() %>%
toggleWidget(tags = "setosa") %>%
toggleWidget(tags = "versicolor") %>%
toggleWidget(tags = "virginica") %>%
toggleWidget(tags = "axes") %>% 
asRow(last = 4)
```

Since we skipped the `label` argument, the buttons are
labelled with the values of the tags.  The `asRow` function is discussed `r linkfn("asRow", "below")`.

`toggleWidget()` is actually a convenient wrapper for
two functions:  `r indexfns("playwidget")` and `r indexfns("subsetControl")`.  `playwidget()` adds
the button to the web page (and can also add sliders,
do animations, etc.), while `subsetControl()` chooses
a subset of objects to display.

### `subsetControl`

For a more general example, we could use a slider to
select several subsets of the data in the iris display.
For example,

```{r Slider}
rglwidget() %>%
playwidget(start = 0, stop = 3, interval = 1,
	   subsetControl(1, subsets = list(
	   			 Setosa = tagged3d("setosa"),
	   			 Versicolor = tagged3d("versicolor"),
	   			 Virginica = tagged3d("virginica"),
	   			 All = tagged3d(c("setosa", "versicolor", "virginica"))
	   			 )))
```

There are several other "control" functions.  

### `par3dinterpControl`

`r indexfns("par3dinterpControl")` approximates the result of `r linkfn("par3dinterp")`.

For example, the following code (similar to the `r linkfn("play3d")`
example) rotates the scene in a complex way.

```{r "par3dinterpControl()"}
M <- r3dDefaults$userMatrix
fn <- par3dinterp(time = (0:2)*0.75, userMatrix = list(M,
                                      rotate3d(M, pi/2, 1, 0, 0),
                                      rotate3d(M, pi/2, 0, 1, 0)) )
rglwidget() %>%
playwidget(par3dinterpControl(fn, 0, 3, steps=15),
 	   step = 0.01, loop = TRUE, rate = 0.5)
```

Some things to note:  The generated Javascript slider has 300 increments,
so that motion appears smooth.  However, storing 300 `userMatrix` values
would take up a lot of space, so we use interpolation
in the Javascript code.  However, the Javascript code can only do 
linear interpolation, not the more complex spline-based SO(3)
interpolation done by `r linkfn("par3dinterp")`.  Because of this,
we need to output 15 steps from `r linkfn("par3dinterpControl")`
so that the distortions of linear interpolation are not visible.  

### `propertyControl`

`r indexfns("propertyControl")` is a more general function to set
the value of properties of the scene. Currently most
properties are supported, but use does require knowledge
of the internal implementation.

### `clipplaneControl`

`r indexfns("clipplaneControl")` allows the user to control
the location of a clipping plane by moving a slider.  

### `vertexControl`

Less general than `r linkfn("propertyControl")` is
`r indexfns("vertexControl")`.  This function sets attributes
of individual vertices in a scene.  For example, to set the
x-coordinate of the closest point in the setosa group, and modify
its colour from black to white,

```{r "vertexControl()"}
setosavals <- subset(iris, Species == "setosa")
which <- which.min(setosavals$Sepal.Width)
init <- setosavals$Sepal.Length[which]
rglwidget() %>%
playwidget(
  vertexControl(values = matrix(c(init,   0,   0,   0, 
                                     8,   1,   1,   1), 
                                nrow = 2, byrow = TRUE),
                attributes = c("x", "red", "green", "blue"),
                vertices = which, tag = "setosa"),
	step = 0.01)
```

### `ageControl`	     

A related function is `r indexfns("ageControl")`, though it uses
a very different specification of the attributes.
It is used when the slider controls the "age" of the scene, 
and attributes of vertices change with their age. 

To illustrate we will
show a point moving along a curve. We
give two `ageControl` calls in a list; the first
one controls the colour of the trail, the second controls
the position of the point:

```{r "ageControl()"}
time <- 0:500
xyz <- cbind(cos(time/20), sin(time/10), time)
lineid <- plot3d(xyz, type="l", col = "black")["data"]
sphereid <- spheres3d(xyz[1, , drop=FALSE], radius = 8, col = "red")
rglwidget() %>%
playwidget(list(
  ageControl(births = time, ages = c(0, 0, 50),
		colors = c("gray", "red", "gray"), objids = lineid),
  ageControl(births = 0, ages = time,
		vertices = xyz, objids = sphereid)),
  start = 0, stop = max(time) + 20, rate = 50,
  components = c("Reverse", "Play", "Slower", "Faster",
                 "Reset", "Slider", "Label"),
  loop = TRUE)
```

### `rglMouse`

While not exactly a control in the sense of the other
functions in this section, the `r indexfns("rglMouse")`
function is used to add an HTML control to a display
to allow the user to select the mouse mode.  

For example, the display below initially allows selection
of particular points, but the mouse mode may be changed
to let the user rotate the display for a another 
view of the scene.

```{r crosstalk,eval = requireNamespace("crosstalk", quietly=TRUE)}
# This example requires the crosstalk package
# We skip it if crosstalk is not available. 

ids <- with(iris, plot3d(Sepal.Length, Sepal.Width, Petal.Length, 
                  type="s", col=as.numeric(Species)))
par3d(mouseMode = "selecting")
rglwidget(shared = rglShared(ids["data"])) %>% 
rglMouse()
```

The `rglShared()` call used here is described `r linkfn("rglShared", "below")`.

## Layout of the display

Many `rgl` displays will contain several elements:  one or more
`rgl` scenes and controls.  Internally `rgl` uses
the `combineWidgets` function from the
[`manipulateWidget`](https://github.com/rte-antares-rpackage/manipulateWidget)
package.  

The `rgl` package provides 3 convenience functions for arranging
displays.  We have already met the first:  the `magrittr` pipe, `%>%`.
When the display is constructed as a
single object using pipes, the objects in the pipeline
will be arranged in a single column.   

The second convenience function is  `r indexfns("asRow")`.  This takes
as input a list of objects or a `combineWidgets` object (perhaps 
the result of a pipe), and rearranges (some of) them into a 
horizontal row.  As in the `r linkfn("toggleWidget", "toggleWidget example")`, 
the `last` argument can be used to limit the actions of `asRow` to the
specified number of components.  (If `last = 0`, all objects are stacked:  this can be useful if some of them are not from the `rgl`
package, so piping doesn't work for them.)

Finally, `r indexfns("getWidgetId")` can be used to extract the 
HTML element ID from an HTML widget.  This is useful when combining
widgets that are not all elements of the same pipe, as in the
`crosstalk` example below.

If these convenience functions are not sufficient, you can call
`r linkfn("combineWidgets", text = "manipulateWidget::combineWidgets", pkg = "manipulateWidget")` or
other functions from `manipulateWidget` for more flexibility in
the display arrangements.


## Integration with `crosstalk`

The [`crosstalk`](https://rstudio.github.io/crosstalk/) package allows
widgets to communicate with each other.  Currently it supports selection
and filtering of observations.

`rgl` can send, receive and display these messages.  An `rgl` display
may have several subscenes, each displaying different datasets.  Each object in the 
scene is potentially a shared dataset in the `crosstalk` sense.

The linking depends on the `r indexfns("rglShared")` function.  Calling `rglShared(id)`,
where `id` is the `rgl` id value for an object in the current scene,
creates a shared data object containing the coordinates of the vertices of 
the `rgl` object.  This object is passed to `r linkfn("rglwidget")` in the `shared` 
argument.  It can also be passed to other widgets that accept shared data,
linking the two displays.

If a shared data object has been created in some other way, it can be linked to
a particular `rgl` `id` value by copying its `key` and `group` properties
as shown in the example below.

```{r "rglShared()",eval=requireNamespace("crosstalk", quietly = TRUE)}
# This example requires the crosstalk package.  
# We skip it if crosstalk is not available. 

library(crosstalk)
sd <- SharedData$new(mtcars)
ids <- plot3d(sd$origData(), col = mtcars$cyl, type = "s")
# Copy the key and group from existing shared data
rglsd <- rglShared(ids["data"], key = sd$key(), group = sd$groupName())
rglwidget(shared = rglsd) %>%
asRow("Mouse mode: ", rglMouse(getWidgetId(.)), 
      "Subset: ", filter_checkbox("cylinderselector", 
		                "Cylinders", sd, ~ cyl, inline = TRUE),
      last = 4, colsize = c(1,2,1,2), height = 60)
```

If multiple objects in the `rgl` scene need to be considered
as shared data, you can pass the results of several `rglShared()`
calls in a list, i.e.  `rglwidget(shared = <list>)`.  The key
values will be assumed to be shared across datasets; if this is
not wanted, use a prefix or some other means to make sure they
differ between objects.

If the same `rgl` id is used in more than one `rglShared()` object,
it will respond to messages from all of them.  This may lead to
undesirable behaviour as one message cancels the previous one.

## Low level controls

We repeat the initial plot from this document:

```{r plot3d2}
plotids <- with(iris, plot3d(Sepal.Length, Sepal.Width, Petal.Length, 
                  type="s", col=as.numeric(Species)))
subid <- currentSubscene3d()
rglwidget(elementId="plot3drgl2")
```

We might like a button on the web page to cause a change to the
display, e.g. a rotation of the plot.  First we add buttons, with
the "onclick" event set to a function described below:

    <button type="button" onclick="rotate(10)">Forward</button>
    <button type="button" onclick="rotate(-10)">Backward</button>

which produces these buttons: 
<button type="button" onclick="rotate(10)">Forward</button>
<button type="button" onclick="rotate(-10)">Backward</button>

We stored the subscene number that is currently active in
`subid` in the code chunk above, and use it as `r rinline("subid")`
in the script below.  `knitr` substitutes the value
when it processes the document.

The `rotate()` function uses the Javascript function `document.getElementById` to retrieve the `<div>` component
of the web page containing the scene.  It will have a
component named `rglinstance` which contains information about the scene that we can modify:

    <script type="text/javascript">
    var rotate = function(angle) {
      var rgl = document.getElementById("plot3drgl2").rglinstance;
      rgl.getObj(`r rinline("subid",
                           script=TRUE)`).par3d.userMatrix.rotate(angle, 0,1,0);
      rgl.drawScene();
    };
    </script>
    
<script type="text/javascript">
var rotate = function(angle) {
  var rgl = document.getElementById("plot3drgl2").rglinstance;
  rgl.getObj(`r subid`).par3d.userMatrix.rotate(angle, 0,1,0);
  rgl.drawScene();
};
</script>

If we had used `webGL=TRUE` in the chunk header,
the `knitr` WebGL support would create a global object with a name of the form `<chunkname>rgl`.  For example,  if the code chunk
was named `plot3d2`, the object
would be called `plot3d2rgl`, and this code would work:

    <script type="text/javascript">
    var rotate = function(angle) {
      plot3d2rgl.getObj(`r rinline("subid",
                           script=TRUE)`).par3d.userMatrix.rotate(angle, 0,1,0);
      plot3d2rgl.drawScene();
    };
    </script>

## Index 

The following functions are described in this document:<br>

```{r echo=FALSE, results="asis"}
writeIndex(cols = 5)
```

---
title: "rgl Overview"
author: "Duncan Murdoch"
date: "`r format(Sys.time(), '%B %d, %Y')`"
output:
  rmarkdown::html_vignette:
    fig_height: 5
    fig_width: 5
    toc: yes
  html_document:
    df_print: paged
    toc: yes
  pdf_document:
    fig_height: 5
    fig_width: 5
    toc: yes
vignette: >
  %\VignetteIndexEntry{rgl Overview}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r setup, echo=FALSE, results="asis"}
source("setup.R")
setupKnitr(autoprint = TRUE)
set.seed(123)
```



## Introduction

The `rgl` package is used to produce interactive 3-D plots.  It contains
high-level graphics commands modelled loosely after classic R graphics,
but working in three dimensions.  It also contains low level structure
inspired by (but incompatible with) the `grid` package.

This document gives an overview.  See the help pages for details.

For installation instructions, see the `README` file in the top
level directory of the source tarball 
`r paste0("rgl_", packageVersion("rgl"), ".tar.gz")` (or a later
version).


### About this document

This document was written in R Markdown, using the `knitr` package
for production.  It corresponds to rgl version 
`r packageVersion("rgl")`.

Most of the highlighted function names are HTML links.
The internal links should work in any browser; the links to
help topics should work if you view the vignette from
within the R help system.

The document includes WebGL figures.  To view these, you must have
Javascript and WebGL enabled in your browser.  Some older browsers
may not support this -- see https://get.webgl.org for tests
and links to a discussion.


## Basics and High Level Functions

The `r indexfns("plot3d")` function
plots points within an rgl window.  It is similar to the classic 
`r linkfn("plot", pkg="graphics")` function,
but works in 3 dimensions.

For example
```{r}
with(iris, plot3d(Sepal.Length, Sepal.Width, Petal.Length, 
                  type="s", col=as.numeric(Species)))
```

\noindent
can be used to plot three columns of the `iris` data.
Allowed plot types include `"p", "l", "h", "s"`,
meaning points, lines, segments from z=0, and spheres.  There's
a lot of flexibility in specifying the coordinates; the 
`r linkfn("xyz.coords", pkg = "grDevices")` function
from the `grDevices` package is used for this.

You can use your mouse to manipulate the plot.  The default is that
if you click and hold with the left mouse button, you can rotate 
the plot by dragging it.  The right mouse button is used to resize
it, and the middle button changes the perspective in the point of view.

If you call `r linkfn("plot3d")` again, it will overwrite the current plot.
To open a new graphics window, use `r linkfn("open3d")`.
 
The other high level function is `r indexfns("persp3d")` to draw surfaces.
It is
similar to the classic `r linkfn("persp", pkg = "graphics")`
function, but with greater flexibility.
First, any of `x`, `y` or `z`
can be specified using matrices, not just `z`.  This allows parametric
surfaces to be plotted. 
An even simpler specification is possible:  `x` may be a function,
in which case `persp3d` will work out the grid itself.  See 
`r linkfn("persp3d.function", text="?persp3d.function", pkg="rgl")`
for details.  For example, the `MASS` package estimates
Gamma parameters using maximum likelihood in a 
`r linkfn("fitdistr", text="?MASS::fitdistr", pkg="MASS")` example.
Here we show the log likelihood surface. 
```{r persp3d, fig.height=3, fig.width=6, fig.keep="last"}
library(MASS)
# from the fitdistr example
set.seed(123)
x <- rgamma(100, shape = 5, rate = 0.1)
fit <- fitdistr(x, dgamma, list(shape = 1, rate = 0.1), lower = 0.001)
loglik <- function(shape, rate) sum(dgamma(x, shape=shape, rate=rate, 
                                           log=TRUE))
loglik <- Vectorize(loglik)
xlim <- fit$estimate[1]+4*fit$sd[1]*c(-1,1)
ylim <- fit$estimate[2]+4*fit$sd[2]*c(-1,1)

mfrow3d(1, 2, sharedMouse = TRUE)
persp3d(loglik, 
        xlim = xlim, ylim = ylim,
        n = 30)
zlim <- fit$loglik + c(-qchisq(0.99, 2)/2, 0)
next3d()
persp3d(loglik, 
        xlim = xlim, ylim = ylim, zlim = zlim,
        n = 30)
```
 On the left, the whole surface over a range of the parameters; on the right, only the 
parts of the surface with log likelihood values near the maximum.

Note:  this example used the `knitr` hook functions 
(see `r linkfn("setupKnitr")`) to insert the
scene into this vignette; the previous example used the `rglwidget`
function.  We generally
recommend the newer `r linkfn("rglwidget")` approach.

Note that both `plot3d` and `persp3d` are generic functions,
with the following methods defined:
```{r}
methods(plot3d)
methods(persp3d)
```


## Adding Graphical Elements

### Primitive shapes

Just as we have `r linkfn("points", pkg="graphics")` and 
`r linkfn("lines", pkg="graphics")` in classic graphics, there are a number
of low level functions in `rgl` to add graphical elements to the
currently active plot.  The "primitive" shapes are those that are 
native to OpenGL:

Function                      | Description
----------------------------- | -----------
`r indexfns("points3d")`:     | adds points
`r indexfns("lines3d")`:      | adds lines
`r indexfns("segments3d")`:   | adds line segments
`r indexfns("triangles3d")`:  | adds triangles
`r indexfns("quads3d")`:      | adds quadrilaterals

Each of the above functions takes arguments `x`, `y` and `z`, again
using `r linkfn("xyz.coords", pkg="grDevices")` for flexibility.
They group successive entries
as necessary.  For example, the `r linkfn("triangles3d")` function takes each
successive triple of points as the vertices of a triangle.

You can use these functions to annotate the current graph, or to 
construct a figure from scratch.

### Constructed shapes

`rgl` also has a number of objects which it constructs
from the primitives.

Function                      | Description
----------------------------- | -----------
`r indexfns(c("text3d", "texts3d"))`: | adds text
`r indexfns("abclines3d")`:   | adds straight lines to plot (like `abline`)
`r indexfns("arc3d")`:        | adds spherical arcs or spirals to plot
`r indexfns("planes3d")`:     | adds planes to plot
`r indexfns("clipplanes3d")`: | add clipping planes to plot
`r indexfns(c("sprites3d", "particles3d"))`: | add sprites (fixed shapes or images) to plot
`r indexfns("spheres3d")`:    | adds spheres
`r indexfns(c("surface3d", "terrain3d"))`:    | a surface (as used in `r linkfn("persp3d")`)
`r indexfns("arrow3d")`:      | add an arrow to a scene
`r indexfns("pch3d")`:	      | draw base-style plotting symbols
`r indexfns("plotmath3d")`:   | used by `r linkfn("text3d")` for math text

### Axes and other "decorations"

The following low-level functions control the look of the graph:

Function                             | Description
------------------------------------ | -----------
`r indexfns(c("axes3d", "axis3d"))`: | add axes to plot
`r indexfns(c("box3d", "bbox3d"))`:               | add box around plot
`r indexfns("title3d")`:             | add title to plot
`r indexfns("mtext3d")`:             | add marginal text to plot
`r indexfns("decorate3d")`:          | add multiple "decorations" (scales, etc.) to plot
`r indexfns("aspect3d")`:            | set the aspect ratios for the plot
`r indexfns(c("bg3d", "bgplot3d"))`: | set the background of the scene
`r indexfns("show2d")`:              | show a 2D plot or image in a 3D scene
`r indexfns("legend3d")`:            | set a legend for the scene
`r indexfns("grid3d")`:              | add a reference grid to a graph
`r indexfns("thigmophobe3d")`:       | choose label positions to avoid overlap

For example, to plot three random triangles, one could use
```{r fig.width=3, fig.height=3}
triangles3d(cbind(x=rnorm(9), y=rnorm(9), z=rnorm(9)), col = "green")
decorate3d()
bg3d("lightgray")
aspect3d(1,1,1)
```

Besides the `*3d` functions mentioned above, there are even lower-level
functions 
`r indexfns(c("rgl.primitive", "rgl.points", "rgl.linestrips", "rgl.lines", "rgl.triangles",
"rgl.quads", "rgl.texts", "rgl.abclines", "rgl.planes", "rgl.bg",
"rgl.clipplanes", "rgl.bbox", "rgl.spheres", "rgl.sprites", "rgl.surface"))`.  
You should avoid using these functions, which do not
work well with the higher level `*3d` functions.  See the 
`r linkfn("r3d", text="?r3d", pkg="rgl")` help
topic for details.

## Controlling the Look of the Scene

### Lighting

In most scenes, objects are "lit", meaning that their appearance
depends on their position and orientation relative to lights
in the scene.  The lights themselves don't normally show up,
but their effect on the objects does.

Use the `r indexfns("light3d")` function to 
specify the position and characteristics of a light.
Lights may be infinitely distant, or may be embedded
within the scene.  Their characteristics include 
`ambient`, `diffuse`, and `specular` components, all
defaulting to white.  The `ambient` component appears
the same from any direction.  The `diffuse` component 
depends on the angle between the surface and the light, 
while the `specular` component also takes the viewer's
position into account.

The `r indexfns("rgl.light")` function is a lower-level
function with different defaults; users should normally
use `r linkfn("light3d")`.

### Materials

The mental model used in `rgl` is that the objects being shown
in scenes are physical objects in space, with material properties
that affect how light reflects from them (or is emitted by them).
These are mainly controlled by the `r indexfns("material3d")` function,
or by arguments to other functions that are passed to it.

The material properties that can be set by calls to `material3d` are
described in detail in the 
`r linkfn("material3d", text="?material3d", pkg="rgl")` help page.
Here we give an overview.

Property  | Default | Meaning 
--------- | ------- | ------------------
color     | white   | vector of surface colors to apply to successive vertices for diffuse light
alpha     | 1       | transparency:  0 is invisible, 1 is opaque
lit       | TRUE    | whether lighting calculations should be done
ambient   | black   | color in ambient  light
specular  | white   | color in specular light
emission  | black   | color emitted by the surface
shininess | 50      | controls the specular lighting:  high values look shiny
smooth    | TRUE    | whether shading should be interpolated between vertices
texture   | NULL    | optional path to a "texture" bitmap to be displayed on the surface
front, back | fill  | should polygons be filled, or outlined? 
size      | 3       | size of points in pixels
lwd       | 1       | width of lines in pixels

Other properties include 
"texmipmap", "texmagfilter", "texminfilter", "texenvmap", "fog", 
"point\_antialias", "line\_antialias", "depth\_mask", "depth\_test" and "polygon_offset"; 
see `r linkfn("material3d", "the help page", pkg = "rgl")` for details.

There is also an `r indexfns("rgl.material")` function that works
at a lower level; users should normally avoid it.

### Textures

As described in the previous section, one of the material
properties is `texture`, the name of a bitmap file (in `.png`
format) containing an image to be displayed on the surface. 
This section gives more details about textures.

In `OpenGL`, each vertex in a polygon may be associated with a
particular location in the bitmap.  The interior of the
polygon interpolates within the bitmap.  There are two 
conventions in `rgl` functions for specifying these coordinates.

Functions which specify primitives (`r linkfn("triangles3d")`, etc.) accept an optional matrix argument
`texcoords` which gives `s` (horizontal) and `t` (vertical)
locations within the bitmap in columns with one row per vertex.
The coordinates are `(0,0)` for the lower left, and `(1,1)`
for the upper right.  If values outside this range are given,
the image repeats, i.e. `(1.1, 1.1)` would specify the same
point in the image as `(0.1, 0.1)`.

Other functions such as `r linkfn("surface3d")` that take
matrices for each vertex coordinate accept texture coordinates
as matrices as well, in arguments `texture_s` and `texture_t`.

For example, the 
following code displays four copies of a 2D plot on a quad, 
because the texture coordinates run from 0 to 2 in both
`s` and `t`:

```{r}
filename <- tempfile(fileext = ".png")
png(filename = filename)
plot(rnorm(1000), rnorm(1000))
dev.off()
open3d()
xyz <- cbind(c(0,1,1,0), 0, c(0,0,1,1))
quads3d(xyz, texture = filename, texcoords = xyz[,c(1, 3)]*2, col = "white", specular = "black")
```

Some other notes:

 - The color in the figure above was specified to be white.  By default,
the colors in the bitmap will modify the colour of the
quad.  If `col` is black (a common default), you won't see anything.
 - On the other hand, you usually don't want specular 
 reflections (which show up as glare).  Setting `specular`
 to black prevents those.
 - How the bitmap is handled is controlled by the material
 property `"textype"`.  The default is `"rgb"`, which 
 takes the red-green-blue colours from the bitmap and
 uses them to modify the corresponding colours in the polygon.  Other possibilities for `"textype"` are described
 in `r linkfn("material3d", "the material3d help page", pkg = "rgl")`.
 - The other `"tex*"` material properties control how 
 the interpolation within the image is done.
 - Modern `OpenGL` supports 1- and 3-dimensional textures;
 these are not supported in `rgl`.
 
### par3d:  Miscellaneous graphical parameters

The `r indexfns("par3d")` function, modelled after the classic
graphics `r linkfn("par", pkg="graphics")` function, sets or reads 
a variety of different `rgl` internal
parameters.  Some parameters are completely read-only; others are
fixed at the time the window is opened, and others may be changed 
at any time.

Name        |  Changeable?     | Description
----------- | ----- | -----------
antialias   | fixed | Amount of hardware antialiasing 
cex         |       | Default size for text
family      |       | Device-independent font family name; see `r linkfn("text3d", text="?text3d", pkg="rgl")`
font        |       | Integer font number
useFreeType |       | Should FreeType fonts be used if available?
fontname    | read-only | System-dependent font name set by `r indexfns("rglFonts")`
FOV         |       | Field of view, in degrees.  Zero means isometric perspective
ignoreExtent |      | Should `rgl` ignore the size of new objects when computing the bounding box?
skipRedraw   |      | Should `rgl` suppress updates to the display? 
maxClipPlanes | read-only | How many clip planes can be defined?
modelMatrix | read-only | The OpenGL ModelView matrix; partly set by `r indexfns("view3d")` or the obsolete `r indexfns("rgl.viewpoint")`
projMatrix  | read-only | The OpenGL Projection matrix
bbox        | read-only | Current bounding-box of the scene
viewport    |       | Dimensions in pixels of the scene within the window
windowRect  |           | Dimensions in pixels of the window on the whole screen
listeners   |           | Which subscenes respond to mouse actions in the current one
mouseMode   |       | What the mouse buttons do.  See `r linkfn("mouseMode", '<code>"mouseMode"</code>')`
observer    | read-only | The position of the observer; set by `r indexfns("observer3d")`
scale       |           | Rescaling for each coordinate; see `r linkfn("aspect3d")`
zoom        |           | Magnification of the scene

### Default settings

The `r indexfns("r3dDefaults")` list and the `r indexfns("getr3dDefaults")` 
function control defaults in new windows opened by `r linkfn("open3d")`.  
The function looks for the variable in the user's global environment,
and if not found there, finds the one in the `rgl` namespace.  This
allows the user to override the default settings for new windows.

Once found, the `r3dDefaults` list provides initial values for `r linkfn("par3d")` parameters, as well as defaults for `r linkfn("material3d")` and `r linkfn("bg3d")` in components `"material"`
and `"bg"` respectively.


## Meshes:  Constructing Shapes 

`rgl` includes a number of functions to construct and display 
various solid shapes.  These generate objects of class `"shape3d"`,
`"mesh3d"` or `"shapelist3d"`.  The details of the classes are 
described below.  We start with functions to generate them.

### Specific solids

These functions generate specific shapes.  Optional arguments allow
attributes such as colour or transformations to be specified.

Function                             | Description
------------------------------------ | -----------
`r indexfns(c("tetrahedron3d", "cube3d", "octahedron3d", "dodecahedron3d", "icosahedron3d"))`: | Platonic solids
`r indexfns(c("cuboctahedron3d", "oh3d"))`:             | other solids

```{r}
cols <- rainbow(7)
layout3d(matrix(1:16, 4,4), heights=c(1,3,1,3))
text3d(0,0,0,"tetrahedron3d"); next3d()
shade3d(tetrahedron3d(col=cols[1])); next3d()
text3d(0,0,0,"cube3d"); next3d()
shade3d(cube3d(col=cols[2])); next3d()
text3d(0,0,0,"octahedron3d"); next3d()
shade3d(octahedron3d(col=cols[3])); next3d()
text3d(0,0,0,"dodecahedron3d"); next3d()
shade3d(dodecahedron3d(col=cols[4])); next3d()
text3d(0,0,0,"icosahedron3d"); next3d()
shade3d(icosahedron3d(col=cols[5])); next3d()
text3d(0,0,0,"cuboctahedron3d"); next3d()
shade3d(cuboctahedron3d(col=cols[6])); next3d()
text3d(0,0,0,"oh3d"); next3d()
shade3d(oh3d(col=cols[7]))
```

A very large collection of polyhedra is contained in the 
`r linkfn("Rpolyhedra-package", text = "Rpolyhedra", pkg = "Rpolyhedra")` 
package.

### Generating new shapes 

These functions generate new shapes:

Function                             | Description
------------------------------------ | -----------
`r indexfns("cylinder3d")`: | generate a tube or cylinder
`r indexfns("polygon3d")`:  | generate a flat polygon by triangulation
`r indexfns("extrude3d")`:  | generate an "extrusion" of a polygon
`r indexfns("turn3d")`:     | generate a solid of rotation
`r indexfns("ellipse3d")`:  | generate an ellipsoid in various ways
`r indexfns(c("tmesh3d", "qmesh3d"))`:  | generate a shape from vertices and faces
`r indexfns("shapelist3d")`: | generate a shape by combining other shapes
`as.mesh3d`:                 | a generic function; see below

A related function is `r indexfns("triangulate")`, which takes a
two dimensional polygon and divides it up into triangles using the 
"ear-clipping" algorithm.

The generic function `r indexfns("as.mesh3d")`
is provided to allow data structures produced by
other code to be converted to mesh
structures.  Currently the following classes
are supported:

Class |  Package | Description
----- |  ------- | -----------
`r indexfns("deldir", pkg = "deldir")` | `deldir` | Delaunay triangulations of irregular point clouds 
`r indexfns("triSht", pkg = "interp")` | `interp` | Also Delaunay triangulations
`r indexfns("tri", pkg = "tripack")` | `tripack` | Generalized Delaunay triangulations
`r indexfns("ashape3d", pkg = "alphashape3d")` | `alphashape3d` | Alpha-shapes
`r linkfn("rglId")` | `rgl` | `rgl` object identifiers

The default `r indexfns("as.mesh3d.default")` method is a simple way
to construct a mesh from a matrix of vertices; it can use
`r indexfns("mergeVertices")` (which can also be used on its own)
to merge repeated vertices within the matrix, allowing 
`r linkfn("addNormals")` to be used to give a smooth appearance.

### The underlying class structure for shapes

`"shape3d"` is the basic abstract type.  Objects of this class can be
displayed by `r indexfns("shade3d")` (which shades faces), 
`r indexfns("wire3d")` (which draws edges), or `r indexfns("dot3d")`
(which draws points at each vertex.)  

`"mesh3d"` is a descendant type.  Objects of this type contain the following 
fields:

Field        | Meaning
------------ | ---------------
vb           | A 4 by n matrix of vertices in homogeneous coordinates.  Each column is a point.
it           | (optional) A 3 by t matrix of vertex indices.  Each column is a triangle.
ib           | (optional) A 4 by q matrix of vertex indices.  Each column is a quadrilateral.
material     | (optional) A list of material properties.
normals      | (optional) A matrix of the same shape as vb, containing normal vectors at each vertex.
texcoords    | (optional) A 2 by n matrix of texture coordinates corresponding to each vertex.
values       | (optional) A vector of length n holding values at each vertex


### Contouring shapes

These functions compute and plot contours of functions
on surfaces, or clip objects along a contour of a function.

Function                         | Description
-------------------------------- | -----------
`r indexfns("contourLines3d")`:  | draw contour lines on surface
`r indexfns("filledContour3d")`: | fill between contours on surface
`r indexfns("clipMesh3d")`:      | clip mesh object using curved boundary
`r indexfns("clipObj3d")`:       | clip general object using curved boundary

### Manipulating shapes

These functions manipulate and modify mesh objects:

Function                             | Description
------------------------------------ | -----------
`r indexfns("addNormals")`: | add normal vectors to make a shape look smooth
`r indexfns("subdivision3d")`: | add extra vertices to make it look even smoother
`r indexfns("merge.mesh3d", backticked("merge"))`: | merge mesh objects

The individual steps in `r linkfn("subdivision3d")` are also available:
`r indexfns(c("deform.mesh3d", "divide.mesh3d", "normalize.mesh3d"))`. These
are mainly intended for internal use.

## Multi-figure Layouts

`rgl` has several functions to support displaying multiple different
"subscenes" in the same window.  The high level functions are

Function                  | Description
------------------------- | -----------
`r indexfns("mfrow3d")`:  | Multiple figures (like `r linkfn("par", text = 'par("mfrow")', pkg="graphics")`
`r indexfns("layout3d")`: | Multiple figures (like `r linkfn("layout", pkg="graphics")`)
`r indexfns("next3d")`:   | Move to the next figure (like `r linkfn("plot.new", pkg="graphics")` or `r linkfn("frame", pkg="graphics")`)
`r indexfns("subsceneList")`: | List all the subscenes in the current layout
`r indexfns("clearSubsceneList")`: | Clear the current list and revert to the previous one

There are also lower level functions.

Function                           | Description
---------------------------------- | -----------
`r indexfns("newSubscene3d")`:     | Create a new subscene, with fine control over what is inherited from the parent
`r indexfns("currentSubscene3d")`: | Report on the active subscene
`r indexfns("subsceneInfo")`:      | Get information on current subscene
`r indexfns("useSubscene3d")`:     | Make a different subscene active
`r indexfns(c("addToSubscene3d", "delFromSubscene3d"))`: | Add objects to a subscene, or delete them
`r indexfns("gc3d")`:              | Do "garbage collection":  delete objects that are not displayed in any subscene

## Documents with `rgl` Scenes

The `rgl` package can produce output that can be embedded
in other documents.  The recommended way to do this has changed several
times over the years.  We will start with the current
recommendation, then list older methods.

### The recommended method

Currently the best way to embed an `rgl` scene in a document
is to produce the document in HTML using R Markdown.  Early in
the document, you should have code like this in one of the
setup code chunks:

````markdown
`r ''````{r echo=FALSE, include=FALSE}
library(rgl)
setupKnitr(autoprint = TRUE)
```
````

The call to `setupKnitr()` will install a number of hooks
and set options in `knitr` so that `rgl` code is handled properly.
The `autoprint = TRUE` argument makes `rgl` act in the document
almost the same way it would act in the console, or the way 
base graphics are handled by `knitr`:  If you print the value
of high level `rgl` functions, a plot will be inserted into the
output, but maybe only after low level modifications to it are
complete.  For example, this code block prints both
triangles and spheres in a single plot at the end:

```{r}
xyz <- matrix(rnorm(27), ncol = 3)
triangles3d(xyz, col = rainbow(9))
spheres3d(xyz, col = rainbow(9), radius = 0.1)
```

There are a few differences if you have a complicated situation:

  -  The mechanism depends on the result of the `rgl` function
  calls being automatically printed.
  If the calls are in a loop or other code block where automatic
  printing doesn't happen, you'll need some trickery to get 
  things to print.  For example, this will print three plots:
  
```{r eval = FALSE}
plots <- NULL
for (i in 1:3) {
  plot3d(rnorm(10), rnorm(10), rnorm(10))
  plots <- htmltools::tagList(plots, rglwidget())
  close3d()
}
plots
```

  - It also depends on the fact that `rgl` functions return
  results using `lowlevel()` or `highlevel()` to mark which
  kind of plot they are.  If you are using a function from
  another package to produce the plot, you may need to insert
  an explicit call to one of those to get it to print.  Use `lowlevel()` if the function just modifies an existing plot, 
  `highlevel()` if it starts a new one.  For example,
  
```{r eval = FALSE}
foreignHigh()   # Produces a high level plot, but doesn't return
                # an appropriate value
highlevel()
foreignLow()    # Modifies the previous plot
lowlevel()
```  

This should display the output at the end of the code chunk, when
modifications are assumed complete.

### Producing PDF output

While some PDF previewers support interactive 3D graphics,
most don't.  To produce a screenshot of an `rgl` scene in an R Markdown
document with PDF output, simply follow the directions given
above.  The auto-printing will detect PDF output and use
`snapshot3d` to produce a PNG file to insert.  (See below
if you want to insert a different format of graphic.)

If you really need interactive output, see the 
`r linkfn("writeASY")` function.

### Manual insertion of plots

You may not want to use the `setupKnitr(autoprint = TRUE)`
method described above.  It is very new, and may still have
bugs; you may have an older document and not want to edit it
to work that way.

In this case, you can insert plots manually.  Use setup
code

````markdown
`r ''````{r echo=FALSE, include=FALSE}
library(rgl)
setupKnitr()
```
````

and call `rglwidget()` at top level whenever you want to 
insert a plot.

There are a couple of other differences in default behaviour if
you are not using `autoprint`:

 - By default, each code chunk continues the `rgl` scene from
earlier chunks.  You'll need an explicit `r linkfn("open3d")`
call to get a clean window.

 - Also by default, the `rgl` window is not closed at the end of
the chunk.  This probably doesn't matter, but you may find
you run out of memory if your scenes are really big.

### Older methods

The original way to insert an `rgl` scene in a document was to use
the `r indexfns("writeWebGL")` function to write HTML code to insert in a document. 
Later, `Sweave` and `knitr` hooks were added.  These are no longer maintained, and it is recommended that
you update old documents to use the newer methods.  See
the `r linkfn("writeWebGL")`, `r linkfn("rgl.Sweave")` and `r linkfn("hook_rgl")` help
topics for details on these if you must use them.  If you are
reading documents that suggest using those methods, let the
author know they need updating!
  
## Utility Functions


### User interaction

By default, `rgl` detects and handles mouse clicks within your scene,
and uses these to control its appearance.  You can find out the current
handlers using the following code:
```{r}
par3d("mouseMode")
```
The labels `c("left", "right", "middle")` refer to the buttons on
a three button mouse, or simulations of them on other mice.  `"wheel"` 
refers to the mouse wheel.

The button actions generally correspond to click and drag operations.
Possible values for `r indexfns("mouseMode", '<code>"mouseMode"</code>')` for buttons or the wheel are as follows:

Mode           | Description
-------------- | ---------
`"none"`       | No action
`"trackball"`  | The mouse acts as a virtual trackball. Clicking and dragging rotates the scene
`"xAxis"`, `"yAxis"`, `"zAxis"`      | Like `"trackball"`, but restricted to rotation about one axis
`"polar"`      | The mouse affects rotations by controlling polar coordinates directly
`"selecting"`  | The mouse is being used by the `r linkfn("select3d")` function
`"zoom"`       | The mouse zooms the display
`"fov"`        | The mouse affects perspective by changing the field of view
`"pull"`       | Rotating the mouse wheel towards the user "pulls the scene closer"
`"push"`       | The same rotation "pushes the scene away"
`"user"`       | A user action set by `r indexfns(c("rgl.setMouseCallbacks", "rgl.setWheelCallback"))`.  Use `r indexfns("rgl.getMouseCallbacks")` and `r indexfns("rgl.getWheelCallback")` to retrieve

The following functions make use of the mouse for selection within a 
scene.

Function                     | Description
---------------------------- | -----------
`r indexfns("identify3d")`:   | like the classic graphics `r linkfn("identify", pkg="graphics")` function
`r indexfns("select3d")`:     | returns a function that tests whether a coordinate was selected
`r indexfns("selectpoints3d")`: | selects from specific objects

`r indexfns("selectionFunction3d")` produces the selection
function from information about the projection and mouse
selection region; it is used internally in the 
functions above.

The `r indexfns("rgl.select3d")` function is an obsolete version of
`select3d`, and `r indexfns("rgl.select")` is a low-level support
function.


### Animations

`rgl` has several functions that can be used to construct
animations.  These are based on functions that update the
scene according to the current real-world time, and repeated
calls to those.  The functions are:

Function               | Description
---------------------- | -------------
`r indexfns("play3d")`: | Repeatedly call the update function 
`r indexfns("spin3d")`: | Update the display by rotating at a constant rate
`r indexfns("par3dinterp")`: | Compute new values of some `r linkfn("par3d")` parameters by interpolation over time

See the `r linkfn("movie3d")` function for a way to output an animation
to a file on disk.  
Animations are not currently supported 
in the HTML written by `r linkfn("rglwidget")`, though the
`playwidget` function provides equivalent functionality.

### Integration with TCL/TK

There are three functions in `rgl` that support control of an `rgl`
scene using the TCL/TK framework.

Function               | Description
---------------------- | -------------
`r indexfns("tkspin3d")`: | Set up buttons in a window to control a scene 
`r indexfns("tkspinControl")`: | Embed the control buttons in a separate TCL/TK frame
`r indexfns("tkpar3dsave")`: | Create a dialog to interactively save mouse actions

These functions were formerly contained (without the `tk` prefixes on 
their names) in the `tkrgl` package.  That package is now deprecated.

### Exporting and importing scenes

`rgl` contains several functions to write scenes to disk for use
by other software, or to read them in.

In order from highest fidelity to lowest, the functions are:

Function    | Description
----------- | -------------
`r indexfns("scene3d")`:    | Save a scene to an R variable, which can be saved and reloaded
`r indexfns("rglwidget")`: | Prints as HTML and Javascript to display a scene in a web browser.  (See also [User Interaction in WebGL](WebGL.html).)
`r linkfn("writeWebGL")`: | Deprecated.
`r indexfns("writeASY")`:   | Write files for Asymptote
`r indexfns("writePLY")`:   | Write PLY files (commonly used in 3D printing)
`r indexfns(c("readOBJ", "writeOBJ"))`:   | Read or write OBJ files (commonly used in 3D graphics)
`r indexfns(c("readSTL", "writeSTL"))`:   | Read or write STL files (also common in 3D printing)

There are also functions to save snapshots or other recordings of a scene,
without any 3D information being saved:

Function     | Description
------------ | -------------
`r indexfns("snapshot3d")`: | Save a PNG file bitmap of the scene
`r indexfns("rgl.postscript")`: | Save a Postscript, LaTeX, PDF, SVG or PGF vector rendering of the scene
`r indexfns("movie3d")`:    | Save a series of bitmaps to be assembled into a movie
`r indexfns("rgl.pixels")`: | Obtain pixel-level information about the scene in an R variable
`r indexfns("rgl.Sweave")`: | Driver function for inserting a snapshot into a Sweave document.
`r indexfns(c("hook_rgl", "hook_webgl"))`: | `knitr` hook functions for inserting images into a document.
`r indexfns("setupKnitr")`: | Function to set up `knitr` hooks

The `r indexfns("rgl.snapshot")` function is a low level
version of `snapshot3d()`; it requires that the `rgl` display
be onscreen and copies from there.  `snapshot3d()` tries to 
use the `webshot2` package so it will work even with no display.
The functions `r indexfns(c("rgl.Sweave.off", "Sweave.snapshot"))` are 
involved in Sweave processing and not normally called by users.


### Default display

There are two ways in which `rgl` scenes are normally
displayed within R.  The older one is in a dedicated
window.  In Unix-alikes this is an X11 window;  it is
a native window in Microsoft Windows.  On macOS, the
XQuartz system (see https://www.xquartz.org) needs to be
installed to support this.

To suppress this display, set `options(rgl.useNULL = TRUE)` before opening a new `rgl` window.  See
the help page for the `r indexfns("rgl.useNULL")` function
for how to set this before starting R.

The newer way to display a scene is by using WebGL
in a browser window or in the Viewer pane in RStudio.
To select this, set `options(rgl.printRglwidget = TRUE)`.
Each operation that would change the scene will 
return a value which triggers a new WebGL display
when printed.

### Working with WebGL scenes

You should use the following scheme for exporting a scene to a web page.  There's also an older scheme, which is no longer 
supported.

The recommended approach works with the `htmlwidgets` framework
(see http://www.htmlwidgets.org/).  In an R Markdown document
in `knitr`, use the 
`r linkfn("rglwidget")` function.  (You can also use
chunk option `webgl=TRUE`; we recommend the explicit use of
`rglwidget`.)  This approach also
allows display of `rgl` scenes in
[RStudio](https://rstudio.com).  Besides 
`rgl` scenes, various controls for them can
be displayed, and there are a few utility functions
that can be useful:

Function                             | Description
------------------------------------ | -------------
`r indexfns("propertyControl")`:       | set individual properties
`r indexfns("clipplaneControl")`:       |  control a clipping plane
`r indexfns("subsetControl")`:       | control which objects are displayed
`r indexfns("ageControl")`:       | "age" vertices of an object
`r indexfns("vertexControl")`:       | control properties of vertices
`r indexfns("par3dinterpControl")`:  | WebGL control like `r linkfn("par3dinterp")`
`r indexfns("playwidget")`:       | display and automate controls
`r indexfns("toggleWidget")`:     | display a button to toggle some items
`r documentedfns <- c(documentedfns, "%>%");indexfns("pipe", text="%>%")`:    | `magrittr` pipe
`r indexfns(c("figHeight", "figWidth"))`:     | Dimensions of figures in R Markdown document
`r indexfns("rglShared")`:        | share data using `crosstalk` package
`r indexfns("rglMouse")`:         | change mouse mode in rgl scene
`r indexfns("asRow")`:            | arrange multiple objects in an HTML display
`r indexfns("getWidgetId")`:      | get the `elementId` from a widget

These functions work with the above scheme in Shiny apps:

Function                             | Description
------------------------------------ | -------------
`r indexfns("sceneChange")`:      | used in `Shiny` for large scene changes
`r indexfns(c("shinyGetPar3d", "shinySetPar3d"))`: | get or set `r linkfn("par3d")` values from Shiny
`r indexfns("shinyResetBrush")`:  | reset the mouse brush in Shiny

The `r linkfn("selectionFunction3d")` function is also likely
to be involved in mouse interactions when using Shiny.

Some functions are mainly for internal use:
`r indexfns(c("elementId2Prefix", "playwidgetOutput", "renderPlaywidget", "rglwidgetOutput", "renderRglwidget", "registerSceneChange"))`. 
More details are given in the vignette 
[User Interaction in WebGL](WebGL.html).
The functions `r indexfns(c("lowlevel", "highlevel", "rglId"))` are also for internal use,
marking function results for automatic
printing.  Finally, the experimental function
`r indexfns("setUserShaders")` allows you to
use hand-written shaders in WebGL.

The older approach uses the `r linkfn("writeWebGL")` function to export a scene 
to HTML and Javascript code.  
These functions write HTML and Javascript for working with the 
exported scene:  `r indexfns("propertySlider")`,
`r indexfns("clipplaneSlider")`, `r indexfns("subsetSlider")`,
`r indexfns("toggleButton")`, `r indexfns("propertySetter")`,
`r indexfns("subsetSetter")`, `r indexfns("ageSetter")`, 
`r indexfns("par3dinterpSetter")`, 
`r indexfns("vertexSetter")`, `r indexfns("matrixSetter")`.
Use the newer functions instead.

### Working with the scene

`rgl` maintains internal structures for all the scenes it displays.
The following functions allow users to find information about them
and manipulate them.  In cases where there are both `*3d`
and `rgl.*` versions of functions, most users should use
the `*3d` version:  the `rgl.*` functions are more primitive
and are mainly intended for internal use.

Function                             | Description
------------------------------------ | -----------
`r indexfns("open3d")`:               | open a new window
`r indexfns(c("close3d", "rgl.close"))`:              | close the current window
`r indexfns(c("cur3d", "rgl.cur"))`:                | id of the active device
`r indexfns(c("set3d", "rgl.set"))`:                | set a particular device to be active
`r indexfns(c("pop3d", "rgl.pop"))`:     | delete an object from the scene
`r indexfns(c("clear3d", "rgl.clear"))`:            | delete all objects of certain classes
`r indexfns(c("ids3d", "rgl.ids"))`:              | ids and types of all current objects

These functions are mainly intended for programming, and 
have no corresponding `*3d` counterparts:

Function | Description
------------- | -----------
`r indexfns("rgl.bringtotop")`: | bring the current window to the top
`r indexfns("rgl.dev.list")`:   | ids of all active devices
`r indexfns(c("rgl.attrib", "rgl.attrib.info", "rgl.attrib.count"))`: | attributes of objects in the scene
`r indexfns("rgl.projection")`: | return information about the current projection
`r indexfns(c("rgl.user2window", "rgl.window2user"))`: | convert between coordinates in the current projection

The `r indexfns("as.triangles3d")` generic function is intended to 
extract coordinates in a form suitable for passing to 
`r linkfn("triangles3d")`.  Currently a method is provided for
`r linkfn("rglId")` objects.

In addition to these, there are some other related functions which
should rarely be called by users: `r indexfns(c("rgl.init", "rgl.open", "rgl.quit"))`.  


### Working with 3-D vectors

Most `rgl` functions work internally with "homogeneous" coordinates.
In this system, 3-D points are represented with 4 coordinates, generally
called (x, y, z, w).  The corresponding Euclidean point is
(x/w, y/w, z/w), if w is nonzero; zero values of w correspond to 
"points at infinity".  The advantage of this system is that
affine transformations including translations and perspective shifts
become linear transformations, with multiplication by a 4 by 4 matrix.

`rgl` has the following functions to work with homogeneous coordinates:

Function                             | Description
------------------------------------ | -----------
`r indexfns(c("asEuclidean", "asHomogeneous"))`: | convert between homogeneous and Euclidean coordinates
`r indexfns(c("rotate3d", "scale3d", "translate3d"))`:  | apply a transformation
`r indexfns("transform3d")`:  | apply a general transformation
`r indexfns(c("rotationMatrix", "scaleMatrix", "translationMatrix"))`:  | compute the transformation matrix
`r indexfns("identityMatrix")`: | return a 4 x 4 identity matrix

There is also a function `r indexfns("GramSchmidt")`, mainly for internal
use:  it does a Gram-Schmidt orthogonalization of a 3x3 matrix,
with some specializations for its use in `r linkfn("cylinder3d")`.

### Working with other packages

Sometimes it may be convenient to interactively
rotate a scene to a particular view, then 
display it in `lattice` or base graphics.
The `r indexfns("rglToLattice")` and `r indexfns("rglToBase")` functions support this.

For example, we first display the volcano data in `rgl`:
```{r echo = 2:3}
close3d()
persp3d(volcano, col = "green")
```
This display is interactive, but we can reproduce the initial view using the
`lattice` `r linkfn("wireframe", pkg = "lattice")` or base graphics `r linkfn("persp", pkg = "graphics")` functions:
```{r}
lattice::wireframe(volcano, col = "green", 
		   screen = rglToLattice())
angles <- rglToBase()
persp(volcano, col = "green", shade = TRUE,
      theta = angles$theta, phi = angles$phi)
```

Note that the `orientlib` package must be 
available for these functions to work.

### Working with Javascript

The WebGL displays created using `r linkfn("rglwidget")`
rely on a large body of Javascript code included in this
package.  To help in development of this code, the
`r indexfns("makeDependency")` function was written.
It may be useful in other packages that include Javascript.

### Other functions

This section is for miscellaneous functions that don't
fall in any of the other categories in this document.

The `r indexfns("setGraphicsDelay")` function is designed
to work around what appears to be a bug on macOS:  if
a standard plot window is opened too quickly after an `rgl`
window, R can crash.  This function inserts a one second 
delay when it appears to be needed.

## Warning:  Work in Progress!

This vignette is always a work in progress.  Some aspects of the `rgl`
package are not described, or do not have examples.  There may
even be functions that are missed completely, if the following
list is not empty:
```{r echo=FALSE}
setdiff(ls("package:rgl"), documentedfns)
```

## Index of Functions

The following functions and constants are described in this document:<br>

```{r echo=FALSE, results="asis"}
writeIndex(cols = if (knitr::is_html_output()) 5 else 4)
```

---
title: "A Note on Transparency"
author: "Duncan Murdoch"
date: "24/10/2020"
output:
  rmarkdown::html_vignette:
    fig_height: 5
    fig_width: 5
    toc: yes
  pdf_document:
    fig_height: 5
    fig_width: 5
    toc: yes
  html_document:
    default
vignette: >
  %\VignetteIndexEntry{A Note on Transparency}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
library(rgl)
options(rgl.useNULL = TRUE)
setupKnitr(autoprint = TRUE)
M <- structure(c(0.997410774230957, 0.0707177817821503, -0.0130676832050085, 
0, -0.0703366547822952, 0.99714070558548, 0.02762770652771, 0, 
0.0149840852245688, -0.0266370177268982, 0.999532878398895, 0, 
0, 0, 0, 1), .Dim = c(4L, 4L))
```

## Introduction

When drawing transparent surfaces, `rgl` tries to sort objects
from back to front to get better rendering of transparency.
However, it doesn't sort each pixel separately, so some pixels
end up drawn in the incorrect order.  This note
describes the consequences of that error, and 
suggests remedies.

## Correct Drawing

We'll assume that the standard `glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)` blending is used.  That is,
when drawing with transparency $\alpha$, the new colour is
mixed at proportion $\alpha$ with the colour previously
drawn (which gets weight $1-\alpha$.)

This note is concerned with what happens when two transparent
objects are drawn at the same location.  We suppose the further
one has transparency $\alpha_1$, and the closer one 
has transparency $\alpha_2$.  If they are drawn in the correct
order (further first), the three colours (background, further
object, closer object) should end up mixed in proportions
$C = [(1-\alpha_1)(1-\alpha_2), \alpha_1(1-\alpha_2), \alpha_2]$
respectively.

## Incorrect sorting

If the objects are drawn in the wrong order, the actual 
proportions of each colour will be $N = [(1-\alpha_1)(1-\alpha_2),
\alpha_1, \alpha_2(1-\alpha_1)]$ if no masking occurs. 

`rgl` currently defaults to depth masking using `glDepthMask(GL_TRUE)`.  This means that depths are saved 
when the objects are drawn, and if an attempt is made to 
draw the further object after the closer one (i.e. as here),
the further object will be culled, and the proportions will be
$M = [(1-\alpha_2), 0, \alpha_2]$.

## Mask or Not?

The question is:  which is better, `glDepthMask(GL_TRUE)` or
`glDepthMask(GL_FALSE)`?  One way to measure this is to 
measure the distance between $C$ and the incorrect proportions.
(This is unlikely to match perceptual distance, which will
depend on the colours as well, but we need something. Some
qualitative comments below.)

So we have 

\[|C-N|^2 = 2\alpha_1^2\alpha_2^2\], 

and 

\[|C-M|^2 = 2\alpha_1^2(1-\alpha_2)^2\].

Thus the error is larger with $N$ when $\alpha_2 > 1/2$, and 
larger with $M$ when $\alpha_2 < 1/2$.  The value of $\alpha_1$
doesn't affect the preference, though small values of $\alpha_1$ will be associated with smaller errors.  

Depending on the colours of the background and the two
objects, this recommendation could be modified.  For example,
if the two objects are the same colour (or very close),
it doesn't really matter how the 2nd and 3rd proportions
are divided up, and $N$ will be best because it gets the
background proportion exactly right.


## Recommendation

Typically in `rgl` we don't know which object will be closer 
and which one will be further, so we can't base our choice on
a single $\alpha_i$.  The recommendation would be to use
all small levels of `alpha` and disable masking, or use all
large values of `alpha` and retain masking.

## Example

The classic example of an impossible to sort scene involves
three triangles arranged cyclicly so each one is behind one and in front
of one of the others (based on https://paroj.github.io/gltut/Positioning/Tut05%20Overlap%20and%20Depth%20Buffering.html).

```{r}
theta <- 2*pi*c(0:2, 4:6, 8:10)/12
x <- cos(theta)
y <- sin(theta)
z <- rep(c(0,0,1), 3)
xyz <- cbind(x, y, z)
xyz <- xyz[c(1,2,6, 4,5,9, 7,8,3),]
open3d()
par3d(userMatrix = M)
triangles3d(xyz, col = rep(c("red", "green", "blue"), each = 3))
```

To see the effect of the calculations above, consider the following four displays.

```{r fig.width=8}
open3d()
par3d(userMatrix = M)
layout3d(matrix(1:9, ncol = 3, byrow=TRUE),
         widths = c(1,2,2), heights = c(1, 3,3), 
         sharedMouse = TRUE)
text3d(0,0,0, " ")
next3d()
text3d(0,0,0, "depth_mask = TRUE")
next3d()
text3d(0,0,0, "depth_mask = FALSE")
next3d()
text3d(0,0,0, "alpha = 0.7")
next3d()
triangles3d(xyz, col = rep(c("red", "green", "blue"), each = 3), alpha = 0.7, depth_mask = TRUE)
next3d()
triangles3d(xyz, col = rep(c("red", "green", "blue"), each = 3), alpha = 0.7, depth_mask = FALSE)
next3d()
text3d(0,0,0, "alpha = 0.3")
next3d()
triangles3d(xyz, col = rep(c("red", "green", "blue"), each = 3), alpha = 0.3, depth_mask = TRUE)
next3d()
triangles3d(xyz, col = rep(c("red", "green", "blue"), each = 3), alpha = 0.3, depth_mask = FALSE)
```

As you rotate the figures, you can see imperfections in
rendering.  On the right, the last drawn appears to be on top,
while on the left, the first drawn appears more opaque than
it should.  

In the figure below, the three triangles each have different transparency, and each use the recommended setting:

```{r}
open3d()
par3d(userMatrix = M)
triangles3d(xyz[1:3,], col = "red", alpha = 0.3, depth_mask = FALSE)
triangles3d(xyz[4:6,], col = "green", alpha = 0.7, depth_mask = TRUE)
triangles3d(xyz[7:9,], col = "blue", depth_mask = TRUE)
```

In this figure, all three triangles are the same colour,
only lighting affects the display:
```{r fig.width=8}
open3d()
par3d(userMatrix = M)
layout3d(matrix(1:9, ncol = 3, byrow=TRUE),
         widths = c(1,2,2), heights = c(1, 3,3), 
         sharedMouse = TRUE)
text3d(0,0,0, " ")
next3d()
text3d(0,0,0, "depth_mask = TRUE")
next3d()
text3d(0,0,0, "depth_mask = FALSE")
next3d()
text3d(0,0,0, "alpha = 0.7")
next3d()
triangles3d(xyz, col = "red", alpha = 0.7, depth_mask = TRUE)
next3d()
triangles3d(xyz, col = "red", alpha = 0.7, depth_mask = FALSE)
next3d()
text3d(0,0,0, "alpha = 0.3")
next3d()
triangles3d(xyz, col = "red", alpha = 0.3, depth_mask = TRUE)
next3d()
triangles3d(xyz, col = "red", alpha = 0.3, depth_mask = FALSE)
```

Here `depth_mask = FALSE` seems to be the right choice in both cases.

---
title: "rgl Demos"
date: "`r Sys.Date()`"
output: 
  rmarkdown::html_vignette:
    toc: true
    fig_height: 5
    fig_width: 5
vignette: >
  %\VignetteIndexEntry{rgl Demos}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = FALSE,  # Bug in knitr 1.42 requires FALSE here
  comment = "#>",
  snapshot = FALSE,
  screenshot.force = FALSE)
library(rgl)
setupKnitr(autoprint = TRUE)
```

## Introduction

This vignette holds code that was previously included as "demos" 
in `rgl`.  Some of the demos require R to be running; those
remain available via `demo(package = "rgl")`.

## hist3d:  3D histogram using basic building blocks

```{r}

##########
### 3D HIST EXAMPLE:
##########

################################################################################

##### Required functions 'binplot' and 'hist3d':

binplot.3d<-function(x,y,z,alpha=1,topcol="#ff0000",sidecol="#aaaaaa") {
  save <- par3d(skipRedraw=TRUE)
  on.exit(par3d(save))
    
  x1<-c(rep(c(x[1],x[2],x[2],x[1]),3),rep(x[1],4),rep(x[2],4))
  z1<-c(rep(0,4),rep(c(0,0,z,z),4))
  y1<-c(y[1],y[1],y[2],y[2],rep(y[1],4),rep(y[2],4),rep(c(y[1],y[2],y[2],y[1]),2))
  x2<-c(rep(c(x[1],x[1],x[2],x[2]),2),rep(c(x[1],x[2],rep(x[1],3),rep(x[2],3)),2))
  z2<-c(rep(c(0,z),4),rep(0,8),rep(z,8) )
  y2<-c(rep(y[1],4),rep(y[2],4),rep(c(rep(y[1],3),rep(y[2],3),y[1],y[2]),2) )
  quads3d(x1,z1,y1,col=rep(sidecol,each=4),alpha=alpha)
  quads3d(c(x[1],x[2],x[2],x[1]),rep(z,4),c(y[1],y[1],y[2],y[2]),
              col=rep(topcol,each=4),alpha=1) 
  segments3d(x2,z2,y2,col="#000000")
}

hist3d<-function(x,y=NULL,nclass="auto",alpha=1,col="#ff0000",scale=10) {
  save <- par3d(skipRedraw=TRUE)
  on.exit(par3d(save))
  xy <- xy.coords(x,y)
  x <- xy$x
  y <- xy$y
  n<-length(x)
  if (nclass == "auto") nclass<-ceiling(sqrt(nclass.Sturges(x)))
  breaks.x <- seq(min(x),max(x),length=(nclass+1))
  breaks.y <- seq(min(y),max(y),length=(nclass+1))
  z<-matrix(0,(nclass),(nclass))
  for (i in seq_len(nclass)) {
    for (j in seq_len(nclass)) {
      z[i, j] <- (1/n)*sum(x < breaks.x[i+1] & y < breaks.y[j+1] & 
                            x >= breaks.x[i] & y >= breaks.y[j])
      binplot.3d(c(breaks.x[i],breaks.x[i+1]),c(breaks.y[j],breaks.y[j+1]),
                 scale*z[i,j],alpha=alpha,topcol=col)
    }
  }
}
################################################################################

open3d()

bg3d(color="gray")
light3d(0, 0)

# Drawing a 'bin' for given coordinates:
binplot.3d(c(-0.5,0.5),c(4.5,5.5),2,alpha=0.6)

# Setting the viewpoint ('theta' and 'phi' have the same meaning as in persp):
view3d(theta=40,phi=40)

##### QUADS FORMING BIN

open3d()

# Defining transparency and colors:
alpha<-0.7; topcol<-"#ff0000"; sidecol<-"#aaaaaa"

# Setting up coordinates for the quads and adding them to the scene:
y<-x<-c(-1,1) ; z<-4; of<-0.3
x12<-c(x[1],x[2],x[2],x[1]); x11<-rep(x[1],4); x22<-rep(x[2],4)
z00<-rep(0,4); z0z<-c(0,0,z,z); zzz<-rep(z,4); y11<-rep(y[1],4)
y1122<-c(y[1],y[1],y[2],y[2]); y12<-c(y[1],y[2],y[2],y[1]); y22<-rep(y[2],4)

quads3d(c(x12,x12,x11-of,x12,x22+of,x12),
          c(z00-of,rep(z0z,4),zzz+of),
          c(y1122,y11-of,y12,y22+of,y12,y1122),
          col=rep(c(rep(sidecol,5),topcol),each=4),alpha=c(rep(alpha,5),1))

# Setting up coordinates for the border-lines of the quads and drawing them:
yl1<-c(y[1],y[2],y[1],y[2]); yl2<-c(y[1]-of,y[1]-of)
xl<-c(rep(x[1],8),rep(x[1]-of,8),rep(c(x[1],x[2]),8),rep(x[2],8),rep(x[2]+of,8))
zl<-c(0,z,0,z,z+of,z+of,-of,-of,0,0,z,z,0,z,0,z,rep(0,4),rep(z,4),rep(-of,4),
      rep(z+of,4),z+of,z+of,-of,-of,rep(c(0,z),4),0,0,z,z)
yl<-c(yl2,y[2]+of,y[2]+of,rep(c(y[1],y[2]),4),y[1],y[1],y[2],y[2],yl2,
      rep(y[2]+of,4),yl2,y[2],y[2],rep(y[1],4),y[2],y[2],yl1,yl2,y[2]+of,
      y[2]+of,y[1],y[1],y[2],y[2],yl1)

lines3d(xl,zl,yl,col="#000000")
view3d(theta=40,phi=40)

##### COMPLETE HISTOGRAM:

open3d()

# Setting the rng to a fixed value:
set.seed(1000)

# Drawing a 3d histogramm of 2500 normaly distributed observations:
hist3d(rnorm(2500),rnorm(2500),alpha=0.4,nclass=7,scale=30)
# Choosing a lightgrey background:
bg3d(col="#cccccc")
view3d(theta=40,phi=40)
```


## bivar:  Bivariate densities: kernel smoothing using surface3d and alpha-channel (requires MASS package)

```{r}
# rgl demo: rgl-bivar.r
# author: Daniel Adler

rgl.demo.bivar <- function() {
  if (!requireNamespace("MASS", quietly = TRUE))
    stop("This demo requires MASS")
  
  # parameters:
  n<-50; ngrid<-40
  
  # generate samples:
  set.seed(31415)
  x<-rnorm(n); y<-rnorm(n)
  
  # estimate non-parameteric density surface via kernel smoothing
  denobj <- MASS::kde2d(x, y, n=ngrid)
  den.z <-denobj$z
  
  # generate parametric density surface of a bivariate normal distribution
  xgrid <- denobj$x
  ygrid <- denobj$y
  bi.z <- dnorm(xgrid)%*%t(dnorm(ygrid))
  
  # visualize:
  zscale<-20
  
  # New window
  open3d()
  
  # clear scene:
  clear3d("all")
  
  # setup env:
  bg3d(color="#887777")
  light3d()
  
  # Draws the simulated data as spheres on the baseline
  spheres3d(x,y,rep(0,n),radius=0.1,color="#CCCCFF")
  
  # Draws non-parametric density
  surface3d(xgrid,ygrid,den.z*zscale,color="#FF2222",alpha=0.5)
  
  # Draws parametric density
  surface3d(xgrid,ygrid,bi.z*zscale,color="#CCCCFF",front="lines") 
}

rgl.demo.bivar()
```

## Abundance:  Animal abundance, visualization of multi-dimension data using multiple techniques

```{r}
# RGL-Demo: animal abundance
# Authors: Oleg Nenadic, Daniel Adler

rgl.demo.abundance <- function() {
  open3d()
  clear3d("all")               # remove all shapes, lights, bounding-box, and restore viewpoint
  
  # Setup environment:
  bg3d(col="#cccccc")     # setup background
  light3d()               # setup head-light
  
  # Importing animal data (created with wisp)
  terrain<-dget(system.file("demodata/region.dat",package="rgl"))
  pop<-dget(system.file("demodata/population.dat",package="rgl"))
  
  # Define colors for terrain
  zlim <- range(terrain)
  colorlut <- terrain.colors(82) 
  col1 <- colorlut[9*sqrt(3.6*(terrain-zlim[1])+2)]
  
  # Set color to (water-)blue for regions with zero 'altitude' 
  col1[terrain==0]<-"#0000FF"
  
  # Add terrain surface shape (i.e. population density):
  surface3d( 
      1:100,seq(1,60,length=100),terrain,
      col=col1,spec="#000000", ambient="#333333", back="lines"
  )
  
  # Define colors for simulated populations (males:blue, females:red):
  col2<-pop[,4]
  col2[col2==0]<-"#3333ff"
  col2[col2==1]<-"#ff3333"
  
  # Add simulated populations as sphere-set shape
  spheres3d(
    pop[,1],
    pop[,2],
    terrain[cbind( ceiling(pop[,1]),ceiling(pop[,2]*10/6) )]+0.5,
    radius=0.2*pop[,3], col=col2, alpha=(1-(pop[,5])/10 )
  )

}
rgl.demo.abundance()
```



## lsystem: Plant modelling using a turtle and L-system

```{r}
# demo: lsystem.r
# author: Daniel Adler

#
# geometry 
#

deg2rad <- function( degree ) {
  return( degree*pi/180 )
}

rotZ.m3x3 <- function( degree ) {
  kc <- cos(deg2rad(degree))
  ks <- sin(deg2rad(degree))
  return( 
    matrix(
      c(
        kc, -ks,   0, 
        ks,  kc,   0,
         0,   0,   1
      ),ncol=3,byrow=TRUE
    ) 
  )
}

rotX.m3x3 <- function( degree ) {
  kc <- cos(deg2rad(degree))
  ks <- sin(deg2rad(degree))
  return(
    matrix(
      c(
         1,   0,   0,
         0,  kc, -ks,
         0,  ks,  kc
      ),ncol=3,byrow=TRUE
    )
  )
}

rotY.m3x3 <- function( degree ) {
  kc <- cos(deg2rad(degree))
  ks <- sin(deg2rad(degree))
  return(
    matrix(
      c(
        kc,   0,   ks,
         0,   1,    0,
       -ks,   0,   kc
      ),ncol=3,byrow=TRUE
    )
  )
}

rotZ <- function( v, degree ) {
  return( rotZ.m3x3(degree) %*% v)
}

rotX <- function( v, degree ) {
  return( rotX.m3x3(degree) %*% v)
}

rotY <- function( v, degree ) {
  return( rotY.m3x3(degree) %*% v)
}


#
# turtle graphics, rgl implementation:
#

turtle.init <- function(pos=c(0,0,0),head=0,pitch=90,roll=0,level=0) {
  clear3d("all")
  bg3d(color="gray")
  light3d()
  return( list(pos=pos,head=head,pitch=pitch,roll=roll,level=level) )
}


turtle.move <- function(turtle, steps, color) {
  
  rm <- rotX.m3x3(turtle$pitch) %*% rotY.m3x3(turtle$head) %*% rotZ.m3x3(turtle$roll)
  
  from <- as.vector( turtle$pos )  
  dir  <- rm %*% c(0,0,-1)
  to   <- from + dir * steps
    
  x <- c( from[1], to[1] )
  y <- c( from[2], to[2] )
  z <- c( from[3], to[3] )
  lines3d(x,y,z,col=color,size=1.5,alpha=0.5)
  turtle$pos <- to
  return(turtle)
}

turtle.pitch <- function(turtle, degree) {
  turtle$pitch <- turtle$pitch + degree
  return(turtle)
}

turtle.head <- function(turtle, degree) {
  turtle$head <- turtle$head + degree
  return(turtle)
}

turtle.roll <- function(turtle, degree) {
  turtle$roll <- turtle$roll + degree
  return(turtle)
}

#
# l-system general
#


lsystem.code <- function( x )
  substitute( x )

lsystem.gen <- function( x, grammar, levels=0 ) {
  code <- eval( substitute( substitute( REPLACE , grammar ), list(REPLACE=x) ) )
  if (levels)
    return( lsystem.gen( code , grammar , levels-1 ) )
  else
    return( code )
}

#
# l-system plot
#

lsystem.plot <- function( expr, level ) {
  turtle <- turtle.init(level=level)
  lsystem.eval( expr, turtle )
}

lsystem.eval <- function( expr, turtle ) {
  if ( length(expr) == 3 ) {
    turtle <- lsystem.eval( expr[[2]], turtle )
    turtle <- lsystem.eval( expr[[3]], turtle )
    turtle <- lsystem.eval( expr[[1]], turtle )
  } else if ( length(expr) == 2 ) {
    saved <- turtle
    turtle <- lsystem.eval( expr[[1]], turtle )
    turtle <- lsystem.eval( expr[[2]], turtle )
    turtle <- saved
  } else if ( length(expr) == 1 ) {
    if ( as.name(expr) == "stem" )      turtle <- turtle.move(turtle, 5, "brown")
    else if ( as.name(expr) == "short") turtle <- turtle.move(turtle, 5, "brown")
    else if ( as.name(expr) == "leaf" ) {
      spheres3d(turtle$pos[1],turtle$pos[2],turtle$pos[3],radius=0.1+turtle$level*0.3,color="green")
      sprites3d(turtle$pos[1],turtle$pos[2],turtle$pos[3],radius=0.5+turtle$level*0.3 ,color="green",texture=system.file("textures/particle.png",package="rgl"),textype="alpha",alpha=0.5)     
    }
    else if ( as.name(expr) == "roll" ) turtle <- turtle.head(turtle, 60)
    else if ( as.name(expr) == "down" ) turtle <- turtle.pitch(turtle,10)
    else if ( as.name(expr) == "up" )   turtle <- turtle.pitch(turtle,-10)
    else if ( as.name(expr) == "left" ) turtle <- turtle.head(turtle, 1)
    else if ( as.name(expr) == "right") turtle <- turtle.head(turtle,-1.5)
    else if ( as.name(expr) == "turnleft") turtle <- turtle.head(turtle,20)
    else if ( as.name(expr) == "turnright") turtle <- turtle.head(turtle,-20)
    else if ( as.name(expr) == "turn") turtle <- turtle.roll(turtle,180)
  }
  return(turtle)
}


#
# example
#

simple <- function(level=0) {
  grammar <- list(
    stem=lsystem.code(
      stem-(up-stem-leaf)-stem-(down-stem-leaf)-stem-leaf
    )
  )
  plant <- lsystem.gen(lsystem.code(stem), grammar, level )
  lsystem.plot(plant,level)
}

rgl.demo.lsystem <- function(level=0) {
  gen   <- list(
    stem=lsystem.code( 
      stem-left-stem-branch( turnleft-down-short-turnleft-down-stem-leaf)-right-right-stem--branch( turnright-up-short-turnright-up-short-turnright-short-stem-leaf)-left-left-left-stem-branch( turnleft-down-short-turnright-down-stem-leaf )-branch( up-turnright-short-up-turnleft-up-stem-leaf ) 
    )
  )    
  plant <- lsystem.gen(lsystem.code(stem), gen, level )
  lsystem.plot(plant,level)  
}

open3d()
rgl.demo.lsystem(level=1)
rglwidget()
```

## subdivision:  Subdivision surfaces using generic meshes (preview of generic 3D interface)

```{r}
# RGL-demo: subdivision surfaces
# author: Daniel Adler

rgl.demo.subdivision <- function() {
  # setup environment
  clear3d("all")
  view3d()
  bg3d(color="gray")
  light3d()

  # generate basic mesh
  obj <- oh3d()

  part <- function( level, tx, ... ) {
    shade3d( translate3d( obj, tx, 0, 0 )
    , color="gray30", front="lines",alpha=0.5,back="lines", override=TRUE
    )
    shade3d( translate3d( subdivision3d( obj, depth=level ), tx, 0, 0 )
    , override=TRUE, ... )
  }
  
  part(0, -5.50, color="blue"   )
  part(1, -1.75, color="yellow" )
  part(2,  1.75, color="red"    )
  part(3,  5.50, color="green"  )

}

open3d()
rgl.demo.subdivision()
```

## envmap:  Environment mapping

```{r}
# RGL-Demo: environment mapping
# Author: Daniel Adler

rgl.demo.envmap <- function() {
  open3d()
  # Clear scene:
  clear3d("all")
  light3d()
  bg3d(sphere=TRUE, color="white", back="filled"
  , texture=system.file("textures/refmap.png",package="rgl")
  )
  data(volcano)

  surface3d( 10*seq_len(nrow(volcano)),10*seq_len(ncol(volcano)),5*volcano
  , texture=system.file("textures/refmap.png",package="rgl")
  , texenvmap=TRUE
  , texmode="modulate"
  , color = "white"
  )
}
rgl.demo.envmap()
```



## shapes3d:  3D shape primitives (cones, ellipsoids, cubes), some taken from qmesh3d

```{r}

cone3d <- function(base=c(0,0,0),tip=c(0,0,1),rad=1,n=30,draw.base=TRUE,qmesh=FALSE,
                   trans = par3d("userMatrix"), ...) {
  ax <- tip-base
  if (missing(trans) && !cur3d()) trans <- diag(4)
  ### is there a better way?
  if (ax[1]!=0) {
    p1 <- c(-ax[2]/ax[1],1,0)
    p1 <- p1/sqrt(sum(p1^2))
    if (p1[1]!=0) {
      p2 <- c(-p1[2]/p1[1],1,0)
      p2[3] <- -sum(p2*ax)
      p2 <- p2/sqrt(sum(p2^2))
    } else {
      p2 <- c(0,0,1)
    }
  } else if (ax[2]!=0) {
    p1 <- c(0,-ax[3]/ax[2],1)
    p1 <- p1/sqrt(sum(p1^2))
    if (p1[1]!=0) {
      p2 <- c(0,-p1[3]/p1[2],1)
      p2[3] <- -sum(p2*ax)
      p2 <- p2/sqrt(sum(p2^2))
    } else {
      p2 <- c(1,0,0)
    }
  } else {
    p1 <- c(0,1,0); p2 <- c(1,0,0)
  }
  degvec <- seq(0,2*pi,length=n+1)[-1]
  ecoord2 <- function(theta) {
    base+rad*(cos(theta)*p1+sin(theta)*p2)
  }
  i <- rbind(1:n,c(2:n,1),rep(n+1,n))
  v <- cbind(sapply(degvec,ecoord2),tip)
  if (qmesh) 
    ## minor kluge for quads -- draw tip twice
    i <- rbind(i,rep(n+1,n))
  if (draw.base) {
    v <- cbind(v,base)
    i.x <- rbind(c(2:n,1),1:n,rep(n+2,n))
    if (qmesh)  ## add base twice
      i.x <-  rbind(i.x,rep(n+2,n))
    i <- cbind(i,i.x)
  }
  if (qmesh) v <- rbind(v,rep(1,ncol(v))) ## homogeneous
  if (!qmesh)
    triangles3d(v[1,i],v[2,i],v[3,i],...)
  else
    return(rotate3d(qmesh3d(v,i,material=list(...)), matrix=trans))
}     


ellipsoid3d <- function(rx=1,ry=1,rz=1,n=30,ctr=c(0,0,0),
                        qmesh=FALSE,
                        trans = par3d("userMatrix"),...) {
  if (missing(trans) && !cur3d()) trans <- diag(4)
  degvec <- seq(0,pi,length=n)
  ecoord2 <- function(p)
    c(rx*cos(p[1])*sin(p[2]),ry*sin(p[1])*sin(p[2]),rz*cos(p[2]))
  v <- apply(expand.grid(2*degvec,degvec),1,ecoord2)
  if (qmesh) v <- rbind(v,rep(1,ncol(v))) ## homogeneous
  e <- expand.grid(1:(n-1),1:n)
  i1 <- apply(e,1,function(z)z[1]+n*(z[2]-1))
  i2 <- i1+1
  i3 <- (i1+n-1) %% n^2 + 1
  i4 <- (i2+n-1) %% n^2 + 1
  i <- rbind(i1,i2,i4,i3)
  if (!qmesh)
    quads3d(v[1,i],v[2,i],v[3,i],...)
  else return(rotate3d(qmesh3d(v,i,material=list(...)),matrix=trans))
}

############
open3d()
ellipsoid3d(ctr=c(2,2,2),rx=3,ry=2,col="red",alpha=0.4)
cone3d(base=c(-2,-2,-2),rad=0.5,tip=c(-3,0,-4),col="blue",front="lines",back="lines")
shade3d(translate3d(cube3d(),3,-2,3,col="purple"))

### now with qmesh()
open3d()
q1 <- cone3d(qmesh=TRUE,trans=diag(4))  ## the "unit cone";
                                        ## height=1,radius=1, base at (0,0,0)
shade3d(q1)
## various transformations and rotations
wire3d(translate3d(q1,3,0,0),col="green")
wire3d(translate3d(scale3d(q1,1,1,2),6,0,0),col="green")
dot3d(translate3d(q1,0,3,0),col="green")
dot3d(translate3d(scale3d(q1,2,1,1),0,6,0),col="green")
shade3d(translate3d(q1,0,0,3),col="red")
shade3d(translate3d(rotate3d(scale3d(q1,1,1,2),pi/4,0,1,0),0,0,6),col="red")

axes3d()
open3d()

s1 <- ellipsoid3d(qmesh=TRUE,trans=diag(4))  ## the "unit sphere";
                                        ## radius=1, ctr at (0,0,0)
shade3d(s1)
## various transformations and rotations
wire3d(translate3d(s1,3,0,0),col="green")
wire3d(translate3d(scale3d(s1,1,1,2),6,0,0),col="green")
dot3d(translate3d(s1,0,3,0),col="green")
dot3d(translate3d(scale3d(s1,2,1,1),0,6,0),col="green")
shade3d(translate3d(s1,0,0,3),col="red")
shade3d(translate3d(rotate3d(scale3d(s1,1,1,2),pi/4,0,1,0),0,0,6),col="red")

axes3d()
```

## lollipop3d:  "Lollipop" plots (3D scatterplot with lines between points and a surface)

```{r}

cone3d <- function(base,tip,rad,n=30,...) {
  degvec <- seq(0,2*pi,length=n)
  ax <- tip-base
  ## what do if ax[1]==0?
  if (ax[1]!=0) {
    p1 <- c(-ax[2]/ax[1],1,0)
    p1 <- p1/sqrt(sum(p1^2))
    if (p1[1]!=0) {
      p2 <- c(-p1[2]/p1[1],1,0)
      p2[3] <- -sum(p2*ax)
      p2 <- p2/sqrt(sum(p2^2))
    } else {
      p2 <- c(0,0,1)
    }
  } else if (ax[2]!=0) {
    p1 <- c(0,-ax[3]/ax[2],1)
    p1 <- p1/sqrt(sum(p1^2))
    if (p1[1]!=0) {
      p2 <- c(0,-p1[3]/p1[2],1)
      p2[3] <- -sum(p2*ax)
      p2 <- p2/sqrt(sum(p2^2))
    } else {
      p2 <- c(1,0,0)
    }
  } else {
    p1 <- c(0,1,0); p2 <- c(1,0,0)
  }
  ecoord2 <- function(theta) {
    base+rad*(cos(theta)*p1+sin(theta)*p2)
  }
    for (i in seq_len(n-1)) {
      li <- ecoord2(degvec[i])
      lj <- ecoord2(degvec[i+1])
      triangles3d(c(li[1],lj[1],tip[1]),c(li[2],lj[2],tip[2]),c(li[3],lj[3],tip[3]),...)
    }   
}

lollipop3d <- function(data.x,data.y,data.z,surf.fun,surf.n=50,
                         xlim=range(data.x),
			 ylim=range(data.y),
			 zlim=range(data.z),
                         asp=c(y=1,z=1),
			 xlab=deparse(substitute(x)),
			 ylab=deparse(substitute(y)),
			 zlab=deparse(substitute(z)),
			 alpha.surf=0.4,
                         col.surf=fg,col.stem=c(fg,fg),
                         col.pt="gray",type.surf="line",ptsize,
                         lwd.stem=2,lit=TRUE,bg="white",fg="black",
                         col.axes=fg,col.axlabs=fg,
                         axis.arrow=TRUE,axis.labels=TRUE,
                         box.col=bg,
                         axes=c("lines","box")) {
  axes <- match.arg(axes)
  col.stem <- rep(col.stem,length=2)
  x.ticks <- pretty(xlim)
  x.ticks <- x.ticks[x.ticks>=min(xlim) & x.ticks<=max(xlim)]
  x.ticklabs <- if (axis.labels) as.character(x.ticks) else NULL
  y.ticks <- pretty(ylim)
  y.ticks <- y.ticks[y.ticks>=min(ylim) & y.ticks<=max(ylim)]
  y.ticklabs <- if (axis.labels) as.character(y.ticks) else NULL
  z.ticks <- pretty(zlim)
  z.ticks <- z.ticks[z.ticks>=min(zlim) & z.ticks<=max(zlim)]
  z.ticklabs <- if (axis.labels) as.character(z.ticks) else NULL
  if (!missing(surf.fun)) {
    surf.x <- seq(xlim[1],xlim[2],length=surf.n)
    surf.y <- seq(ylim[1],ylim[2],length=surf.n)
    surf.z <- outer(surf.x,surf.y,surf.fun)  ## requires surf.fun be vectorized
    z.interc <- surf.fun(data.x,data.y)
    zdiff <- diff(range(c(surf.z,data.z)))
  } else {
    z.interc <- rep(min(data.z),length(data.x))
    zdiff <- diff(range(data.z))
  }
  xdiff <- diff(xlim)
  ydiff <- diff(ylim)
  y.adj <- if (asp[1]<=0) 1 else asp[1]*xdiff/ydiff
  data.y <- y.adj*data.y
  y.ticks <- y.adj*y.ticks
  ylim <- ylim*y.adj
  ydiff <- diff(ylim)
  z.adj <- if (asp[2]<=0) 1 else asp[2]*xdiff/zdiff
  data.z <- z.adj*data.z
  if (!missing(surf.fun)) {
    surf.y <- y.adj*surf.y
    surf.z <- z.adj*surf.z
  }
  z.interc <- z.adj*z.interc
  z.ticks <- z.adj*z.ticks
  zlim <- z.adj*zlim
  open3d()
  clear3d("all")
  light3d()
  bg3d(color=c(bg,fg))
  if (!missing(surf.fun)) 
    surface3d(surf.x,surf.y,surf.z,alpha=alpha.surf,
                front=type.surf,back=type.surf,
                col=col.surf,lit=lit)
  if (missing(ptsize)) ptsize <- 0.02*xdiff
  ## draw points
  spheres3d(data.x,data.y,data.z,r=ptsize,lit=lit,color=col.pt)
  ## draw lollipops
  apply(cbind(data.x,data.y,data.z,z.interc),1,
        function(X) {
          lines3d(x=rep(X[1],2),
                  y=rep(X[2],2),
                  z=c(X[3],X[4]),
                  col=ifelse(X[3]>X[4],col.stem[1],
                    col.stem[2]),lwd=lwd.stem)
        })
  if (axes=="box") {
    bbox3d(xat=x.ticks,xlab=x.ticklabs,
             yat=y.ticks,ylab=y.ticklabs,
             zat=z.ticks,zlab=z.ticklabs,lit=lit)
  } else if (axes=="lines") { ## set up axis lines
    axis3d(edge="x",at=x.ticks,labels=x.ticklabs,
           col=col.axes,arrow=axis.arrow)
    axis3d(edge="y",at=y.ticks,labels=y.ticklabs,
           col=col.axes,arrow=axis.arrow)
    axis3d(edge="z",at=z.ticks,labels=z.ticklabs,
           col=col.axes,arrow=axis.arrow)
    box3d(col=col.axes)
  }
  decorate3d(xlab=xlab, ylab=ylab, zlab=zlab, box=FALSE, axes=FALSE, col=col.axlabs)
}

x <- 1:5
y <- x*10
z <- (x+y)/20
open3d()
spheres3d(x,y,z)
axes3d()
set.seed(1001)
x <- runif(30)
y <- runif(30,max=2)
dfun <- function(x,y)  2*x+3*y+2*x*y+3*y^2 
z <- dfun(x,y)+rnorm(30,sd=0.5)
## lollipops only
lollipop3d(x,y,z)

## lollipops plus theoretical surface

open3d()
lollipop3d(x,y,z,dfun,col.pt="red",col.stem=c("red","blue"))

## lollipops plus regression fit

linmodel <- lm(z~x+y)
dfun <- function(x,y) predict(linmodel,newdata=data.frame(x=x,y=y))
open3d()
lollipop3d(x,y,z,dfun,col.pt="red",col.stem=c("red","blue"))

####
```

---
title: "Using RGL in pkgdown web sites"
author: "Duncan Murdoch"
output:
  markdown::html_format:
    options:
      toc: true
  rmarkdown::html_vignette:
    fig_height: 5
    fig_width: 5
    toc: true
  pdf_document:
    fig_height: 5
    fig_width: 5
    toc: true
  html_document:
    default
vignette: >
  %\VignetteIndexEntry{Using RGL in pkgdown web sites}
  %\VignetteEngine{knitr::knitr}
---

```{r setup, include=FALSE}
if (!requireNamespace("rmarkdown", quietly = TRUE)) {
  warning(call. = FALSE, "These vignettes assume rmarkdown.  This was not found.")
  knitr::knit_exit()
}
knitr::opts_chunk$set(echo = TRUE, snapshot = FALSE, screenshot.force = FALSE)
library(rgl)
options(rgl.useNULL = TRUE)
setupKnitr(autoprint = TRUE)
```


## What is the problem?

[pkgdown](https://pkgdown.r-lib.org/) is an R package that
makes it easy to build a web site for your package.  However, the previous version 1.6.1 on CRAN didn't work so well for
packages whose examples use RGL or other packages like [leaflet](http://rstudio.github.io/leaflet/) that use
[htmlwidgets](http://www.htmlwidgets.org).  This document
described changes that support both of these.  The changes
are now incorporated into `pkgdown` 2.0.0, so everything but
user instructions has been removed.

## Using `RGL in examples.

If you use RGL in example code, they should
appear automatically in examples.
The RGL output is set up to mimic what would happen in a
`knitr` document that uses 
```{r eval=FALSE}
setupKnitr(autoprint = TRUE)
```
i.e. the output RGL commands will automatically be 
included in the display.  Low-level changes will be collected
into a single display:
```{r}
# Show regression plane with z as dependent variable
library(rgl)
open3d()
x <- rnorm(100)
y <- rnorm(100)
z <- 0.2*x - 0.3*y + rnorm(100, sd = 0.3)
fit <- lm(z ~ x + y)

plot3d(x, y, z, type = "s", col = "red", size = 1)

# No plot here, because of the planes3d() call below

coefs <- coef(fit)
a <- coefs["x"]
b <- coefs["y"]
c <- -1
d <- coefs["(Intercept)"]
planes3d(a, b, c, d, alpha = 0.5)
```

## Specifying the size of figures

By default, pkgdown generates standard plots which are wider
than they are high (according to the golden ratio).  Often
RGL plots look better with equal width and height, since
the contents may be rotated by the user.

To specify the size of plots in pkgdown, you use the `figures`
entry in `_pkgdown.yml`.  The defaults are similar to
```
figures:
  dev: ragg::agg_png
  dpi: 96
  dev.args: []
  fig.ext: png
  fig.width: 7
  fig.height: ~
  fig.retina: 2
  fig.asp: 1.618
  bg: NA
  other.parameters: []
```

By default RGL uses these parameters as well, but allows you 
to override any of `fig.width`, `fig.height` and `fig.asp` by
specifying an `rgl` entry in `other.parameters`.  For example:
```
figures:
  fig.width: 5
  other.parameters:
    rgl:
      fig.asp: 1
```
This will make all plots have a width of 5 inches
and will make RGL plots square.

\name{sceneChange}
\alias{sceneChange}
\alias{registerSceneChange}
\title{
Make large change to a scene from Shiny
}
\description{
These functions allow Shiny apps to make relatively large changes
to a scene, adding and removing objects from it.
}
\usage{
sceneChange(elementId, x = scene3d(minimal),
            delete = NULL, add = NULL, replace = NULL,
            material = FALSE, rootSubscene = FALSE,
            delfromSubscenes = NULL, skipRedraw = FALSE,
            minimal = TRUE)
registerSceneChange()
}
\arguments{
  \item{elementId}{
The id of the element holding the \code{rglClass} instance.
}
  \item{x}{
The new scene to use as a source for objects to add.
}
  \item{delete, add, replace}{
Object ids to modify in the scene.  The \code{delete} and
\code{replace} ids must be present in the old scene in the browser;
the \code{add} and \code{replace} ids must be present in \code{x}.
}
  \item{material}{
Logical to indicate whether default material should be updated.
}
  \item{rootSubscene}{
Logical to indicate whether root subscene should be updated.
}
  \item{delfromSubscenes}{
A vector of subscene ids that may have been changed by
deletions.  By default, all subscenes in \code{x} are used,
but the objects may be included in subscenes in the browser
that are different.
}
  \item{skipRedraw}{
If \code{TRUE}, stop the scene from redrawing until \code{skipRedraw=FALSE} is sent.  If \code{NA}, don't redraw this time, but don't
change the state of the \code{skipRedraw} flag.
}
  \item{minimal}{
See \code{\link{scene3d}}.
}
}
\details{
\code{registerSceneChange} must be called in the UI component
of a Shiny app to register the \code{"sceneChange"} custom
message.
}
\value{
\code{registerSceneChange} returns the HTML code to register the
message.

\code{sceneChange} returns a list to be used as the \code{"sceneChange"}
message to change the scene.  Use
\code{\link[shiny:session]{shiny::session$sendCustomMessage}} to send it.
}

\author{
Duncan Murdoch
}
\seealso{
\code{\link{playwidget}} for a different approach to modifying
scenes that can be much faster, but may be less flexible.  The
Shiny demo in this package makes use of all of these approaches.
}
\examples{
\dontrun{
shinyUI(fluidPage(
  registerSceneChange(),
  actionButton("thebutton", "Change")
))

shinyServer(function(input, output, session) {
  observeEvent(input$thebutton, {
    session$sendCustomMessage("sceneChange",
      sceneChange("thewidget", delete = deletes, add = adds))
  })
})
}
}

\name{open3d}
\alias{open3d}
\alias{close3d}
\alias{cur3d}
\alias{set3d}
\alias{getr3dDefaults}
\alias{r3dDefaults}
\alias{rgl.dev.list}
\alias{rgl.quit}

\title{Work with RGL windows}

\description{
  \code{open3d} opens a new RGL window; \code{cur3d} returns the device number of the current
  window; \code{close3d} closes one or more windows.
}
\usage{
open3d(\dots, params = getr3dDefaults(), 
	    useNULL = rgl.useNULL(), silent = FALSE)
	    
close3d(dev = cur3d(), silent = TRUE)
	    
cur3d()

rgl.dev.list()

set3d(dev, silent = FALSE)

getr3dDefaults(class = NULL, value = NULL)

r3dDefaults

rgl.quit()
}

\arguments{
  \item{\dots}{arguments in \code{name = value} form, or a list of named
    values.  The names must come from the graphical
    parameters described in \code{\link{par3d}}.}
  \item{params}{a list of graphical parameters}
  \item{useNULL}{whether to use the null graphics device}
  \item{dev}{which device to close or use}
  \item{silent}{whether report on what was done}
  \item{class, value}{names of components to retrieve}
}

\details{
  \code{open3d} opens a new RGL device, and sets the parameters as
  requested.  The \code{r3dDefaults} list returned by the 
  \code{getr3dDefaults} function will be used as default
  values for parameters.  As installed this sets the point of view to
  'world coordinates' (i.e. x running from left to right, y from front
  to back, z from bottom to top), the \code{mouseMode} to
  \code{(zAxis, zoom, fov)}, and the field of view to 30 degrees.  \code{useFreeType} defaults to \code{FALSE} on 
  Windows; on other systems it indicates the availability
  of FreeType.
  Users may create their own variable named \code{r3dDefaults} in the global
  environment and it will override the installed one.  If there
  is a \code{bg} element in the list or the arguments, it should be
  a list of arguments to pass to the \code{\link{bg3d}} function to 
  set the background.  
  
  The arguments to \code{open3d} may include \code{material}, a list
  of material properties as in \code{\link{r3dDefaults}}, but note
  that high level functions such as \code{\link{plot3d}} normally use
  the \code{r3dDefaults} values in preference to this setting. 
  
  If \code{useNULL} is \code{TRUE}, RGL will use a \dQuote{null}
  device.  This device records objects as they are plotted, but 
  displays nothing. It is intended for use with \code{\link{rglwidget}}.
  

}

\value{
The \code{open3d} function returns the device that 
was opened.  If 
\code{silent = TRUE}, it is returned invisibly.

The \code{cur3d} function returns the current device,
or the value 0 if there isn't one.  \code{rgl.dev.list}
returns a vector of all open devices.  Items are named
according to the type of device:  \code{null} for a hidden
null device, 
\code{wgl} for a Windows device, and \code{glX} for an X windows device.

\code{set3d} returns the device number of the
previously active device.

The \code{close3d} function returns the new current
device, invisibly.

  The \code{r3dDefaults} variable is a list containing default
  settings.  The \code{getr3dDefaults} function searches the user's
  global environment for \code{r3dDefaults} and returns the 
  one in the RGL namespace if it was not found there.
  The components of the list may include any settable \code{par3d}
  parameter, or \code{"material"}, which should include a list
  of default \code{\link{material3d}} properties, or \code{"bg"}, 
  which is a
  list of defaults to pass to the \code{\link{bg3d}} function.
  
  \code{rgl.quit} attempts to unload \pkg{rgl} and then returns
  \code{NULL} invisibly.
}

\seealso{
\code{\link{rgl.useNULL}} for default usage of null device.
}

\examples{
    r3dDefaults
    open3d()
    shade3d(cube3d(color = rainbow(6), meshColor = "faces"))
    cur3d()
}

\keyword{dynamic}

\name{subsceneInfo}
\alias{subsceneInfo}
\title{
Get information on subscenes
}
\description{
This function retrieves information about the tree of subscenes shown in the active
window.
}
\usage{
subsceneInfo(id = NA, embeddings, recursive = FALSE)
}
\arguments{
  \item{id}{
Which subscene to report on; \code{NA} is the current subscene.  Set to \code{"root"} for the root.
}
  \item{embeddings}{
Optional new setting for the embeddings for this subscene.
}
  \item{recursive}{
Whether to report on children recursively.
}
}
\details{
In RGL, each window contains a tree of \dQuote{subscenes}, each containing views of 
a subset of the objects defined in the window.

Rendering in each subscene depends on the viewport, the projection, and the model
transformation.  Each of these characteristics may be inherited from the parent 
(\code{embedding[i] = "inherit"}), may modify the parent (\code{embedding[i] = "modify"}),
or may replace the parent (\code{embedding[i] == "replace"}).  All three must be
specified if \code{embeddings} is used.
}
\value{
\item{id}{The object id of the subscene}
\item{parent}{The object id of the parent subscene, if any}
\item{children}{If \code{recursive}, a list of the information for the children,
otherwise just their object ids.}
\item{embedding}{A vector of 3 components describing how this subscene is
embedded in its parent.
}
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{newSubscene3d}}
}
\examples{
example(plot3d)
subsceneInfo()
}
\keyword{ graphics }

\name{bgplot3d}
\alias{bgplot3d}
\alias{legend3d}
\title{Use base graphics for RGL background}
\description{
Add a 2D plot or a legend in the background of an RGL window.
}
\usage{
bgplot3d(expression, bg.color = getr3dDefaults("bg", "color"), 
         magnify = 1, ...)
legend3d(...)
}
\arguments{
  \item{expression}{
Any plotting commands to produce a plot.
}
  \item{bg.color}{
The color to use for the background.
}
  \item{magnify}{
Multiplicative factor to apply to size of window when producing background 
plot.
  }
  \item{...}{
For \code{legend3d}, arguments to pass to \code{bgplot3d}
or \code{\link{legend}}; for \code{bgplot3d}, 
arguments to pass to \code{\link{bg3d}}.
}
}
\details{
The \code{bgplot3d} function opens a \code{\link{png}} device and executes \code{expression},
producing a plot there.  This plot is then used as a bitmap background for the current
RGL subscene.

The \code{legend3d} function draws a standard 2D legend to the background of the current
subscene by calling \code{bgplot3d} to open a device, and setting up a plot region there 
to fill the whole display.
}
\value{
The \code{bgplot3d} function invisibly returns the ID of the background object that was created,
with attribute \code{"value"} holding the value returned when the \code{expression} was
evaluated.

The \code{legend3d} function does similarly.  The \code{"value"} attribute is the result
of the call to \code{\link{legend}}.   The scaling of the coordinates runs from 0 to 1 
in X and Y.
}
\author{
Duncan Murdoch
}
\note{
Because the background plots are drawn as bitmaps, they do not resize very gracefully.
It's best to size your window first, then draw the background at that size.
}


\seealso{
\code{\link{bg3d}} for other background options.
}
\examples{
x <- rnorm(100)
y <- rnorm(100)
z <- rnorm(100)
open3d()
# Needs to be a bigger window than the default
par3d(windowRect = c(100, 100, 612, 612))
parent <- currentSubscene3d()
mfrow3d(2, 2)
plot3d(x, y, z)
next3d(reuse = FALSE)
bgplot3d(plot(y, z))
next3d(reuse = FALSE)
bgplot3d(plot(x, z))
next3d(reuse = FALSE)
legend3d("center", c("2D Points", "3D Points"), pch = c(1, 16))
useSubscene3d(parent)
}
\keyword{ graphics }

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buffer.R
\name{Buffer}
\alias{Buffer}
\title{R6 Class for binary buffers in glTF files.}
\description{
These files typically have one buffer holding all the
binary data for a scene.
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Buffer-new}{\code{Buffer$new()}}
\item \href{#method-Buffer-load}{\code{Buffer$load()}}
\item \href{#method-Buffer-saveOpenBuffer}{\code{Buffer$saveOpenBuffer()}}
\item \href{#method-Buffer-getBuffer}{\code{Buffer$getBuffer()}}
\item \href{#method-Buffer-setBuffer}{\code{Buffer$setBuffer()}}
\item \href{#method-Buffer-openBuffer}{\code{Buffer$openBuffer()}}
\item \href{#method-Buffer-writeBuffer}{\code{Buffer$writeBuffer()}}
\item \href{#method-Buffer-closeBuffer}{\code{Buffer$closeBuffer()}}
\item \href{#method-Buffer-closeBuffers}{\code{Buffer$closeBuffers()}}
\item \href{#method-Buffer-getBufferview}{\code{Buffer$getBufferview()}}
\item \href{#method-Buffer-addBufferView}{\code{Buffer$addBufferView()}}
\item \href{#method-Buffer-openBufferview}{\code{Buffer$openBufferview()}}
\item \href{#method-Buffer-setBufferview}{\code{Buffer$setBufferview()}}
\item \href{#method-Buffer-getAccessor}{\code{Buffer$getAccessor()}}
\item \href{#method-Buffer-setAccessor}{\code{Buffer$setAccessor()}}
\item \href{#method-Buffer-readAccessor}{\code{Buffer$readAccessor()}}
\item \href{#method-Buffer-readAccessor0}{\code{Buffer$readAccessor0()}}
\item \href{#method-Buffer-addAccessor}{\code{Buffer$addAccessor()}}
\item \href{#method-Buffer-dataURI}{\code{Buffer$dataURI()}}
\item \href{#method-Buffer-as.list}{\code{Buffer$as.list()}}
\item \href{#method-Buffer-clone}{\code{Buffer$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-new"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-new}{}}}
\subsection{Method \code{new()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$new(json = NULL, binfile = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{json}}{list read from glTF file.}

\item{\code{binfile}}{optional External binary filename, or raw vector}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-load"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-load}{}}}
\subsection{Method \code{load()}}{
Load from file.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$load(uri, buf = 0)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{uri}}{Which file to load.}

\item{\code{buf}}{Which buffer number to load.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-saveOpenBuffer"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-saveOpenBuffer}{}}}
\subsection{Method \code{saveOpenBuffer()}}{
Write open buffer to connection.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$saveOpenBuffer(con, buf = 0)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{con}}{Output connection.}

\item{\code{buf}}{Buffer number.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-getBuffer"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-getBuffer}{}}}
\subsection{Method \code{getBuffer()}}{
Get buffer object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$getBuffer(buf, default = list(byteLength = 0))}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{buf}}{Buffer number.}

\item{\code{default}}{Default buffer object if \code{buf} not found.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A list containing components described here:
\url{https://registry.khronos.org/glTF/specs/2.0/glTF-2.0.html#reference-buffer}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-setBuffer"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-setBuffer}{}}}
\subsection{Method \code{setBuffer()}}{
Set buffer object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$setBuffer(buf, buffer)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{buf}}{Buffer number.}

\item{\code{buffer}}{New value to insert.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-openBuffer"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-openBuffer}{}}}
\subsection{Method \code{openBuffer()}}{
Open a connection for the data in a buffer.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$openBuffer(buf)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{buf}}{Buffer number.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
An open binary connection.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-writeBuffer"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-writeBuffer}{}}}
\subsection{Method \code{writeBuffer()}}{
Write data to buffer.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$writeBuffer(values, type, size, buf = 0)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{values}}{Values to write.}

\item{\code{type}}{Type to write.}

\item{\code{size}}{Byte size of each value.}

\item{\code{buf}}{Which buffer to write to.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Byte offset of start of bytes written.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-closeBuffer"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-closeBuffer}{}}}
\subsection{Method \code{closeBuffer()}}{
Close the connection in a buffer.

If there was a connection open, this will save the
contents in the raw vector \code{bytes} within the buffer object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$closeBuffer(buf)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{buf}}{The buffer number.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-closeBuffers"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-closeBuffers}{}}}
\subsection{Method \code{closeBuffers()}}{
Close any open buffers.

Call this after working with a GLTF file to avoid warnings
from R about closing unused connections.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$closeBuffers()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-getBufferview"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-getBufferview}{}}}
\subsection{Method \code{getBufferview()}}{
Get \code{bufferView} object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$getBufferview(bufv)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{bufv}}{\code{bufferView} number.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A list containing components described here:
\url{https://registry.khronos.org/glTF/specs/2.0/glTF-2.0.html#reference-bufferview}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-addBufferView"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-addBufferView}{}}}
\subsection{Method \code{addBufferView()}}{
Add a new buffer view.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$addBufferView(values, type, size, target = NULL, buf = 0)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{values}}{Values to put in the view.}

\item{\code{type}}{Type of values.}

\item{\code{size}}{Size of values in bytes.}

\item{\code{target}}{Optional target use for values.}

\item{\code{buf}}{Which buffer to write to.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
New \code{bufferView} number.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-openBufferview"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-openBufferview}{}}}
\subsection{Method \code{openBufferview()}}{
Open a connection to a buffer view.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$openBufferview(bufv)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{bufv}}{Which \code{bufferView}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A connection.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-setBufferview"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-setBufferview}{}}}
\subsection{Method \code{setBufferview()}}{
Set \code{bufferView} object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$setBufferview(bufv, bufferView)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{bufv}}{\code{bufferView} number.}

\item{\code{bufferView}}{New value to insert.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-getAccessor"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-getAccessor}{}}}
\subsection{Method \code{getAccessor()}}{
Get accessor object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$getAccessor(acc)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{acc}}{Accessor number}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A list containing components described here:
\url{https://registry.khronos.org/glTF/specs/2.0/glTF-2.0.html#reference-accessor}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-setAccessor"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-setAccessor}{}}}
\subsection{Method \code{setAccessor()}}{
Set accessor object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$setAccessor(acc, accessor)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{acc}}{Accessor number.}

\item{\code{accessor}}{New value to insert.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-readAccessor"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-readAccessor}{}}}
\subsection{Method \code{readAccessor()}}{
Read data given by accessor number.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$readAccessor(acc)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{acc}}{Accessor number.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A vector or array as specified in the accessor.  For the \code{MATn} types, the 3rd index
indexes the element.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-readAccessor0"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-readAccessor0}{}}}
\subsection{Method \code{readAccessor0()}}{
Read data given by accessor object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$readAccessor0(accessor)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{accessor}}{Accessor object}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A vector or array as specified in the accessor.  For the \code{MATn} types, the 3rd index
indexes the element.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-addAccessor"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-addAccessor}{}}}
\subsection{Method \code{addAccessor()}}{
Write values to accessor, not including \code{min} and \code{max}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$addAccessor(
  values,
  target = NULL,
  types = "anyGLTF",
  normalized = FALSE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{values}}{Values to write.}

\item{\code{target}}{Optional target use for values.}

\item{\code{types}}{Which types can be used?}

\item{\code{normalized}}{Are normalized integers allowed?}

\item{\code{useDouble}}{Whether to write doubles or singles.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
New accessor number
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-dataURI"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-dataURI}{}}}
\subsection{Method \code{dataURI()}}{
Convert buffer to data URI.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$dataURI(buf = 0)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{buf}}{Buffer to convert.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
String containing data URI.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-as.list"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-as.list}{}}}
\subsection{Method \code{as.list()}}{
Convert to list.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$as.list()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
List suitable for writing using JSON.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Buffer-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Buffer-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Buffer$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Macro to create link to an HTML vignette from a help page.  In
% this version the vignette must be in the same package as the help
% page
% arg 1 is the name of the vignette (without .html)
% arg 2 is an optional anchor within the vignette (without the leading #)
% arg 3 is the text to show on the link.
% Use the RdVignettePath environment variable if vignettes aren't stored
% in ../doc/, e.g. pkgdown puts them in ../articles/.

\newcommand{\HTMLVignette}{\Sexpr[stage=build,results=rd]{
   local({anchor <- "#2"
          if (nchar(anchor))
            anchor <- paste0("#", anchor)
          vigpath <- Sys.getenv("RdVignettePath", unset = "../doc/")
          paste0("\\\\\\\\href{", vigpath, "#1.html",
                 anchor, "}{#3}")
          })}}

% Macro to create link to an HTML vignette in another package on CRAN
% from a help page.
% arg 1 is the package name
% arg 2 is the name of the vignette (without .html)
% arg 3 is an optional anchor within the vignette (without the leading #)
% arg 4 is the text to show on the link.

\newcommand{\HTMLVignetteCRAN}{\Sexpr[stage=build,results=rd]{
  local({anchor <- "#3"
         if (nchar(anchor)) 
           anchor <- paste0("#", anchor)  
         paste0("\\\\\\\\href{https://cran.r-project.org/package=#1/vignettes/#2.html", 
                anchor, "}{#4}")
         })}}

\name{asRow}
\alias{asRow}
\alias{getWidgetId}
\title{
Convenience functions for RGL HTML layouts
}
\description{
The \code{asRow} function arranges objects in a row in the display;
the \code{getWidgetId} function extracts the HTML element ID
from an HTML widget.
}
\usage{
asRow(..., last = NA, height = NULL, colsize = 1)
getWidgetId(widget)
}
\arguments{
  \item{\dots}{
Either a single \code{"combineWidgets"} object produced by
\code{asRow} or a \code{\%>\%} pipe of RGL objects, 
or several objects intended for rearrangement.
}
  \item{last}{
If not \code{NA}, the number of objects from \code{...} that
are to be arranged in a row.  Earlier ones will remain in a column.
}
  \item{height}{
An optional height for the resulting row.  This is normally specified
in pixels, but will be rescaled as necessary to fit the display.
}
  \item{colsize}{
A vector of relative widths for the columns in the row.
}
  \item{widget}{
A single HTML widget from which to extract the HTML element ID.
  }
}
\details{
Using \code{asRow} requires that the \pkg{manipulateWidget}
package is installed.

\code{asRow} produces a \code{"combineWidgets"} object which
is a single column whose last element is another \code{"combineWidgets"} object which is a single row.

If \code{n} objects are given as input and \code{last} is given a value
less than \code{n}, the first \code{n - last} objects will be displayed
in a column above the row containing the \code{last} objects.
}
\value{
\code{asRow} returns a single \code{"combineWidgets"} object suitable for display or nesting
within a more complicated display.

\code{getWidgetId} returns a character string containing the 
HTML element ID of the widget.
}
\author{
Duncan Murdoch
}

\seealso{
\link{pipe} for the \code{\%>\%} operator.  The \HTMLVignette{WebGL}{}{User Interaction in WebGL} vignette
gives more details.
}
\examples{
if (requireNamespace("manipulateWidget", quietly = TRUE) &&
    require("crosstalk", quietly = TRUE)) {
  sd <- SharedData$new(mtcars)
  ids <- plot3d(sd$origData(), col = mtcars$cyl, type = "s")
  # Copy the key and group from existing shared data
  rglsd <- rglShared(ids["data"], key = sd$key(), group = sd$groupName())
  w <- rglwidget(shared = rglsd) \%>\%
       asRow("Mouse mode: ", rglMouse(getWidgetId(.)), 
             "Subset: ", filter_checkbox("cylinderselector", 
		               "Cylinders", sd, ~ cyl, inline = TRUE),
             last = 4, colsize = c(1,2,1,2), height = 60)
  if (interactive() || in_pkgdown_example())
    w
}
}

\name{shapelist3d}
\alias{shapelist3d}
\title{ Create and plot a list of shapes }
\description{
 These functions create and plot a list of shapes.
}
\usage{
shapelist3d(shapes, x = 0, y = NULL, z = NULL, size = 1, matrix = NULL, override = TRUE, 
            ..., plot = TRUE)  
}
\arguments{
  \item{shapes}{ A single \code{shape3d} object, or a list of them. }
  \item{x, y, z}{ Translation(s) to apply }
  \item{size}{ Scaling(s) to apply }
  \item{matrix}{ A single matrix transformation, or a list of them. }
  \item{override}{ Whether the material properties should override the ones in the shapes. }
  \item{\dots}{ Material properties to apply. }
  \item{plot}{ Whether to plot the result. }
}
\details{
 \code{shapelist3d} is a quick way to create a complex object made up of simpler ones.
Each of the arguments \code{shapes} through \code{override} may be a vector of 
values (a list in the case of \code{shapes} or \code{matrix}).  All values
will be recycled to produce a list of shapes as long as the longest of them.

The \code{\link{xyz.coords}} function will be used to process the \code{x},
\code{y} and \code{z} arguments, so a matrix may be used as \code{x} to 
specify all three.   If a vector is used for \code{x} but \code{y} or \code{z}
is missing, default values of \code{0} will be used.

The \code{"shapelist3d"} class is simply a list of \code{"shape3d"} objects.

Methods for \code{\link{dot3d}}, \code{\link{wire3d}}, \code{\link{shade3d}},
\code{\link{translate3d}}, \code{\link{scale3d}}, and \code{\link{rotate3d}}
are defined for these objects.

}
\value{
An object of class \code{c("shapelist3d", "shape3d")}.
}
\author{ Duncan Murdoch }
\seealso{ \code{\link{mesh3d}} }
\examples{

open3d() 
shapelist3d(icosahedron3d(), x = rnorm(10), y = rnorm(10), z = rnorm(10), col = 1:5, size = 0.3)

}
\keyword{ dynamic }

\name{addNormals}
\alias{addNormals}
\alias{addNormals.mesh3d}
\alias{addNormals.shapelist3d}
\title{
Add normal vectors to objects so they render more smoothly
}
\description{
This generic function adds normals at each of the vertices
of a polyhedron by averaging the normals of each incident face.  This has the
effect of making the surface of the object appear smooth rather than faceted
when rendered.
}
\usage{
addNormals(x, ...)
\method{addNormals}{mesh3d}(x, angleWeighted = TRUE, ...)
}
\arguments{
  \item{x}{An object to which to add normals.}
  \item{\dots}{Additional parameters which will be passed to the methods.}
  \item{angleWeighted}{See Details below.}
}
\details{
Currently methods are supplied for \code{\link[=mesh3d]{"mesh3d"}} and 
\code{\link[=shapelist3d]{"shapelist3d"}} classes.

These methods work by averaging the normals on the faces incident at 
each vertex.  By default these are weighted according to the angle 
in the polygon at that vertex.
If \code{angleWeighted = FALSE}, a slightly faster
but less accurate weighting by the triangle area is used.  

Prior to \pkg{rgl} version 0.104.12 an incorrect weighting was used; 
it can be partially reproduced by using \code{angleWeighted = NA}, 
but not all the bugs in that scheme will be kept.
}
\value{
A new object of the same class as \code{x}, with normals added.
}
\author{
Duncan Murdoch
}
\examples{
open3d()
y <- subdivision3d(tetrahedron3d(col = "red"), depth = 3)
shade3d(y) # No normals
y <- addNormals(y)
shade3d(translate3d(y, x = 1, y = 0, z = 0)) # With normals
}
\keyword{dynamic}

\name{contourLines3d}
\alias{contourLines3d}
\alias{contourLines3d.rglId}
\alias{contourLines3d.mesh3d}
\alias{filledContour3d}
\alias{filledContour3d.rglId}
\alias{filledContour3d.mesh3d}
\title{
Draw contours on a surface
}
\description{
\code{contourLines3d} draws contour lines on a surface;
\code{filledContour3d} draws filled contours on it.
}
\usage{
contourLines3d(obj, ...)
\method{contourLines3d}{rglId}(obj, ...)
\method{contourLines3d}{mesh3d}(obj, fn = "z", 
    nlevels = 10, 
    levels = NULL, 
    minVertices = 0,
    plot = TRUE, ... )
filledContour3d(obj, ...)
\method{filledContour3d}{rglId}(obj, plot = TRUE, replace = plot, ...)
\method{filledContour3d}{mesh3d}(obj, fn = "z", 
    nlevels = 20, 
    levels = pretty(range(values), nlevels), 
    color.palette = function(n) hcl.colors(n, "YlOrRd", rev = TRUE),
    col = color.palette(length(levels) - 1),
    minVertices = 0,
    plot = TRUE, 
    keepValues = FALSE, ... )    
}
\arguments{
  \item{obj}{
The object(s) on which to draw contour lines.
}
  \item{fn}{
The function(s) to be contoured.  See Details.
  }
  \item{nlevels}{
Suggested number of contour levels if \code{levels} is not given.
  }
  \item{levels}{
Specified contour values.
  }
  \item{minVertices}{
See Details below.  
  }
  \item{plot}{
Whether to draw the lines or return them in a dataframe.
  }
  \item{\dots}{
For the \code{"mesh3d"} methods, additional parameters to pass to \code{\link{segments3d}}
when drawing the contour lines or to \code{\link{shade3d}}
when drawing the filled contours.  For the \code{"rglId"}
methods, additional parameters to pass to the \code{"mesh3d"}
methods.
}
  \item{replace}{
Whether to delete the objects that are being contoured.
  }
  \item{color.palette}{a color palette function to 
  assign colors in the plot}
  \item{col}{the actual colors to use in the plot.}
  \item{keepValues}{whether to save the function values at 
  each vertex when \code{plot = FALSE}}
}
\details{
For \code{contourLines3d}, the \code{fn} argument can be any
of the following:
\itemize{
\item{ a character vector naming
one or more functions} 
\item{a function}
\item{a numeric vector with one value per vertex}
\item{\code{NULL}, indicating that the numeric values
are saved in \code{obj$values}}
\item{a list containing any of the above.}
}
For \code{filledContour3d}, only one function can be specified.

The special names \code{"x", "y", "z"} may be used in
\code{fn} to specify functions returning one of those
coordinates.  (If you have existing functions \code{x()}, \code{y()}
or \code{z()} they will be masked by this choice; specify
such functions by value rather than name, e.g. \code{fn = x}
instead of \code{fn = "x"}.)

Functions in \code{fn} with formal arguments \code{x},
\code{y} and \code{z} will receive the coordinates of 
vertices in those arguments, otherwise they will receive 
the coordinates in a single n x 3 matrix.   They should
be vectorized and return one value per vertex.

Each of the functions will be evaluated at each vertex
of the surface specified by \code{obj}, and contours will
be drawn assuming the function is linear between vertices.
If contours of a nonlinear function are needed, you may
want to increase \code{minVertices} as described below.

If \code{levels} is not specified, values will be set
separately for each entry in \code{fn}, using 
\code{pretty(range(values, na.rm = TRUE), nlevels)} where
\code{values} are the values on the vertices.

The \code{minVertices} argument is used to improve the 
approximation to the contour when the function is non-linear.
In that case, the interpolation between vertices 
can be inaccurate.  If \code{minVertices} is set to a
positive
number (e.g. \code{10000}), then the mesh is modified
by subdivision to have at least that number of vertices, 
so that pieces are smaller and the linear interpolation
is more accurate. 
}
\note{
To draw contours on a surface, the surface should be drawn
with material property \code{polygon_offset = 1} (or perhaps
some larger positive value) so that the lines of the contour are not
obscured by the surface.

In R versions prior to 3.6.0, the default \code{color.palette}
is \code{grDevices::cm.colors}.
}
\value{
For both \code{contourLines3d} and \code{filledContour3d}
the \code{"rglId"} method converts the given id values to
a mesh, and calls the \code{"mesh3d"} method.  

The \code{"mesh3d"} method returns an object of class
\code{"rglId"} corresponding to what was
drawn if \code{plot} is \code{TRUE},

If \code{plot} is \code{FALSE}, \code{contourLines3d} returns a dataframe containing 
columns \code{c("x", "y", "z", "fn", "level")} giving
the coordinates of the endpoints of each line segment, 
the name (or index) of the function for this contour, and the 
level of the contour.  

If \code{plot} is \code{FALSE}, \code{filledContour3d}
returns a \code{"\link{mesh3d}"} object holding the result.
If \code{keepValues} is \code{TRUE}, the mesh
will contain the values corresponding to each vertex
(with linear approximations at the boundaries).
}

\author{
Duncan Murdoch
}
\seealso{
The \pkg{misc3d} package contains the function \code{\link[misc3d]{contour3d}}
to draw contour surfaces in space instead of contour lines
on surfaces.
}
\examples{

# Add contourlines in "z" to a persp plot

z <- 2 * volcano        # Exaggerate the relief
x <- 10 * (1:nrow(z))   # 10 meter spacing (S to N)
y <- 10 * (1:ncol(z))   # 10 meter spacing (E to W)

open3d()
id <- persp3d(x, y, z, aspect = "iso",
      axes = FALSE, box = FALSE, polygon_offset = 1)
contourLines3d(id)     # "z" is the default function
filledContour3d(id, polygon_offset = 1, nlevels = 10, replace = TRUE)

# Draw longitude and latitude lines on a globe

lat <- matrix(seq(90, -90, length.out = 50)*pi/180, 50, 50, byrow = TRUE)
long <- matrix(seq(-180, 180, length.out = 50)*pi/180, 50, 50)

r <- 6378.1 # radius of Earth in km
x <- r*cos(lat)*cos(long)
y <- r*cos(lat)*sin(long)
z <- r*sin(lat)

open3d()
ids <- persp3d(x, y, z, col = "white", 
        texture = system.file("textures/worldsmall.png", package = "rgl"), 
        specular = "black", axes = FALSE, box = FALSE, xlab = "", ylab = "", zlab = "",
        normal_x = x, normal_y = y, normal_z = z, polygon_offset = 1)
        
contourLines3d(ids, list(latitude = function(x, y, z) asin(z/sqrt(x^2+y^2+z^2))*180/pi,
                         longitude = function(x, y, z) atan2(y, x)*180/pi))
}

\name{polygon3d}
\alias{polygon3d}
\title{
Draw a polygon in three dimensions
}
\description{
This function takes a description of a flat polygon in x, y and z coordinates,
and draws it in three dimensions.
}
\usage{
polygon3d(x, y = NULL, z = NULL, fill = TRUE, plot = TRUE, 
          coords, random = TRUE, ...)
}
\arguments{
  \item{x, y, z}{
Vertices of the polygon in a form accepted by \code{\link{xyz.coords}}.
}
  \item{fill}{
logical; should the polygon be filled?
}
  \item{plot}{
logical; should the polygon be displayed?
}
  \item{coords}{
Which two coordinates (\code{x = 1}, \code{y = 2}, \code{z = 3}) describe the
polygon.  If missing, \code{\link{triangulate}} makes
an automatic choice.
}
  \item{random}{
Currently ignored.  The triangulation is deterministic.
}
  \item{\dots}{
Other parameters to pass to \code{\link{lines3d}} or \code{\link{shade3d}} if \code{plot = TRUE}.
}
}
\details{
The function triangulates the two dimensional polygon described by \code{coords}, then
applies the triangulation to all three coordinates.  No check is made that the polygon
is actually all in one plane, but the results may be somewhat unpredictable
(especially if \code{random = TRUE}) if it is not.

Polygons need not be simple; use \code{NA} to indicate separate closed pieces.  For
\code{fill = FALSE} there are no other restrictions on the pieces, but for 
\code{fill = TRUE} the resulting two-dimensional polygon needs to be one that
\code{\link{triangulate}} can handle.
}
\value{
If \code{plot = TRUE}, the id number of the lines (for \code{fill = FALSE}) or triangles
(for \code{fill = TRUE}) that have been plotted.

If \code{plot = FALSE}, then for \code{fill = FALSE}, a vector of indices into
the XYZ matrix that could be used to draw the polygon.  For \code{fill = TRUE}, 
a triangular mesh object representing the triangulation.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{extrude3d}} for a solid extrusion of a polygon, \code{\link{triangulate}} for
the triangulation.
}
\examples{
theta <- seq(0, 4*pi, length.out = 50)
r <- theta + 1
r <- c(r[-50], rev(theta*0.8) + 1)
theta <- c(theta[-50], rev(theta))
x <- r*cos(theta)
y <- r*sin(theta)
open3d()
plot(x, y, type = "n")
polygon(x, y)
polygon3d(x, y, x + y, col = "blue")
}
\keyword{ graphics }

\name{propertyControl}
\alias{subsetControl}
\alias{propertyControl}
\title{
Controls to use with playwidget()
}
\description{
These are setter functions to produce actions in a Shiny app,
or in an animation.
}
\usage{
subsetControl(value = 1, subsets, subscenes = NULL,
                         fullset = Reduce(union, subsets),
                         accumulate = FALSE)
propertyControl(value = 0, entries, properties, 
                objids = tagged3d(tags), tags, values = NULL,
                param = seq_len(NROW(values)) - 1, interp = TRUE)
}
\arguments{
  \item{value}{The value to use for input (typically \code{input$value}
  in a Shiny app.)}
  \item{subsets}{A list of vectors of object identifiers; the value will
  choose among them.}
  \item{fullset}{Objects in the subscene which are not in \code{fullset}
  will not be touched.}
  \item{subscenes}{The subscenes to be controlled.  If \code{NULL}, the
  root subscene.}
  \item{accumulate}{If \code{TRUE}, the subsets will accumulate (by union)
  as the value increases.}
  \item{entries, properties, objids}{Which properties to set.}
  \item{tags}{
Select objects with matching tags.  Ignored
if \code{objids} is specified.
}
  \item{values}{Values to set.}
  \item{param}{Parameter values corresponding to the rows of \code{value}}
  \item{interp}{Whether to use linear interpolation between \code{param}
  values}
}
\details{
\code{subsetControl} produces data for \code{\link{playwidget}} to
display subsets of the object in one or more subscenes.  This code will not
touch objects in the
subscenes if they
are not in \code{fullset}.  \code{fullset} defaults to the union of
all the object ids mentioned in \code{subsets}, so by default if an id is
not mentioned in one of the subsets, it will not be controlled by the slider.  If \code{value} is specified in R code, it will be a 1-based
index into the \code{subsets} list; when specified internally
in Javascript, 0-based indexing into the corresponding array
will be used.

\code{propertyControl} sets individual properties.  Here the row
of \code{values} is determined by the position of \code{value}
in \code{param}.
}
\value{
These functions return controller data in a list of class
\code{"rglControl"}.
}
\author{
Duncan Murdoch
}

\seealso{
\code{\link{subsetSetter}} for a way to embed a pure Javascript
control, and \code{\link{playwidget}} for a way to use these
in animations (including Shiny), \code{\link{rglShared}}
for linking using the \pkg{crosstalk} package.  The \HTMLVignette{WebGL}{}{User Interaction in WebGL} vignette
gives more details.
}

\name{plot3d.formula}
\alias{plot3d.formula}
\alias{persp3d.formula}
\title{
Methods for formulas
}
\description{
These functions provide a simple formula-based interface
to \code{\link{plot3d}} and \code{\link{persp3d}}.
}
\usage{
\method{plot3d}{formula}(x, data = NULL, xlab, ylab, zlab, ...)
\method{persp3d}{formula}(x, data = NULL, xlab, ylab, zlab, ...)
}
\arguments{
  \item{x}{
A formula like \code{z ~ x + y}.
}
  \item{data}{
An optional dataframe or list in which to find the components
of the formula.
}
  \item{xlab, ylab, zlab}{
Optional axis labels to override the ones automatically
obtained from the formula.
}
  \item{\dots}{
Additional arguments to pass to the default \code{plot3d}
method, or the \code{persp3d} method for \code{"deldir"}
objects.
}
}
\details{
Only simple formulas (the ones handled by the \code{\link{xyz.coords}} function) are supported:  a single variable on the left hand side (which will be plotted on the Z axis),
and a sum of two variables on the right hand side (which
will be the X and Y axis variables in the plot.)
}
\note{
The \code{persp3d} method requires that the suggested package \pkg{deldir} is installed.
}
\value{
These functions are called for the side effect of drawing
the plots. The \code{plot3d} method draws a scatterplot.  The \code{persp3d} method draws a surface plot.

Return values are as given by the \code{\link{plot3d.default}} method or the \code{\link{persp3d.deldir}} 
methods.
}
\author{
Duncan Murdoch
}
\examples{
open3d()
mfrow3d(1, 2, sharedMouse = TRUE)
plot3d(mpg ~ wt + qsec, data = mtcars)
if (checkDeldir())
  persp3d(mpg ~ wt + qsec, data = mtcars)
}

\name{grid3d}
\alias{grid3d}
\title{Add a grid to a 3D plot }
\description{
This function adds a reference grid to an RGL plot.
}
\usage{
grid3d(side, at = NULL, col = "gray", lwd = 1, lty = 1, n = 5)
}
\arguments{
  \item{side}{ Where to put the grid; see the Details section. }
  \item{at}{ How to draw the grid; see the Details section. }
  \item{col}{ The color of the grid lines. }
  \item{lwd}{ The line width of the grid lines. (Currently only
  \code{lty = 1} is supported.)}
  \item{lty}{ The line type of the grid lines. }
  \item{n}{ Suggested number of grid lines; see the Details section. }
}
\details{
This function is similar to \code{\link{grid}} in classic graphics,
except that it draws a 3D grid in the plot.

The grid is drawn in a plane perpendicular to the coordinate axes. The
first letter of the \code{side} argument specifies the direction of
the plane: \code{"x"}, \code{"y"} or \code{"z"} (or uppercase
versions) to specify the coordinate which is constant on the plane.

If \code{at = NULL} (the default), the grid is drawn at the limit of 
the box around the data.  If the second letter of the \code{side} argument
is \code{"-"} or is not present, it is the lower limit; if \code{"+"}
then at the upper limit.  The grid lines are drawn at values
chosen by \code{\link{pretty}} with \code{n} suggested locations.
The default locations should match those chosen by \code{\link{axis3d}}
with \code{nticks = n}.

If \code{at} is a numeric vector, the grid lines are drawn at those values.

If \code{at} is a list, then the \code{"x"} component is used to
specify the x location, the \code{"y"} component specifies the y location, and
the \code{"z"} component specifies the z location.  Missing components
are handled using the default as for \code{at = NULL}.

Multiple grids may be drawn by specifying multiple values for \code{side}
or for the component of \code{at} that specifies the grid location.
}
\note{
If the scene is resized, the grid will not be resized; use \code{\link{abclines3d}}
to draw grid lines that will automatically resize.
}
\value{ A vector or matrix of object ids is returned invisibly. }
\author{ Ben Bolker and Duncan Murdoch }
\seealso{ \code{\link{axis3d}} }
\examples{
x <- 1:10
y <- 1:10
z <- matrix(outer(x - 5, y - 5) + rnorm(100), 10, 10)
open3d()
persp3d(x, y, z, col = "red", alpha = 0.7, aspect = c(1, 1, 0.5))
grid3d(c("x", "y+", "z"))
}
\keyword{ dynamic }

\name{as.triangles3d}
\alias{as.triangles3d}
\alias{as.triangles3d.rglId}
\title{
Convert an object to triangles
}
\description{
This generic and its methods extract or creates a matrix of coordinates
of triangles from an object, suitable for passing to \code{\link{triangles3d}}.
}
\usage{
as.triangles3d(obj, ...)
\method{as.triangles3d}{rglId}(obj,
               attribute = c("vertices", "normals", "texcoords", "colors"),
               subscene = NA,
               ...)
}
\arguments{
  \item{obj}{
The object to convert.
}
  \item{attribute}{Which attribute of an RGL object to extract?}
  \item{subscene}{Which subscene is this object in?}
  \item{\dots}{
Additional arguments used by the methods.
}
}
\details{
The method for \code{"rglId"} objects can extract several different
attributes, organizing them as it would organize the vertices for
the triangles.
}
\value{
An \code{n x 3} matrix containing the vertices of triangles making
up the object.  Each successive 3 rows of the matrix corresponds to
a triangle.

If the attribute doesn't exist, \code{NULL} will be returned.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{as.mesh3d}} to also capture material properties.
}
\examples{
open3d()
x <- surface3d(x = 1:10, y = 1:10, z = rnorm(100), col = "red")
tri <- as.triangles3d(x)
open3d()
triangles3d(tri, col = "blue")
}

\name{shiny}
\alias{rglwidgetOutput}
\alias{renderRglwidget}
\alias{playwidgetOutput}
\alias{renderPlaywidget}
\title{
Functions for integration of RGL widgets into
Shiny app
}
\description{
These functions allow an RGL scene to be embedded
in a Shiny app.
}
\usage{
rglwidgetOutput(outputId, width = "512px", height = "512px")
renderRglwidget(expr, env = parent.frame(), quoted = FALSE, outputArgs = list())

playwidgetOutput(outputId, width = "0px", height = "0px")
renderPlaywidget(expr, env = parent.frame(), quoted = FALSE, outputArgs = list())
}
\arguments{
  \item{outputId}{
The name for the control.
}
  \item{width, height}{
Width and height to display the control.
}
  \item{expr}{An R expression returning
a \code{\link{rglwidget}} (for \code{renderRglwidget}) or a \code{\link{playwidget}} (for \code{renderPlaywidget}) as output.}
  \item{env}{The environment in which to evaluate \code{expr}.}
  \item{quoted}{Is the expression already quoted?}
  \item{outputArgs}{A list containing arguments; see details below.}
}
\details{
Use \code{rglwidgetOutput} or \code{playwidgetOutput} as an output
object in a Shiny user
interface section; use \code{renderRglwidget} or \code{renderPlaywidget}
as the render
function in the server section.

In a dynamic R Markdown document with \code{runtime: shiny}, you 
only call the render function, and may optionally pass \code{width}
and \code{height} to the output function by putting them in
a list in \code{outputArgs}.  See the example below.
}
\value{
Used internally by Shiny.
}
\author{
Duncan Murdoch
}
\examples{
\dontrun{
# This could be used in a dynamic R Markdown document.  See
# demo("shinyDemo") and demo("simpleShinyRgl") for Shiny apps.

inputPanel(
  sliderInput("n", label = "n", min = 10, max = 100, value = 10, step = 10)
)

renderRglwidget({
    n <- input$n
    try(close3d())
    plot3d(rnorm(n), rnorm(n), rnorm(n))
    rglwidget()
  }, outputArgs = list(width = "auto", height = "300px"))

}
}

\name{webGLcontrols}
\alias{subsetSlider}
\alias{subsetSetter}
\alias{clipplaneSlider}
\alias{toggleButton}
\title{
Obsolete functions to write HTML/Javascript code to control a WebGL display.
}
\description{
These functions write out HTML code to control WebGL displays on the same page.  They are deprecated; most documentation has
now been removed.
}
\usage{
subsetSlider(subsets, labels = names(subsets), 
             fullset = Reduce(union, subsets),
             subscenes = currentSubscene3d(), prefixes = "",
             accumulate = FALSE, ...)
             
subsetSetter(subsets, subscenes = currentSubscene3d(), prefixes = "", 
			 fullset = Reduce(union, subsets),
                         accumulate = FALSE)
             
clipplaneSlider(a=NULL, b=NULL, c=NULL, d=NULL, 
                plane = 1, clipplaneids, prefixes = "", 
                labels = signif(values[,1],3), ...)
             
toggleButton(subset, subscenes = currentSubscene3d(), prefixes = "", 
             label = deparse(substitute(subset)), 
             id = paste0(basename(tempfile("input"))), name = id) 
}
\arguments{
  \item{subsets}{A list of vectors of object identifiers; the slider or setter
  will choose among them.}
  \item{labels}{Labels to display corresponding to each subset.  If \code{NULL},
  numeric labels will be shown.}
  \item{fullset}{Objects in the subscene which are not in \code{fullset}
  will not be touched.}
  \item{subscenes}{The subscenes to be controlled.}  
  \item{prefixes}{The prefixes of the WebGL scenes to be controlled.}
  \item{accumulate}{If \code{TRUE}, the subsets will accumulate (by union)
  as the value increases.}
  \item{id}{The \code{id} of the input control that will be generated.}
  \item{name}{The name of the input control that will be generated.}
  \item{...}{Arguments to pass to \code{\link{propertySlider}}.}
  \item{a,b,c,d}{The parameter values to change.  Leave as \code{NULL}
  to hold the parameter constant.}
  \item{plane, clipplaneids}{The identifier of the particular clipplane
  to modify.}
  \item{subset}{The subset that the button should toggle.}
  \item{label}{The button label.}
}
\value{
\code{subsetSetter} returns a length-one character vector
of class \code{"propertySetter"}.  

The other functions 
use \code{\link{cat}} to write their output and invisibly return the 
\code{id} of the control that was generated.
}
\author{
Duncan Murdoch
}

\seealso{
\code{\link{playwidget}} and \code{\link{toggleWidget}}
for a newer, preferred method of inserting controls
into a scene.
}

\name{getBoundary3d}
\alias{getBoundary3d}
\title{
Extract the boundary of a mesh
}
\description{
Constructs a mesh of line segments corresponding to
non-shared (i.e. boundary) edges of triangles or quads
in the original mesh.
}
\usage{
getBoundary3d(mesh, sorted = FALSE, simplify = TRUE, ...)
}
\arguments{
  \item{mesh}{
A mesh object.
}
  \item{sorted}{
Whether the result should have the segments sorted in
sequential order.
}
  \item{simplify}{
Whether to simplify the resulting mesh, dropping all unused
vertices.  If \code{FALSE}, the vertices of the result will 
be identical to the vertices of \code{mesh}; if \code{TRUE},
they will likely be different, even if no vertices were dropped.
  }
  \item{...}{
Material properties to apply to the mesh.
  }
}
\value{
A \code{"mesh3d"} object containing 0 or more segments.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{mesh3d}}
}
\examples{
x <- cube3d(col = "blue")
x$ib <- x$ib[,-(1:2)]
b <- getBoundary3d(x, sorted = TRUE, col = "black")

open3d()
shade3d(x, alpha=0.2)

shade3d(b) 

# Show edge vertices in sequence:
text3d(t(b$vb), texts = 1:ncol(b$vb), adj = 0)
c(b$is[1,1], b$is[2,]) 
}

\name{scene}
\alias{clear3d}
\alias{pop3d}
\alias{ids3d}
\title{Scene management}
\description{
Clear shapes, lights, bbox
}
\usage{
clear3d( type = c("shapes", "bboxdeco", "material"), defaults, subscene = 0 ) 
pop3d( type = "shapes", id = 0, tag = NULL)
ids3d( type = "shapes", subscene = NA, tags = FALSE )
}
\arguments{
  \item{type}{Select subtype(s):
    \describe{
      \item{"shapes"}{shape stack}
      \item{"lights"}{light stack}
      \item{"bboxdeco"}{bounding box}
      \item{"userviewpoint"}{user viewpoint}
      \item{"modelviewpoint"}{model viewpoint}
      \item{"material"}{material properties}
      \item{"background"}{scene background}
      \item{"subscene"}{subscene list}
      \item{"all"}{all of the above}
    }
  }
  \item{defaults}{default values to use after clearing}
  \item{subscene}{which subscene to work with.  \code{NA} means the current one, 
                  \code{0} means the whole scene}  
  \item{id}{vector of ID numbers of items to remove}
  \item{tag}{override \code{id} with objects matching
  these \code{tag} material properties}
  \item{tags}{logical; whether to return \code{tag} column.}
}
\details{
RGL holds several lists of objects in each scene. 
There are lists for shapes, lights, bounding box decorations, subscenes, etc.
\code{clear3d} clears the specified stack, or restores
the defaults for the bounding box (not visible) or viewpoint. 
With \code{id = 0} \code{pop3d} removes 
the last added node on the list (except for subscenes:  there it
removes the active subscene).  The \code{id} argument
may be used to specify arbitrary item(s) to remove; if \code{id != 0}, 
the \code{type} argument is ignored.

\code{clear3d} may also be used to clear material properties
back to their defaults.  

\code{clear3d} has an optional \code{defaults} argument, which defaults to 
\code{\link{r3dDefaults}}.  Only the \code{materials} component of this argument
is currently used by \code{clear3d}.

\code{ids3d} returns a dataframe containing the IDs in the currently active subscene
by default, or a specified subscene, or if \code{subscene = 0}, in the whole
rgl window along with an indicator of their type and if \code{tags = TRUE}, the
\code{tag} value for each.

Note that clearing the light stack leaves the scene in darkness; it should normally
be followed by a call to \code{\link{light3d}}.
}
\seealso{
\code{\link{rgl}},
\code{\link{bbox3d}},
\code{\link{light3d}},
\code{\link{open3d}} to open a new window.
}
\examples{
  x <- rnorm(100)
  y <- rnorm(100)
  z <- rnorm(100)
  p <- plot3d(x, y, z, type = 's', tag = "plot")
  ids3d()
  lines3d(x, y, z)
  ids3d(tags = TRUE)
  if (interactive() && !rgl.useNULL()) {
    readline("Hit enter to change spheres")
    pop3d(id = p["data"])
    spheres3d(x, y, z, col = "red", radius = 1/5)
    box3d()
  }
}
\keyword{dynamic}

\name{mergeVertices}
\alias{mergeVertices}
\title{
Merge duplicate vertices in mesh object
}
\description{
A mesh object can have the same vertex listed twice.  Each copy is
allowed to have separate normals, texture coordinates, and color.
However, it is more efficient to have just a single copy if those
differences aren't needed.  For automatic smoothing using 
\code{\link{addNormals}}, triangles and quads need to share vertices.
This function merges identical (or similar) vertices to achieve this.
}
\usage{
mergeVertices(mesh, 
              notEqual = NULL, 
              attribute = "vertices", 
              tolerance = sqrt(.Machine$double.eps))
}
\arguments{
  \item{mesh}{
A \code{\link{mesh3d}} object.
}
  \item{notEqual}{
A logical matrix indicating that certain pairs should not be merged
even if they appear identical.
}
  \item{attribute}{
Which attribute(s) should be considered in comparing vertices? 
A vector chosen from \code{c("vertices", "colors", "normals", "texcoords"))}
}
  \item{tolerance}{
When comparing vertices using \code{\link{all.equal}}, this tolerance
will be used to ignore rounding error.
}
}

\value{
A new mesh object.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{as.mesh3d.rglId}}, which often constructs mesh
objects containing a lot of duplication.
}
\examples{
open3d()
(mesh1 <- cuboctahedron3d(col = rainbow(14), meshColor = "face"))
id <- shade3d(mesh1)
(mesh2 <- as.mesh3d(id))
shade3d(translate3d(mesh2, 3, 0, 0))
(mesh3 <- mergeVertices(mesh2))
shade3d(translate3d(mesh3, 6, 0, 0))
}

\name{rgl.open}
\title{3D visualization device system}
\alias{rgl.open}
\alias{rgl.close}
\alias{rgl.cur}
\alias{rgl.set}
\alias{rgl.quit}
\alias{rgl.antialias}
\alias{rgl.dev.list}
\description{
  3D real-time rendering system.
}
\usage{
# Low level rgl.* interface
rgl.open(useNULL = rgl.useNULL())     # open new device
rgl.close()    # close current device
rgl.cur()      # returns active device ID
rgl.dev.list() # returns all device IDs
rgl.set(which, silent = FALSE) # set device as active
rgl.quit()     # shutdown rgl device system
}
\arguments{
  \item{useNULL}{whether to open the \dQuote{null} device}
  \item{which}{device ID}
  \item{silent}{whether to suppress update of window titles}
}
\details{
  The \pkg{rgl} device design is oriented towards the R device metaphor. If you send
  scene management instructions, and there's no device open, it will be opened
  automatically.
  Opened devices automatically get the current device focus. The focus may be
  changed by using \code{rgl.set()}.
  \code{rgl.quit()} shuts down the rgl subsystem and all open devices, 
  detaches the package including the shared library and additional system libraries.
  
  The \code{rgl.open()} function attempts to open a new RGL window.
  If the \code{"rgl.antialias"} option is set, it will be used to
  select the requested antialiasing.  (See \code{\link{open3d}} for
  more description of antialiasing and an alternative way to set the
  value.)
  
  If \code{useNULL} is \code{TRUE}, \pkg{rgl} will use a \dQuote{null}
  device.  This device records objects as they are plotted, but 
  displays nothing. It is intended for use with \code{\link{rglwidget}}
  and similar functions.
  
  If \code{rgl.open()} fails (e.g. because X windows is not running, or its
  \code{DISPLAY} variable is not set properly), then you can retry the initialization
  by calling \code{\link{rgl.init}()}.  Do not do this when windows have 
  already been successfully opened:  they will be orphaned, with no way to 
  remove them other than closing R.  In fact, it's probably a good idea not to do this
  at all:  quitting R and restarting it is a better solution.

  This package also includes a higher level interface which is described
  in the \link{r3d} help topic.  That interface is designed to act more like
  classic 2D R graphics.  We recommend that you avoid mixing
  \code{rgl.*} and \code{*3d} calls.
}
\value{
  \code{rgl.open}, \code{rgl.close} and \code{rgl.set} are called for their side effects
  and return no useful value.  Similarly \code{rgl.quit} is not designed
  to return useful values; in fact, users shouldn't call it at all!
  
  \code{rgl.cur} returns the currently active devices, or \code{0} if none is
  active; \code{rgl.dev.list} returns a vector of all open devices.  Both functions name
  the items according to the type of device:  \code{null} for a hidden null device, 
  \code{wgl} for a Windows device, and \code{glX} for an X windows device.
}
\seealso{
\link{r3d},
\code{\link{rgl.init}},
\code{\link{rgl.clear}}, 
\code{\link{rgl.pop}},
\code{\link{rgl.viewpoint}},
\code{\link{rgl.light}},
\code{\link{rgl.bg}},
\code{\link{rgl.bbox}},
\code{\link{rgl.points}},
\code{\link{rgl.lines}},
\code{\link{rgl.triangles}},
\code{\link{rgl.quads}},
\code{\link{rgl.texts}},
\code{\link{rgl.surface}},
\code{\link{rgl.spheres}},
\code{\link{rgl.sprites}},
\code{\link{rgl.snapshot}},
\code{\link{rgl.useNULL}}
}
\keyword{dynamic}

\name{identify3d}
\alias{identify3d}
\title{
Identify points in plot
}
\description{
Identify points in a plot, similarly to the \code{\link{identify}} function
in base graphics.
}
\usage{
identify3d(x, y = NULL, z = NULL, labels = seq_along(x), n = length(x), 
	plot = TRUE, adj = c(-0.1, 0.5), tolerance = 20, 
	buttons = c("right", "middle"))
}
\arguments{
  \item{x, y, z}{coordinates of points in a scatter plot.  Alternatively, any
    object which defines coordinates (see \code{\link{xyz.coords}}) 
    can be given as \code{x}, and \code{y} and \code{z} left missing.}
  \item{labels}{an optional character vector giving labels for the
    points.  Will be coerced using \code{\link{as.character}}, and
    recycled if necessary to the length of \code{x}.}
  \item{n}{the maximum number of points to be identified.}
  \item{plot}{logical: if \code{plot} is \code{TRUE}, the labels are
    printed near the points and if \code{FALSE} they are omitted.}
  \item{adj}{numeric vector to use as \code{adj} parameter to 
    \code{\link{text3d}} when plotting the labels.}
  \item{tolerance}{the maximal distance (in pixels) 
    for the pointer to be \sQuote{close enough} to a point.}
  \item{buttons}{a length 1 or 2 character vector giving the buttons
    to use for selection and quitting.}
}
\details{
If \code{buttons} is length 1, the user can quit by reaching \code{n}
selections, or by hitting the 
escape key, but the result will be lost if escape is used.
}
\value{
A vector of selected indices.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{identify}} for base graphics, \code{\link{select3d}} for selecting regions.
}
\keyword{ graphics }

\name{writeOBJ}
\alias{writeOBJ}
\alias{readOBJ}
\title{
Read and write Wavefront OBJ format files
}
\description{
\code{writeOBJ} writes OBJ files.  This is a file format that
is commonly used in 3D graphics applications.  It does not represent text,
but does represent points, lines, polygons (and many other types that
RGL doesn't support).  \code{readOBJ} reads only some parts
of OBJ files.
}
\usage{
writeOBJ(con, 
         pointRadius = 0.005, pointShape = icosahedron3d(), 
         lineRadius = pointRadius, lineSides = 20, 
         pointsAsPoints = FALSE, linesAsLines = FALSE, 
         withNormals = TRUE, withTextures = TRUE, 
         separateObjects = TRUE,
         ids = NULL)
readOBJ(con, ...)
}
\arguments{
  \item{con}{
A connection or filename.
}
  \item{pointRadius, lineRadius}{
The radius of points and lines relative to the overall scale of the figure,
if they are converted to polyhedra.
}
  \item{pointShape}{
A mesh shape to use for points if they are converted.  It is scaled by the \code{pointRadius}.
}
  \item{lineSides}{
Lines are rendered as cylinders with this many sides.
}
  \item{pointsAsPoints, linesAsLines}{
Whether to convert points and lines to \dQuote{point} and \dQuote{line} records in the OBJ output.
}
  \item{withNormals}{
Whether to output vertex normals for smooth shading.
}
  \item{separateObjects}{
Whether to mark each RGL object as a separate object in the file.
}
  \item{withTextures}{
Whether to output texture coordinates.
}
  \item{ids}{
The identifiers (from \code{\link{ids3d}}) of the
objects to write.  If \code{NULL}, try to write everything.
}

  \item{...}{
Additional arguments (typically just \code{material}) to pass to
\code{\link{tmesh3d}}.
  }

}
\details{

The current \code{writeOBJ} implementation only outputs triangles, quads, planes, spheres,
points, line segments, line strips and surfaces.  It does not output material
properties such as colors, since the OBJ format does not support the per-vertex
colors that RGL uses.  

The \code{readOBJ} implementation can
read faces, normals, and textures coordinates, but ignores
material properties (including the specification of the texture
file to use).  To read a file that uses a single texture,
specify it in the \code{material} argument, e.g.
\code{readOBJ("model.OBJ", material = list(color = "white", texture = "texture.png"))}.  There is no support for files that use
multiple textures.

The defaults for \code{pointsAsPoints} and \code{linesAsLines} have been
chosen because Blender (\url{https://www.blender.org}) does not import
points or lines, only polygons.  If you are exporting to other software you
may want to change them.

If present, texture coordinates are output by default, but the
textures themselves are not.

Individual RGL objects are output as separate objects in the file when 
\code{separateObjects = TRUE}, the default.

The output file should be readable by Blender and Meshlab; the latter can write in
a number of other formats, including U3D, suitable for import into a PDF document.
}
\value{
\code{writeObj} invisibly returns the name of the connection to which the 
data was written.

\code{readObj} returns a mesh object constructed
from the input file.
}
\references{
The file format was found at \url{http://www.martinreddy.net/gfx/3d/OBJ.spec}
on November 11, 2012.  
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{scene3d}} saves a copy of a scene to an R variable; \code{\link{rglwidget}}, \code{\link{writeASY}},
\code{\link{writePLY}} and \code{\link{writeSTL}}
write the scene to a file in various other formats.
}
\examples{
filename <- tempfile(fileext = ".obj")
open3d()
shade3d( icosahedron3d() )
writeOBJ(filename)

# The motivation for writing readObj() was to read a shape
# file of Comet 67P/Churyumov-Gerasimenko, from the ESA.
# The file no longer appears to be online, but may still be
# available on archive.org.  Here was the original URL:
#   cometurl <- "http://sci.esa.int/science-e/www/object/doc.cfm?fobjectid=54726"
# This code would read and display it:
#   open3d()
#   shade3d(readOBJ(url(cometurl), 
#                   material = list(col = "gray")))


# Textures are used in a realistic hand image available from
# https://free3d.com/3d-model/freerealsichand-85561.html
# Thanks to Monte Shaffer for pointing this out.
# Decompress the files into the current directory, convert
# hand_mapNew.jpg to hand_mapNew.png, then use
\dontrun{
open3d()
shade3d(readOBJ("hand.OBJ", material = list(color = "white", 
shininess = 1, texture = "hand_mapNew.png")))
}
}
\keyword{ graphics }

\name{rglFonts}
\alias{rglFonts}
\title{
Specify FreeType fonts
}
\description{
Specify FreeType fonts for use in \pkg{rgl} graphics.
}
\usage{
rglFonts(...)
}
\arguments{
  \item{ ... }{Device dependent font definitions for use with FreeType.}  
}
\details{
FreeType fonts are specified using the \code{rglFonts} function.  This function
takes a vector of four filenames of TrueType font files which
will be used for the four styles regular, bold, italic and bold italic.  The
vector is passed with a name to be used as the family name, e.g.
\code{rglFonts(sans = c("/path/to/FreeSans.ttf", ...))}.  In order to 
limit the file size, the \pkg{rgl} package ships with just 3 font files, for 
regular versions of the \code{serif}, \code{sans} and \code{mono} families.  
Additional free font files were available in the past from the
Amaya project, though currently the
\code{\link{rglExtrafonts}} function provides an easier way
to register new fonts.

On Windows the system fonts are acceptable and are used
when \code{useFreeType = FALSE} (the current default in
\code{\link{r3dDefaults}}).  Mappings to \code{family} names
are controlled by the \code{grDevices::windowsFonts()}
function.

Full pathnames should normally be used to specify font files.  If relative
paths are used, they are interpreted differently by platform.  Currently
Windows fonts are looked for in the Windows fonts folder, while other
platforms use the current working directory.

If FreeType fonts are not used, then bitmapped fonts will be used instead.
On Windows these will be based on the fonts specified using the 
#ifdef windows
\code{\link{windowsFonts}} 
#endif
#ifndef windows
\code{windowsFonts}
#endif
function, and are resizable.
Other platforms will use the default bitmapped font which is not
resizable.  

Bitmapped fonts have a limited number of characters supported; if any
unsupported characters are used, an error will be thrown.
}
\value{
the current set of font definitions.
}
\seealso{\code{\link{text3d}}}
\examples{
\dontrun{
# These FreeType fonts are available from the Amaya project, and are not shipped
# with rgl.  You would normally install them to the rgl/fonts directory
# and use fully qualified pathnames, e.g. 
# system.file("fonts/FreeSerif.ttf", package = "rgl")

rglFonts(serif = c("FreeSerif.ttf", "FreeSerifBold.ttf", "FreeSerifItalic.ttf",
                 "FreeSerifBoldItalic.ttf"),
         sans  = c("FreeSans.ttf", "FreeSansBold.ttf", "FreeSansOblique.ttf",
                 "FreeSansBoldOblique.ttf"),
         mono  = c("FreeMono.ttf", "FreeMonoBold.ttf", "FreeMonoOblique.ttf",
                 "FreeMonoBoldOblique.ttf"),
         symbol= c("ESSTIX10.TTF", "ESSTIX12.TTF", "ESSTIX9_.TTF", 
                 "ESSTIX11.TTF"))
} 
}

\name{plotmath3d}
\alias{plotmath3d}
\title{
Draw text using base graphics math plotting
}
\description{
To plot mathematical text, this function uses base graphics
functions to plot it to a \file{.png} file, then uses that
file as a texture in a sprite.
}
\usage{
plotmath3d(x, y = NULL, z = NULL, text, cex = par("cex"),
           adj = 0.5, pos = NULL, offset = 0.5,
           fixedSize = TRUE, startsize = 480, initCex = 5, 
           margin = "", floating = FALSE, tag = "", ...)
}
\arguments{
  \item{x, y, z}{coordinates.  Any reasonable way of defining the
    coordinates is acceptable.  See the function \code{\link[grDevices]{xyz.coords}}
    for details.}  
  \item{text}{
A character vector or expression.  See
\code{\link[grDevices]{plotmath}} for how expressions are
interpreted.
}
  \item{cex}{
Character size expansion.
}
  \item{adj}{ one value specifying the horizontal adjustment, or two, 
              specifying horizontal and vertical adjustment respectively, or three, for depth as well. }
  \item{pos, offset}{ alternate way to specify \code{adj}; see \code{\link{text3d}}}
  \item{fixedSize}{
Should the resulting sprite behave like the default
ones, and resize with the scene, or like text, and
stay at a fixed size?
}
  \item{startsize, initCex}{
These parameters are unlikely to be needed by users.
\code{startsize} is an over-estimate of the size (in pixels) of the 
largest expression.  Increase this if large expressions
are cut off.  \code{initCex} is the size of text used
to form the bitmap.  Increase this if letters look too blurry
at the desired size.
}
  \item{margin, floating, tag}{
\code{\link{material3d}} properties.
  }
  \item{\dots}{
Additional arguments to pass to  
\code{\link[graphics]{text}} when drawing the text.
}
}
\note{
The \code{\link{text3d}} function passes calls to this
function if its \code{usePlotmath} argument is \code{TRUE}.
This is the default value if its
\code{texts} argument looks like an expression.
}
\value{
Called for the side effect of displaying the sprites.
The shape ID of the displayed object is returned.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{text3d}}
}
\examples{
open3d()
plotmath3d(1:3, 1:3, 1:3, expression(x[1] == 1, x[2] == 2, x[3] == 3))
# This lets the text resize with the plot
text3d(4, 4, 4, "resizeable text", usePlotmath = TRUE, fixedSize = FALSE)
}

\name{viewpoint}
\alias{view3d}
\title{Set up viewpoint}
\description{
  Set the viewpoint orientation.
}
\usage{
view3d( theta = 0, phi = 15, fov = 60, zoom = 1, 
        scale = par3d("scale"), interactive = TRUE, userMatrix, 
        type = c("userviewpoint", "modelviewpoint") )
}
\arguments{
  \item{theta, phi}{polar coordinates in degrees. \code{theta} rotates round the vertical axis. \code{phi} rotates round the horizontal axis.}
  \item{fov}{field-of-view angle in degrees}
  \item{zoom}{zoom factor}
  \item{scale}{real length 3 vector specifying the rescaling to apply to each axis}
  \item{interactive}{logical, specifying if interactive navigation is allowed}
  \item{userMatrix}{4x4 matrix specifying user point of view}
  \item{type}{which viewpoint to set?}
}
\details{
The data model can be rotated using the polar coordinates \code{theta}
and \code{phi}.  Alternatively, it can be set in a completely general way using the 4x4 matrix
\code{userMatrix}.  If \code{userMatrix} is specified, \code{theta} and \code{phi} are ignored.

The pointing device of your graphics user-interface can also be used to 
set the viewpoint interactively. With the pointing device the buttons are by default set as follows:

\describe{
\item{left}{adjust viewpoint position}
\item{middle}{adjust field of view angle}
\item{right or wheel}{adjust zoom factor}
}

The user's view can be set with \code{fov} and \code{zoom}.

If the \code{fov} angle is set to 0, a parallel or orthogonal
projection is used. Small non-zero values (e.g. 0.01 or less, but not
0.0) are likely to lead to rendering errors due to OpenGL limitations.

Prior to version 0.94, all of these characteristics were stored in 
one viewpoint object.  With that release the characteristics are split
into those that affect the projection (the user viewpoint) and those that
affect the model (the model viewpoint).  By default, this function
sets both, but the \code{type} argument can be used to limit the effect.

}
\seealso{\code{\link{par3d}}}
\examples{
\dontrun{
# animated round trip tour for 10 seconds

open3d()
shade3d(oh3d(), color = "red")

start <- proc.time()[3]
while ((i <- 36*(proc.time()[3] - start)) < 360) {
  view3d(i, i/4); 
}
}
}
\keyword{dynamic}

\name{axes3d}
\alias{axes3d}
\alias{axis3d}
\alias{mtext3d}
\alias{title3d}
\alias{box3d}
\title{ Draw boxes, axes and other text outside the data }
\description{
These functions draw axes, boxes and text outside the range of the data.
\code{axes3d}, \code{box3d} and \code{title3d} are the higher level functions; 
normally the others need not be called directly by users.
}
\usage{
axes3d(edges = "bbox", labels = TRUE, tick = TRUE, nticks = 5, 
	box = FALSE, expand = 1.03, ...)
box3d(...) 
title3d(main = NULL, sub = NULL, xlab = NULL, ylab = NULL, 
	zlab = NULL, line = NA, level = NA, floating = NULL, ...) 
axis3d(edge, at = NULL, labels = TRUE, tick = TRUE, line = 0, 
	pos = NULL, nticks = 5, ...) 
mtext3d(text, edge, at = NULL, line = 0, level = 0, 
        floating = FALSE, pos = NA, ...) 
}
\arguments{
  \item{edges}{ a code to describe which edge(s) of the box to use; see Details below }
  \item{labels}{ whether to label the axes, or (for \code{axis3d}) the
  labels to use}
  \item{tick}{ whether to use tick marks }
  \item{nticks}{ suggested number of ticks }
  \item{box}{ draw the full box if \code{"bbox"} axes are used }
  \item{expand}{ how much to expand the box around the data }
  \item{main}{ the main title for the plot }
  \item{sub}{ the subtitle for the plot }
  \item{xlab, ylab, zlab}{ the axis labels for the plot }
  \item{line, level}{ the "line" of the plot margin to draw the label on, and "level" above or below it  }
  \item{floating}{ which mode of axis labels? One of 
  \code{TRUE}, \code{FALSE} or \code{NA}. (\code{NULL}
  may also be used in \code{title3d} calls).
  See Details for how these are handled.}
  \item{edge, pos}{ the position at which to draw the axis or text }
  \item{text}{ the text to draw }
  \item{at}{ the value of a coordinate at which to draw the axis or labels. }
  \item{\dots}{ additional parameters which are passed to \code{\link{bbox3d}} or \code{\link{material3d}} }
}

\details{
The rectangular prism holding the 3D plot has 12 edges.  They are identified
using 3 character strings.  The first character (`x', `y', or `z') selects 
the direction of the axis.  The next two characters are each `-' or `+',
selecting the lower or upper end of one of the other coordinates.  If only
one or two characters are given, the remaining characters normally default to `-' (but with \code{mtext3d(..., floating = TRUE)} the default is `+'; see below).  
For example \code{edge = 'x+'} draws an x-axis at the high level of y and the
low level of z.

By default, \code{axes3d} uses the \code{\link{bbox3d}} function to draw the axes.  
The labels will move so that they do not obscure the data.  Alternatively,
a vector of arguments as described above may be used, in which case
fixed axes are drawn using \code{axis3d}.

As of \pkg{rgl} version 0.106.21, axis drawing has changed
significantly.  Text drawn in the margins 
will
adapt to the margins (see \code{\link{bbox3d}}).  The
\code{edge} and \code{floating} parameters will be recorded
in the \code{margin} and \code{floating} material
properties for the object.

If \code{floating = FALSE}, they will be drawn on the
specified edge.  

If \code{floating = TRUE}, they will
move as the axis labels move when the scene is rotated.
The signs on the edge specification are interpreted as
agreeing with the axis ticks `+' or disagreeing `-'.
For example, \code{"x++"} will draw text on the x axis
in the same edge as the ticks, while \code{"x--"} will draw
on the opposite edge.

The final possible value for \code{floating} in \code{mtext3d} is
\code{NA}, which reproduces legacy \pkg{rgl}
behaviour.  In
this case the labels are not tied to the bounding box,
so they should be drawn last, or they could appear inside
the box, overlapping the data.

In \code{title3d} \code{floating = NULL} (the default) 
indicates the main title 
and subtitle will be fixed while the axis labels will be
floating.  The default locations for title and subtitle
are \code{line = 2} and 
\code{level = 2} on edges \code{"x++"} and \code{"x--"}
respectively.  The axis labels float at \code{line = 4} and \code{level = 1} on the same edge as the ticks.

The \code{at} parameter in \code{axis3d} is the location
of the ticks, defaulting to \code{\link{pretty}} locations.
In \code{mtext3d} the \code{at} parameter is the location
on the specified axis
at which to draw the text, defaulting to the middle of the
bounding box.

The \code{line} parameter is the line counting out
from the box in the same direction as the axis ticks,
and \code{level} is the line out in the orthogonal
direction.  The ticks run from \code{line = 0} to \code{line = 1}, and the
tick labels are drawn at \code{line = 2}.  Both are drawn at level 0.  

The \code{pos} parameter is only supported in legacy mode.  If it is a numeric vector of length 3, \code{edge} determines
the direction of the axis and the tick marks, and the values of the
other two coordinates in \code{pos} determine the position.  
The \code{level} parameter is ignored in legacy mode.

For \code{mtext3d} in \code{floating = TRUE} or \code{floating = FALSE} mode,
there are 3 special values for the \code{at} parameter: it may
be \code{-Inf}, \code{NA} or \code{+Inf}, referring 
to the bottom, middle or top of the given axis respectively.
}
\note{\code{mtext3d} is a wrapper for \code{\link{text3d}}
that sets the \code{margin} and \code{floating} material
properties.  In fact, these properties can be set for many
kinds of objects (most kinds where it would make sense), 
with the effect that the object will be drawn
in the margin, with \code{x} coordinate corresponding to
\code{at}, \code{y} corresponding to \code{line}, and \code{z}
corresponding to \code{level}.

}
\value{
These functions are called for their side effects.  They return the object IDs of
objects added to the scene.
}
\author{ Duncan Murdoch }

\seealso{Classic graphics functions \code{\link{axis}}, \code{\link{box}},
\code{\link{title}}, \code{\link{mtext}} are related.
See RGL functions \code{\link{bbox3d}} for drawing the 
box around the plot, and
\code{\link{setAxisCallbacks}} for customized axes.}

\examples{
  open3d()
  points3d(rnorm(10), rnorm(10), rnorm(10))

  # First add standard axes
  axes3d()  

  # and one in the middle (the NA will be ignored, a number would 
  # do as well)
  axis3d('x', pos = c(NA, 0, 0))

  # add titles
  title3d('main', 'sub', 'xlab', 'ylab', 'zlab')
  
  # Use a log scale for z
    
  open3d()
  
  x <- rnorm(10)
  y <- rnorm(10)
  z <- exp(rnorm(10, mean = 3, sd = 2))
  
  logz <- log10(z)
  zticks <- axisTicks(range(logz), log = TRUE)
  zat <- log10(zticks)
  
  plot3d(x, y, logz, zlab = "z")
  axes3d(zat = zat, zlab = zticks, box = TRUE)
  
}


\keyword{dynamic}%-- one or more ...

\name{rgl.user2window}
\alias{rgl.user2window}
\alias{rgl.window2user}
\alias{rgl.projection}
\title{ Convert between RGL user and window coordinates }
\description{
This function converts from 3-dimensional user coordinates
to 3-dimensional window coordinates.
}
\usage{
rgl.user2window(x, y = NULL, z = NULL, projection = rgl.projection())
rgl.window2user(x, y = NULL, z = 0, projection = rgl.projection())
rgl.projection(dev = cur3d(), subscene = currentSubscene3d(dev))
}
\arguments{
  \item{x, y, z}{Input coordinates.  Any reasonable way of defining the
    coordinates is acceptable.  See the function \code{\link[grDevices]{xyz.coords}}
    for details.}
  \item{projection}{The RGL projection to use }
  \item{dev, subscene}{The RGL device and subscene to work with }
}
\details{
These functions convert between user coordinates and window coordinates.

Window coordinates run from 0 to 1 in X, Y, and Z.  X runs from 0 on the
left to 1 on the right; Y runs from 0 at the bottom to 1 at the top;
Z runs from 0 foremost to 1 in the background.  RGL does not currently
display vertices plotted outside of this range, but in normal circumstances will automatically resize the
display to show them.  In the example below this has been suppressed.

}
\value{
The coordinate conversion functions produce a matrix with columns corresponding 
to the X, Y, and Z coordinates.

\code{rgl.projection()} returns a list containing the following components:
\item{model}{the modelview matrix}
\item{projection}{the projection matrix}
\item{viewport}{the viewport vector}

See \code{\link{par3d}} for more details.
}
\author{ Ming Chen / Duncan Murdoch}
\seealso{\code{\link{select3d}} }
\examples{
open3d()
points3d(rnorm(100), rnorm(100), rnorm(100))
if (interactive() || !.Platform$OS == "unix") {
# Calculate a square in the middle of the display and plot it
square <- rgl.window2user(c(0.25, 0.25, 0.75, 0.75, 0.25), 
                          c(0.25, 0.75, 0.75, 0.25, 0.25), 0.5)
par3d(ignoreExtent = TRUE)
lines3d(square)
par3d(ignoreExtent = FALSE)
}
}
\keyword{ dynamic }

\name{persp3d.deldir}
\alias{persp3d.deldir}
\alias{plot3d.deldir}
\alias{as.mesh3d.deldir}
\title{
Plot a Delaunay triangulation
}
\description{
The \code{\link[deldir]{deldir}()} function in the \pkg{deldir} package computes a Delaunay triangulation of a set
of points.  These functions display it as a surface.
}
\usage{
\method{plot3d}{deldir}(x, ...)
\method{persp3d}{deldir}(x, ..., add = FALSE)
\method{as.mesh3d}{deldir}(x, col = "gray", coords = c("x", "y", "z"), 
    smooth = TRUE, normals = NULL, texcoords = NULL, ...)
}
\arguments{
  \item{x}{
A \code{"deldir"} object, produced by the \code{\link[deldir]{deldir}()} function.  It must contain \code{z} values.
}
  \item{add}{
Whether to add surface to existing plot (\code{add = TRUE})
or create a new plot (\code{add = FALSE}, the default).
}
  \item{col}{
Colors to apply to each vertex in the triangulation.
Will be recycled as needed.
}
  \item{coords}{
See Details below.
}
  \item{smooth}{
Whether to average normals at vertices for
a smooth appearance.
}
  \item{normals}{
User-specified normals at each vertex.  Requires \code{smooth = FALSE}.
}
  \item{texcoords}{
Texture coordinates at each vertex.
}
  \item{...}{
See Details below.
}
}
\details{
These functions construct a \code{\link{mesh3d}} object
corresponding to the triangulation in \code{x}.  The
\code{plot3d} and \code{persp3d} methods plot it.

The \code{coords} parameter allows surfaces to be 
plotted over any coordinate plane.  It should be
a permutation of the column names \code{c("x", "y", "z")}
from the \code{"deldir"} object.  The first will be used
as the x coordinate, the second as the y coordinate, 
and the third as the z coordinate.

The \code{...} parameters in \code{plot3d.deldir} 
are passed to \code{persp3d.deldir}; in \code{persp3d.deldir} they are
passed to both \code{as.mesh3d.deldir} and \code{persp3d.mesh3d}; 
in \code{as.mesh3d.deldir} they are used as material parameters
in a \code{\link{tmesh3d}} call.
}
\examples{
x <- rnorm(200, sd = 5)
y <- rnorm(200, sd = 5)
r <- sqrt(x^2 + y^2)
z <- 10 * sin(r)/r
col <- cm.colors(20)[1 + round(19*(z - min(z))/diff(range(z)))]

save <- options(rgl.meshColorWarning = FALSE)

# This code is awkward:  to work with demo(rglExamples),
# we need auto-printing of the plots.  This means we
# have to repeat the test for deldir.

haveDeldir <- checkDeldir()
              
if (haveDeldir) {
  dxyz <- deldir::deldir(x, y, z = z, suppressMsge = TRUE)
  persp3d(dxyz, col = col)
}

if (haveDeldir) {
  open3d()
  # Do it without smoothing and with a different orientation.
  persp3d(dxyz, col = col, coords = c("z", "x", "y"), smooth = FALSE)
}

options(save)
}
\keyword{graphics}

\name{hover3d}
\alias{hover3d}
\title{
Display hover info in plot.
}
\description{
This adds text to identify points within a plot when the mouse is near 
them.
}
\usage{
hover3d(x, y = NULL, z = NULL, 
        labeller = NULL, 
        tolerance = 20, 
        persist = c("no", "one", "yes"), 
        labels = seq_along(x),
        adj = c(-0.2, 0.5),
        scene = scene3d(minimal = FALSE),
        applyToScene = TRUE,
        ...)
}
\arguments{
  \item{x, y, z}{
Coordinates of point to identify. Any reasonable way of defining the
coordinates is acceptable.  See the function \code{\link[grDevices]{xyz.coords}} for details. 
Alternatively, \code{x} may be the id of a single existing 
object, and its vertices will be used.
}
  \item{labeller}{
A function to display information about identified points. \code{NULL}
indicates the default function, described in Details.
}
  \item{tolerance}{
How close (in pixels) the mouse should be to a point to display the
information.
}
  \item{persist}{
Should the label persist?  If \code{"no"} (the default), it will be removed
when the mouse moves away.  If \code{"one"}, it will be removed when 
another point is closer to the mouse.  If \code{"yes"}, it will 
not be removed.
}
  \item{labels}{
If the default \code{labeller} is used, these labels will be displayed.
}
  \item{adj}{
If the default \code{labeller} is used, this adjustment will be passed
to \code{\link{text3d}} to display the labels.
}
  \item{scene, applyToScene}{
Arguments to pass to \code{\link{setUserCallbacks}}.
The \code{applyToDev} argument to that function is always
\code{TRUE}.
}
  \item{\dots}{
Additional arguments that will be passed to the labeller.
}
}
\details{
If specified, the \code{labeller} argument should specify a
function with arguments
compatible with \code{function(index, ...)}.  It will be called with
\code{index} being the index of the point that was selected.  It should
plot the label, and return the \pkg{rgl} ids of the objects that were 
plotted.

When \code{applyToScene} is \code{TRUE}, all labels
or labelling objects will be created and attached to the scene.  You may want to
delete them (using the ids returned in \code{idverts} and
\code{idtexts}) once \code{\link{rglwidget}} has been called,
as they serve no purpose in the current device.

Only one hover handler is supported per scene or device.
}
\value{
A \code{\link{lowlevel}} vector of ids is returned invisibly.
If \code{applyToScene} is \code{TRUE}, it will contain the
ids of the temporary objects created for Javascript.
It will also have these attributes:
\item{oldPar}{Values of \code{\link{par3d}} parameters that were changed.
Currently only \code{"mouseMode"}.}
\item{oldDev}{The value of \code{cur3d()} at the time of calling, so that
\code{oldPar} can be restored to the right device.}
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{identify3d}} and \code{\link{selectpoints3d}}
work in the \pkg{rgl} device and return information
about the selections. \code{\link{setUserCallbacks}}
is the underlying function used by \code{hover3d}.
}
\examples{
# Create a labeller to show the coordinates of the selected point.
labelLocation <- function(x, y = NULL, z = NULL) {
  xyz <- xyz.coords(x, y, z)
  function(sel, ...) {
    p <- with(xyz, matrix(c(x[sel], y[sel], z[sel]), ncol = 3))
    c(text3d(p, texts = sprintf("x:\%.2f", p[1]), 
                  adj = c(-0.2, -0.6), ...),
      text3d(p, texts = sprintf("y:\%.2f", p[2]),
                  adj = c(-0.2, 0.5), ...),
      text3d(p, texts = sprintf("z:\%.2f", p[3]),
                  adj = c(-0.2, 1.6), ...))
  }
}

xyz <- matrix(rnorm(30), ncol = 3)
open3d()
ids <- plot3d(xyz)
hover3d(xyz, labeller = labelLocation(xyz), col = "red", cex = 0.8)
# The same thing using the data id:
# hover3d(ids["data"], 
#         labeller = labelLocation(rgl.attrib(ids["data"], "vertices")), 
#         col = "red", cex = 0.8)
}

\name{persp3d.triSht}
\alias{persp3d.triSht}
\alias{plot3d.triSht}
\alias{as.mesh3d.triSht}
\title{
Plot an interp Delaunay triangulation
}
\description{
The \code{\link[interp]{tri.mesh}()} function in the
\pkg{interp} package computes a Delaunay triangulation of a set
of points.  These functions display it as a surface.
}
\usage{
\method{plot3d}{triSht}(x, z, ...)
\method{persp3d}{triSht}(x, z, ..., add = FALSE)
\method{as.mesh3d}{triSht}(x, z, col = "gray", coords = c("x", "y", "z"), 
    smooth = TRUE, normals = NULL, texcoords = NULL, ...)
}
\arguments{
  \item{x}{
A \code{"triSht"} object, produced by the \code{\link[interp]{tri.mesh}()} function in the \pkg{interp} package.  
}
  \item{z}{
z coordinate values corresponding to each of the nodes in \code{x}.  
  }
  \item{add}{
Whether to add surface to existing plot (\code{add = TRUE})
or create a new plot (\code{add = FALSE}, the default).
}
  \item{col}{
Colors to apply to each vertex in the triangulation.
Will be recycled as needed.
}
  \item{coords}{
See Details below.
}
  \item{smooth}{
Whether to average normals at vertices for
a smooth appearance.
}
  \item{normals}{
User-specified normals at each vertex.  Requires \code{smooth = FALSE}.
}
  \item{texcoords}{
Texture coordinates at each vertex.
}
  \item{...}{
See Details below.
}
}
\details{
These functions construct a \code{\link{mesh3d}} object
corresponding to the triangulation in \code{x}.  The
\code{plot3d} and \code{persp3d} methods plot it.

The \code{coords} parameter allows surfaces to be 
plotted over any coordinate plane.  It should be
a permutation of the column names \code{c("x", "y", "z")}.
The first will be used
as the x coordinate, the second as the y coordinate, 
and the third as the z coordinate.

The \code{...} parameters in \code{plot3d.triSht} and \code{plot3d.tri} 
are passed to \code{persp3d}; in \code{persp3d.triSht}
and \code{persp3d.tri} they are
passed to both \code{as.mesh3d} and \code{persp3d.mesh3d}; 
in \code{as.mesh3d.triSht} and \code{as.mesh3d.tri} they are used as material parameters
in a \code{\link{tmesh3d}} call.

\code{"tri"} objects may contain constraints.  These appear
internally as extra nodes, representing either the inside
or outside of boundaries on the region being triangulated.
Each of these nodes should also have a \code{z} value, but 
triangles corresponding entirely to constraint nodes will not
be drawn.  In this way complex, non-convex regions can
be triangulated.  See the second example below.
}
\note{
If there are duplicate points, the \code{tri.mesh()} functions
will optionally delete some of them.  If you choose this option,
the \code{z} values must correspond to the nodes \emph{after}
deletion, not before.
}
\examples{
x <- rnorm(200, sd = 5)
y <- rnorm(200, sd = 5)
r <- sqrt(x^2 + y^2)
z <- 10 * sin(r)/r
col <- cm.colors(20)[1 + round(19*(z - min(z))/diff(range(z)))]
save <- NULL
if ((haveinterp <- requireNamespace("interp", quietly = TRUE))) {
  save <- options(rgl.meshColorWarning = FALSE)
  dxy <- interp::tri.mesh(x, y)
  open3d()
  persp3d(dxy, z, col = col, meshColor = "vertices")
}
if (haveinterp) {
  open3d()
  # Do it without smoothing and with a different orientation.
  persp3d(dxy, z, col = col, coords = c("z", "x", "y"), smooth = FALSE)
}

options(save)
}
\keyword{graphics}

\name{setAxisCallbacks}
\alias{setAxisCallbacks}
\title{
User-defined axis labelling callbacks.
}
\description{
This function sets user callbacks to construct axes in R or
\code{\link{rglwidget}} displays.
}
\usage{
setAxisCallbacks(axes, fns, 
                 javascript = NULL, 
                 subscene = scene$rootSubscene$id, 
                 scene = scene3d(minimal = FALSE), 
                 applyToScene = TRUE, 
                 applyToDev = missing(scene))
}
\arguments{
  \item{axes}{
Which axes?  Specify as number in \code{1:3} or letter in
\code{c("x", "y", "z")}.
}
  \item{fns}{
Function or list of functions or character vector giving
names of functions.
}
  \item{javascript}{
Optional block of Javascript code to be
included (at the global level).
}
  \item{subscene}{
Which subscene do these callbacks apply to?
}
  \item{scene}{
Which scene?
}
  \item{applyToScene}{
Should these changes apply to the scene object?
}
  \item{applyToDev}{
Should these changes apply to the current device?
  }
}
\details{
If \code{applyToScene} is \code{TRUE}, this function adds Javascript 
callbacks to the \code{scene} object.  
If \code{applyToDev} is \code{TRUE}, it adds R
callbacks to the current RGL device.

For Javascript, 
the callbacks are specified as strings; these will be
evaluated within the browser in the global context to define the functions, 
which will then be called with the Javascript
\code{this} object set to the current
\code{rglwidgetClass} object.

For R, they may be strings or R functions.

Both options may be \code{TRUE}, in which case the
callbacks must be specified as strings which are 
both valid Javascript and valid R.  The usual way to
do this is to give just a function name, with the
function defined elsewhere, as in the Example below.

The functions should have a header of the form 
\code{function(margin)}.  The \code{margin} argument
will be a string like \code{"x++"} indicating which margin
would be chosen by R.  If RGL would not choose to draw any
axis annotations (which happens with \code{\link{rglwidget}}, though
not currently in R itself), only the letter will be passed,
e.g. \code{"x"}.

}
\value{
Invisibly returns an \code{rglScene} object.  This
object will record the changes if \code{applyToScene}
is \code{TRUE}.

If \code{applyToDev} is \code{TRUE}, it will also 
have the side effect of attempting to install the
callbacks.
}
\seealso{
\code{\link{setUserCallbacks}} for mouse callbacks.
}
\author{
Duncan Murdoch
}
\examples{
# Draw arrows instead of tick marks on axes

arrowAxis <- local({
  ids <- c(NA, NA, NA)
  bbox <- c(NA, NA, NA, NA, NA, NA)
  function(margin) {
    dim <- if (grepl("x", margin)) 1 else
           if (grepl("y", margin)) 2 else
           3
    inds <- 2*dim + (-1):0
    range <- par3d("bbox")[inds]
    if (!identical(bbox[inds], range)) {
      if (!is.na(ids[dim]))
        pop3d(id = ids[dim])
       
      bbox[inds] <<- range 
      center <- mean(range)
      from <- mean(c(range[1], center))
      to <- mean(c(center, range[2]))
      # margin should agree with suggestion, so use "x++" etc.
      margin <- gsub("-", "+", margin)
      ids[dim] <- arrow3d(p0 = c(from, 1, 1), 
                         p1 = c(to, 1, 1),
                         n = 4,
                         type = "lines",
                         margin = margin,
                         floating = TRUE)
    }
  }
})

# Define the Javascript function with the same name to use in WebGL
# Since Javascript won't change the bounding box, this function
# doesn't need to do anything.
  
js <- "
window.arrowAxis = function(margin) {} ;
"

xyz <- matrix(rnorm(60), ncol = 3)
plot3d(xyz, xlab = "x", ylab = "y", zlab = "z")
setAxisCallbacks(1:3, "arrowAxis", javascript = js)
rglwidget()
}

\name{rgl-internal}
\title{Internal RGL functions and data}
\alias{rgl.bool}
\alias{rgl.numeric}
\alias{rgl.vertex}
\alias{rgl.nvertex}
\alias{rgl.color}
\alias{rgl.mcolor}
\alias{rgl.clamp}
\alias{rgl.attr}
\alias{rgl.enum}
\alias{rgl.enum.gl2ps}
\alias{rgl.enum.nodetype}
\alias{rgl.enum.pixfmt}
\alias{rgl.enum.polymode}
\alias{rgl.enum.textype}
\alias{rgl.enum.fogtype}
\alias{rgl.enum.primtype}
\alias{rgl.enum.texmagfilter}
\alias{rgl.enum.texminfilter}
\alias{rgl.selectstate}
\alias{rgl.setselectstate}
\alias{edgemap}
\alias{edgeindex}
\alias{cube3d.ib}
\alias{cube3d.vb}
\alias{oh3d.ib}
\alias{oh3d.vb}
\alias{dev3d}
\description{
Internal RGL functions
}
\usage{
rgl.bool(x)
rgl.numeric(x)
rgl.vertex(x, y = NULL, z = NULL)
rgl.nvertex(vertex)
rgl.color(color)
rgl.mcolor(colors)
rgl.clamp(value, low, high)
rgl.attr(vattr, nvertex) 
rgl.enum(name, ..., multi = FALSE)
rgl.enum.gl2ps(postscripttype) 
rgl.enum.nodetype(type)
rgl.enum.pixfmt(fmt)
rgl.enum.polymode(mode)
rgl.enum.textype(textype)
rgl.enum.fogtype(fogtype)
rgl.enum.primtype(primtype)
rgl.enum.texmagfilter(magfiltertype)
rgl.enum.texminfilter(minfiltertype)
rgl.selectstate(dev, subscene)
rgl.setselectstate(state, dev, subscene)
edgemap(size)
edgeindex(from, to, size, row, col)
}
\details{
  These are not to be called by the user.
}
\keyword{internal}

\name{cylinder3d}
\alias{cylinder3d}
\encoding{UTF-8}
\title{
Create cylindrical or "tube" plots
}
\description{
This function converts a description of a space curve into a \code{\link[=mesh3d]{"mesh3d"}}
object forming a cylindrical tube around the curve.
}
\usage{
cylinder3d(center, radius = 1, twist = 0, e1 = NULL, e2 = NULL, e3 = NULL, 
	sides = 8, section = NULL, closed = 0, 
	rotationMinimizing = is.null(e2) && is.null(e3),
	debug = FALSE, keepVars = FALSE, ...)
}
\arguments{
  \item{center}{An n by 3 matrix whose columns are the x, y and z coordinates of the space curve.}
  \item{radius}{The radius of the cross-section of the tube at each point in the center.}
  \item{twist}{The amount by which the polygon forming the tube is twisted at each point.}
  \item{e1, e2, e3}{The local coordinates to use at each point on the space curve.  These default to a rotation
minimizing frame or Frenet coordinates.}
  \item{sides}{The number of sides in the polygon cross section.}
  \item{section}{The polygon cross section as a two-column matrix, or \code{NULL}.}  
  \item{closed}{Whether to treat the first and last points of the space curve as identical,
  and close the curve, or put caps on the ends.  See the Details.}
  \item{rotationMinimizing}{Use a rotation minimizing local
frame if \code{TRUE}, or a Frenet or user-specified frame if \code{FALSE}.}
  \item{debug}{If \code{TRUE}, plot the local Frenet coordinates at each point.}
  \item{keepVars}{If \code{TRUE}, return the local variables in attribute \code{"vars"}.}
  \item{...}{Additional arguments to set as material properties.}
}
\details{

The number of points in the space curve is determined by the vector lengths in \code{center},
after using \code{\link{xyz.coords}} to convert it to a list.  The other arguments
\code{radius}, \code{twist}, \code{e1}, \code{e2}, and \code{e3} are extended to the same
length.

The \code{closed} argument controls how the ends of the cylinder are
handled. If \code{closed > 0}, it represents the number of points of
overlap in the coordinates.  \code{closed == TRUE} is the same as 
\code{closed = 1}. If \code{closed > 0} but the ends don't actually 
match, a warning will be given and results will be somewhat 
unpredictable.

Negative values of \code{closed} indicate that caps should be put on the
ends of the cylinder.  If \code{closed == -1}, a cap will be put on the
end corresponding to \code{center[1, ]}.  If \code{closed == -2}, caps
will be put on both ends.

If \code{section} is \code{NULL} (the default), a regular
\code{sides}-sided polygon is used, and \code{radius} measures the
distance from the center of the cylinder to each vertex.  If not
\code{NULL}, \code{sides} is ignored (and set internally to
\code{nrow(section)}), and \code{radius} is used as a multiplier to
the vertex coordinates.  \code{twist} specifies the rotation of the
polygon.  Both \code{radius} and \code{twist} may be vectors, with
values recycled to the number of rows in \code{center}, while
\code{sides} and \code{section} are the same at every point along the
curve.

The three optional arguments \code{e1}, \code{e2}, and \code{e3}
determine the local coordinate system used to create the vertices at
each point in \code{center}.  If missing, they are computed by simple
numerical approximations.  \code{e1} should be the tangent coordinate,
giving the direction of the curve at the point.  The cross-section of
the polygon will be orthogonal to \code{e1}. When \code{rotationMinimizing} is \code{TRUE}, \code{e2} and \code{e3} are chosen to give
a rotation minimizing frame (see Wang et al., 2008).  When it is
\code{FALSE},  \code{e2} defaults to an
approximation to the normal or curvature vector; it is used as the
image of the \code{y} axis of the polygon cross-section.  \code{e3}
defaults to an approximation to the binormal vector, to which the
\code{x} axis of the polygon maps.  The vectors are orthogonalized and
normalized at each point.
}
\value{
A \code{\link[=mesh3d]{"mesh3d"}} object holding the cylinder, possibly with
attribute \code{"vars"} containing the local environment of the function.
}
\author{
Duncan Murdoch
}
\references{
Wang, W., Jüttler, B., Zheng, D. and Liu, Y. (2008).  
Computation of rotation minimizing frames.  
ACM Transactions on Graphics, Vol. 27, No. 1, Article 2.
}
\examples{
# A trefoil knot
open3d()
theta <- seq(0, 2*pi, length.out = 25)
knot <- cylinder3d(
      center = cbind(
        sin(theta) + 2*sin(2*theta), 
        2*sin(3*theta), 
        cos(theta) - 2*cos(2*theta)),
      e1 = cbind(
        cos(theta) + 4*cos(2*theta), 
        6*cos(3*theta), 
        sin(theta) + 4*sin(2*theta)),
      radius = 0.8, 
      closed = TRUE,
      color = "green")
                     
shade3d(addNormals(subdivision3d(knot, depth = 2)))  
}

\keyword{ dynamic }

\name{play3d}
\alias{play3d}
\alias{movie3d}
\title{ Play animation of RGL scene }
\description{
\code{play3d} calls a function repeatedly, passing it the elapsed
time in seconds, and using the result of the function to
reset the viewpoint.  \code{movie3d} does the same, but
records each frame to a file to make a movie.
}
\usage{
play3d(f, duration = Inf, dev = cur3d(), ..., startTime = 0)
movie3d(f, duration, dev = cur3d(), ..., fps = 10, 
                    movie = "movie", frames = movie, dir = tempdir(), 
                    convert = NULL, clean = TRUE, verbose = TRUE,
                    top = !rgl.useNULL(), type = "gif", startTime = 0,
                    webshot = TRUE) 

}
\arguments{
  \item{f}{ A function returning a list that may be passed to \code{\link{par3d}} }
  \item{duration}{ The duration of the animation }    
  \item{dev}{ Which RGL device to select }
  \item{\dots}{ Additional parameters to pass to \code{f}. }
  \item{startTime}{ Initial time at which to start the animation }
  \item{fps}{ Number of frames per second }
  \item{movie}{ The base of the output filename, not including .gif }
  \item{frames}{ The base of the name for each frame }
  \item{dir}{ A directory in which to create temporary files for each frame of the movie }  
  \item{convert}{ How to convert to a GIF movie; see Details }
  \item{clean}{ If \code{convert} is \code{NULL} or \code{TRUE}, whether to delete the individual frames }
  \item{verbose}{ Whether to report the \code{convert} command and the output filename }
  \item{top}{ Whether to call \code{\link{rgl.bringtotop}} before each frame }
  \item{type}{ What type of movie to create.  See Details. }
  \item{webshot}{ Whether to use the \pkg{webshot2} package
  for snapshots of frames. See \code{\link{snapshot3d}}.}
}
\details{
The function \code{f} will be called in a loop with the first argument
being the \code{startTime} plus the time in seconds since the start 
(where the start is measured after all arguments have been evaluated).

\code{play3d} is likely to place a high load
on the CPU; if this is a problem, calls to \code{\link{Sys.sleep}}
should be made within the function to release time to other processes.

\code{play3d} will run for the specified \code{duration} (in seconds), but
can be interrupted by pressing \code{ESC} while the RGL window has the focus.

\code{movie3d} saves each frame to disk in a filename of the form \file{framesXXX.png}, where
XXX is the frame number, starting from 0.  
If \code{convert} is \code{NULL} (the default) and the
\pkg{\link[magick]{magick}} package is installed, it will be used
to convert the frames to a GIF movie (or other format if supported).  If 
\pkg{\link[magick]{magick}} is not installed or 
\code{convert} is \code{TRUE}, \code{movie3d} will attempt
to use the external \command{ImageMagick} program to convert the
frames to a movie.  
The newer \command{magick} executable is tried
first, then \command{convert} if that fails.
The \code{type} argument will be passed to \command{ImageMagick} to use as a file extension
to choose the file type.

Finally, \code{convert} can be a template for a command to
execute in the standard shell (wildcards are allowed). The template is
converted to a command using \cr
\code{\link{sprintf}(convert, fps, frames, movie, type, duration, dir)} \cr
For example, \code{convert = TRUE}
uses the template \code{"magick -delay 1x\%d \%s*.png \%s.\%s"}. All work
is done in the directory \code{dir}, so paths should not be needed in
the command.  (Note that \code{\link{sprintf}} does not require all
arguments to be used, and supports formats that use them in an
arbitrary order.)

The \code{top = TRUE} default is designed to work around an OpenGL limitation:
in some implementations, \code{\link{rgl.snapshot}} will fail if the window
is not topmost.

As of \pkg{rgl} version 0.94, the \code{dev} argument is not needed:  the
function \code{f} can specify its device, as \code{\link{spin3d}} does, 
for example.  However, if \code{dev} is specified, it will be selected
as the current device as each update is played.

As of \pkg{rgl} version 0.95.1476, \code{f} can 
include multiple values in a \code{"subscene"}
component, and \code{par3d()} will be called
for each of them.
}
\value{
\code{play3d} is called for the side effect of its repeated calls to \code{f}.
It returns \code{NULL} invisibly.

\code{movie3d} is also normally called for the side effect
of producing the output movie.  It invisibly returns

}
\author{ Duncan Murdoch, based on code by Michael Friendly }
\seealso{ \code{\link{spin3d}} and \code{\link{par3dinterp}} return functions suitable 
          to use as \code{f}. See \code{demo(flag)} for an example that modifies the
          scene in \code{f}.}
\examples{
open3d()
plot3d( cube3d(col = "green") )
M <- par3d("userMatrix")
if (!rgl.useNULL() && interactive())
  play3d( par3dinterp(times = (0:2)*0.5, userMatrix = list(M,
                                     rotate3d(M, pi/2, 1, 0, 0),
                                     rotate3d(M, pi/2, 0, 1, 0) ) ), 
        duration = 2 )
\dontrun{
movie3d( spin3d(), duration = 5 )
}

}
\keyword{ dplot }

\name{rglMouse}
\alias{rglMouse}
\title{
Generate HTML code to select mouse mode
}
\description{
This generates an HTML \code{select} element to 
choose among the mouse modes supported by \code{\link{rglwidget}}.
}
\usage{
rglMouse(sceneId, 
         choices = c("trackball", "selecting", 
                     "xAxis", "yAxis", "zAxis", 
                     "polar", "zoom", "fov", 
                     "none"), 
         labels = choices, 
         button = 1, 
         dev = cur3d(), 
         subscene = currentSubscene3d(dev), 
         default = par3d("mouseMode", dev = dev, subscene = subscene)[button + 1],
         stayActive = FALSE,
         height = 40,
         ...)
}
\arguments{
  \item{sceneId}{
Either an \code{\link{rglwidget}} or the \code{elementId}
from one of them.
}
  \item{choices}{
Which mouse modes to support?
}
  \item{labels}{
How to label each mouse mode.
}
  \item{button}{
Which mouse button is being controlled.
}
  \item{dev}{
The RGL device used for defaults.
}
  \item{subscene}{
Which subscene is being modified.
}
  \item{default}{
What is the default entry to show in the control.
}
  \item{stayActive}{
Whether a selection brush should stay active if the
mouse mode is changed.
}
  \item{height}{
The (relative) height of the item in the output display.
}
  \item{...}{
Additional arguments to pass to \code{htmltools::tags$select()}, e.g. \code{id} or \code{class}.
}
}
\details{
A result of an \code{\link{rglwidget}} call can be passed
as the \code{sceneId} argument.  This allows the widget
to be \dQuote{piped} into the \code{rglMouse} call.
The widget will appear first, the selector next in
a \code{\link[htmltools:tag]{tagList}}.

If the \code{sceneId} is a character string, it should be
the \code{elementId} of a separately constructed
\code{\link{rglwidget}} result.

Finally, the \code{sceneId} can be omitted.  In this case
the \code{rglMouse} result needs to be passed into an
\code{\link{rglwidget}} call as part of the
\code{controllers} argument.  This will place the selector before the
widget on the resulting display.

If the mouse mode is changed while brushing the scene, 
by default the brush will be removed (and so the selection
will be cleared too).  If this is not desired, set
\code{stayActive = TRUE}.
}
\value{
A browsable value to put in a web page.
}
\author{
Duncan Murdoch
}
\seealso{
The \HTMLVignette{WebGL}{}{User Interaction in WebGL} vignette
gives more details.
}
\examples{
if (interactive() || in_pkgdown_example()) {
  open3d()
  xyz <- matrix(rnorm(300), ncol = 3)
  id <- plot3d(xyz, col = "red", type = "s")["data"]
  par3d(mouseMode = "selecting")
  share <- rglShared(id)

# This puts the selector below the widget.
  rglwidget(shared = share, width = 300, height = 300) \%>\% rglMouse()
  
# This puts the selector above the widget.
  rglMouse() \%>\% rglwidget(shared = share, width = 300, height = 300, controllers = .) 
}
}

\name{abclines3d}
\alias{abclines3d}
\title{
Lines intersecting the bounding box
}
\description{
This adds mathematical lines to a scene.  Their intersection with the current
bounding box will be drawn.
}
\usage{
abclines3d(x, y = NULL, z = NULL, a, b = NULL, c = NULL, ...)
}
\arguments{
  \item{x, y, z}{
Coordinates of points through which each line passes.  
} 
  \item{a, b, c}{
Coordinates of the direction vectors for the lines.
}
  \item{...}{
Material properties.
}

}
\details{
Draws the segment of a line that intersects the current
bounding box of the scene using the parametrization
\eqn{ (x, y, z) + (a, b, c) * s } where \eqn{s} is a real number.

Any reasonable way of defining the
coordinates \code{x, y, z} and \code{a, b, c} is acceptable.  
See the function \code{\link[grDevices]{xyz.coords}}
for details.
}
\value{
A shape ID of the object is returned invisibly.
}

\seealso{
\code{\link{planes3d}} for mathematical planes.

\code{\link{segments3d}} draws sections of lines that
do not adapt to the bounding box.

}
\examples{
plot3d(rnorm(100), rnorm(100), rnorm(100))
abclines3d(0, 0, 0, a = diag(3), col = "gray")
}
\keyword{ dynamic }

\name{scene3d}
\alias{scene3d}
\alias{rglscene-class}
\alias{rglobject-class}
\alias{plot3d.rglscene}
\alias{plot3d.rglobject}
\alias{print.rglscene}
\alias{print.rglobject}
\title{
Saves the current scene to a variable, and displays such variables
}
\description{
This function saves a large part of the RGL state associated with the current
window to a variable.  
}
\usage{
scene3d(minimal = TRUE)
\S3method{plot3d}{rglscene}(x, add = FALSE, open3dParams = getr3dDefaults(), ...)
\S3method{plot3d}{rglobject}(x, ...)
\S3method{print}{rglscene}(x, ...)
\S3method{print}{rglobject}(x, ...)
}
\arguments{
\item{minimal}{Should attributes be skipped if they currently have
  no effect?  See Details.}
\item{x}{An object of class \code{"rglscene"}}
\item{add}{Whether to open a new window, or add to the existing one.}
\item{open3dParams}{Default parameters for \code{open3d}}
\item{...}{Additional parameters passed to \code{open3d}
by \code{plot3d(..., add = FALSE)}.  These override
\code{open3dParams}.
}
}
\details{
The components saved are:  the \code{\link{par3d}} settings, the \code{\link{material3d}}
settings, the \code{\link{bg3d}} settings, the lights and the objects in the scene.

In most cases, calling \code{\link{plot3d}} on that variable will
duplicate the scene.  (There are likely to be small differences, mostly internal, but
some aspects of the scene are not currently available.) If textures are used,
the name of the texture will be saved, rather than the contents of the 
texture file.

Other than saving the code
to recreate a scene, saving the result of \code{scene3d} to 
a file will allow it to be reproduced later most accurately.  In roughly decreasing order of fidelity,
\code{\link{writeWebGL}} (now deprecated), \code{\link{writePLY}}, \code{\link{writeOBJ}} and \code{\link{writeSTL}}
write the scene to a file in formats readable by other software.

If \code{minimal = TRUE} (the default), then attributes of objects
will not be saved if they currently have no effect on
the display, thereby reducing the file size.  
Set \code{minimal = FALSE} if the scene is intended
to be used in a context where the appearance could be changed.  
Currently this only affects the inclusion of normals; with 
\code{minimal = TRUE} they are
omitted for objects when the material is not lit.
}
\value{
The \code{scene3d} function returns an object of class
\code{"rglscene"}.  This is a list with some or all of the components:
\item{material}{The results returned from a \code{\link{material3d}} call.}
\item{rootSubscene}{A list containing information about
the main ("root") subscene.  This may include:
\describe{
\item{id}{The scene id.}
\item{type}{"subscene"}
\item{par3d}{The \code{\link{par3d}} settings for the
subscene.}
\item{embeddings}{The \code{\link{subsceneInfo}()$embeddings} for the main subscene.}
\item{objects}{The ids for objects in the subscene.}
\item{subscenes}{A recursive list of child subscenes.}}}
\item{objects}{A list containing the RGL lights, background and objects in the scene.}  

The objects in the \code{objects} component are of class
\code{"rglobject"}.  They are lists containing some or all of the
components
\item{id}{The RGL identifier of the object in the original scene.}
\item{type}{A character variable identifying the type of object.}
\item{material}{Components of the material that differ from the scene material.}
\item{vertices, normals, etc.}{Any of the attributes of the object retrievable by \code{\link{rgl.attrib}}.}
\item{ignoreExtent}{A logical value indicating whether this object contributes to the bounding box.
Currently this may differ from the object in the original scene.}
\item{objects}{Sprites may contain other objects; they will be stored here as a list of  \code{"rglobject"}s.}

Lights in the scene are stored similarly, mixed into the \code{objects} list.

The \code{plot3d} methods invisibly return a vector of RGL object ids
that were plotted.  The \code{print} methods invisibly return the
object that was printed.

}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{rglwidget}}, \code{\link{writePLY}}, \code{\link{writeOBJ}} and \code{\link{writeSTL}}
write the scene to a file in various formats.
}
\examples{
open3d()
z <- 2 * volcano        # Exaggerate the relief
x <- 10 * (1:nrow(z))   # 10 meter spacing (S to N)
y <- 10 * (1:ncol(z))   # 10 meter spacing (E to W)
persp3d(x, y, z, col = "green3", aspect = "iso")

s <- scene3d()
# Make it bigger
s$par3d$windowRect <- 1.5*s$par3d$windowRect
# and draw it again
plot3d(s)
}
\keyword{ graphics }

\name{checkDeldir}
\alias{checkDeldir}
\title{
Check for a compatible version of \pkg{deldir}
}
\description{
Version 1.0-2 of \pkg{deldir} is not compatible with \pkg{rgl}.
This allows code to avoid trying to call it.
}
\usage{
checkDeldir(error = FALSE)
}
\arguments{
  \item{error}{
If \code{TRUE}, stop with an error.
}
}
\value{
Returns \code{TRUE} if \pkg{deldir} is available in a compatible
version.
}
\examples{
checkDeldir()
}

\name{tkspinControl}
\alias{tkspinControl}
\title{Create a spin control in a TCL/TK window}
\description{
    This function may be used to embed a spin control in a TCL/TK window.
}
\usage{
tkspinControl(base, dev = cur3d(), 
	continue=FALSE, speed=30, scale=100, ... )
}
\arguments{
  \item{base}{The TCL/TK frame in which to insert this control.  }
  \item{dev}{A vector of one or more RGL device numbers to control. }
  \item{continue}{Initial setting for continuous rotation checkbox. }
  \item{speed}{Initial setting for speed slider. }
  \item{scale}{Initial setting for scale slider. }
  \item{...}{Additional parameters to pass to \code{\link[tcltk:TkWidgets]{tkframe}}}
}
\author{ Ming Chen and Duncan Murdoch }
\seealso{\code{\link{spin3d}}}
\examples{
if (interactive() && !in_pkgdown_example()) { 
    library(tcltk)
    open3d()
    win1 <- cur3d()
    plot3d(rexp(100), rexp(100), rexp(100), size=3, col='green')
    
    open3d()
    win2 <- cur3d()
    plot3d(rt(100,2), rt(100,2), rt(100, 2), size=3, col='yellow')
    
    open3d()
    win3 <- cur3d()
    plot3d(rexp(100), rexp(100), rexp(100), size=3, col='red')
        
    open3d()
    win4 <- cur3d()
    plot3d(rbinom(100,10,0.5), rbinom(100,10,0.5), rbinom(100,10,0.5), size=3, col='cyan')
    
    base <- tktoplevel()
    tkwm.title(base, "Spinners")
    con1 <- tkspinControl(base, dev=c(win1,win2))
    con2 <- tkspinControl(base, dev=c(win3,win4))
    tkpack(con1, con2)
}
}

\name{selectpoints3d}
\alias{selectpoints3d}
\title{
Select points from a scene
}
\description{
This function uses the \code{\link{select3d}} function to allow the user to choose a
point or region in the scene, then reports on all the vertices in or near that selection.
}
\usage{
selectpoints3d(objects = ids3d()$id, value = TRUE, closest = TRUE, % $
               multiple = FALSE, ...)
}
\arguments{
  \item{objects}{
A vector of object id values to use for the search.
}
  \item{value}{
If \code{TRUE}, return the coordinates of the points; otherwise, return
their indices.
}
  \item{closest}{
If \code{TRUE}, return the points closest to the selection of no points are
exactly within it.
}
  \item{multiple}{
If \code{TRUE} or a function, do multiple selections.  See the Details below.
}
  \item{\dots}{
Other parameters to pass to \code{\link{select3d}}.
}
}
\details{
The \code{multiple} argument may be a logical value or a function.  If logical, 
it controls whether multiple selections will be performed.  If 
\code{multiple} is \code{FALSE}, a single selection will be performed;
it might contain multiple points.  If \code{TRUE}, multiple selections 
will occur and the results will be combined into a single matrix.  

If \code{multiple} is a function, it should take a single argument.
This function will be called with the argument set to a matrix
containing newly added rows to the value, i.e. 
it will contain coordinates of the newly selected points (if 
\code{value = TRUE}), or the indices of the points (if \code{value =
FALSE}).  It should return a logical value, \code{TRUE} to indicate
that selection should continue, \code{FALSE} to indicate that it
should stop.

In either case, if multiple selections are being performed, the \code{ESC} key will 
stop the process.
}
\value{
If \code{value} is \code{TRUE}, a 3-column matrix giving the coordinates of the 
selected points.  All rows in the matrix will be unique even if multiple vertices
have the same coordinates.

If \code{value} is \code{FALSE}, a 2-column matrix containing columns:
\item{id}{The object id containing the point.}
\item{index}{The index of the point within \code{\link{rgl.attrib}(id, "vertices")}.
If multiple points have the same coordinates, all indices will be returned.}

}
\note{This function selects points, not areas.  For example,
if the selection region is in the interior of the triangle, that
will count as a miss for all of the triangle's vertices.}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{select3d}} to return a selection function.
}
\examples{
xyz <- cbind(rnorm(20), rnorm(20), rnorm(20))
ids <- plot3d( xyz )

if (interactive() && !in_pkgdown_example()) {
  # Click near a point to select it and put a sphere there.
  # Press ESC to quit...

  # This version returns coordinates
  selectpoints3d(ids["data"], 
     multiple = function(x) {
        spheres3d(x, color = "red", alpha = 0.3, radius = 0.2)
        TRUE
     })

  # This one returns indices
  selectpoints3d(ids["data"], value = FALSE,
     multiple = function(ids) {
        spheres3d(xyz[ids[, "index"], , drop = FALSE], color = "blue", 
                  alpha = 0.3, radius = 0.2)
        TRUE
     })
}
}
\keyword{ graphics }

\name{writeWebGL}
\alias{writeWebGL}
\title{
Write scene to HTML.
}
\description{
Obsolete function to write the current scene to a collection of files that contain WebGL code to reproduce
it in a browser.  Please use \code{\link{rglwidget}} instead.
}
\usage{
writeWebGL(dir = "webGL", filename = file.path(dir, "index.html"), 
           template = system.file(file.path("WebGL", "template.html"), package = "rgl"),
           prefix = "",
           snapshot = TRUE, commonParts = TRUE, reuse = NULL, 
           font = "Arial", width, height)
}
\arguments{
  \item{dir}{
  Where to write the files.
}
  \item{filename}{
  The filename to use for the main file.  
}
  \item{template}{
The template web page to which to write the Javascript for the scene.  See Details below.
}
  \item{prefix}{
An optional prefix to use on global identifiers in the scene; use different prefixes
for different scenes displayed on the same web page.  If not blank, it should be a legal identifier
in Javascript and HTML.
}
  \item{snapshot}{
  Whether to include a snapshot of the scene, to be displayed in browsers that don't support WebGL, or
a specification of the snapshot to use.  See
details below.
}
  \item{commonParts}{
  Whether to include parts that would be common to several
figures on the same page.  Currently this includes
a reference to and copy of the \file{CanvasMatrix.js}
file in the output.
}
  \item{reuse}{
  When writing several figures on the same page, set this
to a dataframe containing values to reuse.  See the Value section below.
  }
  \item{font}{
  The font to use for text.
}
  \item{width, height}{
  The (optional) width and height in pixels of the image to display.  If omitted,
the \code{par3d("windowRect")} dimensions will be used.
}
}
\details{
This obsolete function writes out a web page containing Javascript that reconstructs the scene in WebGL.  It will
be formally deprecated in an upcoming release; you should use
\code{\link{rglwidget}} instead in any new code, and start
migrating old code there.

The remaining documentation has been removed to discourage 
use of this function.
}
\value{
The \code{filename} is returned.
}
\author{
Duncan Murdoch.
}
\note{
If \code{commonParts} is \code{TRUE}, the output includes a binary copy of
the CanvasMatrix Javascript library.  
Its source (including 
the copyright notice and license for free use) is included in the file named by
\code{system.file("htmlwidgets/lib/CanvasMatrix.src.js", package = "rgl")}.  
}
\seealso{
\code{\link{rglwidget}} should be used instead of \code{writeWebGL}.  Other functions which are related:
\code{\link{scene3d}} saves a copy of a scene to an R variable; \code{\link{writeASY}},
\code{\link{writePLY}}, \code{\link{writeOBJ}} and \code{\link{writeSTL}}
write the scene to a file in various other formats.
}

\name{gltfTypes}
\alias{gltfTypes}
\docType{data}
\title{
Names of glTF types.
}
\description{
The glTF specification packs data into buffers, labelling
the type of each part with an integer.  The first six values
in \code{gltfTypes} are the
integers used there, with \code{"int"} and \code{"double"}
added for completeness; those values are taken from
OpenGL header files.
}
\usage{gltfTypes}
\format{
\code{gltfTypes} is simply a named vector containing integer values.
}
\details{
These are used in the \code{\link{Buffer}} object.
}
\references{
\url{https://registry.khronos.org/glTF/specs/2.0/glTF-2.0.html#_accessor_componenttype}
}
\examples{
gltfTypes
}

\name{arrow3d}
\alias{arrow3d}
\title{
Draw an arrow
}
\description{
Draws various types of arrows in a scene.
}
\usage{
arrow3d(p0 = c(1, 1, 1), p1 = c(0, 0, 0), 
        barblen, s = 1/3, theta = pi/12, 
        type = c("extrusion", "lines", "flat", "rotation"), 
        n = 3, width = 1/3, thickness = 0.618 * width, 
        spriteOrigin = NULL, 
        plot = TRUE, ...)
}
\arguments{
  \item{p0}{
The base of the arrow.
}
  \item{p1}{
The head of the arrow.
}
  \item{barblen}{
The length of the barbs (in display coordinates).
Default given by \code{s}.
}
  \item{s}{
The length of the barbs as a fraction of line length.  Ignored if \code{barblen} is present.
}
  \item{theta}{
Opening angle of barbs
}
  \item{type}{
Type of arrow to draw.  Choose one from the list
of defaults.  Can be abbreviated.  See below.
}
  \item{n}{
Number of barbs.
}
  \item{width}{
Width of shaft as fraction of barb width.
}
  \item{thickness}{
Thickness of shaft as fraction of barb width.
}
  \item{spriteOrigin}{
If arrow is to be replicated as sprites, the origins relative
to which the sprites are drawn.
}
  \item{plot}{
If \code{TRUE} (the default), plot the object;
otherwise return the computed data that
would be used to plot it.
}
  \item{\dots}{
Material properties passed to \code{\link{polygon3d}}, \code{\link{shade3d}} or \code{\link{segments3d}}.
}
}
\details{
Four types of arrows can be drawn.  The shapes
of all of them are affected by \code{p0}, \code{p1}, \code{barblen},
\code{s}, \code{theta}, material properties
in \code{...}, and \code{spriteOrigin}.  Other parameters 
only affect some of the types, as shown.
\describe{
\item{\code{"extrusion"}}{(default) A 3-dimensional
flat arrow, drawn with \code{\link{shade3d}}.  Affected by \code{width}, \code{thickness} and
\code{smooth}.}
\item{\code{"lines"}}{Drawn with lines, similar to \code{\link{arrows}}, drawn with \code{\link{segments3d}}.  Affected by \code{n}.}
\item{\code{"flat"}}{A flat arrow, drawn with \code{\link{polygon3d}}.  Affected by \code{width} and \code{smooth}.}
\item{\code{"rotation"}}{A solid of rotation, 
drawn with \code{\link{shade3d}}.  Affected by 
\code{n} and \code{width}.}
}

Normally this function draws just one arrow from
\code{p0} to \code{p1}, but
if \code{spriteOrigin} is given (in any form
that \code{\link{xyz.coords}(spriteOrigin)} can
handle), arrows will be drawn for each point
specified, with \code{p0} and \code{p1} 
interpreted relative to those origins.  The
arrows will be drawn as 3D sprites which will
maintain their orientation as the scene is rotated, so this is a good way to indicate
particular locations of interest in the scene.
}
\value{
If \code{plot = TRUE} (the default), this is
called mainly for the side effect of drawing
the arrow; invisibly returns the id(s) of 
the objects drawn.

If \code{plot = FALSE}, the data that would be
used in the plot (not including material
properties) is returned.
}
\author{
Design based on \code{heplots::arrow3d}, which contains modifications by Michael Friendly
to a function posted by Barry Rowlingson to R-help on 1/10/2010.  Additions by Duncan Murdoch.
}
\examples{
xyz <- matrix(rnorm(300), ncol = 3)
plot3d(xyz)
arrow3d(xyz[1,], xyz[2,], type = "extrusion", col = "red")
arrow3d(xyz[3,], xyz[4,], type = "flat",      col = "blue")
arrow3d(xyz[5,], xyz[6,], type = "rotation",  col = "green")
arrow3d(xyz[7,], xyz[8,], type = "lines",     col = "black")
arrow3d(spriteOrigin = xyz[9:12,],            col = "purple")
}

\name{rglIds}
\alias{lowlevel}
\alias{highlevel}
\alias{rglId}
\alias{rglLowlevel}
\alias{rglHighlevel}
\alias{print.rglId}
\title{
RGL id values
}
\description{
All objects in an RGL scene have a
numerical id.  These ids are normally stored
in vectors of class \code{c("rglIds", "numeric")}, which
will also have class \code{"rglHighlevel"}
or \code{"rglLowlevel"} depending on 
whether a high level function like 
\code{\link{plot3d}} or \code{\link{persp3d}}, or a low level function created the objects.


}
\usage{
rglId(ids = integer())
lowlevel(ids = integer())
highlevel(ids = integer())
\method{print}{rglId}(x, 
      rglwidget = getOption("rgl.printRglwidget", FALSE),
      ...)
}
\arguments{
  \item{ids}{
A vector of object ids.
}
  \item{x}{
An \code{"rglId"} object to print.
}
  \item{rglwidget}{
Whether to create and print an RGL widget.
If false, nothing is printed.
}
  \item{...}{
Other arguments which will be passed to
\code{\link{rglwidget}} if it is used.
}
}
\details{
These functions and classes are intended to allow
RGL scenes to be automatically 
displayed in R Markdown documents.  See \code{\link{setupKnitr}} for details on enabling auto-printing.

Note that \emph{all} objects in the current
scene will be printed by default, not just
the ids in \code{x}.  (One reason for this
is that lights are also objects; printing
objects without lights would rarely make
sense.)
}
\value{
Objects of class \code{"rglId"}, \code{c("rglHighlevel", "rglId",
"numeric")} or \code{c("rglLowlevel", "rglId",
"numeric")} for \code{rglId}, \code{lowlevel} or \code{highlevel} 
respectively.  
}
\author{
Duncan Murdoch
}
\examples{
x <- matrix(rnorm(30), ncol = 3, dimnames = list(NULL, c("x", "y", "z")))
p <- plot3d(x, type = "s")
str(p)
if (interactive() || in_pkgdown_example())
  print(p, rglwidget = TRUE)
}

\name{all.equal.mesh3d}
\alias{all.equal.mesh3d}
\alias{compare_proxy.mesh3d}
\title{
Compare mesh3d objects in a meaningful way.
}
\description{
These functions allow comparison of mesh3d objects, ignoring
irrelevant differences.  

\code{compare_proxy.mesh3d} can function as a \code{compare_proxy} method for the
\pkg{waldo} package, by stripping out \code{NULL} components
and ordering other components alphabetically by name.

\code{all.equal.mesh3d} compares mesh3d objects by using
\code{compare_proxy.mesh3d} to standardize them, then
using the regular \code{\link{all.equal}} function to
compare them.
}
\usage{
\method{all.equal}{mesh3d}(target, current, ...)
compare_proxy.mesh3d(x, path = "x")
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{target, current}{
Two mesh3d objects to compare.  
}
  \item{x}{
A single mesh3d object to standardize.
}
  \item{path}{
The string to use in a \pkg{waldo} display of this object.
}
  \item{\dots}{
Additional parameters to pass to \code{\link{all.equal}}.
}
}
\value{
\code{all.equal.mesh3d} returns \code{TRUE}, or a character
vector describing (some of) the differences.

\code{compare_proxy.mesh3d} returns a list containing two
components:
\describe{
\item{object}{a copy of \code{x} with
relevant components in alphabetical order.}
\item{path}{a modification of the path label for \code{x}}
}
}
\note{
\pkg{waldo} is not an installation requirement for \pkg{rgl}
and \pkg{rgl} will never cause it to be loaded.
The \code{compare_proxy.mesh3d} function will only be 
registered as a method for \code{waldo::compare_proxy}
if you load \pkg{waldo} before \pkg{rgl}, as would normally 
happen during testing using \pkg{testthat}, or if you load
it before calling \code{\link{mesh3d}},
as might happen if you are doing manual tests.
}

\name{rgl.useNULL}
\alias{rgl.useNULL}
\alias{RGL_USE_NULL}
\title{
Report default use of null device
}
\description{
This function checks the \code{"rgl.useNULL"} option if present, or the
\env{RGL_USE_NULL} environment variable if it is not.  If the value is
\code{TRUE} or a string which matches \dQuote{yes} or \dQuote{true}
in a case-insensitive test, \code{TRUE} is returned.
}
\usage{
rgl.useNULL()
}
\note{
This function is checked by the initialization code when the \pkg{rgl} package
is loaded.  Thus if you want to run RGL on a system where there is no
graphics support, you should run \code{options(rgl.useNULL = TRUE)}
or set the environment variable \code{RGL_USE_NULL=TRUE} *before*
calling \code{library(rgl)} (or other code that 
loads \pkg{rgl}), and it will not fail in its attempt at initialization.
}
\value{
A logical value indicating the current default for use of the null
device.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{open3d}} and \code{\link{rgl.open}}.
}
\examples{
rgl.useNULL()
}

\name{expect_known_scene}
\alias{expect_known_scene}
\title{
Helper for \pkg{testthat} testing.
}
\description{
Gets the current scene using \code{\link{scene3d}}, and
compares the result to a saved value, optionally closing
the window afterwards.
}
\usage{
expect_known_scene(name, 
                   close = TRUE, 
                   file = paste0("testdata/", name, ".rds"),
                   ...)
}
\arguments{
  \item{name}{
By default, the base name of the file to save results in.
Not used if \code{file} is specified.
}
  \item{close}{
Whether to close the \pkg{rgl} window after the comparison.
}
  \item{file}{
The file in which to save the result.
}
  \item{...}{
Other arguments which will be passed to \code{\link[testthat]{expect_known_value}}.
}
}
\details{
This function uses \code{\link[testthat]{expect_known_value}} 
to save a representation of the scene.  During the comparison, the scene
is modified so that non-reproducible aspects are standardized
or omitted:  
\itemize{
\item{object ids are changed to start at 1.}
\item{system-specific font names and texture names are deleted.}
\item{the window is shifted to the top left of the screen.}
}

Calls to \code{expect_known_scene()} enable \code{testthat::\link[testthat]{local_edition}(2)} 
for the duration of the call, so it will work in \pkg{testthat} \dQuote{3rd edition}.

}
\value{
A value describing the changes to the saved object,
suitable for use in \code{\link[testthat]{test_that}()}.
}
\examples{
\dontrun{
# These lines can be included in testthat::test_that() code.
plot3d(1:10, 1:10, 1:10)
expect_known_scene("plot")
}
}

\name{rgl.select}
\alias{rgl.select}
\title{
Switch to select mode, and return the mouse position selected
}
\description{
Mostly for internal use, this function temporarily
installs a handler on a button of the mouse that will 
return the mouse coordinates of one click and drag rectangle.
}
\usage{
rgl.select(button = c("left", "middle", "right"),
           dev = cur3d(), subscene = currentSubscene3d(dev))
}
\arguments{
  \item{button}{
Which button to use?
}
  \item{dev, subscene}{
The RGL device and subscene to work with  
}
}
\value{
A vector of four coordinates:  the X and Y coordinates of the start and end of the
dragged rectangle.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{select3d}}, a version that allows the selection region to be used to
select points in the scene.
}

\name{rgl.surface}
\title{add height-field surface shape}
\alias{rgl.surface}
\description{
  Adds a surface to the current scene. The surface is defined by 
  a matrix defining the height of each grid point and two vectors
  defining the grid.
}
\usage{
rgl.surface(x, z, y, coords = 1:3, ..., 
            normal_x = NULL, normal_y = NULL, normal_z = NULL,
            texture_s = NULL, texture_t = NULL)
}
\arguments{
  \item{ x }{
    values corresponding to rows of \code{y}, or matrix of x coordinates
  }
  \item{ y }{
    matrix of height values
  }
  \item{ z }{
    values corresponding to columns of \code{y}, or matrix of z coordinates
  }
  \item{ coords }{
    See details
  }
  \item{ ... }{Material and texture properties. See \code{\link{rgl.material}} for details.} 
  \item{normal_x, normal_y, normal_z}{
      matrices of the same dimension as \code{y} giving the coordinates of normals at 
      each grid point}
  \item{texture_s, texture_t}{
      matrices of the same dimension as \code{z} giving the coordinates within the current texture
      of each grid point}


}
\details{
  Adds a surface mesh to the current scene. The surface is defined by 
  the matrix of height values in \code{y}, with rows corresponding 
  to the values in \code{x} and columns corresponding to the values in 
  \code{z}.
  
  The \code{coords} parameter can be used to change the geometric
  interpretation of \code{x}, \code{y}, and \code{z}.  The first entry 
  of \code{coords} indicates which coordinate (\code{1 = X}, 
  \code{2 = Y}, \code{3 = Z}) corresponds to the \code{x} parameter.
  Similarly the second entry corresponds to the \code{y} parameter,
  and the third entry to the \code{z} parameter.  In this way 
  surfaces may be defined over any coordinate plane.
  
  If the normals are not supplied, they will be calculated automatically based
  on neighbouring points.
  
  Texture coordinates run from 0 to 1 over each dimension of the texture bitmap.
  If texture coordinates are not supplied, they will be calculated to 
  render the texture exactly once over the grid.  Values greater than 1 can be
  used to repeat the texture over the surface.
  
  \code{rgl.surface} always draws the surface with the `front' upwards
  (i.e. towards higher \code{y} values).  This can be used to render
  the top and bottom differently; see \code{\link{rgl.material}} and
  the example below.
  
  If the \code{x} or \code{z} argument is a matrix, then it must be of the same
  dimension as \code{y}, and the values in the matrix will be used for the corresponding
  coordinates. This is used to plot shapes such as cylinders
  where y is not a function of x and z.
  
  \code{NA} values in the height matrix are not drawn.
}
\note{
It is recommended to use \code{\link{surface3d}} instead
of \code{rgl.surface}; use of the \code{rgl.*} functions 
is discouraged due to their side effects.
}
\value{
  The object ID of the displayed surface is returned invisibly.
}
\examples{

#
# volcano example taken from "persp"
#

data(volcano)

y <- 2 * volcano        # Exaggerate the relief

x <- 10 * (1:nrow(y))   # 10 meter spacing (S to N)
z <- 10 * (1:ncol(y))   # 10 meter spacing (E to W)

ylim <- range(y)
ylen <- ylim[2] - ylim[1] + 1

colorlut <- terrain.colors(ylen) # height color lookup table

col <- colorlut[ y - ylim[1] + 1 ] # assign colors to heights for each point

rgl.open()
rgl.surface(x, z, y, color = col, back = "lines")

}
\seealso{
  \code{\link{rgl.material}}, \code{\link{surface3d}}, \code{\link{terrain3d}}.
  See \code{\link{persp3d}} for a higher level interface.
}
\keyword{dynamic}

\name{callbacks}
\alias{rgl.setMouseCallbacks}
\alias{rgl.getMouseCallbacks}
\alias{rgl.setWheelCallback}
\alias{rgl.getWheelCallback}
\title{ User callbacks on mouse events }
\description{
  Set and get user callbacks on mouse events.
}
\usage{
rgl.setMouseCallbacks(button, begin = NULL, update = NULL, end = NULL, 
                      dev = cur3d(), subscene = currentSubscene3d(dev))
rgl.getMouseCallbacks(button, 
                      dev = cur3d(), subscene = currentSubscene3d(dev))

rgl.setWheelCallback(rotate, 
                      dev = cur3d(), subscene = currentSubscene3d(dev))

rgl.getWheelCallback(dev = cur3d(), subscene = currentSubscene3d(dev))

}
\arguments{
  \item{button}{ Which button? Use 1 for left, 2 for
  right, 3 for middle, 4 for wheel.  Use 0 to set
  an action when no button is pressed. }
  \item{begin}{ Called when mouse down event occurs }
  \item{update}{ Called when mouse moves }
  \item{end}{ Called when mouse is released }
  \item{rotate}{ Called when mouse wheel is rotated }
  \item{dev, subscene}{The RGL device and subscene to work with }
}
\details{
The set functions set event handlers on mouse events that occur within the current RGL window.
The \code{begin} and \code{update} events should be functions taking two arguments; these
will be the mouse coordinates when the event occurs.  The \code{end} event handler
takes no arguments.  The \code{rotate} event takes a single argument, which will be 
equal to \code{1} if the user pushes the wheel away by one click, and \code{2} if 
the user pulls the wheel by one click.

Alternatively, the handlers may be set to \code{NULL}, the default value, in which case
no action will occur.

If a subscene has multiple listeners, the user action will still only be called for the
subscene that received the mouse event.  It should consult \code{\link{par3d}("listeners")}
if it makes sense to take action on the whole group of subscenes.

The get function retrieves the callbacks that are currently set.

The \dQuote{no button} mouse handler may be set by
specifying \code{button = 0}.  The \code{begin}
function will be called the first time the mouse
moves within the subscene, and the \code{update}
function will be called repeatedly as it moves.  The
\code{end} function will never be called.
}
\value{
The set functions are called for the side effect of setting the mouse event handlers.

The \code{rgl.getMouseCallbacks} function returns a list containing the callback functions or \code{NULL} if no user callback is set.
The \code{rgl.getWheelCallback} returns the callback function or \code{NULL}.
}
\author{ Duncan Murdoch }
\seealso{ \code{\link{par3d}} to set built-in handlers,
\code{\link{setUserCallbacks}} to work with \code{\link{rglwidget}}.}
\examples{

 pan3d <- function(button, dev = cur3d(), subscene = currentSubscene3d(dev)) {
   start <- list()
   
   begin <- function(x, y) {
     activeSubscene <- par3d("activeSubscene", dev = dev)
     start$listeners <<- par3d("listeners", dev = dev, subscene = activeSubscene)
     for (sub in start$listeners) {
       init <- par3d(c("userProjection","viewport"), dev = dev, subscene = sub)
       init$pos <- c(x/init$viewport[3], 1 - y/init$viewport[4], 0.5)
       start[[as.character(sub)]] <<- init
     }
   }
   
   update <- function(x, y) {
     for (sub in start$listeners) {
       init <- start[[as.character(sub)]]
       xlat <- 2*(c(x/init$viewport[3], 1 - y/init$viewport[4], 0.5) - init$pos)
       mouseMatrix <- translationMatrix(xlat[1], xlat[2], xlat[3])
       par3d(userProjection = mouseMatrix \%*\% init$userProjection, dev = dev, subscene = sub )
      }
   }
   rgl.setMouseCallbacks(button, begin, update, dev = dev, subscene = subscene)
   cat("Callbacks set on button", button, "of RGL device", dev, "in subscene", subscene, "\n")
 }
 open3d()
 shade3d(icosahedron3d(), col = "yellow")
 # This only works in the internal display...
 pan3d(1)
}
\keyword{ dynamic }

\name{snapshot3d}
\alias{rgl.snapshot}
\alias{snapshot3d}
\alias{RGL_USE_WEBSHOT}
\title{Export screenshot}
\description{
  Saves the screenshot to a file.
}
\usage{
rgl.snapshot( filename, fmt = "png", top = TRUE )
snapshot3d( filename = tempfile(fileext = ".png"), 
            fmt = "png", top = TRUE,
            ..., scene, width = NULL, height = NULL,
            webshot = as.logical(Sys.getenv("RGL_USE_WEBSHOT", "TRUE")) )
}
\arguments{
  \item{filename}{path to file to save.}
  \item{fmt}{image export format, currently supported: png.  Ignored if \code{webshot = TRUE}. }
  \item{top}{whether to call \code{\link{rgl.bringtotop}}.
Ignored if \code{webshot = TRUE}.}
  \item{...}{arguments to pass to \code{webshot2::webshot} }
  \item{scene}{an optional result of \code{\link{scene3d}} 
    or \code{\link{rglwidget}} to plot}
  \item{width, height}{optional specifications of output
    size in pixels}
  \item{webshot}{Use the \pkg{webshot2} package to take the 
  snapshot}  
}
\details{
\code{rgl.snapshot()} is a low-level function
that copies the current RGL window from the screen.  Users
should use \code{snapshot3d()} instead; it is more flexible,
and (if \pkg{webshot2} is installed) can take images even if
no window is showing, and they can be larger than the physical
screen.

Animations can be created in a loop modifying the scene and saving 
each screenshot to a file. Various graphics programs (e.g. ImageMagick)
can put these together into a single animation. (See \code{\link{movie3d}} or
the example below.)
}
\value{
These functions are mainly called for the side effects.  The
filename of the saved file is returned invisibly.
}
\note{
When \code{rgl.useNULL()} is \code{TRUE}, only \code{webshot = TRUE}
will produce a snapshot.  It requires the \pkg{webshot2}
package and a Chrome browser.  If no suitable browser is
found, \code{snapshot3d()} will revert to \code{rgl.snapshot()}.
To override the automatic search, set
environment variable \env{CHROMOTE_CHROME}
to the path to a suitable browser.

\code{rgl.snapshot} works by taking an image from the displayed
window on-screen.
On some systems, the snapshot
will include content from other windows if they cover the active RGL
window.  Setting \code{top = TRUE} (the default) will use 
\code{\link{rgl.bringtotop}} before the snapshot
to avoid this.  

There are likely limits to how large \code{width} and
\code{height} can be set based on the display hardware; if these
are exceeded the results are undefined.  A typical result
is that the snapshot will still be made but at a smaller
size.

There are slight differences between the displays with
\code{webshot = TRUE} and \code{webshot = FALSE}, as the 
former are rendered using WebGL while the latter are rendered
using OpenGL.  Often the \code{webshot = TRUE} displays
have better quality, but they are usually slower to
produce, sometimes drastically so.  

Set the environment
variable \env{RGL_USE_WEBSHOT} to \code{"FALSE"} if
you want \code{rgl.snapshot} to be used by default.
}
\seealso{\code{\link{movie3d}}, \code{\link{view3d}}}

\examples{
if (interactive() && !in_pkgdown_example()) {
  saveopts <- options(rgl.useNULL = TRUE)
  plot3d(matrix(rnorm(300), ncol = 3, dimnames = list(NULL, c("x", "y", "z"))), 
         col = "red")
  options(saveopts)
  browseURL(snapshot3d())
}

\dontrun{

#
# create animation
#

shade3d(oh3d(), color = "red")
rgl.bringtotop()
view3d(0, 20)

olddir <- setwd(tempdir())
for (i in 1:45) {
  view3d(i, 20)
  filename <- paste("pic", formatC(i, digits = 1, flag = "0"), ".png", sep = "")
  snapshot3d(filename)
}
## Now run ImageMagick in tempdir().  Use 'convert' instead of 'magick'
## if you have an older version of ImageMagick:
##    magick -delay 10 *.png -loop 0 pic.gif
setwd(olddir)
}

}

\keyword{dynamic}

\name{makeDependency}
\alias{makeDependency}
\alias{RGL_DEBUGGING}
\title{
Process Javascript for HTML dependency
}
\description{
A utility function to help in development of internal 
Javascript code, this function processes the Javascript
to minify it and report on errors and bad style.
}
\usage{
makeDependency(name, src, script = NULL, package, 
               version = packageVersion(package),
               minifile = paste0(basename(src), ".min.js"),
               debugging = FALSE, ...)
}
\arguments{
  \item{name, src, script, package, version, ...}{
Arguments to pass to \code{htmltools::\link[htmltools]{htmlDependency}.}
}
  \item{minifile}{Basename of minified file.}
  \item{debugging}{
See details below.
}
}
\details{
This is a utility function used by RGL 
to process its Javascript code used when displaying \code{\link{rglwidget}} values.  It may be helpful in other packages to use
in their own installation.

If the \pkg{js} package version 1.2 or greater is installed,
the Javascript code will be minified and stored in the file named
by \code{minifile} in the \code{src} directory.  Syntax errors
in the code will stop the process; unused variables will be
reported.

If \code{debugging} is \code{TRUE}, the locations of Javascript syntax
errors will be reported, along with hints about improvements,
and the original files will be used in the dependency object
that is created.

If \code{debugging} is \code{FALSE} (the default), 
the minified file will be used in the dependency object,
hints won't be given, and syntax errors will lead to an
uninformative failure to minify.
}
\note{
The usual way to use \code{makeDependency} is 
to call it in a \file{.R} file in a package, saving
the result in a variable that will be used when an
HTML widget is created.  This way it is only run
during package installation, when it is safe to
write to the R library holding the package.  

Do not call it to write to the R library from 
code the user can run, as that is not allowed in
general.

If your package uses Roxygen, you may have problems
because by default Roxygen will run the code, and it
is likely to fail.  The current workaround is to
specify Roxygen option \code{load = "installed"}
which prevents it from running your \file{.R} code.
}
\value{
An object that can be included in a list of dependencies passed to \code{htmltools::\link[htmltools:htmlDependencies]{attachDependencies}}.
}
\author{
Duncan Murdoch
}
\examples{
\dontrun{
# This is a slightly simplified version of the code used to 
# produce one of the dependencies for rglwidget().  
# It writes to the system library copy of rgl so 
# has been marked not to run in the example code.

makeDependency("rglwidgetClass", 
               src = "htmlwidgets/lib/rglClass",
               script = c("rglClass.src.js",
                          "utils.src.js",
                          "buffer.src.js",
                          "subscenes.src.js",
                          "shaders.src.js",
                          "textures.src.js",
                          "projection.src.js",
                          "mouse.src.js",
                          "init.src.js",
                          "pieces.src.js",
                          "draw.src.js",
                          "controls.src.js",
                          "selection.src.js",
                          "rglTimer.src.js",
                          "pretty.src.js",
                          "axes.src.js",
                          "animation.src.js"),
               stylesheet = "rgl.css",
               package = "rgl",
               debugging = isTRUE(as.logical(Sys.getenv("RGL_DEBUGGING", "FALSE"))))
}
}

\name{decorate3d}
\alias{decorate3d}
\title{
Add decorations to a 3D plot
}
\description{
\code{decorate3d} adds the usual decorations to a plot:  labels, axes, etc.  
}
\usage{
decorate3d(xlim = NULL, ylim = NULL, zlim = NULL, 
           xlab = "x", ylab = "y", zlab = "z", 
           box = TRUE, axes = TRUE, 
           main = NULL, sub = NULL, 
           top = TRUE, aspect = FALSE, expand = 1.03, 
           tag = material3d("tag"), ...)
}
\arguments{  
  \item{xlim, ylim, zlim}{These are used
for the labels.}
  \item{xlab, ylab, zlab}{labels for the coordinates.}
  \item{box, axes}{whether to draw a box and axes.}
  \item{main, sub}{main title and subtitle.}
  \item{top}{whether to bring the window to the top when done.}
  \item{aspect}{either a logical indicating whether to adjust the aspect ratio, or a new ratio.}
  \item{expand}{how much to expand the box around the data, if it is drawn.}
  \item{tag}{optional label for objects
  being produced.}
  \item{...}{ignored.}
}
\value{
The RGL id values for those items.
}
\examples{
open3d()
shade3d(tetrahedron3d(), col = "red")
decorate3d(main = "A Tetrahedron")
}

\name{aspect3d}
\alias{aspect3d}
\title{Set the aspect ratios of the current plot}
\description{
This function sets the apparent ratios of the x, y, and z axes
of the current bounding box.
}
\usage{
aspect3d(x, y = NULL, z = NULL)
}
\arguments{
  \item{x}{The ratio for the x axis, or all three ratios, or \code{"iso"} }
  \item{y}{The ratio for the y axis }
  \item{z}{The ratio for the z axis }
}
\details{
If the ratios are all 1, the bounding box will be displayed as a cube approximately filling the display.
Values may be set larger or smaller as desired.  Aspect \code{"iso"} signifies that the
coordinates should all be displayed at the same scale, i.e. the bounding box should not be
rescaled.  (This corresponds to the default display before \code{aspect3d} has been called.)
Partial matches to \code{"iso"} are allowed.

\code{aspect3d} works by modifying \code{par3d("scale")}.
}
\value{
The previous value of the scale is returned invisibly.
}
\author{Duncan Murdoch}
\seealso{\code{\link{plot3d}}, \code{\link{par3d}}}
\examples{
  x <- rnorm(100)
  y <- rnorm(100)*2
  z <- rnorm(100)*3
  
  open3d()
  plot3d(x, y, z)
  aspect3d(1, 1, 0.5)
  highlevel()  # To trigger display
  open3d()
  plot3d(x, y, z)
  aspect3d("iso")
  highlevel()
}
\keyword{dynamic}

\name{as.rglscene}
\alias{as.rglscene}
\title{
Convert an object to an \code{rglscene} object.
}
\description{
This is a placeholder generic function, to allow other
packages to create \code{"rglscene"} objects compatible with
the objects produced by \code{\link{scene3d}}.

No methods are currently defined in \pkg{rgl}.
}
\usage{
as.rglscene(x, ...)
}
\arguments{
  \item{x}{
Object to convert.
}
  \item{\dots}{
Other parameters to pass to methods.
}
}

\name{rgl.init}
\title{Initializing RGL}
\alias{rgl.init}
\description{
Initializing the RGL system.
}
\usage{
rgl.init(initValue = 0, onlyNULL = FALSE, 
         debug = getOption("rgl.debug", FALSE))
}
\arguments{
  \item{initValue}{value for internal use only}
  \item{onlyNULL}{only initialize the null (no display) device}
  \item{debug}{enable some debugging messages}
}

\value{
Normally the user doesn't call \code{rgl.init} at all:  it
is called when the package is loaded.  It returns no
useful value.
}
\details{
  If \code{useNULL} is \code{TRUE}, RGL will use a \dQuote{null}
  device.  This device records objects as they are plotted, but 
  displays nothing. It is intended for use with \code{\link{rglwidget}}
  and similar functions.
  
  Currently \code{debug} only controls messages printed by the OpenGL
  library during initialization.  In future \code{debug = TRUE} may
  become more verbose.

For display within an OpenGL window in R, RGL requires the OpenGL
system to be installed and available.  If there is a problem 
initializing it, you may see the message
\verb{'rgl.init' failed, running with 'rgl.useNULL = TRUE'.}  There
are several causes and remedies:
\itemize{
\item{On any system, the OpenGL libraries need to be present for RGL to be able to start an OpenGL device.}
\itemize{
\item{On macOS, you need to install XQuartz.  It is available from
\url{https://www.xquartz.org}.}
\item{On Linux, you need to install Mesa 3D.  One of these commands
may work, depending on your system:
\verb{
  zypper source-install --build-deps-only Mesa # openSUSE/SLED/SLES
  yum-builddep mesa      # yum Fedora, OpenSuse(?)
  dnf builddep mesa      # dnf Fedora
  apt-get build-dep mesa # Debian, Ubuntu and related
  }
}
\item{Windows should have OpenGL installed by default.}
}
\item{On Unix-alike systems (macOS and Linux, for example), RGL
normally
uses the GLX system for creating displays.  If the graphic is created
on a remote machine, it may need to use \dQuote{Indirect GLX} (IGLX).  Due
to security concerns, this is often disabled by default.  See
\url{https://www.x.org/wiki/Development/Security/Advisory-2014-12-09/}
for a discussion of the security issues, and 
\url{https://unix.stackexchange.com/q/317954}
for ways to re-enable IGLX.}
\item{The \url{https://www.virtualgl.org} project is intended to 
be a way to avoid IGLX, by rendering remotely and sending bitmaps
to the local machine.  It's not a simple install...}
\item{If you don't need to see RGL displays on screen, you
can use the \dQuote{NULL device}.  See \code{\link{rgl.useNULL}}.}
\item{If you can't build the \pkg{rgl} package with OpenGL
support, you can
disable it and use the NULL device.  (This may happen automatically
during configuration, but you'll get a tested result if you
specify it explicitly.)  See the instructions in the
\file{README} file in the source tarball.}
}
}

\name{observer3d}
\alias{observer3d}
\title{
Set the observer location
}
\description{
This function sets the location of the viewer.
}
\usage{
observer3d(x, y = NULL, z = NULL, auto = FALSE)
}
\arguments{
  \item{x, y, z}{
The location as a 3 vector, using the usual \code{xyz.coords}
conventions for specification.  If \code{x} is missing
or any coordinate is \code{NA}, no change will be made to the
location.  
}
  \item{auto}{ 
If \code{TRUE}, the location will be set 
automatically by RGL to make the whole bounding box visible.
}
}
\details{
This function
sets the location of the viewer relative to the scene, after the model transformations
(scaling, rotation) have been done, but before lighting or projection have been 
applied.  (See \code{\link{par3d}} for details on the rendering pipeline.)

The coordinate system is a slightly strange one:  the X coordinate
moves the observer location from left to right, and the Y coordinate moves up
and down.  The Z coordinate changes the depth from the viewer.  All are measured
relative to the center of the bounding box (\code{par("bbox")}) of the subscene.
The observer always looks in the positive Z direction
after the model rotation have been done.  The coordinates are in post-scaling
units.
}
\note{
This function is likely to change in future versions of RGL, to allow more 
flexibility in the specification of the observer's location and orientation.
}
\value{
Invisibly returns the previous value.
}
\author{
Duncan Murdoch
}

\examples{
example(surface3d)  # The volcano data
observer3d(0, 0, 440) # Viewed from very close up
}
\keyword{ graphics }

\name{mfrow3d}
\alias{mfrow3d}
\alias{layout3d}
\alias{next3d}
\alias{subsceneList}
\alias{clearSubsceneList}
\title{
Set up multiple figure layouts
}
\description{
The \code{mfrow3d} and \code{layout3d} functions provide functionality in RGL
similar to \code{\link{par}("mfrow")} and \code{\link{layout}} in classic R graphics.
}
\usage{
subsceneList(value, window = cur3d())

mfrow3d(nr, nc, byrow = TRUE, parent = NA, sharedMouse = FALSE, ...)
layout3d(mat, widths = rep.int(1, ncol(mat)), 
              heights = rep.int(1, nrow(mat)),
              parent = NA, sharedMouse = FALSE,
              ...)
next3d(current = NA, clear = TRUE, reuse = TRUE)
clearSubsceneList(delete = currentSubscene3d() \%in\% subsceneList(), 
                  window = cur3d())
}
\arguments{
  \item{value}{
A new subscene list to set.  If missing, return the current one (or \code{NULL}).
}
  \item{window}{
Which window to operate on.
}
  \item{nr, nc}{
Number of rows and columns of figures.
}
  \item{byrow}{
Whether figures progress by row (as with \code{\link{par}("mfrow")}) or by column 
(as with \code{\link{par}("mfcol")}).
}
  \item{mat, widths, heights}{
Layout parameters; see \code{\link{layout}} for their interpretation.
}
  \item{parent}{
The parent subscene.  \code{NA} indicates the current subscene.  See Details below.
}
  \item{sharedMouse}{
Whether to make all subscenes \code{\link{par3d}("listeners")} to each other.
}
  \item{\dots}{
Additional parameters to pass to \code{\link{newSubscene3d}} as each
subscene is created.
}
  \item{current}{
The subscene to move away from.  \code{NA} indicates the current subscene.
}
  \item{clear}{
Whether the newly entered subscene should be cleared upon entry.
}
  \item{reuse}{
Whether to skip advancing if the current subscene has no objects in it.
}
  \item{delete}{
If \code{TRUE}, delete the subscenes in the current window.
}
}
\details{
rgl can maintain a list of subscenes; the \code{mfrow3d} and \code{layout3d} functions create that list.
When the list is in place, \code{next3d} causes RGL to move to the next scene in the list, or cycle
back to the first one.

Unlike the classic R graphics versions of these functions, these functions are completely compatible with
each other.  You can mix them within a single RGL window.

In the default case where \code{parent} is missing, \code{mfrow3d} and \code{layout3d} will
call \code{clearSubsceneList()} at the start.  

By default \code{clearSubsceneList()} checks
whether the current subscene is in the current subscene list; if so, it will delete
all subscenes in the list, and call \code{\link{gc3d}} to delete any objects that are no
longer shown.  The subscene list will be set to a previous value if one was recorded,
or \code{NULL} if not.

If \code{parent} is specified in \code{mfrow3d} or \code{layout3d} 
(even as \code{NA}), the new subscenes will be created within 
the parent.

The \code{next3d()} function first finds out if the current
subscene is in the current list.  If not, it moves to the
previous list, and looks there.  Once it finds a list 
containing the current subscene, it moves to the next entry
in that list.  If it can't find one, it creates a length
one list containing just the current subscene.
}
\value{
\code{mfrow3d} and \code{layout3d} return a vector of subscene id values that have just been
created.  If a previous subscene list was in effect and was not automatically cleared,
it is attached as an attribute \code{"prev"}.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{newSubscene3d}}, \code{\link{par}}, \code{\link{layout}}.
}
\examples{
shapes <- list(Tetrahedron = tetrahedron3d(), Cube = cube3d(), Octahedron = octahedron3d(),
               Icosahedron = icosahedron3d(), Dodecahedron = dodecahedron3d(),
               Cuboctahedron = cuboctahedron3d())
col <- rainbow(6)
open3d()
mfrow3d(3, 2)
for (i in 1:6) {
  next3d()   # won't advance the first time, since it is empty
  shade3d(shapes[[i]], col = col[i])
}
highlevel(integer()) # To trigger display as rglwidget

open3d()
mat <- matrix(1:4, 2, 2)
mat <- rbind(mat, mat + 4, mat + 8)
layout3d(mat, height = rep(c(3, 1), 3), sharedMouse = TRUE)
for (i in 1:6) {
  next3d()
  shade3d(shapes[[i]], col = col[i])
  next3d()
  text3d(0, 0, 0, names(shapes)[i])
}
highlevel(integer())
}
\keyword{graphics }

\name{import}
\alias{\%>\%}
\alias{pipe}
\docType{import}
\title{Imported from magrittr}
\description{
This object is imported from \pkg{magrittr}. Follow the link to its documentation.
\describe{
  \item{magrittr}{\code{\link[magrittr:pipe]{\%>\%}}}
}

Pipes can be used to string together
\code{\link{rglwidget}} calls and \code{\link{playwidget}}
calls.  See \code{\link{ageControl}} for an example.

}

\name{ageSetter}
\alias{ageSetter}
\title{
Set WebGL scene properties based on the age of components
of objects.
}
\description{
Many \pkg{rgl} shapes contain lists of vertices with various
attributes (available via \code{\link{rgl.attrib}}).  This function
modifies the data for those attributes in a WebGL scene.
}
\usage{
ageSetter(births, ages, 
         colors = NULL, alpha = NULL, radii = NULL, 
         vertices = NULL, normals = NULL, origins = NULL, 
         texcoords = NULL, 
         objids, prefixes = "", digits = 7, 
         param = seq(floor(min(births)), ceiling(max(births))))
}
\arguments{
  \item{births}{
Numeric vector with one value per vertex, used to determine the
\dQuote{age} of the vertex when displaying it.
}
  \item{ages}{
A non-decreasing sequence of \dQuote{ages}.  
}
  \item{colors, alpha, radii, vertices, normals, origins, texcoords}{
Attributes of the vertices.  Non-\code{NULL} attributes will be 
interpolated from these values.  See the Details section below.
}
  \item{objids, prefixes}{
The object ids and scene prefixes to modify.  These are recycled to
the same length.
}
  \item{digits}{
How many digits to output in the generated Javascript code.
}
  \item{param}{
Default values to be used by a slider control calling the generated
function.
}
}
\details{
The vertex attributes
are specified as follows:
\describe{
\item{colors}{A vector of colors in a format suitable for input to \code{\link{col2rgb}}}
\item{alpha}{A numeric vector of alpha values between 0 and 1.}
\item{radii}{A numeric vector of sphere radii.}
\item{vertices}{A 3-column matrix of vertex coordinates.}
\item{normals}{A 3-column matrix of vertex normals.}
\item{origins}{A 2-column matrix of origins for text or sprites.}
\item{texcoords}{A 2-column matrix of texture coordinates.}
}

All attributes must have the same number of entries (rows for the matrices)
as the \code{ages} vector.  The \code{births} vector must have the
same number of entries as the number of vertices in the object.

Not all objects contain all attributes listed here; if one is
chosen that is not a property of the corresponding object, a
Javascript \code{alert()} will be generated.  
}
\value{
A character vector of class \code{c("ageSetter", "propertySetter")} 
containing Javascript
code defining a function suitable for use in a \code{\link{propertySlider}}.

The function takes a single argument, \code{time}, and uses it to compute the
\dQuote{age} of vertex \code{i} as \code{time - births[i]}.  
Those are then used with the \code{ages}
argument to linearly interpolate settings of the specified attributes.
Extrapolation is constant.  Repeated values in \code{ages} can be used
to obtain discontinuities in the settings.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{propertySlider}}; more detailed control is available in
\code{\link{vertexSetter}}.
}

\name{textureSource}
\alias{textureSource}
\title{
Retrieve source code used to produce texture file.
}
\description{
Internally, \pkg{rgl} works with PNG files for textures.
If a texture is requested using a different format, a temporary
PNG file of the image will be saved.  This function allows
you to retrieve the original expression used to produce
the texture.
}
\usage{
textureSource(texture)
}
\arguments{
  \item{texture}{
The filename of a texture file.  If missing, the directory
where texture files are stored will be returned.
}
}
\details{
\pkg{rgl} creates a new file in the temporary directory
whenever a non-PNG texture is used.  It will delete them
when it knows there are no references and 
at the end of the session, but conceivably there will be 
situations where you need to delete them earlier.  Calling
\code{textureSource()} with no arguments will give
you the directory holding the textures so that they can
be deleted sooner.
}
\value{
If \code{texture} is specified and it is the name of a 
temporary PNG texture file produced by \pkg{rgl}, the 
expression used to specify the texture will be returned.  
If it is the name of some other file, \code{texture} will
be returned.

If no argument is given, the session-specific directory
holding the temporary texture files will be returned.
}

\seealso{
\code{\link{material3d}}
}
\examples{
xyz <- cbind(c(0,1,1,0), c(0,0,1,1), c(0,0,0,0))
st <- xyz[,1:2]

open3d()
id <- quads3d(xyz, texcoords = st, 
               texture = as.raster(matrix(colors()[1:120], ncol = 10)), 
               col="white")
material3d(id = id, "texture")
textureSource(material3d(id = id, "texture"))
}

\name{rgl.incrementID}
\alias{rgl.incrementID}
\title{
Increment ID
}
\description{
This function is mainly for internal use.  It simply 
increments the internal object ID number and returns
the new value.  Negative values have no effect.
}
\usage{
rgl.incrementID(n = 1L)
}
\arguments{
  \item{n}{
An integer increment to use.
}
}
\value{
The resulting ID value.
}
\examples{
# Get the current ID value
rgl.incrementID(0)

# Increment it
rgl.incrementID()
}

\name{rgl.primitive}
\alias{rgl.primitive}
\alias{rgl.points}
\alias{rgl.lines}
\alias{rgl.linestrips}
\alias{rgl.triangles}
\alias{rgl.quads}
\title{add primitive set shape}
\description{
  Adds a shape node to the current scene
}
\usage{
rgl.points(x, y = NULL, z = NULL, ... )
rgl.lines(x, y = NULL, z = NULL, ... )
rgl.linestrips(x, y = NULL, z = NULL, ...)
rgl.triangles(x, y = NULL, z = NULL, normals = NULL, texcoords = NULL, ... )
rgl.quads(x, y = NULL, z = NULL, normals = NULL, texcoords = NULL, ... )
}
\arguments{
  \item{x, y, z}{coordinates.  Any reasonable way of defining the
    coordinates is acceptable.  See the function \code{\link[grDevices]{xyz.coords}}
    for details.}
  \item{normals}{Normals at each point.}
  \item{texcoords}{Texture coordinates at each point.}
  \item{ ... }{Material properties. See \code{\link{rgl.material}} for details.}  
}
\details{
  Adds a shape node to the scene. The appearance is defined by the material properties.
  See \code{\link{rgl.material}} for details.
  
  The names of these functions correspond to OpenGL primitives.  They
  all take a sequence of vertices in \code{x, y, z}.  The only non-obvious 
  ones are \code{rgl.lines} which draws line segments based on
  pairs of vertices, and \code{rgl.linestrips} which joins the
  vertices.
  
  For triangles and quads, the normals at each vertex may be specified
  using \code{normals}.  These may be given in any way that would be
  acceptable as a single argument to \code{\link[grDevices]{xyz.coords}}.  
  These need not match the actual normals to the polygon:
  curved surfaces can be simulated by using other choices of normals.
  
  Texture coordinates may also be specified.  These may be given in
  any way that would be acceptable as a single argument to 
  \code{\link[grDevices]{xy.coords}}, and are interpreted in terms
  of the bitmap specified as the material texture, with \code{(0, 0)} 
  at the lower left, \code{(1, 1)} at the upper right.  The texture
  is used to modulate the color of the polygon.
  
  These are the lower level functions called by
  \code{\link{points3d}}, \code{\link{segments3d}},
  \code{\link{lines3d}}, etc.  The two principal differences between
  the \code{rgl.*} functions and the \code{*3d} functions are that the
  former set all unspecified material properties to defaults, whereas
  the latter use current values as defaults; the former make
  persistent changes to material properties with each call, whereas
  the latter make temporary changes only for the duration of the call.

}
\value{
   Each primitive function returns the integer object ID of the shape that
   was added to the scene.  These can be passed to \code{\link{pop3d}}
   to remove the object from the scene.
}
\examples{
rgl.open()
rgl.points(rnorm(1000), rnorm(1000), rnorm(1000), color = heat.colors(1000))
}
\seealso{
\code{\link{rgl.material}},
\code{\link{rgl.spheres}},
\code{\link{rgl.texts}},
\code{\link{rgl.surface}},
\code{\link{rgl.sprites}}
}
\keyword{dynamic}

\name{facing3d}
\alias{facing3d}
\alias{projectDown}
\title{
Subset an object to parts facing in a particular direction
}
\description{
\code{facing3d} subsets an object by converting it to a triangle mesh,
then subsetting to those triangles that are counterclockwise
(for \code{front = TRUE}) when projected into a plane.

\code{projectDown} computes a projection that \dQuote{looks down} the specified direction.
}
\usage{
facing3d(obj, up = c(0, 0, 1), 
         P = projectDown(up), 
         front = TRUE, strict = TRUE)
projectDown(up)
}
\arguments{
  \item{obj}{
An object that can be converted to a triangular mesh object.
}
  \item{up}{
The direction that is to be considered \dQuote{up}.  It may
be either a 3 vector in Euclidean coordinates or a 4 vector in
homogeneous coordinates.
  }
  \item{P}{
The projection to use for draping, a 4x4 matrix.  See \code{\link{drape3d}} for details on how \code{P} is used.
}
  \item{front}{
If \code{front = TRUE}, retains triangles that are counterclockwise after projection by \code{P}, otherwise
retains those that are clockwise.
}
  \item{strict}{If \code{TRUE}, drops indeterminate triangles
(those that are annihilated by \code{P}).}
}
\details{
By default 
the returned subset will be those triangles whose upper side
matches \code{front}.  Change \code{up} or use an
arbitrary projection for different subsets.

\code{\link{drape3d}} and \code{\link{shadow3d}} project objects onto meshes; these functions can be used to
project only onto the top or front.
}
\value{
\code{facing3d} returns a mesh object made of those triangles which face in the 
desired direction.

\code{projectDown} computes a 4x4 matrix.   The first two
coordinates of \code{asEuclidean(x \%*\% projectDown(up))} 
give a projection of \code{x} from above into a plane, where 
\code{up} determines which direction is taken to be \dQuote{up}.
}
\seealso{\code{\link{drape3d}}, \code{\link{shadow3d}}}
\examples{
open3d()
d <- rnorm(3)
d <- d/sqrt(sum(d^2))
shade3d( facing3d( icosahedron3d(), up = d, strict = FALSE), 
         col = "yellow")
wire3d( facing3d( icosahedron3d(), up = d, front = FALSE), 
         col = "black")
# Show the direction:
arrow3d(-2*d , -d)
}

\name{mesh3d}
\alias{shape3d}
\alias{mesh3d}
\alias{qmesh3d}
\alias{tmesh3d}

\title{Construct 3D mesh objects}
\description{
Creates meshes containing points, segments, triangles and quads.
}
\usage{
  mesh3d( x, y = NULL, z = NULL, vertices,
                    material = NULL,
                    normals = NULL, texcoords = NULL,
                    points = NULL, segments = NULL,
                    triangles = NULL, quads = NULL,
                    meshColor = c("vertices", "edges", "faces", "legacy"))
  qmesh3d(vertices, indices, homogeneous = TRUE, material = NULL, 
         normals = NULL, texcoords = NULL,
         meshColor = c("vertices", "edges", "faces", "legacy"))
  tmesh3d(vertices, indices, homogeneous = TRUE, material = NULL, 
         normals = NULL, texcoords = NULL,
         meshColor = c("vertices", "edges", "faces", "legacy"))
  
}
\arguments{
  \item{x, y, z}{coordinates.  Any reasonable way of defining the
    coordinates is acceptable.  See the function \code{\link[grDevices]{xyz.coords}}
    for details.}
  \item{vertices}{A 4 row matrix of homogeneous coordinates; takes
  precedence over \code{x, y, z}.}
  \item{material}{material properties for later rendering}
  \item{normals}{normals at each vertex}
  \item{texcoords}{texture coordinates at each vertex}
  \item{points}{vector of indices of vertices to draw as points}
  \item{segments}{2 x n matrix of indices of vertices to draw as
  segments}
  \item{triangles}{3 x n matrix of indices of vertices to draw
  as triangles}
  \item{quads}{4 x n matrix of indices of vertices to draw as quads}
  \item{indices}{(obsolete) 3 or 4 x n matrix of vertex indices}
  \item{homogeneous}{(obsolete) should \code{tmesh3d} and \code{qmesh3d} vertices be assumed to be homogeneous?}
  \item{meshColor}{how should colours be interpreted?  See details in \code{\link{shade3d}}.}
  }
\details{
  These functions create \code{mesh3d} objects, which consist of a matrix
  of vertex coordinates together with a matrices of indices indicating how the vertices should be displayed, and 
  material properties.

  The \code{"shape3d"} class is a general class for shapes that can be plotted
  by \code{dot3d}, \code{wire3d} or \code{shade3d}.
  
  The \code{"mesh3d"} class is a class of objects that form meshes: the vertices
  are in member \code{vb}, as a 4 by \code{n} matrix using
  homogeneous coordinates.  Indices of these vertices are
  contained in optional components \code{ip} for points,
  \code{is} for line segments, \code{it} for triangles, and
  \code{ib} for quads.
  Individual meshes may have any combination of these.

The functions \code{tmesh3d} and \code{qmesh3d} are included
for back-compatibility; they produce meshes of triangles and
quads respectively.
}
\value{
Objects of class \code{c("mesh3d",
  "shape3d")}. 
  
  See \code{\link{points3d}} for a discussion of texture coordinates.
}
\examples{

  # generate a quad mesh object

  vertices <- c( 
     -1.0, -1.0, 0,
      1.0, -1.0, 0,
      1.0,  1.0, 0,
     -1.0,  1.0, 0
  )
  indices <- c( 1, 2, 3, 4 )
  
  open3d()  
  wire3d( mesh3d(vertices = vertices, quads = indices) )
  
}
\seealso{
  \code{\link{shade3d}}, \code{\link{shapelist3d}} for multiple shapes
}
\keyword{dynamic}

\name{show2d}
\alias{show2d}
\title{
Draw a 2D plot on a rectangle in a 3D scene
}
\description{
This function uses a bitmap of a standard 2D graphics plot
as a texture on a quadrilateral.  Default arguments are
set up so that it will appear on the face of the bounding
box of the current 3D plot, but optional arguments allow
it to be placed anywhere in the scene.
}
\usage{
show2d(expression, 
  face = "z-", line = 0, 
  reverse = FALSE, rotate = 0, 
  x = NULL, y = NULL, z = NULL, 
  width = 480, height = 480, 
  filename = NULL, 
  ignoreExtent = TRUE, 
  color = "white", specular = "black", lit = FALSE, 
  texmipmap = TRUE, texminfilter = "linear.mipmap.linear",
  expand = 1.03,
  texcoords = matrix(c(0, 1, 1, 0, 0, 0, 1, 1), ncol = 2),  ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{expression}{
Any plotting commands to produce a plot in standard graphics.
Ignored if \code{filename} is not \code{NULL}.
}
  \item{face}{
A character string defining which face of the bounding
box to use.  See Details below.
}
  \item{line}{
How far out from the bounding box should the quadrilateral
be placed?  Uses same convention as \code{\link{mtext3d}}:
not lines of text, but fraction of the bounding box size.
}
  \item{reverse, rotate}{ 
Should the image be reversed or rotated?  See Details below.
}
  \item{x, y, z}{
Specific values to use to override \code{face}.
}
  \item{width,height}{
Parameters to pass to \code{\link{png}} when creating the
bitmap.  See Details below.
}
  \item{filename}{
A \file{.png} file image to use as the texture.
}
  \item{ignoreExtent}{
Whether the quadrilateral should be ignored when computing
the bounding box of the scene.
}
  \item{color, specular, lit, texmipmap, texminfilter, \dots}{
Material properties to use for the quadrilateral.
}
  \item{expand}{Amount by which the quadrilateral is 
expanded outside the bounding box of the data.
}
  \item{texcoords}{Coordinates on the image.  Lower left
of the bitmap is \code{c(0,0)}, upper right is \code{c(1,1)}.}
}  
\details{
The default arguments are chosen to make it easy to place
a 2D image on the face of the bounding box.  If \code{x}, 
\code{y} and \code{z} are \code{NULL} (the defaults),
\code{face} will be used as a code for one of the
six faces of the bounding box.  The first letter should
be \code{"x"}, \code{"y"} or \code{"z"}; this defines 
the axis perpendicular to the desired face.  If the
second letter is \code{"-"} or is missing, the face
will be chosen to be the face with the lower value 
on that axis.  Any other letter will use the opposite 
face.

If any of  \code{x}, \code{y} or \code{z} is given, the
specified value will be used to replace the value calculated 
above.  Usually four values should be given, corresponding to
the coordinates of the lower left, lower right, upper right
and upper left of the destination for the image before
\code{reverse} and \code{rotate} are used.  Fewer values
can be used for one or two coordinates; \code{\link{cbind}}
will be used to put together all 3 coordinates into
a 4 by 3 matrix (which will be returned as an attribute
of the result).

The bitmap plot will by default be oriented so that it is
properly oriented when viewed from the
direction of the higher values of the perpendicular coordinate,
and its lower left corner is at the lower value of the two
remaining coordinates.  The argument \code{reverse} causes
the orientation to be mirrored, and \code{rotate} causes it
to be rotated by multiples of 90 degrees.  \code{rotate} should be an integer,
with \code{0} for no rotation, \code{1} for a 90 degree
counter-clockwise rotation, etc.

The \code{width} and \code{height} arguments control the shape
and resolution of the bitmap.  The defaults give a square
bitmap, which is appropriate with the usual \code{c(1,1,1)}
aspect ratios (see \code{aspect3d}).  Some tuning may be 
needed to choose the resolution.  The plot will look best when
displayed at its original size; shrinking it smaller tends to make
it look faded, while expanding it bigger will make it look 
blurry.  If \code{filename} is given, the width and height
will be taken from the file, and \code{width} and \code{height}
arguments will be ignored.
}
\value{
Invisibly returns the id value of the quadrilateral,
with the following attributes:
\item{value}{The value returned by \code{expression}.}
\item{xyz}{A 4 by 3 matrix giving the coordinates of the corners
as used in plotting.}
\item{texcoords}{A 4 by 2 matrix giving the texture coordinates
of the image.}
\item{filename}{The filename for the temporary file holding
the bitmap image.}
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{bgplot3d}} uses a plot as the background for the window.
}
\examples{
example(plot3d, ask = FALSE)
show2d({
  par(mar=c(0,0,0,0))
  plot(x, y, col = rainbow(1000), axes=FALSE)
  })
}

\name{primitives}
\alias{primitives}
\alias{points3d}
\alias{lines3d}
\alias{segments3d}
\alias{triangles3d}
\alias{quads3d}
\alias{3dobjects}
\title{Add primitive shape}
\description{
Adds a shape node to the current scene.
}
\usage{
points3d(x, y = NULL, z = NULL,  ...)
lines3d(x, y = NULL, z = NULL,  ...)
segments3d(x, y = NULL, z = NULL, ...)
triangles3d(x, y = NULL, z = NULL, ...)
quads3d(x, y = NULL, z = NULL, ...)
}
\arguments{
  \item{x, y, z}{coordinates. Any reasonable way of defining the
    coordinates is acceptable.  See the function \code{\link[grDevices]{xyz.coords}}
    for details.}
  \item{ ... }{Material properties (see \code{\link{material3d}}), \code{normals},
\code{texcoords} or \code{indices}; see details below.}  
}
\details{

The functions \code{points3d}, \code{lines3d}, \code{segments3d},
\code{triangles3d} and \code{quads3d} add points, joined lines, line segments,
filled triangles or quadrilaterals to the plots.  They correspond to the OpenGL types
\code{GL_POINTS, GL_LINE_STRIP, GL_LINES, GL_TRIANGLES} and \code{GL_QUADS} respectively.  

Points are taken in pairs by \code{segments3d}, triplets as the vertices
of the triangles, and quadruplets for the quadrilaterals.  Colors are applied vertex by vertex; 
if different at each end of a line segment, or each vertex of a polygon, the colors
are blended over the extent of the object.  Polygons
must be non-degenerate and quadrilaterals must be entirely 
in one plane and convex, or the results are undefined.

The appearance of the new objects are defined by the material properties.
See \code{\link{material3d}} for details.

  For triangles and quads, the normals at each vertex may be specified
  using \code{normals}.  These may be given in any way that would be
  acceptable as a single argument to \code{\link[grDevices]{xyz.coords}}.  
  These need not match the actual normals to the polygon:
  curved surfaces can be simulated by using other choices of normals.
  
  Texture coordinates may also be specified.  These may be given in
  any way that would be acceptable as a single argument to 
  \code{\link[grDevices]{xy.coords}}, and are interpreted in terms
  of the bitmap specified as the material texture, with \code{(0, 0)} 
  at the lower left, \code{(1, 1)} at the upper right.  The texture
  is used to modulate the color of the polygon.

All of these functions support an argument called
\code{indices}, which allows vertices (and other attributes)
to be re-used, as they are in objects created by \code{\link{mesh3d}} and related functions.  This is
intended to be used on smooth surfaces, where each shared
vertex has just one value for normals, colors and texture
coordinates.  

For shapes with flat-looking faces (e.g. polyhedra like \code{\link{cube3d}}),
the vertices \bold{must} be duplicated
to be rendered properly.


}
\value{
   Each function returns the integer object ID of the shape that
   was added to the scene.  These can be passed to \code{\link{pop3d}}
   to remove the object from the scene.
}
\author{
  Ming Chen and Duncan Murdoch
}

\examples{
# Show 12 random vertices in various ways. 

M <- matrix(rnorm(36), 3, 12, dimnames = list(c('x', 'y', 'z'), 
                                       rep(LETTERS[1:4], 3)))

# Force 4-tuples to be convex in planes so that quads3d works.

for (i in c(1, 5, 9)) {
    quad <- as.data.frame(M[, i + 0:3])
    coeffs <- runif(2, 0, 3)
    if (mean(coeffs) < 1) coeffs <- coeffs + 1 - mean(coeffs)
    quad$C <- with(quad, coeffs[1]*(B - A) + coeffs[2]*(D - A) + A)
    M[, i + 0:3] <- as.matrix(quad)
}

open3d()

# Rows of M are x, y, z coords; transpose to plot

M <- t(M)
shift <- matrix(c(-3, 3, 0), 12, 3, byrow = TRUE)

points3d(M)
lines3d(M + shift)
segments3d(M + 2*shift)
triangles3d(M + 3*shift, col = 'red')
quads3d(M + 4*shift, col = 'green')  
text3d(M + 5*shift, texts = 1:12)

# Add labels

shift <- outer(0:5, shift[1, ])
shift[, 1] <- shift[, 1] + 3
text3d(shift, 
       texts = c('points3d', 'lines3d', 'segments3d',
         'triangles3d', 'quads3d', 'text3d'),
       adj = 0)
 rgl.bringtotop()
}
\keyword{dynamic}

\name{plot3d.lm}
\alias{plot3d.lm}
\title{
Method for plotting simple linear fit
}
\description{
This function provides several plots of the result of
fitting a two-predictor model.
}
\usage{
\method{plot3d}{lm}(x, 
       which = 1, 
       plane.col = "gray", plane.alpha = 0.5,
       sharedMouse = TRUE, 
       use_surface3d, 
       do_grid = TRUE,
       grid.col = "black", 
       grid.alpha = 1,
       grid.steps = 5,
       sub.steps = 4,
       vars = get_all_vars(terms(x), x$model),
       clip_to_density = 0,
       ...)
}
\arguments{
  \item{x}{
An object inheriting from class \code{"lm"} obtained by fitting
a two-predictor model.
}
  \item{which}{
Which plot to show?  See Details below.
}
  \item{plane.col, plane.alpha}{
These parameters control
the colour and transparency of a plane or surface.
}
  \item{sharedMouse}{
If multiple plots are requested, should they share
mouse controls, so that they move in sync?
}
  \item{use_surface3d}{
Use the \code{\link{surface3d}} function to plot the surface
rather than \code{\link{planes3d}}.  This allows curved
surfaces to be shown.  The default is \code{FALSE} if the model
looks like a simple 2 parameter linear fit, otherwise \code{TRUE}.
  }
  \item{do_grid}{Plot a grid.}
  \item{grid.col, grid.alpha, grid.steps}{
Characteristics of the grid.
  }
  \item{sub.steps}{If \code{use_surface3d} is \code{TRUE}, 
use an internal grid of \code{grid.steps*sub.steps} to draw
the surface.  \code{sub.steps > 1} allows curvature within facets.
Similarly, if \code{do_grid} is \code{TRUE}, it allows 
curvature within grid lines.}
  \item{vars}{A dataframe containing the variables to plot
in the first three columns, with the response assumed to be in
column 1.  See the Note below.}
  \item{clip_to_density}{
If positive, the surface, plane or grid will be clipped to a region with sufficient data.
  }
  \item{\dots}{
Other parameters to pass to the default \code{\link{plot3d}} method, to control the appearance of aspects of the plot
other than the plane.
}
}
\details{
Three plots are possible, depending on the value(s) in \code{which}:
\enumerate{
\item{(default) Show the points and the fitted plane or surface.}
\item{Show the residuals and the plane at \code{z = 0}.}
\item{Show the predicted values on the fitted plane or surface.}
}

If \code{clip_to_density} is positive, then the
surface, plane or grid will be clipped to 
the region where a non-parametric density estimate
(using \code{MASS::\link[MASS]{kde2d}}), normalized to have a
maximum value of 1, is greater than the given value.
This will suppress parts of the plot that aren't 
supported by the observed data.  
}
\value{
Called for the side effect of drawing one or more plots.

Invisibly returns a high-level vector of object ids.  Names of
object ids have the plot number (in drawing order) appended.
}
\note{
The default value for the \code{vars} argument will handle
simple linear models with a response and two predictors, and
some models with functions of those two predictors.  For 
models that fail (e.g. models using \code{\link{poly}}), you
can include the observed values as in the third example below.

If \code{clip_to_density > 0},
\enumerate{
\item{The clipping is
approximate, so it may not agree perfectly between
surfaces, planes and grids.}
\item{This option requires the suggested packages
\pkg{MASS} and \pkg{interp}, and will be ignored with
a warning if either is not installed.}
}
}
\author{
Duncan Murdoch
}
\examples{
open3d()
ids <- plot3d(lm(mpg ~ wt + qsec, data = mtcars), which = 1:3)
names(ids)

open3d()
plot3d(lm(mpg ~ wt + I(wt^2) + qsec, data = mtcars))

open3d()
# Specify vars in the order:  response, pred1, pred2.
plot3d(lm(mpg ~ poly(wt, 3) + qsec, data = mtcars), 
       vars = mtcars[,c("mpg", "wt", "qsec")])
       
open3d()
# Clip parts of the plot with few (wt, qsec) points
plot3d(lm(mpg ~ poly(wt, 3) + qsec, data = mtcars), 
       vars = mtcars[,c("mpg", "wt", "qsec")],
       clip_to_density = 0.1)       
}

\name{plot3d}
\alias{plot3d}
\alias{plot3d.default}
\alias{plot3d.mesh3d}
\title{3D scatterplot}
\description{
Draws a 3D scatterplot.
}
\usage{
plot3d(x, ...)
\method{plot3d}{default}(x, y, z,  
	xlab, ylab, zlab, type = "p", 
	col, size, lwd, radius,
	add = FALSE, aspect = !add, 
	xlim = NULL, ylim = NULL, zlim = NULL, 
	forceClipregion = FALSE, 
	decorate = !add, ...)
\method{plot3d}{mesh3d}(x, xlab = "x", ylab = "y", zlab = "z", type = c("shade", "wire", "dots"),
	add = FALSE, aspect = !add, ...)	
}
\arguments{
  \item{x, y, z}{vectors of points to be plotted. Any reasonable way of defining the
    coordinates is acceptable.  See the function \code{\link[grDevices]{xyz.coords}}
    for details.}
  \item{xlab, ylab, zlab}{labels for the coordinates.}
  \item{type}{For the default method, a single character indicating the type of item to plot.  
  Supported types are: 'p' for points, 's' for spheres, 
  'l' for lines, 'h' for line segments 
  from \code{z = 0}, and 'n' for nothing.  For the \code{mesh3d} method, one of 
  'shade', 'wire', or 'dots'.  Partial matching is used.
  }
  \item{col}{the color to be used for plotted items.}
  \item{size}{the size for plotted points.}
  \item{lwd}{the line width for plotted items.}
  \item{radius}{the radius of spheres:  see Details below.}
  \item{add}{whether to add the points to an existing plot.}
  \item{aspect}{either a logical indicating whether to adjust the aspect ratio, or a new ratio.}
  \item{xlim, ylim, zlim}{If not \code{NULL}, set clipping
limits for the plot.}
  \item{forceClipregion}{Force a clipping region to be used,
whether or not limits are given.}
  \item{decorate}{Whether to add bounding axes and other
  decorations.}
  \item{\dots}{additional parameters which will be passed to \code{\link{par3d}}, \code{\link{material3d}}
  or \code{\link{decorate3d}}.}
}
\value{
  \code{plot3d} is called for the side effect of drawing the plot; a vector
  of object IDs is returned.
}
\details{
\code{plot3d} is a partial 3D analogue of plot.default.

Missing values in the data are skipped, as in standard graphics.

If \code{aspect} is \code{TRUE}, aspect ratios of \code{c(1, 1, 1)} are passed to
\code{\link{aspect3d}}.  If \code{FALSE}, no aspect adjustment is done.  In other
cases, the value is passed to \code{\link{aspect3d}}.

With \code{type = "s"}, spheres are drawn centered at the specified locations.
The radius may be controlled by \code{size} (specifying the size relative
to the plot display, with the default \code{size = 3} giving a radius 
about 1/20 of the plot region) or \code{radius} (specifying it on the data scale
if an isometric aspect ratio is chosen, or on an average scale
if not). 
}
\section{Clipping}{
If any of \code{xlim}, \code{ylim} or \code{zlim} are specified,
they should be length two vectors giving lower and upper 
clipping limits for the corresponding coordinate.  \code{NA}
limits will be ignored.

If any clipping limits are given, then the data will be
plotted in a newly created subscene within the current one;
otherwise plotting will take place directly in the current
subscene.  This subscene is named \code{"clipregion"}
in the results.  This may affect the appearance of transparent
objects if some are drawn in the \code{plot3d} call and others after, 
as RGL will not attempt to depth-sort objects if they are
in different subscenes.  It is best to draw all overlapping 
transparent objects in the same subscene.  See the example 
in \code{\link{planes3d}}.  It will also affect the use
of \code{\link{clipplanes3d}}; clipping planes need to be in the 
same subscene as the objects being clipped.

Use \code{forceClipregion = TRUE} to force creation of this 
subscene even without specifying limits. 
}
\author{
  Duncan Murdoch
}
\seealso{
  \code{\link{plot.default}},  
  \code{\link{open3d}}, \code{\link{par3d}}.
  There are  \code{\link{plot3d.function}} and \code{\link{plot3d.deldir}} methods for plotting surfaces.
}
\examples{
  open3d()
  x <- sort(rnorm(1000))
  y <- rnorm(1000)
  z <- rnorm(1000) + atan2(x, y)
  plot3d(x, y, z, col = rainbow(1000))
}
\keyword{dynamic}

\name{tkrgl}
\alias{tkrgl}
\title{
The former tkrgl package
}
\description{
Functions from the former \pkg{tkrgl} package.
}
\details{
The \pkg{tkrgl} package contained functions to use TCL/TK to control
an RGL scene on screen.  These functions have now been merged
into the \pkg{rgl} package, and the \pkg{tkrgl} package has
been archived.

To avoid conflicts with RGL names and to indicate the TCL/TK 
nature of these functions, they have all been prefixed with \code{tk}:

\describe{
\item{\code{\link{tkpar3dsave}}}{Formerly \code{tkrgl::par3dsave}, allows interactive
saving of scene parameters.}
\item{\code{\link{tkspin3d}, \link{tkspinControl}}}{Formerly 
\code{tkrgl::spin3d} and \code{tkrgl::spinControl}, create buttons
to spin the scene.}
}

History:
\tabular{ll}{
0.2-2 \tab First public release \cr
0.3   \tab Added possibility to control multiple windows \cr
0.4   \tab Compatibility with 2.0.0 tcltk package \cr
0.5   \tab Added continuous rotation \cr
0.6   \tab Added par3dsave \cr
0.7   \tab Added parameters to \code{\link{tkspinControl}}, fixed startup \cr
0.8   \tab Minor fixes to pass checks \cr
0.9   \tab Merge functions into \pkg{rgl} \cr
}
}

\name{par3dinterp}
\alias{par3dinterp}
\title{Interpolator for par3d parameters}
\description{
Returns a function which interpolates \code{par3d} parameter values,
suitable for use in animations.
}
\usage{
par3dinterp(times = NULL, userMatrix, scale, zoom, FOV, 
            method = c("spline", "linear"), 
            extrapolate = c("oscillate", "cycle", "constant", "natural"),
            dev = cur3d(), subscene = par3d("listeners", dev = dev))
}
\arguments{
  \item{times}{ Times at which values are recorded or a list; see below }
  \item{userMatrix}{ Values of \code{par3d("userMatrix")} }
  \item{scale}{ Values of \code{par3d("scale")} }
  \item{zoom}{ Values of \code{par3d("zoom")} }
  \item{FOV}{ Values of \code{par3d("FOV")} }
  \item{method}{ Method of interpolation }
  \item{extrapolate}{ How to extrapolate outside the time range }
  \item{dev}{ Which RGL device to use }
  \item{subscene}{ Which subscene to use }
}
\details{
This function is intended to be used in constructing animations.  It produces
a function that returns a list suitable to pass to \code{\link{par3d}}, to set 
the viewpoint at a given point in time.

All of the parameters are optional.  Only those \code{par3d} parameters that are 
specified will be returned.

The input values other than \code{times} may each be specified as
lists, giving the parameter value settings at a fixed time, or as
matrices or arrays.  If not lists, the following formats should be
used: \code{userMatrix} can be a \code{4 x 4 x n} array, or a \code{4
x 4n} matrix; \code{scale} should be an \code{n x 3} matrix;
\code{zoom} and \code{FOV} should be length \code{n} vectors.

An alternative form of input is to put all of the above arguments into
a list (i.e. a list of lists, or a list of arrays/matrices/vectors),
and pass it as the first argument.  This is the most convenient way to
use this function with the  function
\code{\link{tkpar3dsave}}.

Interpolation is by cubic spline or linear interpolation in an appropriate
coordinate-wise fashion.  Extrapolation may oscillate
(repeat the sequence forward, backward, forward, etc.), cycle (repeat
it forward), be constant (no repetition outside the specified time
range), or be natural (linear on an appropriate scale).   
In the case of cycling, the first and last specified values
should be equal, or the last one will be dropped.  Natural extrapolation
is only supported with spline interpolation.

}
\value{
A function is returned.  The function takes one argument, and returns a list of 
\code{par3d} settings interpolated to that time.
}
\note{
Prior to \pkg{rgl} version 0.95.1476, the
\code{subscene} argument defaulted to the
current subscene, and any additional entries would
be ignored by \code{\link{play3d}}.  The current
default value of \code{par3d("listeners", dev = dev)} means that all subscenes that share 
mouse responses will also share modifications
by this function.
}
\author{Duncan Murdoch }
\seealso{ \code{\link{play3d}} to play the animation. }
\examples{
f <- par3dinterp( zoom = c(1, 2, 3, 1) )
f(0)
f(1)
f(0.5)
\dontrun{
play3d(f)
}
}
\keyword{ dplot }

\name{clipMesh3d}
\alias{clipMesh3d}
\alias{clipObj3d}
\title{
Clip mesh or RGL object to general region
}
\description{
Modifies a mesh3d object so that values of 
a function are bounded.
}
\usage{
clipMesh3d(mesh, fn, bound = 0, greater = TRUE, 
           minVertices = 0, plot = FALSE, keepValues = FALSE,
           keepTags = FALSE)
clipObj3d(ids = tagged3d(tags), fn, bound = 0, greater = TRUE,
           minVertices = 0,
           replace = TRUE, tags)
}
\arguments{
  \item{mesh}{
A \code{\link{mesh3d}} object.
}
  \item{fn}{
A function used to determine clipping, or a vector of values from such
a function, with one value per vertex.
}
  \item{bound}{
The value(s) of \code{fn} on the clipping boundary.
}
  \item{greater}{
Logical; whether to keep \code{fn >= bound} or not.
}
  \item{minVertices}{
See Details below.
  }
  \item{plot}{Logical; whether or not to plot the mesh.}
  \item{keepValues}{Logical; whether to save the function values at 
  each vertex when \code{plot = FALSE}.}
  \item{keepTags}{Whether to keep the \code{"tags"} component
  of the result; see details below.}
  \item{ids}{
The RGL id value(s) of objects to clip.
  }
  \item{tags}{
Object tags; an alternate way to specify \code{ids}.  Ignored if \code{ids} is
given.
  }
  \item{replace}{
Should the \code{ids} objects be deleted after the clipped
ones are drawn?
  }
}
\details{
These functions transform a mesh3d object or other
RGL objects by removing parts where \code{fn} violates
the bound.

For \code{clipMesh3d} the \code{fn} argument can be any
of the following:
\itemize{
\item{a character vector naming a function (with special
names \code{"x"}, \code{"y"}, and \code{"z"} corresponding
to functions returning those coordinates)}
\item{a function}
\item{a numeric vector with one value per vertex}
\item{\code{NULL}, indicating that the numeric values
are saved in \code{mesh$values}}
}
For \code{clipObj3d} any of the above except \code{NULL}
may be used.

If \code{fn}
is a numeric vector, with one value per vertex, those values will be
used in the test.
If it is a function with formal arguments \code{x},
\code{y} and \code{z}, it will receive the coordinates of 
vertices in those arguments, otherwise it will receive 
the coordinates in a single n x 3 matrix.   The function
should be vectorized and return one value per vertex, 
to check against the bound.

These operations are performed on the mesh:

First, all quads are converted to triangles.

Next, each vertex is checked against the condition.

Modifications to triangles depend
on how many of the vertices satisfy the condition 
(\code{fn >= bound} or \code{fn <= bound}, depending on \code{greater})
for inclusion.
\itemize{
\item If no vertices in a triangle satisfy the condition, the triangle is omitted.
\item If one vertex satisfies the condition, the other two vertices
in that triangle are shrunk towards it by assuming \code{fn}
is locally linear.
\item If two vertices satisfy the condition, the third vertex
is shrunk along each edge towards each other vertex, forming a
quadrilateral made of two new triangles.
\item If all vertices satisfy the condition, they are included
with no modifications.
}
Modifications to line segments are similar:  the segment
will be shortened if it crosses the boundary, or omitted
if it is entirely out of bounds.  Points, spheres, text
and sprites will just be kept or rejected.

The \code{minVertices} argument is used to improve the 
approximation to the boundary when \code{fn} is a non-linear
function.  In that case, the interpolation described above
can be inaccurate.  If \code{minVertices} is set to a
positive
number (e.g. \code{10000}), then each object is modified
by subdivision to have at least that number of vertices, 
so that pieces are smaller and the linear interpolation
is more accurate.  In the \code{clipObj3d} function, 
\code{minVertices} can be a vector, with entries corresponding
to each of the entries in \code{ids}.
}
\section{The \code{keepTags} argument}{

If \code{keepTags = TRUE}, a \code{"tags"} element will be 
added to the result.  It will be a vector with one entry per
point, segment, triangle and quad in the output mesh.  (These tags are not related
to the tags used to identify \pkg{rgl} objects.)
The mesh tags may be used to show the correspondence between
the parts of 
the input mesh and output mesh.
By default, the tags are constructed as a numerical sequence
over points,
segments, triangles and
quads in the input mesh, in that order, starting from one.  This is the same order 
used for colours when 
shading with 
\code{meshColor == "faces"}.  

For example, start with a mesh with one point, two segments, three
triangles and four
quads, but no \code{tags} member.  
It would implicitly tag the parts from one to ten as

\verb{
  c(1,    # the point
    2:3,  # the two segments
    4:6,  # the three triangles
    7:10) # the four quads
}

If clipping deleted the segments
and the first triangle, the output would contain the seven
element result

\verb{
  mesh$tags <- c(1,    # the point remains
                       # no segments now
                 5:6,  # the two remaining triangles
                       # were previously items 5 and 6
                 7:10) # the four quads
}

The \code{tags} output may contain repetitions. For example,
when a triangle is partially clipped and replaced by several
smaller triangles, entries for all of them will contain the value
corresponding to the original triangle.

The \code{mesh$tags} component may be supplied as part of the input
mesh as any type of vector; 
the output will propagate values to the new mesh.  The input length
must match the total number of points, segments, triangles and 
quads in the input mesh or an error will be raised.
}

\value{
If \code{plot = FALSE},
\code{clipMesh3d} returns new mesh3d object in which all vertices (approximately) satisfy the
clipping condition.  Note that the order of vertices will likely
differ from the original order, and new vertices will be added near
the boundary (and if \code{minVertices > 0}, in the 
interior).  If in addition \code{keepValues = TRUE},
a component named \code{"values"} will be added to the 
mesh containing the values for each vertex.
If \code{keepTags = TRUE}, the \code{tags} component
described below will be added to the output mesh.

If \code{plot = TRUE}, the result will be 
plotted with \code{\link{shade3d}} and its result returned.

\code{clipObj3d} is called for the side effect of modifying
the scene.  It returns a list of new RGL id values
corresponding to the \code{ids} passed as arguments.
}
\author{
Duncan Murdoch
}
\references{
See \url{https://stackoverflow.com/q/56242470/2554330} and
\url{https://laustep.github.io/stlahblog/posts/MeshClipping.html} for a
motivating example.
}
\seealso{
See \code{\link{contourLines3d}} and
\code{\link{filledContour3d}} for ways to display function 
values without clipping.
}
\examples{

# Show the problem that minVertices solves:

cube <- cube3d(col = rainbow(6), meshColor = "faces")

# This function only has one argument, so it will 
# be passed x, y and z in columns of a matrix
vecnorm <- function(vals) apply(vals, 1, function(row) sqrt(sum(row^2)))

open3d()
mfrow3d(2, 2, sharedMouse = TRUE)
id1 <- shade3d(cube)
# All vertices have norm sqrt(3), so this clips nothing:
clipObj3d(id1, fn = vecnorm, bound = sqrt(2))
next3d()
id2 <- wire3d(cube, lit = FALSE)
clipObj3d(id2, fn = vecnorm, bound = sqrt(2))

# This subdivides the cube, and does proper clipping:
next3d()
id3 <- shade3d(cube)
clipObj3d(id3, fn = vecnorm, bound = sqrt(2), minVertices = 200)
next3d()
id4 <- wire3d(cube, lit = FALSE)
clipObj3d(id4, fn = vecnorm, bound = sqrt(2), minVertices = 200)

}

\name{as.mesh3d}
\alias{as.mesh3d}
\alias{as.mesh3d.default}
\title{
Convert object to mesh object
}
\description{
The \code{as.mesh3d} generic function converts various objects
to \code{\link{mesh3d}} objects.  

The default method takes takes a matrix of vertices 
as input and (optionally) merges repeated vertices, producing a \code{\link{mesh3d}}
object as output.  It will contain either triangles or quads or segments or points
according to the \code{type} argument.

If the generic is called without any argument, it will pass
all RGL ids from the current scene to the 
\code{\link{as.mesh3d.rglId}} method.
}
\usage{
as.mesh3d(x, ...)
\method{as.mesh3d}{default}(x, y = NULL, z = NULL, 
             type = c("triangles", "quads", "segments", "points"),
             smooth = FALSE, 
             tolerance = sqrt(.Machine$double.eps), 
             notEqual = NULL, 
             merge = TRUE,
             ...,
             triangles)
}
\arguments{
  \item{x, y, z}{
For the generic, \code{x} is the object to convert.
For the default method, \code{x}, \code{y} and \code{z} are
coordinates. Any reasonable way of defining the coordinates is
acceptable. See the function \code{\link{xyz.coords}} for details.
}
  \item{type}{
What type of things should be in the mesh?  Tries
this list in order until it finds one that works.
  }
  \item{smooth}{
If \code{TRUE}, \code{\link{addNormals}} will be called on the mesh
object to make it render smoothly.
}  
  \item{tolerance}{
The numerical tolerance to be used in \code{\link{all.equal}} to
determine whether two vertices should be merged.
}
  \item{notEqual}{
If not \code{NULL}, an n by n matrix of logical values, where n is the
number of vertices as input.  \code{TRUE} entries indicate that the
corresponding pair of vertices should not be merged even if they appear equal.
}
  \item{merge}{
Should apparently equal vertices be merged?
  }
  \item{\dots}{
Material properties to pass to \code{\link{tmesh3d}} or \code{\link{qmesh3d}}.
}
  \item{triangles}{
Deprecated.  If present, \code{TRUE} indicates \code{type = "triangles"} and \code{FALSE} indicates \code{type = "quads"}.
}
}
\details{
The motivation for this function is the following problem:  I was
asked whether RGL could render a surface made up of triangles or 
quadrilaterals to look smooth.  It can do that, but needs normals at 
each vertex; they should be the average of the normals for each polygon 
sharing that vertex.  Then OpenGL will interpolate the normals across 
the polygons and give the illusion of smoothness.

To do this, it needs to know which polygons share each vertex.  If the 
surface is described as a list of triangles or quadrilaterals, that 
means identifying vertices that are in multiple polygons, and converting 
the representation to a \code{"\link{mesh3d}"} object (which is a matrix of vertices 
and a matrix of vertex numbers making up triangles or quads).  Then the 
\code{\link{addNormals}} function will add the normals.

Sometimes two polygons will share vertices (within numerical 
tolerance) without the user wanting them to be considered internal to 
the surface, or might want one sharp edge in an otherwise smooth 
surface.  This means I needed a way to declare that two vertices from 
the original list of vertices in the triangles or quads are "not equal", 
even when they test numerically equal.  That's what the \code{notEqual} matrix specifies.
}
\value{
A \code{"\link{mesh3d}"} object with the same faces as in the 
input, but (if \code{merge=TRUE}) with vertices that test equal to
within \code{tolerance} merged.  
}
\author{
Duncan Murdoch
}
\examples{
xyz <- matrix(c(-1, -1, -1,
                -1,  1, -1,
                 1,  1, -1,
                 1, -1, -1,
                -1,  1, -1,
                -1,  1,  1,
                 1,  1,  1,
                 1,  1, -1,
                 1, -1, -1,
                 1,  1, -1,
                 1,  1,  1,
                 1, -1,  1), byrow = TRUE, ncol = 3)
mesh <- as.mesh3d(xyz, type = "quads", col = "red")
mesh$vb
mesh$ib
open3d()
shade3d(mesh)

# Stop vertices 2 and 5 from being merged
notEQ <- matrix(FALSE, 12, 12)
notEQ[2, 5] <- TRUE
mesh <- as.mesh3d(xyz, type = "quads", notEqual = notEQ)
mesh$vb
mesh$ib
}

\name{surface3d}
\title{Add surface}
\alias{surface3d}
\description{
  Adds a surface to the current scene. The surface is defined by 
  a matrix defining the height of each grid point and two vectors or matrices
  defining the grid.
}
\usage{
surface3d(x, y, z, ..., 
         normal_x = NULL, normal_y = NULL, normal_z = NULL,
         texture_s=NULL, texture_t=NULL, flip = FALSE)
}
\arguments{
  \item{ x, y, z }{
    vectors or matrices of values.  See Details.
  }
  \item{ ... }{Material properties. See \code{\link{material3d}} for details.}    
  \item{normal_x, normal_y, normal_z}{
    matrices giving the coordinates of normals at 
    each grid point
  }
  \item{texture_s, texture_t}{
    matrices giving the texture coordinates at each
    grid point
  }
  \item{flip}{
    flip definition of \dQuote{up}
  }
}
\details{
Adds a surface mesh to the current scene. The surface is 
typically defined by a matrix of height values in \code{z}
(as in \code{\link{persp}}), 
but any of \code{x}, \code{y}, or \code{z} may be matrices or
vectors, as long as at least one is a matrix.  (One 
historical exception is allowed:  if all are vectors but
the length of \code{z} is the product of the lengths of
\code{x} and \code{y}, \code{z} is converted to a matrix.)

Dimensions of all matrices must match.

If any of the coordinates are vectors, they are interpreted as follows:
\itemize{
\item If \code{x} is a vector, it corresponds to rows of the matrix.
\item If \code{y} is a vector, it corresponds to columns
of the matrix.
\item If \code{z} is a vector, it corresponds to columns
unless \code{y} is also a vector, in which case it corresponds
to rows.
}

  If the normals are not supplied, they will be calculated automatically based
  on neighbouring points.

  Texture coordinates run from 0 to 1 over each dimension of the texture bitmap.
  If texture coordinates are not supplied, they will be calculated to 
  render the texture exactly once over the grid.  Values greater than 1 can be
  used to repeat the texture over the surface.
  
  \code{surface3d} always tries to draw the surface with the `front' upwards
  (typically towards higher \code{z} values).  This can be used to render
  the top and bottom differently; see \code{\link{material3d}} and
  the example below.  If you don't like its choice, set
  \code{flip = TRUE} to use the opposition definition.
  
  \code{NA} values in the height matrix are not drawn.
  
}
\examples{
#
# volcano example taken from "persp"
#

z <- 2 * volcano        # Exaggerate the relief

x <- 10 * (1:nrow(z))   # 10 meter spacing (S to N)
y <- 10 * (1:ncol(z))   # 10 meter spacing (E to W)

zlim <- range(z)
zlen <- zlim[2] - zlim[1] + 1

colorlut <- terrain.colors(zlen) # height color lookup table

col <- colorlut[ z - zlim[1] + 1 ] # assign colors to heights for each point

open3d()
surface3d(x, y, z, color = col, back = "lines")

}
\seealso{
  See \code{\link{persp3d}} for a higher level interface.
}
\keyword{dynamic}

\name{in_pkgdown_example}
\alias{in_pkgdown_example}
\title{
Are we running in a \pkg{pkgdown} example?
}
\description{
This is mainly for internal use to decide whether results
should be automatically included in a \pkg{pkgdown} web
page.
}
\usage{
in_pkgdown_example()
}
\value{
\code{TRUE} or \code{FALSE}
}

\examples{
in_pkgdown_example()
}

\name{bbox3d}
\alias{bbox3d}
\title{Set up bounding box decoration}
\description{
  Set up the bounding box decoration.
}
\usage{
bbox3d(xat = NULL, yat = NULL, zat = NULL, 
	xunit = "pretty", yunit = "pretty", zunit = "pretty", 
	expand = 1.03,
	draw_front = FALSE, 
	xlab=NULL, ylab=NULL, zlab=NULL,
  xlen=5, ylen=5, zlen=5,
  marklen=15.0, marklen.rel=TRUE, ...) 
}
\arguments{
  \item{xat, yat, zat}{vector specifying the tickmark positions}
  \item{xlab, ylab, zlab}{character vector specifying the tickmark labeling}
  \item{xunit, yunit, zunit}{value specifying the tick mark base for uniform tick mark layout}
  \item{xlen, ylen, zlen}{value specifying the number of tickmarks}
  \item{marklen}{value specifying the length of the tickmarks}
  \item{marklen.rel}{logical, if TRUE tick mark length is calculated using 1/\code{marklen} * axis length, otherwise tick mark length is \code{marklen} in coordinate space}
  \item{expand}{value specifying how much to expand the bounding box around the data}
  \item{draw_front}{draw the front faces of the bounding box}
  \item{ ... }{Material properties (or other \code{rgl.bbox} parameters
  in the case of \code{bbox3d}). See \code{\link{material3d}} for details.}
}
\details{
  Four different types of tick mark layouts are possible.
This description applies to the X axis; other axes are similar:
  If \code{xat} is not \code{NULL}, the ticks are set up at custom positions.
  If \code{xunit} is numeric but not zero, it defines the tick mark base.  If
  it is \code{"pretty"} (the default in \code{bbox3d}), ticks are set at \code{\link{pretty}}
  locations.
  If \code{xlen} is not zero, it specifies the number of ticks
  (a suggestion if \code{xunit} is \code{"pretty"}).

  The first color specifies the bounding box, while the second one specifies the tick mark and font color.
  
  \code{bbox3d} defaults to \code{\link{pretty}} locations for the axis labels and a slightly larger
  box, whereas \code{rgl.bbox} covers the exact range.
  
  \code{\link{axes3d}} offers more flexibility in the specification of the axes, but 
  they are static, unlike those drawn by \code{\link{bbox3d}}.
}
\value{
  This function is called for the side effect of setting the bounding box decoration.
  A shape ID is returned to allow \code{\link{pop3d}} to delete it.
}
\examples{
  open3d()
  points3d(rnorm(100), rnorm(100), rnorm(100))
  bbox3d(color = c("#333377", "black"), emission = "#333377", 
         specular = "#3333FF", shininess = 5, alpha = 0.8)
}
\seealso{
  \code{\link{material3d}}, \code{\link{axes3d}} 
}
\keyword{dynamic}

\name{figWidth}
\alias{figWidth}
\alias{figHeight}
\title{
Get R Markdown figure dimensions in pixels
}
\description{
In an R Markdown document, figure dimensions are normally
specified in inches; these are translated into pixel
dimensions when HTML output is requested and \code{\link{rglwidget}} is used.  These functions reproduce that
translation.
}
\usage{
figWidth()
figHeight()
}
\value{
When used in an R Markdown document, these functions
return the requested current dimensions of figures in pixels.  Outside such a document, \code{NULL}
is returned.
}
\author{
Duncan Murdoch
}

\examples{
# No useful return value outside of R Markdown:
figWidth()
figHeight()
}

\name{cube3d}
\alias{cube3d}
\alias{oh3d}
\alias{tetrahedron3d}
\alias{octahedron3d}
\alias{icosahedron3d}
\alias{dodecahedron3d}
\alias{cuboctahedron3d}

\title{Sample 3D mesh objects}
\description{
A collection of sample mesh objects.
}
\usage{
  cube3d(trans = identityMatrix(), ...)  
  tetrahedron3d(trans = identityMatrix(), ...)  
  octahedron3d(trans = identityMatrix(), ...)  
  icosahedron3d(trans = identityMatrix(), ...)
  dodecahedron3d(trans = identityMatrix(), ...)
  cuboctahedron3d(trans = identityMatrix(), ...)
  
  oh3d(trans = identityMatrix(), ...)    # an 'o' object

}

\arguments{
  \item{trans}{transformation to apply to objects}
  \item{...}{additional parameters to pass to \code{\link{mesh3d}}}
}
\details{

  These sample objects optionally take a 4x4 matrix transformation \code{trans} as 
  an argument.  This transformation is applied to all vertices of the default shape.
  The default is an identity transformation. 
}
\value{
Objects of class \code{c("mesh3d",
  "shape3d")}. 
}
\examples{
  # render all of the Platonic solids
  open3d()
  shade3d( translate3d( tetrahedron3d(col = "red"), 0, 0, 0) )
  shade3d( translate3d( cube3d(col = "green"), 3, 0, 0) )
  shade3d( translate3d( octahedron3d(col = "blue"), 6, 0, 0) )
  shade3d( translate3d( dodecahedron3d(col = "cyan"), 9, 0, 0) )
  shade3d( translate3d( icosahedron3d(col = "magenta"), 12, 0, 0) )
}
\seealso{
  \code{\link{mesh3d}}
}
\keyword{dynamic}

\name{text3d}
\alias{text3d}
\alias{texts3d}
\title{Add text to plot}
\description{
  Adds text to the scene. The text is positioned in 3D space. 
  Text is always oriented towards the camera.
}
\usage{
text3d(x, y = NULL, z = NULL, texts, adj = 0.5, pos = NULL, offset = 0.5,
       usePlotmath = is.language(texts), 
       family = par3d("family"), font = par3d("font"), 
			 cex = par3d("cex"), useFreeType = par3d("useFreeType"),
			...) 
texts3d(x, y = NULL, z = NULL, texts, adj = 0.5, pos = NULL, offset = 0.5,
       usePlotmath = is.language(texts),
       family = par3d("family"), font = par3d("font"), 
			 cex = par3d("cex"), useFreeType = par3d("useFreeType"),
			...)       
}
\arguments{
  \item{x, y, z}{point coordinates.  Any reasonable way of defining the
    coordinates is acceptable.  See the function \code{\link[grDevices]{xyz.coords}}
    for details.}
  \item{texts}{text character vector to draw}
  \item{adj}{ one value specifying the horizontal adjustment, or two, 
              specifying horizontal and vertical adjustment respectively, or three, specifying adjustment in all
              three directions.}
  \item{pos}{ a position specifier for the text.  If specified, this
  overrides any \code{adj} value given. Values of 0, 1, 2, 3, 4, 5, 6 respectively indicate positions at, below, to the left of, above, to the right of, in front of, and behind the specified coordinates.}
  \item{offset}{ when \code{pos} is specified, this value gives the offset of the label from the specified coordinate in fractions of a character width.}
  \item{ usePlotmath }{logical.  Should \code{\link{plotmath3d}} be used for the text?}
  \item{ family }{A device-independent font family name, or "" }
  \item{ font }{A numeric font number from 1 to 4 }
  \item{ cex }{A numeric character expansion value }
  \item{ useFreeType }{logical.  Should FreeType be used to draw text? (See details below.)}
  \item{ ... }{Material properties; see \code{\link{material3d}} for details.} 
}
\details{
The \code{adj} parameter determines the position of the text relative to the 
specified coordinate.  Use \code{adj = c(0, 0)} to place the left bottom corner at
\code{(x, y, z)}, \code{adj = c(0.5, 0.5)} to center the text there, and \code{adj = c(1, 1)}
to put the right top corner there. The optional second coordinate for vertical 
adjustment defaults to \code{0.5}.  Placement is done using the "advance" of
the string and the "ascent" of the font relative to the baseline, when these metrics
are known.

\code{text3d} and \code{texts3d} draw text using the \link{r3d}
conventions. These are synonyms; the former is singular to be
consistent with the classic 2-D graphics functions, and the latter is
plural to be consistent with all the other graphics primitives.  Take
your choice!

If any coordinate or text is \code{NA}, that text is not plotted.

If \code{usePlotmath} is \code{TRUE}, the work will be done
by the \code{\link{plotmath3d}} function.  This is the default
if the \code{texts}
parameter is \dQuote{language}, e.g. the result of
a call to \code{\link{expression}} or \code{\link{quote}}.
}
\section{Fonts}{
Fonts are specified using the \code{family}, \code{font}, \code{cex}, 
and \code{useFreeType} arguments.  Defaults for the currently active
device may be set using \code{\link{par3d}}, or for future devices
using \code{\link{r3dDefaults}}.

The \code{family} specification is the same as for standard graphics, i.e.
families \cr
\code{c("serif", "sans", "mono", "symbol")} \cr
are normally
available, but users may add additional families.  \code{font} numbers
are restricted to the range 1 to 4 for standard, bold, italic and bold italic
respectively.  Font 5 is recoded as family \code{"symbol"}
font 1, but that is not supported unless specifically
installed, so should be avoided.

Using an unrecognized value for \code{"family"} will result in
the system standard font as used in RGL up to version 0.76.  That font
is not resizable and \code{font} values are ignored.

If \code{useFreeType} is \code{TRUE}, then RGL will use the FreeType
anti-aliased fonts for drawing.  This is generally desirable, and it is the
default on non-Windows systems if RGL was built to support FreeType.

FreeType fonts are specified using the \code{\link{rglFonts}} function. 
}
\value{
  The text drawing functions return the object ID of the text object (or sprites, in case of \code{usePlotmath = TRUE})
invisibly.
}
\examples{
open3d()
famnum <- rep(1:3, 8)
family <- c("serif", "sans", "mono")[famnum]
font <- rep(rep(1:4, each = 3), 2)
cex <- rep(1:2, each = 12)
text3d(font, cex, famnum, texts = paste(family, font), adj = 0.5, 
       color = "blue", family = family, font = font, cex = cex)
}
\seealso{
  \code{\link{r3d}}, \code{\link{plotmath3d}}, \code{\link{rglExtrafonts}} for adding fonts
}
\keyword{dynamic}

\name{rgl.attrib.info}
\alias{rgl.attrib.info}
\alias{rgl.attrib.count}
\title{
Get information about attributes of objects
}
\description{
These functions give information about the attributes of
RGL objects.  \code{rgl.attrib.info} is the more 
\dQuote{user-friendly} function; \code{rgl.attrib.count} is a 
lower-level function more likely to be used in programming.
}
\usage{
rgl.attrib.info(id = ids3d("all", 0)$id, attribs = NULL, showAll = FALSE)
rgl.attrib.count(id, attrib)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{id}{
One or more RGL object ids.
}
  \item{attribs}{
A character vector of one or more attribute names.
}
  \item{showAll}{
Should attributes with zero entries be shown?
}
  \item{attrib}{
A single attribute name.
}
}
\details{
See the first example below to get the full list of attribute
names.
}
\value{
A dataframe containing the following columns:
\item{id}{The id of the object.}
\item{attrib}{The full name of the attribute.}
\item{nrow, ncol}{The size of matrix that would be returned
by \code{\link{rgl.attrib}} for this attribute.}
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{rgl.attrib}} to obtain the attribute values.
}
\examples{
open3d()
id <- points3d(rnorm(100), rnorm(100), rnorm(100), col = "green")
rgl.attrib.info(id, showAll = TRUE)
rgl.attrib.count(id, "vertices")

merge(rgl.attrib.info(), ids3d("all"))
}

\name{bg3d}
\alias{bg3d}
\title{Set up background}
\description{
  Set up the background of the scene.
}
\usage{
bg3d(color,
    sphere=FALSE, 
    back="lines",
    fogtype="none",
    fogScale = 1, 
    col, ...)

}
\arguments{
  \item{color, col}{
  See Details below.
  }
  \item{sphere}{
  logical:  if \code{TRUE}, an environmental sphere geometry is used for the background decoration.
  }
  \item{back}{
  Specifies the fill style of the sphere geometry. See \code{\link{material3d}} for details.
  }
  \item{fogtype}{fog type:
    \describe{
      \item{"none"}{no fog}
      \item{"linear"}{linear fog function}
      \item{"exp"}{exponential fog function}
      \item{"exp2"}{squared exponential fog function}
    }
    Fog only applies to objects with \code{\link{material3d}} 
    property \code{fog} set to \code{TRUE}.
  }
  \item{fogScale}{
  Scaling for fog.  See Details.
  }
  \item{ ... }{Additional material properties. See \code{\link{material3d}} for details.}  
}
\details{

The background color is taken from \code{color} or \code{col}
if \code{color} is missing.
The first entry
is used for background clearing and as the fog color.
The second (if present) is used for background sphere geometry.

If \code{color} and \code{col} are both missing, the default is found in 
the \code{\link{r3dDefaults}$bg} list, or \code{"white"} is used
if nothing is specified there.

If \code{sphere} is set to \code{TRUE}, an environmental
sphere enclosing the whole scene is drawn.

If not, but the material properties include a bitmap as a texture, the bitmap is drawn in the 
background of the scene.  (The bitmap colors modify the general color setting.)

If neither a sphere nor a bitmap background is drawn, the background is filled with a 
solid color.

The \code{fogScale} parameter should be a positive value
to change the density of the fog in the plot.  For 
\code{fogtype = "linear"} it multiplies the density of the
fog; for the exponential fog types it multiplies the density
parameter used in the display.  

See 
\href{https://registry.khronos.org/OpenGL-Refpages/gl2.1/xhtml/glFog.xml}{the OpenGL 2.1 reference} 
for the formulas used in the fog calculations within \R (though the
\code{"exp2"} formula appears to be wrong, at least on my
system).  In WebGL displays,
the following rules are used.  They appear to match the
rules used in \R on my system.
\itemize{
\item{For \code{"linear"} fog, the near clipping plane is
taken as \eqn{c=0}, and the 
far clipping plane is taken as \eqn{c=1}.  The
amount of fog is \eqn{s * c} clamped to a 0 to 1
range, where \eqn{s = fogScale}.}
\item{For \code{"exp"} and \code{"exp2"} fog, the observer location
is negative at a distance depending on the field of view.
The formula for the distance is 
\deqn{c = [1-sin(theta)]/[1 + sin(theta)]}
where \eqn{theta = FOV/2}.
We calculate \deqn{c' = d(1-c) + c}
so \eqn{c'} runs from 0 at the observer to
1 at the far clipping plane.}
\item{For \code{"exp"} fog, the amount of fog is 
\eqn{1 - exp(-s * c')}.}
\item{For \code{"exp2"} fog, the amount of fog is
\eqn{1 - exp[-(s * c')^2]}.}
}
}
\examples{
  open3d()
  
  # a simple white background
  
  bg3d("white")

  # the holo-globe (inspired by star trek):

  bg3d(sphere = TRUE, color = c("black", "green"), lit = FALSE, back = "lines" )

  # an environmental sphere with a nice texture.

  bg3d(sphere = TRUE, texture = system.file("textures/sunsleep.png", package = "rgl"), 
         back = "filled" )
         
  # The same texture as a fixed background
  
  open3d()
  bg3d(texture = system.file("textures/sunsleep.png", package = "rgl"), col = "white")
}
\seealso{
  \code{\link{material3d}}, \code{\link{bgplot3d}} to add a 2D plot as background.
}
\keyword{dynamic}

\name{rglShared}
\alias{rglShared}
\title{
Create shared data from an RGL object
}
\description{
The \pkg{crosstalk} package provides a way for
different parts of an interactive display to 
communicate about datasets, using \dQuote{shared data} objects.  When selection or filtering
is performed in one view, the result is mirrored 
in all other views.

This function allows vertices of RGL objects
to be treated as shared data.
}
\usage{
rglShared(id, key = NULL, group = NULL,
          deselectedFade = 0.1, 
          deselectedColor = NULL,
	  selectedColor = NULL,
          selectedIgnoreNone = TRUE,
	  filteredFade = 0,
          filteredColor = NULL)
}
\arguments{
  \item{id}{
An existing RGL id.
}
  \item{key}{
Optional unique labels to apply to each vertex.
If missing, numerical keys will be used.
}
  \item{group}{
Optional name of the shared group to which
this data belongs.  If missing, a random name
will be generated.
}
  \item{deselectedFade, deselectedColor}{
Appearance of points that are not selected. See Details.
}
  \item{selectedColor}{
Appearance of points that are selected.
}
  \item{selectedIgnoreNone}{
If no points are selected, should the points be shown
in their original colors (\code{TRUE}), or in the 
deselected colors (\code{FALSE})?
}
  \item{filteredFade, filteredColor}{
Appearance of points that have been filtered out.
}
}
\details{
Some functions which normally work on 
dataframe-like datasets will accept shared data
objects in their place. 

If a selection is in progress, the alpha value for
unselected points is multiplied by \code{deselectedFade}.
If \code{deselectedColor} is \code{NULL}, the color is left
as originally specified; if not, the point is changed to
the color given by \code{deselectedColor}.

If no points have been selected, then by default points
are shown in their original colors.  However, if 
\code{selectedIgnoreNone = FALSE}, all points are displayed
as if unselected.

The \code{selectedColor} argument is similarly used to 
change the color (or not) of selected points, and \code{filteredFade} 
and \code{filteredColor} are used for points that
have been filtered out of the display.
}
\value{
An object of class \code{"SharedData"} (from the 
optional \pkg{crosstalk} package) which 
contains the x, y and z coordinates of the RGL object 
with the given \code{id}.
}
\references{
\url{https://rstudio.github.io/crosstalk/index.html}
}
\author{
Duncan Murdoch
}
\seealso{
The \HTMLVignette{WebGL}{}{User Interaction in WebGL} vignette
gives more details.
}

\examples{
  save <- options(rgl.useNULL = TRUE)
  
  #  rglShared requires the crosstalk package,
  #  and the slider and rglMouse require manipulateWidget
  
  if (requireNamespace("crosstalk", quietly = TRUE) &&
      requireNamespace("manipulateWidget", quietly = TRUE)) {
    open3d()
    x <- sort(rnorm(100))
    y <- rnorm(100)
    z <- rnorm(100) + atan2(x, y)
    ids <- plot3d(x, y, z, col = rainbow(100))

    # The data will be selected and filtered, not the axes.
    sharedData <- rglShared(ids["data"])
  
    # Also add some labels that are only displayed
    # when points are selected
  
    sharedLabel <- rglShared(text3d(x, y, z, text = 1:100,
                                    adj = -0.5),
                             group = sharedData$groupName(),
                             deselectedFade = 0,
                             selectedIgnoreNone = FALSE) 
    if (interactive() || in_pkgdown_example()) 
      crosstalk::filter_slider("x", "x", sharedData, ~x) \%>\%
      rglwidget(shared = list(sharedData, sharedLabel), controller = .) \%>\% 
      rglMouse()
  }    
  options(save)
}

\name{as.tmesh3d}
\alias{as.tmesh3d}
\alias{as.tmesh3d.default}
\alias{as.tmesh3d.mesh3d}
\title{
Convert object to a triangular mesh
}
\description{
Converts the quads in a mesh version of an object to
triangles by splitting them up.  Optionally drops any
point or segment components.
}
\usage{
as.tmesh3d(x, ...)
\method{as.tmesh3d}{default}(x, drop = FALSE, ...)
\method{as.tmesh3d}{mesh3d}(x, drop = FALSE, ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{
An object from which to create a triangular mesh object.
}
  \item{drop}{
If \code{TRUE}, drop any point or segment components.
  }
  \item{\dots}{
Ignored in the \code{mesh3d} method, passed to 
\code{as.mesh3d} in the default method.
}
}
\details{
The default method simply calls \code{\link{as.mesh3d}(x, ...)}
and passes the result to the \code{"mesh3d"} method.
}
\note{
Older versions of \pkg{rgl} had a \code{"tmesh3d"} class
for meshes of triangles.  That class is no longer used:
\code{as.tmesh3d} and \code{\link{tmesh3d}} both produce
\code{"mesh3d"} objects.
}
\value{
A \code{"mesh3d"} object containing no quads.  If \code{drop = TRUE}, it will only contain triangles.
}
\author{
Duncan Murdoch
}

\seealso{
\code{as.triangles3d} to get just the coordinates.
}
\examples{
x <- cuboctahedron3d()
x             # has quads and triangles
as.tmesh3d(x) # has only triangles
}

\name{propertySetter}
\alias{propertySlider}
\alias{propertySetter}
\alias{par3dinterpSetter}
\alias{matrixSetter}
\alias{vertexSetter}
\title{
Obsolete function to write HTML/Javascript code to control a WebGL display.
}
\description{ These functions write 
out HTML code to control WebGL displays based
using the obsolete \code{writeWebGL}.  Use
\code{\link{playwidget}} and related functions instead.
}
\usage{
propertySlider(setter = propertySetter,
             minS = NULL, maxS = NULL, step = 1, init = NULL, 
	     labels, 
	     id = basename(tempfile("input")), name = id,
	     outputid = paste0(id, "text"),
	     index = NULL,
	     ...) 
	     
propertySetter(values = NULL, entries, properties, objids, prefixes = "",
               param = seq_len(NROW(values)), interp = TRUE, digits = 7) 
               
par3dinterpSetter(fn, from, to, steps, subscene, omitConstant = TRUE, 
                  rename = character(), ...)

matrixSetter(fns, from, to, steps, subscene = currentSubscene3d(), 
             matrix = "userMatrix", omitConstant = TRUE, prefix = "", ...) 
             
vertexSetter(values, vertices = 1, attributes, objid, prefix = "",
	     param = seq_len(NROW(values)), interp = TRUE,
	     digits = 7)
}
\arguments{
  \item{setter}{A function to write Javascript code, or its output, or
   a list containing several of these.}
  \item{minS, maxS, step, init}{Slider values to be displayed. Reasonable defaults are used if missing.}
  \item{labels}{Labels to display for each slider value.  The
  defaults are calculated using internal variables.  If \code{NULL},
  no labels will be shown.}
  \item{id}{The \code{id} of the input control that will be generated.}
  \item{name}{The name of the input control that will be generated.}
  \item{outputid}{The \code{id} of the output control that
will display the slider value, or \code{NULL} for none.}
  \item{index}{The 1-based index of this slider:  it controls the
  corresponding entry in an indexed setter such as \code{matrixSetter}.}
  \item{...}{Other parameters.}
  \item{values}{An array of values; rows correspond to slider positions.  Alternatively, \code{NULL}; the generated function takes a
  single value or array of values and applies them directly.}
  \item{entries, properties, objids, prefixes}{Vectors describing
  the columns of \code{values}.}
  \item{param}{Parameter values corresponding to each row of \code{values}.}
  \item{interp}{Whether to interpolate values.  If \code{FALSE}, 
  the Javascript function will expect non-negative integer values.
  Ignored if \code{values} is \code{NULL}.}
  \item{digits}{How many significant digits to emit in the Javascript code.}
  \item{fn}{A function returned from \code{\link{par3dinterp}}.}
  \item{from, to, steps}{Values where \code{fn} should be evaluated.}
  \item{subscene}{Which subscene's properties should be modified?}
  \item{omitConstant}{If \code{TRUE}, do not set values that are
  constant across the range.}
  \item{rename}{A named character vector of names of Javascript properties
  to modify.}
  \item{fns}{A list containing functions returned from 
    \code{\link{par3dinterp}}.}
  \item{matrix}{A character string giving the Javascript property
    name of the matrix to modify.}
  \item{prefix}{The prefix of the scene containing \code{matrix}.}
  \item{vertices}{A vector of vertex numbers (1-based) within 
  an object.}
  \item{attributes}{A vector of attributes of a vertex, 
  from \code{c("x", "y", "z", "r", "g", "b", "a", "nx", "ny", "nz",
               "radius", "ox", "oy", "oz", "ts", "tt")}.}
  \item{objid}{The object containing the vertices to be modified.}
}
\author{
Duncan Murdoch
}

\seealso{
\code{\link{rglwidget}}, \code{\link{playwidget}}
}

\name{rgl.postscript}
\alias{rgl.postscript}
\title{Export vector graphics}
\description{
  Saves the screenshot to a file in PostScript or other vector graphics format.
}
\usage{
rgl.postscript( filename, fmt = "eps", drawText = TRUE )
}
\arguments{
  \item{filename}{full path to filename.}
  \item{fmt}{export format, currently supported: ps, eps, tex, pdf, svg, pgf }
  \item{drawText}{logical, whether to draw text}
}
\details{
Animations can be created in a loop modifying the scene and saving 
a screenshot to a file. (See example below)

This function is a wrapper for the GL2PS library by Christophe Geuzaine,
and has the same limitations as that library:  not all OpenGL features
are supported, and some are only supported in some formats.
See the reference for full details.
}
\references{
GL2PS: an OpenGL to PostScript printing library by Christophe Geuzaine,
\url{http://www.geuz.org/gl2ps/}, version 1.4.0.
}
\author{ Christophe Geuzaine / Albrecht Gebhardt }
\examples{
# Create new files in tempdir
savedir <- setwd(tempdir())

x <- y <- seq(-10, 10, length.out = 20)
z <- outer(x, y, function(x, y) x^2 + y^2)
persp3d(x, y, z, col = 'lightblue')

title3d("Using LaTeX text", col = 'red', line = 3)
rgl.postscript("persp3da.ps", "ps", drawText = FALSE)
rgl.postscript("persp3da.pdf", "pdf", drawText = FALSE)
rgl.postscript("persp3da.tex", "tex")
pop3d()
title3d("Using ps/pdf text", col = 'red', line = 3)
rgl.postscript("persp3db.ps", "ps")
rgl.postscript("persp3db.pdf", "pdf")
rgl.postscript("persp3db.tex", "tex", drawText = FALSE)

setwd(savedir)

\dontrun{

#
# create a series of frames for an animation
#

open3d()
shade3d(oh3d(), color = "red")
view3d(0, 20)

for (i in 1:45) {
  view3d(i, 20)
  filename <- paste("pic", formatC(i, digits = 1, flag = "0"), ".eps", sep = "") 
  rgl.postscript(filename, fmt = "eps")
}

}

}
\seealso{
  \code{\link{view3d}}, \code{\link{snapshot3d}}
}
\keyword{dynamic}

\name{planes3d}
\alias{planes3d}
\alias{clipplanes3d}
\title{Add planes}
\description{
\code{planes3d} adds mathematical planes 
to a scene.  Their intersection with the current
bounding box will be drawn.  \code{clipplanes3d} 
adds clipping planes to a scene.
}
\usage{
planes3d(a, b = NULL, c = NULL, d = 0, ...)
clipplanes3d(a, b = NULL, c = NULL, d = 0)
}
\arguments{
  \item{a, b, c}{
Coordinates of the normal to the plane.  Any reasonable way of defining the
coordinates is acceptable.  See the function \code{\link[grDevices]{xyz.coords}}
for details.
}
  \item{d}{
Coordinates of the "offset".  See the details.
}
  \item{\dots}{
Material properties. See \code{\link{material3d}} for details.  
}
}
\details{
\code{planes3d} draws planes using the parametrization
\eqn{a x + b y + c z + d = 0}.  Multiple planes may be specified
by giving multiple values for any of \code{a, b, c, d}; the other 
values will be recycled as necessary.

\code{clipplanes3d} defines clipping planes
using the same equations.  Clipping planes suppress the display of
other objects (or parts of them) in the subscene, based on their coordinates.
Points (or parts of lines or surfaces) where the coordinates \code{x, y, z}
satisfy \eqn{a x + b y + c z + d < 0} will be suppressed.

The number of clipping planes
supported by the OpenGL driver is implementation dependent; use
\code{par3d("maxClipPlanes")} to find the limit.
}
\value{
A shape ID of the planes or clipplanes object is returned invisibly.
}
\seealso{
\code{\link{abclines3d}} for mathematical lines.

\code{\link{triangles3d}} or the corresponding
functions for quadrilaterals may be used to draw sections of planes that
do not adapt to the bounding box.

The example in \link{subscene3d} shows how to combine clipping planes
to suppress complex shapes.
}
\examples{

# Show regression plane with z as dependent variable

open3d()
x <- rnorm(100)
y <- rnorm(100)
z <- 0.2*x - 0.3*y + rnorm(100, sd = 0.3)
fit <- lm(z ~ x + y)
plot3d(x, y, z, type = "s", col = "red", size = 1)

coefs <- coef(fit)
a <- coefs["x"]
b <- coefs["y"]
c <- -1
d <- coefs["(Intercept)"]
planes3d(a, b, c, d, alpha = 0.5)

open3d()
ids <- plot3d(x, y, z, type = "s", col = "red", size = 1, forceClipregion = TRUE) 
oldid <- useSubscene3d(ids["clipregion"])
clipplanes3d(a, b, c, d)
useSubscene3d(oldid)
}
\keyword{dynamic}

\name{rglExtrafonts}
\alias{rglExtrafonts}
\title{
Register extra fonts
}
\description{
This function uses the \href{https://github.com/wch/extrafont}{\pkg{extrafont}} 
package to help register system fonts for use with
FreeType in \pkg{rgl}.
}
\usage{
rglExtrafonts(..., quiet = TRUE)
}
\arguments{
  \item{\dots}{
Vectors of fonts to try.  See the Details.
}
  \item{quiet}{
Whether to print information on progress.
}
}
\details{
The \pkg{extrafont}
package collects information on
installed fonts from the system.  When you first
install \pkg{extrafont}, or after new fonts have been installed on your
system, run \code{extrafont::font_import()} to build
its database of system fonts.  

Fonts can be installed in \pkg{rgl} using
\code{rglExtrafonts(rglname = familyname)} or \code{rglExtrafonts(familyname)}.  In this call
\code{familyname} is a vector of family names to look
for in the \pkg{extrafont} database using \code{extrafont::choose_font(familyname)}; the first one found
will be registered with \pkg{rgl}.  The optional name \code{rglname} will also be usable to refer to the font family.

If none of the given family names is found, no change
will be made to the registered fonts in \pkg{rgl}.

During startup, \pkg{rgl} detects whether \pkg{extrafont}
is installed, and if so runs 
\verb{

rglExtrafonts(sans = c("Helvetica", "Arial"), 
              serif = c("Times", "Times New Roman"), 
              mono = c("Courier", "Courier New"))
              
} 
to attempt to set up the default fonts.

Fonts found by \pkg{extrafont} can also be used in
some other graphics devices besides \pkg{rgl}; see 
\href{https://github.com/wch/extrafont}{the \pkg{extrafont} documentation}
for details.
}
\note{
Each font in a display needs a unique \pkg{rgl} name;
if the associated font for a given name is changed,
all previously plotted text will also change.

Currently \code{\link{rglwidget}} displays will not 
respect the new definitions.
}
\value{
Invisibly returns a vector giving the \pkg{rgl} name
and the family name for the newly installed font.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{text3d}}, \code{\link{rglFonts}}
}
\examples{
if (requireNamespace("extrafont") && !in_pkgdown_example()) {
  
  open3d()
  text3d(1,1,1, "Default", family = "sans", cex = 2)  
  
  # Attempt to register new sans-serif font:
  newfamily <- rglExtrafonts(newsans = c("Comic Sans MS", "Impact", 
                                         "Verdana", "Tahoma"))
  
  text3d(2,2,2, newfamily, family = "newsans", cex = 2)
  
}
}

\name{sprites}
\alias{sprites3d}
\alias{particles3d}
\title{Add sprites}
\description{
  Adds a sprite set shape node to the scene.
}
\usage{
sprites3d(x, y = NULL, z = NULL, radius = 1, 
          shapes = NULL, userMatrix, 
          fixedSize = FALSE, 
          adj = 0.5, pos = NULL, offset = 0.25, 
          rotating = FALSE, ...)
particles3d(x, y = NULL, z = NULL, radius = 1, ...)
}
\arguments{
  \item{ x, y, z }{point coordinates.  Any reasonable way of defining the
    coordinates is acceptable.  See the function \code{\link[grDevices]{xyz.coords}}
    for details.}
  \item{ radius }{vector or single value defining the sprite radius}
  \item{ shapes }{\code{NULL} for a simple square, or a vector of
    identifiers of shapes in the scene}
  \item{ userMatrix }{if \code{shape} is not \code{NULL}, the transformation matrix 
    for the shapes}
  \item{ fixedSize }{should sprites remain at a fixed size,
or resize with the scene?}
  \item{ adj, pos, offset }{positioning arguments; see Details}
  \item{ rotating }{should sprites remain at a fixed orientation, or rotate with the scene?}
  \item{ ... }{material properties when \code{shapes = NULL}, texture mapping is supported}
}
\details{
  Simple sprites (used when \code{shapes} is \code{NULL}) are 1 by 1 squares
  that are directed towards the viewpoint. Their primary use is for
  fast (and faked) atmospherical effects, e.g. particles and clouds
  using alpha blended textures. Particles are sprites using an alpha-blended 
  particle texture giving the illusion of clouds and gases.  
  The centre of each square will by default be at the
coordinates given by \code{x, y, z}.  This may be adjusted
using the \code{adj} or \code{pos} parameters.

\code{adj} and \code{pos} are treated similarly to the same
parameters for \code{\link{text3d}}.  \code{adj} has 3
entries, for adjustment to the \code{x}, \code{y} and
\code{z} coordinates respectively.  For \code{x}, a value
of 0 puts the sprite to the right of the specified point,
0.5 centers it there, and 1 puts it to the left.  The other
coordinates are similar.  By default, each value is 0.5 and
the sprites are centered at the points given by \code{(x, y, z)}.

The \code{pos} parameter overrides \code{adj}.  It should
be an integer or vector of integers (one per point),
interpreted as in \code{\link{text3d}} to position the
sprite relative to the \code{(x, y, z)} point:  0 is centered on it, 1 is below,
2 is to the left, 3 is above, 4 is to the right, 5 is in
front, and 6 is behind.  \code{offset} is the
fraction of the sprite size to separate it from the point.

  When \code{shapes} is not \code{NULL}, it should be a vector of
  identifiers of objects to plot in the scene (e.g. as returned by
  plotting functions or by \code{\link{ids3d}}).  These objects will
  be removed from the scene and duplicated as a sprite image in a
  constant orientation, as specified by \code{userMatrix}.  By default the 
  origin \code{(0, 0, 0)} will be plotted at the coordinates given by \code{(x, y, z)}, perhaps modified by \code{adj}
  or \code{pos}.
  
  The \code{userMatrix} argument is ignored for \code{shapes = NULL}.  For
  shapes, \code{sprites3d} defaults the matrix to \code{r3dDefaults$userMatrix}.
  
  If any coordinate is \code{NA}, the sprite is not plotted.
  
  The id values of the shapes may be retrieved after plotting
using \code{rgl.attrib(id, "ids")};
the user matrix is retrieved using \code{rgl.attrib(id, "usermatrix")}.
}
\value{
  These functions are called for the side effect of displaying the sprites.
  The shape ID of the displayed object is returned.
}
\examples{
open3d()
particles3d( rnorm(100), rnorm(100), rnorm(100), color = rainbow(100) )
# is the same as
sprites3d( rnorm(100), rnorm(100), rnorm(100), color = rainbow(100),
  lit = FALSE, alpha = .2,
  textype = "alpha", texture = system.file("textures/particle.png", package = "rgl") )
sprites3d( rnorm(10) + 6, rnorm(10), rnorm(10), shape = shade3d(tetrahedron3d(), col = "red") )
}
\seealso{
  \code{\link{material3d}}, \code{\link{text3d}}
}
\keyword{dynamic}

\name{shadow3d}
\alias{shadow3d}
\title{
Project shadows of mesh onto object.
}
\description{
Project a mesh onto a surface in a scene so that it appears to cast a shadow onto the surface.
}
\usage{
shadow3d(obj, mesh, plot = TRUE, up = c(0, 0, 1),
         P = projectDown(up), outside = FALSE,
         ...)
}
\arguments{
  \item{obj}{
The target object which will show the shadow.
}
  \item{mesh}{
The mesh which will cast the shadow.
}
  \item{plot}{
Whether to plot the result.
}
  \item{up}{
Which direction is \dQuote{up}?
  }
  \item{P}{
The projection to use for draping, a 4x4 matrix.  See \code{\link{drape3d}} for details on how \code{P} is used.
}
  \item{outside}{
Should the function compute and (possibly) plot the region
outside of the shadow?
}
  \item{\dots}{
Other arguments to pass to \code{\link{filledContour3d}},
which will do the boundary calculations and plotting.
}
}
\details{
\code{shadow3d} internally constructs a function that 
is zero on the boundary of the shadow and positive inside, 
then draws filled contours of that function.  Because the
function is nonlinear, the boundaries will be approximate,
with the best approximation resulting from a large
value of \code{\link{filledContour3d}} parameter \code{minVertices}.

If \code{outside = TRUE}, the first color used by
\code{\link{filledContour3d}} will indicate the inside
of the shadow, and the second color will indicate the exterior.
}
\value{
The returned value from \code{\link{filledContour3d}}.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{drape3d}}, \code{\link{facing3d}}
}
\examples{
open3d()
obj <- translate3d(scale3d(oh3d(), 0.3, 0.3, 0.3), 0,0,2)
shade3d(obj, col = "red")
target <- icosahedron3d()

# We offset the target using polygon_offset = 1 so that the
# shadow on its surface will appear clearly.

shade3d(target, col = "white", polygon_offset = 1)

# minVertices = 1000 leaves noticeable artifacts on the edges
# of the shadow.  A larger value gives a better result, but is
# slower.

# We use facing3d(target) so the shadow and outside part only 
# appear on the upper side of the target

shadow3d(facing3d(target), obj, minVertices = 1000, plot=TRUE,
         col = c("yellow", "blue"), outside = TRUE)
}

\name{rgl.bringtotop}
\alias{rgl.bringtotop}
\title{Assign focus to an RGL window }
\description{
'rgl.bringtotop' brings the current RGL window to the
front of the window stack (and gives it focus).}
\usage{
rgl.bringtotop(stay = FALSE)
}
\arguments{
  \item{stay}{whether to make the window stay on top.}
}
\details{
If \code{stay} is \code{TRUE}, then the window will stay on top
of normal windows.
}
\author{ Ming Chen/Duncan Murdoch }
\note{not completely implemented for X11 graphics (stay not implemented;
window managers such as KDE may block this action
(set "Focus stealing prevention level" to None in Control Center/Window Behavior/Advanced)).
Not currently implemented under OS/X.}
#ifdef windows
\seealso{\code{\link[grDevices]{bringToTop}}}
#endif
\examples{
open3d()
points3d(rnorm(1000), rnorm(1000), rnorm(1000), color = heat.colors(1000))
rgl.bringtotop(stay = TRUE)    
}
\keyword{ dynamic }

\name{par3d}
\alias{par3d}
\alias{rgl.par3d.names}
\alias{rgl.par3d.readonly}
\concept{activeSubscene}
\concept{antialias}
\concept{bbox}
\concept{cex}
\concept{family}
\concept{font}
\concept{fontname}
\concept{FOV}
\concept{ignoreExtent}
\concept{maxClipPlanes}
\concept{modelMatrix}
\concept{listeners}
\concept{mouseMode}
\concept{projMatrix}
\concept{scale}
\concept{skipRedraw}
\concept{useFreeType}
\concept{userMatrix}
\concept{userProjection}
\concept{viewport}
\concept{windowRect}
\concept{zoom}

\title{Set or query RGL parameters}

\description{
  \code{par3d} can be used to set or query graphical parameters in RGL.
  Parameters can be set by specifying them as arguments to \code{par3d} in
  \code{name = value} form, or by passing them as a list of named
  values.
}
\usage{
par3d(\dots, no.readonly = FALSE, dev = cur3d(), 
      subscene = currentSubscene3d(dev))
rgl.par3d.names
rgl.par3d.readonly
}

\arguments{
  \item{\dots}{arguments in \code{name = value} form, or a list of tagged
    values.  The names must come from the graphical parameters described
    below.}
  \item{no.readonly}{logical; if \code{TRUE} and there are no other
    arguments, only those parameters which can be set by a
    subsequent \code{par3d()} call are returned.}
  \item{dev}{integer; the RGL device.}
  \item{subscene}{integer; the subscene.}
}

\details{
  Parameters are queried by giving one or more character vectors to
  \code{par3d}.

  \code{par3d()} (no arguments) or \code{par3d(no.readonly = TRUE)} is used to
  get \emph{all} the graphical parameters (as a named list).  

  By default, queries and modifications apply to the current subscene
  on the current device; specify \code{dev} and/or \code{subscene} to 
  change this.  Some parameters apply to the device as a whole;
  these are marked in the list below.

}

\value{
  When parameters are set, their former values are returned in an
  invisible named list.  Such a list can be passed as an argument to
  \code{par3d} to restore the parameter values.
  Use \code{par3d(no.readonly = TRUE)} for the full list of parameters
  that can be restored.

  When just one parameter is queried, its value is returned directly.
  When two or more parameters are queried, the result is a list of
  values, with the list names giving the parameters.

  Note the inconsistency: setting one parameter returns a list, but
  querying one parameter returns an object.
}

\section{Parameters}{

The \code{rgl.par3d.names} variable contains the
full list of names of \code{par3d} properties.
\code{rgl.par3d.readonly} contains the list of
read-only properties.

In the list below, \emph{\bold{R.O.}} indicates the
read-only arguments: These may only be used in
queries, they do not set anything.
  
  \describe{
    \item{\code{activeSubscene}}{\emph{\bold{R.O.}} integer.  Used
    with \code{\link{rgl.setMouseCallbacks}}:  during a 
    callback, indicates the id of the subscene that was clicked.}
    \item{\code{antialias}}{\emph{\bold{R.O.}} in \code{par3d}, may be
    set in \code{open3d}. The (requested) number of hardware
    antialiasing planes to use (with multisample antialiasing).  The
    OpenGL driver may not support the requested number, in which case
    \code{par3d("antialias")} will report what was actually set. Applies to the whole device.}    
    \item{\code{cex}}{real.  The default size for text.}
    \item{\code{family}}{character.  The default device independent family name; see \code{\link{text3d}}.
    Applies to the whole device.}
   \item{\code{font}}{integer.  The default font number (from 1 to 4; see \code{\link{text3d}}).
    Applies to the whole device.}
    \item{\code{useFreeType}}{logical.  Should FreeType fonts be used?
    Applies to the whole device.}
    \item{\code{fontname}}{\emph{\bold{R.O.}}; the system-dependent name of the current font.
    Applies to the whole device.}
    \item{\code{FOV}}{real.  The field of view, from 0 to 179 degrees.  This controls
      the degree of parallax in the perspective view.
      Isometric perspective corresponds to \code{FOV = 0}.}
    \item{\code{ignoreExtent}}{logical.  Set to \code{TRUE} so that subsequently plotted objects
      will be ignored in calculating the bounding box of the scene.  Applies to the
      whole device.}
    \item{\code{maxClipPlanes}}{\emph{\bold{R.O.}}; an integer giving the maximum
      number of clip planes that can be defined in the current system. Applies to the whole device.}
    \item{\code{modelMatrix}}{\emph{\bold{R.O.}}; a 4 by 4 matrix describing the position
      of the user data.  See the Note below.}
    \item{\code{listeners}}{integer.  A vector of subscene id values.  If a subscene receives
      a mouse event (see \code{mouseMode} just below), the same action will be carried out
      on all subscenes in this list.  (The subscene itself is normally listed as a listener.
      If it is not listed, it will not respond to its own mouse events.)}
    \item{\code{mouseMode}}{character.  A vector of 5 strings describing
      mouse actions. The 5 entries are named
      \code{c("none", "left", "right", "middle", "wheel")}, corresponding to actions for no button, 
      the left, right or middle button, and the 
      mouse wheel.
      Partial matching to action names is used.  Possible values for the actions are:
	\describe{
	 \item{\code{"none"}}{No action for this button.}
	 \item{\code{"trackball"}}{Mouse acts as a virtual trackball, rotating the scene.}
	 \item{\code{"xAxis"}}{Similar to \code{"trackball"}, but restricted to X axis rotation.}
	 \item{\code{"yAxis"}}{Y axis rotation.}
	 \item{\code{"zAxis"}}{Z axis rotation.}
	 \item{\code{"polar"}}{Mouse rotates the scene by moving in polar coordinates.}
	 \item{\code{"selecting"}}{Mouse is used for selection.  This is not normally set by the
	 user, but is used internally by the \code{\link{select3d}} function.}
	 \item{\code{"zoom"}}{Mouse is used to zoom the display.}
	 \item{\code{"fov"}}{Mouse changes the field of view of the display.}
	 \item{\code{"user"}}{Used  when a user handler is set by 
	 \code{\link{rgl.setMouseCallbacks}}.}
	}
      Possible values for the last entry corresponding to the mouse wheel also include
        \describe{
	  \item{\code{"pull"}}{Pulling on the mouse wheel increases magnification, i.e. \dQuote{pulls the scene closer}.}
	  \item{\code{"push"}}{Pulling on the mouse wheel decreases magnification, i.e. \dQuote{pushes the scene away}.}
	  \item{\code{"user2"}}{Used when a user handler is set by 
	    \code{\link{rgl.setWheelCallback}}.}
	}
      A common default on Mac OSX is to convert a two finger drag on a trackpad to a mouse wheel rotation.
      
      The first entry is for actions to take when
no mouse button is pressed.  Legal values are the same
as for the mouse buttons.

      The first entry was added after \pkg{rgl} version 0.106.8.  For back compatibility, if the vector
      of actions is less than 5 entries, \code{"none"}
      will be added at the start of it.
    }
    \item{\code{observer}}{\emph{\bold{R.O.}}; the position of the observer relative to
      the model.  Set by \code{\link{observer3d}}.  See the Note below.}
    \item{\code{projMatrix}}{\emph{\bold{R.O.}}; a 4 by 4 matrix describing the current
      projection of the scene.}
    \item{\code{scale}}{real.  A vector of 3 values indicating the amount by which to
      rescale each axis before display.  Set by \code{\link{aspect3d}}.}
    \item{\code{skipRedraw}}{whether to update the display.  Set to \code{TRUE} to suspend
      updating while making multiple changes to the scene.  See \code{demo(hist3d)} for an example.
      Applies to the whole device.}
    \item{\code{userMatrix}}{a 4 by 4 matrix describing user actions to
      display the scene.}
    \item{\code{userProjection}}{a 4 by 4 matrix describing
changes to the projection.}
    \item{\code{viewport}}{real.  A vector giving the dimensions of the
      window in pixels.  The entries are taken to be \code{c(x, y, width, height)} where
      \code{c(x, y)} are the coordinates in pixels of the lower left corner within the window.}
    \item{\code{zoom}}{real. A positive value indicating the current
      magnification of the scene.}
      \item{\code{bbox}}{\emph{\bold{R.O.}}; real. A vector of six values indicating the current values of the
      bounding box of the scene (xmin, xmax, ymin, ymax, zmin, zmax)}
    \item{\code{windowRect}}{integer.  A vector of four values indicating the left, top, right and 
      bottom of the displayed window (in pixels).  Applies to the whole device.}
  }
}

\note{
The \code{"xAxis"}, \code{"yAxis"} and \code{"zAxis"} mouse modes rotate 
relative to the coordinate system of the data, regardless of the current 
orientation of the scene.

When multiple parameters are set, they are set in the order given.  In some 
cases this may lead to warnings and ignored values; for example, some font 
families only support \code{cex = 1}, so changing both \code{cex} and 
\code{family} needs to be done in the right order.  For example, when using the
\code{"bitmap"} family on Windows, \code{par3d(family = "sans", cex = 2)}
will work, but \code{par3d(cex = 2, family = "sans")} will leave \code{cex}
at 1 (with a warning that the \code{"bitmap"} family only supports that size).

Although \code{par3d("viewport")} names the entries of the reported vector,
names are ignored when setting the viewport and entries must be specified
in the standard order.

In \pkg{rgl} versions 0.94.x the \code{modelMatrix} entry had a changed meaning; 
before and after that it contains a copy of the OpenGL MODELVIEW matrix. 

As of version 0.100.32, when changing the \code{"windowRect"} parameter, the 
\code{"viewport"} for the root (or specified) subscene
is changed immediately.  This fixes a bug where in earlier
versions it would only be changed when the window was redrawn, potentially after another command making use of the value.

Default values are not described here, as several of them
are changed by the \code{\link{r3dDefaults}} variable when
the window is opened by \code{\link{open3d}}.
}

\section{Rendering}{
The parameters returned by \code{par3d} are sufficient to determine where RGL would render
a point on the screen.  Given a column vector \code{(x, y, z)} in a subscene \code{s}, 
it performs the equivalent of 
the following operations:
\enumerate{
\item It converts the point to homogeneous coordinates by appending \code{w = 1}, giving
the vector \code{v = (x, y, z, 1)}.
\item It calculates the \code{M = par3d("modelMatrix")} as a product from right to left
of several matrices:
  \itemize{  
  \item A matrix to translate the centre of the bounding box to the origin.
  \item A matrix to rescale according to \code{par3d("scale")}.
  \item The \code{par3d("userMatrix")} as set by the user.
  \item A matrix which may be set by mouse movements.
  \item The description above applies to the usual case where there 
  is just one subscene, or where the subscene's \code{"model"} is set to
  \code{"replace"}.  If it is set to \code{"modify"}, the first step is
  skipped, and at the end the procedure is followed for the parent subscene.  If it is set to \code{"inherit"} only the parent settings
  are used.
  }
\item It multiplies the point by \code{M} giving \code{u = M \%*\% v}. 
\item It multiplies that point by a matrix based on the observer position
to translate the origin to the centre of the viewing region.
\item Using this location and information on the
normals (which have been similarly transformed), it performs lighting calculations. 
\item It obtains the projection matrix \code{P = par3d("projMatrix")} based on the bounding box and field of view
or observer location, multiplies that by the \code{userProjection} matrix to give \code{P}.  It multiplies the point
by it giving \code{P \%*\% u = (x2, y2, z2, w2)}.
\item It converts back to Euclidean coordinates by dividing the first 3 coordinates by \code{w2}.
\item The new value \code{z2/w2} represents the depth into the scene of the point.  Depending
on what has already been plotted, this depth might be obscured, in which case nothing more is plotted.
\item If the point is not culled due to depth, the \code{x2} and \code{y2} values are used
to determine the point in the image.  The \code{par3d("viewport")} values are used to translate
from the range \code{(-1, 1)} to pixel locations, and the point is plotted.
\item If hardware antialiasing is enabled, then the whole process is
repeated multiple times (at least conceptually) with different
locations in each pixel sampled to determine what is plotted there,
and then the images are combined into what is displayed.
}
See ?\link{matrices} for more information on homogeneous and Euclidean coordinates.

Note that many of these calculations are done on the graphics card using single precision;
you will likely see signs of rounding error if your scene requires more than 4 or 5 
digit precision to distinguish values in any coordinate.  
}

\seealso{
\code{\link{view3d}} to set \code{FOV} and \code{zoom}.

\code{\link{open3d}} for how to open a new window
with default settings for these parameters.
}

\references{
OpenGL Architecture Review Board (1997).  OpenGL Programming Guide.  Addison-Wesley.
}

\examples{
    open3d()
    shade3d(cube3d(color = rainbow(6), meshColor = "faces"))
    save <- par3d(userMatrix = rotationMatrix(90*pi/180, 1, 0, 0))
    highlevel()  # To trigger display
    save
    par3d("userMatrix")    
    par3d(save)
    highlevel()
    par3d("userMatrix")
}

\keyword{dynamic}

\name{select3d}
\alias{select3d}
\alias{selectionFunction3d}
\title{ Select a rectangle in an RGL scene }
\description{
  This function allows the user to use the mouse to
  select a region in an RGL scene.  
}
\usage{
select3d(button = c("left", "middle", "right"),
             dev = cur3d(), subscene = currentSubscene3d(dev))
selectionFunction3d(proj, region = proj$region)
}
\arguments{
  \item{ button }{ Which button to use for selection.}
  \item{ dev, subscene}{ The RGL device and subscene to work with }
  \item{proj}{An object returned from \code{\link{rgl.projection}}
  containing details of the current projection.}
  \item{region}{Corners of a rectangular region in the display.}
}

\details{
    
\code{select3d} selects 3-dimensional regions by allowing the
user to use a mouse to draw a rectangle showing
the projection of the region onto the screen.  It returns
a function which tests points for inclusion in the selected region.

\code{selectionFunction3d} constructs such a test function given 
coordinates and current transformation matrices.

If the scene is later moved or rotated, the selected region will 
remain the same, though no longer corresponding to a rectangle on the screen.
  
}
\value{
These return a function \code{f(x, y, z)} which tests whether each
of the points \code{(x, y, z)} is in the selected region, returning
a logical vector.  This function accepts input in a wide
variety of formats as it uses \code{\link[grDevices]{xyz.coords}} 
to interpret its parameters.
}
\author{ Ming Chen / Duncan Murdoch }
\seealso{ \code{\link{selectpoints3d}}, \code{\link{locator}} }
\examples{

# Allow the user to select some points, and then redraw them
# in a different color

if (interactive()) {
 x <- rnorm(1000)
 y <- rnorm(1000)
 z <- rnorm(1000)
 open3d()
 points3d(x, y, z)
 f <- select3d()
 if (!is.null(f)) {
   keep <- f(x, y, z)
   pop3d()
   points3d(x[keep], y[keep], z[keep], color = 'red')
   points3d(x[!keep], y[!keep], z[!keep])
 }
}
}
\keyword{dynamic}

\name{rgl.getAxisCallback}
\alias{rgl.getAxisCallback}
\title{
Get user-defined axis labelling callbacks.
}
\description{
This function gets a user-defined axis labelling callback in R.
}
\usage{
rgl.getAxisCallback(axis, dev = cur3d(), subscene = currentSubscene3d(dev))
}

\arguments{
  \item{axis}{
Which axis? Can be value from \code{1:3}.
}
  \item{dev, subscene}{
The RGL device and subscene to work with.
}
}
\value{
The callback function.
}
\seealso{\code{\link{setAxisCallbacks}} to
work with \code{\link{rglwidget}}.
} 
\author{
Duncan Murdoch
}

\name{shade3d}
\alias{dot3d}
\alias{dot3d.mesh3d}
\alias{wire3d}
\alias{wire3d.mesh3d}
\alias{shade3d}
\alias{shade3d.mesh3d}

\title{Draw 3D mesh objects}
\description{
Draws 3D mesh objects in full, or just the edges, or just 
the vertices.
}
\usage{
  dot3d(x, ...)   # draw dots at the vertices of an object
  \method{dot3d}{mesh3d}(x, ...,
                         front = "points", back = "points")
  wire3d(x, ...)  # draw a wireframe object
  \method{wire3d}{mesh3d}(x, ...,
                          front = "lines", back = "lines")
  shade3d(x, ...) # draw a shaded object
  \method{shade3d}{mesh3d}(x, override = TRUE, 
                           meshColor = c("vertices", "edges", "faces", "legacy"), 
                           texcoords = NULL, ...,
                           front = "filled", back = "filled")
}
\arguments{
  \item{x}{a \code{mesh3d} object.}
  \item{...}{additional rendering parameters, or for
  \code{dots3d} and \code{wire3d}, parameters to pass
  to \code{shade3d}}
  \item{override}{should the parameters specified here override those stored in the object?}
  \item{meshColor}{how should colours be interpreted?  See details 
  below}
  \item{texcoords}{texture coordinates at each vertex.}
  \item{front, back}{Material properties for rendering.}
}
\details{
  
  The \code{meshColor} argument controls how material colours and textures are interpreted.  This parameter
was added in \pkg{rgl} version 0.100.1 (0.100.27 for \code{dot3d}).  Possible values are:
\describe{
\item{\code{"vertices"}}{Colours and texture coordinates are applied by vertex, in the order
they appear in the \code{x$vb} matrix.}
\item{\code{"edges"}}{Colours are applied to each edge:  first to the segments in the \code{x$is} matrix, then
the 
3 edges of each triangle in the \code{x$it} matrix, then the 4
edges of each quad in the \code{x$ib} matrix.  This mode
is only supported if both front and back materials are
\code{"lines"}, and the mesh contains no points.}
\item{\code{"faces"}}{Colours are applied to each object
in the mesh:  first to the points, then the segments, 
triangles and finally quads.  The entire whole face (or point or
segment) receives one colour from the specified colours.}
\item{\code{"legacy"}}{Colours and textures are applied in the same way
as in \pkg{rgl} versions earlier than 0.100.1.}
}
Unique partial matches of these values will be recognized.

If colours are specified but \code{meshColor} is not
and \code{options(rgl.meshColorWarning = TRUE)},
a warning will be given that their
interpretation may have changed.  In versions 0.100.1 to 0.100.26
of \pkg{rgl}, the default
was to give the warning; now the default is for no warning.

Note that since version 0.102.10, \code{meshColor = 
"edges"} is only allowed when drawing lines (the
\code{wire3d} default), and it may draw
edges more than once.  In general, if any rendering
draws twice at the same location, which copy is visible 
depends on the order of drawing and the
\code{\link{material3d}("depth_test")}
setting.  

Whether points, lines or solid faces are drawn is
determined in 3 steps:
\enumerate{
\item{If arguments \code{"front"} or \code{"back"} are
specified in the call, those are used.}
\item{If one or both of those arguments are not
specified, but the material properties are present in
the object, those are used.}
\item{If values are not specified in either of those
places, \code{shade3d} draws filled surfaces,
\code{wire3d} draws lines, and \code{dot3d} draws points.
}
}
Note:  For some versions of rgl up to version 0.107.15, 
rule 2 above was not respected.
}
\value{
  \code{dot3d}, \code{wire3d}, and \code{shade3d} are called for their side effect
  of drawing an object into the scene; they return an object ID (or vector of IDs) invisibly.
  
  See \link{primitives} for a discussion of texture coordinates.
}
\examples{

  # generate a quad mesh object

  vertices <- c( 
     -1.0, -1.0, 0,
      1.0, -1.0, 0,
      1.0,  1.0, 0,
     -1.0,  1.0, 0
  )
  indices <- c( 1, 2, 3, 4 )
  
  open3d()  
  wire3d( mesh3d(vertices = vertices, quads = indices) )
  
  # render 4 meshes vertically in the current view

  open3d()  
  bg3d("gray")
  l0 <- oh3d(tran = par3d("userMatrix"), color = "green" )
  shade3d( translate3d( l0, -6, 0, 0 ))
  l1 <- subdivision3d( l0 )
  shade3d( translate3d( l1 , -2, 0, 0 ), color = "red", override = FALSE )
  l2 <- subdivision3d( l1 )
  shade3d( translate3d( l2 , 2, 0, 0 ), color = "red", override = TRUE )
  l3 <- subdivision3d( l2 )
  shade3d( translate3d( l3 , 6, 0, 0 ), color = "red" )
  
  # render all of the Platonic solids
  open3d()
  shade3d( translate3d( tetrahedron3d(col = "red"), 0, 0, 0) )
  shade3d( translate3d( cube3d(col = "green"), 3, 0, 0) )
  shade3d( translate3d( octahedron3d(col = "blue"), 6, 0, 0) )
  shade3d( translate3d( dodecahedron3d(col = "cyan"), 9, 0, 0) )
  shade3d( translate3d( icosahedron3d(col = "magenta"), 12, 0, 0) )
}
\seealso{
  \code{\link{mesh3d}}, \code{\link{par3d}}, \code{\link{shapelist3d}} for multiple shapes
}
\keyword{dynamic}

\name{matrices}
\alias{matrices}
\alias{identityMatrix}
\alias{scaleMatrix}
\alias{translationMatrix}
\alias{rotationMatrix}
\alias{scale3d}
\alias{translate3d}
\alias{rotate3d}
\alias{transform3d}
\alias{asHomogeneous}
\alias{asEuclidean}
\alias{asHomogeneous2}
\alias{asEuclidean2}
\title{Work with homogeneous coordinates }
\description{
These functions construct 4x4 matrices for transformations
in the homogeneous coordinate system used by OpenGL, and translate
vectors between homogeneous and Euclidean coordinates.
}
\usage{
identityMatrix()
scaleMatrix(x, y, z)
translationMatrix(x, y, z)
rotationMatrix(angle, x, y, z, matrix)
asHomogeneous(x)
asEuclidean(x)
asHomogeneous2(x)
asEuclidean2(x)

scale3d(obj, x, y, z, ...)
translate3d(obj, x, y, z, ...)
rotate3d(obj, angle, x, y, z, matrix, ...)

transform3d(obj, matrix, ...)

}
\arguments{
  \item{x, y, z, angle, matrix}{See details} 
  \item{obj}{An object to be transformed}
  \item{...}{Additional parameters to be passed to methods}
}
\details{
OpenGL uses homogeneous coordinates to handle perspective and affine
transformations.  The homogeneous point \code{(x, y, z, w)} corresponds
to the Euclidean point \code{(x/w, y/w, z/w)}.  The matrices produced by
the functions \code{scaleMatrix}, \code{translationMatrix}, and \code{rotationMatrix}
are to be left-multiplied by a row vector 
of homogeneous coordinates; alternatively, the transpose of the result
can be right-multiplied by a column vector.  The generic functions
\code{scale3d}, \code{translate3d} and \code{rotate3d} apply these transformations
to the \code{obj} argument.  The \code{transform3d} function is a synonym 
for \code{rotate3d(obj, matrix = matrix)}.

By default, it is assumed that \code{obj} is a row vector
(or a matrix of row vectors) which will be multiplied on the right by
the corresponding matrix, but users may write methods for these generics
which operate differently.  Methods are supplied for \code{\link{mesh3d}}
objects.

To compose transformations, use matrix multiplication.  The effect is
to apply the matrix on the left first, followed by the one on the right.

\code{identityMatrix} returns an identity matrix.

\code{scaleMatrix} scales each coordinate by the given factor.  In Euclidean
coordinates, \code{(u, v, w)} is transformed to \code{(x*u, y*v, z*w)}.

\code{translationMatrix} translates each coordinate by the given translation, i.e.
\code{(u, v, w)} is transformed to \code{(u + x, v + y, w + z)}.

\code{rotationMatrix} can be called in three ways.  With 
arguments \code{angle, x, y, z} it represents a rotation 
of \code{angle} radians about the axis 
\code{x, y, z}.  If \code{matrix} is a 3x3 rotation matrix,
it will be converted into the corresponding matrix in 4x4 homogeneous
coordinates.  Finally, if a 4x4 matrix is given, it will be returned unchanged.
(The latter behaviour is used to allow \code{transform3d} to act like a 
generic function, even though it is not.)

Use \code{asHomogeneous(x)} to convert the Euclidean vector \code{x} to
homogeneous coordinates, and \code{asEuclidean(x)} for the reverse transformation.  These functions accept the following
inputs:
\itemize{
\item n x 3 matrices:  rows are assumed to be Euclidean
\item n x 4 matrices:  rows are assumed to be homogeneous
\item vectors of length 3n or 4n:  assumed to be vectors
concatenated. For the ambiguous case
of vectors that are length 12n (so both 3n and 4n are possible),
the assumption is that the conversion is necessary:  \code{asEuclidean} assumes the vectors are homogeneous,
and \code{asHomogeneous} assumes the vectors are Euclidean.
}
Outputs are n x 4 or n x 3 matrices for \code{asHomogeneous}
and \code{asEuclidean} respectively.  

The functions \code{asHomogeneous2} and \code{asEuclidean2}
act similarly, but they assume inputs are 3 x n or 4 x n
and outputs are in similar shapes.


}
\value{
\code{identityMatrix},
\code{scaleMatrix}, \code{translationMatrix}, and \code{rotationMatrix} produce
a 4x4 matrix representing the requested transformation
in homogeneous coordinates.  

\code{scale3d}, \code{translate3d} and \code{rotate3d} transform the object
and produce a new object of the same class.
}
\author{ Duncan Murdoch }
\seealso{\code{\link{par3d}} for a description of how RGL uses matrices in
rendering.}

\examples{
# A 90 degree rotation about the x axis:

rotationMatrix(pi/2, 1, 0, 0)

# Find what happens when you rotate (2, 0, 0) by 45 degrees about the y axis:

x <- asHomogeneous(c(2, 0, 0))
y <- x \%*\% rotationMatrix(pi/4, 0, 1, 0)
asEuclidean(y)

# or more simply...

rotate3d(c(2, 0, 0), pi/4, 0, 1, 0)

}
\keyword{ dynamic }

\name{rgl.pixels}
\alias{rgl.pixels}
\title{ Extract pixel information from window }
\description{
This function extracts single components of the pixel information from the 
topmost window.
}
\usage{
rgl.pixels(component = c("red", "green", "blue"), 
        viewport = par3d("viewport"), top = TRUE)
}
\arguments{  
  \item{component}{ Which component(s)? }
  \item{viewport}{ Lower left corner and size of desired region. }
  \item{top}{ Whether to bring window to top before reading. }
}
\details{
The possible components are \code{"red"}, \code{"green"}, \code{"blue"}, 
\code{"alpha"}, \code{"depth"}, and \code{"luminance"} (the sum of the three
colors).  All are scaled from 0 to 1.

Note that the luminance is kept below 1 by truncating the sum; this is the 
definition used for the \code{GL_LUMINANCE} component in OpenGL.
}
\value{
A vector, matrix or array containing the desired components.  If one component
is requested, a vector or matrix will be returned depending on the size of
block requested (length 1 dimensions are dropped); if more, an array, whose last dimension
is the list of components.
}
\author{ Duncan Murdoch }
\seealso{ \code{\link{rgl.snapshot}} to write a copy to a file,
\code{demo("stereo")} for functions that make use of this to draw
a random dot stereogram and an anaglyph. }
\examples{
example(surface3d)
depth <- rgl.pixels(component = "depth")
if (length(depth) && is.matrix(depth)) # Protect against empty or single pixel windows
    contour(depth)
}
\keyword{ dynamic }

\name{vertexControl}
\alias{vertexControl}
\title{
Set attributes of vertices
}
\description{
This is a function to produce actions in a web display.  A
\code{\link{playwidget}} or Shiny
input control (e.g. a \code{\link[shiny]{sliderInput}} control)
sets a value which controls attributes of a selection of vertices.
}
\usage{
vertexControl(value = 0, values = NULL, vertices = 1, attributes,
              objid = tagged3d(tag), tag,
              param = seq_len(NROW(values)) - 1, interp = TRUE)
}
\arguments{
  \item{value}{The value to use for input (typically \code{input$value}
  in a Shiny app.)  Not needed with \code{\link{playwidget}}.}
  \item{values}{
A matrix of values, each row corresponding to an input value.
}
  \item{vertices}{
Which vertices are being controlled?  Specify \code{vertices} as a number from 1 to
the number of vertices in the \code{objid}.
}
  \item{attributes}{A vector of attributes of a vertex,
  from \code{c("x", "y", "z", "red", "green", "blue", "alpha", "nx", "ny", "nz",
               "radii", "ox", "oy", "oz", "ts", "tt", "offset")}.  See
               Details.}
  \item{objid}{
A single RGL object id.
}
  \item{tag}{
An alternate way to specify \code{objid}.
  }
  \item{param}{
Parameter values corresponding to each row of \code{values}.
}
  \item{interp}{
Whether to interpolate between rows of \code{values}.
}
}
\details{
This function
modifies attributes of vertices in a single
object.  The \code{attributes} are properties of each vertex
in a scene; not all are applicable to all objects. In order,
they are: coordinates of the vertex \code{"x", "y", "z"}, color
of the vertex \code{"red", "green", "blue", "alpha"}, normal at the vertex
\code{"nx", "ny", "nz"}, radius of a sphere at the vertex
\code{"radius"}, origin within a texture \code{"ox", "oy"} and
perhaps \code{"oz"}, texture coordinates \code{"ts", "tt"}.

Planes are handled specially.  The coefficients \code{a, b, c}
in the \code{\link{planes3d}} or \code{\link{clipplanes3d}} 
specification are controlled 
using \code{"nx", "ny", "nz"}, and \code{d} is handled as
\code{"offset"}.  The \code{vertices} argument is interpreted
as the indices of the planes when these attributes are set.

If only one attribute of one vertex is specified, \code{values} may
be given as a vector and will be treated as a one-column matrix.  Otherwise
\code{values} must be given as a matrix with \code{ncol(values) == max(length(vertices),
length(attributes))}.  The \code{vertices} and \code{attributes} vectors will be recycled to the same length, and entries from column \code{j} of \code{values} will be applied to 
vertex \code{vertices[j]}, attribute \code{attributes[j]}.

The \code{value} argument is translated into a row (or two rows if
\code{interp = TRUE}) of \code{values}
by finding its location in \code{param}.
}
\value{
A list of class \code{"rglControl"} of cleaned up parameter
values, to be used in an RGL widget.
}
\author{
Duncan Murdoch
}

\examples{
  saveopts <- options(rgl.useNULL = TRUE)

  theta <- seq(0, 6*pi, length.out = 100)
  xyz <- cbind(sin(theta), cos(theta), theta)
  plot3d(xyz, type="l")
  id <- spheres3d(xyz[1,,drop=FALSE], col="red")

  widget <- rglwidget(width=500, height=300) \%>\%
  playwidget(vertexControl(values=xyz,
                           attributes=c("x", "y", "z"),
                           objid = id, param=1:100),
             start = 1, stop = 100, rate=10)
  if (interactive() || in_pkgdown_example())
    widget
  options(saveopts)
}

\name{rgl.Sweave}
\alias{rgl.Sweave}
\alias{rgl.Sweave.off}
\alias{Sweave.snapshot}
\title{
Integrating RGL with Sweave
}
\description{
As of \R 2.13.0, it is possible to include RGL graphics into
a \link{Sweave} document.  These functions support that
integration.
}
\usage{
Sweave.snapshot()
rgl.Sweave(name, width, height, options, ...)
rgl.Sweave.off()
}
\arguments{
  \item{name, width, height, options, ...}{
These arguments are passed by \code{\link{Sweave}} to \code{rgl.Sweave}
when it opens the device.
}
}
\details{
The \code{rgl.Sweave} function is not normally called by the user.  The user
specifies it as the graphics driver when opening the code chunk, e.g. by
using

\preformatted{<<fig = TRUE, pdf = FALSE, grdevice = rgl.Sweave, resolution = 100>>=}

When the RGL device is closed at the end of the code chunk,
\code{rgl.Sweave.off()} will be called automatically.  It
will save a snapshot
of the last image (by default in \file{.png} format) for inclusion in the 
Sweave document and (by default) close the device.  
Alternatively, the \code{Sweave.snapshot()} function
can be called to save the image before the end of the chunk.  Only one
snapshot will be taken per chunk.

Several chunk options are used by the \code{rgl.Sweave} device:
\describe{
\item{stayopen}{(default \code{FALSE}).  If \code{TRUE} then the RGL
device will \emph{not} be closed at the end of the chunk,
instead a call to \code{Sweave.snapshot()}
will be used if it has not been called explicitly.  Subsequent chunks can add
to the scene.}
\item{outputtype}{(default \code{png}).  The output may be specified 
as \code{outputtype = pdf} or \code{outputtype = eps} instead, in which case
the \code{\link{rgl.postscript}} function will be used to write output in the
specified format.  Note that  \code{\link{rgl.postscript}} has limitations
and does not always render scenes correctly.}
\item{delay}{(default 0.1).  After creating the display window, \code{\link{Sys.sleep}}
will be called to delay this many seconds, to allow the display system to 
initialize.  This is needed in X11 systems which open the display
asynchronously.  If the default time is too short, \code{rgl.Sweave} may
falsely report that the window is too large to open.}
}
}

\note{
We recommend turning off all other graphics drivers in a chunk that uses
\code{grdevice = rgl.Sweave}.  The RGL functions do not write to
a standard graphics device.
}

\note{
The \pkg{rgl} package relies on your graphics hardware to render OpenGL scenes,
and the default \file{.png} output copies a bitmap from the hardware device.  All 
such devices have limitations on the size of the bitmap, but they do not 
always signal these limitations in a way that RGL will detect.  If you find
that images are not being produced properly, try reducing the size using 
the \code{resolution}, \code{width} or \code{height} chunk options.
}

\value{
These functions are called for their side effects.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{RweaveLatex}} for a description of alternate graphics drivers
in Sweave, and standard options that can be used in code chunks.  

\code{\link{hook_rgl}}
and \code{\link{hook_webgl}} allow fixed or interactive RGL
scenes to be embedded in \pkg{knitr} 
documents.
}
\keyword{ utilities }

\name{arc3d}
\alias{arc3d}
\title{
Draw arcs
}
\description{
Given starting and ending points on a sphere and the center of the
sphere, draw the
great circle arc between the starting and ending points.  If
the starting and ending points have different radii, a segment of a
logarithmic spiral will join them.
}
\usage{
arc3d(from, to, center, radius, n, circle = 50, base = 0,
plot = TRUE, ...)
}
\arguments{
  \item{from}{
One or more points from which to start arcs.
}
  \item{to}{
One or more destination points.
}
  \item{center}{
One or more center points.
}
  \item{radius}{
If not missing, a vector of length \code{n} giving the
radii at each point between \code{from} and \code{to}.
If missing, the starting and ending points will be joined by
a logarithmic spiral.
}
  \item{n}{
If not missing, how many segments to use between the first and last point.  If missing, a value will be calculated based on the
angle between starting and ending points as seen from the center.
}
  \item{circle}{
How many segments would be used if the arc went completely around
a circle.
}
  \item{base}{
See Details below.
  }
  \item{plot}{
Should the arcs be plotted, or returned as a matrix?
  }
  \item{\dots}{
Additional parameters to pass to \code{\link{points3d}}.
}
}
\details{
If any of \code{from}, \code{to} or \code{center} is an n by 3
matrix with n > 1, multiple arcs will be drawn by recycling
each of these parameters to the number of rows of the longest
one.

If the vector lengths of \code{from - center} and
\code{to - center} differ, then instead of a spherical
arc, the function will draw a segment of a logarithmic spiral joining
the two points.

By default, the arc is drawn along the shortest great 
circle path from \code{from} to \code{to}, but the
\code{base} parameter can be used to modify this.  
If \code{base = 1} is used, the longer arc will be followed.
Larger positive integer values will result in \code{base - 1}
loops in that direction completely around the sphere.  
Negative values will draw the curve in the same direction
as the shortest arc, but with \code{abs(base)} full loops.
It doesn't make much sense to ask for such loops unless
the radii of \code{from} and \code{to} differ, because
spherical arcs would overlap.
Normally the \code{base} parameter is left at its default
value of \code{0}.  

When \code{base} is non-zero, the curve will be constructed in multiple
pieces, between \code{from}, \code{to}, \code{-from} and \code{-to},
for as many steps as necessary.  If \code{n} is specified, it
will apply to each of these pieces.
}
\value{
If \code{plot = TRUE}, called mainly for the side effect of drawing arcs.  Invisibly
returns the object ID of the collection of arcs. 

If \code{plot = FALSE}, returns a 3 column matrix containing
the points that would be drawn as the arcs.
}
\author{
Duncan Murdoch
}

\examples{
normalize <- function(v) v/sqrt(sum(v^2))

# These vectors all have the same length

from <- t(apply(matrix(rnorm(9), ncol = 3), 1, normalize))
to <- normalize(rnorm(3))
center <- c(0, 0, 0)

open3d()
spheres3d(center, radius = 1, col = "white", alpha = 0.2)

arc3d(from, to, center, col = "red")
arc3d(from, 2*to, center, col = "blue")

text3d(rbind(from, to, center, 2*to), 
       texts = c(paste0("from", 1:3), "to", "center", "2*to"),
       depth_mask = FALSE, depth_test = "always")
}

\name{.check3d}
\alias{.check3d}
\alias{check3d}
\title{
Check for an open RGL window.
}
\description{
Mostly for internal use, this function returns the current device
number if one exists, or opens a new device and returns that.
}
\usage{
.check3d()
}
\value{
The device number of an RGL device.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{open3d}}
}
\examples{
rgl.dev.list()
.check3d()
rgl.dev.list()
.check3d()
rgl.dev.list()
close3d()
}

\name{subscene3d}
\alias{subscene3d}
\alias{newSubscene3d}
\alias{currentSubscene3d}
\alias{useSubscene3d}
\alias{addToSubscene3d}
\alias{delFromSubscene3d}
\alias{gc3d}
\title{
Create, select or modify a subscene
}
\description{
This creates a new subscene, or selects one by \code{id} value, or adds objects to one.
}
\usage{
newSubscene3d(viewport = "replace", 
              projection = "replace", 
              model = "replace",
              mouseMode = "inherit", 
              parent = currentSubscene3d(), 
              copyLights = TRUE, 
              copyShapes = FALSE, 
              copyBBoxDeco = copyShapes, 
              copyBackground = FALSE, newviewport,
              ignoreExtent)
currentSubscene3d(dev = cur3d())
useSubscene3d(subscene)
addToSubscene3d(ids = tagged3d(tags), tags, subscene = currentSubscene3d())
delFromSubscene3d(ids = tagged3d(tags), tags, subscene = currentSubscene3d())
gc3d(protect = NULL)
}
\arguments{
  \item{viewport, projection, model, mouseMode}{
How should the new subscene be embedded?  Possible values are
\code{c("inherit", "modify", "replace")}.  See Details below.
}
  \item{parent}{
The parent subscene (defaults to the current subscene).
}
  \item{copyLights, copyShapes, copyBBoxDeco, copyBackground}{
Whether lights, shapes, bounding box decorations and background should be copied to the new subscene.
}
  \item{newviewport}{
Optionally specify the new subscene's viewport (in pixels).
}
  \item{ignoreExtent}{
Whether to ignore the subscene's bounding box when calculating the parent bounding
box.  Defaults to \code{TRUE} if \code{model} is not \code{"inherit"}.
}
  \item{dev}{
Which RGL device to query for the current subscene.
}
  \item{subscene}{
Which subscene to use or modify.
}
  \item{ids}{
A vector of integer object ids to add to the subscene.
}
  \item{tags}{
Alternate way to specify \code{ids}.  Ignored if \code{ids} is
given.
  }
  \item{protect}{
Object ids to protect from this garbage collection.
}
}
\details{
The \pkg{rgl} package allows multiple windows to be open; each one corresponds to a \dQuote{scene}.
Within each scene there are one or more \dQuote{subscenes}.  Each subscene corresponds
to a rectangular region in the window, and may have its own projection, transformation and behaviour in response to 
the mouse.

There is always a current subscene:  most graphic operations make changes there, e.g. by adding
an object to it.

The scene \dQuote{owns} objects; \code{addToSubscene3d} and \code{delFromSubscene3d} 
put their ids into or remove them from the list being displayed within a particular subscene.
The \code{gc3d} function deletes objects from the scene if they are not visible in any
subscene, unless they are protected by having their id included in \code{protect}.

The \code{viewport}, \code{projection} and \code{model} parameters each have three possible settings:
\code{c("inherit", "modify", "replace")}.  \code{"inherit"} means that the corresponding value
from the parent subscene will be used.  \code{"replace"} means that the new subscene will
have its own value of the value, independent of its parent.  \code{"modify"} means that the 
child value will be applied first, and then the parent value will be applied.  For viewport, this
means that if the parent viewport is changed, the child will maintain its relative position.  For
the two matrices, \code{"modify"} is unlikely to give satisfactory results, but it is available
for possible use.  

The \code{mouseMode} parameter 
can only be one of \code{c("inherit", "replace")}.  If
it is \code{"inherit"}, the subscene will use the mouse
controls of the parent, and any change to them will
affect the parent and all children that inherit from it.
This is the behaviour that was present before \pkg{rgl}
version 0.100.13.  If it is \code{"replace"}, then
it will receive a copy of the parent mouse controls,
but modifications to them will affect only this 
subscene, not the parent.  Note that this is orthogonal
to the \code{\link{par3d}("listeners")} setting:
if another subscene is listed as a listener, it will respond
to mouse actions using the same mode as the one receiving
them.

The \code{viewport} parameter controls the rectangular region in which the subscene is displayed.
It is specified using \code{newviewport} (in pixels relative to the whole window), or set to 
match the parent viewport.

The \code{projection} parameter controls settings corresponding to the observer.  These
include the field of view and the zoom; they also include the position of the observer relative to
the model.  The \code{par3d("projMatrix")} matrix is determined by the projection.

The \code{model} parameter controls settings corresponding to the model.  Mouse rotations affect
the model, as does scaling.  The \code{par3d("modelMatrix")} matrix is determined by these as
well as by the position of the observer (since OpenGL assumes that the observer is at
(0, 0, 0) after the MODELVIEW transformation).  Only those parts concerning the model are
inherited when \code{model} specifies inheritance, the observer setting is controlled by
\code{projection}.

If \code{copyBackground} is \code{TRUE}, the background of the newly created child will
overwrite anything displayed in the parent subscene, regardless of depth.
}
\value{
If successful, each function returns the object id of the subscene, with the exception of
\code{gc3d}, which returns the count of objects which have been deleted, 
and \code{useSubscene3d}, which returns the previously active subscene id.
}
\author{
Duncan Murdoch and Fang He.
}
\seealso{
\code{\link{subsceneInfo}} for information about a subscene, 
\code{\link{mfrow3d}} and \code{\link{layout3d}} to set up 
multiple panes of subscenes.
}
\examples{

# Show the Earth with a cutout by using clipplanes in subscenes

lat <- matrix(seq(90, -90, length.out = 50)*pi/180, 50, 50, byrow = TRUE)
long <- matrix(seq(-180, 180, length.out = 50)*pi/180, 50, 50)

r <- 6378.1 # radius of Earth in km
x <- r*cos(lat)*cos(long)
y <- r*cos(lat)*sin(long)
z <- r*sin(lat)

open3d()
obj <- surface3d(x, y, z, col = "white", 
       texture = system.file("textures/worldsmall.png", package = "rgl"), 
       specular = "black", axes = FALSE, box = FALSE, xlab = "", ylab = "", zlab = "",
       normal_x = x, normal_y = y, normal_z = z)
       
cols <- c(rep("chocolate4", 4), rep("burlywood1", 4), "darkgoldenrod1")
rs <- c(6350, 5639, 4928.5, 4207, 3486, 
                         (3486 + 2351)/2, 2351, (2351 + 1216)/2, 1216)
for (i in seq_along(rs)) 
  obj <- c(obj, spheres3d(0, 0, col = cols[i], radius = rs[i]))
  
root <- currentSubscene3d()

newSubscene3d("inherit", "inherit", "inherit", copyShapes = TRUE, parent = root)
clipplanes3d(1, 0, 0, 0)

newSubscene3d("inherit", "inherit", "inherit", copyShapes = TRUE, parent = root)
clipplanes3d(0, 1, 0, 0)

newSubscene3d("inherit", "inherit", "inherit", copyShapes = TRUE, parent = root)
clipplanes3d(0, 0, 1, 0)

# Now delete the objects from the root subscene, to reveal the clipping planes
useSubscene3d(root)
delFromSubscene3d(obj)
}
\keyword{ graphics }

\name{toggleWidget}
\alias{toggleWidget}
\title{
An HTML widget to toggle display of elements of a scene
}
\description{
This function produces a button in an HTML scene
that will toggle the display of items in the scene.
}
\usage{
toggleWidget(sceneId, 
             ids = tagged3d(tags), tags = NULL, hidden = integer(),
             subscenes = NULL, 
             label, 
             ...)
}
\arguments{
  \item{sceneId}{
The HTML id of the RGL scene being controlled,
or an object as in \code{\link{playwidget}}.
}
  \item{ids, hidden}{
The RGL id numbers of the objects to toggle.  Those in \code{ids}
are initially shown; those in \code{hidden} are initially hidden.
}
  \item{tags}{
  Alternate way to specify \code{ids}.  Ignored if \code{ids} is
given.
  }
  \item{subscenes}{
The subscenes in which to toggle the objects.
}
  \item{label}{
The label to put on the button.  The default is set
from the expression passed to \code{ids} or the value of \code{tags}. 
}
  \item{\dots}{
Additional arguments to pass to \code{\link{playwidget}}.
}
}
\details{
Like \code{\link{playwidget}}, this function is designed
to work within the \pkg{htmlwidgets} framework.  
If the value is printed, the button will be inserted
into the output.

It is also designed to work with \pkg{magrittr}-style
pipes:  the result of \code{\link{rglwidget}} or
other widgets can be piped into it to add it 
to a display.  It can also appear first in the 
pipeline, if \code{sceneId} is set to \code{NA}.
}
\value{
A widget suitable for use in an \pkg{Rmarkdown}-generated web
page, or elsewhere.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{toggleButton}} for the older style
of HTML control.
}
\examples{
theplot <- plot3d(rnorm(100), rnorm(100), rnorm(100), col = "red")
widget <- rglwidget(height = 300, width = 300) \%>\%
  toggleWidget(theplot["data"], 
               hidden = theplot[c("xlab", "ylab", "zlab")], 
               label = "Points")
if (interactive() || in_pkgdown_example())
  widget
}

\name{setUserShaders}
\alias{setUserShaders}
\alias{getShaders}
\title{
Set user-defined shaders for RGL objects, or get
shaders.
}
\description{
\code{setUserShaders} sets user-defined shaders (programs written in GLSL)
for customized display of RGL objects.  Currently
only supported in WebGL displays, as the regular
displays do not support GLSL.  \code{getShaders} gets
the user defined shader, or if it is not present, the
automatically generated one.
}
\usage{
setUserShaders(ids, vertexShader = NULL, fragmentShader = NULL, 
               attributes = NULL, uniforms = NULL, textures = NULL,
               scene = scene3d(minimal), minimal = TRUE)

getShaders(id, scene = scene3d(minimal), minimal = TRUE)
}
\arguments{
  \item{ids, id}{
Which objects should receive the shaders, or which object
should be queried?
}
  \item{vertexShader, fragmentShader}{
The vertex and fragment shader source code.  If \code{NULL}, the
automatically generated shader will be used instead.
}
  \item{attributes}{
A named list of \dQuote{attributes} to attach to each vertex.
}
  \item{uniforms}{
A named list of \dQuote{uniforms}.
}
  \item{textures}{
A named list of textures.
  }
  \item{scene}{
A \code{\link{scene3d}} object to work with.
}
  \item{minimal}{
See \code{\link{scene3d}}.  
}
}
\details{
Modern versions of OpenGL work with \dQuote{shaders}, 
programs written to run on the graphics processor.  The
vertex shader does the calculations to move vertices and
set their intrinsic colours.  The fragment shader
computes how each pixel in the display will be shown,
taking into account lighting, material properties, etc.
(More precisely, it does the computation for each \dQuote{fragment}; a fragment is a pixel within an object to display.
There may be many objects at a particular location, and
each will result in a fragment calculation unless culled
by z-buffering or being discarded in some other way.)

Normally the WebGL Javascript code uses the default
shaders stored in 
\code{system.file("htmlwidgets/lib/rglClass/shaders", 
package = "rgl")}.  This function allows them to be written
by hand, for testing new features, hand optimization, etc.
The defines used by the default shaders will also be prepended
to user shaders, which can use them for customization on
an object-by-object basis.

The names used for the \code{attributes}, \code{uniforms}
and \code{textures} should match names in the shaders
for corresponding variables.  (The texture names should be
names of \code{uniform sampler2D} variables.)
}
\note{
The \code{getShaders} function requires the \pkg{V8} package
to extract auto-generated shaders, since the defines
are generated by Javascript code.
}
\value{
A modified version of the \code{scene}.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{rglwidget}} for display of the scene in WebGL.
}
\examples{
open3d()
id <- shade3d(octahedron3d(), col = "red")

# For each triangle, set weights on the 3 vertices.
# This will be replicated to the appropriate size in Javascript.
wts <- diag(3)

# This leaves out the centres of each face
vs <- "
  attribute vec3 aPos;
  attribute vec4 aCol;
  uniform mat4 mvMatrix;
  uniform mat4 prMatrix;
  varying vec4 vCol;
  varying vec4 vPosition;
  attribute vec3 aNorm;
  uniform mat4 normMatrix;
  varying vec3 vNormal;
  attribute vec3 wts;
  varying vec3 vwts;
  void main(void) {
    vPosition = mvMatrix * vec4(aPos, 1.);
    gl_Position = prMatrix * vPosition;
    vCol = aCol;
    vNormal = normalize((normMatrix * vec4(aNorm, 1.)).xyz);
    vwts = wts;
  }
"  
fs <- "
  #ifdef GL_ES
  precision highp float;
  #endif
  varying vec4 vCol; // carries alpha
  varying vec4 vPosition;
  varying vec3 vNormal;
  uniform mat4 mvMatrix;
  uniform vec3 emission;
  uniform float shininess;
  uniform vec3 ambient[NLIGHTS];
  uniform vec3 specular[NLIGHTS]; // light*material
  uniform vec3 diffuse[NLIGHTS];
  uniform vec3 lightDir[NLIGHTS];
  uniform bool viewpoint[NLIGHTS];
  uniform bool finite[NLIGHTS];
  varying vec3 vwts;
  uniform vec2 wtrange;
  void main(void) {
    float minwt = min(vwts.x, min(vwts.y, vwts.z));
    if (minwt < wtrange.x || minwt > wtrange.y) discard;
    vec3 eye = normalize(-vPosition.xyz);
    vec3 lightdir;
    vec4 colDiff;
    vec3 halfVec;
    vec4 lighteffect = vec4(emission, 0.);
    vec3 col;
    float nDotL;
    vec3 n = normalize(vNormal);
    n = -faceforward(n, n, eye);
    colDiff = vec4(vCol.rgb * diffuse[0], vCol.a);
    lightdir = lightDir[0];
    if (!viewpoint[0])
      lightdir = (mvMatrix * vec4(lightdir, 1.)).xyz;
    if (!finite[0]) {
      halfVec = normalize(lightdir + eye);
    } else {
      lightdir = normalize(lightdir - vPosition.xyz);
      halfVec = normalize(lightdir + eye);
    }
    col = ambient[0];
    nDotL = dot(n, lightdir);
    col = col + max(nDotL, 0.) * colDiff.rgb;
    col = col + pow(max(dot(halfVec, n), 0.), shininess) * specular[0];
    lighteffect = lighteffect + vec4(col, colDiff.a);
    gl_FragColor = lighteffect;
  }
"
x <- setUserShaders(id, vs, fs, attributes = list(wts=wts),
                    uniforms = list(wtrange = c(-0.01, 0.15)))
if (interactive() || in_pkgdown_example())
  rglwidget(x)
}

\name{points3d}
\alias{points3d}
\alias{lines3d}
\alias{segments3d}
\alias{triangles3d}
\alias{quads3d}
\title{add primitive set shape}
\description{
Adds a shape node to the current scene
}
\usage{
points3d(x, y = NULL, z = NULL,  ...)
lines3d(x, y = NULL, z = NULL,  ...)
segments3d(x, y = NULL, z = NULL, ...)
triangles3d(x, y = NULL, z = NULL, ...)
quads3d(x, y = NULL, z = NULL, ...)
}
\arguments{
  \item{x, y, z}{coordinates. Any reasonable way of defining the
    coordinates is acceptable.  See the function \code{\link[grDevices]{xyz.coords}}
    for details.}
  \item{ ... }{Material properties (see \code{\link{rgl.material}}).
  For normals use \code{normals} and for
  texture coordinates use \code{texcoords}; see \code{\link{rgl.primitive}}
  for details.}  
}
\details{

The functions \code{points3d}, \code{lines3d}, \code{segments3d},
\code{triangles3d} and \code{quads3d} add points, joined lines, line segments,
filled triangles or quadrilaterals to the plots.  They correspond to the OpenGL types
\code{GL_POINTS, GL_LINE_STRIP, GL_LINES, GL_TRIANGLES} and \code{GL_QUADS} respectively.  

Points are taken in pairs by \code{segments3d}, triplets as the vertices
of the triangles, and quadruplets for the quadrilaterals.  Colors are applied vertex by vertex; 
if different at each end of a line segment, or each vertex of a polygon, the colors
are blended over the extent of the object.  Polygons
must be non-degenerate and quadrilaterals must be entirely 
in one plane and convex, or the results are undefined.

These functions call the lower level functions \code{\link{rgl.points}}, \code{\link{rgl.linestrips}},
and so on, and are provided for convenience.  

The appearance of the new objects are defined by the material properties.
See \code{\link{rgl.material}} for details.

The two principal differences between the \code{rgl.*} functions and
the \code{*3d} functions are that the former set all unspecified
material properties to defaults, whereas the latter use current values
as defaults; the former make persistent changes to material properties
with each call, whereas the latter make temporary changes only for the
duration of the call.  

}
\value{
   Each function returns the integer object ID of the shape that
   was added to the scene.  These can be passed to \code{\link{pop3d}}
   to remove the object from the scene.
}
\author{
  Ming Chen and Duncan Murdoch
}

\examples{
# Show 12 random vertices in various ways. 

M <- matrix(rnorm(36), 3, 12, dimnames = list(c('x', 'y', 'z'), 
                                       rep(LETTERS[1:4], 3)))

# Force 4-tuples to be convex in planes so that quads3d works.

for (i in c(1, 5, 9)) {
    quad <- as.data.frame(M[, i + 0:3])
    coeffs <- runif(2, 0, 3)
    if (mean(coeffs) < 1) coeffs <- coeffs + 1 - mean(coeffs)
    quad$C <- with(quad, coeffs[1]*(B - A) + coeffs[2]*(D - A) + A)
    M[, i + 0:3] <- as.matrix(quad)
}

open3d()

# Rows of M are x, y, z coords; transpose to plot

M <- t(M)
shift <- matrix(c(-3, 3, 0), 12, 3, byrow = TRUE)

points3d(M)
lines3d(M + shift)
segments3d(M + 2*shift)
triangles3d(M + 3*shift, col = 'red')
quads3d(M + 4*shift, col = 'green')  
text3d(M + 5*shift, texts = 1:12)

# Add labels

shift <- outer(0:5, shift[1, ])
shift[, 1] <- shift[, 1] + 3
text3d(shift, 
       texts = c('points3d', 'lines3d', 'segments3d',
         'triangles3d', 'quads3d', 'text3d'),
       adj = 0)
 rgl.bringtotop()
}
\keyword{dynamic}

\name{extrude3d}
\alias{extrude3d}
\title{
Generate extrusion mesh
}
\description{
Given a two-dimensional polygon, this generates a three-dimensional extrusion
of the shape by triangulating the polygon and creating a cylinder with
that shape as the end faces.
}
\usage{
extrude3d(x, y = NULL, thickness = 1, smooth = FALSE, ...)
}
\arguments{
  \item{x, y}{
A polygon description in one of the forms supported by \code{\link{triangulate}}.
}
  \item{thickness}{
The extrusion will have this thickness.
}
  \item{smooth}{
logical; should normals be added so that the edges of the extrusion appear
smooth?
}
  \item{\dots}{
Other parameters to pass to \code{\link{tmesh3d}} when constructing the
mesh.
}
}
\details{
The extrusion is always constructed with the polygon in the xy plane at
\code{z = 0} and another copy at \code{z = thickness}.  Use
the transformation functions (e.g. \code{\link{rotate3d}}) to 
obtain other orientations and placements.
}
\value{
A mesh object containing a triangulation of the polygon
for each face, and quadrilaterals for the sides.
}
\author{
Duncan Murdoch
}

\seealso{
\code{\link{polygon3d}} for a simple polygon, \code{\link{triangulate}} for
the triangulation, \code{\link{turn3d}} for a solid of rotation.
}
\examples{
x <- c(1:10, 10:1)
y <- rev(c(rep(c(0, 2), 5), rep(c(1.5, -0.5), 5)))
plot(x, y, type = "n")
polygon(x, y)
open3d()
shade3d( extrude3d(x, y), col = "red" )
}
\keyword{ graphics }

\name{shinyGetPar3d}
\alias{shinyGetPar3d}
\alias{shinySetPar3d}
\alias{shinyResetBrush}
\title{
Communicate RGL parameters between R and Javascript in Shiny
}
\description{
These functions allow Shiny apps to read and write the \code{par3d}
settings that may have been modified by user interaction in the browser.

}
\usage{
shinyGetPar3d(parameters, session, subscene = currentSubscene3d(cur3d()), tag = "")
shinySetPar3d(..., session, subscene = currentSubscene3d(cur3d()))
shinyResetBrush(session, brush)
}
\arguments{
  \item{parameters}{
A character vector naming the parameters to get.
}
  \item{session}{
The Shiny session object.
}
  \item{subscene}{
The subscene to which the parameters apply.  Defaults to 
the currently active subscene in the R session.
}
  \item{tag}{
An arbitrary string or value which will be sent as part of the
response.
}
  \item{...}{
A number of \code{name = value} pairs to be modified, or a
single named list of parameters.  Entries named \code{tag}
or \code{subscene} will be ignored.
  }
  \item{brush}{The name of a Shiny input element corresponding to
  the \code{shinyBrush} argument to \code{\link{rglwidget}}.}
}
\details{
Requesting information from the browser is a complicated process.
The \code{shinyGetPar3d} function doesn't return the requested value,
it just submits a request for the value to be returned later in
\code{input$par3d}, a reactive input.  No action will result
except when a reactive observer depends on \code{input$par3d}.
See the example code below.

The \code{shinySetPar3d} function sends a message to the browser
asking it to change a particular parameter.  The change will be
made immediately, without sending the full scene to the browser,
so should be reasonably fast.
}
\value{
These functions are called for their side effects, and don't return
useful values.

The side effect of \code{shinyGetPar3d} is to cause \code{input$par3d}
to be updated sometime later.  
Besides the requested parameter values, \code{input$par3d} will
contain a copy of the \code{subscene} and \code{tag} arguments.

The side effect of \code{shinySetPar3d} is to send a message to the
browser to update its copy of the \code{par3d} parameters 
immediately.
}
\note{
R and the browser don't maintain a perfect match between the way
parameters are stored internally.  The browser version of parameters
will be returned by \code{shinyGetPar3d} and should be supplied to
\code{shinySetPar3d}.
}
\references{
\url{https://shiny.rstudio.com/articles/communicating-with-js.html}
describes the underlying mechanisms used by these two functions.
}
\seealso{The \code{\link{rglwidget}} argument \code{shinySelectionInput} allows information about mouse selections
to be returned to R.}
\author{
Duncan Murdoch
}

\examples{
if (interactive() && !in_pkgdown_example() && requireNamespace("shiny")) {
  save <- options(rgl.useNULL = TRUE)

  xyz <- matrix(rnorm(300), ncol = 3)

  app = shiny::shinyApp(
    ui = shiny::bootstrapPage(
      shiny::actionButton("redraw", "Redraw"),
      rglwidgetOutput("rglPlot")
    ),
    server = function(input, output, session) {
      # This waits until the user to click on the "redraw" 
      # button, then sends a request for the current userMatrix
      shiny::observeEvent(input$redraw, {
        shinyGetPar3d("userMatrix", session)
      })
    
      # This draws the plot whenever input$par3d changes,
      # i.e. whenever a response to the request above is
      # received.
      output$rglPlot <- renderRglwidget({
        if (length(rgl.dev.list())) close3d()
        col <- sample(colors(), 1)
        plot3d(xyz, col = col, type = "s", main = col)
        par3d(userMatrix = input$par3d$userMatrix)
        rglwidget()
      })
    })
  shiny::runApp(app)
  options(save)
}
}

\name{thigmophobe3d}
\alias{thigmophobe3d}
\title{
Find the direction away from the closest point in a 3d projection
}
\description{
Jim Lemon's \code{\link[plotrix]{thigmophobe}} function in the 
\code{\link[plotrix:plotrix-package]{plotrix}} package
computes good directions for labels in a 2D plot.  This function
does the same for a particular projection in a 3D plot by
projecting down to 2D and calling his function.
}
\usage{
thigmophobe3d(x, y = NULL, z = NULL, 
              P = par3d("projMatrix"), 
              M = par3d("modelMatrix"),
              windowRect = par3d("windowRect"))
}
\arguments{
  \item{x, y, z}{point coordinates.  Any reasonable way of defining
  the coordinates is acceptable.  See the function \code{\link[grDevices]{xyz.coords}} for details.
}
  \item{P, M, windowRect}{
The projection and modelview matrices, and the size and position of 
the display in pixels.
}
}
\details{
Since \code{thigmophobe3d} projects using fixed \code{P} and \code{M},
it will not necessarily choose good directions if the user rotates
the display or makes any other change to the projection.}

\note{
The example
below shows how to update the directions during an animation; 
I find that the moving labels are distracting, and prefer to live with 
fixed ones.
}
\value{
A vector of values from 1 to 4 to be used as the \code{pos} argument
in \code{\link{text3d}}.
}
\references{
\CRANpkg{plotrix}
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{text3d}}
}
\examples{
if (requireNamespace("plotrix", quietly = TRUE)) {
  # Simulate some data
  xyz <- matrix(rnorm(30), ncol = 3)
  
  # Plot the data first, to establish the projection
  plot3d(xyz)
  
  # Now thigmophobe3d can choose directions
  textid <- text3d(xyz, texts = 1:10, pos = thigmophobe3d(xyz))
  
  # Update the label positions during an animation
  if (interactive() && !rgl.useNULL()) {
    spin <- spin3d(rpm = 5)
    f <- function(time) {
      par3d(skipRedraw = TRUE)
      on.exit(par3d(skipRedraw = FALSE))
      pop3d(id = textid)
      # Need to rotate before thigmophobe3d is called
      result <- spin(time)
      par3d(userMatrix = result$userMatrix)
      textid <<- text3d(xyz, texts = 1:10, pos = thigmophobe3d(xyz))
      result
    }
    play3d(f, duration = 5)
  } else
    textid   # just print the static display
}
}

\name{triangulate}
\alias{triangulate}
\title{
Triangulate a two-dimensional polygon
}
\description{
This algorithm decomposes a general polygon into simple
polygons and uses the \dQuote{ear-clipping} algorithm to triangulate it.
Polygons with holes are supported.
}
\usage{
triangulate(x, y = NULL, z = NULL, random = TRUE, plot = FALSE, partial = NA)
}
\arguments{
  \item{x, y, z}{
Coordinates of a two-dimensional polygon in a format supported by \code{\link{xyz.coords}}.  
See Details for a description of proper input and
how \code{z} is handled.
}
  \item{random}{
Currently ignored, the triangulation
is deterministic.
}
  \item{plot}{
Whether to plot the triangulation; mainly for debugging purposes.
}
  \item{partial}{
Currently ignored. Improper input will lead to undefined 
results.
}
}
\details{
Normally \code{triangulate} looks only at the \code{x} and \code{y}
coordinates.  However, if one of those is constant, it is replaced
with the \code{z} coordinate if present.

The algorithm works as follows.  First, it breaks the polygon into 
pieces separated by \code{NA} values in \code{x} or \code{y}.
Each of these pieces should be a simple, non-self-intersecting
polygon, not intersecting the other pieces. 
(Though some minor exceptions to this rule may work, none
are guaranteed).  The nesting of these pieces is
determined:  polygons may contain holes, and the
holes may contain other polygons.  

Vertex order around the polygons does not affect
the results:  whether a polygon is on the outside 
or inside of a region is determined by nesting.

Polygons should not repeat vertices.  An attempt is 
made to detect if the final vertex matches the first
one.  If so, it will be deleted with a warning.

The \dQuote{outer} polygon(s) are then merged with the
polygons that they immediately contain, and each of these
pieces is triangulated using the ear-clipping algorithm from the references.

Finally, all the triangulated pieces are put together into one
result.
}
\value{
A three-by-n array giving the indices of the vertices of 
each triangle.  (No vertices are added; only the original
vertices are used in the triangulation.)

The array has an integer vector attribute \code{"nextvert"}
with one entry per vertex, giving the index of the next 
vertex to proceed counter-clockwise around outer
polygon boundaries, clockwise around inner boundaries.
}
\references{
This function uses the C++ version of the
\code{earcut} library from \url{https://github.com/mapbox/earcut.hpp}.
}
\author{
R wrapper code written by Duncan Murdoch; the
\code{earcut} library has numerous authors.
}
\note{
Not all inputs will succeed, though inputs satisfying
the rules listed in the Details section should.
}
\seealso{
\code{\link{extrude3d}} for a solid extrusion of a polygon, \code{\link{polygon3d}} for
a flat display; both use \code{triangulate}.
}
\examples{
theta <- seq(0, 2*pi, length.out = 25)[-25]
theta <- c(theta, NA, theta, NA, theta, NA, theta, NA, theta)
r <- c(rep(1.5, 24), NA, rep(0.5, 24), NA, rep(0.5, 24), NA, rep(0.3, 24), NA, rep(0.1, 24))
dx <- c(rep(0, 24), NA, rep(0.6, 24), NA, rep(-0.6, 24), NA, rep(-0.6, 24), NA, rep(-0.6, 24))
x <- r*cos(theta) + dx
y <- r*sin(theta)
plot(x, y, type = "n")
polygon(x, y)
triangulate(x, y, plot = TRUE)
open3d()
polygon3d(x, y, x - y, col = "red")
}
\keyword{ graphics }

\name{rglwidget}
\alias{rglwidget}
\alias{rgl.printRglwidget}
\title{
An htmlwidget to hold an RGL scene
}
\description{
The \pkg{htmlwidgets} package provides a framework for embedding
graphical displays in HTML documents of various types.  This function
provides the necessities to embed an RGL scene in one.
}
\usage{
rglwidget(x = scene3d(minimal), width = figWidth(), height = figHeight(),
          controllers = NULL,
          elementId = NULL, 
          reuse = FALSE,
          webGLoptions = list(preserveDrawingBuffer = TRUE), 
          shared = NULL, minimal = TRUE, 
          webgl, snapshot, 
          shinyBrush = NULL, 
          ...,
          oldConvertBBox = FALSE)
}
\arguments{
  \item{x}{
An RGL scene produced by the \code{\link[rgl]{scene3d}} function.
}
  \item{width, height}{
The width and height of the display in pixels.
}
  \item{controllers}{Names of \code{\link{playwidget}} objects
associated with this scene, or objects (typically piped in).  See Details below.
}
  \item{snapshot,webgl}{Control of mode of display of scene.
  See Details below.
}
  \item{elementId}{The id to use on the HTML \code{div}
component that will hold the scene.
}
  \item{reuse}{Ignored.  See Details below.
}
  \item{webGLoptions}{A list of options to pass
  to WebGL when the drawing context is created.  See the Details below.}
  \item{shared}{An object produced by \code{\link{rglShared}}, or a list of such objects.}
  \item{minimal}{Should attributes be skipped if they currently have
  no effect?  See \code{\link{scene3d}}.}
  \item{shinyBrush}{The name of a Shiny \code{input} element
  to receive information about mouse selections.}
  \item{oldConvertBBox}{See Details below.}
  \item{...}{Additional arguments
to pass to \code{htmlwidgets::\link{createWidget}}.}
}
\details{
This produces a WebGL version of an RGL scene using the \pkg{htmlwidgets}
framework.  This allows display of the scene in the RStudio IDE, a browser, an \pkg{rmarkdown} document
or in a \pkg{shiny} app.

\code{options(rgl.printRglwidget = TRUE)} will cause
\code{rglwidget()} to be called and displayed
when the result of an RGL call that changes the
scene is printed.

In RMarkdown or in standalone code, you can use a \pkg{magrittr}-style
\dQuote{pipe} command to join an \code{rglwidget} with a
\code{\link{playwidget}} or \code{\link{toggleWidget}}.  If the control widget comes
first, it should be piped into the \code{controllers}
argument.  If the \code{rglwidget} comes first, it
can be piped into the first argument of \code{playwidget} or \code{toggleWidget}.

In earlier versions, the \code{reuse} argument let one output scene share data from earlier ones.  This is no longer supported.

If \code{elementId} is \code{NULL} and we are not in a Shiny app,
\code{elementId} is set to a random value to facilitate re-use
of information.

To save the display to a file, use \code{htmlwidgets::\link{saveWidget}}.
This requires \command{pandoc} to be installed.  
For a snapshot, you can use
\code{htmltools::save_html(img(src=rglwidget(snapshot=TRUE)), file = ...)}.

The \code{webGLoptions} argument is a list which will
be passed when the WebGL context is created.  See
the WebGL 1.0 specification on \url{https://registry.khronos.org/webgl/specs/} for possible
settings.  The default in \code{rglwidget} differs
from the WebGL default by setting \code{preserveDrawingBuffer = TRUE} in order to allow other tools to read
the image, but please note that some implementations
of WebGL contain bugs with this setting.  We have
attempted to work around them, but may change our
default in the future if this proves unsatisfactory.

The \code{webgl} argument controls
whether a dynamic plot is displayed in HTML.  In LaTeX
and some other formats
dynamic plots can't be
displayed, so if the \code{snapshot} argument is \code{TRUE},
\code{webgl} must be \code{FALSE}.  (In previous versions
of the \pkg{rgl} package, both \code{webgl} and \code{snapshot} could be
\code{TRUE}; that hasn't worked for a while and is no longer
allowed as of version 0.105.6.)

The \code{snapshot} argument controls whether a snapshot is
displayed:  it must be \code{!webgl} if both are specified.

Prior to \pkg{rgl} 0.106.21, \code{rglwidget} converted 
bounding box decorations into separate objects: a box, text
for the labels, segments for the ticks.  By default it now
generates these in Javascript, allowing axis labels to move as
they do in the display in \R.  If you prefer the old conversion,
set \code{oldConvertBBox = TRUE}.
}
\section{Shiny specifics}{

This widget is designed to work with Shiny for interactive
displays linked to a server running R.

In a Shiny app, there will often be one or more
\code{\link{playwidget}} objects in the app, taking input from
the user.  In order to be sure that the initial value of the user control
is reflected in the scene, you should list all players in the
\code{controllers} argument.  See the sample application in
\code{system.file("shinyDemo", package = "rglwidget")} for an example.

In Shiny, it is possible to find out information about mouse selections
by specifying the name of an \code{input} item in the
\code{shinyBrush} argument.  For example, with
\code{shinyBrush = "brush3d"}, each change
to the mouse selection will send data to \code{input$brush3d} in an
object of class \code{"rglMouseSelection"} with the
following components:
\describe{
\item{subscene}{The ID of the subscene where the mouse is selecting.}
\item{state}{Either \code{"changing"} or \code{"inactive"}.}
\item{region}{The coordinates of the corners of the selected region in the window,
in order \code{c(x1, y1, x2, y2)}.}
\item{model, proj, view}{The model matrix, projection matrix and viewport in effect at that location.}
}
This object can be used as the first argument to
\code{\link{selectionFunction3d}} to produce a test function for
whether a particular location is in the selected region.  If the 
brush becomes inactive, an object containing only the \code{state}
field will be sent, with value \code{"inactive"}.
}

\value{
An object of class \code{"htmlwidget"} (or \code{"shiny.tag.list"}
if pipes are used) that will intelligently print itself into
HTML in a variety of contexts including the R console, within R Markdown
documents, and within Shiny output bindings.

If objects are passed in the \code{shared} argument,
then the widget will respond to selection and filtering 
applied to those as shared datasets.  See \code{\link{rglShared}} for more details and an example.
}
\section{Appearance}{
The appearance of the display is set by the stylesheet
in \code{system.file("htmlwidgets/lib/rglClass/rgl.css")}.

The widget is of class \code{rglWebGL}, with id
set according to \code{elementId}.  (As of this writing,
no special settings are given for class \code{rglWebGL},
but you can add your own.)
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{hook_webgl}} for an earlier approach to this problem. \code{\link{rglwidgetOutput}} for Shiny details.
}
\examples{
save <- options(rgl.useNULL=TRUE)
example("plot3d", "rgl")
widget <- rglwidget()
if (interactive() || in_pkgdown_example())
  widget
  
\donttest{
if (interactive() && !in_pkgdown_example()) {
  # Save it to a file.  This requires pandoc
  filename <- tempfile(fileext = ".html")
  htmlwidgets::saveWidget(rglwidget(), filename)
  browseURL(filename)
}
}

options(save)
}

\name{merge.mesh3d}
\alias{merge.mesh3d}
\title{
Merge RGL mesh objects
}
\description{
Attempts to merge \code{"mesh3d"} objects.  Objects need
to be similar enough; see Details.
}
\usage{
\method{merge}{mesh3d}(x, y, ..., attributesMustMatch = FALSE)
}
\arguments{
  \item{x, y}{
\code{"mesh3d"} objects to merge.
}
  \item{\dots}{
Optional additional objects.
}
  \item{attributesMustMatch}{ See Details.}
}
\details{
To allow objects to be merged, they need to be similar enough
in terms of having the same list of material properties,
normals, texture coordinates, etc.  

If \code{attributesMustMatch} is \code{TRUE}, it is an error
to have attributes in one mesh but not in another, and 
those attributes that only specify a single value must have equal
values in all meshes.

If \code{attributesMustMatch} is \code{FALSE}, any non-matching
attributes will be dropped from the final result.
}
\value{
A single \code{"mesh3d"} object merging the contents
of the arguments.
}
\author{
Duncan Murdoch
}
\examples{
open3d()
# Notice that the alpha setting for the cube is dropped, because
# the other shapes don't specify alpha.
shade3d(merge(cube3d(col="red", alpha = 0.5),
              translate3d(tetrahedron3d(col="green"), 2, 0, 0),
              translate3d(octahedron3d(col="blue"), 4, 0, 0)))
}

\name{playwidget}
\alias{playwidget}
\title{
Add a widget to play animations
}
\description{
This is a widget that can be put in a web page to allow
animations with or without Shiny.
}
\usage{
playwidget(sceneId, controls,
            start = 0, stop = Inf, interval = 0.05, rate = 1,
            components = c("Reverse", "Play", "Slower", "Faster",
                           "Reset", "Slider", "Label"),
            loop = TRUE,
            step = 1, labels = NULL,
            precision = 3,
            elementId = NULL, respondTo = NULL,
            reinit = NULL,
            buttonLabels = components, pause = "Pause",
            height = 40,
            ...)

}
\arguments{
  \item{sceneId}{
The HTML id of the RGL scene being controlled,
or an object.  See the Details below.
}
  \item{controls}{
A single \code{"rglControl"} object, e.g. \code{\link{propertyControl}}, or a list of several.
}
  \item{start, stop}{
The starting and stopping values of the animation.  If
\code{labels} is supplied \code{stop} will default to step
through the labels.
}
  \item{interval}{
The requested interval (in seconds) between updates.  Updates
may occur at longer intervals.
}
  \item{rate}{
The number of units of \dQuote{nominal} time per real world
second.
}
  \item{components}{
Which components should be displayed?  See Details below.
}
  \item{loop}{
When the player reaches the end of the interval,
should it loop back to the beginning?
}
  \item{step}{
Step size in the slider.
}
  \item{labels}{
Optional labels to use, corresponding to slider steps.  Set to
\code{NULL} for auto-generated labels.
}
  \item{precision}{
If \code{labels=NULL}, the precision to use when displaying
timer values.
}
  \item{elementId}{
The HTML id of the generated widget, containing buttons,
slider, etc.
}
  \item{respondTo}{
The HTML ID of a Shiny input control (e.g. a \code{\link[shiny]{sliderInput}}
control) to respond to.}

  \item{reinit}{
A vector of ids that will need re-initialization before
being drawn again.
}

  \item{buttonLabels, pause}{
These are the labels that will be shown on the buttons if they are displayed.  \code{pause} will
be shown on the \code{"Play"} button while playing.
}
  \item{height}{
The height of the widget in pixels.  In a pipe, this is a relative height.
  }
  \item{...}{Additional arguments to pass to
to \code{htmlwidgets::\link[htmlwidgets]{createWidget}}.}
}
\details{
The \code{components} are buttons to control the animation,
a slider for manual control, and a label to show the current
value.  They will be displayed in the order given in \code{components}.  Not all need be included.

The buttons have the following behaviour:
\describe{
\item{Reverse}{Reverse the direction.}
\item{Play}{Play the animation.}
\item{Slower}{Decrease the playing speed.}
\item{Faster}{Increase the playing speed.}
\item{Reset}{Stop the animation and reset to the start value.}
}

If \code{respondTo} is used, no \code{components} are shown, as it is assumed Shiny (or whatever control is being referenced) will provide the UI components.

The \code{sceneId} component can be another \code{playwidget}, a \code{\link{rglwidget}} result, or a result of
\code{htmltools::\link[htmltools:builder]{tags}} or \code{htmltools::\link[htmltools:tag]{tagList}}.  This allows you
to use a \pkg{magrittr}-style
\dQuote{pipe} command to join an \code{rglwidget} with one or more \code{\link{playwidget}}s.  If a \code{playwidget} comes
first, \code{sceneId} should be set to \code{NA}.  If
the \code{\link{rglwidget}} does not come first,
previous values should be piped into its \code{controllers}
argument.  Other HTML code (including other widgets)
can be used in the chain if wrapped in \code{htmltools::\link[htmltools:tag]{tagList}}.

Each control should inherit from \code{"rglControl"}.  They
can have the following components in addition to any 
private ones:
\describe{
\item{\code{labels}}{default labels for the slider.}
\item{\code{param}}{values to include on the slider.}
\item{\code{dependencies}}{additional HTML dependencies to
include, after the default \code{rglwidgetClass}.}
}
}
\section{Appearance}{
The appearance of the controls is set by the stylesheet
in \code{system.file("htmlwidgets/lib/rglClass/rgl.css")}.

The overall widget is of class \code{rglPlayer}, with id
set according to \code{elementId}.

The buttons are of HTML class \code{rgl-button}, the
slider is of class \code{rgl-slider}, and the label is of
class \code{rgl-label}.  Each element has an id prefixed
by the widget id, e.g. \code{elementId-button-Reverse},
\code{elementId-slider}, etc. (where \code{elementId}
should be replaced by the actual id).

The \code{reinit} parameter handles the case where
an object needs re-initialization after each change.  For
example, plane objects may need this if their intersection
with the bounding box changes shape.  Note that
re-initialization is generally incompatible with
the \code{\link{vertexControl}} as it modifies values
which are set during initialization.
}
\value{
A widget suitable for use in an \pkg{Rmarkdown}-generated web
page, or elsewhere.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{subsetControl}},
\code{\link{propertyControl}}, \code{\link{ageControl}} and
\code{\link{vertexControl}} are possible controls to use.

\code{\link{toggleWidget}} is a wrapper for 
\code{playwidget} and \code{\link{subsetControl}}
to insert a single button to toggle some elements in a display.
}
\examples{
saveopts <- options(rgl.useNULL = TRUE)

objid <- plot3d(1:10, 1:10, rnorm(10), col=c("red", "red"), type = "s")["data"]

control <- ageControl(value=0,
             births=1:10,
             ages = c(-5,0,5),
             colors = c("green", "yellow", "red"),
             objids = objid)

\donttest{
# This example uses explicit names
rglwidget(elementId = "theplot", controllers = "theplayer",
          height = 300, width = 300)
playwidget("theplot", control, start = -5, stop = 5,
           rate = 3, elementId = "theplayer",
           components = c("Play", "Slider"))
}

# This example uses pipes, and can skip the names

widget <- rglwidget(height = 300, width = 300) \%>\%
playwidget(control, start = -5, stop = 5,
           rate = 3, components = c("Play", "Slider"))
if (interactive() || in_pkgdown_example())
  widget

options(saveopts)
}

\name{rgl-package}
\title{3D visualization device system}
\alias{rgl-package}
\alias{rgl}
\alias{RGL}
\description{
  3D real-time rendering system.
}
\details{
  RGL is a 3D real-time rendering system for R.
  Multiple windows are managed at a time. Windows
  may be divided into \dQuote{subscenes}, where one has the current focus
  that receives instructions from the R command-line. 
  The device design is oriented towards the R device metaphor. If you send
  scene management instructions, and there's no device open, it will be opened
  automatically.
  Opened devices automatically get the current device focus. The focus may be
  changed by using \code{\link{set3d}()} or \code{\link{useSubscene3d}()}.
  
RGL provides medium to high level functions for 3D interactive graphics, including
  functions modelled on base graphics (\code{\link{plot3d}()}, etc.) as well as functions for 
  constructing geometric objects (\code{\link{cube3d}()}, etc.).  Output
  may be on screen using OpenGL, or to various standard 3D file formats including 
  WebGL, PLY, OBJ, STL as well as 2D image formats, including PNG, Postscript, SVG, PGF.
  
  The \code{\link{open3d}()} function attempts to open a new RGL window,
  using default settings specified by the user.

  See the first example below to display the ChangeLog.
}
\seealso{\link{r3d} for a description of the \code{*3d} interface;
\code{\link{par3d}} for a description of scene properties and the rendering
pipeline; \code{\link{rgl.useNULL}} for a description of how to 
use RGL on a system with no graphics support.}
\examples{
if (!in_pkgdown_example())
  file.show(system.file("NEWS", package = "rgl"))
example(surface3d)
example(plot3d)
}
\keyword{dynamic}

\name{setGraphicsDelay}
\alias{setGraphicsDelay}
\alias{RGL_SLOW_DEV}
\title{
Set a one-time slowdown on opening standard graphics
}
\description{
This function is mainly for internal use, to work around  
a bug in macOS Catalina:  if base plotting happens
too quickly after opening RGL and the first call to quartz, R crashes.

This inserts a delay after the first call to open the 
graphics device.  The default is
no delay, unless on Catalina with no graphics device
currently open but the \code{\link{quartz}} device set
as the default, when a 1 second delay will be added.
Use environment variable "RGL_SLOW_DEV = value" to set
a different default delay. 

It works by changing the value of \code{\link{options}("device")},
so explicit calls to the device will not be affected.

It is called automatically when the \pkg{rgl} package is loaded.  
}
\usage{
setGraphicsDelay(delay = Sys.getenv("RGL_SLOW_DEV", 0), 
                 unixos = "none")
}
\arguments{
  \item{delay}{
Number of seconds for the delay.
}
  \item{unixos}{
The name of the Unix OS.  If set to \code{"Darwin"}, 
and the version is 19.0.0 or greater, the default delay is
changed to 1 second.
}
}
\value{
Called for the side effect of adding the delay to the first
opening of the graphics device.
}

\name{elementId2Prefix}
\alias{elementId2Prefix}
\title{
Use widget with old-style controls
}
\description{
The \code{\link{rglwidget}} control is designed to work in the
\pkg{htmlwidgets} framework.  Older RGL web pages that
used the deprecated \code{\link{writeWebGL}} or \pkg{knitr} used a different
method of linking the controls to the scene.  This is a partial
bridge between the two systems.  You should adopt the new system, not use this function.
}
\usage{
elementId2Prefix(elementId, prefix = elementId)
}
\arguments{
  \item{elementId}{
An element identifier from a \code{\link{rglwidget}} call.
}
  \item{prefix}{
The prefix to use in the old-style control.
}
}
\value{
This function generates Javascript code, so it should be
used in an \code{results = "asis"} block in a \pkg{knitr} document.
}
\author{
Duncan Murdoch
}

\name{clipplaneControl}
\alias{clipplaneControl}
\title{
Sets attributes of a clipping plane
}
\description{
This is a function to produce actions in a web display.  A
\code{\link{playwidget}} or Shiny
input control (e.g. a \code{\link[shiny]{sliderInput}} control)
sets a value which controls attributes of one or more clipping
planes.
}
\usage{
clipplaneControl(a = NULL, b = NULL, c = NULL, d = NULL, 
                 plane = 1, clipplaneids = tagged3d(tag), tag, ...)
}
\arguments{
  \item{a, b, c, d}{
Parameter values for the clipping planes.
}
  \item{plane}{
Which plane in the clipplane object?
}
  \item{clipplaneids}{
The id of the clipplane object.
}
  \item{tag}{
Select clipplane with matching tag.  Ignored
if \code{clipplaneid} is specified.
}  
  \item{\dots}{
Other parameters passed to \code{\link{propertyControl}}.
}
}
\value{
A list of class \code{"rglControl"} of cleaned up parameter
values, to be used in an RGL widget.
}
\author{
Duncan Murdoch
}
\examples{
  open3d()
  saveopts <- options(rgl.useNULL = TRUE)
  xyz <- matrix(rnorm(300), ncol = 3)
  id <- plot3d(xyz, type="s", col = "blue", zlim = c(-3,3))["clipplanes"]
  dvals <- c(3, -3)
  widget <- rglwidget() \%>\%
    playwidget(clipplaneControl(d = dvals, clipplaneids = id),
               start = 0, stop = 1, step = 0.01,
               rate = 0.5)
  if (interactive() || in_pkgdown_example())
    widget
  options(saveopts)
}

\name{tkpar3dsave}
\alias{tkpar3dsave}
\alias{par3dsave}
\title{ Modal dialog for saving par3d settings }
\description{
This function opens a TCL/TK modal dialog to allow particular views of an RGL scene
to be saved.
}
\usage{
tkpar3dsave(params = c("userMatrix", "scale", "zoom", "FOV"),
          times = FALSE, dev = cur3d(), ...)
}
\arguments{
  \item{params}{ Which parameters to save }
  \item{times}{ Should times be saved as well? }
  \item{dev}{ Which RGL device to work with }
  \item{...}{ Additional parameters to pass to \code{\link[tcltk:TkWidgets]{tktoplevel}}}
}
\details{
This opens a TCL/TK modal dialog box with \code{Record} and \code{Quit} buttons.  Each time
\code{Record} is clicked, a snapshot is taken of current \code{\link[rgl]{par3d}} settings.  When
\code{Quit} is clicked, the dialog closes and the values are returned in a list.

If \code{times == TRUE}, then the times at which the views are recorded will also be saved, so that
the \code{\link[rgl]{play3d}} function will play back with the same timing.
}
\value{
A list of the requested components.  Each one will consist of a list of values that were
current when the \code{Record} button was clicked.  These are suitable to be passed directly
to the \code{\link[rgl]{par3dinterp}} function.
}
\author{ Duncan Murdoch }
\seealso{ \code{\link{par3d}}, \code{\link{par3dinterp}}}
\examples{
if (interactive() && !in_pkgdown_example()) {

  # Record a series of positions, and then play them back immediately
  # at evenly spaced times, in an oscillating loop
  example(plot3d)
  play3d( par3dinterp( tkpar3dsave() ) )

  # As above, but preserve the click timings

  # play3d( par3dinterp( tkpar3dsave(times=TRUE) ) )
}
}

\name{setupKnitr}
\alias{hook_rgl}
\alias{hook_webgl}
\alias{hook_rglchunk}
\alias{setupKnitr}
\title{
Displaying RGL scenes in \pkg{knitr} documents
}
\description{
These functions allow RGL graphics to be embedded in \pkg{knitr}
documents.

The simplest method is to run \code{setupKnitr(autoprint = TRUE)}
early in the document.  That way RGL commands act a lot
like base graphics commands:  plots will be automatically inserted
where appropriate, according to the \code{fig.keep} chunk option.
By default (\code{fig.keep = "high"}), only high-level plots
are kept, after low-level changes have been merged into them.
See the \pkg{knitr} documentation 
\url{https://yihui.org/knitr/options/#plots} for more details.  To suppress auto-printing, the RGL calls
can be wrapped in \code{\link{invisible}()}.  
Similarly to \pkg{grid} graphics (used by \pkg{lattice}
and \pkg{ggplot2}), automatic inclusion requires the object
to be printed:  only the last statement in a code block
in braces is automatically printed.  Unlike those 
packages, auto-printing is the only way to get this to
work:  calling \code{\link{print}} explicitly doesn't
work.

Other functions allow embedding either as bitmaps (\code{hook_rgl} with format \code{"png"}),
fixed vector graphics (\code{hook_rgl} with format \code{"eps"}, \code{"pdf"} or
\code{"postscript"}), or interactive WebGL graphics (\code{hook_webgl}).  \code{hook_rglchunk} is not normally invoked by the 
user; it is the hook that supports automatic creation and 
deletion of RGL scenes.
}

\note{The \code{setupKnitr(autoprint = TRUE)} method assumes
\emph{all} printing of RGL objects happens through 
auto-printing of objects produced by the \code{\link{lowlevel}}
or \code{\link{highlevel}} functions.  All RGL functions that
produce graphics do this, but functions in other packages that
call them may not return values appropriately.

If you have multiple calls to \code{setupKnitr()}, all
should have the same arguments.  If any differ, a 
warning will be issued, and the first set of arguments
will be used.

Mixing explicit calls
to \code{\link{rglwidget}} with auto-printing is likely to lead to failure of some
scenes to display.  To avoid this, set \code{options(rgl.printRglwidget = FALSE)} before using such explicit calls.  Similarly,
use that option before calling the \code{\link{example}} function
in a code chunk if the example prints RGL objects.
}

\usage{
setupKnitr(autoprint = FALSE,
           rgl.newwindow = autoprint,
           rgl.closewindows = autoprint)
hook_rgl(before, options, envir)
hook_webgl(before, options, envir)
hook_rglchunk(before, options, envir)
}
\arguments{
  \item{autoprint}{If true, RGL commands
  automatically plot (with low level plots suppressed by
  the default value of the \code{fig.keep} chunk option.)}
  \item{rgl.newwindow, rgl.closewindows}{Default values
  for the \pkg{knitr} chunk options.}
  \item{before, options, envir}{
Standard \pkg{knitr} hook function arguments.
}
}
\details{
The \code{setupKnitr()} function needs to be called once
at the start of the document to install the \pkg{knitr} hooks.
If it is called twice in the same session the second call 
will override the first.

The following chunk options are supported:
\itemize{
  \item \code{rgl.newwindow}:  Whether to open a new window for the chunk.  Default is set by \code{setupKnitr} argument.
  
  \item \code{rgl.closewindows}: Whether
to close windows at the end of the chunk.  Default is set by \code{setupKnitr} argument.

  \item \code{rgl.margin} (default 100):  number of pixels by which to indent the
  WebGL window.
  
  \item \code{snapshot}: Logical value:  when autoprinting
  in HTML, should a snapshot be used instead of the dynamic 
  WebGL display?  Corresponds to \code{rglwidget(snapshot = TRUE, webgl = FALSE)}.  Ignored in LaTeX, where a snapshot
  will always be produced (unless \code{fig.keep} specifies
  no figure at all).

  \item \code{dpi}, \code{fig.retina}, \code{fig.width}, \code{fig.height}:  standard \pkg{knitr} chunk
  options used to set the size of the output.
  
  \item \code{fig.keep}, \code{fig.hold}, \code{fig.beforecode}:  standard \pkg{knitr} chunk options used to control the
display of plots.
  
  \item \code{dev}:  used by \code{hook_rgl} to set 
  the output format.  May be \code{"eps"}, \code{"postscript"},
  \code{"pdf"} or \code{"png"} (default:  \code{"png"}).
  
  \item \code{rgl.keepopen}:  no longer used.  Ignored
  with a warning.

}
}
\value{
A string to be embedded into the output, or \code{NULL} if called
when no output is available.
}
\author{
The \code{hook*} functions are originally by Yihui Xie in the \pkg{knitr} package; and have been modified by
Duncan Murdoch.  Some parts of the \code{setupKnitr} function
duplicate source code from \pkg{knitr}.
}
\keyword{ utilities }

\name{rgl.fns}
\alias{rgl.abclines}
\alias{rgl.planes}
\alias{rgl.clipplanes}
\alias{rgl.sprites}
\alias{rgl.spheres}
\title{
Low level functions that should not be called by users.
}
\description{
These functions provide the implementation for various
\code{*3d} functions designed for users to call.
}
\usage{
rgl.abclines(x, y = NULL, z = NULL, a, b = NULL, c = NULL, ...)
rgl.planes(a, b = NULL, c = NULL, d = 0, ...)
rgl.clipplanes(a, b = NULL, c = NULL, d = 0)
rgl.sprites(x, y = NULL, z = NULL, radius = 1, 
            shapes = NULL, userMatrix, fixedSize = FALSE,
            adj = 0.5, pos = NULL, offset = 0.25, 
            rotating = FALSE, ...)
rgl.spheres(x, y = NULL, z = NULL, radius, fastTransparency = TRUE, ...)
}
\arguments{
  \item{x, y, z}{
Coordinates of points
} 
  \item{a, b, c}{
Coordinates of the direction vectors for the lines,
or normals for the planes.
}
  \item{d}{Plane offset.}
  \item{radius}{Size of sprites or spheres.}
  \item{fixedSize, rotating}{Fixed size, changing orientation.}
  \item{fastTransparency}{Sphere drawing strategy.}
  \item{adj, pos, offset}{Positioning.}
  \item{...}{
Material properties.
}

}
\details{
See the corresponding \code{*3d} function: \code{\link{abclines3d}}, \code{\link{planes3d}}, \code{\link{clipplanes3d}}, \code{\link{sprites3d}}, \code{\link{spheres3d}}.
}
\keyword{internal}

\name{rgl-deprecated}
\alias{rgl-deprecated}
\alias{rgl.antialias}
\alias{rgl.bbox}
\alias{rgl.primitive}
\alias{rgl.points}
\alias{rgl.light}
\alias{rgl.lines}
\alias{rgl.linestrips}
\alias{rgl.material}
\alias{rgl.primitive}
\alias{rgl.select3d}
\alias{rgl.triangles}
\alias{rgl.quads}
\alias{rgl.viewpoint}
\alias{rgl.open}
\alias{rgl.close}
\alias{rgl.set}
\alias{rgl.setAxisCallback}
\alias{writeWebGL}
\alias{rgl.bg}
\alias{rgl.surface}
\alias{rgl.clear}
\alias{rgl.texts}
\title{
Deprecated functions in \pkg{rgl}
}
\description{
These functions are provided for compatibility with older versions of
\pkg{rgl} only, and may be defunct as soon as the next release.  See the comments for the replacements.
}
\usage{
rgl.bbox(xat = NULL, xlab = NULL, xunit = 0, xlen = 5, 
         yat = NULL, ylab = NULL, yunit = 0, ylen = 5, 
         zat = NULL, zlab = NULL, zunit = 0, zlen = 5, 
         marklen = 15, marklen.rel = TRUE, 
         expand = 1, draw_front = FALSE, ...)  # bbox3d
rgl.bg( sphere = FALSE, fogtype = "none", color = c("black", "white"), 
	back = "lines", fogScale = 1, ...)           # bg3d
rgl.clear( type = "shapes", subscene = 0 )     # clear3d

rgl.close()                                    # close3d
rgl.light( theta = 0, phi = 0, viewpoint.rel = TRUE, 
           ambient = "#FFFFFF", diffuse = "#FFFFFF", specular = "#FFFFFF", 
           x = NULL, y = NULL, z = NULL)       # light3d
rgl.lines(x, y = NULL, z = NULL, ... )         # segments3d
rgl.linestrips(x, y = NULL, z = NULL, ...)     # lines3d
rgl.open(useNULL = rgl.useNULL())              # open3d

rgl.points(x, y = NULL, z = NULL, ... )        # points3d
rgl.primitive(...)                             
rgl.quads(x, y = NULL, z = NULL, 
          normals = NULL, 
          texcoords = NULL, ... )              # quads3d
rgl.select3d(button = c("left", "middle", "right"),
            dev = cur3d(), 
            subscene = currentSubscene3d(dev)) # select3d
rgl.set(which, silent = FALSE)                 # set3d
rgl.setAxisCallback(axis, draw = NULL, 
            dev = cur3d(), 
            subscene = currentSubscene3d(dev)) # setAxisCallback
rgl.surface(x, z, y, coords = 1:3, ..., 
            normal_x = NULL, normal_y = NULL, normal_z = NULL,
            texture_s = NULL, texture_t = NULL) # surface3d
rgl.texts(x, y = NULL, z = NULL, text, 
          adj = 0.5, pos = NULL, offset = 0.5, 
          family = par3d("family"), 
          font = par3d("font"), 
          cex = par3d("cex"), 
          useFreeType = par3d("useFreeType"), 
          ...)                                  # text3d
rgl.triangles(x, y = NULL, z = NULL, 
              normals = NULL, texcoords = NULL, 
              ... )                             # triangles3d
rgl.viewpoint( theta = 0, phi = 15, 
               fov = 60, zoom = 1, 
               scale = par3d("scale"), 
               interactive = TRUE, 
               userMatrix, 
               type = c("userviewpoint",
                        "modelviewpoint") )     # view3d
writeWebGL(dir = "webGL", 
           filename = file.path(dir, "index.html"), 
           template = system.file(file.path("WebGL", "template.html"), package = "rgl"),
           prefix = "",
           snapshot = TRUE, 
           commonParts = TRUE, 
           reuse = NULL, 
           font = "Arial", 
           width, height)                       # rglwidget
}
\seealso{
  \code{\link{Deprecated}, \link{bbox3d}, \link{bg3d},
  \link{clear3d}, \link{close3d}, \link{light3d}, 
  \link{lines3d}, \link{open3d}, \link{points3d}, 
  \link{quads3d}, \link{rglwidget}, \link{segments3d},
  \link{select3d}, \link{set3d}, \link{surface3d},
  \link{text3d}, \link{triangles3d}, \link{view3d}}
}
\keyword{internal}
\keyword{misc}

\name{r3d}
\alias{r3d}
\title{Generic 3D interface}
\description{Generic 3D interface for 3D rendering and computational geometry.}
\details{  
  R3d is a design for an interface for 3d rendering and computation without dependency
  on a specific rendering implementation. R3d includes a collection
  of 3D objects and geometry algorithms.
  All r3d interface functions are named \code{*3d}.  They represent generic functions that delegate 
  to implementation functions.
  
  The interface can be grouped into 8 categories: Scene Management, Primitive Shapes,
  High-level Shapes, Geometry Objects, Visualization, Interaction, Transformation,
  Subdivision.  
  
  The rendering interface gives an abstraction to the underlying rendering model. It can
  be grouped into four categories:    
  \describe{
    \item{Scene Management:}{A 3D scene consists of shapes, lights and background environment.} 
    \item{Primitive Shapes:}{Generic primitive 3D graphics shapes such as points, lines, triangles, quadrangles and texts.} 
    \item{High-level Shapes:}{Generic high-level 3D graphics shapes such as spheres, sprites and terrain.}
    \item{Interaction:}{Generic interface to select points in 3D space using the pointer device.}
  }
  
  In this package we include an implementation of r3d using the underlying \code{rgl.*} functions.
  
  3D computation is supported through the use of object structures that live entirely in R.
  \describe{
    \item{Geometry Objects:}{Geometry and mesh objects allow to define high-level geometry for computational purpose such as triangle or 
                             quadrangle meshes (see \code{\link{mesh3d}}).}
    \item{Transformation:}{Generic interface to transform 3d objects.}
    \item{Visualization:}{Generic rendering of 3d objects such as dotted, wired or shaded.}
    \item{Computation:}{Generic subdivision of 3d objects.}
  }
  
  At present, the main practical differences between the r3d functions
  and the \code{rgl.*} functions are as follows. 
  
  The r3d functions call
  \code{\link{open3d}} if there is no device open, and the
  \code{rgl.*} functions call \code{\link{rgl.open}}. By default
  \code{\link{open3d}} sets the initial orientation of the coordinate
  system in 'world coordinates', i.e. a right-handed coordinate system
  in which the x-axis increases from left to right, the y-axis
  increases with depth into the scene, and the z-axis increases from
  bottom to top of the screen.  \code{rgl.*} functions, on the other
  hand, use a right-handed coordinate system similar to that used in
  OpenGL.  The x-axis matches that of r3d, but the y-axis increases
  from bottom to top, and the z-axis decreases with depth into the
  scene.  Since the user can manipulate the scene, either system can
  be rotated into the other one.  
  
  The r3d functions also preserve the \code{rgl.material} setting across
  calls (except for texture elements, in the current implementation), whereas
  the deprecated \code{rgl.*} functions leave it as set by the last call.
  
  The example code below illustrates the two coordinate systems.
}

\seealso{
\code{\link{points3d}},
\code{\link{lines3d}},
\code{\link{segments3d}},
\code{\link{triangles3d}},
\code{\link{quads3d}},
\code{\link{text3d}},
\code{\link{spheres3d}},
\code{\link{sprites3d}},
\code{\link{terrain3d}},
\code{\link{select3d}},
\code{\link{dot3d}},
\code{\link{wire3d}},
\code{\link{shade3d}},
\code{\link{transform3d}},
\code{\link{rotate3d}},
\code{\link{subdivision3d}},
\code{\link{mesh3d}},
\code{\link{cube3d}},
\code{\link{rgl}}
}
\examples{
    
     x <- c(0, 1, 0, 0)
     y <- c(0, 0, 1, 0)
     z <- c(0, 0, 0, 1)
     labels <- c("Origin", "X", "Y", "Z")
     i <- c(1, 2, 1, 3, 1, 4)

     # *3d interface
     
     open3d()
     text3d(x, y, z, labels)
     text3d(1, 1, 1, "*3d coordinates")
     segments3d(x[i], y[i], z[i])
}

\keyword{dynamic}

\name{safe.dev.off}
\alias{safe.dev.off}
\title{
Close graphics device in a safe way.
}
\description{
The \code{\link{dev.off}} function in \pkg{grDevices} doesn't restore
the previous graphics device when called.  This function does.
}
\usage{
safe.dev.off(which = dev.cur(), prev = dev.prev())
}
\arguments{
  \item{which}{
Which device to close.
}
  \item{prev}{
Which device to set as current after closing.
}
}
\details{
This function closes device \code{which} if it is not device 1,
then calls \code{\link{dev.set}(prev)} if there are any devices still
open.
}
\value{
The number and name of the new active device.  It will not 
necessarily be \code{prev} if that device isn't already open.
}
\references{
\url{https://bugs.r-project.org/show_bug.cgi?id=18604}
}
\author{
Duncan Murdoch
}
\examples{
# Open a graphics device
dev.new()
first <- dev.cur()

# Open a second graphics device
dev.new()
second <- dev.cur()
second

# Open another one, and close it using dev.off()
dev.new()
dev.off()
dev.cur() == second # Not the same as second!

# Try again with safe.dev.off()
dev.set(second)
dev.new()
safe.dev.off()
dev.cur() == second

# Close the other two devs
safe.dev.off()
safe.dev.off()
}

\name{ageControl}
\alias{ageControl}
\title{
Set attributes of vertices based on their age
}
\description{
This is a function to produce actions in response to
a \code{\link{playwidget}} or Shiny input control.  The
mental model is that each of the vertices of some object has
a certain birth time; a control sets the current time, so that
vertices have ages depending on the control setting.
Attributes of those vertices can then be changed.
}
\usage{
ageControl(births, ages, objids = tagged3d(tags), tags, value = 0,
           colors = NULL, alpha = NULL, radii = NULL, vertices = NULL,
           normals = NULL, origins = NULL, texcoords = NULL,
           x = NULL, y = NULL, z = NULL,
           red = NULL, green = NULL, blue = NULL)
}
\arguments{
  \item{births}{Numeric birth times of vertices.}
  \item{ages}{Chosen ages at which the following attributes
  will apply.}
  \item{objids}{Object ids to which the changes apply.}
  \item{tags}{
  Alternate way to specify \code{objids}.  Ignored if \code{objids} is
given.
  }
  \item{value}{Initial value; typically overridden by input.}
  \item{colors, alpha, radii, vertices, normals, origins, texcoords}{
  Attributes of the vertices that can be changed.  There should be
  one entry or row for each entry in \code{ages}.}
  \item{x, y, z, red, green, blue}{These one-dimensional
  components of vertices and colors are provided for convenience.}
}
\details{
All attributes must have the same number of entries (rows for the matrices)
as the ages vector. The births vector must have the same number of entries as
the number of vertices in the object.

Not all objects contain all attributes; if one is chosen that is
not a property of the corresponding object, a Javascript \code{alert()} will be
generated.  (This restriction may be removed in the future by attempting
to add the attribute when it makes sense.)

If a \code{births} entry is \code{NA}, no change will be made to that
vertex.
}
\value{
A list of class \code{"rglControl"} of cleaned up parameter
values, to be used in an RGL widget.
}
\author{
Duncan Murdoch
}

\examples{
  saveopts <- options(rgl.useNULL = TRUE)

  theta <- seq(0, 4*pi, length.out = 100)
  xyz <- cbind(sin(theta), cos(theta), sin(theta/2))
  lineid <- plot3d(xyz, type="l", alpha = 0, lwd = 5, col = "blue")["data"]

  widget <- rglwidget() \%>\%
  playwidget(ageControl(births = theta,
                        ages = c(-4*pi, -4*pi, 1-4*pi, 0, 0, 1),
                        objids = lineid,
                        alpha = c(0, 1, 0, 0, 1, 0)),
             start = 0, stop = 4*pi,
             step = 0.1, rate = 4)
  if (interactive() || in_pkgdown_example())
    widget
  options(saveopts)
}

\name{GramSchmidt}
\alias{GramSchmidt}
\title{
The Gram-Schmidt algorithm
}
\description{
Generate a 3x3 orthogonal matrix using the Gram-Schmidt algorithm.
}
\usage{
GramSchmidt(v1, v2, v3, order = 1:3)
}
\arguments{
  \item{v1, v2, v3}{
Three length 3 vectors (taken as row vectors).
}
  \item{order}{
The precedence order for the vectors; see Details.
}
}
\details{
This function orthogonalizes the matrix \code{rbind(v1, v2, v3)}
using the Gram-Schmidt algorithm.  It can handle rank 2 matrices
(returning a rank 3 matrix).  If the original is rank 1, it is likely
to fail.

The \code{order} vector determines the precedence of the original
vectors. For example, if it is \code{c(i, j, k)}, then row \code{i} 
will be unchanged (other than normalization); row \code{j} will
normally be transformed within the span of rows \code{i} and \code{j}.  
Row \code{k} will be transformed orthogonally to the span of
the others.
}
\value{
A 3x3 matrix whose rows are the orthogonalization of the original
row vectors.
}
\author{
Duncan Murdoch
}
\examples{
# Proceed through the rows in order
print(A <- matrix(rnorm(9), 3, 3))
GramSchmidt(A[1, ], A[2, ], A[3, ])

# Keep the middle row unchanged
print(A <- matrix(c(rnorm(2), 0, 1, 0, 0, rnorm(3)), 3, 3, byrow = TRUE))
GramSchmidt(A[1, ], A[2, ], A[3, ], order = c(2, 1, 3))
}

\name{rglToLattice}
\alias{rglToLattice}
\alias{rglToBase}
\title{
Convert RGL userMatrix to lattice or base angles
}
\description{
These functions take a user orientation matrix
from an RGL scene and approximate the
parameters to either \pkg{lattice} or base
graphics functions.
}
\usage{
rglToLattice(rotm = par3d("userMatrix"))
rglToBase(rotm = par3d("userMatrix"))
}
\arguments{
  \item{rotm}{
A matrix in homogeneous coordinates to convert.
}
}
\details{
The \pkg{lattice} package can use Euler angles in
the ZYX scheme to describe the rotation of a scene
in its \code{\link[lattice:cloud]{wireframe}} or
\code{\link[lattice]{cloud}} functions.  The
\code{rglToLattice} function computes these angles
based on \code{rotm}, which defaults to the current
user matrix.  This allows RGL to be used
to interactively find a decent viewpoint and
then reproduce it in \pkg{lattice}.

The base graphics \code{\link{persp}} function does
not use full Euler angles; it
uses a viewpoint angle, and assume the z axis
remains vertical.  The \code{rglToBase} function
computes the viewpoint angle accurately if the
RGL scene is displayed with a vertical
z axis, and does an approximation otherwise.
}
\value{
\code{rglToLattice} returns a list suitable to
be used as the \code{screen} argument to 
\code{\link[lattice:cloud]{wireframe}}.

\code{rglToBase} returns a list containing
\code{theta} and \code{phi} components which
can be used as corresponding arguments in \code{\link{persp}}.
}
\author{
Duncan Murdoch
}
\examples{
persp3d(volcano, col = "green")
if ((hasorientlib <- requireNamespace("orientlib", quietly = TRUE)) && 
    requireNamespace("lattice", quietly = TRUE)) 
    lattice::wireframe(volcano, screen = rglToLattice())
if (hasorientlib) {
  angles <- rglToBase()
  persp(volcano, col = "green", border = NA, shade = 0.5,
        theta = angles$theta, phi = angles$phi)
}
}

\name{ellipse3d}
\alias{ellipse3d}
\alias{ellipse3d.default}
\alias{ellipse3d.lm}
\alias{ellipse3d.glm}
\alias{ellipse3d.nls}
\title{
Make an ellipsoid
}
\description{
A generic function and several methods
returning an ellipsoid or other outline of a confidence region
for three parameters.  
}
\usage{
ellipse3d(x, \dots)
\method{ellipse3d}{default}(x, scale = c(1, 1, 1), centre = c(0, 0, 0), level = 0.95, 
	t = sqrt(qchisq(level, 3)), which = 1:3, subdivide = 3, smooth = TRUE, ...)
\method{ellipse3d}{lm}(x, which = 1:3, level = 0.95, t = sqrt(3 * qf(level, 
                                                3, x$df.residual)), ...)     
\method{ellipse3d}{glm}(x, which = 1:3, level = 0.95, t, dispersion, ...) 
\method{ellipse3d}{nls}(x, which = 1:3, level = 0.95, t = sqrt(3 * qf(level, 
                                                3, s$df[2])), ...) 
}
\arguments{
\item{x}{
An object. In the default method the parameter \code{x} should be a 
square positive definite matrix at least 3x3
in size. It will be treated as the correlation or covariance 
of a multivariate normal distribution.
}
\item{\dots}{
Additional parameters to pass to the default method or to \code{\link{qmesh3d}}.
}
\item{scale}{
If \code{x} is a correlation matrix, then the standard deviations of each
parameter can be given in the scale parameter.  This defaults to \code{c(1, 1, 1)},
so no rescaling will be done.
}
\item{centre}{
The centre of the ellipse will be at this position.
}
\item{level}{
The confidence level of a simultaneous confidence region.  The default is
0.95, for a 95\% region.  This is used to control the size of the ellipsoid.
}
\item{t}{
The size of the ellipse may also be controlled by specifying the value
of a t-statistic on its boundary.  This defaults to the appropriate
value for the confidence region.
}
\item{which}{
This parameter selects which variables from the object will be
plotted.  The default is the first 3.
}
\item{subdivide}{
This controls the number of subdivisions (see \code{\link{subdivision3d}})
used in constructing the ellipsoid.  Higher numbers give a smoother shape.
}
\item{smooth}{
If \code{TRUE}, smooth interpolation of normals is used; if \code{FALSE}, 
a faceted ellipsoid will be displayed.
}
\item{dispersion}{
The value of dispersion to use.  If specified, it is treated as fixed,
and chi-square limits for \code{t} are used. If missing, it is 
taken from \code{summary(x)}.
}
}
\value{
A \code{\link{mesh3d}} object representing the ellipsoid.
}
\examples{
# Plot a random sample and an ellipsoid of concentration corresponding to a 95\% 
# probability region for a
# trivariate normal distribution with mean 0, unit variances and 
# correlation 0.8.
if (requireNamespace("MASS", quietly = TRUE)) {
  Sigma <- matrix(c(10, 3, 0, 3, 2, 0, 0, 0, 1), 3, 3)
  Mean <- 1:3
  x <- MASS::mvrnorm(1000, Mean, Sigma)
  
  open3d()
  
  plot3d(x, box = FALSE)
  
  plot3d( ellipse3d(Sigma, centre = Mean), col = "green", alpha = 0.5, add = TRUE)
}  

# Plot the estimate and joint 90\% confidence region for the displacement and cylinder
# count linear coefficients in the mtcars dataset

data(mtcars)
fit <- lm(mpg ~ disp + cyl , mtcars)

open3d()
plot3d(ellipse3d(fit, level = 0.90), col = "blue", alpha = 0.5, aspect = TRUE)
}
\keyword{dplot}

\name{light}
\alias{light3d}
\title{Add light source}
\description{
  add a light source to the scene.
}
\usage{
light3d(theta=0, phi=15,
        x=NULL, y = NULL, z = NULL,
        viewpoint.rel = TRUE, 
        ambient = "#FFFFFF", 
        diffuse = "#FFFFFF", 
        specular = "#FFFFFF")
}
\arguments{
  \item{theta, phi}{direction to infinitely distant light}
  \item{x, y, z}{position of finitely distant light}
  \item{viewpoint.rel}{logical, if TRUE light is a viewpoint light that is positioned relative to the current viewpoint}
  \item{ambient, diffuse, specular }{ light color values used for lighting calculation }
}
\details{
  Up to 8 light sources are supported. They are positioned either in
  world space or relative to the camera. By providing polar
  coordinates to \code{theta} and \code{phi} a directional light source is used. If
  numerical values are given to x, y and z, a point-like light source with
  finite distance to the objects in the scene is set up.
  
  If \code{x} is non-null, \code{\link{xyz.coords}} will
  be used to form the location values, so all three coordinates
  can be specified in \code{x}.
  
  If no lights have been added to a subscene, lights from the parent
  subscene will be used.
  
See \code{\link{material3d}} for a discussion of how
the components of the light affect the display of objects.
}
\value{
  This function is called for the side effect of adding a light.  A light ID is
  returned to allow \code{\link{pop3d}} to remove it.
}
\seealso{
  \code{\link{clear3d}}
  \code{\link{pop3d}}
}
\examples{

#
# a lightsource moving through the scene
#
data(volcano)
z <- 2 * volcano # Exaggerate the relief
x <- 10 * (1:nrow(z)) # 10 meter spacing (S to N)
y <- 10 * (1:ncol(z)) # 10 meter spacing (E to W)
zlim <- range(z)
zlen <- zlim[2] - zlim[1] + 1
colorlut <- terrain.colors(zlen) # height color lookup table
col <- colorlut[ z - zlim[1] + 1 ] # assign colors to heights for each point

open3d()
bg3d("gray50")
surface3d(x, y, z, color = col, back = "lines")
r <- max(y) - mean(y)
lightid <- spheres3d(1, 1, 1, alpha = 0)
frame <- function(time) {
    a <- pi*(time - 1)
    save <- par3d(skipRedraw = TRUE)
    clear3d(type = "lights")
    pop3d(id = lightid)
    xyz <- matrix(c(r*sin(a) + mean(x), r*cos(a) + mean(y), max(z)), ncol = 3)
    light3d(x = xyz, diffuse = "gray75", 
            specular = "gray75", viewpoint.rel = FALSE) 
    light3d(diffuse = "gray10", specular = "gray25")
    lightid <<- spheres3d(xyz, emission = "white", radius = 4)
    par3d(save)
    Sys.sleep(0.02)
    NULL
}
play3d(frame, duration = 2)
}
\keyword{dynamic}

\name{as.mesh3d.ashape3d}
\alias{as.mesh3d.ashape3d}
\title{
Convert alpha-shape surface of a cloud of points to
RGL mesh object
}
\description{
The \code{\link[alphashape3d:ashape3d]{alphashape3d::ashape3d}}
function computes the
3D \eqn{\alpha}-shape of a cloud of points. This is an approximation
to the visual outline of the cloud.  It may include isolated
points, line segments, and triangular faces:  this function
converts the triangular faces to an RGL \code{\link{tmesh3d}} object.
}
\usage{
\method{as.mesh3d}{ashape3d}(x, 
                   alpha = x$alpha[1], 
                   tri_to_keep = 2L,
                   col = "gray", 
                   smooth = FALSE, normals = NULL,
                   texcoords = NULL, ...)
}
\arguments{
  \item{x}{
An object of class \code{"ashape3d"}.
}
  \item{alpha}{
Which \code{alpha} value stored in \code{x} should be converted?
}
  \item{tri_to_keep}{
Which triangles to keep.  Expert use only:  see \code{triang}
entry in \bold{Value} section of \link[alphashape3d]{ashape3d}
for details.
}
  \item{col}{
The surface colour.
}
  \item{smooth}{
Whether to attempt to add normals to make the surface
look smooth.  See the Details below.
}
  \item{normals, texcoords}{
Normals and texture coordinates at each vertex can be specified.
}
  \item{\dots}{
Additional arguments to pass to use as \code{\link{material3d}} properties on the resulting mesh.
}
}
\details{
Edelsbrunner and Mucke's (1994) \eqn{\alpha}-shape algorithm is 
intended to compute a surface of a general cloud of points.
Unlike the convex hull, the cloud may have voids, isolated points, and other oddities.  This function is designed to 
work in the case where the surface is made up of 
simple polygons.

If \code{smooth = TRUE}, this method attempts to orient all
of the triangles in the surface consistently and add normals
at each vertex by averaging the triangle normals.  
However, for some point clouds, the \eqn{\alpha}-shape will contain 
sheets of polygons with a few solid polyhedra embedded.
This does not allow a consistent definition of "inside"
and outside.  If this is detected, a warning is issued
and the resulting mesh will likely contain boundaries
where the assumed orientation of triangles changes, resulting
in ugly dark lines through the shape.  Larger values
of \code{alpha} in the call to \code{\link[alphashape3d:ashape3d]{alphashape3d::ashape3d}} may help.

Methods for \code{\link{plot3d}} and \code{\link{persp3d}}
are also defined:  they call the \code{\link{as.mesh3d}} method and then plot the result.  
}
\value{
A \code{"mesh3d"} object, suitable for plotting.
}
\references{
Edelsbrunner, H., Mucke, E. P. (1994). Three-Dimensional Alpha Shapes. ACM Transactions on Graphics, 13(1), pp.43-72.

Lafarge, T. and Pateiro-Lopez, B. (2017).
  alphashape3d: Implementation of the 3D Alpha-Shape
  for the Reconstruction of 3D Sets from a Point Cloud.
  R package version 1.3.
}
\author{
Duncan Murdoch
}

\examples{
if (requireNamespace("alphashape3d", quietly = TRUE)) {
  set.seed(123)
  n <- 400    # 1000 gives a nicer result, but takes longer
  xyz <- rbind(cbind(runif(n), runif(n), runif(n)),
               cbind(runif(n/8, 1, 1.5), 
                     runif(n/8, 0.25, 0.75), 
                     runif(n/8, 0.25, 0.75)))
  ash <- suppressMessages(alphashape3d::ashape3d(xyz, alpha = 0.2))
  m <- as.mesh3d(ash, smooth = TRUE)
  open3d()
  mfrow3d(1, 2, sharedMouse = TRUE)
  plot3d(xyz, size = 1)
  plot3d(m, col = "red", alpha = 0.5)
  points3d(xyz, size = 1)
}
}

\name{pch3d}
\alias{pch3d}
\title{
Plot symbols similar to base graphics
}
\description{
This function plots symbols similarly to 
what the base graphics function \code{\link{points}} does when \code{pch} is specified.
}
\usage{
pch3d(x, y = NULL, z = NULL, pch = 1, 
      bg = material3d("color")[1], cex = 1, radius, 
      color = "black", lit = FALSE, ...)
}
\arguments{
  \item{x, y, z}{
The locations at which to plot in a form
suitable for use in \code{\link{xyz.coords}}.
}
  \item{pch}{
A vector of integers or single characters
describing the symbols to plot.
}
  \item{bg}{
The fill color(s) to use for \code{pch} from 21 
to 25.
}
  \item{cex}{
A relative size of the symbol to plot.
}
  \item{radius}{
An absolute size of the symbol to plot 
in user coordinates.
}
  \item{color}{ 
The color(s) to use for symbols.
}
  \item{lit}{
Whether the object responds to lighting or just shows the
displayed color directly.
}
  \item{\dots}{
Other material properties.
}
}
\details{
The list of symbols encoded by numerical
\code{pch} values is given in the \code{\link{points}} help page.  
}
\note{
This function is not a perfect match to how the \code{\link{points}} function works due
to limitations in RGL and OpenGL.  In
particular:

Symbols with numbers from 1 to 25 are drawn
as 3D sprites (see \code{\link{sprites3d}}),
so they will resize as the window is zoomed.
Letters and numbers from 32 to 255 (which are
mapped to letters) are drawn using \code{\link{text3d}}, 
so they maintain a fixed size.

A calculation somewhat like the one in \code{\link{plot3d}} that sets the size
of spheres is used to choose the size of
sprites based on \code{cex} and the 
current scaling.  This will likely need manual
tweaking.  Use the \code{radius} argument
for a fixed size.

No special handling is done for the case of 
\code{pch = "."}.  Use \code{points3d} for 
small dots.

As of \pkg{rgl} version 0.100.10, background and foreground 
colors can vary from symbol to symbol.
}
\value{
A vector of object id values is returned invisibly.  Separate objects will be drawn
for each different combination of \code{pch} value from 
0 to 25, \code{color} and  \code{bg}, and another holding all the character
symbols.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{points3d}}, \code{\link{text3d}},
\code{\link{plot3d}}, \code{\link{points}}.
}
\examples{
open3d()
i <- 0:25; x <- i \%\% 5; y <- rep(0, 26); z <- i \%/\% 5
pch3d(x, y, z, pch = i, bg = "gray", color = rainbow(26))
text3d(x, y, z + 0.3, i)
pch3d(x + 5, y, z, pch = i+65)
text3d(x + 5, y, z + 0.3, i+65)
}

\name{writeASY}
\alias{writeASY}
\title{
Write Asymptote code for an RGL scene
}
\description{
Asymptote is a language for 3D graphics that
is highly integrated with LaTeX.  This is
an experimental function to
write an Asymptote program to approximate 
an RGL scene.
}
\usage{
writeASY(scene = scene3d(),
         title = "scene", 
         outtype = c("pdf", "eps", "asy", "latex", "pdflatex"), 
         prc = TRUE, 
         runAsy = "asy \%filename\%", 
         defaultFontsize = 12,
         width = 7, height = 7,
         ppi = 100,
         ids = tagged3d(tags),
         tags = NULL, 
         version = "2.65")
}
\arguments{
  \item{scene}{RGL scene object}
  \item{outtype}{
What type of file to write?  See Details.
}
  \item{prc}{
Whether to produce an interactive PRC scene.
}
  \item{title}{
The base of the filename to produce.
}
  \item{runAsy}{
Code to run the Asymptote program.
}
  \item{defaultFontsize}{
The default fontsize for text. 
  }
  \item{width, height}{
Width and height of the output image, in inches.
  }
  \item{ppi}{
\dQuote{Pixels per inch} to assume when converting
line widths and point sizes (which RGL measures in 
pixels).
  }
  \item{ids}{
If not \code{NULL}, write out just these
RGL objects.
}
  \item{tags}{
Alternate way to specify \code{ids}.  Ignored if \code{ids} is
given.
  }
  \item{version}{
Asymptote version 2.44 had a definition for its \dQuote{light()} function
that was incompatibly changed in versions 2.47 and 2.50.  
The current code has been tested with version 2.65.
If you are using an older version, set \code{version}
to your version number and it may work better.
}
}
\details{
Asymptote is both a language describing a 
2D or 3D graphic, and a program to interpret
that language and produce output in a variety
of formats including EPS, PDF (interactive
or static), etc.

The interactive scene produced with \code{prc = TRUE}
requires \code{outtype = "pdf"}, and (as of this
writing) has a number of limitations:
\itemize{
\item{As far as we know, only Adobe Acrobat Reader of a sufficiently recent version can display 
these scenes.}
\item{Current versions ignore lighting settings.}
}
}
\value{
The filename of the output file is returned
invisibly.
}
\references{
J. C. Bowman and A. Hammerlindl (2008). Asymptote: A vector graphics language, 
TUGBOAT: The Communications of the TeX Users Group, 29:2, 288-294.
}
\author{
Duncan Murdoch
}
\note{
This function is currently under development and limited in the quality
of output it produces.  Arguments will likely
change.

There are a number of differences between the interactive display in 
Asymptote and the display in RGL.  In particular, many objects that are
a fixed size in RGL will scale with the image in Asymptote.  Defaults
have been chosen somewhat arbitrarily; tweaking will likely be needed.

Material properties of surfaces are not yet implemented.

On some systems, the program  \command{asy}  used
to process the output has bugs and may fail.  Run the example
at your own risk!
}
\seealso{
\code{\link{scene3d}} saves a copy of a scene to an R variable; \code{\link{rglwidget}},
\code{\link{writePLY}}, \code{\link{writeOBJ}} and \code{\link{writeSTL}}
write the scene to a file in various other formats.
}
\examples{
\dontrun{
# On some systems, the program  "asy"  used
# to process the output has bugs, so this may fail.
x <- rnorm(20)
y <- rnorm(20)
z <- rnorm(20)
plot3d(x, y, z, type = "s", col = "red")
olddir <- setwd(tempdir())
writeASY(title = "interactive")  # Produces interactive.pdf
writeASY(title = "noninteractive", prc = FALSE) # Produces noninteractive.pdf
setwd(olddir)
}
}

\name{tagged3d}
\alias{tagged3d}
\title{
Find tags on rgl objects.
}
\description{
Objects with material properties may have an arbitrary
string set as a tag.  This function retrieves the id
values associated with a given tag, or the tags set on 
given ids.
}
\usage{
tagged3d(tags = NULL, ids = NULL, full = FALSE, subscene = 0)
}
\arguments{
  \item{tags}{
A vector of tags to use for selection.
}
  \item{ids}{
A vector of ids to report the tags on.
}
  \item{full}{logical; whether to return 
a dataframe containing \code{id}, \code{type}, \code{tag},
or a vector of ids or tags.}
  \item{subscene}{
Where to look:  by default, the whole scene is searched.
\code{NA} restricts the search to the current subscene,
or a subscene id can be given.
}
}
\details{
Exactly one of \code{tags} and \code{ids} must be specified.
}
\value{
A dataframe is constructed with columns
\item{id}{item id}
\item{type}{item type}
\item{tag}{item tag}
matching the specified \code{tags} or \code{ids} value.
If \code{full = TRUE}, the full dataframe is returned, otherwise
just the requested ids or tags.

If \code{ids} is specified, the return value will be
in the same order
as \code{ids}).
}
\author{
Duncan Murdoch
}

\examples{
open3d()
ids <- plot3d(rnorm(10), rnorm(10), rnorm(10), tag = "plot")
unclass(ids)
tagged3d("plot")
tagged3d(ids = ids, full = TRUE)
}

\name{par3dinterpControl}
\alias{par3dinterpControl}
\title{
Control RGL widget like par3dinterp()
}
\description{
This control works with \code{\link{playwidget}} to 
change settings in a WebGL display in the same way
as \code{\link{par3dinterp}} does within R.
}
\usage{
par3dinterpControl(fn, from, to, steps, subscene = NULL, omitConstant = TRUE, ...)
}
\arguments{
  \item{fn}{A function returned from \code{\link{par3dinterp}}.}
  \item{from, to, steps}{Values where \code{fn} should be evaluated.}
  \item{subscene}{Which subscene's properties should be modified?}
  \item{omitConstant}{If \code{TRUE}, do not set values that are
  constant across the range.}
  \item{...}{Additional parameters which will be passed
  to \code{\link{propertyControl}}.}
}
\details{
\code{par3dinterpSetter} sets parameters corresponding to values produced by the result of
\code{par3dinterp}.  
}
\value{
Returns controller data in a list of class "rglControl".
}
\author{
Duncan Murdoch
}
\examples{
example(plot3d)
M <- r3dDefaults$userMatrix
fn <- par3dinterp(times = (0:2)*0.75, userMatrix = list(M,
                                      rotate3d(M, pi/2, 1, 0, 0),
                                      rotate3d(M, pi/2, 0, 1, 0)),
                                      scale = c(0.5, 1, 2))
control <- par3dinterpControl(fn, 0, 3, steps = 15)
control      
if (interactive() || in_pkgdown_example()) 
  rglwidget(width = 500, height = 250) \%>\%
  playwidget(control,
       step = 0.01, loop = TRUE, rate = 0.5)
}

\name{writePLY}
\alias{writePLY}
\title{
Write Stanford PLY format files
}
\description{
This function writes PLY files.  This is a simple file format that
is commonly used in 3D printing.  It does not represent text,
only edges and polygons.  The \code{writePLY} function 
does the necessary conversions.

}
\usage{
writePLY(con, format = c("little_endian", "big_endian", "ascii"), 
         pointRadius = 0.005, pointShape = icosahedron3d(), 
         lineRadius = pointRadius, lineSides = 20, 
         pointsAsEdges = FALSE, linesAsEdges = pointsAsEdges, 
         withColors = TRUE, withNormals = !(pointsAsEdges || linesAsEdges), 
         ids = tagged3d(tags), tags = NULL)
}
\arguments{
  \item{con}{
A connection or filename.
}
  \item{format}{
Which output format.  Defaults to little-endian binary.
}
  \item{pointRadius, lineRadius}{
The radius of points and lines relative to the overall scale of the figure,
if they are converted to polyhedra.
}
  \item{pointShape}{
A mesh shape to use for points if they are converted.  It is scaled by the \code{pointRadius}.
}
  \item{lineSides}{
Lines are rendered as cylinders with this many sides.
}
  \item{pointsAsEdges, linesAsEdges}{
Whether to convert points and lines to \dQuote{Edge} records in the PLY output.
}
  \item{withColors}{
Whether to output vertex color information.
}
  \item{withNormals}{
Whether to output vertex normals for smooth shading.
}
  \item{ids}{
The identifiers (from \code{\link{ids3d}}) of the
objects to write.  If \code{NULL}, try to write everything.
}
  \item{tags}{
Select objects with matching tags.  Ignored
if \code{ids} is specified.
}

}
\details{
The current implementation only outputs triangles, quads, planes, spheres,
points, line segments, line strips and surfaces.

The defaults for \code{pointsAsEdges} and \code{linesAsEdges} have been
chosen because Blender (\url{https://www.blender.org}) does not import
lines, only polygons.  If you are exporting to other software you
may want to change them.

Since the PLY format only allows one object per file, all RGL objects
are combined into a single object when output.

The output file is readable by Blender and Meshlab; the latter can write in
a number of other formats, including U3D, suitable for import into a PDF document.

}
\value{
Invisibly returns the name of the connection to which the 
data was written.
}
\references{
The file format was found on \code{www.mathworks.com} on November 10, 2012
at a URL that no longer exists; currently the
format is described at \code{www.mathworks.com/help/vision/ug/the-ply-format.html}.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{scene3d}} saves a copy of a scene to an R variable;
\code{\link{rglwidget}}, \code{\link{writeASY}}, \code{\link{writeOBJ}} and \code{\link{writeSTL}}
write the scene to a file in various other formats.
}
\examples{
filename <- tempfile(fileext = ".ply")
open3d()
shade3d( icosahedron3d(col = "magenta") )
writePLY(filename)
}
\keyword{ graphics }

\name{rgl-duplicated}
\alias{rgl.pop}
\alias{rgl.cur}
\alias{rgl.ids}
\alias{terrain3d}
\title{
Duplicated functions in \pkg{rgl}
}
\description{
These functions are provided for compatibility with older versions of
\pkg{rgl}.  See the functions in the comments for the 
full documentation.
}
\usage{
rgl.cur()                             # cur3d      
rgl.ids(type = "shapes", 
        subscene = NA, tags = FALSE)  # ids3d 
rgl.pop(type = "shapes", 
        id = 0, tag = NULL)           # pop3d
terrain3d(x, y = NULL, z = NULL, ...,
          normal_x = NULL, normal_y = NULL, normal_z = NULL,
          texture_s = NULL, texture_t=NULL,
          flip = FALSE)               # surface3d
}
\seealso{\code{\link{cur3d}}, \code{\link{ids3d}}, \code{\link{pop3d}}, \code{\link{surface3d}}}
\keyword{internal}
\keyword{misc}

\name{persp3d}
\alias{persp3d}
\alias{persp3d.default}
\title{ Surface plots }
\description{
  This function draws plots of surfaces in 3-space. \code{persp3d} is a generic function.}
\usage{
persp3d(x, \dots)

\method{persp3d}{default}(x = seq(0, 1, length.out = nrow(z)),
    y = seq(0, 1, length.out = ncol(z)), z,
    xlim = NULL, ylim = NULL, zlim = NULL,
    xlab = NULL, ylab = NULL, zlab = NULL, add = FALSE, aspect = !add, 
    forceClipregion = FALSE, ...)
}
\arguments{
  \item{x, y, z}{points to plot on surface.  See Details below.}
  \item{xlim, ylim, zlim}{x-, y-  and z-limits.  If present,
  the plot is clipped to this region.}
  \item{xlab, ylab, zlab}{titles for the axes.  N.B. These must be
    character strings; expressions are not accepted.  Numbers will be
    coerced to character strings.}
  \item{add}{whether to add the points to an existing plot.}    
  \item{aspect}{either a logical indicating whether to adjust the aspect ratio, or a new ratio.}
  \item{forceClipregion}{force a clipping region to be used,
whether or not limits are given.}
  \item{\dots}{additional material parameters to be passed to \code{\link{surface3d}}
  and \code{\link{decorate3d}}.}
}
\details{
The default method plots a surface defined as a grid of \code{(x,y,z)} locations in space.  The grid may be specified in several ways:
\itemize{
\item{As with \code{\link[graphics]{persp}}, \code{x} and \code{y} may be given as vectors in ascending order, with \code{z} given as a matrix. There
should be one \code{x} value for each row of \code{z} and
one \code{y} value for each column.  The surface drawn will
have \code{x} constant across rows and \code{y} constant
across columns.  This is the most convenient format when
\code{z} is a function of \code{x} and \code{y} which are
measured on a regular grid.}

\item{\code{x} and \code{y} may also be given as matrices, in which
case they should have the same dimensions as \code{z}.  The
surface will combine corresponding points in each matrix
into locations \code{(x,y,z)} and draw the surface through those.  This
allows general surfaces to be drawn, as in the example of a spherical Earth shown below.}

\item{If \code{x} is a \code{list}, its components \code{x$x}, \code{x$y} and \code{x$z} are used for \code{x}, \code{y} and
\code{z} respectively, though an explicitly specified
\code{z} value will have priority.}}
  
  One difference from \code{\link[graphics]{persp}} is that colors are specified on each
  vertex, rather than on each facet of the surface.  To emulate the \code{\link[graphics]{persp}}
  color handling, you need to do the following.  First, convert the color vector to 
  an \code{(nx - 1)} by \code{(ny - 1)} matrix; then add an extra row before row 1, 
  and an extra column after the last column, to convert it to \code{nx} by \code{ny}.
  (These extra colors will not be used).  For example, 
  \code{col <- rbind(1, cbind(matrix(col, nx - 1, ny - 1), 1))}.
  Finally, call \code{persp3d} with material property \code{smooth = FALSE}.
  
  See the \dQuote{Clipping} section in \code{\link{plot3d}}
for more details on \code{xlim, ylim, zlim} and \code{forceClipregion}.
}
\value{
  This function is called for the side effect of drawing the plot.  A vector 
  of shape IDs is returned invisibly.
}
\author{Duncan Murdoch}
\seealso{\code{\link{plot3d}}, \code{\link{persp}}. There is
a \code{\link{persp3d.function}} method for drawing functions, and \code{\link{persp3d.deldir}} can be used
to draw surfaces defined by an irregular collection of points.  A formula method \code{\link{persp3d.formula}} draws 
surfaces using this method.

The \code{\link{surface3d}} function is used to draw the surface without the axes etc. }
\examples{

# (1) The Obligatory Mathematical surface.
#     Rotated sinc function.

x <- seq(-10, 10, length.out = 20)
y <- x
f <- function(x, y) { r <- sqrt(x^2 + y^2); 10 * sin(r)/r }
z <- outer(x, y, f)
z[is.na(z)] <- 1
open3d()

# Draw the surface twice:  the first draws the solid part, 
# the second draws the grid.  Offset the first so it doesn't
# obscure the lines.

persp3d(x, y, z, aspect = c(1, 1, 0.5), col = "lightblue",
        xlab = "X", ylab = "Y", zlab = "Sinc( r )", 
        polygon_offset = 1)
persp3d(x, y, z, front = "lines", back = "lines", 
        lit = FALSE, add = TRUE)
highlevel()   # trigger the plot

# (2) Add to existing persp plot:

xE <- c(-10, 10); xy <- expand.grid(xE, xE)
points3d(xy[, 1], xy[, 2], 6, col = "red")
lines3d(x, y = 10, z = 6 + sin(x), col = "green")

phi <- seq(0, 2*pi, length.out = 201)
r1 <- 7.725 # radius of 2nd maximum
xr <- r1 * cos(phi)
yr <- r1 * sin(phi)
lines3d(xr, yr, f(xr, yr), col = "pink", lwd = 2)

# (3) Visualizing a simple DEM model

z <- 2 * volcano        # Exaggerate the relief
x <- 10 * (1:nrow(z))   # 10 meter spacing (S to N)
y <- 10 * (1:ncol(z))   # 10 meter spacing (E to W)

open3d()
invisible(bg3d("slategray")) # suppress display
material3d(col = "black")
persp3d(x, y, z, col = "green3", aspect = "iso",
      axes = FALSE, box = FALSE)

# (4) A globe

lat <- matrix(seq(90, -90, length.out = 50)*pi/180, 50, 50, byrow = TRUE)
long <- matrix(seq(-180, 180, length.out = 50)*pi/180, 50, 50)

r <- 6378.1 # radius of Earth in km
x <- r*cos(lat)*cos(long)
y <- r*cos(lat)*sin(long)
z <- r*sin(lat)

open3d()
persp3d(x, y, z, col = "white", 
       texture = system.file("textures/worldsmall.png", package = "rgl"), 
       specular = "black", axes = FALSE, box = FALSE, xlab = "", ylab = "", zlab = "",
       normal_x = x, normal_y = y, normal_z = z)

\dontrun{
# This looks much better, but is slow because the texture is very big
persp3d(x, y, z, col = "white", 
       texture = system.file("textures/world.png", package = "rgl"), 
       specular = "black", axes = FALSE, box = FALSE, xlab = "", ylab = "", zlab = "",
       normal_x = x, normal_y = y, normal_z = z)
}


}
\keyword{ dynamic }
\keyword{ graphics }

\name{spin3d}
\alias{spin3d}
\title{ Create a function to spin a scene at a fixed rate }
\description{
This creates a function to use with \code{\link{play3d}} to
spin an RGL scene at a fixed rate.
}
\usage{
spin3d(axis = c(0, 0, 1), rpm = 5,
       dev = cur3d(), subscene = par3d("listeners", dev = dev))
}
\arguments{
  \item{axis}{ The desired axis of rotation }
  \item{rpm}{ The rotation speed in rotations per minute }
  \item{dev}{ Which RGL device to use }
  \item{subscene}{ Which subscene to use }
}
\value{
A function with header \code{function(time, base = M)}, where \code{M}
is the result of \code{par3d("userMatrix")} at the time the function
is created.  This function
calculates and returns a list containing \code{userMatrix} updated by
spinning the base matrix for \code{time} seconds at \code{rpm} revolutions per
minute about the specified \code{axis}.
}
\note{
Prior to \pkg{rgl} version 0.95.1476, the
\code{subscene} argument defaulted to the
current subscene, and any additional entries would
be ignored by \code{\link{play3d}}.  The current
default value of \code{par3d("listeners", dev = dev)} means that all subscenes that share 
mouse responses will also share modifications
by this function.
}
\author{ Duncan Murdoch }
\seealso{ \code{\link{play3d}} to play the animation }
\examples{
# Spin one object
open3d()
plot3d(oh3d(col = "lightblue", alpha = 0.5))
if (!rgl.useNULL()  && interactive())
  play3d(spin3d(axis = c(1, 0, 0), rpm = 30), duration = 2)

# Show spinning sprites, and rotate the whole view
open3d()
spriteid <- NULL

spin1 <- spin3d(rpm = 4.5 ) # the scene spinner
spin2 <- spin3d(rpm = 9 ) # the sprite spinner

f <- function(time) {
    par3d(skipRedraw = TRUE) # stops intermediate redraws
    on.exit(par3d(skipRedraw = FALSE)) # redraw at the end

    pop3d(id = spriteid) # delete the old sprite
    cubeid <- shade3d(cube3d(), col = "red")
    spriteid <<- sprites3d(0:1, 0:1, 0:1, shape = cubeid,
                   userMatrix = spin2(time, 
                     base = spin1(time)$userMatrix)$userMatrix)
    spin1(time)
}
if (!rgl.useNULL() && interactive())
  play3d(f, duration = 2)
}
\keyword{ dplot }

\name{turn3d}
\alias{turn3d}
\title{
Create a solid of rotation from a two-dimensional curve
}
\description{
This function \dQuote{turns} the curve (as on a lathe) to form a 
solid of rotation along the x axis.
}
\usage{
turn3d(x, y = NULL, n = 12, smooth = FALSE, ...)
}
\arguments{
  \item{x, y}{
Points on the curve, in a form suitable for \code{\link{xy.coords}}.
The \code{y} values must be non-negative.
}
  \item{n}{
How many steps in the rotation?
}
  \item{smooth}{
logical; whether to add normals for a smooth appearance.
}
  \item{...}{
Additional parameters to pass to \code{\link{tmesh3d}}.
}
}
\value{
A mesh object containing triangles and/or quadrilaterals.
}
\author{
Fang He and Duncan Murdoch
}
\seealso{
\code{\link{extrude3d}}
}
\examples{
x <- 1:10
y <- rnorm(10)^2
open3d()
shade3d(turn3d(x, y), col = "green")
}
\keyword{ graphics }

\name{subdivision3d}
\alias{subdivision3d}
\alias{subdivision3d.mesh3d}
\alias{divide.mesh3d}
\alias{normalize.mesh3d}
\alias{deform.mesh3d}
\title{Subdivide a mesh}
\description{
The subdivision surface algorithm divides and refines (deforms) a
given mesh recursively to certain degree (depth).
The mesh3d algorithm consists of two stages: divide and deform.
The divide step generates for each triangle or quad four new triangles or quads, the deform
step drags the points (refinement step).
}
\usage{
  subdivision3d( x,  ...)
  \method{subdivision3d}{mesh3d}( x, depth = 1, normalize = FALSE,
                deform = TRUE, keepTags = FALSE, ... )
  divide.mesh3d(mesh,  vb = mesh$vb, 
                ib = mesh$ib, it = mesh$it, is = mesh$is,
                keepTags = FALSE)
  normalize.mesh3d(mesh)
  deform.mesh3d(mesh, vb = mesh$vb, ib = mesh$ib, it = mesh$it,
                is = mesh$is)
}
\arguments{
  \item{x}{3d geometry mesh}
  \item{mesh}{3d geometry mesh}
  \item{depth}{recursion depth}
  \item{normalize}{normalize mesh3d coordinates after division if \code{deform} is \code{TRUE}}
  \item{deform}{deform mesh}
  \item{keepTags}{if \code{TRUE}, add a \code{"tags"}
  component to the output.}
  \item{is}{indices for segments}
  \item{it}{indices for triangular faces}
  \item{ib}{indices for quad faces}
  \item{vb}{matrix of vertices: 4 x n matrix (rows x, y, z, h) or equivalent
    vector, where h indicates scaling of each plotted quad}
  \item{...}{other arguments (unused)}
}
\details{
\code{subdivision3d} takes
a mesh object and replaces each segment with two new ones, and 
each triangle or quad with 4 new ones
by adding vertices half-way along the edges (and one in the
centre of a quad).  The positions of the vertices are 
deformed so that the resulting surface is smoother than the original.  These operations are repeated \code{depth} times.

The other functions do the individual steps of the subdivision.
\code{divide.mesh3d} adds the extra vertices.  \code{deform.mesh3d}
does the smoothing by replacing each vertex with the average of each of its neighbours. \code{normalize.mesh3d} 
normalizes the homogeneous coordinates, by setting the 
4th coordinate to 1.  (The 4th coordinate is used as a
weight in the deform step.)
}
\value{
A modified \code{\link{mesh3d}} object.  If \code{keepTags} is
\code{TRUE}, it will contain a \code{tags} component.  For
details, see the \code{\link{clipMesh3d}} help topic.
}

\examples{
  open3d()
  shade3d( subdivision3d( cube3d(), depth = 3 ), color = "red", alpha = 0.5 )
}
\seealso{
  \code{\link{r3d}}
  \code{\link{mesh3d}}
}
\keyword{dynamic}

\name{readSTL}
\alias{readSTL}
\alias{writeSTL}
\title{
Read and write STL (stereolithography) format files
}
\description{
These functions read and write STL files.  This is a simple file format that
is commonly used in 3D printing.  It does not represent text,
only triangles.  The \code{writeSTL} function converts some RGL object
types to triangles.
}
\usage{
readSTL(con, ascii = NA, plot = TRUE, ...)
writeSTL(con, ascii = FALSE, 
         pointRadius = 0.005, 
         pointShape = icosahedron3d(),
         lineRadius = pointRadius,
         lineSides = 20,
         ids = tagged3d(tags),
         tags = NULL)
}
\arguments{
  \item{con}{
A connection or filename.
}
  \item{ascii}{
Whether to use the ASCII format or the binary format.  The
default \code{NA} setting for \code{readSTL()} causes it
to detect the format.  This only works for files, not other
connections, which default to binary.
}
  \item{plot}{
On reading, should the object be plotted?
}
  \item{\dots}{
If plotting, other parameters to pass to \code{\link{triangles3d}}
}
  \item{pointRadius, lineRadius}{
The radius of points and lines relative to the overall scale of the figure.
}
  \item{pointShape}{
A mesh shape to use for points.  It is scaled by the \code{pointRadius}.
}
  \item{lineSides}{
Lines are rendered as cylinders with this many sides.
}
  \item{ids}{
The identifiers (from \code{\link{ids3d}}) of the
objects to write.  If \code{NULL}, try to write everything.
}
 \item{tags}{
Alternate way to specify \code{ids}.  Ignored if \code{ids} is
given. 
 }
}
\details{
The current implementation is limited.  For reading, it ignores normals
and color information.  For writing, it only outputs triangles, quads, planes, spheres,
points, line segments, line strips and surfaces, and does not write color information.
Lines and points are rendered in an isometric scale:  if your data scales vary, they
will look strange.

Since the STL format only allows one object per file, all RGL objects
are combined into a single object when output.

The output file is readable by Blender and Meshlab; the latter can write in
a number of other formats, including U3D, suitable for import into a PDF document.

}
\value{
\code{readSTL} invisibly returns the object id if \code{plot = TRUE}, or (visibly) 
a matrix of vertices of the triangles if not.

\code{writeSTL} invisibly returns the name of the connection to which the 
data was written.
}
\references{
The file format was found on Wikipedia on October 25, 2012.  I learned
about the STL file format from David Smith's blog reporting on 
Ian Walker's \code{r2stl} function.  
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{scene3d}} saves a copy of a scene to an R variable;
\code{\link{rglwidget}}, \code{\link{writeASY}}, \code{\link{writePLY}}, \code{\link{writeOBJ}} and \code{\link{writeSTL}}
write the scene to a file in various other formats.
}
\examples{
filename <- tempfile(fileext = ".stl")
open3d()
shade3d( icosahedron3d(col = "magenta") )
writeSTL(filename)
open3d()
readSTL(filename, col = "red")
}
\keyword{ graphics }

\name{spheres3d}
\alias{spheres3d}
\title{Add spheres}
\description{
  Adds a sphere set shape node to the scene
}
\usage{
spheres3d(x, y = NULL, z = NULL, radius = 1, fastTransparency = TRUE, ...)
}
\arguments{
  \item{x, y, z}{Numeric vector of point coordinates corresponding to
   the center of each sphere.  Any reasonable way of defining the
    coordinates is acceptable.  See the function \code{\link[grDevices]{xyz.coords}}
    for details.}
  \item{radius}{Vector or single value defining the sphere radius/radii}
  \item{fastTransparency}{logical value indicating whether
fast sorting should be used for transparency.  See the Details.}
  \item{ ... }{Material properties. See \code{\link{material3d}} for details.}
}
\details{
If a non-isometric aspect ratio is chosen, these functions will still draw
objects that appear to the viewer to be spheres.  Use \code{\link{ellipse3d}} 
to draw shapes that are spherical in the data scale.

When the scale is not isometric, the radius is measured in an average scale.
In this case the bounding box calculation is iterative, since rescaling the
plot changes the shape of the spheres in user-coordinates, which changes
the bounding box.  Versions of \pkg{rgl} prior to 0.92.802 did not do 
this iterative adjustment.

If any coordinate or radius is \code{NA}, the sphere is not plotted.

If a texture is used, its bitmap is wrapped around the 
sphere, with the top edge at the maximum y coordinate, 
and the left-right edges joined at the maximum in the
z coordinate, centred in x.

If the \code{alpha} material value of the spheres is less than 
the default \code{1}, they need to be drawn in order from
back to front.  When \code{fastTransparency} is \code{TRUE}, this is approximated by sorting the centers and drawing
complete spheres in that order.  This produces acceptable
results in most cases, but artifacts may be visible, 
especially if the \code{radius} values vary, or they 
intersect other transparent objects.  Setting 
\code{fastTransparency = FALSE} will cause the sorting 
to apply to each of the 480 facets of individual spheres.
This is much slower, but may produce better
output.
} 
\value{
  A shape ID of the spheres object is returned.
}
\examples{
open3d()
spheres3d(rnorm(10), rnorm(10), rnorm(10), 
          radius = runif(10), color = rainbow(10))
}
\seealso{
  \code{\link{material3d}}, \code{\link{aspect3d}} for setting non-isometric scales
}
\keyword{dynamic}

\name{drape3d}
\alias{drape3d}
\alias{drape3d.default}
\alias{drape3d.mesh3d}
\title{
Drape lines over a scene.
}
\description{
Project a line onto the surface in a scene so that it appears to drape itself
onto the surface.
}
\usage{
drape3d(obj, ...)
\method{drape3d}{mesh3d}(obj, x, y = NULL, z = NULL, plot = TRUE,
                         up = c(0, 0, 1),
                         P = projectDown(up), ...)
\method{drape3d}{default}(obj, ...)
}
\arguments{
  \item{obj}{
The object(s) upon which to drape lines.
}
  \item{x,y,z}{Coordinates of the line segments
to be draped.  Any reasonable way of defining the
coordinates is acceptable.  See the function \code{\link[grDevices]{xyz.coords}} for details.
}
  \item{plot}{
Should the result be plotted, or returned as a data frame?
}
  \item{up}{
The direction to consider as \dQuote{up}.
  }
  \item{P}{
The projection to use for draping, a 4x4 matrix.
  }
  \item{\dots}{
For the \code{"mesh3d"} method, additional parameters to pass to \code{\link{segments3d}}
when drawing the draped lines.
For the \code{"default"} method, additional parameters to pass to the
\code{"mesh3d"} method.
}
}
\details{
The default method converts \code{obj} to a mesh
using \code{\link{as.mesh3d}}, then uses the \code{"mesh3d"}
method.

The current implementation constructs the segments to drape
across the surface using the same method as
\code{\link{lines3d}} uses:  each successive point is
joined to the previous one.  Use \code{NA} coordinates to
indicate breaks in the line.

The \code{P} matrix is used to project points to a plane as
follows:  They are transformed by \code{P} in 
homogeneous coordinates, then only first two (Euclidean)
coordinates are kept.
}
\value{
If \code{plot = TRUE}, 
plots the result and invisibly returns the object ID of the collection of segments. 

If \code{plot = FALSE}, returns a matrix containing "x", "y" and "z"
values for the line(s) 
(for use with \code{\link{segments3d}}),
}
\author{
George Helffrich and Duncan Murdoch
}
\seealso{\code{\link{shadow3d}}, \code{\link{facing3d}}}
\examples{
#
# volcano example taken from "persp"
#

z <- 2 * volcano        # Exaggerate the relief

x <- 10 * (1:nrow(z))   # 10 meter spacing (S to N)
y <- 10 * (1:ncol(z))   # 10 meter spacing (E to W)

zlim <- range(z)
zlen <- zlim[2] - zlim[1] + 1

colorlut <- terrain.colors(zlen) # height color lookup table

col <- colorlut[ z - zlim[1] + 1 ] # assign colors to heights for each point

open3d()
id <- surface3d(x, y, z, color = col, polygon_offset = 1)

segs <- data.frame(x = range(x) + c(100, -100),
                   y = range(y) + c(150, -100), z = 325)
drape3d(id, segs, col = 'yellow', lwd = 3)
lines3d(segs, col='red', lwd=3)

p <- c(350, 205)         # (x,y) of strike & dip reading
off <- 20*c(-1, +1)      # X-marks-the-spot offset
segs <- data.frame(
    x = c(p[1] + off, NA, p[1] + off),
    y = c(p[2] + off, NA, p[2] - off),
    z = rep(350, 5)
    )
drape3d(id, segs, col = "yellow", lwd = 3)
}

\name{as.mesh3d.rglId}
\alias{as.mesh3d.rglId}
\title{
Convert object in plot to RGL mesh object
}
\description{
This method attempts to read the attributes of objects in the
rgl display and construct a mesh3d object to approximate them.
}
\usage{
\method{as.mesh3d}{rglId}(x, type = NA, subscene = NA, ...)
}
\arguments{
  \item{x}{
A vector of RGL identifiers of objects in the specified subscene.
}
  \item{type}{
A vector of names of types of shapes to convert.  Other shapes 
will be ignored.
  }
  \item{subscene}{
Which subscene to look in; the default \code{NA} specifies the 
current subscene.
  }
  \item{\dots}{
Ignored.
}
}
\details{
This function attempts to construct a triangle mesh to approximate 
one or more objects from the current display.  It can handle objects of types from \code{c("points", "lines", "linestrip", "triangles", "quads", "planes", "surface")}.

Since this method only produces meshes containing points,
segments and triangles, they won't necessarily
be an exact match to the original object.

If the generic \code{\link{as.mesh3d}} is called with no \code{x}
argument, this method will be called with \code{x} set to the 
ids in the current scene.  
}
\value{
A mesh object.
}
\author{
Duncan Murdoch
}

\seealso{
\code{\link{as.triangles3d.rglId}} for extracting the triangles,
\code{\link{clipMesh3d}} to apply complex clipping to a mesh object.
}
\examples{
# volcano example taken from "persp"
#
data(volcano)

z <- 2 * volcano        # Exaggerate the relief

x <- 10 * (1:nrow(z))   # 10 meter spacing (S to N)
y <- 10 * (1:ncol(z))   # 10 meter spacing (E to W)

zlim <- range(y)
zlen <- zlim[2] - zlim[1] + 1

colorlut <- terrain.colors(zlen) # height color lookup table

col <- colorlut[ z - zlim[1] + 1 ] # assign colors to heights for each point

open3d(useNULL = TRUE)
surface3d(x, y, z, color = col)
m <- as.mesh3d()
close3d()

open3d()
shade3d(m)
}

\name{rgl.attrib}
\alias{rgl.attrib}
\title{
Get information about shapes
}
\description{
Retrieves information about the shapes in a scene.
}
\usage{
rgl.attrib(id, attrib, first = 1, 
	last = rgl.attrib.count(id, attrib))
}
\arguments{
  \item{id}{
A shape identifier, as returned by \code{\link{ids3d}}.
}
  \item{attrib}{
An attribute of a shape.  Currently supported: one of \cr
\code{"vertices"}, \code{"normals"}, \code{"colors"}, \code{"texcoords"}, 
\code{"dim"}, \code{"texts"}, \code{"cex"}, \code{"adj"}, \code{"radii"}, 
\code{"centers"}, \code{"ids"}, \code{"usermatrix"}, \code{"types"},
\code{"flags"}, \code{"offsets"}, \code{"family"}, \code{"font"},
\code{"pos"}\cr
or unique prefixes
to one of those. 
}
  \item{first, last}{
Specify these to retrieve only those rows of the result.
}
}
\details{
If the identifier is not found or is not a shape that has the given attribute, 
zero will be returned by \code{rgl.attrib.count}, and an empty matrix
will be returned by \code{rgl.attrib}.

The first four \code{attrib} names correspond to the usual OpenGL
properties; \code{"dim"} is used just for surfaces, defining the rows
and columns in the rectangular grid; \code{"cex"}, \code{"adj"},
\code{"family"}, \code{"font"} and \code{"pos"} apply only to text objects.
}
\value{
\code{rgl.attrib}
returns the values of the attribute.  Attributes
are mostly real-valued, with the following sizes:
\tabular{lll}{
\code{"vertices"}	\tab 3 values 	\tab x, y, z \cr
\code{"normals"}	\tab 3 values 	\tab x, y, z \cr
\code{"centers"}	\tab 3 values 	\tab x, y, z \cr
\code{"colors"}	\tab 4 values 	\tab r, g, b, a \cr
\code{"texcoords"}	\tab 2 values 	\tab s, t \cr
\code{"dim"}	\tab 2 values 	\tab r, c \cr
\code{"cex"}	\tab 1 value 	\tab cex \cr
\code{"adj"}	\tab 2 values 	\tab x, y \cr
\code{"radii"}	\tab 1 value 	\tab r \cr
\code{"ids"}	\tab 1 value	\tab id \cr
\code{"usermatrix"}	\tab 4 values 	\tab x, y, z, w \cr
\code{"texts"}	\tab 1 value	\tab text \cr
\code{"types"}	\tab 1 value	\tab type \cr
\code{"flags"}	\tab 1 value	\tab flag \cr
\code{"family"} \tab 1 value    \tab family \cr
\code{"font"}   \tab 1 value    \tab font \cr
\code{"pos"}    \tab 1 value \tab pos \cr
}
The \code{"texts"}, \code{"types"} and \code{"family"}
attributes are character-valued; the \code{"flags"}
attribute is logical valued, with named rows.

These are returned as matrices with the row count equal to the count for the attribute,
and the columns as listed above.
}
\author{
Duncan Murdoch
}
\seealso{
\code{\link{ids3d}}, \code{\link{rgl.attrib.info}}
}
\examples{
p <- plot3d(rnorm(100), rnorm(100), rnorm(100), type = "s", col = "red")
rgl.attrib(p["data"], "vertices", last = 10)
}
\keyword{ graphics }

\name{setUserCallbacks}
\alias{setUserCallbacks}
\title{
Set mouse callbacks in R or Javascript code
}
\description{
This function sets user mouse callbacks in R or
\code{\link{rglwidget}} displays.
}
\usage{
setUserCallbacks(button, 
                 begin = NULL, 
                 update = NULL, 
                 end = NULL, 
                 rotate = NULL,
                 javascript = NULL, 
                 subscene = scene$rootSubscene$id,
                 scene = scene3d(minimal = FALSE),
                 applyToScene = TRUE,
			           applyToDev = missing(scene))
}
\arguments{
  \item{button}{
Which button should this callback apply to? Can
be numeric from \code{0:4}, or character from \code{"none", "left", "right", "center", "wheel"}.
}
  \item{begin, update, end, rotate}{
Functions to call when events occur.  See Details.
}
  \item{javascript}{
Optional block of Javascript code to be
included (at the global level).
}
  \item{subscene}{
Which subscene do these callbacks apply to?
}
  \item{scene}{
Which scene?
}
  \item{applyToScene}{
Should these changes apply to the scene object?
}
  \item{applyToDev}{
Should these changes apply to the current device?
  }
}
\details{
If \code{applyToScene} is \code{TRUE}, this function adds Javascript 
callbacks to the \code{scene} object.  
If \code{applyToDev} is \code{TRUE}, it adds R
callbacks to the current RGL device.

For Javascript, 
the callbacks are specified as strings; these will be
evaluated within the browser in the global context to define the functions, 
which will then be called with the Javascript
\code{this} object set to the current
\code{rglwidgetClass} object.

For R, they may be strings or R functions.

Both options may be \code{TRUE}, in which case the
callbacks must be specified as strings which are 
both valid Javascript and valid R.  The usual way to
do this is to give just a function name, with the
function defined elsewhere, as in the Example below.

The \code{begin} and \code{update} functions should be 
of the form
\code{function(x, y) { ... }}.  The \code{end} function
will be called with no arguments.  

The \code{rotate} callback can only be set on the 
mouse wheel.  It is called when the mouse 
wheel is rotated.  It should be of the form
\code{function(away)}, where \code{away} will be 1
while rotating the wheel \dQuote{away} from you,
and 2 while rotating it towards you.  If \code{rotate}
is not set but other callbacks are set on the wheel
\dQuote{button}, then each click of the mouse wheel
will trigger all \code{start}, \code{update}, 
then \code{end} calls in sequence.

The \code{javascript} argument is an optional block 
of code which will be evaluated once during the 
initialization of the widget.  It can define functions
and assign them as members of the \code{window} object,
and then the names of those functions can be given 
in the callback arguments; this allows the callbacks
to share information.
}
\value{
Invisibly returns an \code{rglScene} object.  This
object will record the changes if \code{applyToScene}
is \code{TRUE}.

If \code{applyToDev} is \code{TRUE}, it will also 
have the side effect of attempting to install the
callbacks using \code{\link{rgl.setMouseCallbacks}}
and \code{\link{rgl.setWheelCallback}}.
}
\seealso{
\code{\link{setAxisCallbacks}} for user defined axes.
}
\author{
Duncan Murdoch
}
\examples{
  # This example identifies points in both the rgl window and
  # in WebGL
  
  verts <- cbind(rnorm(11), rnorm(11), rnorm(11))
  idverts <- plot3d(verts, type = "s", col = "blue")["data"]
  
  # Plot some invisible text; the Javascript will move it
  idtext <- text3d(verts[1,,drop = FALSE], texts = 1, adj = c(0.5, -1.5), alpha = 0)
  
  # Define the R functions to use within R
  fns <- local({
    idverts <- idverts
    idtext <- idtext
    closest <- -1
    update <- function(x, y) {
      save <- par3d(skipRedraw = TRUE)
      on.exit(par3d(save))
      rect <- par3d("windowRect")
      size <- rect[3:4] - rect[1:2]
      x <- x / size[1];
      y <- 1 - y / size[2];
      verts <- rgl.attrib(idverts, "vertices")
      # Put in window coordinates
      vw <- rgl.user2window(verts)
      dists <- sqrt((x - vw[,1])^2 + (y - vw[,2])^2)
      newclosest <- which.min(dists)
      if (newclosest != closest) {
        if (idtext > 0)
          pop3d(id = idtext)
        closest <<- newclosest
        idtext <<- text3d(verts[closest,,drop = FALSE], texts = closest, adj = c(0.5, -1.5))
      }
    }
    end <- function() {
      if (idtext > 0)
        pop3d(id = idtext)
      closest <<- -1
      idtext <<- -1
    }
    list(rglupdate = update, rglend = end)
  })
  rglupdate <- fns$rglupdate
  rglend <- fns$rglend
  
  # Define the Javascript functions with the same names to use in WebGL
  js <-
   ' var idverts = \%id\%, idtext = \%idtext\%, closest = -1,
         subid = \%subid\%;
   
     window.rglupdate = function(x, y) { 
       var   obj = this.getObj(idverts), i, newdist, dist = Infinity, pt, newclosest;
       x = x/this.canvas.width;
       y = y/this.canvas.height;
       
       for (i = 0; i < obj.vertices.length; i++) {
         pt = obj.vertices[i].concat(1);
         pt = this.user2window(pt, subid);
         pt[0] = x - pt[0];
         pt[1] = y - pt[1];
         pt[2] = 0;
         newdist = rglwidgetClass.vlen(pt);
         if (newdist < dist) {
           dist = newdist;
           newclosest = i;
         }
       }

       if (newclosest !== closest) {
         closest = newclosest
         var text = this.getObj(idtext);
         text.vertices[0] = obj.vertices[closest];
         text.colors[0][3] = 1; // alpha is here!
         text.texts[0] = (closest + 1).toString();
         text.initialized = false;
         this.drawScene();
       }
     };
     window.rglend = function() {
       var text = this.getObj(idtext);
       closest = -1;
       text.colors[0][3] = 0;
       text.initialized = false;
       this.drawScene();
     }'
  js <- sub("\%id\%", idverts, js)  
  js <- sub("\%subid\%", subsceneInfo()$id, js)
  js <- sub("\%idtext\%", idtext, js)
    
  # Install both
  setUserCallbacks("left",
                    begin = "rglupdate",
                    update = "rglupdate",
                    end = "rglend",
                    javascript = js)
  rglwidget()
  
  # This example doesn't affect the rgl window, it only modifies
  # the scene object to implement panning
  
  # Define the Javascript functions to use in WebGL
  js <-
  '  window.subid = \%subid\%;
   
     window.panbegin = function(x, y) {
       var activeSub = this.getObj(subid),
           viewport = activeSub.par3d.viewport,
           activeModel = this.getObj(this.useid(activeSub.id, "model")),
           l = activeModel.par3d.listeners, i;

        this.userSave = {x:x, y:y, viewport:viewport,
                            cursor:this.canvas.style.cursor};
        for (i = 0; i < l.length; i++) {
          activeSub = this.getObj(l[i]);
          activeSub.userSaveMat = new CanvasMatrix4(activeSub.par3d.userMatrix);
        }
        this.canvas.style.cursor = "grabbing";
     };
     
     window.panupdate = function(x, y) { 
        var objects = this.scene.objects,
            activeSub = this.getObj(subid),
            activeModel = this.getObj(this.useid(activeSub.id, "model")),
            l = activeModel.par3d.listeners,
            viewport = this.userSave.viewport,
            par3d, i, zoom;
        if (x === this.userSave.x && y === this.userSave.y)
          return;
        x = (x - this.userSave.x)/this.canvas.width;
        y = (y - this.userSave.y)/this.canvas.height;
        for (i = 0; i < l.length; i++) {
          activeSub = this.getObj(l[i]);
          par3d = activeSub.par3d;
          /* NB:  The right amount of zoom depends on the scaling of the data
                  and the position of the observer.  This might
                  need tweaking.
          */
          zoom = rglwidgetClass.vlen(par3d.observer)*par3d.zoom;
          activeSub.par3d.userMatrix.load(objects[l[i]].userSaveMat);
          activeSub.par3d.userMatrix.translate(zoom*x, zoom*y, 0);
        }
        this.drawScene();
     };
     
     window.panend = function() {
       this.canvas.style.cursor = this.userSave.cursor;
     };
'

js <- sub("\%subid\%", subsceneInfo()$id, js)

scene <- setUserCallbacks("left", 
                 begin = "panbegin", 
                 update = "panupdate", 
                 end = "panend", 
                 applyToDev = FALSE, javascript = js)
rglwidget(scene)
}

\name{tkspin3d}
\alias{tkspin3d}
\title{Create TCL/TK controller for RGL window}
\description{
  This function creates a TCL/TK window containing buttons to spin and 
  resize one or more RGL windows.
}
\usage{
tkspin3d(dev = cur3d(), ...)
}
\arguments{
  \item{dev}{A vector of one or more RGL device numbers to control }
  \item{...}{Named parameters in that match named formal arguments
to \code{\link{tkspinControl}} are passed there, while others are passed to \code{\link[tcltk:TkWidgets]{tktoplevel}}}}
\author{ Ming Chen and Duncan Murdoch }
\seealso{ \code{\link{tkspinControl}}}
\examples{
if (interactive() && !in_pkgdown_example()) {
    open3d()
    points3d(rnorm(100), rnorm(100), rnorm(100), size=3)
    axes3d()
    box3d()
    tkspin3d()
}
}

\name{persp3d.function}
\alias{plot3d.function}
\alias{persp3d.function}
\title{
Plot a function of two variables
}
\description{
Plot a function \code{z(x, y)} or a parametric function
\code{(x(s, t), y(s, t), z(s, t))}.
}
\usage{
\method{persp3d}{`function`}(x, 
  xlim = c(0, 1), ylim = c(0, 1), 
  slim = NULL, tlim = NULL, 
  n = 101, 
  xvals = seq.int(min(xlim), max(xlim), length.out = n[1]), 
  yvals = seq.int(min(ylim), max(ylim), length.out = n[2]), 
  svals = seq.int(min(slim), max(slim), length.out = n[1]), 
  tvals = seq.int(min(tlim), max(tlim), length.out = n[2]), 
  xlab, ylab, zlab, 
  col = "gray", otherargs = list(), 
  normal = NULL, texcoords = NULL, \dots)
\S3method{plot3d}{`function`}(x, \ldots)
}
\arguments{
  \item{x}{
A function of two arguments.  See the details below.
}
  \item{xlim, ylim}{
By default, the range of x and y values.  For a parametric
surface, if these are not missing, they are used as limits
on the displayed x and y values.
}
  \item{slim, tlim}{
If not \code{NULL}, these give the range of s and t in the
parametric specification of the surface.  If
only one is given, the other defaults to \code{c(0, 1)}.
}
  \item{n}{
A one or two element vector giving the number of 
steps in the x and y (or s and t) grid.
}
  \item{xvals, yvals}{
The values at which to evaluate x and y.  Ignored for
a parametric surface.  If used, \code{xlim} and/or \code{ylim}
are ignored.
}
  \item{svals, tvals}{
The values at which to evaluate s and t for a parametric
surface.  Only used if
\code{slim} or \code{tlim} is not \code{NULL}.  As with 
\code{xvals} and \code{yvals}, these override the corresponding
\code{slim} or \code{tlim} specification.
}
  \item{xlab, ylab, zlab}{
The axis labels.  See the details below for the defaults.
}
  \item{col}{
The color to use for the plot.  See the details below.
}
  \item{otherargs}{
Additional arguments to pass to the function.
}
  \item{normal, texcoords}{
Functions to set surface normals or texture coordinates.  See
the details below.
}
  \item{\dots}{
Additional arguments to pass to \code{\link{persp3d}}.
}
}
\details{
The \code{"function"} method for \code{plot3d} simply passes
all arguments to \code{persp3d}.  Thus this description applies
to both.

The first argument \code{x} is required to be a function.  It
is named \code{x} only because of the requirements of the S3
system; in the remainder of this help page, we will assume
that the assignment \code{f <- x} has been made, and will 
refer to the function \code{f()}.

\code{persp3d.function} evaluates \code{f()} on a two-dimensional
grid of values, and displays the resulting surface.  The values
on the grid will be passed in as vectors in the 
first two arguments to the function, so \code{f()} needs
to be vectorized.  Other optional arguments to \code{f()}
can be specified in the \code{otherargs} list.

In the default form where \code{slim} and \code{tlim} are both
\code{NULL}, it is assumed that \code{f(x, y)} returns heights,
which will be plotted in the z coordinate.  The default axis labels
will be taken from the argument names to \code{f()} and the
expression passed as argument \code{x} to this function.

If \code{slim} or \code{tlim} is specified, a parametric
surface is plotted.  The function
\code{f(s, t)} must return a 3-column matrix, giving x, y and z 
coordinates of points on the surface.  The default axis
labels will be the column names if those are present.
In this case \code{xlim}, \code{ylim} and \code{zlim}
are used to define a clipping region only if specified;
the defaults are ignored.

The color of the surface may be specified as the name of a
color, or a vector or matrix of color names. In this case
the colors will be recycled across the points on the grid
of values.

Alternatively, a function may be given:  it should be a function
like \code{\link{rainbow}} that takes an integer argument and
returns a vector of colors.  In this case the colors are mapped
to z values.

The \code{normal} argument allows specification of a function
to compute normal vectors to the surface.  This function is
passed the same arguments as \code{f()} (including \code{otherargs}
if present), and should produce a 3-column matrix containing the
x, y and z coordinates of the normals.

The \code{texcoords} argument is a function similar to \code{normal}, but
it produces a 2-column matrix containing texture coordinates.

Both \code{normal} and \code{texcoords} may also contain matrices,
with 3 and 2 columns respectively, and rows corresponding to the
points that were passed to \code{f()}.
}
\value{
This function constructs a call to \code{\link{persp3d}}
and returns the value from that function.
}
\author{
Duncan Murdoch
}

\seealso{
The \code{\link{curve}} function in base graphics does
something similar for functions of one variable.  See the
example below for space curves.
}
\examples{
# (1) The Obligatory Mathematical surface.
#     Rotated sinc function, with colors

f <- function(x, y) { 
  r <- sqrt(x^2 + y^2)
  ifelse(r == 0, 10, 10 * sin(r)/r)
}
open3d()
plot3d(f, col = colorRampPalette(c("blue", "white", "red")), 
       xlab = "X", ylab = "Y", zlab = "Sinc( r )", 
       xlim = c(-10, 10), ylim = c(-10, 10),
       aspect = c(1, 1, 0.5))
       
# (2) A cylindrical plot

f <- function(s, t) {
  r <- 1 + exp( -pmin( (s - t)^2, 
                       (s - t - 1)^2, 
                       (s - t + 1)^2 )/0.01 )
  cbind(r*cos(t*2*pi), r*sin(t*2*pi), s)
}

open3d()
plot3d(f, slim = c(0, 1), tlim = c(0, 1), col = "red", alpha = 0.8)

# Add a curve to the plot, fixing s at 0.5.

plot3d(f(0.5, seq.int(0, 1, length.out = 100)), type = "l", add = TRUE, 
       lwd = 3, depth_test = "lequal")
}
\keyword{ graphics }

\name{rgl-defunct}
\alias{rgl-defunct}
\alias{subsetSlider}
\alias{subsetSetter}
\alias{clipplaneSlider}
\alias{toggleButton}
\alias{propertySlider}
\alias{propertySetter}
\alias{par3dinterpSetter}
\alias{matrixSetter}
\alias{vertexSetter}
\alias{elementId2Prefix}
\alias{rgl.texts}
\alias{rgl.sprites}
\alias{rgl.abclines}
\alias{rgl.bbox}
\alias{rgl.bg}
\alias{rgl.clear}
\alias{rgl.clipplanes}
\alias{rgl.close}
\alias{rgl.light}
\alias{rgl.lines}
\alias{rgl.linestrips}
\alias{rgl.material}
\alias{rgl.open}
\alias{rgl.planes}
\alias{rgl.points}
\alias{rgl.primitive}
\alias{rgl.quads}
\alias{rgl.select3d}
\alias{rgl.set}
\alias{rgl.setAxisCallback}
\alias{rgl.surface}
\alias{rgl.triangles}
\alias{rgl.viewpoint}
\alias{writeWebGL}
\title{
Defunct functions in \pkg{rgl}
}
\description{
These functions have been removed from \pkg{rgl}.
Comments indicate the replacements.
}
\usage{
# Removed in version 1.4.0:
elementId2Prefix(elementId, prefix = elementId)
rgl.abclines(x, y=NULL, z=NULL, a, b=NULL, c=NULL, ...)
rgl.bbox(xat = NULL, xlab = NULL, xunit = 0, xlen = 5, 
         yat = NULL, ylab = NULL, yunit = 0, ylen = 5, 
         zat = NULL, zlab = NULL, zunit = 0, zlen = 5, 
         marklen = 15, marklen.rel = TRUE, 
         expand = 1, draw_front = FALSE, ...)  # bbox3d
         
rgl.bg( sphere = FALSE, fogtype = "none", color = c("black", "white"), 
	back = "lines", fogScale = 1, ...)           # bg3d
	
rgl.clear( type = "shapes", subscene = 0 )     # clear3d
rgl.clipplanes( a, b=NULL, c=NULL, d=0)        # clipplanes3d
rgl.close()                                    # close3d
rgl.light( theta = 0, phi = 0, viewpoint.rel = TRUE, 
           ambient = "#FFFFFF", diffuse = "#FFFFFF", specular = "#FFFFFF", 
           x = NULL, y = NULL, z = NULL)       # light3d
rgl.lines(x, y = NULL, z = NULL, ... )         # segments3d
rgl.linestrips(x, y = NULL, z = NULL, ...)     # lines3d
rgl.material(...)                              # material3d
rgl.open(useNULL = rgl.useNULL())              # open3d
rgl.planes( a, b=NULL, c=NULL, d=0,...)        # planes3d
rgl.points(x, y = NULL, z = NULL, ... )        # points3d
rgl.primitive(...)                             
rgl.quads(x, y = NULL, z = NULL, 
          normals = NULL, 
          texcoords = NULL, ... )              # quads3d
rgl.select3d(button = c("left", "middle", "right"),
            dev = cur3d(), 
            subscene = currentSubscene3d(dev)) # select3d
rgl.set(which, silent = FALSE)                 # set3d
rgl.setAxisCallback(axis, draw = NULL, 
            dev = cur3d(), 
            subscene = currentSubscene3d(dev)) # setAxisCallback

rgl.sprites(x, y = NULL, z = NULL, radius = 1, 
            shapes = NULL, userMatrix, fixedSize = FALSE,
            adj = 0.5, pos = NULL, offset = 0.25, 
            rotating = FALSE, ...)
            
rgl.surface(x, z, y, coords = 1:3, ..., 
            normal_x = NULL, normal_y = NULL, normal_z = NULL,
            texture_s = NULL, texture_t = NULL) # surface3d
            
rgl.texts(x, y = NULL, z = NULL, text, 
          adj = 0.5, pos = NULL, offset = 0.5, 
          family = par3d("family"), 
          font = par3d("font"), 
          cex = par3d("cex"), 
          useFreeType = par3d("useFreeType"), 
          ...)                                  # text3d               

rgl.triangles(x, y = NULL, z = NULL, 
              normals = NULL, texcoords = NULL, 
              ... )                             # triangles3d
rgl.viewpoint( theta = 0, phi = 15, 
               fov = 60, zoom = 1, 
               scale = par3d("scale"), 
               interactive = TRUE, 
               userMatrix, 
               type = c("userviewpoint",
                        "modelviewpoint") )     # view3d
writeWebGL(dir = "webGL", 
           filename = file.path(dir, "index.html"), 
           template = system.file(file.path("WebGL", "template.html"), package = "rgl"),
           prefix = "",
           snapshot = TRUE, 
           commonParts = TRUE, 
           reuse = NULL, 
           font = "Arial", 
           width, height)                       # rglwidget


          
# Removed in version 1.0.0:

subsetSlider(subsets, labels = names(subsets), 
             fullset = Reduce(union, subsets),
             subscenes = currentSubscene3d(), prefixes = "",
             accumulate = FALSE, ...)
             
subsetSetter(subsets, subscenes = currentSubscene3d(), prefixes = "", 
			 fullset = Reduce(union, subsets),
                         accumulate = FALSE)
             
clipplaneSlider(a=NULL, b=NULL, c=NULL, d=NULL, 
                plane = 1, clipplaneids, prefixes = "", 
                labels = signif(values[,1],3), ...)
             
toggleButton(subset, subscenes = currentSubscene3d(), prefixes = "", 
             label = deparse(substitute(subset)), 
             id = paste0(basename(tempfile("input"))), name = id) 
propertySlider(setter = propertySetter,
             minS = NULL, maxS = NULL, step = 1, init = NULL, 
	     labels, 
	     id = basename(tempfile("input")), name = id,
	     outputid = paste0(id, "text"),
	     index = NULL,
	     ...) 
	     
propertySetter(values = NULL, entries, properties, objids, prefixes = "",
               param = seq_len(NROW(values)), interp = TRUE, digits = 7) 
               
par3dinterpSetter(fn, from, to, steps, subscene, omitConstant = TRUE, 
                  rename = character(), ...)

matrixSetter(fns, from, to, steps, subscene = currentSubscene3d(), 
             matrix = "userMatrix", omitConstant = TRUE, prefix = "", ...) 
             
vertexSetter(values, vertices = 1, attributes, objid, prefix = "",
	     param = seq_len(NROW(values)), interp = TRUE,
	     digits = 7)             
}
\seealso{
  \code{\link{Defunct}}
}
\keyword{internal}
\keyword{misc}

\name{material3d}
\alias{material3d}
\alias{rgl.material.names}
\alias{rgl.material.readonly}
\concept{color}
\concept{lit}
\concept{ambient}
\concept{specular}
\concept{emission}
\concept{shininess}
\concept{alpha}
\concept{smooth}
\concept{texture}
\concept{textype}
\concept{texmode}
\concept{texmipmap}
\concept{texmagfilter}
\concept{texminfilter}
\concept{texenvmap}
\concept{front}
\concept{back}
\concept{size}
\concept{lwd}
\concept{fog}
\concept{point_antialias}
\concept{line_antialias}
\concept{depth_mask}
\concept{depth_test}
\concept{polygon_offset}
\concept{margin}
\concept{floating}
\concept{tag}
\concept{blend}
\concept{isTransparent}
\concept{rgl.warnBlackTexture}

\title{Get or set material properties}
\description{
  Get or set material properties for geometry appearance.
}
\usage{
material3d(..., id = NULL)

rgl.material.names
rgl.material.readonly
}
\arguments{
  \item{...}{Material properties to set or query.}
  \item{id}{
  the \pkg{rgl} id of an object to query, or \code{NULL}
  to query or set the defaults.
  }
}
\details{
In an \pkg{rgl} scene, each object has \dQuote{material properties} that control how it is rendered and (in the case of \code{tag}) 
that can be used to store a 
label or other information.  \code{material3d} sets defaults
for these properties and queries the defaults or specific
values for an individual object.

To set values, use \code{name = value} settings, e.g.
\code{material3d(color = "red")}.  To query values, 
specify the property or properties in a character vector,
e.g. \code{material3d("color")}.  

  Only one side at a time can be culled.

  The \code{material} member of the \code{\link{r3dDefaults}} list may be used to 
set default values for material properties.
  
}

\section{Material Properties}{

The \code{rgl.material.names} variable contains the
full list of material names.
The following read-write material properties control
the appearance
of objects in an \pkg{rgl} scene.

\describe{

  \item{color}{
    vector of R color characters. Represents the diffuse component in case of lighting calculation (lit = TRUE),
    otherwise it describes the solid color characteristics.
  }
  \item{lit}{
    logical, specifying if lighting calculation should take place on geometry
  }
  \item{ambient, specular, emission, shininess}{
    properties for lighting calculation. ambient, specular, emission are R color character string values; shininess represents a
    numerical.
  }
  \item{alpha}{ 
    vector of alpha values between 0.0 (fully transparent) .. 1.0 (opaque).
  }
  \item{smooth}{
    logical, specifying whether smooth shading or flat shading
    should be used.  For smooth shading, Gouraud shading is
    used in \pkg{rgl} windows, while Phong shading is used 
    in WebGL.
  }
  \item{texture}{ 
    path to a texture image file. See the Textures section
    below for details.
  }
  \item{textype}{
    specifies what is defined with the pixmap
    \describe{
      \item{"alpha"}{alpha values}
      \item{"luminance"}{luminance}
      \item{"luminance.alpha"}{luminance and alpha}
      \item{"rgb"}{color}
      \item{"rgba"}{color and alpha texture}
    }
    Note that support for these modes is slightly different
in the display within R versus the WebGL display using
\code{rglwidget()}.  In particular, in WebGL 
\code{textype = "alpha"} will always take the alpha value from
the luminance (i.e. the average of the R, G and B channels) 
of the texture, whereas the R display bases
the choice on the internal format of the texture file.
  }
  \item{texmode}{
    specifies how the texture interacts with the existing
    color
    \describe{
      \item{"replace"}{texture value replaces existing value}
      \item{"modulate"}{default; texture value multiplies existing value}
      \item{"decal"}{for \code{textype = "rgba"}, texture
      is mixed with existing value}
      \item{"blend"}{uses the texture to blend the existing
      value with black}
      \item{"add"}{adds the texture value to the existing.
      May not be available in the R display with very old
      OpenGL drivers.}
    }
  }
  \item{texmipmap}{
    Logical, specifies if the texture should be mipmapped.
  }
  \item{texmagfilter}{
    specifies the magnification filtering type (sorted by ascending quality):
    \describe{
      \item{"nearest"}{texel nearest to the center of the pixel}
      \item{"linear"}{weighted linear average of a 2x2 array of texels}
    }
  }
  \item{texminfilter}{
    specifies the minification filtering type (sorted by ascending quality):
    \describe{
      \item{"nearest"}{texel nearest to the center of the pixel}
      \item{"linear"}{weighted linear average of a 2x2 array of texels}
      \item{"nearest.mipmap.nearest"}{low quality mipmapping}
      \item{"nearest.mipmap.linear"}{medium quality mipmapping}
      \item{"linear.mipmap.nearest"}{medium quality mipmapping}
      \item{"linear.mipmap.linear"}{high quality mipmapping}
    }
  }
  \item{texenvmap}{
    logical, specifies if auto-generated texture coordinates for environment-mapping 
    should be performed on geometry.
  }
  \item{front, back}{
    Determines the polygon mode for the specified side:
    \describe{
      \item{"filled"}{filled polygon}
      \item{"lines"}{wireframed polygon}
      \item{"points"}{point polygon}
      \item{"culled"}{culled (hidden) polygon}
    }
  }
  \item{size}{ 
    numeric, specifying the size of points in pixels
  }
  \item{lwd}{
    numeric, specifying the line width in pixels
  }
  \item{fog}{logical, specifying if fog effect should be applied on the corresponding shape.  Fog type is set in \code{\link{bg3d}}.}
  \item{point_antialias, line_antialias}{logical, specifying if points should be round and lines 
     should be antialiased, but see Note below.}
  \item{depth_mask}{logical, specifying whether the object's depth
     should be stored.}
  \item{depth_test}{Determines which depth test is used to see if this
     object is visible, depending on its apparent depth in the scene
     compared to the stored depth. Possible values are \code{"never"},
     \code{"less"} (the default), \code{"equal"}, \code{"lequal"}
     (less than or equal), \code{"greater"}, \code{"notequal"},
     \code{"gequal"} (greater than or equal), \code{"always"}.}
  \item{polygon_offset}{A one or two element
vector giving the \samp{factor} and \samp{units} values
to use in a \code{glPolygonOffset()} call in OpenGL.  If
only one value is given, it is used for both elements.
The \samp{units} value is added to the depth of all pixels in
a filled polygon,
and the \samp{factor} value is multiplied by an estimate of
the slope of the polygon and then added to the depth.  Positive values \dQuote{push} polygons back slightly for the purpose
of depth testing, to allow points, lines or other polygons
to be drawn on the surface without being obscured due 
to rounding error.  Negative values pull the object forward.
A typical value to use is \code{1} (which
is automatically expanded to \code{c(1,1)}).
If values are too large, objects which should be behind
the polygon will show through, and if values are too small,
the objects on the surface will be partially obscured. 
Experimentation may be needed to get it right.  The first
example in \code{?\link{persp3d}} uses this property to add 
grid lines to a surface.}
  \item{margin, floating}{Used mainly for text to
  draw annotations in the margins, but supported by
  most kinds of objects:  see \code{\link{mtext3d}}.}
  \item{tag}{A length 1 string value.  These
  may be used to identify objects, or encode other meta
  data about the object.}
  \item{blend}{Two string values from the list below
  describing how transparent objects are blended with
  colors behind them.
  The first determines the coefficient applied to the 
  color of the current object (the source); the second determines
  the coefficient applied to the existing color 
  (the destination). The resulting
  color will be the sum of the two resulting colors.
  The allowed strings
  correspond to OpenGL constants:
  \describe{
  \item{"zero"}{Zero; color has no effect.}
  \item{"one"}{One; color is added to the other term.}
  \item{"src_color", "one_minus_src_color"}{Multiply by source color or its opposite.}
  \item{"dst_color", "one_minus_dst_color"}{Multiply by destination color or its opposite.}
  \item{"src_alpha", "one_minus_src_alpha"}{Multiply by
source alpha or its opposite.  Default values.}
  \item{"dst_alpha", "one_minus_dst_alpha"}{Multiply by
destination alpha or its opposite.}
  \item{"constant_color", "one_minus_constant_color",
  "constant_alpha", "one_minus_constant_alpha",
  "src_alpha_saturate"}{These are allowed, but to be useful
  they require other settings which \pkg{rgl} doesn't
  support.}
  }
  }
  \item{col}{An allowed abbreviation of \code{color}.}
}

The \code{rgl.material.readonly} variable contains
the subset of material properties that are read-only
so they can be queried
but not set.
Currently there is only one:

\describe{
\item{isTransparent}{Is the current color transparent?}
}

}
\value{
\code{material3d()} returns values similarly
to \code{\link{par3d}}:
When setting properties, it returns the previous values
invisibly in a named list.  When querying multiple values, a
named list is returned. When a single value is queried it is
returned directly.
} 
\section{Textures}{
The \code{texture} material property may be \code{NULL} or the name of
a bitmap file to be displayed on the surface being rendered.
Currently only PNG format files are supported.

By default, the colors in the bitmap will modify the color of the
object being plotted.  If the color is black (a common default), you
won't see anything, so a warning may be issued.  You can
suppress the warning by specifying 
the color explicitly, or calling
\code{options{rgl.warnBlackTexture = FALSE}}.

Other aspects of texture display are controlled by the material
properties \code{textype, texmode, texmipmap, texmagfilter,
texminfilter} and \code{texenvmap} described above.

For an extensive discussion of textures, see the \href{../articles/rgl.html#textures}{Textures} section of the 
\href{../articles/rgl.html}{rgl Overview}
vignette.
}
\section{Display of objects}{
  
  Object display colors are determined as follows:
\itemize{
\item{If \code{lit = FALSE}, an element of the \code{color} vector property is displayed without modification.  See documentation for individual objects for information on which element is chosen.}
\item{If \code{lit = TRUE}, the color is determined as follows.}
\enumerate{
\item{The color is set to the \code{emission} property
of the object. }
\item{For each defined light, the following are added:
\itemize{
\item{the product of the \code{ambient} color
of the light and the \code{ambient} color of the object is added.}
\item{the \code{color} of the
object is multiplied by the \code{diffuse} color
of the light and by a constant depending on
the angle between the surface and the direction to 
the light, and added.}
\item{the \code{specular} color of the object
is multiplied by the \code{specular} color of the 
light and a constant depending on the \code{shininess}
of the object and the direction to the light, and
added.  The \code{shininess} property mainly
determines the size of the shiny highlight; adjust
one or both of the \code{specular} colors to change
its brightness.}
}

}
}
}
  If \code{point_antialias} is \code{TRUE}, points will be drawn as circles in WebGL; otherwise, they
will be drawn as squares.  Within R, the behaviour depends
on your graphics hardware:  for example, I see circles for 
both settings on my laptop.

Within R, lines tend to appear heavier with \code{line_antialias == TRUE}.  There's no difference at all
in WebGL.
}
\seealso{
\code{\link{bbox3d}},
\code{\link{bg3d}},
\code{\link{light3d}}
}
\examples{
save <- material3d("color")
material3d(color = "red")
material3d("color")
material3d(color = save)

# this illustrates the effect of depth_test
x <- c(1:3); xmid <- mean(x)
y <- c(2, 1, 3); ymid <- mean(y)
z <- 1
open3d()
tests <- c("never", "less", "equal", "lequal", "greater", 
                  "notequal", "gequal", "always")
for (i in 1:8) {
  triangles3d(x, y, z + i, col = heat.colors(8)[i])
  texts3d(xmid, ymid, z + i, paste(i, tests[i], sep = ". "), depth_test = tests[i]) 
}
highlevel()  # To trigger display

# this illustrates additive blending
open3d()
bg3d("darkgray")
quad <- cbind(c(-1, 1, 1, -1), 1, c(-1, -1, 1, 1))
quads3d(rbind(translate3d(quad, -0.5, 0, -0.5),
              translate3d(quad, 0.5,  0.5, -0.5),
              translate3d(quad, 0, 1, 0.5)), 
        col = rep(c("red", "green", "blue"), each = 4),
        alpha = 0.5, 
        blend = c("src_alpha", "one"))
}
\keyword{dynamic}
